/*!**********************************************************************

  module:	Reg_Win.c

  -----------------------------------------------------------------------

  responsible:  Markus zgen

  special area: Registry
  description : Handling of registry entries


  -----------------------------------------------------------------------

                         Copyright by SAP AG, 2000



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

************************************************************************/

/*
 * first check if WIN32 platform is defined
 */
#if !defined WIN32
#error WIN32 not defined !
#endif


/*===========================================================================*/
/*===
/*===	Includes
/*===
/*===========================================================================*/

#include <WINDOWS.H>	/* GetPrivateProfileString */
#include <shlwapi.h>	/* Extended registry API */

#include "SAPDB/Registry/Reg_Win.h"

/*===========================================================================*/
/*===
/*===	Defines
/*===
/*===========================================================================*/

/*===========================================================================*/
/*===
/*===	Types & Structs
/*===
/*===========================================================================*/

/*===========================================================================*/
/*===
/*===	Prototypes
/*===
/*===========================================================================*/

SAPDB_Bool ReadRegistry ( Reg_RegistryP	 registry,
						  const char	*sectionPath	);

SAPDB_Bool WriteSection ( Reg_SectionP	section	);

SAPDB_Bool WriteKey ( HKEY	keyHandle,
					  Reg_KeyP	key			);

SAPDB_Bool DropRegistry( const char	*path	);

/*===========================================================================*/
/*===
/*===	Public Methods
/*===
/*===========================================================================*/

/*!***************************************************************************

	Reg_WinOpenRegistry

******************************************************************************/

SAPDB_Bool Reg_WinOpenRegistry ( Reg_RegistryP	*registry,
							  const char	*path		)
{
	SAPDB_Bool		ok = SAPDB_FALSE;
	Reg_RegistryP	newRegistry = NULL;

	if ( !registry )
		return SAPDB_FALSE;

	sqlallocat( sizeof(Reg_Registry), registry, &ok );

	if ( !ok )
		return SAPDB_FALSE;

	newRegistry = *registry;

	strncpy( newRegistry->path, path, REG_MAX_PATH_LEN );
	newRegistry->path[REG_MAX_PATH_LEN] = '\0';
	newRegistry->sections = NULL;
	newRegistry->enumKeys = NULL;
	newRegistry->enumSections = NULL;

	if ( !ReadRegistry( newRegistry, newRegistry->path )) {
		Reg_WinCloseRegistry( newRegistry );
		*registry = NULL;

		return SAPDB_FALSE;
	}

	return SAPDB_TRUE;

}

/*!***************************************************************************

	Reg_WinCloseRegistry

******************************************************************************/

SAPDB_Bool Reg_WinCloseRegistry ( Reg_RegistryP	registry	)
{

	Reg_SectionNodeP	currentSectionNode = NULL;
	Reg_SectionNodeP	nextSectionNode = NULL;

	if ( !registry )
		return SAPDB_TRUE;

	currentSectionNode = registry->sections;
	while ( currentSectionNode ) {
		nextSectionNode = currentSectionNode->nextSectionNode;

		Reg_CommonDropSection( currentSectionNode->section );
		currentSectionNode->section = NULL;
		sqlfree( currentSectionNode );

		currentSectionNode = nextSectionNode;
	}

	sqlfree( registry );

	return SAPDB_TRUE;

}

/*!***************************************************************************

	Reg_WinReadRegistry

******************************************************************************/

SAPDB_Bool Reg_WinReadRegistry ( Reg_RegistryP	registry	)
{

	return ReadRegistry( registry, registry->path );

}

/*!***************************************************************************

	Reg_WinWrite

******************************************************************************/

SAPDB_Bool Reg_WinWriteRegistry ( Reg_RegistryP	registry	)
{

	Reg_SectionNodeP	currentSectionNode = NULL;

	/* Make some checks */
	if ( !registry )
		return SAPDB_FALSE;

	/* Drop registry */
	if ( !DropRegistry( registry->path ))
		return SAPDB_FALSE;

	/* Write all sections */
	currentSectionNode = registry->sections;
	while ( currentSectionNode ) {
		/* Write current section */
		if ( !WriteSection( currentSectionNode->section ))
			return SAPDB_FALSE;

		/* Get next section */
		currentSectionNode = currentSectionNode->nextSectionNode;
	}

	return SAPDB_TRUE;

}

/*===========================================================================*/
/*===
/*===	Private Methods
/*===
/*===========================================================================*/

/*!***************************************************************************

	ReadRegistry

******************************************************************************/

SAPDB_Bool ReadRegistry ( Reg_RegistryP	 registry,
						  const char	*section	)
{

	HKEY		sectionHandle = 0;
	HKEY		keyHandle = 0;
	DWORD		sectionIndex = 0;
	DWORD		keyIndex = 0;
	LONG		regRC;
	DWORD		sectionLen = 0;
	DWORD		keyLen = 0;
	DWORD		valueLen = 0;
	DWORD		type;
    FILETIME	lastWriteTime;
	SAPDB_Bool	nextSectionOk = SAPDB_FALSE;
	SAPDB_Bool	nextKeyOk = SAPDB_FALSE;
	char		currentSection	[REG_MAX_SECTION_LEN+1];
	char		newSection		[REG_MAX_SECTION_LEN+1];
	char		key				[REG_MAX_KEY_LEN+1];
	char		value			[REG_MAX_VALUE_LEN+1];
	
	if ( !registry )
		return SAPDB_FALSE;

	/* Open path */
	if ( RegOpenKeyEx( HKEY_LOCAL_MACHINE, section, 0,
					   KEY_READ, &sectionHandle ) != ERROR_SUCCESS)
		return SAPDB_FALSE;

	/* Get first section */
	sectionIndex = 0;
	sectionLen = REG_MAX_SECTION_LEN+1;
	regRC = RegEnumKeyEx( sectionHandle, sectionIndex++, currentSection,
						  &sectionLen, NULL, NULL, NULL, &lastWriteTime );

	/* Insert all sections */
	while ( regRC == ERROR_SUCCESS ) {
		/* Can we extend the path? */
		if ( strlen(section) + sectionLen + 1 > REG_MAX_SECTION_LEN ) {
		        RegCloseKey( sectionHandle );
				return SAPDB_FALSE;
		}

		/* Extend the path */
		sp77sprintf( newSection, REG_MAX_SECTION_LEN+1, "%s\\%s", section, currentSection );

		/* Insert section into internal tree */
		if ( !Reg_CommonInsertSection( registry, (const char*)newSection, NULL )) {
		    RegCloseKey( sectionHandle );
			return SAPDB_FALSE;
		}

		/* Open key */
		if ( RegOpenKeyEx( HKEY_LOCAL_MACHINE, newSection, 0,
						   KEY_READ, &keyHandle ) != ERROR_SUCCESS ) {
		    RegCloseKey( sectionHandle );
			return SAPDB_FALSE;
		}

		/* Get first key */
		keyIndex = 0;
		keyLen = REG_MAX_KEY_LEN+1;
		valueLen = REG_MAX_VALUE_LEN+1;
		regRC = RegEnumValue( keyHandle, keyIndex++, key, &keyLen,
							  NULL, &type, value, &valueLen );

		/* Insert all keys */
		while ( regRC == ERROR_SUCCESS ) {
			/* Insert key into internal tree */
			if ( !Reg_CommonInsertKey( registry, (const char*)newSection, key, value )) {
			    RegCloseKey( sectionHandle );
			    RegCloseKey( keyHandle );
				return SAPDB_FALSE;
			}

			/* Get next key */
			keyLen = REG_MAX_KEY_LEN+1;
			valueLen = REG_MAX_VALUE_LEN+1;
			regRC = RegEnumValue( keyHandle, keyIndex++, key, &keyLen,
								  NULL, &type, value, &valueLen );
		}

		/* Close key */
		RegCloseKey( keyHandle );

		/* Read child sections */
		if ( !ReadRegistry( registry, newSection )) {
		    RegCloseKey( sectionHandle );
			return SAPDB_FALSE;
		}

		/* Get next section */
		sectionLen = REG_MAX_SECTION_LEN+1;

		regRC = RegEnumKeyEx( sectionHandle, sectionIndex++, currentSection,
							  &sectionLen, NULL, NULL, NULL, &lastWriteTime );
	};

	/* Close key */
	RegCloseKey( sectionHandle );

	return SAPDB_TRUE;
}

/*!***************************************************************************

	WriteSection

******************************************************************************/

SAPDB_Bool WriteSection ( Reg_SectionP	section	)
{

	HKEY			keyHandle = NULL;
	Reg_KeyNodeP		currentKeyNode = NULL;
	Reg_SectionNodeP	currentChildSectionNode = NULL;

	/* Check */
	if ( !section )
		return SAPDB_FALSE;

	/* Open and write section */
	if ( RegCreateKeyEx( HKEY_LOCAL_MACHINE, section->sectionPath, 0, "",
						 REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL,
						 &keyHandle, NULL) != ERROR_SUCCESS)
		return SAPDB_FALSE;

	/* Get current key node */
	currentKeyNode = section->keys;
	while( currentKeyNode ) {
		/* Write current key */
		if ( !WriteKey( keyHandle, currentKeyNode->key ))
			return SAPDB_FALSE;

		/* Get next key node */
		currentKeyNode = currentKeyNode->nextKeyNode;
	}

	/* Close section */
    RegCloseKey( keyHandle );
    keyHandle = NULL;

	/* Write all child sections */
	currentChildSectionNode = section->childSections;
	while( currentChildSectionNode ) {
		/* Write current child section */
		if ( !WriteSection ( currentChildSectionNode->section ))
			return SAPDB_FALSE;

		/* Get next child section node */
		currentChildSectionNode = currentChildSectionNode->nextSectionNode;
	}

	return SAPDB_TRUE;

}

/*!***************************************************************************

	WriteKey

******************************************************************************/

SAPDB_Bool WriteKey ( HKEY	keyHandle,
					  Reg_KeyP	key			)
{

	/* Check */
	if ( !keyHandle || !key )
		return SAPDB_FALSE;

	/* Write key */
    if ( RegSetValueEx( keyHandle, key->key, 0,	REG_SZ,
						key->value,	strlen(key->value) + 1 ) != ERROR_SUCCESS )
		return SAPDB_FALSE;

	return SAPDB_TRUE;

}

/****************************************************************************
/***
/***	DropRegistry
/***
/****************************************************************************/

SAPDB_Bool DropRegistry( const char	*path	)
{

	HKEY	regHandle;
	LONG	regRC;
	char	sectionName[REG_MAX_SECTION_LEN+1];
	char	parentSectionPath[REG_MAX_SECTION_LEN+1];

	/* Checks */
	if ( !path )
		return SAPDB_FALSE;

	/* Get parent section */
	if ( !SplitSectionPath ( path, sectionName, parentSectionPath ))
		return SAPDB_FALSE;

	/* Open registry */
	regRC = RegOpenKeyEx( HKEY_LOCAL_MACHINE, parentSectionPath, 0, KEY_ALL_ACCESS, &regHandle );
	if ( regRC != ERROR_SUCCESS)
		return SAPDB_FALSE;

	/* Drop registry */
    regRC = SHDeleteKey( regHandle, sectionName );
    if ( regRC != ERROR_SUCCESS ) {
		RegCloseKey( regHandle );
		return SAPDB_FALSE;
    };

	/* Close registry */
    RegCloseKey( regHandle );

    return SAPDB_TRUE;

}

/*===========================================================================*/
/*===
/*===	End
/*===
/*===========================================================================*/
