/*! 
  -----------------------------------------------------------------------------
 
  module: SAPDBWeb_String.hpp
 
  -----------------------------------------------------------------------------
 
  responsible:  BerndV
 
  special area: universal stuff for web access
   
  description:  a string class like STL string

  version:      7.3 and higher 
  -----------------------------------------------------------------------------
 
                          Copyright by SAP AG, 2000
 
  -----------------------------------------------------------------------------


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*/

#ifndef _SAPDBWeb_String_HPP_
#define _SAPDBWeb_String_HPP_

#include "SAPDBCommon/SAPDB_Types.hpp"
#include "SAPDBCommon/SAPDB_DynamicUTF8String.hpp"

/*
  -----------------------------------------------------------------------------
  Class:        SAPDBWeb_Array
  -----------------------------------------------------------------------------

  description:  Array

  -----------------------------------------------------------------------------
 */
template <class E>
class SAPDBWeb_Array
{
public:

  typedef int       size_type;
  typedef       E & reference;
  typedef const E & const_reference;

  explicit SAPDBWeb_Array  ( );
           SAPDBWeb_Array  (const SAPDBWeb_Array & aArray);
          ~SAPDBWeb_Array  ( );

  inline reference       operator [] ( size_type nPos);
  inline const_reference operator [] ( size_type nPos) const;

  SAPDBWeb_Array & operator = (const SAPDBWeb_Array & aArray);

protected:
  
  void reallocate(size_type nNewSize = 0);

  E           m_theEmpty;
  E         * m_theArray;
  size_type   m_theSize;

};

/* EndClass: SAPDBWeb_Array */

/*
  =============================================================================
  implementation class SAPDBWeb_Array
  =============================================================================
 */

/*
  -----------------------------------------------------------------------------
  private macros
  -----------------------------------------------------------------------------
 */
#define SAPDBWEB_ARRAY_SIZE_STEP 10

/*
  -----------------------------------------------------------------------------
  public constructor SAPDBWeb_Array :: SAPDBWeb_Array
  -----------------------------------------------------------------------------
 */
template <class E>
inline SAPDBWeb_Array<E> :: SAPDBWeb_Array ( )
                                : m_theArray(0),
                                  m_theSize(0)
{

} // end SAPDBWeb_Array :: SAPDBWeb_Array
/*
  -----------------------------------------------------------------------------
 */
template <class E>
inline SAPDBWeb_Array<E> :: SAPDBWeb_Array (const SAPDBWeb_Array & aArray)
                                : m_theArray(0),
                                  m_theSize(0)
{
  *this = aArray;
} // end SAPDBWeb_Array :: SAPDBWeb_Array

/*
  -----------------------------------------------------------------------------
  public operator SAPDBWeb_Array :: =
  -----------------------------------------------------------------------------
 */
template <class E>
inline SAPDBWeb_Array<E> & SAPDBWeb_Array<E>::operator = (  const SAPDBWeb_Array & aArray )
{
  reallocate(aArray.m_theSize);

  for (size_type nIndex = 0; nIndex < aArray.m_theSize; ++nIndex) {
    this->m_theArray[nIndex] = aArray[nIndex];
  } // end if

  return *this;
} // end SAPDBWeb_Array::operator =

/*
  -----------------------------------------------------------------------------
  public destructor SAPDBWeb_Array :: ~SAPDBWeb_Array
  -----------------------------------------------------------------------------
 */
template <class E>
inline SAPDBWeb_Array<E> :: ~SAPDBWeb_Array ( )
{
  if (this->m_theArray != 0) {
    delete [] this->m_theArray;
  } // end if
  this->m_theArray = 0;
  this->m_theSize   = 0;
} // end SAPDBWeb_Array :: ~SAPDBWeb_Array

/*
  -----------------------------------------------------------------------------
  public operator SAPDBWeb_Array :: []
  -----------------------------------------------------------------------------
 */
template <class E>
inline typename SAPDBWeb_Array<E>::reference SAPDBWeb_Array<E>::operator [] ( size_type nPos ) 
{
  if (nPos >= this->m_theSize) {
    reallocate(nPos);
  } // end if
  return this->m_theArray[nPos];
} // end SAPDBWeb_Array<E>::operator []
/*
  -----------------------------------------------------------------------------
 */
template <class E>
inline typename  SAPDBWeb_Array<E>::const_reference SAPDBWeb_Array<E>::operator [] ( size_type nPos ) const
{
  if (nPos >= this->m_theSize) {
    return this->m_theEmpty;
  } else {
    return this->m_theArray[nPos];
  } // end if
} // end SAPDBWeb_Array<E>::operator []

/*
  -----------------------------------------------------------------------------
  private function SAPDBWeb_Array :: reallocate
  -----------------------------------------------------------------------------
 */
template <class E>
inline void SAPDBWeb_Array<E> :: reallocate (size_type nNewSize )
{
  size_type newSize = ((nNewSize / SAPDBWEB_ARRAY_SIZE_STEP) * SAPDBWEB_ARRAY_SIZE_STEP) + SAPDBWEB_ARRAY_SIZE_STEP;

  if (newSize != this->m_theSize) {
    E* newArray = new E[newSize + 1];

    for (size_type nIndex = 0;nIndex < newSize && nIndex < this->m_theSize; ++nIndex) {
      newArray[nIndex] = this->m_theArray[nIndex];
    } // end for

    if (this->m_theArray != 0) {
      delete [] this->m_theArray;
    } // end if

    this->m_theSize  = newSize;
    this->m_theArray = newArray;
  } // end if
} // end SAPDBWeb_Array :: reallocate

/*
  -----------------------------------------------------------------------------
  Class:        SAPDBWeb_BasicString
  -----------------------------------------------------------------------------

  description:  String

  -----------------------------------------------------------------------------
 */
template <class E>
class SAPDBWeb_BasicString : public SAPDBWeb_Array<E>
{
public:

  explicit SAPDBWeb_BasicString  ( );
  SAPDBWeb_BasicString  
        ( const E                               * aString);
  SAPDBWeb_BasicString  
        ( const E                               * aString,
          typename SAPDBWeb_Array<E>::size_type   nLength);
  SAPDBWeb_BasicString  
        ( const SAPDBWeb_BasicString & aString);
  SAPDBWeb_BasicString  
        ( const SAPDBWeb_BasicString & aString,
          typename SAPDBWeb_Array<E>::size_type   nLength);
  SAPDBWeb_BasicString  
        ( const SAPDB_DynamicUTF8String & aString);
  SAPDBWeb_BasicString  
        ( int   nNumber,
          int   nBase = 10);

  ~SAPDBWeb_BasicString  
        ( );

  SAPDBWeb_BasicString  & append 
        ( const SAPDBWeb_BasicString            & aString );
  SAPDBWeb_BasicString  & append 
        ( const SAPDBWeb_BasicString            & aString,
          typename SAPDBWeb_Array<E>::size_type   nLength);
  SAPDBWeb_BasicString  & append 
        ( const E                               * aString );
  SAPDBWeb_BasicString  & append 
        ( const E                               * aString, 
          typename SAPDBWeb_Array<E>::size_type   nLength);

  SAPDBWeb_BasicString  & assign 
        ( const SAPDBWeb_BasicString            & aString );
  SAPDBWeb_BasicString  & assign 
        ( const SAPDBWeb_BasicString            & aString,
          typename SAPDBWeb_Array<E>::size_type   nLength);
  SAPDBWeb_BasicString  & assign 
        ( const E                               * aString );
  SAPDBWeb_BasicString  & assign 
        ( const E                               * aString, 
          typename SAPDBWeb_Array<E>::size_type   nLength);
  
  int                     compare
        ( const SAPDBWeb_BasicString            & aString );

  SAPDBWeb_BasicString  & operator =  
        ( const SAPDBWeb_BasicString            & aString );
  SAPDBWeb_BasicString  & operator =  
        ( const E                               * aString );

  bool                                  empty  ( ) const {return (this->m_theLength == 0);}
  typename SAPDBWeb_Array<E>::size_type length ( ) const {return this->m_theLength;}
  const E *                             c_str  ( ) const;

protected:
  
  typename SAPDBWeb_Array<E>::size_type   m_theLength;

};

/* EndClass: SAPDBWeb_BasicString */

typedef SAPDBWeb_BasicString<SAPDB_Char> SAPDBWeb_String;

/*
  =============================================================================
  implementation
  =============================================================================
 */

/*
  -----------------------------------------------------------------------------
  private macros
  -----------------------------------------------------------------------------
 */
#define SAPDBWEB_STRING_SIZE_STEP 100

/*
  -----------------------------------------------------------------------------
  public constructor SAPDBWeb_BasicString :: SAPDBWeb_BasicString
  -----------------------------------------------------------------------------
 */
template <class E>
inline SAPDBWeb_BasicString<E> :: SAPDBWeb_BasicString 
      ( )
      : SAPDBWeb_Array<E>(),
        m_theLength(0)
{
  this->reallocate();
  this->m_theArray [this->m_theLength] = E(0);
} // end SAPDBWeb_BasicString :: SAPDBWeb_BasicString
/*
  -----------------------------------------------------------------------------
 */
template <class E>
inline SAPDBWeb_BasicString<E> :: SAPDBWeb_BasicString 
      ( const E *aString )
      : SAPDBWeb_Array<E>(),
        m_theLength(0)
{
  if (aString != 0) {
    assign(aString);
  } else {
    this->reallocate();
    this->m_theArray [this->m_theLength] = E(0);
  } // end if
} // end SAPDBWeb_BasicString :: SAPDBWeb_BasicString
/*
  -----------------------------------------------------------------------------
 */
template <class E>
inline SAPDBWeb_BasicString<E> :: SAPDBWeb_BasicString 
      ( const E                               * aString,
        typename SAPDBWeb_Array<E>::size_type   nLength)
      : SAPDBWeb_Array<E>(),
        m_theLength(0)
{
  if (aString != 0) {
    assign(aString, nLength);
  } else {
    this->reallocate();
    this->m_theArray [this->m_theLength] = E(0);
  } // end if
} // end SAPDBWeb_BasicString :: SAPDBWeb_BasicString
/*
  -----------------------------------------------------------------------------
 */
template <class E>
inline SAPDBWeb_BasicString<E> :: SAPDBWeb_BasicString 
      ( const SAPDBWeb_BasicString & aString )
      : SAPDBWeb_Array<E>(),
        m_theLength(0)
{
  assign(aString);
} // end SAPDBWeb_BasicString :: SAPDBWeb_BasicString
/*
  -----------------------------------------------------------------------------
 */
template <class E>
inline SAPDBWeb_BasicString<E> :: SAPDBWeb_BasicString 
      ( const SAPDBWeb_BasicString            & aString,
        typename SAPDBWeb_Array<E>::size_type   nLength)
      : SAPDBWeb_Array<E>(),
        m_theLength(0)
{
  assign(aString, nLength);
} // end SAPDBWeb_BasicString :: SAPDBWeb_BasicString
/*
  -----------------------------------------------------------------------------
 */
template <class E>
inline SAPDBWeb_BasicString<E> :: SAPDBWeb_BasicString 
      ( const SAPDB_DynamicUTF8String & aString )
      : SAPDBWeb_Array<E>(),
        m_theLength(0)
{
  assign(REINTERPRET_CAST(const E *, aString.Data()), aString.BasisSize());
} // end SAPDBWeb_BasicString :: SAPDBWeb_BasicString
/*
  -----------------------------------------------------------------------------
 */
template <class E>
inline SAPDBWeb_BasicString<E> :: SAPDBWeb_BasicString 
      ( int nNumber,
        int nBase)
      : SAPDBWeb_Array<E>(),
        m_theLength(0)
{
  E   aBuffer[30];
  E*  pCurrent = 0;
  E   Sign(0);
  int nDigit   = 0;

  pCurrent = &(aBuffer[29]);

  *pCurrent = E(0);

  --pCurrent;

  if (nNumber < 0) {
    nNumber = nNumber * (-1);
    Sign = E('-');
  } // end if

  if (nNumber > 0) {
    while (nNumber > 0) {
      nDigit = nNumber % nBase;
      nNumber = (nNumber - nDigit) / nBase;
      switch (nDigit) {
        case  0:*pCurrent = E('0');break;
        case  1:*pCurrent = E('1');break;
        case  2:*pCurrent = E('2');break;
        case  3:*pCurrent = E('3');break;
        case  4:*pCurrent = E('4');break;
        case  5:*pCurrent = E('5');break;
        case  6:*pCurrent = E('6');break;
        case  7:*pCurrent = E('7');break;
        case  8:*pCurrent = E('8');break;
        case  9:*pCurrent = E('9');break;
        case 10:*pCurrent = E('A');break;
        case 11:*pCurrent = E('B');break;
        case 12:*pCurrent = E('C');break;
        case 13:*pCurrent = E('D');break;
        case 14:*pCurrent = E('E');break;
        case 15:*pCurrent = E('F');break;
      } // end switch
      --pCurrent;
    } // end while
    if (Sign != E(0)) {
      *pCurrent = Sign;
    } else {
      ++pCurrent;
    } // end if
  } else {
    *pCurrent = E('0');
  } // end if

  assign(pCurrent);
} // end SAPDBWeb_BasicString :: SAPDBWeb_BasicString

/*
  -----------------------------------------------------------------------------
  public destructor SAPDBWeb_BasicString :: ~SAPDBWeb_BasicString
  -----------------------------------------------------------------------------
 */
template <class E>
inline SAPDBWeb_BasicString<E> :: ~SAPDBWeb_BasicString 
      ( )
{
  this->m_theLength = 0;
} // end SAPDBWeb_BasicString :: ~SAPDBWeb_BasicString

/*
  -----------------------------------------------------------------------------
  private function SAPDBWeb_BasicString :: assign
  -----------------------------------------------------------------------------
 */
template <class E>
inline SAPDBWeb_BasicString<E>  & SAPDBWeb_BasicString<E> :: assign 
      ( const SAPDBWeb_BasicString & aString )
{
  return assign(aString.c_str(), aString.length());
} // end SAPDBWeb_BasicString :: assign
/*
  -----------------------------------------------------------------------------
 */
template <class E>
inline SAPDBWeb_BasicString<E>  & SAPDBWeb_BasicString<E> :: assign 
      ( const SAPDBWeb_BasicString            & aString,
        typename SAPDBWeb_Array<E>::size_type   nLength)
{
  return assign(aString.c_str(), nLength);
} // end SAPDBWeb_BasicString :: assign
/*
  -----------------------------------------------------------------------------
 */
template <class E>
inline SAPDBWeb_BasicString<E>  & SAPDBWeb_BasicString<E> :: assign 
      ( const E * aString )
{
  if (aString != 0) {
    typename SAPDBWeb_Array<E>::size_type nLength = 0;
    while (aString[nLength] != E(0)) ++nLength;
    assign(aString, nLength);
  } else {
    this->reallocate();
    this->m_theLength = 0;
    this->m_theArray [this->m_theLength] = E(0);
  } // end if

  return (*this);
} // end SAPDBWeb_BasicString :: assign
/*
  -----------------------------------------------------------------------------
 */
template <class E>
inline SAPDBWeb_BasicString<E>  & SAPDBWeb_BasicString<E> :: assign 
      ( const E                               * aString, 
        typename SAPDBWeb_Array<E>::size_type   nLength)
{
  if (aString != 0) {
    if ((nLength > this->m_theSize)  || (nLength < (this->m_theSize/2)) || (this->m_theSize == 0)) {
      reallocate(nLength);
    } // end if

    for (typename SAPDBWeb_Array<E>::size_type nIndex = 0;nIndex < nLength;++nIndex) {
      this->m_theArray[nIndex] = aString[nIndex];
    } // end for

    this->m_theLength = nLength;
    this->m_theArray[this->m_theLength] = E(0);
  } else {
    this->reallocate();
    this->m_theLength = 0;
    this->m_theArray [this->m_theLength] = E(0);
  } // end if

  return (*this);
} // end SAPDBWeb_BasicString :: assign

/*
  -----------------------------------------------------------------------------
  private function SAPDBWeb_BasicString :: append
  -----------------------------------------------------------------------------
 */
template <class E>
inline SAPDBWeb_BasicString<E>  & SAPDBWeb_BasicString<E> :: append 
      ( const SAPDBWeb_BasicString & aString )
{
  return append(aString.c_str(), aString.length());
} // end SAPDBWeb_BasicString :: append
/*
  -----------------------------------------------------------------------------
 */
template <class E>
inline SAPDBWeb_BasicString<E>  & SAPDBWeb_BasicString<E> :: append 
      ( const SAPDBWeb_BasicString            & aString,
        typename SAPDBWeb_Array<E>::size_type   nLength)
{
  return append(aString.c_str(), nLength);
} // end SAPDBWeb_BasicString :: append
/*
  -----------------------------------------------------------------------------
 */
template <class E>
inline SAPDBWeb_BasicString<E>  & SAPDBWeb_BasicString<E> :: append 
      ( const E * aString )
{
  if (aString != 0) {
    typename SAPDBWeb_Array<E>::size_type nLength = 0;
    while (aString[nLength] != E(0)) ++nLength;
    append(aString, nLength);
  } // end if

  return (*this);
} // end SAPDBWeb_BasicString :: append
/*
  -----------------------------------------------------------------------------
 */
template <class E>
inline SAPDBWeb_BasicString<E>  & SAPDBWeb_BasicString<E> :: append 
      ( const E                               * aString, 
        typename SAPDBWeb_Array<E>::size_type   nLength)
{
  if (aString != 0) {
    int nWholeLength = nLength + this->m_theLength;

    if ((nWholeLength > this->m_theSize)  || (this->m_theSize == 0)) {
      this->reallocate(nWholeLength);
    } // end if

    for (typename SAPDBWeb_Array<E>::size_type nIndex = this->m_theLength; nIndex < nWholeLength; ++nIndex) {
      this->m_theArray[nIndex] = aString[nIndex - this->m_theLength];
    } // end for

    this->m_theLength = nWholeLength;
    this->m_theArray[this->m_theLength] = E(0);
  } // end if

  return (*this);
} // end SAPDBWeb_BasicString :: append

/*
  -----------------------------------------------------------------------------
  public function SAPDBWeb_BasicString :: c_str
  -----------------------------------------------------------------------------
 */
template <class E>
inline const E * SAPDBWeb_BasicString<E> :: c_str ( ) const
{
  return &(this->m_theArray[0]);
} // end SAPDBWeb_BasicString :: c_str

/*
  -----------------------------------------------------------------------------
  public operator SAPDBWeb_BasicString :: =
  -----------------------------------------------------------------------------
 */
template <class E>
inline SAPDBWeb_BasicString<E> & SAPDBWeb_BasicString<E>::operator = (  const SAPDBWeb_BasicString & aString  )
{
  return assign(aString);
} // end SAPDBWeb_BasicString::operator =

/*
  -----------------------------------------------------------------------------
  public operator SAPDBWeb_BasicString :: =
  -----------------------------------------------------------------------------
 */
template <class E>
inline SAPDBWeb_BasicString<E>  & SAPDBWeb_BasicString<E>::operator = (  const E * aString  )
{
  return assign(aString);
} // end SAPDBWeb_BasicString::operator =

/*
  -----------------------------------------------------------------------------
  public function SAPDBWeb_BasicString :: compare
  -----------------------------------------------------------------------------
 */
template <class E>
inline int SAPDBWeb_BasicString<E> :: compare ( const SAPDBWeb_BasicString & aString )
{
  typename SAPDBWeb_Array<E>::size_type nMinLen = (this->m_theLength < aString.length()) ? this->m_theLength : aString.length();
  int       nReturn = 0;
  
  for (int nIndex = 0; nIndex < nMinLen && nReturn == 0; ++nIndex) {
    nReturn = this->m_theArray[nIndex] - aString[nIndex];
  } // end if
  
  return ((nReturn == 0) ? this->m_theLength - aString.length() : nReturn);
} // end SAPDBWeb_BasicString :: compare

#endif // _SAPDBWeb_String_HPP_
