/*
    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*/

/*!**********************************************************************

  module: StudioOAL_Result.cpp

  -----------------------------------------------------------------------

  responsible: WolfgangA

  special area: OAL

  description:  Result Class for OAL

  see also:

  -----------------------------------------------------------------------

  copyright:                Copyright by SAP AG, 2000

************************************************************************/



#include "SAPDB/SAPDBCommon/SAPDB_UTF8Basis.h"
#include "SAPDB/SQLStudio/OAL/StudioOAL_Result.hpp"

const SAPDB_Int4 MX_TIME_COLUMN_VALUE = 6;

//max length of column name
const SAPDB_Int4 MX_COLUMN_NAME			= 256;

//max length of date value
const SAPDB_Int4 MX_TIMESTAMP_COLUMN_VALUE = 16;
const SAPDB_Int4 MX_DOUBLE_COLUMN_VALUE	= 8;

const SAPDB_Int4 MX_COLUMN_GENERIC_NUMERIC		 = 64;
const SAPDB_Int4 MX_COLUMN_VALUE				 = 4096;

//number of bytes to fetch for a long value
const SAPDB_Int4 MX_COLUMN_LONG_VALUE			 = 32768;

//number of bytes for copying into clipboard
const SAPDB_Int4 MX_COLUMN_LONG_VALUE_CLIPBOARD = 1048576;



//define string constants
#define REPOFNULLVALUE		"?"
#define S1092				"S1092"	
#define STRTRUE				"TRUE"	
#define STRFALSE			"FALSE"	


StudioOAL_Result::StudioOAL_Result(SQLHANDLE hDbc,SQLHENV hEnv)
{

	SAPDB_ULong lRun;
	
	m_hDbc = hDbc;
	m_hEnv = hEnv;
	m_enumResultCursorType = stoalCursorForwardOnly;

    //init statement handle
	m_hStmt = 0;    
	m_bNoRows = SAPDB_TRUE;	
	m_bDBProcedureResult = SAPDB_FALSE;
	m_lLimitForLong = 32768;	

	m_pError = new StudioOAL_Error(m_hDbc,m_hEnv);	

	for( lRun = 0; lRun <= 1023;lRun++ ){

		m_pColumn[lRun] = NULL;
	}
	
}

StudioOAL_Result::~StudioOAL_Result()
{
	
	clearParameters() ;

	if( m_pError != NULL) {

		delete	m_pError;
		m_pError = NULL;
	}

	SAPDB_UInt lRun;
	for( lRun = 0; lRun <= 1023;lRun++ )
	{
		if (m_pColumn[lRun] != NULL ){

			delete m_pColumn[lRun];
		}
		m_pColumn[lRun] = NULL;
	}

	closeResult();
}


SAPDB_Bool StudioOAL_Result::openResult(const SAPDB_DynamicUTF8String& strCmd)
{  

    //check it
	if( m_pError == NULL) return SAPDB_FALSE;
    if( m_hDbc == SQL_NULL_HDBC || m_hEnv == SQL_NULL_HENV)  return SAPDB_FALSE;
	

    //init
	SQLRETURN		nRc = SQL_SUCCESS;
    m_bDBProcedureResult = SAPDB_FALSE;
    
    m_sStatement = strCmd;
    m_lFetchedRows = 0;
    m_bNoRows = SAPDB_TRUE;	

    //if there is a open result the close it
    if(m_hStmt != 0) closeResult();

    //allocate handle for statement
	nRc = SQLAllocHandle(SQL_HANDLE_STMT,m_hDbc,&m_hStmt);
	if( nRc != SQL_SUCCESS)
	{
		if(! (m_pError->checkSQLReturnCode(nRc,m_hStmt))) {

			SQLFreeStmt(m_hStmt, SQL_DROP);
			m_hStmt = SQL_NULL_HSTMT;
			return SAPDB_FALSE;
		}

	}		
    
    nRc = SQLSetStmtOption(m_hStmt, SQL_ATTR_ROW_BIND_TYPE, SQL_BIND_BY_COLUMN);
    if( nRc != SQL_SUCCESS ) {
    
		if(! (m_pError->checkSQLReturnCode(nRc,m_hStmt))) {
    
            SQLFreeStmt(m_hStmt, SQL_DROP);
			m_hStmt = SQL_NULL_HSTMT;
            return SAPDB_FALSE;
        }
    }

    switch (m_enumResultCursorType)
    {
        case stoalCursorForwardOnly:
    
            nRc = SQLSetStmtOption(m_hStmt, SQL_ATTR_CONCURRENCY, SQL_CONCUR_READ_ONLY);
            if( nRc != SQL_SUCCESS ) {
                
                if(! (m_pError->checkSQLReturnCode(nRc,m_hStmt))) {
        
                    SQLFreeStmt(m_hStmt, SQL_DROP);
					m_hStmt = SQL_NULL_HSTMT;
                    return SAPDB_FALSE;
                }
            }
            nRc = SQLSetStmtOption(m_hStmt, SQL_ATTR_CURSOR_TYPE, SQL_CURSOR_FORWARD_ONLY);
            if( nRc != SQL_SUCCESS ) {
              if(! (m_pError->checkSQLReturnCode(nRc,m_hStmt))) {
    
                    SQLFreeStmt(m_hStmt, SQL_DROP);
					m_hStmt = SQL_NULL_HSTMT;
                    return SAPDB_FALSE;
                }
            }

			break;
        
        case stoalCursorStatic:

            nRc = SQLSetStmtOption(m_hStmt, SQL_ATTR_CONCURRENCY, SQL_CONCUR_ROWVER);
            if( nRc != SQL_SUCCESS ) {
                if(! (m_pError->checkSQLReturnCode(nRc,m_hStmt))) {
        
                    SQLFreeStmt(m_hStmt, SQL_DROP);
					m_hStmt = SQL_NULL_HSTMT;
                    return SAPDB_FALSE;
                }
            }
            
            nRc = SQLSetStmtOption(m_hStmt, SQL_ATTR_CURSOR_TYPE, SQL_CURSOR_STATIC);
            if( nRc != SQL_SUCCESS ) {
                
                if(! (m_pError->checkSQLReturnCode(nRc,m_hStmt))) {
                
                    SQLFreeStmt(m_hStmt, SQL_DROP);
					m_hStmt = SQL_NULL_HSTMT;
                    return SAPDB_FALSE;
                }
            }
			break;
            
		case stoalOptimizeCursor:
    
			nRc = SQLSetStmtOption(m_hStmt, SQL_ATTR_CONCURRENCY, SQL_CONCUR_READ_ONLY);
			if( nRc != SQL_SUCCESS ) {
            
				if(! (m_pError->checkSQLReturnCode(nRc,m_hStmt))) {
    
					SQLFreeStmt(m_hStmt, SQL_DROP);
					m_hStmt = SQL_NULL_HSTMT;
					return SAPDB_FALSE;
				}
			}
        
			nRc = SQLSetStmtOption(m_hStmt, SQL_ATTR_OPTIMIZE_CURSOR, 1);
			if( nRc != SQL_SUCCESS ) {
        
				if(! (m_pError->checkSQLReturnCode(nRc,m_hStmt))) {

					 // Out of Option range						
					if( m_pError->getSQLStateErrorNumber() == (SAPDB_UTF8*)S1092) { 
                
						SQLFreeStmt(m_hStmt, SQL_DROP);
						m_hStmt = SQL_NULL_HSTMT;
						return SAPDB_FALSE;
					}
				}
			}
			nRc = SQLSetStmtOption(m_hStmt, SQL_ATTR_CURSOR_TYPE, SQL_CURSOR_KEYSET_DRIVEN);
			if( nRc != SQL_SUCCESS ) {
        
				if(! (m_pError->checkSQLReturnCode(nRc,m_hStmt))) {

					SQLFreeStmt(m_hStmt, SQL_DROP);
					m_hStmt = SQL_NULL_HSTMT;
					return SAPDB_FALSE;
				}
			}
			break;
         
		default:
			return SAPDB_FALSE;
			break;
   
   }

   nRc = SQLSetStmtOption(m_hStmt, SQL_ATTR_USE_BOOKMARKS, SQL_UB_ON);
   if( nRc != SQL_SUCCESS ) {

        if(! (m_pError->checkSQLReturnCode(nRc,m_hStmt))) {
    
            SQLFreeStmt(m_hStmt, SQL_DROP);
			m_hStmt = SQL_NULL_HSTMT;
            return SAPDB_FALSE;
        }
    }
    
    nRc = SQLSetStmtOption(m_hStmt, SQL_ROWSET_SIZE, 1);
    if( nRc != SQL_SUCCESS ) {

        if(! (m_pError->checkSQLReturnCode(nRc,m_hStmt))) {
    
            SQLFreeStmt(m_hStmt, SQL_DROP);
			m_hStmt = SQL_NULL_HSTMT;
            return SAPDB_FALSE;
        }
    }
    
    //open the result set
    if( ! execStmt() ) {

        //free statement
        SQLFreeStmt(m_hStmt, SQL_DROP);
		m_hStmt = SQL_NULL_HSTMT;
        return SAPDB_FALSE;
    }
    
    if(! getColDescriptions() ) 	return SAPDB_FALSE;
	

    //is there a result
	m_lRowCount = rowCount(&m_hStmt);
    if( m_lRowCount == 0 ) {
        
        return SAPDB_TRUE;
	}
	else{
        m_bNoRows = SAPDB_FALSE;
    } 
    
    return SAPDB_TRUE;

}


SAPDB_Long StudioOAL_Result::rowCount(SQLHANDLE *hStmt)
{

    SAPDB_Long		 lRowCount = 0;
    SQLRETURN	 nRc = 0;

    
     /*check it*/
    if( m_hDbc == SQL_NULL_HDBC || m_hEnv == SQL_NULL_HENV || m_pError == NULL) return 0;
	

    /* Try to get the number of rows*/
    nRc = SQLRowCount(*hStmt, &lRowCount);
	if( nRc != SQL_SUCCESS){
    
        if(!(m_pError->checkSQLReturnCode(nRc,*hStmt)))  return 0;        
    }
    
	return lRowCount;


}

void StudioOAL_Result::closeResult()
{       
    
    //check it
    if( m_hStmt == SQL_NULL_HSTMT) return;	    
	SQLFreeStmt(m_hStmt, SQL_DROP);    
    m_hStmt = SQL_NULL_HSTMT;

}
SAPDB_Bool StudioOAL_Result::getColDescriptions() 
{

    SQLUSMALLINT		lRun;    
    SQLSMALLINT			nColNameLen;
    SQLSMALLINT			nColType;
    SQLUINTEGER			lColLength;
    SQLRETURN			nRc;
    SQLSMALLINT			nColDecimal;
    SQLSMALLINT			nColNullable;

	SQLWCHAR  			strColNameBuffer[MX_COLUMN_NAME];
	SAPDB_DynamicUTF8String strColName;

    

    
    /*check it*/
	if( m_pError == NULL) return SAPDB_FALSE;
    if( m_hDbc == SQL_NULL_HDBC || m_hEnv == SQL_NULL_HENV) return SAPDB_FALSE;	
	

    /*get number of columns*/
    nRc = SQLNumResultCols(m_hStmt, &m_nColCount);
    if( nRc != SQL_SUCCESS) {
    
        if(! (m_pError->checkSQLReturnCode(nRc,m_hStmt))) {    

			return SAPDB_FALSE;
		}
	}    
    
    /*get the column descriptions*/
    for( lRun = 1; lRun <= m_nColCount;lRun++ )
	{
    
        m_pColumn[lRun-1] = new StudioOAL_Column;

 		/*SQLRETURN  SQL_API SQLDescribeCol(SQLHSTMT StatementHandle,
										    SQLUSMALLINT ColumnNumber,
										    SQLCHAR *ColumnName,
										    SQLSMALLINT BufferLength, 
										    SQLSMALLINT *NameLength,
										    SQLSMALLINT *DataType, 
										    SQLUINTEGER *ColumnSize,
										    SQLSMALLINT *DecimalDigits, 
										    SQLSMALLINT *Nullable);*/

        /*get the individual column description*/
        nRc = SQLDescribeColW(m_hStmt,
							 lRun,
							 (SQLWCHAR *)strColNameBuffer, 
							 MX_COLUMN_NAME,
							 &nColNameLen, 
							 &nColType,
							 &lColLength,
							 &nColDecimal,
							 &nColNullable);
        if( nRc != SQL_SUCCESS ){        

               if( !(m_pError->checkSQLReturnCode(nRc,m_hStmt)) ) {    

				   return SAPDB_FALSE;
			   }
        }
        
		strColName.ConvertFromUCS2((SAPDB_UCS2*)strColNameBuffer, (SAPDB_UCS2*)(strColNameBuffer+nColNameLen));        
		m_pColumn[lRun-1]->setColName (strColName); 
        
        m_pColumn[lRun-1]->setODBCColType (nColType);
        m_pColumn[lRun-1]->setColLength(lColLength);
        m_pColumn[lRun-1]->setColDecimal(nColDecimal);
        
        if (nColNullable != 0){
        
            m_pColumn[lRun-1]->setColNullable(SAPDB_TRUE);		
        }
		else{
            m_pColumn[lRun-1]->setColNullable(SAPDB_FALSE);
        }      
		
    }

	for (lRun = 1023; lRun > m_nColCount;lRun-- ){

		if(m_pColumn[lRun] != NULL ){

			delete m_pColumn[lRun];
		}
		m_pColumn[lRun] = NULL;
	}
    
    
	return SAPDB_TRUE;

}

SAPDB_Bool StudioOAL_Result::getColumnName(SAPDB_UInt2 nCol,SAPDB_DynamicUTF8String &strColName)
{
	
	/*init*/
	strColName.Erase();
	if (nCol > m_nColCount)	return SAPDB_FALSE;
	
	m_pColumn[nCol-1]->getColName (strColName); 

	return SAPDB_TRUE;
	
}
SAPDB_Bool StudioOAL_Result::execStmt() 
{ 
	/*-------------------------------------------------------------------------*
     * check if pointer to error class is not null 						       *
     *-------------------------------------------------------------------------*/
	if (m_pError == NULL)	return SAPDB_FALSE;	

	/*-------------------------------------------------------------------------*
     * check if dbc handle and enviroment handle are not null 			       *
     *-------------------------------------------------------------------------*/
    if( m_hDbc == SQL_NULL_HDBC || m_hEnv == SQL_NULL_HENV) return SAPDB_FALSE;	

	/*-------------------------------------------------------------------------*
     * local variables										 			       *
     *-------------------------------------------------------------------------*/
	SQLRETURN	nRc = 0;	
	SAPDB_UCS2* pConvertOut = NULL;
	
   /*-------------------------------------------------------------------------*
    * use a copy for conversion	to ucs2pointer								  *
    *-------------------------------------------------------------------------*/
	SAPDB_DynamicUTF8String strTemp = m_sStatement;

	/*-------------------------------------------------------------------------*
     * init conversion buffer								 			       *
     *-------------------------------------------------------------------------*/
	memset(m_strCmdBuffer,0,sizeof(m_strCmdBuffer));

	/*-------------------------------------------------------------------------*
     * convert utf8 to ucs2									 			       *
     *-------------------------------------------------------------------------*/
	if (strTemp.ConvertToUCS2(m_strCmdBuffer,
							  (m_strCmdBuffer+(sizeof(m_strCmdBuffer)/sizeof(SAPDB_UCS2))),
							  pConvertOut) != SAPDB_UTF8Basis::Success ) return SAPDB_FALSE;		
	
	/*-------------------------------------------------------------------------*
     * check if something was converted						 			       *
     *-------------------------------------------------------------------------*/
	if(pConvertOut<=m_strCmdBuffer) return SAPDB_FALSE;

	/*-------------------------------------------------------------------------*
     * execute statement									 			       *
     *-------------------------------------------------------------------------*/
	nRc = SQLExecDirectW(m_hStmt,(SQLWCHAR*)m_strCmdBuffer, ((SQLINTEGER)(pConvertOut-m_strCmdBuffer)));
	if (nRc != SQL_SUCCESS) {
    
        if (!(m_pError->checkSQLReturnCode(nRc,m_hStmt))) {    

			return SAPDB_FALSE;
		}
	}    

	return SAPDB_TRUE;	
}


SAPDB_Bool StudioOAL_Result::fetch() 
{

	

	/*-------------------------------------------------------------------------*
     * check if pointer to error class is not null 						       *
     *-------------------------------------------------------------------------*/
	if (m_pError == NULL)	return SAPDB_FALSE;

	/*-------------------------------------------------------------------------*
     * check if dbc handle and enviroment handle are not null 			       *
     *-------------------------------------------------------------------------*/
    if (m_hDbc == SQL_NULL_HDBC || m_hEnv == SQL_NULL_HENV || m_hStmt == SQL_NULL_HSTMT) return SAPDB_FALSE;	
    
	/*-------------------------------------------------------------------------*
     * local variable										 			       *
     *-------------------------------------------------------------------------*/ 
	SQLRETURN	nRc = 0;
    
    /*-------------------------------------------------------------------------*
     * init member											 			       *
     *-------------------------------------------------------------------------*/ 
    m_bNoRows = SAPDB_FALSE;
    
	/*-------------------------------------------------------------------------*
     * fetch row											 			       *
     *-------------------------------------------------------------------------*/ 
	nRc = SQLFetch(m_hStmt);
    
    if (!(m_pError->checkSQLReturnCode(nRc,m_hStmt))) {    

		m_bNoRows = SAPDB_TRUE;			
		return SAPDB_FALSE;
	}
	    
        
	if (m_pError->getReturnCode() == SQL_NO_DATA_FOUND)
    
        m_bNoRows = SAPDB_TRUE;				
	else
        m_lFetchedRows ++;
	

	return SAPDB_TRUE;

}

SAPDB_Bool StudioOAL_Result::extendedFetch(SAPDB_UInt2 lFetchType,SAPDB_ULong lRow,SAPDB_ULong *NoRowsFetched,SAPDB_UInt2 *nRowStatus) 
{

	SQLRETURN	nRc;
    
    /*inits*/
    m_bNoRows = SAPDB_FALSE;
	*NoRowsFetched = 0;

	/*check it*/
	if( m_pError == NULL) return SAPDB_FALSE;	
	
    if( m_hDbc == SQL_NULL_HDBC || 
		m_hEnv == SQL_NULL_HENV || 
		m_hStmt == SQL_NULL_HSTMT)  return SAPDB_FALSE;
	

	if (lFetchType == SQL_FETCH_LAST  || 
		lFetchType == SQL_FETCH_FIRST ||
		lFetchType == SQL_FETCH_ABSOLUTE )	m_lFetchedRows = 0;
		    
 
    /*fetch*/
	nRc =SQLExtendedFetch(m_hStmt, lFetchType, lRow, NoRowsFetched, nRowStatus);
    
    if(! (m_pError->checkSQLReturnCode(nRc,m_hStmt))) {    

		m_bNoRows = SAPDB_TRUE;			
		return SAPDB_FALSE;
	}

        
	if( m_pError->getReturnCode() == SQL_NO_DATA_FOUND)
    
        m_bNoRows = SAPDB_TRUE;			
	else
        m_lFetchedRows ++;
	

	return SAPDB_TRUE;

}

SAPDB_Bool StudioOAL_Result::getColumnDataAsString(SAPDB_UInt2 nCol,SAPDB_DynamicUTF8String &outValue)
{
 

    SAPDB_Bool			bBoolValue;
    SAPDB_Long			lGetDataLen;
	SQLRETURN			nRc = 0;

	char 				*pValue = NULL;
	SAPDB_Long			lValueLen = 0;
    
	outValue.Erase();
    
    if( m_hDbc == SQL_NULL_HDBC || 
		m_hEnv == SQL_NULL_HENV ||
		m_hStmt == SQL_NULL_HSTMT) 	return SAPDB_FALSE;	
        
    if (m_pColumn[nCol-1] == NULL)	return SAPDB_FALSE;	
	if( m_pError == NULL)	return SAPDB_FALSE;	

    //get the data for the current col
    switch (m_pColumn[nCol-1]->getODBCColType())
	{    
        case SQL_DATE:
        
			pValue = new char [MX_COLUMN_GENERIC_NUMERIC];
			if (pValue == NULL) return SAPDB_FALSE;			
			
            //get the current value
            nRc = SQLGetData(m_hStmt, nCol, SQL_CHAR, pValue, MX_COLUMN_GENERIC_NUMERIC, &lValueLen);
            if( nRc != SQL_SUCCESS) {
    
				if(! (m_pError->checkSQLReturnCode(nRc,m_hStmt))) {    

					delete[] pValue;	
					return SAPDB_FALSE;
				};
			};    
                            
            //set value
            if( lValueLen >= 0 ){
            
                m_pColumn[nCol-1]->setColIsNull(SAPDB_FALSE);
            }
			else{
				strcpy(pValue,REPOFNULLVALUE);
				lValueLen = strlen(REPOFNULLVALUE);
                m_pColumn[nCol-1]->setColIsNull(SAPDB_TRUE);
            }
			
			break;

        case SQL_TIME:
        
            //get the current value            
			pValue = new char [MX_COLUMN_GENERIC_NUMERIC];
			if (pValue == NULL) return SAPDB_FALSE;
			
			nRc = SQLGetData(m_hStmt, nCol, SQL_CHAR, pValue, MX_COLUMN_GENERIC_NUMERIC, &lValueLen);
            if( nRc != SQL_SUCCESS) {
    
				if(! (m_pError->checkSQLReturnCode(nRc,m_hStmt))) {    

					delete[] pValue;
					return SAPDB_FALSE;
				};
			};    
            
            //set value
            if( lValueLen >= 0 ){
            
                //oCurrentCol.ColValue = GetFormattedTime(m_sValueTime)
                m_pColumn[nCol-1]->setColIsNull(SAPDB_FALSE);
            }
			else{
                strcpy(pValue,REPOFNULLVALUE);
				lValueLen = strlen(REPOFNULLVALUE);
                m_pColumn[nCol-1]->setColIsNull(SAPDB_TRUE);
            };

			break;

       case SQL_TIMESTAMP:
            //get the current value
                        
		    pValue = new char [MX_COLUMN_GENERIC_NUMERIC];
			if (pValue == NULL) return SAPDB_FALSE;
			
		    nRc = SQLGetData(m_hStmt, nCol, SQL_CHAR, pValue, MX_COLUMN_GENERIC_NUMERIC, &lValueLen);
            if( nRc != SQL_SUCCESS) {
    
				if(! (m_pError->checkSQLReturnCode(nRc,m_hStmt))) {    

					delete[] pValue;	
					return SAPDB_FALSE;
				};
			};    
            //set value
            if( lValueLen >= 0 ){
            
                //oCurrentCol.ColValue = GetFormattedTimeSamp(m_sValueTimeStamp)
                m_pColumn[nCol-1]->setColIsNull(SAPDB_FALSE);
            }
			else{
                strcpy(pValue,REPOFNULLVALUE);
				lValueLen = strlen(REPOFNULLVALUE);
                m_pColumn[nCol-1]->setColIsNull(SAPDB_TRUE);
            };

			break;
                                   
	    case SQL_LONGVARBINARY:
        case SQL_LONGVARCHAR: 
            
            if( m_lLimitForLong >= 0 ){

                 lGetDataLen = m_lLimitForLong;                 

				 //Limit for long column can overwrite max value for long column
                 if( m_lLimitForLong > MX_COLUMN_LONG_VALUE ) lGetDataLen = MX_COLUMN_LONG_VALUE;   
				 
            }
			else
                lGetDataLen = MX_COLUMN_LONG_VALUE;
            
				pValue = new char [lGetDataLen];
				if (pValue == NULL) return SAPDB_FALSE;			

				m_pError->setNoSQLState(SQLSTATE_DATA_TRUNCATED);
				nRc = SQLGetData(m_hStmt, nCol, SQL_CHAR,pValue, lGetDataLen, &lValueLen);
				if( nRc != SQL_SUCCESS ){            
					
					if(! (m_pError->checkSQLReturnCode(nRc,m_hStmt))) {    

						m_pError->setNoSQLState (SQLSTATE_SUCCESS);

						delete[] pValue;
						return SAPDB_FALSE;
					}
					m_pError->setNoSQLState (SQLSTATE_SUCCESS);                
				 };

				//set value
				if( lValueLen == SQL_NO_TOTAL )            
					m_pColumn[nCol-1]->setColIsNull(SAPDB_FALSE);                				
				else{
					if( lValueLen >= 0 )                    
						m_pColumn[nCol-1]->setColIsNull(SAPDB_FALSE);					
					else{
						strcpy(pValue,REPOFNULLVALUE);
						lValueLen = strlen(REPOFNULLVALUE);
						m_pColumn[nCol-1]->setColIsNull(SAPDB_TRUE);
					};
				};

			break;

        case SQL_DECIMAL:
		case SQL_NUMERIC:
        case SQL_REAL:
		case SQL_FLOAT:
        case SQL_DOUBLE:
		case SQL_INTEGER:
        

            //get the current value			
			pValue = new char [MX_COLUMN_GENERIC_NUMERIC];
			if (pValue == NULL) return SAPDB_FALSE;
			

            nRc = SQLGetData(m_hStmt, nCol, SQL_CHAR, pValue, MX_COLUMN_GENERIC_NUMERIC, &lValueLen);
            if( nRc != SQL_SUCCESS) {
    
				if(! (m_pError->checkSQLReturnCode(nRc,m_hStmt))) {    

					delete[] pValue;	
					return SAPDB_FALSE;
				}
			}
            //set value
            if( lValueLen >= 0 )            
                m_pColumn[nCol-1]->setColIsNull(SAPDB_FALSE);            
			else{
                strcpy(pValue,REPOFNULLVALUE);
				lValueLen = strlen(REPOFNULLVALUE);
                m_pColumn[nCol-1]->setColIsNull(SAPDB_TRUE);
            }

			break;

        case SQL_BIT:

			//get the current value			
			pValue = new char [6];
			if (pValue == NULL) return SAPDB_FALSE;			
            
            //get the current value
            nRc = SQLGetData(m_hStmt, nCol, SQL_BIT, &bBoolValue,sizeof(SAPDB_Bool), &lValueLen);
            if( nRc != SQL_SUCCESS) {
    
				if(! (m_pError->checkSQLReturnCode(nRc,m_hStmt)))  return SAPDB_FALSE;				
			}
            //set value
            if(lValueLen > 0){
                if(bBoolValue){
                
                    strcpy(pValue,"TRUE");
					lValueLen = 4;
                }
				else{
                    strcpy(pValue,"FALSE");
					lValueLen = 5;
                }                
                m_pColumn[nCol-1]->setColIsNull(SAPDB_FALSE);
            }
			else{
                strcpy(pValue,REPOFNULLVALUE);
				lValueLen = strlen(REPOFNULLVALUE);
                m_pColumn[nCol-1]->setColIsNull(SAPDB_TRUE);
            }

			break;

        default:
            //SQL_VARBINARY

            if( m_pColumn[nCol-1]->getODBCColType() == SQL_VARBINARY ||
				m_pColumn[nCol-1]->getODBCColType() == SQL_BINARY ){

                //will be transformed from hex to ascii 
                lGetDataLen = m_pColumn[nCol-1]->getColLength() * 2 + 1;
            }
			else
                lGetDataLen = m_pColumn[nCol-1]->getColLength() + 1;            

			pValue = new char [lGetDataLen];
			if (pValue == NULL) return SAPDB_FALSE;			
            
            //get the current value
            nRc = SQLGetData(m_hStmt, nCol, SQL_CHAR, pValue, lGetDataLen, &lValueLen);
            if( nRc != SQL_SUCCESS) {
    
				if(! (m_pError->checkSQLReturnCode(nRc,m_hStmt))) {    

					delete[] pValue;
					return SAPDB_FALSE;
				}
			}

            //set value
            if( lValueLen > 0 )
                
                m_pColumn[nCol-1]->setColIsNull(SAPDB_FALSE);
            else{
                if( lValueLen != SQL_NULL_DATA ){
            
					delete[] pValue;
                    pValue  = 0;
					lValueLen = 0;
                    m_pColumn[nCol-1]->setColIsNull(SAPDB_FALSE);
                }
				else{
                    strcpy(pValue,REPOFNULLVALUE);
					lValueLen = strlen(REPOFNULLVALUE);
                    m_pColumn[nCol-1]->setColIsNull(SAPDB_TRUE);
                }
            }		
	};//switch

	//if data was truncated then get the possible size lGetDataLen
	if(lValueLen == SQL_NO_TOTAL) lValueLen = lGetDataLen;

	//return value
	if (pValue != NULL && lValueLen > 0){

		if(outValue.ConvertFromASCII_Latin1(
					(SAPDB_DynamicUTF8String::ASCIIConstPointer)pValue,
					(SAPDB_DynamicUTF8String::ASCIIConstPointer)pValue+lValueLen) 
					!= SAPDB_UTF8Basis::Success){

			delete[] pValue;
			return SAPDB_FALSE;
		}
		delete[] pValue;		
	}
        
	return SAPDB_TRUE;
}

const SAPDB_DynamicUTF8String& StudioOAL_Result::getErrorText()
{

	if( m_pError == NULL) return m_strEmty;	
	return m_pError->getErrorText();
}


const SAPDB_DynamicUTF8String& StudioOAL_Result::getSQLState()
{

	if( m_pError == NULL) return m_strEmty;	
	return m_pError->getSQLState();
}

const SAPDB_DynamicUTF8String& StudioOAL_Result::getSQLStateErrorNumber()
{

	if( m_pError == NULL) return m_strEmty;	
	return m_pError->getSQLStateErrorNumber();
}


SAPDB_Long StudioOAL_Result::getNativeError()
{

	if( m_pError == NULL) return 0;	
	return m_pError->getNativeError();
}


const SAPDB_DynamicUTF8String& StudioOAL_Result::getNativeErrorAsString()
{

	if( m_pError == NULL) return m_strEmty;	
	return m_pError->getNativeErrorAsString();
}


SAPDB_Bool StudioOAL_Result::fullExecStmt(const SAPDB_DynamicUTF8String& sCmdStr)
{

	//check it
	if( m_pError == NULL) return SAPDB_FALSE;
    if( m_hDbc == SQL_NULL_HDBC || m_hEnv == SQL_NULL_HENV)  return SAPDB_FALSE;
	

    //Init	
	SQLRETURN		nRc = SQL_SUCCESS;
    m_bDBProcedureResult = SAPDB_FALSE;
    
    m_sStatement = sCmdStr;
    m_lFetchedRows = 0;
    m_bNoRows = SAPDB_TRUE;	

    //if there is a open result the close it
    if(m_hStmt != 0) closeResult();

    //allocate handle for statement    
	nRc = SQLAllocHandle(SQL_HANDLE_STMT,m_hDbc,&m_hStmt);
	if( nRc != SQL_SUCCESS)
	{
		if(! (m_pError->checkSQLReturnCode(nRc,m_hStmt))) {

			SQLFreeStmt(m_hStmt, SQL_DROP);
			m_hStmt = SQL_NULL_HSTMT;
			return SAPDB_FALSE;
		}
	}		

    //open the result set
    if( ! execStmt() ) {

        //free statement
        SQLFreeStmt(m_hStmt, SQL_DROP);
		m_hStmt = SQL_NULL_HSTMT;
        return SAPDB_FALSE;
    };
    

    m_bNoRows = SAPDB_TRUE;
    
    return SAPDB_TRUE;
	
}

void StudioOAL_Result::clearParameters() 
{
    
    StudioOAL_Parameters  *poCurrentParam;    
	poCurrentParam = (StudioOAL_Parameters*)m_ParamList.first();
    while (poCurrentParam != NULL){
		
		delete poCurrentParam;
		poCurrentParam = (StudioOAL_Parameters*)m_ParamList.next();
	} 

	m_ParamList.clear();		

    return;
}

/*!-------------------------------------------------------------------------
		function:     isColNull
--------------------------------------------------------------------------*/
SAPDB_Bool StudioOAL_Result::isColNull(SAPDB_UInt2 nCol,SAPDB_Bool &bIsNull)
{
	if (nCol > m_nColCount || nCol < 1)	return SAPDB_FALSE;	

	bIsNull = m_pColumn[nCol-1]->getColIsNull();

	return SAPDB_TRUE;	
}


SAPDB_Bool StudioOAL_Result::getODBCColType (SAPDB_UInt2 nCol,SQLSMALLINT &nColType)	
{
	if (nCol > m_nColCount || nCol < 1)	return SAPDB_FALSE;	

	nColType = m_pColumn[nCol-1]->getODBCColType();
	return SAPDB_TRUE;	
}