/*
    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end


*/

/*!**********************************************************************

  module: StudioOAL_Result.hpp

  -----------------------------------------------------------------------

  responsible: WolfgangA

  special area: OAL

  description:  Result Class for OAL


************************************************************************/

#ifndef StudioOAL_RESULT_HPP
#define StudioOAL_RESULT_HPP


/*=======================================================================
 *  INCLUDES
 *=====================================================================*/

#include "SAPDBCommon/SAPDB_DynamicUTF8String.hpp"
//Includes different enum definitons
#include "SAPDB/SQLStudio/OAL/StudioOAL_Common.h"


#include "SAPDB/SQLStudio/OAL/StudioOAL_Error.hpp"
#include "SAPDB/SQLStudio/OAL/StudioOAL_Column.hpp"
#include "SAPDB/SAPDBCommon/SAPDB_List.hpp"

#include "SAPDB/SQLStudio/OAL/StudioOAL_Parameters.hpp"


class StudioOAL_Result  
{


public:
	StudioOAL_Result(SQLHANDLE hDbc,SQLHENV hEnv);
	virtual ~StudioOAL_Result();

	enum StudioOAL_ColumnCodeType
	{
		stoalColCodeTypeUnknown = 0,
		stoalColCodeTypeASCII = 1,
		stoalColCodeTypeBYTE = 2,
		stoalColCodeTypeEBCDIC = 3,
		stoalColCodeTypeUNICODE = 4
	};


	enum StudioOAL_ResultCursorType
	{
		stoalCursorForwardOnly = 0,
		stoalCursorStatic = 1,
		stoalOptimizeCursor = 2
	};

	enum StudioOAL_AutoCommit
	{
		AutoCommitOff = 0,
		AutoCommitOn = 1   
	};

private:	

	SQLSMALLINT				m_nColCount;
	SAPDB_Bool				m_bNoRows;
	SAPDB_ULong				m_lFetchedRows;
	SAPDB_Long				m_lRowCount;
	SAPDB_Long				m_lLimitForLong;

	StudioOAL_ResultCursorType	m_enumResultCursorType;

	SQLHANDLE					m_hStmt;
	SQLHENV						m_hEnv ;
	SQLHDBC						m_hDbc; 

	//statement to execute
	SAPDB_DynamicUTF8String		m_sStatement;	
	SAPDB_DynamicUTF8String		m_strEmty;	

	
	StudioOAL_Error			*m_pError;
	StudioOAL_Column		*m_pColumn[1024];

	SAPDB_UCS2				m_strCmdBuffer[4096];	


	//flag which shows that the result is build to represent the values of a DB procedure
	SAPDB_Bool				m_bDBProcedureResult;

	SAPDB_List				m_ParamList;
	

//Functions
private:
	/*!-------------------------------------------------------------------------
		function:     setDbc
		description:  set dbc handle
		arguments:	  [in] hDbc, dbc handle
					  
		return value: /
    --------------------------------------------------------------------------*/
	void setDbc(SQLHANDLE hDbc) {m_hDbc = hDbc;}

	/*!-------------------------------------------------------------------------
		function:     setEnv
		description:  set enviroment handle
		arguments:	  [in] hEnv, enviroment handle
					  
		return value: /
    --------------------------------------------------------------------------*/
	void setEnv(SQLHENV hEnv);

	/*!-------------------------------------------------------------------------
    function:     execStmt
    description:  execute a statement
	arguments:	  [in] hStmt, statement handle
				  [in] strCmd, reerence on utf8 string class, which contains statement to execute
				  
    return value: SAPDB_TRUE, if success
    --------------------------------------------------------------------------*/
	SAPDB_Bool execStmt();

	/*!-------------------------------------------------------------------------
    function:     getColDescriptions
    description:  get column description
	arguments:	  /
				  
				  
    return value: true if success
    --------------------------------------------------------------------------*/
	SAPDB_Bool getColDescriptions();

	/*!-------------------------------------------------------------------------
    function:     rowCount
    description:  get numbers of rows in result 
	arguments:	  [in] hStmt, statement handle
				  
				  
    return value: numbers of rows in result
    --------------------------------------------------------------------------*/
	SAPDB_Long rowCount(SQLHANDLE *hStmt);


	SAPDB_Bool fullParameterExecute(){ return FALSE;}
	SAPDB_Bool getParameterDescription(){return FALSE;}
	SAPDB_Bool prepareForParamaters(){return FALSE;}
	SAPDB_Bool bindParameters(){return FALSE;}


	/*!-------------------------------------------------------------------------
		function:     clearParameters
		description:  clear paramter list
		arguments:	  /
					  
					  
		return value: /
    --------------------------------------------------------------------------*/
	void clearParameters() ;
public:	

	SAPDB_Bool	isDBProcedureResult() { return 	m_bDBProcedureResult;}

	SAPDB_Bool fullExecStmt(const SAPDB_DynamicUTF8String& sCmdStr);

	/*!-------------------------------------------------------------------------
		function:     getRowCount
		description:  numbers of rows in result
		arguments:	  /
					  
					  
		return value: numbers of rows in result
    --------------------------------------------------------------------------*/
	SAPDB_ULong getRowCount()		{return m_lRowCount;}

	/*!-------------------------------------------------------------------------
		function:     getColCount
		description:  numbers of cols in result
		arguments:	  /
					  
					  
		return value: numbers of cols in result
    --------------------------------------------------------------------------*/
	SQLSMALLINT getColCount()		{return m_nColCount;}
	
	/*!-------------------------------------------------------------------------
		function:     setLimitForLong
		description:  limit for data in long column
		arguments:	  lLimitForLong, number of byzes to be retreived for long column
					  
					  
		return value: /
    --------------------------------------------------------------------------*/
	void setLimitForLong(SAPDB_ULong lLimitForLong) {m_lLimitForLong = lLimitForLong;}

	//get statement handle
	const SQLHANDLE getStmt() const {return m_hStmt;}
	
	//which kind of cursor should be used in result object
	void setResultCursorType(StudioOAL_ResultCursorType ResultCursorType){m_enumResultCursorType = ResultCursorType;}

	SAPDB_Bool openResult(const SAPDB_DynamicUTF8String &strCmd);

	void closeResult();	

	SAPDB_Bool fetch();
	SAPDB_Bool extendedFetch(SAPDB_UInt2 lFetchType,SAPDB_ULong lRow,SAPDB_ULong *NoRowsFetched,SAPDB_UInt2 *nRowStatus);

	SAPDB_Bool getColumnDataAsString(SAPDB_UInt2 nCol,SAPDB_DynamicUTF8String &Value);
	SAPDB_Bool getColumnName(SAPDB_UInt2 nCol,SAPDB_DynamicUTF8String &strColName);

	/*!-------------------------------------------------------------------------
		function:     isColNull
		description:  get if col of result row contains null value
		arguments:	  [in] nCol, number of column
					  [out] bIsNull, flag if col of result row contains null value
					  
					  
		return value: true if succesfully
    --------------------------------------------------------------------------*/
	SAPDB_Bool isColNull(SAPDB_UInt2 nCol,SAPDB_Bool &bIsNull);


	/*!-------------------------------------------------------------------------
		function:     getODBCColType
		description:  return the sql (odbc) data type of the column
		arguments:	  [in] nCol, number of column
					  [out] nColType, data type of the column
					  
					  
		return value: SAPDB_TRUE if succesfully
    --------------------------------------------------------------------------*/
	SAPDB_Bool getODBCColType (SAPDB_UInt2 nCol,SQLSMALLINT &nColType);

	//return statement string
	const SAPDB_DynamicUTF8String &getStatement() const {return m_sStatement;}	

	//are there any rows in the result?
	const SAPDB_Bool noRows() const {return m_bNoRows;}

	const SAPDB_DynamicUTF8String& getErrorText();
	const SAPDB_DynamicUTF8String& getSQLState();
	const SAPDB_DynamicUTF8String& getSQLStateErrorNumber();
	const SAPDB_DynamicUTF8String& getNativeErrorAsString();

	SAPDB_Long getNativeError();

	/*!-------------------------------------------------------------------------
		function:     isError
		description:  get if executing result failed
		arguments:	  /
					  
					  
		return value: SAPDB_TRUE if executing result failed
    --------------------------------------------------------------------------*/
	SAPDB_Bool isError(){	return m_pError->isError();}

};

#endif
