/*!===========================================================================

  module:       StudioTD_ColumnDef.cpp

  ----------------------------------------------------------------------------

  responsible:  RalfS,WolfgangA,Hans-GeorgB

  special area: SQL Studio

  description:  Implementation of the column definition class StudioTD_ColumnDef.

  see:          www.sapdb.org

  version:      7.3

  created:      2000-05-11

  updated:      2000-05-11

  ----------------------------------------------------------------------------

  copyright:    Copyrighted by SAP AG, 2000
  


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

============================================================================*/


/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/
#include "SQLStudio/TableDefinition/StudioTD_ColumnDef.h"


/*===========================================================================*
 *     StudioTD_ColumnDef::StudioTD_ColumnDef                                *
 *===========================================================================*/
StudioTD_ColumnDef::StudioTD_ColumnDef ()
    : mKey(false), mNotNull(false), mUnique(false), mAlterType(AlterNoChange)
{
  setType (ClassColumnDef);
}

/*===========================================================================*
 *     StudioTD_ColumnDef::~StudioTD_ColumnDef                               *
 *===========================================================================*/
StudioTD_ColumnDef::~StudioTD_ColumnDef()
{
}


/*===========================================================================*
 *     StudioTD_ColumnDef::findIncomplete                                    *
 *===========================================================================*/
StudioTD_String StudioTD_ColumnDef::findIncomplete ()
{
  /*-------------------------------------------------------------------------*
   * If all field have their initial values the definition is totally empty  *
   * --> it will be ignored.                                                 *
   *-------------------------------------------------------------------------*/
  if (   mName.Empty() && mDataType.Empty() && mDimension.Empty()
      && mCodeType.Empty() && !mKey && !mNotNull && mDefault.Empty()
      && !mUnique && mComment.Empty())
    return "";

  /*-------------------------------------------------------------------------*
   * If name or datatype is empty its index is returned.                     *
   *-------------------------------------------------------------------------*/
  if (mName.Empty())
    return "0";

  if (mDataType.Empty())
    return "1";

  return "";
}


/*===========================================================================*
 *     StudioTD_ColumnDef::genCreateStatement                                *
 *===========================================================================*/
StudioTD_String StudioTD_ColumnDef::genCreateStatement (StudioTD_Mode mode, SAPDB_Bool casesensitive)
{
  StudioTD_String  sql;
  StudioTD_String  attributes;
  StudioTD_String  delimiter;
  
  /*-------------------------------------------------------------------------*
   * If name or datatype is empty the column is not taken into account.      *
   *-------------------------------------------------------------------------*/
  if (mName.Empty() || mDataType.Empty())
    return "";

  /*-------------------------------------------------------------------------*
   * Insert a TAB and the name of the column.                                *
   *-------------------------------------------------------------------------*/
  sql += "\t\"";
  sql += CASE(mName);
  sql += "\"";

  /*-------------------------------------------------------------------------*
   * If a query was used as a template, that's all --> return.               *
   *-------------------------------------------------------------------------*/
  if (mode == CreateFromQuery)
  {
    sql += ",\n";
    return sql;
  }
  
  /*-------------------------------------------------------------------------*
   * Insert the datatype fo the column.                                      *
   *-------------------------------------------------------------------------*/
  sql += "               ";
  sql += mDataType;
  
  /*-------------------------------------------------------------------------*
   * Depending on the datatype the dimension and codetype is inserted.       *
   *-------------------------------------------------------------------------*/
  if ((mDataType == "Char") || (mDataType == "Varchar"))
  {
    if (!mDimension.Empty())
    {
      sql += " (";
      sql += mDimension;
      sql += ")";
    }
    
    if (!mCodeType.Empty())
    {
      sql += " ";
      sql += mCodeType;
    }
  }
  
  if ((mDataType == "Fixed") || (mDataType == "Float"))
  {
    if (!mDimension.Empty())
    {
      sql += " (";
      sql += mDimension;
      sql += ")";
    }
  }
  
  if (mDataType == "Long")
  {
    if (!mCodeType.Empty())
    {
      sql += " ";
      sql += mCodeType;
    }
  }

  /*-------------------------------------------------------------------------*
   * Insert the NOT NULL clause.                                             *
   *-------------------------------------------------------------------------*/
  if (mNotNull)
    attributes += "    NOT NULL";

  /*-------------------------------------------------------------------------*
   * Insert the default value.                                               *
   *-------------------------------------------------------------------------*/
  if ((mDataType == "Char") || (mDataType == "Varchar") || (mDataType == "Long") || 
	  (mDataType == "Date") || (mDataType == "Time")    || (mDataType == "Timestamp"))
  {
    if (   (mDefault != "NULL") && (mDefault != "USER") && (mDefault != "USERGROUP") 
        && (mDefault != "DATE") && (mDefault != "TIME") && (mDefault != "TIMESTAMP")
        && (mDefault != "STAMP"))
      delimiter += "'";
  }
  
  if (!mDefault.Empty())
  {
    attributes += "    DEFAULT ";
	attributes += (mCodeType == "BYTE" ? "x" : "" );
    attributes += delimiter;
    attributes += mDefault;
    attributes += delimiter;
  }
  
  /*-------------------------------------------------------------------------*
   * Insert the unique clause.                                               *
   *-------------------------------------------------------------------------*/
  if (mUnique)
    attributes += "    UNIQUE";
  
  /*-------------------------------------------------------------------------*
   * Append the column attributes to the statement.                          *
   *-------------------------------------------------------------------------*/
  if (!attributes.Empty())
    sql += attributes;
  
  sql += ",\n";
  
  return sql;
}

/*===========================================================================*
 *     StudioTD_ColumnDef::genCreateXML
 *===========================================================================*/
StudioTD_String StudioTD_ColumnDef::genCreateXML( SAPDB_Bool casesensitive )
{

  StudioTD_String  attributes;
  StudioTD_String  delimiter;

  StudioTD_String  deployedcolumn;
  
  /*-------------------------------------------------------------------------*
   * Insert the name of the column.                              
   *-------------------------------------------------------------------------*/

   // SQLDescribeCol : column name, type, column size, decimal digits, and nullability
   // SQLColAttribute: all see sysodbctypes

   // dbm:DeployedColumn 
   deployedcolumn += "<dbm:DeployedColumn oim:id=\"_"; 
   deployedcolumn += string2XML(mTableOwner);
   deployedcolumn += ".";
   deployedcolumn += string2XML(mTableName );
   deployedcolumn += ".";
   deployedcolumn += string2XML(CASE(mName));
   deployedcolumn += "\" oim:label=\"";
   deployedcolumn += string2XML(CASE(mName));
   deployedcolumn += "\" oim:name=\"";
   deployedcolumn += string2XML(CASE(mName));
   deployedcolumn += "\" dbm:comments=\""; 
   deployedcolumn += string2XML(mComment);
   deployedcolumn += "\" dbm:IdentityIncrement=\"";
   deployedcolumn += "";
   deployedcolumn += "\" dbm:initialValue=\"";
   deployedcolumn += mDefault;
   deployedcolumn += "\" dbm:initialValueLanguage=\"";
   deployedcolumn += "";
   deployedcolumn += "\" dbm:IsConstant=\"";
   deployedcolumn += "";
   deployedcolumn += "\" dbm:IsNullable=\"";          // SQL_DESC_NULLABLE(ODBC 3.0)     ,SQLColAttribute.
   if ( mNotNull == SAPDB_TRUE  )
   deployedcolumn += "0";
   if ( mNotNull == SAPDB_FALSE )
   deployedcolumn += "1";
   deployedcolumn += "\" dbm:IsReadOnly=\"";
   deployedcolumn += "";
   deployedcolumn += "\" dbm:Length=\"";              // SQL_DESC_LENGTH (ODBC 3.0)      ,SQLColAttribute.
   deployedcolumn += mDimension;
   deployedcolumn += "\" dbm:MinimumLength=\"";
   deployedcolumn += "";
   deployedcolumn += "\" dbm:name=\"";                // SQL_DESC_NAME(ODBC 3.0)         ,SQLColAttribute.
   deployedcolumn += string2XML(CASE(mName));
   deployedcolumn += "\" dbm:NumericPrecision=\"";    // SQL_DESC_PRECISION(ODBC 3.0)    ,SQLColAttribute.
   deployedcolumn += "";
   deployedcolumn += "\" dbm:NumericScale=\"";        // SQL_DESC_SCALE(ODBC 3.0)        ,SQLColAttribute.
   deployedcolumn += "";
   deployedcolumn += "\" dbm:OctetLength=\"";         // SQL_DESC_OCTET_LENGTH(ODBC 3.0) ,SQLColAttribute
   deployedcolumn += "";
   deployedcolumn += "\" dbm:Ordinal=\"";
   deployedcolumn += "";
   deployedcolumn += "\" dbm:TimePrecision=\"";
   deployedcolumn += "";
   deployedcolumn += "\">\n";

   // sapdbm:<data_typ>

  /*-------------------------------------------------------------------------*
   * Insert the datatype fo the column.                                      *
   *-------------------------------------------------------------------------*/

   deployedcolumn += "<sapdbm:";                         // SQLColAttribute 
   deployedcolumn += mDataType.ToUpper();                // SQL_DESC_TYPE(ODBC 3.0), SQL_DESC_TYPE_NAME(ODBC 1.0)
   deployedcolumn += " oim:href=\"#_";
   deployedcolumn += "sapdbm:";           
   deployedcolumn += mDataType.ToUpper(); 
   deployedcolumn += "\"/>\n";

   deployedcolumn += "</dbm:DeployedColumn>\n";
  
   return deployedcolumn;
}

/*===========================================================================*
 *     StudioTD_ColumnDef::genAlterStatement                                 *
 *===========================================================================*/
StudioTD_String StudioTD_ColumnDef::genAlterStatement (StudioTD_ColumnDef *oldColumn,
													   StudioTD_String owner,
                                                       StudioTD_String table,
													   SAPDB_Bool casesensitive)
{
  StudioTD_String  sql;
  StudioTD_String  attributes;
  StudioTD_String  delimiter;
  StudioTD_String  intro;

  intro  = "ALTER TABLE \"";

  if(owner != ""){
	  intro += owner;
	  intro += "\".\"";
  }

  intro += table;
  intro += "\" ";

  if (mAlterType == AlterChange)
  {
    /*-----------------------------------------------------------------------*
     * Compare the datatypes, dimensions and code types.                     *
     *-----------------------------------------------------------------------*/
    if (   (mDataType != oldColumn->dataType())
        || (mDimension != oldColumn->dimension())
        || (mCodeType != oldColumn->codeType()))
    {
      sql += intro;
      sql += "MODIFY (\"";
      sql += CASE(oldColumn->name());
      sql += "\" ";
      sql += mDataType;
        
      if ((mDataType == "Char") || (mDataType == "Varchar"))
      {
        if (!mDimension.Empty())
        {
          sql += " (";
          sql += mDimension;
          sql += ")";
        }
        
        if (!mCodeType.Empty())
        {
          sql += " ";
          sql += mCodeType;
        }
      }
        
      if ((mDataType == "Fixed") || (mDataType == "Float"))
      {
        if (!mDimension.Empty())
        {
          sql += " (";
          sql += mDimension;
          sql += ")";
        }
      }

      sql += ")";
      sql += SEP;
    }

    /*-----------------------------------------------------------------------*
     * Compare the Not Null values.                                          *
     *-----------------------------------------------------------------------*/
    if (mNotNull != oldColumn->notNull())
    {
      sql += intro;
      sql += "COLUMN \"";
      sql += CASE(oldColumn->name());
      sql += "\" ";
      sql += (mNotNull ? "NOT" : "DEFAULT");
      sql += " NULL";
      sql += SEP;
    }

    /*-----------------------------------------------------------------------*
     * Compare the default values.                                           *
     *-----------------------------------------------------------------------*/
    if (mDefault != oldColumn->defaultValue())
    {
      if ((mDataType == "Char") || (mDataType == "Varchar"))
      {
        if (   (mDefault != "NULL")      && (mDefault != "USER") && (mDefault != "STAMP") 
            && (mDefault != "USERGROUP") && (mDefault != "DATE") && (mDefault != "TIME") 
            && (mDefault != "TIMESTAMP"))
          delimiter = "'";
      }

      sql += intro;
      sql += "COLUMN \"";
      sql += CASE(oldColumn->name());
      sql += "\"";

      if (mDefault == "")
      {
        sql += " DROP DEFAULT";
        sql += SEP;
      }
      
      if (oldColumn->defaultValue() == "")
        sql += " ADD DEFAULT ";
      else
        sql += " ALTER DEFAULT ";

      sql += delimiter;
      sql += mDefault;
      sql += delimiter;
      sql += SEP;
    }

    /*-----------------------------------------------------------------------*
     * Compare the comments.                                                 *
     *-----------------------------------------------------------------------*/
    if (mComment != oldColumn->comment())
    {
      sql += "COMMENT ON COLUMN \"";
      sql += CASE(table);
      sql += "\".\"";
      sql += CASE(oldColumn->name());
      sql += "\" IS '";
      sql += mComment;
      sql += "'";
      sql += SEP;
    }

    /*-----------------------------------------------------------------------*
     * Compare the column names.                                             *
     *-----------------------------------------------------------------------*/
    if (mName != oldColumn->name())
    {
      sql += "RENAME COLUMN \"";
      sql += CASE(table);
      sql += "\".\"";
      sql += CASE(oldColumn->name());
      sql += "\" ";
      sql += "TO \"";
      sql += CASE(mName);
      sql += "\"";
      sql += SEP;
    }
  }
  else  /*  AlterAdd  */
  {
    /*-----------------------------------------------------------------------*
     * If name or datatype is empty the column is not taken into account.    *
     *-----------------------------------------------------------------------*/
    if (mName.Empty() || mDataType.Empty())
      return "";

    /*-----------------------------------------------------------------------*
     * Insert the column name and the datatype of the added column.          *
     *-----------------------------------------------------------------------*/
    sql += intro;
    sql += "ADD (\"";
    sql += CASE(mName);
    sql += "\" ";
    sql += mDataType;

    /*-----------------------------------------------------------------------*
     * Depending on the datatype, insert a dimension and/or a code           *
     * specification into the statement.                                     *
     *-----------------------------------------------------------------------*/
    if ((mDataType == "Char") || (mDataType == "Varchar"))
    {
      if (!mDimension.Empty())
      {
        sql += " (";
        sql += mDimension;
        sql += ")";
      }
      
      if (!mCodeType.Empty())
      {
        sql += " ";
        sql += mCodeType;
      }
    }
      
    if ((mDataType == "Fixed") || (mDataType == "Float"))
    {
      if (!mDimension.Empty())
      {
        sql += " (";
        sql += mDimension;
        sql += ")";
      }
    }

    /*-----------------------------------------------------------------------*
     * Insert the NOT NULL clause.                                           *
     *-----------------------------------------------------------------------*/
    if (mNotNull)
      attributes += " NOT NULL";

    /*-----------------------------------------------------------------------*
     * Insert the default value.                                             *
     *-----------------------------------------------------------------------*/
    if ((mDataType == "Char") || (mDataType == "Varchar"))
    {
      if (   (mDefault != "NULL") && (mDefault != "USER") && (mDefault != "USERGROUP") 
          && (mDefault != "DATE") && (mDefault != "TIME") && (mDefault != "TIMESTAMP")
          && (mDefault != "STAMP"))
        delimiter = "'";
    }
  
    if (!mDefault.Empty())
    {
      attributes += " DEFAULT ";
      attributes += delimiter;
      attributes += mDefault;
      attributes += delimiter;
    }
    
    /*-----------------------------------------------------------------------*
     * Insert the unique clause.                                             *
     *-----------------------------------------------------------------------*/
    if (mUnique)
      attributes += " UNIQUE";
  
    /*-----------------------------------------------------------------------*
     * Append the column attributes to the statement.                        *
     *-----------------------------------------------------------------------*/
    if (!attributes.Empty())
      sql += attributes;
  
    sql += ")";
    sql += SEP;
  
    /*-----------------------------------------------------------------------*
     * Append the comment statement.                                         *
     *-----------------------------------------------------------------------*/
    if (!mComment.Empty())
    {
      sql += "COMMENT ON COLUMN \"";
      sql += CASE(table);
      sql += "\".\"";
      sql += CASE(mName);
      sql += "\" IS '";
      sql += mComment;
      sql += "'";
      sql += SEP;
    }
  }
  
  return sql;
}


/*===========================================================================*
 *     StudioTD_ColumnDef::genDescription                                    *
 *===========================================================================*/
StudioTD_String StudioTD_ColumnDef::genDescription ()
{
  StudioTD_String  desc;

  desc  = mName;
  desc += SEP_LEVEL_3;
  desc += mDataType;
  desc += SEP_LEVEL_3;
  desc += mDimension;
  desc += SEP_LEVEL_3;
  desc += mCodeType;
  desc += SEP_LEVEL_3;
  desc += (mKey     ? "true" : "false");
  desc += SEP_LEVEL_3;
  desc += (mNotNull ? "true" : "false");
  desc += SEP_LEVEL_3;
  desc += mDefault;
  desc += SEP_LEVEL_3;
  desc += (mUnique  ? "true" : "false");
  desc += SEP_LEVEL_3;
  desc += mComment;
  desc += SEP_LEVEL_3;
  
  return desc;
}


/*===========================================================================*
 *     StudioTD_ColumnDef::parseDescription                                  *
 *===========================================================================*/
void StudioTD_ColumnDef::parseDescription (StudioTD_String description)
{
  SAPDB_UInt   pos = 0;
  
  mName      = NextToken (description, SEP_LEVEL_3, &pos);
  mDataType  = NextToken (description, SEP_LEVEL_3, &pos);
  mDimension = NextToken (description, SEP_LEVEL_3, &pos);
  mCodeType  = NextToken (description, SEP_LEVEL_3, &pos);
  mKey       = (NextToken (description, SEP_LEVEL_3, &pos) == "true") ? true : false;
  mNotNull   = (NextToken (description, SEP_LEVEL_3, &pos) == "true") ? true : false;
  mDefault   = NextToken (description, SEP_LEVEL_3, &pos);
  mUnique    = (NextToken (description, SEP_LEVEL_3, &pos) == "true") ? true : false;
  mComment   = NextToken (description, SEP_LEVEL_3, &pos);
}
