/****************************************************************************

  module      : gbd920.h

  -------------------------------------------------------------------------

  responsible : AlexanderK

  special area: container for object file info
  description : the container bases on a chained list hash


  last changed: 2000-01-26  9:33
  see also    : example.html ...

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 1999



    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/


#ifndef GBD920_H
#define GBD920_H

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include <new.h>

#include "gsp00.h"
#include "ggg00.h"
#include "heo55k.h" // vbegexcl
#include "heo57.h"  // vmalloc
#include "hgg08.h"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

/*===========================================================================*
 *  FORWARD CLASS DECLARATION                                                *
 *===========================================================================*/

// K: item key type
// T: type of the item to be contained 
// N: width of the hash
// M: number of item frames the hash allocates if a new Item should be inserted but no emty frame is available

template <class K, class T, int N, int M> class cbd920_ChainedListHash;
template <class K, class T, int N, int M> class cbd920_HashIter;

/*===========================================================================*
 *  CLASS DECLARATION                                                        *
 *===========================================================================*/

template <class K, class T, int N, int M> class cbd920_HashItemFrame 
{	
public:
	T                        m_Item;     
	K                        m_ItemKey;
	cbd920_HashItemFrame<K,T,N,M> *m_pNextItemFrame;
	bool                     m_FrameIsEmpty; 
	cbd920_HashItemFrame() :m_FrameIsEmpty (true), m_pNextItemFrame (NULL) {};
};

/*---------------------------------------------------------------------------*/


template <class K, class T, int N, int M> class cbd920_ChainedListHash
{
	friend class cbd920_HashIter <K,T,N,M>;

public: // class definitions

	/* --- local class CItemFrame  ---- */
	typedef cbd920_HashItemFrame<K,T,N,M> CItemFrame;

private: // class definitions

	/* --- local class CSynchObject --- */
	class CSynchObject
	{
	private:
		tsp00_TaskId m_TaskId;
	public:
		CSynchObject  (tsp00_TaskId TaskId) :m_TaskId(TaskId) {vbegexcl (m_TaskId, g08garbage);};
		~CSynchObject ()                                      {vendexcl (m_TaskId, g08garbage);}; 
	};
    /* --- local class CItemFrameChunk ---*/
	struct CItemFrameChunk
	{
        CItemFrameChunk * m_pNextItemFrameChunk;
		tsp00_Int4        m_ChunkIndex;
		CItemFrame        m_ItemFrame [M];
		CItemFrameChunk(){};
	};
	/* --------------------------------- */
	
public: // methods
	T       *bd920InsItem     (const tsp00_TaskId TaskId, const K &Key, const T &Item, tgg00_BasisError &TrError);  
	void     bd920DelItem     (const tsp00_TaskId TaskId, const K &Key);
	void     bd920DelAllItems (const tsp00_TaskId TaskId);
	void     bd920Verify      (bool  &bHashIsOk) const;	
	T       &operator[]       (const K &Key);
	int      bd920NumItems    ();
	cbd920_ChainedListHash    ();
	
private:// members
	int              m_NumFreeFrames;
	int              m_NumItemFrameChunks;
	CItemFrame       m_NullItemFrame;
	CItemFrame      *m_pFirstItemFrame [N];
	CItemFrame      *m_pFirstFreeItemFrame;
	CItemFrameChunk *m_FirstItemFrameChunk;
	
private:// methods
	int    Hash  (const K & Key) const {return ( Key % N);};
	void   bd920_DelItem (CItemFrame* &pItemFrame);
	void   bd920_AllocateNewFreeItems ();
};

/*---------------------------------------------------------------------------*/

template <class K, class T, int N, int M> class cbd920_HashIter
{
	typedef cbd920_ChainedListHash<K,T,N,M> C;
	typedef cbd920_HashItemFrame  <K,T,N,M> CItemFrame;

private: 
	C              *m_pChainedHashList;
    int             m_CurrHashSlotNo;
    CItemFrame     *m_pCurrItemFrame;

public:
    cbd920_HashIter        (C & ChainedHashList);
	      T * operator  () ();
	      T * operator  ++ ();
	      T * bd920First   (); 
	const K & bd920Key     () const; 
};

/*---------------------------------------------------------------------------*/
/*
class cbd920_HashError
{
public:
	tsp00_Int4       m_ErrNo;
	cbd920_HashError (ErrNo) :m_ErrNo(ErrNo){};
};
*/
/*===========================================================================*
 *  CLASS METHOD IMPLEMENTATION                                              *
 *===========================================================================*/

/*---------------------------------------------------------------------------*/
/* class cbd920_HashIter method implementation                               */
/*---------------------------------------------------------------------------*/

template <class K, class T, int N, int M> inline
cbd920_HashIter <K,T,N,M>::cbd920_HashIter (C & ChainedHashList)
: m_pChainedHashList (&ChainedHashList),
  m_pCurrItemFrame   (NULL)
{};

/*---------------------------------------------------------------------------*/

template <class K, class T, int N, int M> inline T *
cbd920_HashIter <K,T,N,M>::operator () ()
{
	return &(m_pCurrItemFrame->m_Item);
}

/*---------------------------------------------------------------------------*/

template <class K, class T, int N, int M> inline T *
cbd920_HashIter <K,T,N,M>::operator ++ ()
{
	if (NULL == m_pCurrItemFrame)
		return NULL;

	m_pCurrItemFrame = m_pCurrItemFrame->m_pNextItemFrame;
	while (NULL == m_pCurrItemFrame)
	{
		if(N - 1 <= m_CurrHashSlotNo)
			return NULL;
		else
		{
			++ m_CurrHashSlotNo; 
			m_pCurrItemFrame = m_pChainedHashList->m_pFirstItemFrame [m_CurrHashSlotNo];
		}
	}
   return &(m_pCurrItemFrame->m_Item);
}

/*---------------------------------------------------------------------------*/

template <class K, class T, int N, int M> inline T *
cbd920_HashIter <K,T,N,M>::bd920First()
{
	m_CurrHashSlotNo = 0;

	while (NULL == m_pChainedHashList->m_pFirstItemFrame [m_CurrHashSlotNo])
	{
		if(N - 1 <= m_CurrHashSlotNo)
			return NULL;
		else
			++ m_CurrHashSlotNo; 
	}

   m_pCurrItemFrame = m_pChainedHashList->m_pFirstItemFrame [m_CurrHashSlotNo];
   return &(m_pCurrItemFrame->m_Item);
}

/*---------------------------------------------------------------------------*/

template <class K, class T, int N, int M> inline const K &
cbd920_HashIter <K,T,N,M>::bd920Key() const
{
	return m_pCurrItemFrame->m_ItemKey;
}

/*---------------------------------------------------------------------------*/
/* class cbd920_ChainedListHash method implementation                        */
/*---------------------------------------------------------------------------*/

template <class K, class T, int N, int M> inline
cbd920_ChainedListHash <K,T,N,M>::cbd920_ChainedListHash()
:
m_NumFreeFrames       (0),
m_NumItemFrameChunks  (0), 
m_pFirstFreeItemFrame (NULL),
m_FirstItemFrameChunk (NULL)
{
	/* initialize hash table */
	for (int i = 0; i < N; ++i)
		m_pFirstItemFrame[i] = NULL;
}

/*---------------------------------------------------------------------------*/

template <class K, class T, int N, int M> inline T* 
cbd920_ChainedListHash <K,T,N,M>::bd920InsItem (const tsp00_TaskId  TaskId, 
												const K            &Key,
												const T            &Item, 
												tgg00_BasisError   &TrError)
{
	/* determine chain where the items has to be inserted (outside a critical section) */
	CItemFrame * &pFirstItemFrame = (m_pFirstItemFrame [Hash(Key)]);
	
	/* enter critical section */
	CSynchObject  SynchObj(TaskId);
	
	/* check if the key is unique */
	for (CItemFrame* pItemFrame = pFirstItemFrame;
	pItemFrame != NULL;
	pItemFrame  = pItemFrame->m_pNextItemFrame)
	{
		if (Key == pItemFrame->m_ItemKey)
		{
			TrError = e_duplicate_hash_key; //?? in future this should be done by exceptions
			return NULL;
		}
	}
	
	/* check if there is still a free frame. if not try to allocate new free items*/
	if (m_pFirstFreeItemFrame == NULL) 
	{
		/* allocate new free items */
		bd920_AllocateNewFreeItems ();
		
		/* return if allocation failed */
		if (m_pFirstFreeItemFrame == NULL )
		{
			TrError = e_hash_memory_exceeded; //?? in future this should be done by exceptions
			return NULL;
		}
	}
	
	/* get a free item frame from the chain of free frames */
	--m_NumFreeFrames;
	CItemFrame* pNewItemFrame = m_pFirstFreeItemFrame;
	m_pFirstFreeItemFrame = pNewItemFrame->m_pNextItemFrame;
	
	/* insert new item frame at the beginning of the appropriate hash chain */
	pNewItemFrame->m_pNextItemFrame = pFirstItemFrame;
	pFirstItemFrame                 = pNewItemFrame;
	
	/* copy item and key into item frame and mark frame as used */
	pNewItemFrame->m_Item         = Item;
	pNewItemFrame->m_ItemKey      = Key;
	pNewItemFrame->m_FrameIsEmpty = false;
	
	return &(pNewItemFrame->m_Item);	
}

/*---------------------------------------------------------------------------*/

template <class K, class T, int N, int M> inline T &
cbd920_ChainedListHash <K,T,N,M>::operator[] (const K & Key) 
{
	bool      bTryAgain;
	const int HashSlotNo                  = Hash(Key);
	CItemFrame* volatile &pFirstItemFrame = m_pFirstItemFrame [HashSlotNo];
	
	do
	{
		bTryAgain = false;
		for (CItemFrame* pItemFrame = pFirstItemFrame;
	    	pItemFrame != NULL;
	    	pItemFrame  = pItemFrame->m_pNextItemFrame)
		{				 
			if (Key == pItemFrame->m_ItemKey) 
				return pItemFrame->m_Item;
			
			/* check for writer/reader collision */
			if (pItemFrame->m_FrameIsEmpty || HashSlotNo != Hash (pItemFrame->m_ItemKey))
			{
				/* return control to an other task hoping */
				/* that the writer finishs his work       */  
				// ??? Sleep (0);
				
				/* repeat search for key */
				bTryAgain = true;
				break;
			}
		}
	}
	while (bTryAgain);

	/* no item for this key found */
	return m_NullItemFrame.m_Item;
}

/*---------------------------------------------------------------------------*/

template <class K, class T, int N, int M> inline void  
cbd920_ChainedListHash <K,T,N,M>::bd920_DelItem (CItemFrame* &pItemFrame)
{
	CItemFrame* pItemFrameToDelete = pItemFrame;
	
	/* mark item as unused */
	pItemFrameToDelete->m_FrameIsEmpty = true;
	
	/* remove frame from chained hash list */
	pItemFrame = pItemFrameToDelete->m_pNextItemFrame;
	
	/* insert frame into free chain */
	pItemFrameToDelete->m_pNextItemFrame = m_pFirstFreeItemFrame;
	m_pFirstFreeItemFrame                = pItemFrameToDelete;
	++m_NumFreeFrames;
}

/*---------------------------------------------------------------------------*/

template <class K, class T, int N, int M> inline void  
cbd920_ChainedListHash <K,T,N,M>::bd920DelItem (const tsp00_TaskId TaskId, const K & Key)
{
	CItemFrame**   ppItemFrame = &m_pFirstItemFrame [Hash(Key)];

	/* enter critical section */
	// already done outside the hash CSynchObject   SynchObj(TaskId);
	
#   if COMPILEMODE_MEO00 >= SLOW_MEO00 
    g08excl_check (TaskId, g08garbage);
#   endif

	for (;*ppItemFrame != NULL; ppItemFrame = &((*ppItemFrame)->m_pNextItemFrame))
	{
		if (Key == (*ppItemFrame)->m_ItemKey)
		{
			bd920_DelItem (*ppItemFrame);
			return;
		}
	}
}

/*---------------------------------------------------------------------------*/

template <class K, class T, int N, int M> inline void  
cbd920_ChainedListHash <K,T,N,M>::bd920DelAllItems (const tsp00_TaskId TaskId)
{
	/* enter critical section */
	CSynchObject   SynchObj(TaskId);
	
	for (int HashSlot=0; HashSlot<N; ++ HashSlot)
	{
		while (NULL != m_pFirstItemFrame [HashSlot])
		{
			bd920_DelItem (m_pFirstItemFrame [HashSlot]);
		}
	}
}

/*---------------------------------------------------------------------------*/


template <class K, class T, int N, int M> inline int  
cbd920_ChainedListHash <K,T,N,M>::bd920NumItems ()
{
	return (m_NumItemFrameChunks*M - m_NumFreeFrames);
}

/*---------------------------------------------------------------------------*/

template <class K, class T, int N, int M> inline void  
cbd920_ChainedListHash <K,T,N,M>::bd920Verify(bool & bHashIsOk) const
{
	int       i;
	int       NumUsedFrames  = 0;
	int       NumFreeFrames  = 0;	
	int       Num2FreeFrames = 0;

	/* initialize output parameter */
	bHashIsOk = true;
	
	CItemFrame*  pItemFrame;

    /* check the NullItemFrame */
    if (NULL != m_NullItemFrame.m_pNextItemFrame)
      bHashIsOk = false;

	if (!m_NullItemFrame.m_FrameIsEmpty)
		bHashIsOk = false;
	
	/* goto chains of occupied frames */
	for (i=0; i<N; ++i)
	{
		for (pItemFrame=m_pFirstItemFrame [i];
		     NULL != pItemFrame;
		     pItemFrame = pItemFrame->m_pNextItemFrame)
		{
			++NumUsedFrames;
			if (NumUsedFrames > (M*m_NumItemFrameChunks))
				bHashIsOk = false;
			
			/* check if this items belongs really to this chain */
			if (i != Hash(pItemFrame->m_ItemKey)) 
				bHashIsOk = false;
		}
	}
	
	/* goto chain of free frames */
	for (pItemFrame = m_pFirstFreeItemFrame;
	     NULL != pItemFrame;
	     pItemFrame = pItemFrame->m_pNextItemFrame)
	{
		++NumFreeFrames;
		if (NumFreeFrames > (M*m_NumItemFrameChunks))
			bHashIsOk = false;
		
		/* check if item is free,i.e.  key is null */
		if (!pItemFrame->m_FrameIsEmpty)
			bHashIsOk = false;
	};

	
    /* go through all chunks of items and count free Items */
	CItemFrameChunk * pItemFrameChunk = m_FirstItemFrameChunk;
	int NumFoundChunks = 0;
	while (NULL != pItemFrameChunk)
	{
		/* check index of chunk */
		if ((pItemFrameChunk ->m_ChunkIndex) != (m_NumItemFrameChunks - NumFoundChunks -1))
			bHashIsOk = false;

		++NumFoundChunks;
		
		for (i=0;i<M;++i)
		{
			if (pItemFrameChunk->m_ItemFrame[i].m_FrameIsEmpty)
				++Num2FreeFrames;
		}
		pItemFrameChunk = pItemFrameChunk->m_pNextItemFrameChunk;		
	}
	
	/* check counters */	
	if (NumFreeFrames  != m_NumFreeFrames)
		bHashIsOk = false;
	
	if (Num2FreeFrames != m_NumFreeFrames)
		bHashIsOk = false;
	
	if ((NumFreeFrames + NumUsedFrames) != (M*m_NumItemFrameChunks))
		bHashIsOk = false;
}

/*---------------------------------------------------------------------------*/


template <class K, class T, int N, int M> inline void  
cbd920_ChainedListHash <K,T,N,M>::bd920_AllocateNewFreeItems ()
{
	CItemFrameChunk* pItemFrameChunk = NULL;
	tsp00_Bool       bOk;                 
	
	/* get memory block */
    vmalloc (sizeof (CItemFrameChunk), (void **) &pItemFrameChunk, &bOk);
	
	if (bOk)
	{
		/* initialize new memory block */
		new (pItemFrameChunk) CItemFrameChunk;
		
		/* insert chunk into chain of chunks */
		pItemFrameChunk->m_pNextItemFrameChunk = m_FirstItemFrameChunk;
		m_FirstItemFrameChunk                  = pItemFrameChunk;
		pItemFrameChunk->m_ChunkIndex          = m_NumItemFrameChunks;
		++m_NumItemFrameChunks;
		
		/*  concatinate free item frames */
		for (int i = 0; i < M - 1; ++i) 
		{
			pItemFrameChunk->m_ItemFrame[i].m_pNextItemFrame =
				&(pItemFrameChunk->m_ItemFrame[i+1]);
		}
		
		/* insert new free items frames into chain of free items frame */
		pItemFrameChunk->m_ItemFrame[M-1].m_pNextItemFrame = m_pFirstFreeItemFrame;
		m_pFirstFreeItemFrame = &(pItemFrameChunk->m_ItemFrame[0]);
		
		/* update counter of free item frames */
		m_NumFreeFrames += M;
	}
}

/*---------------------------------------------------------------------------*/

#endif  /* GBD920_H */
