/****************************************************************************

  module      : vbd910.cpp

  -------------------------------------------------------------------------

  responsible : AlexanderK

  special area: object node classes
  description : non inline methods of the classes cbd910_ObjDataNode 
                cbd910_ObjNode

  last changed: 2000-03-03  13:12
  see also    : example.html ...

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2000



    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gbd900.h"
#include "gbd910.h"
#include "gbd930.h"

#include "hgg06.h"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/

/*===========================================================================*
 *  EXTERNAL VARIABLES                                                       *
 *===========================================================================*/

extern cbd930_GarbCollController  bd91GarbCollController;

/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/

/*===========================================================================*
 *  LOCAL FUNCTIONS (PROTOTYPES)                                             *
 *===========================================================================*/

static void
bd910_InitObjPageBody (cbd910_ObjNode        &NptrsPage,
					   const tsp00_Int4       AlignedObjFrameLen,
					   const bool             bPageIsOld);

/*---------------------------------------------------------------------------*/

static void
bd910_InitContObjPageBody (cbd910_ObjNode      &NptrsPage,
						   const tsp00_Int4     AlignedObjFrameLen);

/*===========================================================================*
 *  GLOBAL FUNCTIONS (CODE)                                                  *
 *===========================================================================*/

/*---------------------------------------------------------------------------*/
/* METHODS of cbd910_ObjNode                                                 */
/*---------------------------------------------------------------------------*/

void
cbd910_ObjNode::bd910GetAnotherObjPage (tsp00_PageNo                  NewPageNo,
				                        const tbd00_LockRequestSeqNo  NewPageSeqNo,
						                cbd910_ObjNode               &NptrsNewPage)
{
	ROUTINE_DBG_MEO00 ("bd910GetAnotherObjPage");
	
	/* This function determines the pointer to a second page under the */
	/* constraint that the order of page pointer accesses has to       */
	/* follow the inverse order of page sequence numbers. In case the  */
	/* first page has a lower sequence number an optimistic            */
	/* page access is tried first                                      */

    tgg00_BasisError    &TrError = m_ptrCurrent->curr_trans->trError_gg00;

#   if COMPILEMODE_MSP00 >= QUICK_MSP00 
	if (this->np_ptr() == NULL)
		g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, "BD910GetNode: Nptr==NULL", NewPageNo);
#   endif
	
	if (this->np_ptr()->ndLockRequestSeqNo_bd00() > NewPageSeqNo)
		NptrsNewPage.bd910AssignToPage (NewPageNo);
	else
	{
		/* optimistic try */
        NptrsNewPage.bd910AssignToPage (NewPageNo, plmLockNoWait_ebd00); 
		
		if (TrError == e_bd_page_collision)
		{
			TrError                    = e_ok;
			const tsp00_PageNo PageNo  = this->np_ptr()->nd_id();

	        /* increase counter  which forbids the GC to remove this */
			/* page since this task wants to read this page again    */
			++(this->np_ptr()->nd_num_chain_updater());

			this->bd910ReleasePage();
			
			/* get pages in correct order */
			NptrsNewPage.bd910AssignToPage (NewPageNo);
			if (TrError != e_ok) return;
			this->bd910AssignToPage (PageNo);   

			/* reset counter of task which wait for a second page */
			--(this->np_ptr()->nd_num_chain_updater());
		}
	}

	/* check if both pages belong to the same chain since */
	/* one can only hold two pages of the same chain      */
#   if COMPILEMODE_MSP00 >= QUICK_MSP00 
	if (this->np_ptr()->nd_chain_no() != NptrsNewPage.np_ptr()->nd_chain_no())
	{
		g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, 
			"BD910GetNode: Nptr==NULL", NewPageNo);
	}
#   endif
}

/*---------------------------------------------------------------------------*/

void
cbd910_ObjNode::bd910InitObjPage (tsp00_Int4      ObjBodyLen,
								  tsp00_Int4      ObjKeyLen,
								  tsp00_PageNo    SubRoot,
								  tsp00_Int4      ChainNo,
								  tsp00_Int4      PageSeqNo)
{
	ROUTINE_DBG_MEO00 ("bd910InitObjPage");
	
	
	const tgg00_FileId  &FileId  = m_ptrCurrent->curr_tree_id;
    tgg00_BasisError    &TrError = m_ptrCurrent->curr_trans->trError_gg00;
	tbd_nodeptr         &pPage   = this->np_ptr();
	
	tsp00_Int4      ObjFrameLen;
	tsp00_Int4      AlignedObjFrameLen;
	bool            bPageIsOld;
	
	tgg00_PageType  OldPageType    = pPage->nd_pt();          
	tgg00_PageType2 OldPageType2   = pPage->nd_pt2();         
	tgg00_ObjFileNo OldObjFileNo   = pPage->nd_obj_file_no(); 
	
	/* ---------------------- */
	/* initialize page header */
	/* ---------------------- */
	
	// determine object frame length
	switch (FileId.fileObjFileType_gg00())
	{
	case oftVarLenObjFile_egg00:
		
		if  (PRIM_CONT_OBJ_FILE_NO_BD900 != FileId.fileContObjFileNo_gg00())
		{	// continuation object 
			ObjFrameLen = ObjBodyLen + sizeof (tbd900_ContObjHeader);
			pPage->nd_pt2().becomes(pt2ContObject_egg00);
		}
		else
		{	// variable length object 
			ObjFrameLen = ObjBodyLen + sizeof (tgg92_ObjHeader) + sizeof (tbd900_VarObjBodyHeader);
			pPage->nd_pt2().becomes(pt2VarObject_egg00);
		} 
		break;
	case oftFixLenObjFile_egg00:
		// fixed length object 
		ObjFrameLen = ObjBodyLen + sizeof (tgg92_ObjHeader);
		pPage->nd_pt2().becomes(pt2Object_egg00);
		break;
		
	default:
		TrError = e_wrong_obj_file_id;
		break;
	}
	
	if  ((ptData_egg00 == OldPageType) && 
		(OldPageType2 == (tgg00_PageType2_Enum) pPage->nd_pt2()))
		bPageIsOld = (OldObjFileNo == FileId.fileObjFileNo_gg00());
	else
		bPageIsOld = false;
	
	/* write subroot of this page into the header */
	if  (NIL_PAGE_NO_GG00 == SubRoot)
		pPage->nd_sub_root() = pPage->nd_id();
	else
		pPage->nd_sub_root() = SubRoot;	
	
	/* find position of first object on this page */
	if ((pPage->nd_sub_root() == pPage->nd_id()) || (pPage->nd_root() == pPage->nd_id()))
		pPage->nd_first_obj() = BD92ALIGN_OBJ_FRAME (NODE_HEADER_MXBD00 + sizeof (cbd900_RootAndSubRootPageData));
	else
		pPage->nd_first_obj() = NODE_HEADER_MXBD00;
	
	/* determine maximum number of objects which can be stored on this page */
	AlignedObjFrameLen = BD92ALIGN_OBJ_FRAME(ObjFrameLen);
	if  ((BODY_END_BD00 - pPage->nd_first_obj()) >= ObjFrameLen)
	{
		pPage->nd_max_obj_cnt()    = 1 + (BODY_END_BD00 - pPage->nd_first_obj() - ObjFrameLen)/AlignedObjFrameLen;
		pPage->nd_first_free_obj() = pPage->nd_first_obj(); /* PTS 1107145 AK 10/07/2000 */
	}
	else
	{
		pPage->nd_max_obj_cnt() = 0;
		pPage->nd_first_free_obj() = NIL_OBJ_PAGE_POS_GG92; /* PTS 1107145 AK 10/07/2000 */
		/* pPage->nd_first_obj()   = NIL_OBJ_PAGE_POS_GG92;  PTS 1107145 AK 10/07/2000 */

		if (FULLCOVERING_BD00 < ObjFrameLen)
		{
			g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj,
					"BD910InitObjP:TooLongObj", ObjFrameLen);
			TrError = e_too_long_objects;
			return;
		}
	}
	
	pPage->nd_occ_obj_cnt()          = 0;
	pPage->nd_chain_no()             = ChainNo; 
	pPage->ndLockRequestSeqNo_bd00() = PageSeqNo;
	pPage->nd_obj_frame_len()        = tsp00_Uint2 (ObjFrameLen);
	pPage->nd_obj_key_len()          = tsp00_Uint2 (ObjKeyLen);
	pPage->nd_obj_file_no()          = FileId.fileObjFileNo_gg00();
	pPage->nd_next_free()            = NIL_PAGE_NO_GG00;
	pPage->nd_in_free_chain()        = false;
	pPage->nd_empty_page()           = false;
	pPage->nd_space1()               = 0;
    pPage->nd_num_chain_updater()    = 0;
    pPage->nd_space2()               = 0;
    pPage->nd_space3()               = 0;
	pPage->nd_free_eot_obj_cnt()     = 0;
	pPage->nd_free_obj_cnt()         = pPage->nd_max_obj_cnt();
	
	/* ---------------------- */
	/* initialize page bodies */
	/* ---------------------- */
	
	if (pPage->nd_root() == pPage->nd_id())
	{
	    /* get handler for the data accomodated on a */
	    /* root page body and initialize these data  */
		cbd900_RootPageInfo  RootPageInfo (*this);
		RootPageInfo.bd900InitRootPageInfo();
	}

	if (pPage->nd_sub_root() == pPage->nd_id())
	{
	    /* get handler for the data accomodated on a sub root page body and initialize these data */
		cbd900_SubRootPageInfo  SubRootPageInfo (*this);
		SubRootPageInfo.bd900InitSubRootPageInfo();

		/* a subroot is the header of the free chain and therefore  */
		/* automatically member of the free chain                   */
		if (pPage->nd_max_obj_cnt() > 0)
		{
		   pPage->nd_in_free_chain()           = true;
	       pPage->nd_empty_page()              = true;
		   SubRootPageInfo.bd900FreePageCnt()  = 1;
	       SubRootPageInfo.bd900EmptyPageCnt() = 1;
		}
		else
		{
		   pPage->nd_in_free_chain()           = false;
	       pPage->nd_empty_page()              = false;
		   SubRootPageInfo.bd900FreePageCnt()  = 0;
	       SubRootPageInfo.bd900EmptyPageCnt() = 0;
		}
	}

	if (FileId.fileContObjFileNo_gg00() == PRIM_CONT_OBJ_FILE_NO_BD900)
		bd910_InitObjPageBody (*this, AlignedObjFrameLen, bPageIsOld);
	else
		bd910_InitContObjPageBody (*this, AlignedObjFrameLen);
			
	/* ----------------------- */
	/* initialize page trailor */
	/* ----------------------- */
			
	pPage->nd_checksum() = 0;
	pPage->nd_trailer()  = pPage->nd_header();
}
				
/*---------------------------------------------------------------------------*/
/* METHODS of cbd910_ObjDataNode                                             */
/*---------------------------------------------------------------------------*/

void  
cbd910_ObjDataNode::bd910AssignToNewPageInDataChain (cbd910_SubRootNode  &NptrsSubRoot)
{
	ROUTINE_DBG_MEO00 ("bd910AssignToNewPageInDataChain");
	
	/* this function inserts a new and empty object page into  */
	/* the chain of all pages and into the chain of free pages */
	
    tgg00_BasisError    &TrError = m_ptrCurrent->curr_trans->trError_gg00;
	tbd_nodeptr         &pSubRoot = NptrsSubRoot.np_ptr();
	tsp00_Int4           ObjBodySize;
	
	/*  determine object len from root page  */
	switch (NptrsSubRoot.np_ptr()->nd_pt2())
	{
	case pt2Object_egg00: 
		ObjBodySize = pSubRoot->nd_obj_frame_len() - sizeof(tgg92_ObjHeader);
		break;
		
	case pt2ContObject_egg00: 
		ObjBodySize = pSubRoot->nd_obj_frame_len() - sizeof(tbd900_ContObjHeader);
		break;
		
	case pt2VarObject_egg00: 
		ObjBodySize = pSubRoot->nd_obj_frame_len() - sizeof(tgg92_ObjHeader) - sizeof (tbd900_VarObjBodyHeader);;
		break;
		
	default:
		TrError = e_inconsistent_nodetype;
		g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, 
			"BD910Insert:      PageId", pSubRoot->nd_id());
		g01opmsg (sp3p_knldiag, sp3m_error, csp3_bd_msg, csp3_n_obj, 
			"BD910Insert:Wrong PType2", pSubRoot->nd_pt2());
		return;
	}

	/* get and initialize new object page */
	cbd900_SubRootPageInfo SubRootPageInfo (NptrsSubRoot);
	tbd00_LockRequestSeqNo PageSeqNo = ++SubRootPageInfo.bd900MaxPageSeqNo(); 
	   
	this->bd910AssignToNewPage (ObjBodySize, pSubRoot->nd_obj_key_len(), pSubRoot->nd_id(), pSubRoot->nd_chain_no(), PageSeqNo);
	if (TrError != e_ok) return;
	   
	/* update statistic information at sub root page */   
	++SubRootPageInfo.bd900AllPageCnt();
	++SubRootPageInfo.bd900EmptyPageCnt();  

	/* for faster access of monitoring data the           */
	/* number of all pages is copied into the global      */
	/* structures bd90ObjFileDir                          */
	bd910_SetAllPageCntInChainInfo (m_ptrCurrent->curr_tree_id, pSubRoot->nd_chain_no(), SubRootPageInfo.bd900AllPageCnt());

    /* introduce page into chain of all pages, i.e. update nd_left_nd_right pointers */
    if (pSubRoot->nd_right() != NIL_PAGE_NO_GG00) 
    {
        cbd910_ObjDataNode  NptrsPageRight (*m_ptrCurrent, pSubRoot->nd_right(), nr_for_update);
        if (TrError != e_ok) return;
        
        NptrsPageRight.np_ptr()->nd_left() = this->np_ptr()->nd_id();
        this->np_ptr()->nd_right()         = NptrsPageRight.np_ptr()->nd_id();           
    };

    this->np_ptr()->nd_empty_page()    = true;
    this->np_ptr()->nd_left()          = NptrsSubRoot.np_ptr()->nd_id();
    NptrsSubRoot.np_ptr()->nd_right()  = this->np_ptr()->nd_id();

    /* update free chain, i.e put new page at the beginning */ 
	/* of the chain of pages with free object frames        */
	this->bd910InsertPageIntoFreeChain (NptrsSubRoot);
}

/*---------------------------------------------------------------------------*/

bool  
cbd910_ObjDataNode::bd910RemovePageFromDataChain    (cbd910_SubRootNode      &NptrsSubRoot,
					                                 tsp00_PageNo             PrevFreePage,
					                                 tbd00_LockRequestSeqNo   PrevFreePageSeqNo)
{
	ROUTINE_DBG_MEO00 ("bd910RemovePageFromDataChain");
		
	tgg00_BasisError    &TrError    = m_ptrCurrent->curr_trans->trError_gg00;
	cbd910_ObjDataNode   NptrsPrevFreePage (*m_ptrCurrent, nr_for_update);
	cbd910_ObjNode      *pNptrsPrevFreePage; 
	
	/* remove page from free chain */
	if(NptrsSubRoot.np_ptr()->nd_id() == PrevFreePage)
		pNptrsPrevFreePage = &NptrsSubRoot;
	else
	{			
		this->bd910GetAnotherObjPage (PrevFreePage, PrevFreePageSeqNo, NptrsPrevFreePage);
		if (TrError != e_ok) return (false);
		pNptrsPrevFreePage = &NptrsPrevFreePage;
	}
	
	/* check if the page can be removed from the free chain */
	if ( this->np_ptr()->nd_empty_page()              &&
		(this->np_ptr()->nd_num_chain_updater() == 0) &&
		(TrError == e_ok))
	{
#       if COMPILEMODE_MSP00 >= QUICK_MSP00 
		if ((this->np_ptr()->nd_free_obj_cnt() != this->np_ptr()->nd_max_obj_cnt()) ||
			(!this->np_ptr()->nd_in_free_chain()))  
	 		g01abort (csp3_bd_msg, csp3_n_obj, "BD910RP: PageNotEmpty", this->np_ptr()->nd_id());
#       endif
		      
		/* remove page from free chain  */
        this->bd910RemovePageFromFreeChain (*pNptrsPrevFreePage, NptrsSubRoot);

        /* relesaes previous page in free chain */
        NptrsPrevFreePage.bd910ReleasePage();

		/* remove page from chain of all pages if the page is not any more in the free chain*/
		if ( ! this->np_ptr()->nd_in_free_chain())
		{	
			/* get left and right neighbour pages */
			cbd910_ObjDataNode   NptrsLeftPage (*m_ptrCurrent, nr_for_update);
			cbd910_ObjDataNode   NptrsRightPage(*m_ptrCurrent, nr_for_update);
			
			if(this->np_ptr()->nd_left() != NptrsSubRoot.np_ptr()->nd_id())
			{
				this->bd910GetAnotherObjPage (this->np_ptr()->nd_left(), MAX_LOCK_REQUEST_SEQ_NO_BD00, NptrsLeftPage);
				if (TrError != e_ok) return (false);
			}
			
			if (this->np_ptr()->nd_right() != NIL_PAGE_NO_GG00)
			{
				this->bd910GetAnotherObjPage (this->np_ptr()->nd_right(), MIN_LOCK_REQUEST_SEQ_NO_BD00, NptrsRightPage);
				if (TrError != e_ok) return (false);
			}
			
			/* update statistics information in root page */
			cbd900_SubRootPageInfo SubRootPageInfo (NptrsSubRoot);
			--SubRootPageInfo.bd900AllPageCnt();
			--SubRootPageInfo.bd900EmptyPageCnt();
			
			/* for faster access of monitoring data the           */
			/* number of all pages is copied into the global      */
			/* structure  bd90ObjFileDir                          */
			bd910_SetAllPageCntInChainInfo (m_ptrCurrent->curr_tree_id, 
				NptrsSubRoot.np_ptr()->nd_chain_no(), SubRootPageInfo.bd900AllPageCnt());

			/* remove page from chain ,i.e. update neighbor pages */
			if(this->np_ptr()->nd_left() != NptrsSubRoot.np_ptr()->nd_id())	
				NptrsLeftPage.np_ptr()->nd_right() = this->np_ptr()->nd_right();
			else
				NptrsSubRoot.np_ptr()->nd_right() = this->np_ptr()->nd_right();
			
			if (this->np_ptr()->nd_right() != NIL_PAGE_NO_GG00)
				NptrsRightPage.np_ptr()->nd_left() = this->np_ptr()->nd_left();
			
			/* write a vtrace entry */
#           if COMPILEMODE_MEO00 >= SLOW_MEO00 
				t01name (bi,"rem page from file");
				t01int4 (bi,"rem page no=",this->np_ptr()->nd_id());  
#           endif

			/* release page */
            this->bd910FreePage();
			
			return (true);
		}
	}
	return (false);
}

/*---------------------------------------------------------------------------*/

void
cbd910_PrimObjDataNode::bd910InsertPageIntoFreeChainAndUpdateEmptyPageCount (cbd900_ObjFileInfo &ObjFileInfo)
{
	
	tgg00_BasisError    &TrError = m_ptrCurrent->curr_trans->trError_gg00;

	/* get pointer to subroot page */	
	cbd910_SubRootNode    NptrsSubRoot  (*m_ptrCurrent, nr_for_update);
	
	if (this->np_ptr()->nd_sub_root() == this->np_ptr()->nd_id())
		NptrsSubRoot.bd910SetTo (*this);
	else
	{
		this->bd910GetAnotherObjPage (this->np_ptr()->nd_sub_root(), MAX_LOCK_REQUEST_SEQ_NO_BD00, NptrsSubRoot);
		if (e_ok != TrError) 
		{
			if ((e_page_in_wrong_tree == TrError) || (e_no_converter_entry == TrError))
			{
				/* the page was removed while the page was released to wait */
				/* for the subroot therefore nothing needs to be done       */
				TrError = e_ok;
			}
			return;
		}
	}
	
	/* update free chain and free page counter */
	if (this->bd910PageMustBeInsertedIntoFreeChain()) 
		this->bd910InsertPageIntoFreeChain (NptrsSubRoot);
	
	/* update empty page counter on subroot  */
	if (this->bd910EmptyPageCountMustBeIncreased()) 
		bd910IncreaseEmptyPageCount (NptrsSubRoot);
	
	/* check if the object file contains too many empty pages */
	/* and start garbage collection if neccessary             */
	if (NptrsSubRoot.bd910ChainContainsTooManyEmptyPages())
	{
		cbd900_ObjChainInfo  &ChainInfo = ObjFileInfo[NptrsSubRoot.np_ptr()->nd_chain_no()];
		
		/* release page pointers before starting the garbage   */ 
		/* collector what requires entering a critical section */
		NptrsSubRoot.bd910ReleasePage();
		this->bd910ReleasePage(); 
		
		bd91GarbCollController.bd930StartGCToRemoveEmptyPages (*m_ptrCurrent->curr_trans, ChainInfo);
	}
}

/*===========================================================================*
 *  LOCAL FUNCTIONS (CODE)                                                   *
 *===========================================================================*/

static void
bd910_InitContObjPageBody (cbd910_ObjNode        &NptrsPage,
						   const tsp00_Int4       AlignedObjFrameLen)
{
    ROUTINE_DBG_MSP00 ("bd90_InitContObjPageBody");
    
    /* this function initializes the continuation object */
    /* frames at an object page                          */
    
    tsp00_Int4           ContObjFramePos = NptrsPage.np_ptr()->nd_first_obj();
    const tsp00_Int4     MaxObjCnt       = NptrsPage.np_ptr()->nd_max_obj_cnt();
    tbd900_ContObjFrame *pContObjFrame;
    
    for (int iObjCnt=1; iObjCnt<=MaxObjCnt; ++iObjCnt)
    {
        /* set pointer on current objectframe */
        pContObjFrame = (tbd900_ContObjFrame*) ((char*) NptrsPage.np_ptr() + ContObjFramePos);

        /* calculate position of next object */
        ContObjFramePos += AlignedObjFrameLen;
        
        /* initialize next free obj pointers */
        if  (iObjCnt < MaxObjCnt)
            pContObjFrame->cofHeader_bd900.cvoNextFreeObj_bd900 = ContObjFramePos ;
        else
            pContObjFrame->cofHeader_bd900.cvoNextFreeObj_bd900 = NIL_OBJ_PAGE_POS_GG92;
        
        pContObjFrame->cofHeader_bd900.cvoRootOIdPno_bd900  = NIL_PAGE_NO_GG00;
        pContObjFrame->cofHeader_bd900.cvoRootOIdPos_bd900  = NIL_OBJ_PAGE_POS_GG92;      
    }
}

/*---------------------------------------------------------------------------*/

static void
bd910_InitObjPageBody (cbd910_ObjNode        &NptrsPage,
					   const tsp00_Int4       AlignedObjFrameLen,
					   const bool             bPageIsOld)
{
	ROUTINE_DBG_MEO00 ("bd910_InitObjPageBody");
	
    tsp00_Int4         ObjFramePos = NptrsPage.np_ptr()->nd_first_obj();
    const tsp00_Int4   MaxObjCnt   = NptrsPage.np_ptr()->nd_max_obj_cnt();
    tbd900_ObjFrame     *pObjFrame;
    
    for (int iObjCnt=1; iObjCnt<=MaxObjCnt; ++iObjCnt)
    {
        /* set pointer on current objectframe */        
        pObjFrame = cbd910_PrimObjDataNode::bd910ObjFramePtr((tsp00_Byte *)NptrsPage.np_ptr(), ObjFramePos);

        /* calculate position of next object */
        ObjFramePos += AlignedObjFrameLen;
        
        gg06InitObjTransInfo (pObjFrame->objHeader_bd900.ohdTransInfo_gg92);     
        pObjFrame->objHeader_bd900.ohdState_gg92.becomes(obsFree_egg00);
        
        /* set object version */
        if (bPageIsOld)
        {
            if(pObjFrame->objHeader_bd900.ohdFrameVersion_gg92 >= MAX_OBJ_FRAME_VERS_GG92)
                pObjFrame->objHeader_bd900.ohdFrameVersion_gg92 = MIN_OBJ_FRAME_VERS_GG92;
            else
                pObjFrame->objHeader_bd900.ohdFrameVersion_gg92 ++; 
        }
        else 
			pObjFrame->objHeader_bd900.ohdFrameVersion_gg92 = NIL_OBJ_FRAME_VERS_GG92; 
        
        /* initialize next free obj pointers */
        if  (iObjCnt < MaxObjCnt)
            pObjFrame->objHeader_bd900.ohdNextFreeObj_gg92 = ObjFramePos ;
        else
            pObjFrame->objHeader_bd900.ohdNextFreeObj_gg92 = NIL_OBJ_PAGE_POS_GG92;
		
        /* initialize continuation info for variable object pages */
		if (pt2VarObject_egg00 == NptrsPage.np_ptr()->nd_pt2())
		{
			tbd900_VarObjBody & VarObjBody = *((tbd900_VarObjBody*) (pObjFrame->objBody_bd900));
			
			VarObjBody.vbdHeader_bd900.vhdContObjPno_bd900    = NIL_PAGE_NO_GG00;
			VarObjBody.vbdHeader_bd900.vhdContObjPos_bd900    = NIL_OBJ_PAGE_POS_GG92;
            VarObjBody.vbdHeader_bd900.vhdContObjFileNo_bd900 = PRIM_CONT_OBJ_FILE_NO_BD900;
			VarObjBody.vbdHeader_bd900.vhdObjSize_bd900       = 0;
		}
    }
}

/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/