/*!***************************************************************************

  module      : vbd94.cpp
  responsible : UweH
  special area: History Files
  last changed: 2000-08-09  17:06
  copyright   : Copyright by SAP AG, 2000
  see also    :
  description : Implements the History Image Management.



    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gsp00.h"    // All Types
#include "ggg92.h"    // tgg92_KernelOid
#include "ggg00.h"    // Kernel Types

#include "gbd94.h"    // Class HistoryManager
#include "hbd01.h"    // b01pdestroy_perm_file, b01niltree_id
#include "hbd09.h"    // ChainContainer Interface
#include "hbd12.h"    // b12vtrace
#include "hbd91.h"    // register history file at the garbage collector
#include "heo55k.h"   // vbegexcl, vendexcl
#include "heo57.h"    // vallocat
#include "hgg01.h"    // g01maxuser
#include "hgg01_1.h"  // g01allocate, g01abort, g01opmsg
#include "hgg08.h"    // g08history
#include "hgg10.h"    // g10fil
#include "gkb08.h"    // tkb08_* Logentries, tkb08_LogentryParam
#include "hkb85.h"    // Read LogEntries
#include "gsp03.h"    // Opmsg Codes
#include "gsp03_3.h"  // Opmsg Codes
#include "hsp77.h"    // sp77sprintf


#if COMPILEMODE_MEO00 >= SLOW_MEO00 
#include "hta99.h"
#include "hta01_2.h"
#endif

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#ifdef NO_INLINES
#  define _INLINE
#else
#  define _INLINE   inline
#endif

#define NIL_OBJ_FILLCHAR       0
#define REMEMBER_LAST_POSITION TRUE
#define COUNT_PAGES            TRUE
 
/*===========================================================================*
 *  GLOBAL OBJECTS                                                           * 
 *===========================================================================*/

cbd94_HistoryManager bd94History;

/*===========================================================================*
 *  LOCAL FUNCTIONS (PROTOTYPES)                                             *
 *===========================================================================*/

inline void
bd94_WriteTrace (tgg00_TransContext &Trans,
                 const char         *Line,
                 tsp00_Int4          LineLen);

/*---------------------------------------------------------------------------*/

inline void
bd94Trace (tgg00_TransContext &Trans,
           tgg91_PageRef      &HistEntryRef,
           tkb08_HistoryEntry &HistEntry,
           const char         *Label);

/*===========================================================================*
 *  GLOBAL FUNCTIONS (CODE)                                                  * 
 *===========================================================================*/

/*---------------------------------------------------------------------------*/

externPascal
void bd94InitHistoryManagement (tgg00_TransContext &Trans)
{
    /* called by k56restart_log */

	ROUTINE_DBG_MSP00 ("bd94InitHistoryManagement");
    bd94History.bd94Initialize (Trans);
}

/*---------------------------------------------------------------------------*/

externPascal
void bd94GetHistFileInfo (tgg00_TransContext &Trans,
                          tsp00_Int4         &HistFileNo,
                          tgg91_TransNo      &FirstTransId,
                          pasbool            &FirstTransIsUsed,
                          tgg91_TransNo      &LastTransId,
                          pasbool            &LastTransIsUsed,
                          tsp00_Int4         &PageCount,
                          pasbool            &bFileIsEmpty)
{
    bd94History.bd94GetHistFileInfo (Trans,        HistFileNo,
                                     FirstTransId, FirstTransIsUsed,
                                     LastTransId,  LastTransIsUsed,
                                     PageCount,    bFileIsEmpty);
}

/*---------------------------------------------------------------------------*/

externPascal
void bd94SaveHistfileInfo (tgg00_TransContext &Trans)
{
    /* used by savepoint/checkpoint */
    bd94History.bd94SaveHistfileInfo (Trans);
}
 
/*---------------------------------------------------------------------------*/

externPascal
void bd94TraceHistoryEntry (tgg00_TransContext &Trans,
                            char               *Label, /* tsp00_Sname */
                            tgg91_PageRef      &HistEntryRef,
                            tkb08_HistoryEntry &HistEntry)
{
    bd94Trace (Trans, HistEntryRef, HistEntry, Label);
}

/*---------------------------------------------------------------------------*/

externPascal
void bd94TraceTransId (tgg00_TransContext  &Trans,
                       char                *Label, /* tsp00_Sname */
                       tsp00_Int4           HistFileNo,
                       tsp00_PageNo         PageNo,
                       tgg91_TransNo       &TransId)
{
    tsp00_Int4  LineLen;
    tsp00_Line  Line;
    const char *AuxLabel = (NULL == Label) ? "HIS:TransId " : Label;

    sp77sprintf (Line, sizeof(Line), "%12s: HFILNO=%d PNO=%d TRANSID= ", AuxLabel, HistFileNo, PageNo);
    LineLen = (tsp00_Int4)strlen(Line);
    TransId.gg90AppendToLine (LineLen, Line);
    bd94_WriteTrace (Trans, Line, LineLen);
}

/*---------------------------------------------------------------------------*/

externPascal void
bd94Dump (tgg00_VfFileref &HostFile,
		  tsp00_Int4       BufferSize,
		  tsp00_Page      &Buffer,
		  tsp00_Int4      &OutPno,
		  tsp00_Int4      &OutPos,
		  tsp00_VfReturn  &HostError,
		  tsp00_ErrText   &ErrText)
{
    bd94History.bd94Dump (HostFile,BufferSize,Buffer,OutPno,OutPos,HostError,ErrText);
}

/*===========================================================================*
 *  CLASS METHODS IMPLEMENTATION                                             * 
 *===========================================================================*/

/*---------------------------------------------------------------------------*/

inline cbd94_HistoryFile&
cbd94_HistoryManager::bd94_GetHistoryFileOfTaskId (tsp00_TaskId TaskId)
{
    const tsp00_Int4 HistFileNo = TaskId % m_MaxHistoryFiles;
#   if COMPILEMODE_MEO00 >= SLOW_MEO00
    t01p2int4 (bd, "TaskId =>   ", TaskId, "HistFileNo  ", HistFileNo);
#   endif
    return bd94_GetHistoryFileOfHistFileNo (HistFileNo);
}

/*---------------------------------------------------------------------------*/

void
cbd94_HistoryManager::bd94Initialize (tgg00_TransContext &Trans)
{
    /* PTS 1107331 UH 2000-10-05 rewritten */

    ROUTINE_DBG_MSP00 ("bd94Initialize");

    /* within LOG-region: called during RESTART (k56restart_log) */

    tsp00_Int4 HistFileNo;

    if ( (m_MaxHistoryFiles == 0) && (m_pHistoryFiles == NULL) )
    {

        m_MaxHistoryFiles = g01maxuser();

        /* PTS 1106791 UH 2000-06-05 */

        g01opmsg (sp3p_knldiag, sp3m_info, csp3_bd_msg, csp3_n_history,
            "History Files           ", m_MaxHistoryFiles);

        tsp00_Bool bOkay          = FALSE;
        tsp00_Int4 AllocationSize = sizeof(cbd94_HistoryFile) * m_MaxHistoryFiles;

        /* allocate HistoryFileInfos */

        vallocat (AllocationSize, (void**)(&m_pHistoryFiles), &bOkay);
        if ( TRUE != bOkay )
        {
            Trans.trError_gg00 = e_allocate_failed;
            return;
        }
        g01allocate_msg (csp3_n_dynpool, "DYNP_B94_HISTORY:       ", AllocationSize);

        for (HistFileNo = 0; HistFileNo < m_MaxHistoryFiles; ++HistFileNo)
        {
            m_pHistoryFiles[HistFileNo].bd94FileId() = b01niltree_id;
        }
    }

    /* initialize alle HistoryFileInfos */
    bool bRegisterFileToGarbageCollector = false;

    for (HistFileNo = 0; HistFileNo < m_MaxHistoryFiles; ++HistFileNo)
    {
        cbd94_HistoryFile &HistFile = bd94_GetHistoryFileOfHistFileNo (HistFileNo);

        HistFile.bd94CheckTaskId()   = 0;
        HistFile.bd94RightmostPage() = NIL_PAGE_NO_GG00; /* must always be determined */

        tgg00_FileId &FileId = HistFile.bd94FileId();

        if ( NIL_PAGE_NO_GG00 == FileId.fileRoot_gg00() )
        {
            FileId.fileType_gg00().clear();
            FileId.fileType_gg00().addElement (ftsPerm_egg00);
            FileId.fileType_gg00().addElement (ftsConcurrent_egg00);
            FileId.fileTfn_gg00().becomes     (tfnObjBefore_egg00);
            FileId.fileTransId_gg00().becomes (HistFileNo);
            bRegisterFileToGarbageCollector = true;
        }
        b01lg_create_log_file (Trans, FileId);

        if ( e_ok == Trans.trError_gg00 )
        {
            /* never before a history file was created */
            HistFile.bd94RightmostPage() = FileId.fileRoot_gg00();
        }
        else if ( e_duplicate_filename == Trans.trError_gg00 )
        {
            /* a history file exists, so determine rightmost, which was saved during checkpoint */
            Trans.trError_gg00 = e_ok;
            b09get_rightmost_id (Trans, FileId, HistFile.bd94RightmostPage());
            if ( e_ok != Trans.trError_gg00 ) return;
        }

        if ( bRegisterFileToGarbageCollector )
        {
            bd91RegisterHistoryFile (Trans, &HistFile);
            if ( e_ok != Trans.trError_gg00 ) return;
        }
    }
}


/*---------------------------------------------------------------------------*/

void cbd94_HistoryManager
::bd94GetOldestTransIdToBeRemoved (tgg00_TransContext      &Trans,
                                   tbd94_HistoryFileHandle  HistFileHandle,
                                   tgg91_TransNo           &OldestTransId)
{
	ROUTINE_DBG_MSP00 ("bd94GetOldestTransIdToBeRemoved");
    tsp00_Int4    DummyCount;
    tsp00_Bool    OldestTransIsUsed = FALSE;
    tsp00_Bool    LastTransIsUsed   = FALSE;
    tsp00_Bool    bFileIsEmpty      = FALSE;
    tgg91_TransNo LastTransId;

    b09histfile_statistic (Trans, HistFileHandle->bd94FileId(),
                           HistFileHandle->bd94RightmostPage(),
                           OldestTransId, OldestTransIsUsed,
                           LastTransId,  LastTransIsUsed,
                           bFileIsEmpty, ! COUNT_PAGES, DummyCount);

    if ( e_ok != Trans.trError_gg00 ) return;

    if ( (OldestTransIsUsed == TRUE) || (bFileIsEmpty == TRUE) )
    {
        OldestTransId.gg90SetNil();
    }
}

/*---------------------------------------------------------------------------*/

void cbd94_HistoryManager
::bd94GetHistFileInfo (tgg00_TransContext &Trans,
                       tsp00_Int4         &HistFileNo,
                       tgg91_TransNo      &FirstTransId,
                       pasbool            &FirstTransIsUsed,
                       tgg91_TransNo      &LastTransId,
                       pasbool            &LastTransIsUsed,
                       tsp00_Int4         &PageCount,
                       pasbool            &bFileIsEmpty)
{
	ROUTINE_DBG_MSP00 ("bd94GetHistFileInfo");
	if ( (HistFileNo < 0) || (HistFileNo >= m_MaxHistoryFiles) )
    {
        HistFileNo = -1;
        PageCount  =  0;
        FirstTransId.gg90SetNil();
        LastTransId.gg90SetNil();
        return;
    }
    cbd94_HistoryFile &HistFile = bd94_GetHistoryFileOfHistFileNo (HistFileNo);

    b09histfile_statistic (Trans, HistFile.bd94FileId(), HistFile.bd94RightmostPage(),
                           FirstTransId, FirstTransIsUsed,
                           LastTransId,  LastTransIsUsed,
                           bFileIsEmpty, COUNT_PAGES,  PageCount);

    ++HistFileNo;
}

/*---------------------------------------------------------------------------*/

void cbd94_HistoryManager
::bd94RemoveOldestTransId (tgg00_TransContext      &Trans,
                           tbd94_HistoryFileHandle  HistFileHandle)
{
	ROUTINE_DBG_MSP00 ("bd94RemoveOldestTransId");
    tsp00_Bool AnythingDeleted;
    b09remove_oldest_history (Trans, HistFileHandle->bd94FileId(), HistFileHandle->bd94RightmostPage(), AnythingDeleted);
    if (AnythingDeleted != TRUE)
    {
        g01opmsg (sp3p_syscon, sp3m_error, csp3_bd_msg, csp3_n_history,
            "bd94RemoveOldestTransId ", HistFileHandle->bd94FileId().fileRoot_gg00());
    }
}

/*---------------------------------------------------------------------------*/

void cbd94_HistoryManager
::bd94SaveHistfileInfo (tgg00_TransContext &Trans)
{
	ROUTINE_DBG_MSP00 ("bd94SaveHistfileInfo");
    for (tsp00_Int4 HistFileNo = 0; HistFileNo < m_MaxHistoryFiles; ++HistFileNo)
    {
        cbd94_HistoryFile &HistFile = bd94_GetHistoryFileOfHistFileNo (HistFileNo);
        b09save_rightmost_id (Trans, HistFile.bd94FileId(), HistFile.bd94RightmostPage());
    }
}

/*---------------------------------------------------------------------------*/

bool cbd94_HistoryManager
::bd94BeforeImageExists (tgg00_TransContext &Trans,
                         tgg92_KernelOid     TestOid,
                         tgg91_PageRef       WantedPageRef)
{
	ROUTINE_DBG_MSP00 ("bd94BeforeImageExists");
    tgg00_MessType     Action;
    tgg91_PageRef      DummyNextPageRef;
    tgg91_TransNo      DummyConsistentView;
    tgg91_TransNo      DummyPrevUpdTrans;
    tgg00_ObjFrameVers DummyFrameVers;
    tsp00_Int4         DummyBodyLen;

    Action.becomes (m_nil);

    bd94GetHistory (Trans, TestOid, false, WantedPageRef,
                    DummyNextPageRef,
                    Action,
                    DummyConsistentView,
                    DummyPrevUpdTrans,
                    DummyFrameVers,
                    DummyBodyLen);

    /* PTS 1107965 UH 2000-10-04 begin */
    if ( e_ok != Trans.trError_gg00 ) 
    {
        if ( e_illegal_page_no != Trans.trError_gg00 )
            Trans.trError_gg00 = e_ok;
        return false;
    }
    else
        return m_drop != Action;
    /* PTS 1107965 UH 2000-10-04 end */
}


/*---------------------------------------------------------------------------*/

void cbd94_HistoryManager
::bd94RemoveHistory (tgg00_TransContext       &Trans,
                     tbd94_HistoryFileHandle  &HistFileHandle,
                     bool                      bDoExactUsageCheck,
                     bool                     &bAnythingDeleted)
{
	ROUTINE_DBG_MSP00 ("bd94RemoveHistory");

    if ( m_MaxHistoryFiles > 0 )
    {
        bAnythingDeleted = TRUE == b09remove_history (Trans,
                                                      HistFileHandle->bd94FileId(),
                                                      HistFileHandle->bd94RightmostPage(),
                                                      bDoExactUsageCheck);
    }
}

/*---------------------------------------------------------------------------*/

void cbd94_HistoryManager
::bd94InsertImage (tgg00_TransContext &Trans,
                   tgg92_KernelOid     TestOid,
                   tgg00_ObjFileNo     ObjFileNo,
                   tgg00_MessType      MessType,
                   tgg91_TransNo       ConsistentView,
                   tgg91_TransNo       PrevUpdTrans,
                   tgg91_PageRef       BeforePageRef,
                   tgg91_PageRef      &NextPageRef,
                   tsp00_Int4          BodySize,
                   void               *pBody)
{
	ROUTINE_DBG_MSP00 ("bd94InsertImage");

    cbd94_HistoryFile &HistFile = bd94_GetHistoryFileOfTaskId (Trans.trTaskId_gg00);

#   if COMPILEMODE_MEO00 >= QUICK_MEO00
    if ( 0 == HistFile.bd94CheckTaskId() )
    {
        HistFile.bd94CheckTaskId() = Trans.trTaskId_gg00;
    }
    else if ( HistFile.bd94CheckTaskId() != Trans.trTaskId_gg00 )
    {
        Trans.trError_gg00 = e_not_implemented;
        g01abort (csp3_bd_msg, csp3_n_history, "TaskMismatch            ", HistFile.bd94CheckTaskId());
    }
#   endif

    tkb08_HistoryEntry         HistEntry;
    tkb08_LogEntryShortHeader &HistHeader = HistEntry.leHead_kb08;

    HistHeader.leLen_kb08 = sizeof(tkb08_HistoryEntry) + BodySize;
    HistHeader.leLen_kb08 = ((HistHeader.leLen_kb08 + ALIGNMENT_KB08 - 1) / ALIGNMENT_KB08) * ALIGNMENT_KB08;
    HistHeader.leType_kb08.becomes  (MessType);
    HistHeader.leType2_kb08.becomes (mm_outcopy);
    HistHeader.leSequence_kb08.gg90SetNil();

    HistEntry.leOid_kb08          = TestOid;
    HistEntry.leUpdTrans_kb08     = Trans.trTransId_gg00;
    HistEntry.leConsistView_kb08  = ConsistentView;
    HistEntry.lePrevUpdTrans_kb08 = PrevUpdTrans;
    HistEntry.leBeforeRef_kb08    = BeforePageRef;
    HistEntry.leBodyLen_kb08      = BodySize;    // tsp00_Int4 := tsp00_Uint2 !!
    HistEntry.leObjFileNo_kb08    = ObjFileNo;

    NextPageRef.gg91SetNilRef ();

    /* e_wait_for_savepoint check is not needed, because there are no savepoints */
    /* only checkpoints with no open transactions are done                       */

    tgg00_LogPos NewEntryPos;

    b09append_logentry (Trans,
                        HistFile.bd94FileId(),
                        mm_first,                          /* first of more parts      */
                        sizeof (HistEntry), &HistEntry, 1, /* MoveObject: Len,Ptr,Pos  */
                        sizeof (HistEntry),                /* MoveLength               */
                        HistHeader.leLen_kb08,             /* complete RestLength       */
                        ! REMEMBER_LAST_POSITION,
                        NewEntryPos,                       /* out: Position written to */               
                        HistFile.bd94RightmostPage());     /* out: new RightMostPage   */

    if ( Trans.trError_gg00 != e_ok )
    {
        g01abort (csp3_bd_msg, csp3_n_history, "AppendHead:             ", HistFile.bd94RightmostPage());
    }

    if ( BodySize > 0 )
    {
        b09append_logentry (Trans,
                            HistFile.bd94FileId(),
                            mm_last,                                    /* last of several parts     */
                            BodySize, pBody, 1,                         /* MoveObject: Len,Ptr,Pos   */
                            BodySize,                                   /* MoveLength w/o alignment  */
                            HistHeader.leLen_kb08 - sizeof (HistEntry), /* complete RestLength       */
                            ! REMEMBER_LAST_POSITION,
                            NewEntryPos,                                /* NOT out: Position written */
                            HistFile.bd94RightmostPage());              /* out: new RightMostPage    */

        if ( Trans.trError_gg00 != e_ok )
        {
            g01abort (csp3_bd_msg, csp3_n_history, "AppendBody:             ", HistFile.bd94RightmostPage());
        }
    }

    NextPageRef.gg91BuildRef ( NewEntryPos.logLpno_gg00(), NewEntryPos.logPpos_gg00() );
    if ( g01vtrace.vtrKbLock_gg00 )
    {
        bd94Trace (Trans, NextPageRef, HistEntry, "bd94Ins");
    }
}

/*---------------------------------------------------------------------------*/

void cbd94_HistoryManager
::bd94GetHistory (tgg00_TransContext &Trans,
                  tgg92_KernelOid     TestOid,
                  bool                ChangeDeleteEntry,
                  tgg91_PageRef       WantedPageRef,
                  tgg91_PageRef      &NextPageRef,
                  tgg00_MessType     &Action,
                  tgg91_TransNo      &ConsistentView,
                  tgg91_TransNo      &PrevUpdTrans,
                  tgg00_ObjFrameVers &FrameVers,
                  tsp00_Int4         &BodyLen,
                  tsp00_Int4          BodySize,
                  void               *pBody)
{
	ROUTINE_DBG_MSP00 ("bd94GetHistory");

    cbd94_HistoryFile &HistFile = bd94_GetHistoryFileOfTaskId (Trans.trTaskId_gg00);
    tgg00_LogPos       CurrEntryPos;
    const bool         bWithBody = (BodySize > 0);

    if ( g01vtrace.vtrKbLock_gg00 )
    {
        const tsp00_Int4 LineSize = sizeof(tsp00_Line)+1;
        char             Line [LineSize];

        sp77sprintf (Line, LineSize, "HIS:Wanted REF=%d.%d OID=%d.%d:%d CHANGE=%d",
            WantedPageRef.gg91RefPno(), WantedPageRef.gg91RefPos(),
            TestOid.gg92GetPno(), TestOid.gg92GetPos(), TestOid.gg92GetFrameVers(),
            ChangeDeleteEntry);
        bd94_WriteTrace (Trans, Line, (tsp00_Int4)strlen(Line));
    }

    if ( NIL_PAGE_NO_GG00 == WantedPageRef.gg91RefPno() )
    {
        g01opmsg (sp3p_syscon, sp3m_error,
            csp3_bd_msg, csp3_n_history,
            "bd94Get: WntdPageRef=NIL", WantedPageRef.gg91RefPno());
        Trans.trError_gg00 = e_obj_history_not_found;
        return;
    }

    /* build parameters for b09 Interface */

    CurrEntryPos.logPageType_gg00().becomes (ptData_egg00);
    CurrEntryPos.logLpno_gg00() = WantedPageRef.gg91RefPno();
    CurrEntryPos.logPpos_gg00() = WantedPageRef.gg91RefPos();

    tkb08_LogentryParam Log;

    kb85ParamLogEntryInit (Log);

    Log.lgeForUpdate_kb08 = ChangeDeleteEntry;
    Log.lgeFile_kb08      = b01niltree_id;
    Log.lgeFile_kb08.fileTfn_gg00().becomes (tfnObjBefore_egg00);

    kb85ReadLogEntry (Trans, CurrEntryPos, Log);

    if ( e_ok == Trans.trError_gg00 )
    {
        tkb08_HistoryEntry &HistEntry = *(REINTERPRET_CAST(tkb08_HistoryEntry*, Log.lgeEntry_kb08));
        const bool          bWithHistBody = ((m_update == HistEntry.leHead_kb08.leType_kb08) ||
                                             (m_delete == HistEntry.leHead_kb08.leType_kb08)  );

        if ( g01vtrace.vtrKbLock_gg00 )
        {
            bd94Trace (Trans, WantedPageRef, HistEntry, "bd94Get");
        }

        /* It may happen that the history must have been truncated, due to lack of space. */
        /* Then a reference is invalid and not an error.                                  */

        if ((ptData_egg00 != Log.lgePage_kb08->pageType_kb00 ()) ||
            (pt2Log_egg00 != Log.lgePage_kb08->pageType2_kb00()))
        {
            Trans.trError_gg00 = e_page_in_wrong_tree;
        }
        else if ( mm_outcopy != HistEntry.leHead_kb08.leType2_kb08 )
        {
            Trans.trError_gg00 = e_log_error;
        }
        else if ( ! HistEntry.leOid_kb08.gg92IsEqIgnoreFrameVers(TestOid) )
        {
            Trans.trError_gg00 = e_invalid_oid;
        }
        else
        {
            Action         = HistEntry.leHead_kb08.leType_kb08;
            BodyLen        = HistEntry.leBodyLen_kb08;
            ConsistentView = HistEntry.leConsistView_kb08;
            PrevUpdTrans   = HistEntry.lePrevUpdTrans_kb08;
            FrameVers      = HistEntry.leOid_kb08.gg92GetFrameVers();
            NextPageRef    = HistEntry.leBeforeRef_kb08;

            if ( ChangeDeleteEntry && (Action == m_delete) )
            {
                /* this avoids that the GarbageCollector wants to remove the rollbacked object */
                HistEntry.leHead_kb08.leType_kb08.becomes (m_drop);
            }

            if ( NULL != pBody )
            {
                if ( (m_delete == Action) || (m_update == Action) )
                {
                    kb85MoveFromLogPage (Trans, Log,
                        sizeof(tkb08_HistoryEntry)+1,
                        BodySize, pBody, 1, BodyLen);
                }
                else
                {
                    if ( (m_new == Action) || (m_insert == Action) )
                    { 
                        g10fil ( "VBD94." , 1 , 
                            BodySize, pBody, 1, BodyLen,
                            NIL_OBJ_FILLCHAR, Trans.trError_gg00);
                    }
                }
            }
        }
    }
    if ( Trans.trError_gg00 != e_ok )
    {
        g01opmsg (sp3p_nil, sp3m_info, csp3_bd_msg, csp3_n_history, "bd94Get: Error          ", Trans.trError_gg00);
        Trans.trError_gg00 = e_obj_history_not_found;
    }
    const tgg00_BasisError AuxError = Trans.trError_gg00;
    kb85LogPageRelease (Trans, Log);
    Trans.trError_gg00 = AuxError;
}
						  
/*---------------------------------------------------------------------------*/

void cbd94_HistoryManager
::bd94GetObjectBody (tgg00_TransContext &Trans,
                     tgg92_KernelOid     TestOid,
                     tgg00_MessType      RollbackAction,
                     tgg91_PageRef       WantedPageRef,
                     tsp00_Int4         &BodyLen,
                     tsp00_Int4          BodySize,
                     void               *pBody)
{
    tgg00_MessType     Action;
    tgg91_PageRef      DummyNextPageRef;
    tgg91_TransNo      DummyConsistentView;
    tgg91_TransNo      DummyPrevUpdTrans;
    tgg00_ObjFrameVers DummyFrameVers;

    Action.becomes (m_nil);

    bd94GetHistory (Trans, TestOid, RollbackAction == m_delete, WantedPageRef,
        DummyNextPageRef, Action, DummyConsistentView, DummyPrevUpdTrans,
        DummyFrameVers, BodyLen, BodySize, pBody);

    if ( tgg00_MessType_Enum (RollbackAction) != tgg00_MessType_Enum (Action))
    { 
        Trans.trError_gg00 = e_log_error;
        g01opmsg (sp3p_syscon, sp3m_error, csp3_bd_msg, csp3_n_history,
            "bd94GetBody: WrongAction", Action);
    }

    /* PTS 1110006 UH 2001-04-18 begin */
    if ( Trans.trError_gg00 != e_ok )
    {
        g01opmsg (sp3p_syscon, sp3m_error, csp3_bd_msg, csp3_n_history,
            "bd94GetBody:TrError     ", Trans.trError_gg00);

        // switch on more trace
        g01vtrace.vtrKbLock_gg00 = true;

        // try again
        Trans.trError_gg00 = e_ok;

        Action.becomes (m_nil);

        bd94GetHistory (Trans, TestOid, RollbackAction == m_delete, WantedPageRef,
            DummyNextPageRef, Action, DummyConsistentView, DummyPrevUpdTrans,
            DummyFrameVers, BodyLen, BodySize, pBody);

        if ( tgg00_MessType_Enum (RollbackAction) != tgg00_MessType_Enum (Action))
        { 
            Trans.trError_gg00 = e_log_error;
            g01opmsg (sp3p_syscon, sp3m_error, csp3_bd_msg, csp3_n_history,
                "bd94GetBody:WrongAction2", Action);
        }
        g01abort (csp3_bd_msg, csp3_n_history, "Rollb: no before image  ", Trans.trError_gg00);

    }
    /* PTS 1110006 UH 2001-04-18 end */
}

/*---------------------------------------------------------------------------*/

inline void
bd94Trace (tgg00_TransContext &Trans,
           tgg91_PageRef      &HistEntryRef,
           tkb08_HistoryEntry &HistEntry,
           const char         *Label)
{
    const tsp00_Int4  LineSize = sizeof(tsp00_Line)+1;
    char              Line [LineSize];
    const char       *AuxLabel = (NULL == Label) ? "HIS:ENTRY" : Label;

    sp77sprintf (Line, LineSize,
	"%s: (%d.%d) OID=%d.%d:%d ACT=%d UPD-T=%d CONS-V=%d PUPD-T=%d BREF=%d.%d",
	AuxLabel,
        HistEntryRef.gg91RefPno(),
	HistEntryRef.gg91RefPos(),
        HistEntry.leOid_kb08.gg92GetPno(),
	HistEntry.leOid_kb08.gg92GetPos(),
	HistEntry.leOid_kb08.gg92GetFrameVers(),
        (int)(HistEntry.leHead_kb08.leType_kb08),
        HistEntry.leUpdTrans_kb08.gg90GetInt4(),
        HistEntry.leConsistView_kb08.gg90GetInt4(),
        HistEntry.lePrevUpdTrans_kb08.gg90GetInt4(),
        HistEntry.leBeforeRef_kb08.gg91RefPno(),
	HistEntry.leBeforeRef_kb08.gg91RefPos() );

    bd94_WriteTrace (Trans, Line, (tsp00_Int4)strlen(Line));
}

/*---------------------------------------------------------------------------*/

void
cbd94_HistoryManager::bd94Dump (tgg00_VfFileref &HostFile,
								tsp00_Int4       BufferSize,
								tsp00_Page      &Buffer,
								tsp00_Int4      &OutPno,
								tsp00_Int4      &OutPos,
								tsp00_VfReturn  &HostError,
								tsp00_ErrText   &ErrText)
{
	ROUTINE_DBG_MEO00 ("bd94Dump");
	
    tsp00_Int2        CodeVars    = 255;
    const tsp00_Int4  MarkVarSize = 8;
    const char       *MarkVars    = "B94HISTM";
    tgg00_BasisError  MoveError   = e_ok;
	
    // OutPos is handled as a Pascal (!) Position
    // only for C-Routines 1 must be subtracted
	
    HostError.becomes(vf_ok);
	
    g01new_dump_page (HostFile, Buffer, OutPno, OutPos, HostError, ErrText);
    
    if ( vf_ok == HostError )
    {
        /* write header mark  */

        g10mv1 ("VBD94 ", 1,    
            MarkVarSize, BufferSize, MarkVars, POS_OFF_DIFF_BD00,
            Buffer, OutPos, MarkVarSize, MoveError);
        MoveError = e_ok; // ignore error 
        OutPos += MarkVarSize;

        memcpy (Buffer + OutPos - POS_OFF_DIFF_BD00, &CodeVars, sizeof(tsp00_Int2));
        OutPos += sizeof(tsp00_Int2);
		
        /* write contents */

		memcpy (Buffer + OutPos - POS_OFF_DIFF_BD00, &m_MaxHistoryFiles, sizeof(m_MaxHistoryFiles));
        OutPos += sizeof(m_MaxHistoryFiles);
		
		for (tsp00_Int4 HistFileNo=0; HistFileNo<m_MaxHistoryFiles; ++HistFileNo)
		{
            cbd94_HistoryFile &HistFile       = bd94_GetHistoryFileOfHistFileNo (HistFileNo);
			tsp00_PageNo       HFRoot         = HistFile.bd94FileId().fileRoot_gg00();
			tsp00_PageNo       RightmostPage  = HistFile.bd94RightmostPage();
			tsp00_TaskId       CheckTaskId    = HistFile.bd94CheckTaskId();
			
			memcpy (Buffer + OutPos - POS_OFF_DIFF_BD00, &(*(m_pHistoryFiles+HistFileNo)),
				sizeof(tbd94_HistoryFileHandle));
			OutPos += sizeof(tbd94_HistoryFileHandle);

            memcpy (Buffer + OutPos - POS_OFF_DIFF_BD00, &HFRoot, sizeof(HFRoot));
			OutPos += sizeof(HFRoot);
			
			memcpy (Buffer + OutPos - POS_OFF_DIFF_BD00, &RightmostPage, sizeof(RightmostPage));
			OutPos += sizeof(RightmostPage);
			
			memcpy (Buffer + OutPos - POS_OFF_DIFF_BD00, &CheckTaskId, sizeof(CheckTaskId));
			OutPos += sizeof(CheckTaskId);
		}
	}
}

/*===========================================================================*
 *  LOCAL FUNCTIONS (CODE)                                                   *
 *===========================================================================*/

/*---------------------------------------------------------------------------*/

inline void
bd94_WriteTrace (tgg00_TransContext &Trans,
                 const char         *Line,
                 tsp00_Int4          LineLen)
{
    b12vtrace (Trans, ta, td_none, LineLen, Line);
}

