/* @lastChanged: "1998-07-14  8:07"

  * @filename:   hco002.h
  * @purpose:    "OMS Memory Management"
  * @release:    7.1.0.0
  * @see:        "http://www.bea.sap-ag.de/doku/rel-7.1/gco10.html"
  *
  * @copyright:  "Copyright by   SAP AG, 1998"


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*/

#if !defined (__GCO16_H)
#define __GCO16_H
#include "ggg00.h"
#include "gco00.h"
#include "ilivecachesink.h" /* no check */
#include "hco10.h"
#include "gco01.h"
#include "gco001.h"

#define CHUNK_SIZE_CO16 64000
#define ALIGNMENT_CO16  sizeof(void*)

struct co16_MemChunk {
    co16_MemChunk*  m_next;
    unsigned char m_mem[CHUNK_SIZE_CO16 - sizeof(co16_MemChunk*)];
    long memSize() { return sizeof(m_mem); }
};

class co16_Allocator {
public :
  co16_Allocator() : m_sharedMemUsed(0), m_free(NULL), m_chunkCount(0) {
  }
  co16_MemChunk* newChunk(IliveCacheSink* lcSink);
  
  void freeChunk(IliveCacheSink* lcSink, co16_MemChunk* p); 

  void freeChunks(IliveCacheSink* lcSink, co16_MemChunk* p);

  long usedChunks() { return m_chunkCount; }

  void Release(IliveCacheSink* lcSink);

  inline tsp00_8ByteCounter EvalFreeSpace(IliveCacheSink* lcSink) const; 
  
private :
  tsp00_8ByteCounter m_sharedMemUsed;
  long               m_chunkCount;
  co16_MemChunk*     m_free;
};

class co16_OmsStackHeap {
public :
  co16_OmsStackHeap( co16_Allocator* allocator ) : 
    m_allocator(allocator), m_lcSink(NULL), m_usedChunks(NULL), m_rest(0), m_used(0) {}
  void* Malloc( long requiredSize);
  void  Free();
  tsp00_8ByteCounter HeapUsed() const { return m_used; } /* PTS 1107744 */
  void  SetSession(IliveCacheSink* lcSink) { m_lcSink = lcSink; }
private :
  co16_Allocator*    m_allocator;
  IliveCacheSink*    m_lcSink;
  co16_MemChunk*     m_usedChunks;
  unsigned char*     m_curr;
  long               m_rest;
  tsp00_8ByteCounter m_used; /* PTS 1107744 */
};


class co16_SharedMemAllocator {
public :
  co16_SharedMemAllocator();
  inline void* allocate   (size_t sz);
  inline void* allocate   (IliveCacheSink* lcSink, size_t sz);
  inline void  deallocate (void* pz);
  inline void  deallocate (IliveCacheSink* lcSink, void* p);
private :
  tsp00_8ByteCounter m_sharedMemUsed;
};

extern co16_SharedMemAllocator SharedMemAllocator_co16;

/*----------------------------------------------------------------------*/

inline long co16Aligned(long val) {
  return ((val + (ALIGNMENT_CO16 - 1)) & ~(ALIGNMENT_CO16 - 1)); 
}

/*----------------------------------------------------------------------*/

inline co16_MemChunk* co16_Allocator::newChunk( IliveCacheSink* lcSink) {
  CriticalSection_co001 s(lcSink, RGN_STACK_HEAP);
  s.Enter();
  if (m_free != NULL) { 
    co16_MemChunk* p = m_free;
    m_free = p->m_next;
    return p;
  }
  s.Leave();
  m_chunkCount++;
  if ((m_chunkCount <= 16) || (0 == m_chunkCount % 16)) {
    char    buffer[256];
    short   DBError;
    tsp00_Int4    BufLen;
    if (lcSink->IsDebugEnvironment()) {
      sprintf(buffer, "Proxy Chunks allocated : %d", m_chunkCount);
    }
    else {
      sprintf(buffer, "OMS Chunks allocated : %d", m_chunkCount);
    }
    BufLen = strlen(buffer);
    lcSink->MultiPurpose (m_show, mm_log, &BufLen, 
      (unsigned char*) &buffer[0], &DBError);
  }
  return REINTERPRET_CAST(co16_MemChunk*, SharedMemAllocator_co16.allocate (lcSink, sizeof(co16_MemChunk)));
}

/*----------------------------------------------------------------------*/

inline void co16_Allocator::freeChunk(IliveCacheSink* lcSink, co16_MemChunk* p) {
  CriticalSection_co001 s(lcSink, RGN_STACK_HEAP);
  s.Enter();
  p->m_next = m_free;
  m_free    = p;
}

/*----------------------------------------------------------------------*/

inline void co16_Allocator::freeChunks(IliveCacheSink* lcSink, co16_MemChunk* p) {
  co16_MemChunk* last = p;
  while (last->m_next != NULL) {
    last = last->m_next;
  }
  CriticalSection_co001 s(lcSink, RGN_STACK_HEAP);
  s.Enter();
  last->m_next = m_free;
  m_free       = p;
}

/*----------------------------------------------------------------------*/

inline tsp00_8ByteCounter co16_Allocator::EvalFreeSpace(IliveCacheSink* lcSink) const {
  tsp00_8ByteCounter freeBytes = 0;
  CriticalSection_co001 s(lcSink, RGN_SHARED_HEAP);
  s.Enter();
  co16_MemChunk* curr = m_free;
  while (curr != NULL) {
    freeBytes += CHUNK_SIZE_CO16;
    curr = curr->m_next;
  }
  return freeBytes;
};

/*----------------------------------------------------------------------*/

inline void co16_Allocator::Release(IliveCacheSink* lcSink) {
  co16_MemChunk* p = m_free;
  while (NULL != p) {
    co16_MemChunk* toFree = p;
    p                     = p->m_next;
    SharedMemAllocator_co16.deallocate (lcSink, toFree);
  }
  m_free       = NULL;
  m_chunkCount = 0;
}

/*----------------------------------------------------------------------*/

inline void* co16_OmsStackHeap::Malloc(long requiredSize) {
  void * p;
  long alignedSize = co16Aligned(requiredSize);
  if (m_rest >= requiredSize) {
    m_rest -= alignedSize;
    p       = m_curr;
    m_curr += alignedSize;
    return p;
  }
  else {
    co16_MemChunk* c = m_allocator->newChunk(m_lcSink);
    if (c != NULL) {
      c->m_next    = m_usedChunks;
      m_usedChunks = c;
      m_rest       = c->memSize() - alignedSize;
      m_curr       = c->m_mem + alignedSize;
      m_used      += c->memSize(); /* PTS 1107744 */
      return c->m_mem;
    }
    else {
      throw DbpError (DbpError::RTE_ERROR, e_new_failed);
    }
  }
}

/*----------------------------------------------------------------------*/

inline void co16_OmsStackHeap::Free() {
  if (m_usedChunks != NULL) {
    m_allocator->freeChunks(m_lcSink, m_usedChunks);
    m_usedChunks = NULL;
    m_rest       = 0;
    m_used       = 0; /* PTS 1107744 */
  }
}

/*----------------------------------------------------------------------*/

inline void* co16_SharedMemAllocator::allocate(IliveCacheSink* lcSink, size_t sz) {
  void* p = co10_malloc(lcSink, sz);
  if (NULL != p) {
    CriticalSection_co001 s(lcSink, RGN_SHARED_HEAP);
    m_sharedMemUsed += sz;
  }
  return p;
}

/*----------------------------------------------------------------------*/

inline void* co16_SharedMemAllocator::allocate(size_t sz) {
  IliveCacheSink* lcSink = co10_GetLcSink();
  return this->allocate(lcSink, sz);
}

/*----------------------------------------------------------------------*/

inline void co16_SharedMemAllocator::deallocate(IliveCacheSink* lcSink, void*p) {
  co10_free(lcSink, p);
}

/*----------------------------------------------------------------------*/

inline void co16_SharedMemAllocator::deallocate(void *p) {
  IliveCacheSink* lcSink = co10_GetLcSink();
  this->deallocate(lcSink, p);
}

#endif
