/*


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end


*/
#include "gsp00.h"
#include "gco00.h"
#include "gco102.h"
#include "hco10.h"
#include "gco10.h"
#include "hco12.h"

OmsContext_co12::ContextHeap::ContextHeap(IliveCacheSink** pSink) : m_sysLcSink(pSink), ceo381_Heap(16, 16) 
{
}

/*----------------------------------------------------------------------*/

OmsContext_co12::ContextHeap::~ContextHeap() {
  this->freeHeap();
}

/*----------------------------------------------------------------------*/

teo381_ChunkPtr OmsContext_co12::ContextHeap::GetMem(size_t sz) {
  return REINTERPRET_CAST(teo381_ChunkPtr, SharedMemAllocator_co16.allocate(*m_sysLcSink, sz));
}

/*----------------------------------------------------------------------*/

void OmsContext_co12::ContextHeap::FreeMem(teo381_ChunkPtr p) {
  SharedMemAllocator_co16.deallocate(*m_sysLcSink, p);
}

/*----------------------------------------------------------------------*/

OmsContext_co12::OmsContext_co12(OmsSession_co12* session, const OmsVersionId* vid,
                                        const tgg00_OmsVersionContext* vctxt) :
       m_session(session), m_oidDir(), m_containerDir(), m_consistentView(), m_pVersionContext(NULL), 
       m_isOpen(false), m_isDropped(false), m_isVersion(false), m_heap(&m_sysLcSink), 
       m_varObjHeapUsed(0), m_lastDropId(0), m_boundToTrans(false),
       m_marked(false), m_date(), m_time(), m_lastOpenDate(), m_lastOpenTime(), m_currLcSink(session->m_lcSink),
       m_sysLcSink(co10_GetLcSink())
{
  session->m_lcSink->GetDateTime(&m_date, &m_time);
  ResetConsistentView();
  m_oidDir.Create(this);
  m_containerDir.Create(this);
  if (NULL != vid) {
    m_isVersion      = true;
    m_versionContext = *vctxt;
    memcpy (m_version, vid, sizeof(*vid));
    m_isOpen          = true;
  }
};

/*----------------------------------------------------------------------*/

OmsContext_co12::~OmsContext_co12() {
}

/*----------------------------------------------------------------------*/

void OmsContext_co12::CleanContainerDir() {
  tsp00_Uint4 lastDropId = co10_GetLastDropId();
  if (lastDropId != m_lastDropId) {
    m_containerDir.Clean();
    m_lastDropId = lastDropId;
  }
}

/*----------------------------------------------------------------------*/

void OmsContext_co12::DeleteSelf() {
  tgg00_BasisError DBError;
  if (m_isVersion) {                /* FF 16.Nov.99  PTS 1104668 */
    m_versionContext.ovc_trans_version = m_consistentView;
    CHResult_co001 hr = m_currLcSink->DropVersion((unsigned char*) &m_versionContext, &DBError);
    this->CheckDBError(DBError, "OmsContext_co12::DropVersion", OmsObjectId());
  }
  this->OmsContext_co12::~OmsContext_co12();
  SharedMemAllocator_co16.deallocate(m_sysLcSink, this);
}

/*----------------------------------------------------------------------*/

void OmsContext_co12::DumpObjCache(OmsHandle& h) {
  char                  flags[4];
  flags[sizeof(flags)-1] = 0;
  OmsObjectContainerPtr curr;
  for (cco13_OidHash::OidIter iter = m_oidDir.First(); iter; ++iter) {
    curr = iter();
    if (curr->DeletedFlag()) {
        flags[0] = 'D';
      }
      else {
        flags[0] = ' ';
      }
      if (curr->StoredFlag()) {
        flags[1] = 'U';
      }
      else {
        flags[1] = ' ';
      }
      if (curr->LockedFlag()) {
        flags[2] = 'L';
      }
      else {
        flags[2] = ' ';
      }
      curr->m_pobj.omsDump(h, flags);
  }
}

/*----------------------------------------------------------------------*/

void OmsContext_co12::EmptyObjCache(tsp00_Uint4 containerHandle) {
  OmsObjectContainerPtr curr;
  tco13_ClsIdEntry* pContainerInfo ;
  cco13_OidHash::OidIter iter = m_oidDir.First();
  while (iter) {
    curr = iter();
    ++iter;
    if ((0xFFFFFFFF == containerHandle) || (curr->GetContainerHandle() == containerHandle)) {
      pContainerInfo = this->GetContainerInfo(containerHandle);
      DeleteObjInContext(curr, pContainerInfo);
    }
  }
};

/*----------------------------------------------------------------------*/

void OmsContext_co12::FlushObjCache(bool emptyObjCache) {
  bool                   errorOccured = false;
  bool                   checkHash    = false;
  OmsObjectContainerPtr  curr;
  tco13_ClsIdEntry*      clsinfo;
  DbpError               error(0);
  cco13_OidHash::OidIter iter = m_oidDir.First();
  while (iter) {
    if (checkHash) {
    m_oidDir.HashCheck();
    }
    curr = iter();
    ++iter;
    curr->m_beforeImages = 0;
    if (!errorOccured) {
      try {
        if (curr->DeletedFlag()) { 
          FlushDelete(curr, !emptyObjCache);
        }
        else {
          if (curr->StoredFlag()) { 
            FlushObj (curr);
          }
          else { /* CR 1103570 FF 12-Aug-1999 */
            if ((curr->LockedFlag()) &&
              (!IsVersion())         &&
              (!curr->VarObjFlag())) { 
              FlushObj (curr);
            }
          }
        }
      }
      catch (DbpError& e) {
        error        = e;
        errorOccured = true;
      }
    }
    if (emptyObjCache) {
      clsinfo = curr->GetContainerInfo(this);
      clsinfo->chainFree(*this, curr);
    }
    else {
      if (curr->VarObjFlag()) {
        REINTERPRET_CAST(OmsVarObjInfo_co12*, &curr->m_pobj)->freeVarObj(this);
      }
    }
  }
  m_oidDir.SetEmpty();
  if (errorOccured) {
    throw error;
  }
}

/*----------------------------------------------------------------------*/

void OmsContext_co12::FlushObj(OmsObjectContainerPtr p) {
  if (p->VarObjFlag()) {
    FlushVarObject (p);
  }
  else {
    tsp00_Int2  DBError;
    tco13_ClsIdEntry* clsinfo = p->GetContainerInfo(this);
    CHResult_co001 hr = m_currLcSink->UpdateObj (
      (unsigned char*) &m_consistentView,
      (unsigned char*) &clsinfo->GetFileId(),
      VersionContext(),
      (unsigned char*) &p->m_oid,
      (unsigned char*) &p->m_objseq, 
      clsinfo->GetPersistentSize(), 
      ((unsigned char*) &p->m_pobj) + sizeof(void*), 
      (unsigned char*) &clsinfo->GetInvColDesc(), 
      &DBError);
    this->CheckDBError(DBError, "OmsContext_co12::FlushObj", p->m_oid); 
    p->RemoveStoredFlag();
    m_session->IncStore_lC();
  }
}

/*----------------------------------------------------------------------*/

void OmsContext_co12::FlushDelete (OmsObjectContainerPtr pObj, bool deleteInHash) {
  
  tsp00_Int2     DBError;
  CHResult_co001 hr;
  tco13_ClsIdEntry* clsinfo = pObj->GetContainerInfo(this);
  if (pObj->VarObjFlag()) {
    bool                    isFirst = true;
    OmsVarObjInfoNode_co12* pInfoNode; 
    OmsVarObjInfo_co12*     objInfo = REINTERPRET_CAST (OmsVarObjInfo_co12*, &pObj->m_pobj);
    pInfoNode = &objInfo->m_vobjInfo;
    while (NULL != pInfoNode) {
      if (!isFirst) {
        // ADIS 1001523, T.A. 24.09.1998
        hr = m_currLcSink->LockObj (
          (unsigned char*) &m_consistentView,
          (unsigned char*) &clsinfo->GetFileId(),
          REINTERPRET_CAST(OmsTypeOid*, &pInfoNode->m_vobjOid),
          (unsigned char*) &pInfoNode->m_vobjSeq,
          &DBError);
        this->CheckDBError(DBError, "OmsContext_co12::FlushDelete (VarObj)", pInfoNode->m_vobjOid);
      }
      hr = m_currLcSink->DeleteObj (
        (unsigned char*) &m_consistentView,
        (unsigned char*) &clsinfo->GetFileId(),
        VersionContext(),
        (unsigned char*) &pInfoNode->m_vobjOid,
        (unsigned char*) &pInfoNode->m_vobjSeq, 
        (unsigned char*) &clsinfo->GetInvColDesc(),
        &DBError);
      this->CheckDBError(DBError, "OmsContext_co12::FlushDelete (VarObj)", pInfoNode->m_vobjOid);
      pInfoNode = pInfoNode->m_vobjNext;
      isFirst   = false;
    }
  } 
  else { // no Var Object
    m_session->IncDelete_lC();
    hr = m_currLcSink->DeleteObj (
      (unsigned char*) &m_consistentView,
      (unsigned char*) &clsinfo->GetFileId(),
      VersionContext(),
      (unsigned char*) &pObj->m_oid, 
      (unsigned char*) &pObj->m_objseq,
      (unsigned char*) &clsinfo->GetInvColDesc(),
      &DBError);
    this->CheckDBError(DBError, "OmsContext_co12::FlushDelete", pObj->m_oid);
  }
  if (deleteInHash) {
    DeleteObjInContext (pObj, clsinfo);
  }
}


/*----------------------------------------------------------------------*/

void OmsContext_co12::FlushVarObject (OmsObjectContainerPtr pObj) {

  CHResult_co001 hr;
  tsp00_Int2     DBError;
  tsp00_Uint4    moveLen;
  int            vObjLen;
  tgg91_PageRef  newObjSeq;
  bool           isNew    = true;
  bool           isLocked = false;
  bool           isFirst  = true;
  tsp00_Uint4    objLength;
  OmsVarObjContainer_co001 varObjContainer;
  OmsVarObjInfoNode_co12*  pInfoNode; 
  OmsVarObjInfoNode_co12*  pLastNode;

  tco13_ClsIdEntry* clsinfo = pObj->GetContainerInfo(this);
  m_session->IncStoreVarObj(); 
  OmsVarObjInfo_co12* objInfo  = REINTERPRET_CAST (OmsVarObjInfo_co12*, &pObj->m_pobj);
  unsigned char* p             = objInfo->m_pvobj;
  pInfoNode                    = &objInfo->m_vobjInfo;
  objLength                    = objInfo->m_vobjSize;
  varObjContainer.m_vobjSize   = objInfo->m_vobjSize;
  do {
    isLocked = isNew;
    isNew    = false;
    moveLen = co001_Min (VAR_OBJ_CHUNK_SIZE, objLength);
    if (moveLen >= objLength) {
      varObjContainer.m_vobjNext.setNil();
    }
    else {
      if (pInfoNode->m_vobjNext) {
        memcpy (&varObjContainer.m_vobjNext, &pInfoNode->m_vobjNext->m_vobjOid, 
          sizeof (pInfoNode->m_vobjNext->m_vobjOid));
      }
      else {
        hr = m_currLcSink->NewObj (
          (unsigned char*) &m_consistentView,
          (unsigned char*) &clsinfo->GetFileId(), 
          VersionContext(),
          0,
          0,
          NULL,
          REINTERPRET_CAST(OmsTypeOid*, &varObjContainer.m_vobjNext),
          (unsigned char*) &newObjSeq,
          &DBError);
        this->CheckDBError(DBError, "OmsContext_co12::FlushVarObject", pObj->m_oid); 
        isNew                 = true;
        pInfoNode->m_vobjNext = new(*this) OmsVarObjInfoNode_co12 (varObjContainer.m_vobjNext, newObjSeq);
      }
    }
    memcpy (&varObjContainer.vobjInfo[0], p, moveLen);
    if ((!isLocked) && (!IsVersion())) {
      hr = m_currLcSink->LockObj (
        (unsigned char*) &m_consistentView,
        (unsigned char*) &clsinfo->GetFileId(),
        REINTERPRET_CAST(OmsTypeOid*, &pInfoNode->m_vobjOid),
        (unsigned char*) &pInfoNode->m_vobjSeq, &DBError);
      this->CheckDBError(DBError, "OmsContext_co12::FlushVarObject", pInfoNode->m_vobjOid); 
    }
    objLength -= moveLen;
    vObjLen    = (objLength > 0) ? sizeof(varObjContainer) : moveLen;
    hr = m_currLcSink->UpdateObj (
      (unsigned char*) &m_consistentView,
      (unsigned char*) &clsinfo->GetFileId(),
      VersionContext(),
      (unsigned char*) &pInfoNode->m_vobjOid,
      (unsigned char*) &pInfoNode->m_vobjSeq,
      vObjLen, 
      (unsigned char*) &varObjContainer, 
      (unsigned char*) &clsinfo->GetInvColDesc(),
      &DBError);
    this->CheckDBError(DBError, "OmsContext_co12::FlushVarObject", pInfoNode->m_vobjOid); 
    if (isFirst) {
      pObj->m_objseq = pInfoNode->m_vobjSeq;
      isFirst        = false;
    }
    p         += moveLen;
    pLastNode  = pInfoNode;
    pInfoNode  = pInfoNode->m_vobjNext;
  }
  while (objLength > 0);
  if (!IsVersion()) {
    objInfo->freeVarObj(this);
  }
  pLastNode->m_vobjNext = NULL;
  while (pInfoNode) {
    if (!IsVersion()) {
      hr = m_currLcSink->LockObj (
        (unsigned char*) &m_consistentView,
        (unsigned char*) &clsinfo->GetFileId(),
        REINTERPRET_CAST(OmsTypeOid*, &pInfoNode->m_vobjOid),
        (unsigned char*) &pInfoNode->m_vobjSeq, &DBError);
    }
    if (0 == DBError) {
      hr = m_currLcSink->DeleteObj (
        (unsigned char*) &m_consistentView,
        (unsigned char*) &clsinfo->GetFileId(),
        VersionContext(),
        (unsigned char*) &pInfoNode->m_vobjOid,
        (unsigned char*) &pInfoNode->m_vobjSeq, 
        (unsigned char*) &clsinfo->GetInvColDesc(),
        &DBError);
    }
    this->CheckDBError(DBError, "OmsContext_co12::FlushVarObject", pInfoNode->m_vobjOid); 
    pLastNode = pInfoNode;
    pInfoNode = pInfoNode->m_vobjNext;
    //OmsVarObjInfoNode_co12::deleteVarObjInfoNode(*this, *pLastNode);
  }
}

/*----------------------------------------------------------------------*/

void OmsContext_co12::MarkNotBoundToTrans() { 
  m_boundToTrans = false;
  if (!m_isOpen) {
    co10_MarkUnloadable(this);
    this->ResetLcSink();
  }
}

/*----------------------------------------------------------------------*/

void  OmsContext_co12::OpenVersion(OmsSession_co12* session) {
  AssignLcSink(session->m_lcSink);
  m_currLcSink->GetDateTime(&m_lastOpenDate, &m_lastOpenTime);
  m_session    = session;
  m_isOpen     = true;
}

/*----------------------------------------------------------------------*/

void OmsContext_co12::ResetLcSink() {
  m_currLcSink = NULL;
  m_sysLcSink  = NULL;
}

/*----------------------------------------------------------------------*/

void OmsContext_co12::UnLoad() {
  const bool emptyObjCache = true;
  m_pVersionContext        = &m_versionContext;
  FlushObjCache(emptyObjCache);
  m_containerDir.ClearFreeLists();
  for (cco13_ClsIdHash::Iter iter = m_containerDir.First(); iter; ++iter) {
    iter()->VersionDelIndex();
  }
}

/*----------------------------------------------------------------------*/

void  OmsContext_co12::VersionClearObjCache() {

  OmsObjectContainerPtr curr;
  for (cco13_OidHash::OidIter iter = m_oidDir.First(); iter; ++iter) {
    curr = iter();
    if ((curr->StoredFlag()) || (curr->DeletedFlag())) {
      this->CheckDBError(e_transaction_end_required, "OmsContext_co12::VersionClearObjCache", OmsObjectId()); 
    }
    if (curr->VarObjFlag()) {
      REINTERPRET_CAST(OmsVarObjInfo_co12*, &curr->m_pobj)->freeVarObj(this);
    }
    else {
      if (curr->LockedFlag()) {
        this->FlushObj(curr);
      }
    }
  }  
  m_oidDir.SetEmpty();
}

/*----------------------------------------------------------------------*/

OmsSession_co12::OmsSession_co12(IliveCacheSink* lcSink, int cntRegions) :
m_callbackInterfaceGuid(), m_callbackInterface(NULL),
m_lcSink(lcSink), m_sizeVarobj(0), m_cntVarobj(0), m_timeout(0), m_refCnt(0),
m_subtrans_lvl(1),m_stream_io(true), m_read_only(false), m_lockObjects(NULL),
m_context(NULL), m_defaultContext(NULL), m_beforeImages(),
m_stackHeap(&Allocator_co16), m_heapUsed(0), m_monitor_curr(NULL),
m_versionsBoundToTrans(NULL), m_heapAtMethodBegin(0)
{
  int ix;
  m_beforeImages.init(this);
  m_stackHeap.SetSession(lcSink);
  m_versionsBoundToTrans.advise_allocator(this);
  m_critical_section_in_use = REINTERPRET_CAST(bool*, allocate(cntRegions * sizeof(bool)));
  for (ix = 0; ix < cntRegions; ++ix) {
    m_critical_section_in_use[ix] = false;
  }
}

/*----------------------------------------------------------------------*/

OmsSession_co12::~OmsSession_co12() {
  OmsLockObject* pLockObj = m_lockObjects;
  while (pLockObj != NULL) {
    pLockObj->InvalidateHandle();
    pLockObj = pLockObj->Next();
  }
  m_lockObjects = NULL;
  if (this->InVersion()) {
    OmsContext_co12* pVersion = m_context;
    this->CloseVersion();
    pVersion->MarkNotBoundToTrans();
  }
  m_lcSink->SetDefaultContext(NULL);
  m_defaultContext->DeleteSelf();
  deallocate(m_critical_section_in_use);
}

/*----------------------------------------------------------------------*/

const OmsObjectId& OmsSession_co12::CastOid(const GUID& castToGuid, const OmsObjectId& oid) {
  
  if (!oid) { // Null Oid given
    return oid;
  }
  OmsObjectContainerPtr p = m_context->GetObj(oid, false);
  tco13_ClsIdEntry*     pFrom = p->GetContainerInfo(m_context);
  if (pFrom->GetGuid() == castToGuid) {
    return oid;
  }
  if (m_context->m_containerDir.IsDerivedClassOf(pFrom->GetGuid(), castToGuid)) {
    return oid;
  }
  if (m_context->m_containerDir.IsBaseClassOf(castToGuid, pFrom->GetGuid())) {
    return oid;
  }
  char msg[80];
  sprintf (msg, "guid : %8X-%4X-%4X-%1X%1X%1X%1X%1X%1X%1X%1X", castToGuid.Data1, castToGuid.Data2, castToGuid.Data3,
    castToGuid.Data4[0], castToGuid.Data4[1], castToGuid.Data4[2], castToGuid.Data4[3], castToGuid.Data4[4], castToGuid.Data4[5],
    castToGuid.Data4[6], castToGuid.Data4[7]);
  this->ThrowDBError (e_incompatible_oid, msg, oid);
  return oid;
}

/*----------------------------------------------------------------------*/

void OmsSession_co12::ChangedConsistentView() {
  m_currVarObjChunk.Clear();
}

/*----------------------------------------------------------------------*/

void OmsSession_co12::CleanAfterDropSchema() {
  m_defaultContext->m_oidDir.Clean();
  m_defaultContext->CleanContainerDir();
}

/*----------------------------------------------------------------------*/

void OmsSession_co12::ClearDefaultContext() {
  m_defaultContext->ClearObjCache();
  this->ClearFreeLists();
  this->FreeStackHeap();
}

/*----------------------------------------------------------------------*/

bool OmsSession_co12::IsLocked (const OmsObjectId& oid) {
 
  if (m_context->IsVersion()) {
    return true;
  }
  OmsObjectContainerPtr found = m_context->FindObjInContext (&oid);
  if ((NULL != found) && (found->LockedFlag())) {
    return true;
  }
  tgg00_BasisError         DBError;
  CUnknownContainerId_co12 UnknownFileId;
  CHResult_co001 hr = m_lcSink->IsLocked ((unsigned char*) &UnknownFileId, 
    REINTERPRET_CAST(OmsTypeOid*, CONST_CAST(OmsObjectId*, &oid)), &DBError);
  if (e_ok == DBError) {
    return true;
  } 
  else {
    if (e_object_not_locked == DBError) {
      return false;
    }
    else {
      this->ThrowDBError (DBError, "IsLocked", oid);
    }
  }
  return false;
}

/*----------------------------------------------------------------------*/


void OmsSession_co12::CreateVersion(const OmsVersionId& versionId) {

  const char* msg = "OmsSession_co12::CreateVersion";

  tsp00_Int2              DBError;
  tgg00_OmsVersionContext versionContext;
  OmsContext_co12*        context = NULL;

  this->IncCreateVersion   ();
  CHResult_co001 hr = m_lcSink->CreateOpenVersion (
      (unsigned char*) &m_context->m_consistentView,
      (OmsVersionId*)  &versionId,
      true,
      (unsigned char*) &versionContext,
      &DBError);
  if (DBError != 0) {
      this->ThrowDBError (DBError, msg, OmsObjectId());
  }
  /* create a new context for the version */
  try {
    context = new(m_lcSink) OmsContext_co12(this, &versionId, &versionContext);
    context->NewConsistentView();
    context->ResetLcSink();
    this->OpenVersion(context);
  }
  catch (DbpError& e) {
    if (context != NULL) {
      context->DeleteSelf();
    }
    OMS_TRACE(omsTrError, m_lcSink, "Create Version Error : " << e.dbpError());
    throw e;
  }
}

/*----------------------------------------------------------------------*/

void OmsSession_co12::DeleteAll(const GUID& guid, OmsSchemaHandle Schema, OmsContainerNo ContainerNo) {

  tco13_ClsIdEntry* clsinfo = this->GetClsInfo (guid, Schema, ContainerNo);
  tsp00_Int4 cntDeleted;
  tsp00_Int2 DBError;
  CHResult_co001 hr = m_lcSink->DeleteAllObj ((unsigned char*) &clsinfo->GetFileId(), m_context->VersionContext(),
    &cntDeleted, &DBError);
  if ( DBError != 0 ) {
    this->ThrowDBError (DBError, "omsDeleteAll");
  }
  m_context->EmptyObjCache(clsinfo->GetContainerHandle());
  this->IncDelete_lC (cntDeleted);
}

/*----------------------------------------------------------------------*/

void OmsSession_co12::DeleteVarObject(const OmsVarOid& oid) {
  OmsObjectContainerPtr found = this->LoadVarObject(oid, VarObjExclusiveLock, -1, NULL);
  if (!found) {
    this->ThrowDBError (e_object_not_found, "OmsSession_co12::DeleteVarObject", oid);
  }
  OmsVarObjInfo_co12* pObjInfo = REINTERPRET_CAST (OmsVarObjInfo_co12*, &found->m_pobj); /* PTS 1106727 */
  pObjInfo->unlock();
  if (!found->LockedFlag()) {
    if (!IsLocked(oid)) {
      this->ThrowDBError (e_object_not_locked, "DeleteVarObject", oid);
    }
  }
  if (!this->InsertBeforeImage (found)) {
    m_context->FreeVarObj(pObjInfo->m_pvobj, pObjInfo->m_vobjSize);
  }
  pObjInfo->m_pvobj     = NULL;
  pObjInfo->m_vobjSize  = 0;
  found->MarkDeleted();
  if (oid == m_currVarObjChunk.m_oid) {
    m_currVarObjChunk.Clear();
  }
}

/*----------------------------------------------------------------------*/

const  void* OmsSession_co12::DeRefVarObject(const OmsVarOid& oid) {
 
  this->IncLoadVarObj();
  OmsObjectContainerPtr p = m_context->FindVarObjInContext(oid, VarObjShareLock, 0xFFFFFFFF, NULL);
  if (NULL == p) {
    tsp00_Uint4 objSize = m_context->GetVarObjFromLiveCacheBase(oid, false);
    p = m_context->LoadVarObject(oid, VarObjShareLock, objSize, m_currVarObjChunk, NULL);
    m_currVarObjChunk.Clear();
  }
  OmsVarObjInfo_co12* pObjInfo = REINTERPRET_CAST (OmsVarObjInfo_co12*, &p->m_pobj);
  return pObjInfo->m_pvobj;
}

/*----------------------------------------------------------------------*/

void OmsSession_co12::DropVersion(OmsContext_co12* context) {

  const char* msg = "OmsSession_co12::DropVersion";
  if (context->IsBoundToTrans()) {
    /* version is bound to trans, bound by myself ? */
    if ((context != CurrentContext()) && (!this->VersionBoundByMe(context))) {
      this->ThrowDBError (e_missing_privilege, msg);
    }
    context->CloseVersion();
    this->RemoveFromTransVersion(context);
  }
  else {
    context->AssignLcSink(m_lcSink);
  }
  m_beforeImages.removeContext(context);
  /* free context */
  context->DeleteSelf();
  /* return to default context */
  this->ActivateDefaultContext();
  this->IncDropVersion();
}

/*----------------------------------------------------------------------*/

bool OmsSession_co12::HistoryInUse(const OmsObjectId& oid) {
  tgg00_BasisError         LCError;
  bool                     isUsed = true;
  CUnknownContainerId_co12 UnknownFileId;    
  CHResult_co001 hr = m_lcSink->IsObjHistoryUsed(UnknownFileId, *REINTERPRET_CAST(const OmsTypeOid*, &oid), isUsed, LCError);
  if (0 != LCError) {
    this->ThrowDBError(LCError, "OmsSession_co12::HistoryInUse", oid);
  }
  return isUsed;
}

/*----------------------------------------------------------------------*/

void OmsContext_co12::NewConsistentView() {
  tsp00_Int2 DBError;
  tgg91_TransNo consistentView = m_consistentView;
  CHResult_co001 hr = m_currLcSink->NewConsistentView ((unsigned char*) &m_consistentView, &DBError);
  if (0 != DBError) {
    m_consistentView = consistentView;
    m_session->ThrowDBError (DBError, "omsNewConsistentView", OmsObjectId());
    return;
  }
  m_session->ChangedConsistentView();
}

/*----------------------------------------------------------------------*/

void OmsSession_co12::NewConsistentView (OmsObjectId* pOids, int cntOids, short timeout, OmsObjectId* pErrOids, int& cntErrOids) {
  tsp00_Int2        DBError = e_ok;
  int               ix;
  tsp00_Int4        oldTimeout = m_timeout;
  int               maxErrOids = cntErrOids;
  CHResult_co001    hr;
  tgg91_PageRef     pageRef;
  tsp00_Int4        start_time_sec; 
  tsp00_Int4        start_time_msec;
  tsp00_Int4        stop_time_sec; 
  tsp00_Int4        stop_time_msec;
  tsp00_Int4        diff_time_sec; 
  tsp00_Int4        diff_time_msec;
  tsp00_Int4        runTime = 0;
  
  OmsObjectId* pCurrOid = CONST_CAST(OmsObjectId*, pOids);
  cntErrOids = 0;
  if (this->CurrentContext()->IsVersion()) {
    return;
  }
  this->CurrentContext()->RemoveUnlockedObjFromCache();
  this->SetTimeout(timeout);
  m_timeout  = timeout; 
  m_lcSink->GetClock(&start_time_sec, &start_time_msec);
  for (ix = 0; ix < cntOids; ++ix) {
    m_lcSink->GetClock(&stop_time_sec, &stop_time_msec);
    diff_time_sec  = stop_time_sec - start_time_sec;
    diff_time_msec = stop_time_msec - start_time_msec;
    if (stop_time_msec < start_time_msec)
    {
      diff_time_sec  = diff_time_sec - 1;
      diff_time_msec = diff_time_msec + 1000000 ;
    }
    runTime += diff_time_sec;
    if ((runTime > timeout) && (timeout > 0)) {
      m_timeout = oldTimeout;
      this->MonitorWaitNewConsistentView(runTime);
      this->SetTimeout(oldTimeout);
      this->ThrowDBError(e_lock_collision, "OmsSession_co12::NewConsistentView", *pCurrOid);
    }
    CUnknownContainerId_co12 UnknownFileId;    
    pageRef.gg91SetDummyRef();
    hr = m_lcSink->LockObj (
      (unsigned char*) &this->CurrentContext()->m_consistentView,
      (unsigned char*) &UnknownFileId,
      REINTERPRET_CAST(OmsTypeOid*, pCurrOid),
      (unsigned char*) &pageRef,
      &DBError);
    if (e_object_dirty == DBError) {
      if (cntErrOids < maxErrOids) {
        DBError   = 0;
        *pErrOids = *pCurrOid;
        ++pErrOids;
        ++cntErrOids;
      }
    }
    if (e_ok != DBError) {
      int jx;
      for (jx = 0; jx < ix; jx++) {
        this->UnlockObj (*pOids);
        ++pOids;
      }
     
      break;
    }
    ++pCurrOid;
  }
  this->SetTimeout(oldTimeout);
  this->MonitorWaitNewConsistentView(runTime); 
  m_timeout = oldTimeout;
  if (e_ok == DBError) {
    this->CurrentContext()->NewConsistentView();
  }
  else {
    if (e_object_dirty == DBError) {
      this->IncOutOfDate();
      throw OmsOutOfDate(*pCurrOid);
    }
    else {
      this->ThrowDBError(DBError, "OmsSession_co12::NewConsistentView", *pCurrOid);
    }
  }
}

/*----------------------------------------------------------------------*/

OmsAbstractObject* OmsSession_co12::NewKeyedObject(const GUID&          guid, 
                                                   const unsigned char* key, 
                                                   OmsSchemaHandle      Schema, 
                                                   OmsContainerNo       ContainerNo) {
  
  const char* msg = "OmsSession_co12::NewKeyedObject";

  bool        newOid           = true;
  bool        writeBeforeImage = true;
  tsp00_Int2  DBError;
  OmsObjectId errOid;
  
  HRESULT hr;
  tco13_ClsIdEntry*   pContainerInfo = this->GetClsInfo (guid, Schema, ContainerNo);
  OmsObjectContainer* po      = this->GetMemory(*pContainerInfo);
  if (m_context->IsVersion()) {
    int                   ObjHistReadCount;
    OmsObjectContainerPtr found;
    found = pContainerInfo->VersionFindKey(key);
    if (NULL != found) {
      if (!found->DeletedFlag()) {
        throw OmsDuplicateKey (e_duplicate_hash_key, found->m_oid);
      }
      else {
        writeBeforeImage = false;
        this->InsertBeforeImage (found);
        found->InitState(pContainerInfo);
        newOid         = false;
        po             = found;
      }
    }
    else {
      hr = m_lcSink->GetObjWithKey (
        (unsigned char*) &m_context->m_consistentView,
        (unsigned char*) &pContainerInfo->GetFileId(),
        m_context->VersionContext(),
        pContainerInfo->GetKeyLen(), 
        (unsigned char*) key, 
        false, pContainerInfo->GetPersistentSize(), ((unsigned char*) &po->m_pobj) + sizeof(void*),
        REINTERPRET_CAST(OmsTypeOid*, &po->m_oid), (unsigned char*) &po->m_objseq, &ObjHistReadCount, &DBError);
      if FAILED (hr) {
        pContainerInfo->chainFree(*m_context, po);
      }
      if ( DBError != 0 ) {
        if ( e_hash_key_not_found != DBError ) {
          pContainerInfo->chainFree(*m_context, po);
          this->ThrowDBError (DBError, msg);
        }
      }
      else {
        this->IncLogHop(ObjHistReadCount);
        OmsObjectContainerPtr found = m_context->FindObjInContext (&po->m_oid);
        if (found) {
          pContainerInfo->chainFree(*m_context, po);
          if (!found->DeletedFlag()) {
            throw OmsDuplicateKey (e_duplicate_hash_key, found->m_oid);
          }
          else {
            writeBeforeImage = false;
            this->InsertBeforeImage (found);
            found->InitState(pContainerInfo);
            newOid         = false;
            po             = found;
          }
        }
        else {
          throw OmsDuplicateKey (e_duplicate_hash_key, po->m_oid);
        }
      }
    }
    if (newOid) {
      po->m_oid = m_context->VersionNewOid();
      po->m_objseq.gg91SetNilRef();
      memcpy (pContainerInfo->GetKeyPtr(po), key, pContainerInfo->GetKeyLen());
      pContainerInfo->VersionAddKey(po);
      m_context->PutObjectIntoContext (po, pContainerInfo->GetContainerHandle());
      newOid = false;
    }
  }
  if (newOid) {
    HRESULT hr = m_lcSink->NewObj (
      (unsigned char*) &m_context->m_consistentView,
      (unsigned char*) &pContainerInfo->GetFileId(),
      m_context->VersionContext(),
      pContainerInfo->GetKeyDesc().epos(),
      pContainerInfo->GetKeyLen(), 
      (unsigned char*) key,
      REINTERPRET_CAST(OmsTypeOid*, &po->m_oid), 
      (unsigned char*) &po->m_objseq,
      &DBError);
    if FAILED (hr) { 
      pContainerInfo->chainFree(*m_context, po);
      //omsThrowHResult (hr); 
      return NULL; 
    }
    else {
      if ( DBError != 0 ) {
        OmsObjectContainerPtr found = NULL;
        if (e_duplicate_hash_key == DBError) {
          // check, if current transcation already deleted object. If true, reuse it
          tsp00_Int2 getDBError;
          int        dummy;
          hr = m_lcSink->GetObjWithKey (
            (unsigned char*) &m_context->m_consistentView,
            (unsigned char*) &pContainerInfo->GetFileId(),
            m_context->VersionContext(),
            pContainerInfo->GetKeyLen(),
            CONST_CAST(unsigned char*, key), 
            false,
            pContainerInfo->GetPersistentSize(), 
            ((unsigned char*) &po->m_pobj) + sizeof(void*),
            REINTERPRET_CAST(OmsTypeOid*, &po->m_oid),
            (unsigned char*) &po->m_objseq,
            &dummy,
            &getDBError);
          if (!FAILED(hr) && (0 == getDBError)) {
            errOid = po->m_oid;
            found  = m_context->FindObjInContext (&po->m_oid);
            if (found) {
              if (found->DeletedFlag()) {
                writeBeforeImage = false;
                this->InsertBeforeImage (found);
              }
              else {
                found = NULL;
              }
            }
          }
        }
        pContainerInfo->chainFree(*m_context, po);
        if (found) {
          po = found;
          po->InitState(pContainerInfo);
        }
        else {
          if ((e_duplicate_hash_key == DBError) || ( 400 == DBError)) {
            throw OmsDuplicateKey (DBError, errOid);
          }
          else {
            this->ThrowDBError (DBError, "omsNewKeyedObject");
          }
          return NULL;
        }
      }
      else {
        m_context->PutObjectIntoContext (po, pContainerInfo->GetContainerHandle());
      }
    }
    memcpy (pContainerInfo->GetKeyPtr(po), key, pContainerInfo->GetKeyLen());
  }
  po->MarkLocked();
  po->MarkStored();
  if (writeBeforeImage) {
    this->InsertNewBeforeImage (po, pContainerInfo);
  }
  _ASSERTE (_CrtCheckMemory ());
  return &po->m_pobj;
}

/*----------------------------------------------------------------------*/

OmsAbstractObject* OmsSession_co12::NewObject(const GUID& guid, OmsSchemaHandle Schema, OmsContainerNo ContainerNo) {
  
  tco13_ClsIdEntry* pContainerInfo = this->GetClsInfo (guid, Schema, ContainerNo);
  OmsObjectContainerPtr po  = this->GetMemory(*pContainerInfo);
  if (this->InVersion()) {
    po->m_oid = m_context->VersionNewOid();
    po->m_objseq.gg91SetNilRef();
  }
  else {
    tsp00_Int2 DBError;
    CHResult_co001 hr = m_lcSink->NewObj (
      (unsigned char*) &m_context->m_consistentView,
      (unsigned char*) &pContainerInfo->GetFileId(),
      m_context->VersionContext(),
      0,
      0,
      NULL,
      REINTERPRET_CAST(OmsTypeOid*, &po->m_oid), 
      (unsigned char*) &po->m_objseq,
      &DBError);
    if ( DBError != 0 ) {
      pContainerInfo->chainFree(*this->CurrentContext(), po);
      this->ThrowDBError (DBError, "omsNewObject");
    }
  }
  m_context->PutObjectIntoContext (po, pContainerInfo->GetContainerHandle());
  po->MarkLocked();
  po->MarkStored();
  this->InsertNewBeforeImage (po, pContainerInfo);
  return &po->m_pobj;
}

/*----------------------------------------------------------------------*/

const OmsVarOid& OmsSession_co12::NewVarObject(OmsSchemaHandle Schema, OmsContainerNo ContainerNo) {

  tco13_ClsIdEntry* pContainerInfo = this->GetClsInfo (VAR_OBJ_GUID_GCO10, Schema, ContainerNo);
  OmsObjectContainerPtr po  = this->GetMemory(*pContainerInfo);
  if (this->InVersion()) {
    po->m_oid = m_context->VersionNewOid();
    po->m_objseq.gg91SetNilRef();
  }
  else {
    tgg00_BasisError DBError;
    CHResult_co001 hr = m_lcSink->NewObj (
      (unsigned char*) &m_context->m_consistentView,
      (unsigned char*) &pContainerInfo->GetFileId(),
      m_context->VersionContext(),
      0, 0, NULL,
      REINTERPRET_CAST(OmsTypeOid*, &po->m_oid),
      (unsigned char*) &po->m_objseq,
      &DBError);
    if ( DBError != 0 ) {
      pContainerInfo->chainFree(*CurrentContext(), po);
      this->ThrowDBError (DBError, "omsNewVarObject");
    }
  }
  m_context->PutObjectIntoContext (po, pContainerInfo->GetContainerHandle());
  po->MarkLocked();
  po->MarkVarObj();
  OmsVarObjInfo_co12* pObjInfo = new(&po->m_pobj) OmsVarObjInfo_co12(po->m_oid, 0, po->m_objseq); 
  return *(REINTERPRET_CAST(OmsVarOid*, CONST_CAST(OmsObjectId*, &po->m_oid)));
}

/*----------------------------------------------------------------------*/

void OmsSession_co12::RegContainer (OmsSchemaHandle Schema,
                             const GUID&        guid,
                             const char*        ClassName, 
                             size_t             PersistentSize,
                             size_t             ObjectSize,
                             const GUID*        pBaseClass,
                             OmsAbstractObject* Obj, 
                             OmsContainerNo     ContainerNo) {
  
  struct VirtualObject {
    void* vtptr;
  };  
  tco13_ClsIdEntry* pContainerInfo; 
  bool              isVarObject;
  bool              isRegistered = false;
  isVarObject = (0 == memcmp (&VAR_OBJ_GUID_GCO10, &guid, sizeof(guid)));
  try {
    pContainerInfo = this->GetClsInfo (guid, Schema, ContainerNo);
    isRegistered = true;
  }
  catch (DbpError& e) {
    if (e_unknown_guid != e.dbpError()) {
      throw e;
    }
  }
  if (!isRegistered) {
    tgg00_ObjColDesc    InvColDesc;
    ObjectSize     += OmsObjectContainer::headerSize();
    PersistentSize -= /* virtual table ptr */ sizeof (void*);
    InvColDesc.colHashCount_gg00 = 0;
    InvColDesc.colInvCount_gg00  = 0;
    InvColDesc.colHashList_gg00  = NULL;
    InvColDesc.colInvList_gg00   = NULL;
    if  (isVarObject) {
      PersistentSize = sizeof(OmsVarObjContainer_co001);
    }
    ContainerInfo_co18* p = co10_RegisterContainer (m_lcSink, guid, ClassName, PersistentSize,
      ObjectSize, pBaseClass, InvColDesc, Schema, ContainerNo, (REINTERPRET_CAST(VirtualObject*, Obj))->vtptr); 
    pContainerInfo = new(m_context) tco13_ClsIdEntry (p, m_context);
    this->AddContainerInfo (pContainerInfo);
  }
  if (!isVarObject) {
    pContainerInfo->chainFree(*this->CurrentContext(), Obj);
  }
};

/*----------------------------------------------------------------------*/

void OmsSession_co12::RegVarObjectContainer(OmsSchemaHandle Schema, OmsContainerNo cno)
{
  OmsVarObjInfo_co12 info;
  RegContainer (Schema, VAR_OBJ_GUID_GCO10, "VarObjContainer", sizeof(info), sizeof(info), NULL, &info, cno);  
}

/*----------------------------------------------------------------------*/

void OmsSession_co12::RegisterCallbackInterface(const GUID& guid, OmsCallbackInterface* pInterface) {
  if (NULL != m_callbackInterface) {
    if (m_callbackInterfaceGuid == guid) {
      return;
    }
    m_callbackInterface->omsDestroySelf(); 
  }
  m_callbackInterface     = pInterface;
  m_callbackInterfaceGuid = guid;
}

/*----------------------------------------------------------------------*/

void OmsSession_co12::ReleaseVarObject(const OmsVarOid& oid)
{
  OmsObjectContainerPtr found = m_context->FindObjInContext (&oid);
  if (NULL != found) {
    OmsVarObjInfo_co12* pObjInfo = REINTERPRET_CAST (OmsVarObjInfo_co12*, &found->m_pobj);
    pObjInfo->unlock();
  }
}

/*----------------------------------------------------------------------*/

void OmsSession_co12::RemoveFromTransVersion(OmsContext_co12* context) {
  for (cgg251dclIterator<OmsContext_co12*,OmsSession_co12> iter = m_versionsBoundToTrans.begin(); iter; ++iter) {
    if (*iter() == context) {
      m_versionsBoundToTrans.remove(iter);
      return;
    }
  }
}

/*----------------------------------------------------------------------*/

void OmsSession_co12::ReleaseCallbackInterface() {
  if (NULL != m_callbackInterface) {
    m_callbackInterface->omsDestroySelf();
    m_callbackInterface = NULL;
  }
}

/*----------------------------------------------------------------------*/

void OmsSession_co12::SetCurrentMonitorMethod(tgg00_COMMonitorInfo* p) {
  /* PTS 1107744 */
  tsp00_8ByteCounter heapUsed = m_heapUsed + m_stackHeap.HeapUsed() + m_defaultContext->HeapUsed();
  if (NULL == p) {
    // end of procedure call
    if ((heapUsed > m_heapAtMethodBegin) &&
      ( NULL != m_monitor_curr)          &&
      ( heapUsed > m_monitor_curr->cmiCounters_gg00[cmiCacheSize]))
    {
      m_monitor_curr->cmiCounters_gg00[cmiCacheSize] = heapUsed;
    }
  }
  else {
    // start of procedure call
    m_heapAtMethodBegin = heapUsed;
    m_monitor_curr      = p;
  }
}
 
/*----------------------------------------------------------------------*/

void OmsSession_co12::SetTimeout(short timeout) {
  short       DummyError;
  tsp00_Int4  BufLen = sizeof(timeout);
  this->m_timeout = timeout;
  CHResult_co001 hr = m_lcSink->MultiPurpose (m_set, mm_lock_timeout, &BufLen, 
    (unsigned char*) &timeout, &DummyError);
}

/*----------------------------------------------------------------------*/

void OmsSession_co12::ThrowDBError(tsp00_Int2 e, const char* msg) {
  this->ThrowDBError (e, msg, OmsObjectId());
}

/*----------------------------------------------------------------------*/

void OmsSession_co12::ThrowDBError(tsp00_Int2 e, const char* msg, const OmsObjectId& oid) {
  tsp00_Int4  BufLen = sizeof(e);
  tsp00_Int2  SqlError;
  // transform basis error into sql return code
  m_lcSink->MultiPurpose (m_change, mm_nil, &BufLen, 
    (unsigned char*) &e, &SqlError);
  switch (SqlError) {
  case 400 :
  case 500 : /* PTS 1102829 */
    OMS_TRACE(omsTrError, m_lcSink, "throw LockTimeout");
    IncTimeout();
    throw OmsLockTimeout(oid);
  default :
    OMS_TRACE(omsTrError, m_lcSink, "Error : " << e << "sqlError : " << SqlError << ", " << msg << ", " << oid);
    IncExceptions();
    throw DbpError (DbpError::DB_ERROR, SqlError, msg, oid);
  }
}

/*----------------------------------------------------------------------*/

void OmsSession_co12::ThrowReturnCode (tsp00_Int2 e, const char* msg) {
  OMS_TRACE(omsTrError, m_lcSink, "SqlError : " << e << ", " << msg);
  throw DbpError (DbpError::DB_ERROR, e, msg);
}

/*----------------------------------------------------------------------*/

void OmsSession_co12::Trace( const char*fmt, ... ) {
#if defined(WIN32)
  char buffer[256];
#else
  char buffer[1024];
#endif

  va_list args;
  va_start (args, fmt);
#if defined(WIN32)
  _vsnprintf ( buffer, sizeof(buffer), fmt, args );
#else
  vsprintf ( buffer, fmt, args );
  buffer[256] = '\0';
#endif
  m_lcSink->Vtrace (strlen(buffer), (unsigned char*) &buffer[0]);
  va_end (args);
}

/*----------------------------------------------------------------------*/

void OmsSession_co12::TransEnd() {
  short DBError = 0;
  if (m_defaultContext->ExistsConsistentView()) {
    CHResult_co001 hr = m_lcSink->EndConsistentView(
      (unsigned char*) &m_defaultContext->m_consistentView, &DBError);
  }
  m_currVarObjChunk.Clear();
  this->ClearFreeLists();
  this->ClearTransVersions();
  m_defaultContext->CleanContainerDir();
  if ( 0 != DBError ) { 
    this->ThrowReturnCode (DBError, "omsTransEnd");
  }
}

/*----------------------------------------------------------------------*/

void OmsSession_co12::UnlockObj (const OmsObjectId& oid)
{
  tsp00_Int2               DBError;
  CUnknownContainerId_co12 UnknownFileId; 
  CHResult_co001 hr = m_lcSink->UnlockObj ((unsigned char*) &UnknownFileId,
          REINTERPRET_CAST(OmsTypeOid*, CONST_CAST(OmsObjectId*, &oid)), &DBError);
  if ( DBError != 0 ) { 
    this->ThrowDBError (DBError, "OmsSession_co12::Unlock", oid);
  }
}

/*----------------------------------------------------------------------*/

bool OmsSession_co12::VersionBoundByMe(OmsContext_co12* context) {
  for (cgg251dclIterator<OmsContext_co12*,OmsSession_co12> iter = m_versionsBoundToTrans.begin(); iter; ++iter) {
    if (*iter() == context) {
      return true;
    }
  }
  return false;
}