/*!***************************************************************************

  module      : gen71.h

  -------------------------------------------------------------------------

  responsible : JoergM

  special area: dispatcher
  description : Dispatcher and Queue Management


  last changed: 1999-01-27  17:45
  see also    : example.html ...

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 1998



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/


#ifndef GEN71_H
#define GEN71_H



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gsp00.h"
#include "gen50.h"

/*===========================================================================*
 *  FUNCTION PROTOTYPES                                                      *
 *===========================================================================*/
/*!
  Chapter: Dispatcher
 */
/*!
  Function:     en71Dispatcher
  description:  Select the next running task in a UKT
    
  This function is called within the RTE code, whenever a new task selection
  is performed. It therefore scans the UKT specific queues for candidates and
  moves them into the running queue if they are allowed to run.

  It begins scanning the queue that queues request from other UKTs (XXX)

  Second scan is done on the queue of outstanding asynchronous IO (AIOC)

  Third scan is done on the queue of outstanding communication request (COM)

  Forth scan is done on the queue of internal requests to UKT (IOC)

  Then the top element of the run queue is selected.

  If run queue was empty after all these scans UKT starts sleeping on its semaphore.

  arguments:
    this_ukt [in] Pointer to current UKT Control structure

  return value: none
 */
externC void en71Dispatcher ( ten50_UKT_Control *this_ukt );

/*!
  EndChapter: Dispatcher
 */
/*!
  Chapter: Queuing
 */
/*!
  Function:     e71_ioc
  description:  Enqueue element in IOC queue
  arguments:
    tid [in] Pointer to current Task Control structure
    quu [in] Pointer to request element to be enqueued
  return value: none
 */
externC void e71_ioc ( struct TASK_TYPE     * tid ,
           struct DOUBLY_LINKED * quu );

/*!
  Function:     e71_com
  description:  Enqueue task in COM queue
  arguments:
    tcb [in] Pointer to current Task Control Block
  return value: none
 */
externC void e71_com ( struct TASK_TYPE     * tcb );

/*!
  Function:     e71_EnquRav
  description:  Enqueue task in RUN queue using BASE_PRIO_RAV

  If Task was waiting in a vsuspend() call, this call is used.

  arguments:
    tcb [in] Pointer to current Task Control Block
    Prio [in] If set increase dynamic priority
  return value: none
 */
externC void e71_EnquRav ( struct TASK_TYPE * tcb ,
                           tsp00_Bool         Prio );

/*!
  Function:     e71_EnquRex
  description:  Enqueue task in RUN queue using BASE_PRIO_REX

  If Task was waiting rescheduling itself, this call is used.
  Base priority is lower than BASE_PRIO_RAV.

  arguments:
    tcb [in] Pointer to current Task Control Block
    Prio [in] If set increase dynamic priority
  return value: none
 */
externC void e71_EnquRex ( struct TASK_TYPE * tcb ,
                           tsp00_Bool         Prio );

/*!
  Function:     e71_ukp_to_ukp
  description:  Enqueue function request from UKT to UKT

  Enqueue a function call to be executed on other UKT.
  Function is called with two argument

    void Func(void *FuncParam, struct TASK_TYPE *FromTask)

  arguments:
    TargetUkt [in] Pointer to target User Kernel Thread
    TargetTcb [in] Pointer to target Task Control Block
    Func [in] Function to execute in target UKT
    FuncParam [in] Function func parameter
  return value: none
 */
externC void e71_ukp_to_ukp ( ten50_UKT_Control    * TargetUkt ,
                              struct TASK_TYPE     * TargetTcb ,
                              VOIDFUNC_PTR           Func ,
                              void                 * FuncParam );

/*!
  EndChapter: Queuing
 */

/*
 * Assembler exports (ven71a): Very machine dependend ;-)
 */

/*
 * typedef struct context    
 * {
 *     unsigned char * sp ;
 *     unsigned char * top ;
 * } ten50_UContext;
 *
 * Each task control block (tcb) has a struct context.
 * When we alloc a stack for a task 
 *     context->sp  = task stack begin
 *     context->top = task stack begin + size
 * The 1st time we change to a task stack, we will return not to the caller,
 *    but to the start of task_proc().
 *
 * Initialize: call e71_initstack(context, task_proc) once per task stack
 *     push task_proc()'s start addr onto task stack
 *     push regs that should be saved onto task stack
 *    put task stack ptr into context->sp
 *    to prepare for stack check, mark end of task stack
 *    return to caller
 * When all stacks initialized, change to 1st stack: e71_firststack(context)
 *    context = 1st task's context
 *    change stack ptr to context->sp (the task's stack)
 *    pop regs that were saved in e71_inittask() 
 *    pop task_proc()'s start addr, which was saved in e71_inittask()
 *    return to start of task_proc(), not caller
 * Note that stack ptr is at begin of 1st task stack.
 * To change to the next task, call e71_chgstack(new_context, old_context)
 *    new_context = 2nd task's context
 *    old_context = 1st task's context
 *    push regs onto 1st task stack (current stack ptr)
 *    push caller's return addr onto 1st task's stack
 *    update old_context->sp with stack ptr
 *    change stack ptr to new_context->sp (the 2nd task's stack)
 *    pop regs that were saved in e71_inittask() 
 *    pop task_proc()'s start addr, which was saved in e71_inittask()
 *    update new_context->sp with stack ptr
 *    return to start of task_proc(), not caller
 * Suppose we have 2 tasks.
 * To change back to the 1st task, call e71_chgstack(new_context, old_context)
 *    new_context = 1st task's context
 *    old_context = 2nd task's context
 *    push regs onto 2nd task stack (current stack ptr)
 *    push caller's return addr onto 2nd task's stack
 *    update old_context->sp with stack ptr
 *    change stack ptr to new_context->sp (the 1st task's stack)
 *    pop regs that were saved in e71_chgstack() 
 *    pop return addr that was saved in e71_chgstack()
 *    update new_context->sp with stack ptr
 *    return to 1st task's caller
 * From now on, changing tasks follows the pattern in the last example: we
 *    resume the new task's identity and return to its caller.
 */
#if defined(NMP)
externC void    e71_chgstack ( ten50_UContext   * newC , ten50_UContext    * oldC );
externC void    e71_getregs ( ten50_UContext    * context );
externC void    e71_first_task ( ten50_UContext * context );
externC void    e71_make_core ( ten50_UContext  * context );
#endif

#if defined(I386)
externC void    e71_chgstack ( ten50_UContext   * newC , ten50_UContext    * oldC );
externC void    e71_initstack ( ten50_UContext  * context , void (* func) () );
externC void    e71_make_core ( ten50_UContext  * context );
#endif

#if defined(_IBMR2)
externC struct STACK_TYPE * e71a_own_frame ();
#endif

#if defined(_IBMR2) && !defined(USE_UCONTEXT)
externC void    e71_chgstack ( ten50_UContext   * newC , ten50_UContext    * oldC );
externC void    e71_first_task ( ten50_UContext * context );
#endif

#if defined(HP9)
externC void    e71_chgstack ( ten50_UContext   * newC , ten50_UContext    * oldC );
externC void    e71_make_core ( ten50_UContext  * context );
externC void    e71_first_task ( ten50_UContext * context );
#endif

#if defined(ALPHA)
externC void    e71_initstack ( ten50_UContext  * context , void (* func) () );
/* already defined in gen50.h: cxx does not like that... */
/*
    externC void e71_firststack ( ten50_UContext * context );

    externC void e71_chgstack ( ten50_UContext * newC, ten50_UContext * oldC );
 */
#endif

#endif  /* GEN71_H */
