/*!***************************************************************************
  module:       heo381.h

  responsible:  ThomasA

  special area: Kernel

  description:  Implementation of a general purpose allocator.

  copyright:    Copyrighted by SAP AG, 1999

  created:      Wed Jan 12 14:35:56  2000

  updated:      Thu Dec 30 11:39:45  1999


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/

#ifndef __HEO381_H
#define __HEO381_H

#include "gsp00.h"

#ifdef _DEBUG
#include <assert.h>
#endif

#define INLINE inline

#if _MSC_VER > 1100
#define USE_OPERATOR_DELETE_EO381 1
#endif  
#if defined(OSF1)
#define USE_OPERATOR_DELETE_EO381 1
#endif 

typedef teo00_Longuint teo381_Size;
typedef tsp00_Uint4    teo381_ChunkSize;

class  ceo381_Chunk;
class  ceo381_Heap;
struct TRawNextChunk;
typedef ceo381_Chunk* teo381_ChunkPtr;

class CFreeChunkLists 
{
  enum { NO_OF_FREELISTS = 128 };

  struct FreeListHeader {
    teo381_ChunkPtr m_next;
    teo381_ChunkPtr m_prev;
  };
  FreeListHeader m_freeLists[NO_OF_FREELISTS + 2];
public :
  CFreeChunkLists();
  INLINE teo381_ChunkPtr operator[] (int ix);
  INLINE static int GetFreeListIndex(teo381_ChunkSize sz);
  INLINE static int GetSmallFreeListIndex(teo381_ChunkSize size);
};

/*!
  class: ceo381_Heap

  description: Implementation of general purpose heap

    This class provides an implementation of a general purpose heap
    based on the ideas of Doug Lea. 
    
    The algorithm seems to be a good compromise between performance and space-conserving.
*/

class ceo381_Heap {
public :
  struct THeapInfo {
    teo381_Size heapUsed;
    teo381_Size allocated;
    tsp00_Uint4 extends;
  };

  class HeapIterator {
    friend class ceo381_Heap;
  public :
    HeapIterator();
    operator bool() const;
    void operator++();
    const teo381_ChunkPtr operator()() const;
  private :
    ceo381_Heap*    m_pHeap;
    TRawNextChunk*  m_currBlock;
    teo381_ChunkPtr m_currChunk;
  };
/*! 
	function: ceo381_Heap 

	description: Constructs the heap
*/
  ceo381_Heap(tsp00_Uint4 FirstMalloc_KB, tsp00_Uint4 Supplement_KB);
  ceo381_Heap(tsp00_Uint4 FirstMalloc_KB, tsp00_Uint4 Supplement_KB, tsp00_Uint4 MaxHeap_KB);
  virtual ~ceo381_Heap();
  void*                   allocate (teo381_ChunkSize);
  bool                    checkHeap();
  static teo381_ChunkSize chunkSize(void* p);
  void                    deallocate(void* mem);
  HeapIterator            first();
  void                    freeHeap();
  void                    getHeapInfo(THeapInfo&);
  void                    setHeapLimit(teo381_Size limit_KB);
  void*  operator new(size_t sz)  { return ::operator new(sz); }
  void   operator delete(void* p) { ::operator delete(p); }
  void*  operator new(size_t sz, void* p) { return p; }
private :
#if defined(USE_OPERATOR_DELETE_EO381)
  void operator delete(void* p, void*q) {}
#endif
  INLINE static teo381_ChunkSize AlignedSize    (teo381_ChunkSize req);
         void                    CheckConstraints();
  virtual void                   FreeMem(teo381_ChunkPtr);
  INLINE void                    FrontLink(teo381_ChunkPtr pChunk, teo381_ChunkSize chunkSize);
  virtual teo381_ChunkPtr        GetMem(size_t sz);
  INLINE unsigned long           Idx2BinBlock(int);
         bool                    MallocExtend(teo381_ChunkSize required);
  INLINE void                    MarkBinBlock(int);
  static bool                    IsSmallRequest (teo381_ChunkSize req);
#ifdef _DEBUG
  void CheckInUseChunk(teo381_ChunkPtr);
  void CheckMallocedChunk(teo381_ChunkPtr p, teo381_ChunkSize s);
  void CheckChunk(teo381_ChunkPtr);
  void CheckFreeChunk(teo381_ChunkPtr);
#endif
  char*              m_base;
  teo381_ChunkPtr    m_lastRemainder;
  TRawNextChunk*     m_first;
  tsp00_Uint4        m_firstSize;
  tsp00_Uint4        m_supplement;
  tsp00_Uint4        m_maxHeap;
  teo381_Size        m_used;
  teo381_Size        m_mallocedSize;
  tsp00_Uint4        m_extends;
  CFreeChunkLists    m_freeLists;
  tsp00_Uint4        m_binBlocks;
  friend class ceo381_Heap::HeapIterator;
  friend class ceo381_Chunk;
};

/*! endclass: ceo381_Heap */

#endif // __HEO381_H


