/*!***************************************************************************

  module      : heo670.h

  -------------------------------------------------------------------------

  responsible : JoergM, RaymondR

  special area: Stack Back Trace
  description : Interface that allows to produce a stack backtrace

  first created:2000-03-27  12:07
  last changed: 2000-03-27  14:03
  see also    : example.html ...

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2000



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/


#ifndef HEO670_H
#define HEO670_H



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "geo00.h"
#include "SAPDBCommon/SAPDB_Types.h"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/


/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/


/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*!---------------------------------------------------------------------------
  Declaration: eo670AddressInfo
  Desscription: Address information used to to find text symbol for given address
  
  The eo670AddressInfo is filled in the following sequence

  void *callArray[MAX_CALL_ARRAY];
  void *context;
  SAPDB_Int4 depth;
  SAPDB_Int4 loopIndex;

  depth = eo670_FillCallArray(MAX_CALL_ARRAY, callArray);

  if ( depth > 0 
    && eo670_GetAddressInfoProlog(&context) )
  {
    for ( loopIndex = 0; loopIndex < depth; loopIndex++ )
    {
        eo670AddressInfo localAddrInfo;

        if ( eo670_GetAddressInfo(callArray[loopIndex],
                                  &localAddrInfo,
                                  context) )
        {
            ShowAddressInfo(loopIndex, callArray[loopIndex], &localAddrInfo); see below

            eo670_FreeAddressInfo(&localAddrInfo);
        }
    }
    eo670_GetAddressInfoEpilog(context);
  }


  First fill the callArray. This can be done at any time before, but notice that symbols of unloaded
  dynamic loaded libraries maybe no longer available... (should not be a problem for SAPDB, which does
  no dynamic unload...).

  Now the symbol information is taken frame by a prologue and epilogue function.
  Each addressInfo is retrieved, and after usage freed. If you forget eo670_FreeAddressInfo you produce a
  significant memory leak....

  Showing the address info should use such a coding

  void ShowAddressInfo(SAPDB_Int4 loopIndex, void *pcAddr, eo670AddressInfo *addrInfo)
  {
    if ( addrInfo->offsetToSymbol < 0 )
    {
      printf("%d: %s - 0x%lx %s\n",
            loopIndex,
            addrInfo->moduleName,
            (SAPDB_Long)((char *)pcAddr-(char *)0),
            addrInfo->symbol);
    }
    else
    {
      printf("%d: %s - 0x%lx %s + 0x%x\n",
            loopIndex,
            addrInfo->moduleName,
            (SAPDB_Long)((char *)pcAddr-(char *)0), 
            (0 != *addrInfo->demangledSymbol) ? addrInfo->demangledSymbol : addrInfo->symbol, 
            addrInfo->offsetToSymbol );

      if ( addrInfo->lineNumber >= 0 )
      {
        printf(" %s - %s:%d\n",
            addrInfo->moduleName,
            addrInfo->sourceFileName,
            addrInfo->lineNumber );
      }
      else if ( 0 != *addrInfo->sourceFileName )
      {
        printf(" %s - %s\n",
            addrInfo->moduleName,
            addrInfo->sourceFileName );
      }
    }
  }
 */
typedef struct eo670AddressInfo_
{
    SAPDB_Char *symbol;          /* the undemangled symbol or the reason why it was not able to get it */
    SAPDB_Char *demangledSymbol; /* the demangled symbol or "" if no demangled symbol available        */
    SAPDB_Char *moduleName;      /* the module where the symbol was found or "" if not available       */
    SAPDB_Char *sourceFileName;  /* the source file name of the file defining the symbol or "" if NA   */
    SAPDB_Long offsetToSymbol;   /* the offset from symbol address to given pcAddr or -1 if NA         */
    SAPDB_Long lineNumber;       /* the line number inside the source file or -1 if NA                 */
} eo670AddressInfo;

/*!---------------------------------------------------------------------------
  Function: eo670_CTraceStack
  Description: Create a stack backtrace on stdout

  This call is usefull for debugging 'crashing' code that already detected a severe problem.
  It does not use 'normal' memory allocation and tries to prevent anything else, that could
  lead to inflictions with other modules.
  The stack backtrace is normally visible in 'knldiag' if compiled for kernel
  The amount of output visible depends on optimization level used. At least module overlapping
  call sequences are visible, that means all inner module calls must not necessarily be visibile.
  Be carefull to keep this in mind while interpreting the stack backtrace!
 */
externC void eo670_CTraceStack(void);
externC void vtracestack(void); /* alias for PASCAL sources */

/*!---------------------------------------------------------------------------
  Function: eo670_FillCallArray(int level, void **callArray)
  Description: Fill the given array with at most level trace back entries.

  Supported on most platforms (except HPUX 32 where always 0 is returned)

  Arguments: level [in] the number of level to fill
             callArray [out] the array to fill
  Return value: The actual number of entries filled
 */
externC SAPDB_Int4 eo670_FillCallArray(SAPDB_Int4 level,
                                       void **callArray);

/*!---------------------------------------------------------------------------
  Function: eo670_GetAddressInfo
  Description: Get mapping information about a given PC address

  Supported on most platforms (except on HPUX 32 + 64 where a dummy address info is filled)

  Arguments: pcAddr [in] a hopefully valid program counter
             addrInfo [out] structure to be filled with address information
             context [in] context received by eo670_GetAddressInfoProlog
  Return value: true if mapping returns valid data, false if not
 */
externC SAPDB_Bool eo670_GetAddressInfo(void *pcAddr,
                                        eo670AddressInfo *addrInfo,
                                        void *context);

/*!---------------------------------------------------------------------------
  Function: eo670_GetAddressInfoEpilog
  Description: finishes a sequences of eo670_GetAddressInfo calls and
               release all resources hold by eo670_GetAddressInfoProlog.
  Arguments: 
    context [in] pointer returned by eo670_GetAddressInfoProlog
  Return value: none
 */

externC void eo670_GetAddressInfoEpilog (void* context);  

/*!---------------------------------------------------------------------------
  Function: eo670_GetAddressInfoProlog
  Description: prepares for following calls of eo670_GetAddressInfo
  Arguments: 
    context [out] pointer to context required for eo670_GetAddressInfo.
                  may be null on several plattforms
  Return value: true if mapping returns valid data, false if not
 */
externC SAPDB_Bool eo670_GetAddressInfoProlog (void** context);  

/*!---------------------------------------------------------------------------
  Function: eo670_FreeAddressInfo
  Description: Free mapping information

  Necessary call to prevent memory leaks

  Arguments: addrInfo [in] structure filled by a previous call to eo670_GetAddressInfo
 */
externC void eo670_FreeAddressInfo(eo670AddressInfo *addrInfo);

/*!---------------------------------------------------------------------------
  Function: eo670_CTraceContextStack
  Description: Create a stack backtrace using given context

  This call is usefull for debugging 'crashing' code that already detected a severe problem.
  It does not use 'normal' memory allocation and tries to prevent anything else, that could
  lead to inflictions with other modules.
  The stack backtrace is normally visible in 'knldiag' if compiled for kernel
  The amount of output visible depends on optimization level used. At least module overlapping
  call sequences are visible, that means all inner module calls must not necessarily be visibile.
  Be carefull to keep this in mind while interpreting the stack backtrace!

  The given context is retrieved as third argument of a sigaction handler on UNIX and as
  exception context on NT.

  If 0 the current context is retrieved.

  Arguments: exceptionContext [in] a CONTEXT on NT or a struct sigcontext * on UNIX
 */
externC void eo670_CTraceContextStack( void* exceptionContext );

#ifdef WIN32
externC bool eo670_WriteDump ( bool WriteFullMemoryDump,
                               bool IfDumpFileExists );
#endif 

/*!---------------------------------------------------------------------------
  Function: eo670WriteString
  Description: Write a string into stack back trace output channel
  Arguments: str [in] string to write
 */
externC void eo670WriteString(const char *str);

#endif  /* HEO670_H */
