/*!**************************************************************************

  module       : heo76.h

  -------------------------------------------------------------------------

  Responsible : Joerg Mensing - D027923

  Special area: Spinlocks
  Description : System Independend RTE_Spinlock and RTE_LockedScope Interface

  The RTE_Spinlock is a class that support Lock() and Unlock() of a spinlock.
  The RTE_LockedScope is a class that is used to automatically protect a scope.

  A spinlock is a special synchronisation object used to synchronize threads
  or processes. A spinlock can have only to states - locked or unlocked. 
  If it is already locked, a thread or processes, that tries to
  lock it again will spin (or poll) until the locking thread or process
  relaeses the hook. Therefore it is important to use this construct only
  in situtation, where the 'locker' is able to guarantee a short time for
  locking.

  The spinlock resource needs a synchronisation object in memory called a
  'teo76_Lock'. Size of this object may vary on different operating systems,
  since it is highly processor dependend. The teo76_Lock may already be
  created by another process or thread. To access such an external lock,
  the consturctor can be provided with an optionaly external lock pointer.
  
  Backported to Version 7.2
 
  Renamed RTESync_Spinlock    to  ceo76Spinlock
  Renamed RTESync_LockedScope to  ceo76LockedScope
  Renamed RTE_ISystem::member to  eo76member

  first created:2000-02-29  9:27
  last changed: 2001-01-02  12:42

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2000



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/


#ifndef HEO76_H
#define HEO76_H

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gsp00.h"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/

/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/* PTS 1108524 */
/*!
  Declaration: teo76spinlockStatistic
  Description: SpinLoop instance statistic record

  The statistic contains the number of locks and collisions. For each collision the
  number of spin loops (if any) and the number of yield loops are counted and the
  total number as well as the maimum number is kept. All these values are written
  after the lock succeeded, so they are consistent, even if another task is already
  spinning. The currentLoops count is the only value that is 'dirty written', that
  means each time a spinning tasks tests the lock, it updates this counter to its
  individual value. This value is consistent only for the task.
 */
struct teo76spinlockStatistic {
  tsp00_Int8 locks;           /* total number of locks */
  tsp00_Int8 collisions;      /* total number of collisions */
  tsp00_Int4 totalSpinLoops;  /* total number of spin loops needed */
  tsp00_Int4 totalYieldLoops; /* total number of yield loops needed */
  tsp00_Int4 maxSpinLoops;    /* maximum number of spin loops needed */
  tsp00_Int4 maxYieldLoops;   /* maximum number of yield loops needed */
  tsp00_Int4 currentLoops;    /* current number of loops needed (spin loops or yield loops) */
};

#if defined (__cplusplus)

/*--------------------------------------------------------------------------------------------
  The following system depend typedefinitions are needed to support backward compatibility
  with Interlocks created i.e. on shared memory segments allocated by older clients (<7.3...)
 
  The following typedef should not be used directly with one exception: For communication with
  older clients via shared memory segments. In such a case this type is used to describe the
  lock.
  New implementation should use the classes defined in RTESync_Spinlock.hpp
  --------------------------------------------------------------------------------------------*/
#if defined(_WIN32)
  typedef long       teo76_Lock;
#endif

#if defined(ALPHA)
  typedef tsp00_Longint teo76_Lock[1]; /* atomic_p is a pointer to a long on ALPHA */
#endif

#if defined(AIX) || defined(SUN) || defined(LINUX) 
  typedef tsp00_Int4 teo76_Lock[1]; /* atomic_p is a pointer to a 32bit word on AIX */
#endif

#if defined(PA20W) || defined(PA11)
  /* Warning: If HP uses new architecture, cacheline size may change... */
# define CACHELINE_SIZE 16          /* Spinlock must be in seperate chacheline to */
                                    /* guarantie read/write consistency */
  /* 32bit words are ok in 64bit mode too */
  typedef tsp00_Int4 teo76_Lock[CACHELINE_SIZE*2];
#endif /* PA20W || PA11 */

#if defined(NMP)
#  include <ulocks.h>
  typedef spinlock_t teo76_Lock;
#endif
/*-----------------------------------------------------------------------------------------
  End of this ugly teo76_Lock...
  ----------------------------------------------------------------------------------------- */

/*!
  Function: InitLock
  description: Initialization of Atomic lock area

  Arguments: lock [ in ] The lock space
 */
void eo76InitLock( teo76_Lock &lock );

/*!
  Function: TestAndLock
  description: Atomic test and lock

  Arguments: lock [ in ] The lock space
  Return value: true if the lock space now locked by caller, false if not
 */
bool eo76TestAndLock( teo76_Lock &lock );

/*!
  Function: Unlock
  description: Unlock Atomic lock area

  Arguments: lock [ in ] The lock space
 */
void eo76Unlock( teo76_Lock &lock );

/*!
  Function: GiveUpTimeSlice
  Descriptions: Give up the time slice

  In multithreaded environment this allows the scheduler to select the next thread in the
  current process.

  In singlethreaded environment this allows the system to schedule the next process.

  It makes no difference weather this is called within a task context (in term of internal tasking)
  or within a 'normal' context.
 */
void eo76GiveUpTimeSlice();

/*!---------------------------------------------------------------------------
  Class:       ceo76Spinlock
  Description: This class implements the OS dependend spinlock hook

  This class allows to Lock() and Unlock() a spinlock. The spinlock may be
  accessed via internal or external teo76_Lock. The constructor decides, if the
  Unlock() is called during destruction, or the spinlock is left in its last
  state. The RTE_ISystem implementation decides how to yield, when a spinlock
  cannot be locked on first attempt.
 */
class ceo76Spinlock
{
public:
  /*!
    Function: ceo76Spinlock
    Description: Construct a spinlock

    The constructor has several options. The teo76_Lock can be internal or external.
    An external teo76_Lock is i.e. usefull to synchronize via an external provided
    shared memory segment that already contains an teo76_Lock. If external, it may
    not be usefull to initialze the teo76_Lock into a defined state. Therefore this
    is optional. For the same reason the teo76_Lock may not be necessarily unlock 
    during destruction of this object. So this is also optional.

    If the spinlock is internal, it cannot be exported!

    If an array of spinlocks is needed, it can only have default behaviour. An array
    of external spinlocks is not supported. If you need such a thing use a loop
    over individual ceo76_SpinlockHock constructors...

    The internal statistics can be enabled via a call to SetStatisticRecord() later on.
 
    Arguments: ExternalHook [in] If non NULL used instead of internal storage
               DoInitialize [in] If set this flags forces initialization of teo76_Lock
               DoUnlock     [in] If set this flags forces unlock of teo76_Lock in destructor
               DoStatistics [in] If set this flags forces usage of internal statics record
   */
  ceo76Spinlock(teo76_Lock * ExternalHook = (teo76_Lock *)0,
                bool DoInitialize=true,
                bool DoUnlock=true)
              : m_DoUnlock(DoUnlock), m_Status(0)
  {
    m_Spinlock = ( ExternalHook ? ExternalHook : &m_SpinlockSpace );

    if ( DoInitialize )
    {
      eo76InitLock(*m_Spinlock);
    }
  }

  /*!
    Function: ~ceo76Spinlock
    Description: Destructor should unlock the Spinlock space if indicated by flag m_DoUnlock
   */
  ~ceo76Spinlock()
  {
    if ( m_DoUnlock ) eo76Unlock(*m_Spinlock);
  }

  /*!
    Function: GetStatistic
    Description: Retrieve the statistic record pointer for this spinlock

    NOTE that this routine does not lock the spinlock. This should be done
    from outside if necessary !!!!

    Return value: current statistic record pointer ( NULL if currently no statistics )
   */
  teo76spinlockStatistic * GetStatisticRecord() const
  {
    return m_Status;
  }

  /*!
    Function: SetStatisticRecord
    Description: Set the statistic record pointer for this spinlock

    This call is needed to enable statistics. The newStatisticRecord is filled. The space
    should be on heap or shared memory...

    NOTE that this routine does not lock the spinlock. This should be done
    from outside if necessary !!!!

    Arguments: newStatisticRecord [in] New statistic record (or NULL if no statistic wanted)
    Return value: replaced statistic record pointer
   */
  teo76spinlockStatistic * SetStatisticRecord( teo76spinlockStatistic * newStatisticRecord )
  {
    teo76spinlockStatistic * currentStatisticRecord = m_Status;

    m_Status = newStatisticRecord;

    return currentStatisticRecord;
  }

  /*!
   Function: Lock
   Description: Lock the Spinlock space unconditional

   After return from this call, the spinlock space is locked. If already
   locked the GiveUpTimeSlice() function of RTE_ISystem() is called in a 
   tight loop.
   
   Arguments: SpinLoops [in] Number of loops to take before each GiveUpTimeSlice() call
   */
  void Lock(tsp00_Int4 SpinLoops=0)
  {
    if ( m_Status )
    {
      /* PTS 1108524 */
      if ( eo76TestAndLock(*m_Spinlock) )
      {
        tsp00_Int4 currentLoops = 1;
        tsp00_Int4 i;
        /* Loop before sleeping */
        for ( i = 1; i <= SpinLoops; i++ )
        {
          if ( !eo76TestAndLock(*m_Spinlock) )
          {
            /* Now in locked state... */
            m_Status->currentLoops = 0;

            if ( i > m_Status->maxSpinLoops )
            {
              m_Status->maxSpinLoops = i;
            }
            m_Status->totalSpinLoops += i;
            ++m_Status->collisions;
            return;
          }
          else
          {
            /* dirty write, if more then one Spinner. But at least each */
            /* Spinner writes its correct value... */
            m_Status->currentLoops = currentLoops;
            ++currentLoops;
          }
        }

        /* SpinLoops did not help us, so we do YieldLoops now */

        do {
          eo76GiveUpTimeSlice();

          /* dirty write, if more then one Spinner. But at least each */
          /* Spinner writes its correct value... */
          m_Status->currentLoops = currentLoops;
          ++currentLoops;
        } while ( eo76TestAndLock(*m_Spinlock) );

        m_Status->currentLoops = 0;

        if ( (currentLoops - SpinLoops) > m_Status->maxYieldLoops )
        {
          m_Status->maxYieldLoops = currentLoops - SpinLoops;
        }

        if ( SpinLoops > m_Status->maxSpinLoops )
        {
          m_Status->maxSpinLoops = SpinLoops;
        }

        m_Status->totalYieldLoops += i;

        /* Now in locked state... */
        ++m_Status->collisions;
      }

      ++m_Status->locks;
    }
    else
    {
      /* No Statistic record... */

      if ( eo76TestAndLock(*m_Spinlock) )
      {
        /* Loop before sleeping */
        for ( tsp00_Int4 i = 1; i <= SpinLoops; i++ )
        {
          if ( !eo76TestAndLock(*m_Spinlock) )
          {
            /* Now in locks state... */
            return;
          }
        }

        /* SpinLoops did not help us, so we do YieldLoops now */

        do {
          eo76GiveUpTimeSlice();
        } while ( eo76TestAndLock(*m_Spinlock) );

      }
    }
  }

  /*!
    Function: Unlock
    Description: Unlocks the Spinlock space unconditional
   */
  void Unlock() { eo76Unlock(*m_Spinlock); }
  
private:
  teo76_Lock  m_SpinlockSpace; /* Spinlock space if no external space given */
  teo76_Lock *m_Spinlock;      /* Pointer to used spinlock space */
  bool        m_DoUnlock;      /* Flag for destructor */
  teo76spinlockStatistic *m_Status; /* Spinloop statistic record */
};
/*! EndClass: ceo76Spinlock */

/*!---------------------------------------------------------------------------
  Class:        ceo76LockedScope
  Description:  Class for spinlocked scope
  
  This class allows to synchronize access to regions of code that are not 
  allowed to be entered by several threads of context at a time. 

  NOTE This interface does busy waiting until lock is granted! It is more
  than other classes only to be used to protect SHORT sections of code or
  otherwise PERFORMANCE WILL DROP DRAMATICALLY.

  VERY IMPORTANT!!! This Interface is not reentrant! If you create a second
  ceo76LockedScope on the same ceo76Spinlock you produce a non detectable
  deadlock!!! The ceo76LockedScope does not provide any information who did the
  lock, therefore it cannot check on reentrancy.

  If an ceo76LockedScope is created on the heap, it locks the associated region
  until deleted. Placement new / calling of destructor can replace 'Lock' and 
  'Unlock'. But multithreaded access is not synchronized in such a case! If such
  a heap class is used, it should be synchronized from outside.

  Example:
  // NOTE colon is replced by .. not to confuse cppDoc only

  // Primitive StackObject for stacking tsp00_Int4 ...
  class RTE_StackObject {
  public..
    RTE_StackObject(tsp00_Int4 val=0) .. m_val(val), m_Next(0) {}

    tsp00_Int4 GetVal()               { return m_val; }
    tsp00_Int4 SetVal(tsp00_Int4 val) { m_val = val; }

    RTE_StackObject * GetNext()                         { return m_Next; }
    void              SetNext(RTE_StackObject *theNext) { m_Next = theNext; }

  private..
    tsp00_Int4 m_val;
    RTE_StackObject *m_Next;
  };

  // Sample code for implementation of simple stack

  // TRICK
  // Using ceo76Spinlock as base class allows to Lock and Unlock a RTE_Stack!

  class RTE_Stack .. public ceo76Spinlock {

  private..
    RTE_StackObject *m_TopOfStack;

  public..

    RTE_Stack().. m_TopOfStack(0) { EndOfStack = (RTE_StackObject *)-1; }

    ~RTE_Stack() { } // The spinlock is destroyed automatically

    void Push(RTE_StackObject & Object) {

      // The whole member function scope is locked
      ceo76LockedScope Lock(*this);

      Object.SetNext(m_TopOfStack);
      m_TopOfStack = &Object;
    }

    RTE_StackObject & Pop() {

      RTE_StackObject * ReturnObject;

      { // Only this scope is locked

        ceo76LockedScope Lock(*this);
  
        ReturnObject = m_TopOfStack;

        if ( m_TopOfStack )
        {
          m_TopOfStack = m_TopOfStack->GetNext();
        }
      } // after this scope unlocked action can continue

      return ( ReturnObject ? *EndOfStack .. *ReturnObject );
    }

    RTE_StackObject *EndOfStack;

  };
*/
class ceo76LockedScope
{
public:
/*!
  Function: ceo76LockedScope
  
  Description: Constructor locks the spinlock.
  Arguments: Hook [in] The Spinlock hook used
             SpinLoops [in] Number of Spinlock loops before Yield() call is done
 */
  ceo76LockedScope(ceo76Spinlock & Hook, tsp00_Int4 SpinLoops=0) : m_Hook(Hook)
  {
    m_Hook.Lock(SpinLoops);
  }

/*! 
  Function: ~ceo76LockedScope
  
  Description: Destructor gives the lock free
 */
  ~ceo76LockedScope()
  {
    m_Hook.Unlock(); 
  }

private:
  ceo76Spinlock & m_Hook;
};
/*! EndClass:        ceo76LockedScope */

#endif /* defined(__cplusplus) */

#endif  /* HEO76_H */
