/****************************************************************************

  module      : veo15.cpp
  
  -------------------------------------------------------------------------
    
  responsible : JoergM
      
  special area: Message Output Module (for buffered Message I/O)
  description : Contains Message buffering for use inside the kernel
  PTS 1103824
        
  last changed: 2001-01-03  14:32
  see also    : example.html ...
          
  -------------------------------------------------------------------------
            
  copyright:    Copyright by SAP AG, 1999
              


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/


/*===========================================================================*
*  INCLUDES                                                                 *
*===========================================================================*/

#include "gsp00.h"
#include "heo07.h"
#include "heo15.h"
#include "geo007_1.h"
#include "geo007_2.h"
#include "geo57.h"
#include "geo001.h"
#include "hsp77.h" // sp77sprintf()
&if $OS in [ WIN32 ]
#include "vos00kc.h"
&else
#include "hen40.h" // sqlerrs()
&endif
/*===========================================================================*
*  DEFINES                                                                  *
*===========================================================================*/

#ifdef TEST
#include <malloc.h>
#undef ALLOC_MEM_EO57
#define ALLOC_MEM_EO57(_pBuf,_s) \
    ( (*_pBuf = malloc(_s)) != NULL ? NO_ERROR_EO001 : ERROR_NOT_ENOUGH_MEMORY_EO001 )
#undef FREE_MEM_EO57
#define FREE_MEM_EO57(_Buf) free(_Buf)

int DbgPrint = 0;
#define TEST_PRINT(fmt) printf(fmt);
#define TEST_PRINT1(fmt,arg1) printf(fmt,arg1);
#define TEST_PRINT3(fmt,arg1,arg2,arg3) printf(fmt,arg1,arg2,arg3);
#define TEST_DPRINT3(fmt,arg1,arg2,arg3) if ( DbgPrint ) printf(fmt,arg1,arg2,arg3);
#else
#define TEST_PRINT(fmt)
#define TEST_PRINT1(fmt,arg1)
#define TEST_PRINT3(fmt,arg1,arg2,arg3)
#define TEST_DPRINT3(fmt,arg1,arg2,arg3)
#endif /* TEST */

#define END_LINE_EO15 "----- Last Line Written ----- Cycle:" // followed by max 12 digits
#define CYCLE_COUNT_SIZE_EO15                              12 // space for 12 digits
#define CLR_LINE_EO15 "===== Last Line in File. More Information above!" \
"                                                "
#define FIRST_LINE_EO15 "===== Current Cycle:"
#define CRASH_MARK_EO15 "===== WARNING: Crash may reproduce information"

/* Size used for eo15_CopyFile */
# define COPY_SIZE_EO15 ((16*1024)/sizeof(long))

/*===========================================================================*
*  MACROS                                                                   *
*===========================================================================*/

#if !defined(_WIN32)
extern "C" void vabort(bool);
#  define ABORT()                    vabort(true)
#endif /* _WIN32 */

#define WORKER_STACK_EO15 (64*1024)
/*===========================================================================*
*  LOCAL VARIABLES                                                          *
*===========================================================================*/
#if defined(_WIN32)
#define USE_DYNAMIC_MESSAGE_BUFFER
#else
#undef USE_DYNAMIC_MESSAGE_BUFFER
#endif

#ifdef USE_DYNAMIC_MESSAGE_BUFFER

  static teo07_Mutex                eo15_BufferMapMutex;
  static teo15MessageBufferSpace ** eo15_BufferMap; // Each entry in this list is filled on demand
  static tsp00_Int4                 eo15_BufferMapSize;
#  define MAP_INCREASE_AMOUNT_EO15 64 /* If eo15_BuffsListMapSize must ever get increased, use this size */
#  define MESSAGE_BUFFER_PTR(_i)     (eo15_BufferMap[_i])
#  define NUMBER_OF_MESSAGE_BUFFERS   eo15_BufferMapSize

#else

// Due to usage of a shared memory segment, the number of buffers is fixed...
  static teo15MessageBufferSpace *  eo15_BufferArray;
#  define MESSAGE_BUFFER_PTR(_i)    (eo15_BufferArray+_i)
#  define NUMBER_OF_MESSAGE_BUFFERS   eo15_NumberOfBuffers

#endif /* USE_DYNAMIC_MESSAGE_BUFFER */

static tsp00_Int4                 eo15_NumberOfBuffers; // Maximum number of buffers 
static teo07_Thread               eo15_WorkerThread;

/*===========================================================================*
*  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
*===========================================================================*/
/*---------------------------------------------------------------------------*/
static void eo15_ShowLastOsError( const tsp00_Char *    Routine,
                                  const tsp00_Char *    Filename );
/*---------------------------------------------------------------------------*/
static void eo15_MapAndLockMessageFile( const tsp00_Char *    Filename,
                                       tsp00_Int4           &FileIndex,
                                       tsp00_VfReturn_Param &ReturnStatus );
/*---------------------------------------------------------------------------*/
static bool eo15_CheckFilename(const tsp00_Char *Filename);
/*---------------------------------------------------------------------------*/
#define LAST_ERROR_MESSAGE_SIZE_EO15 64
typedef char teo15_LastErrorMessage[LAST_ERROR_MESSAGE_SIZE_EO15];

/*---------------------------------------------------------------------------*/

class ceo15_LockMutex
{
public:
  ceo15_LockMutex(teo07_Mutex &Mutex) : m_Mutex(Mutex) { sqlbeginmutex(&Mutex); }
  ~ceo15_LockMutex() { sqlendmutex(&m_Mutex); }
private:
  teo07_Mutex &m_Mutex;
};

class ceo15_MessageFile
{
private:
  teo15MessageBufferSpace &m_Space;
  
public:
  static bool eo15CheckIndex(const tsp00_Int4 FileIndex)
  {
#ifdef USE_DYNAMIC_MESSAGE_BUFFER
    return ( (FileIndex > 0)
          && (FileIndex <= NUMBER_OF_MESSAGE_BUFFERS)
          && (MESSAGE_BUFFER_PTR(FileIndex-1) != NULL) );
#else
    return ( (FileIndex > 0) 
          && (FileIndex <= NUMBER_OF_MESSAGE_BUFFERS) );
#endif
  }
  
  ceo15_MessageFile(teo15MessageBufferSpace &Space)
    : m_Space(Space)
  {}
  
  void eo15Open( const char *FileName,
    teo15_Position MaxFileSize,
    tsp00_VfReturn_Param &ReturnStatus );
  
  void eo15Append( const char *FileName,
    tsp00_VfReturn_Param &ReturnStatus );
  
#if !defined(_WIN32)
  // This is not not called reguarly... Only if reopen is needed after crash.
  // In such a case the old m_HostFileNo is invalid. We override it simply.
  void eo15Reopen( tsp00_VfReturn_Param &ReturnStatus )
  {
    if ( eo15IsUsed() )
    {
      // Reset refrence count
      m_Space.m_Used = 1;
#ifdef TEST
      // Test program must simulate 'close' done by the crash
      eo15_Close ( m_Space.m_HostFileNo, ReturnStatus );
#endif /* TEST */
      eo15_Open ( m_Space.m_HostFileNo,
        m_Space.m_Filename.asCharp(),
        eo15_WriteOnly,
        false,
        DEFAULT_SECURITY_EO15,
        ReturnStatus );
      if ( vf_ok != ReturnStatus )
      {
        eo15_ShowLastOsError("ReOpen", m_Space.m_Filename.asCharp() );
      }
    }
  }
#endif /* WIN32 */
  
  // Writing tries to store in current buffer. If no more space, first it is checked
  // that the second buffer is not in use already. But if , it is forced to be flushed.
  // Then the m_BytesStored counter is copied to m_BytesToFlush and the current buffer
  // is switched. The new buffer is used to store the message.
  void eo15Write(const char *Message, tsp00_VfReturn_Param &ReturnStatus );
  /* Flush both buffers */
  void eo15Sync( tsp00_VfReturn_Param &ReturnStatus )
  {
    if ( eo15IsUsed() )
    {
TEST_PRINT("Flush first\n")
      eo15Flush( ReturnStatus);
      m_Space.m_BytesToFlush  = m_Space.m_BytesStored;
      m_Space.m_CurrentBuffer = 1 - m_Space.m_CurrentBuffer;
      m_Space.m_BytesStored   = 0;
TEST_PRINT("Flush second\n")
      eo15Flush( ReturnStatus);
    }
  }
  
  void eo15AddRef()
  {
    ++m_Space.m_Used;
  }
  
  bool eo15IsUsed() const
  {
    return ( m_Space.m_Used > 0 );
  }
  
  bool eo15Release()
  {
    --m_Space.m_Used;
    if ( m_Space.m_Used < 0 )
    {
      m_Space.m_Used = 0;
    }
    return (m_Space.m_Used == 0);
  }
  
  void eo15Close( tsp00_VfReturn_Param &ReturnStatus )
  {
    if ( eo15IsUsed() )
    {
      eo15Sync( ReturnStatus );
    }

    if ( eo15Release() )
    {
      eo15_Close( m_Space.m_HostFileNo , ReturnStatus );
      if ( vf_ok != ReturnStatus )
      {
        eo15_ShowLastOsError("Close", m_Space.m_Filename.asCharp() );
      }
    }
  }
  
  void eo15Copy(const char *CopyFileName, tsp00_VfReturn_Param &ReturnStatus)
  {
    eo15Sync(ReturnStatus);
    eo15_CopyFile( m_Space.m_Filename.asCharp(), CopyFileName, ReturnStatus );
  }
  
  void eo15Flush( tsp00_VfReturn_Param &ReturnStatus );
  
  bool eo15Lock( tsp00_VfReturn_Param &ReturnStatus );
  
  void eo15Unlock();
  
  bool eo15CompareFilename(const char *Filename) const
  {
    return ( eo15IsUsed() 
      && (strcmp(m_Space.m_Filename.asCharp(), Filename) == 0) );
  }
  
  void eo15SetTerminating( )
  {
    m_Space.m_Terminating++;
  }
};

/*---------------------------------------------------------------------------*/

void ceo15_MessageFile::eo15Open( const char *FileName,
                                 teo15_Position MaxFileSize,
                                 tsp00_VfReturn_Param &ReturnStatus )
{
  ROUTINE_DBG_MEO60 (_T("ceo15_MessageFile::eo15Open"));

  if ( MaxFileSize != UNLIMITED_FILE_SIZE_EO15 )
  {
    m_Space.m_MaxFileSize = 
      ALIGN_EO00(MaxFileSize, MESSAGE_FILE_PAGE_MXEO15) + strlen(CLR_LINE_EO15);
  }
  else
  {
    m_Space.m_MaxFileSize = UNLIMITED_FILE_SIZE_EO15;
  }
  
  eo15_Open (  m_Space.m_HostFileNo,
    FileName,
    eo15_WriteOnly,
    true,
    DEFAULT_SECURITY_EO15,
    ReturnStatus );
  if ( vf_ok == ReturnStatus )
  {
    strcpy(&(m_Space.m_Filename[0]), FileName);
    eo15AddRef();
    m_Space.m_CurrentBuffer = 0;
    m_Space.m_BytesStored = 0;
    m_Space.m_BytesToFlush = 0;
    m_Space.m_NextWritePosition = 0;
    m_Space.m_CycleCounter = 0;
    m_Space.m_TotalBytesWritten = 0;
  }
  else
  {
    eo15_ShowLastOsError("Open", FileName);
  }
}

/*---------------------------------------------------------------------------*/
/* PTS 1105164 */
//
// Append open does NOT uses POSIX Append mode, but seeks to end of file instead 
//
void ceo15_MessageFile::eo15Append( const char *FileName,
                                   tsp00_VfReturn_Param &ReturnStatus )
{
  ROUTINE_DBG_MEO60 (_T("ceo15_MessageFile::eo15Append"));

  m_Space.m_MaxFileSize = UNLIMITED_FILE_SIZE_EO15;
  
  bool IsDirectory;
  bool DoesExist;

  eo15CheckExist( FileName,
    DoesExist,
    IsDirectory,
    ReturnStatus );
  if ( vf_ok == ReturnStatus )
  {
    if ( IsDirectory )
    {
      MSGD(( ERR_TRACEIO_OS_ERROR, "Append", FileName, "found directory" ));

  TEST_PRINT("eo15CheckExist found directory\n")  
      ReturnStatus = vf_notok;
    }
  }
  TEST_PRINT("eo15CheckExist returned\n")  

  if ( vf_ok == ReturnStatus )
  {
  TEST_PRINT("eo15CheckExist found file\n")  
    eo15_Open (  m_Space.m_HostFileNo,
      FileName,
      eo15_ReadWrite,
      DoesExist ? false : true, // create only if not exists
      DEFAULT_SECURITY_EO15,
      ReturnStatus );

    if ( vf_ok == ReturnStatus )
    {
  TEST_PRINT("eo15_Open succeeded\n") 
      teo15_Position EndPosition;
      
      eo15_Seek( m_Space.m_HostFileNo,
        (teo15_Position)0,
        eo15_SeekEnd,
        EndPosition,
        ReturnStatus );
      
      if ( ReturnStatus == vf_ok )
      {
  TEST_PRINT("eo15_Seek succeeded\n")  
        strcpy(&(m_Space.m_Filename[0]), FileName);
        eo15AddRef();
        m_Space.m_CurrentBuffer = 0;
        m_Space.m_BytesStored = 0;
        m_Space.m_BytesToFlush = 0;
        m_Space.m_NextWritePosition = EndPosition;
        m_Space.m_CycleCounter = 0;
        m_Space.m_TotalBytesWritten = 0;
      }
      else
      {
        eo15_ShowLastOsError("Append(Seek)", FileName);
      }
    }
    else
    {
      eo15_ShowLastOsError("Append(Open)", FileName);
    }
  }
}

/*===========================================================================*
*  PUBLIC FUNCTIONS (CODE)                                                  *
*===========================================================================*/
extern "C"
{
  void *eo15DumpMessageFile( void * );
}
/*---------------------------------------------------------------------------*/
//
// Initialize the message buffer space due the number of preconfigured buffers
//
externC void eo15InitMessageBufferSpace(const tsp00_Int4 NumberOfBuffers,
                                        teo15MessageBufferSpace *BufferArray)
{
  ROUTINE_DBG_MEO60 (_T("eo15InitMessageBufferSpace"));
/* PTS 1106793 */

  MSGD(( INFO_TRACEIO_MAX_MESSAGE_FILES, NumberOfBuffers ));

  /* PTS 1109715 */
  if ( NumberOfBuffers <= 0 )
  {
#ifdef USE_DYNAMIC_MESSAGE_BUFFER
    eo15_BufferMap = NULL;
    eo15_BufferMapSize = 0;
#else
    eo15_BufferArray     = NULL;
#endif
    eo15_NumberOfBuffers = 0;
    MSGD(( INFO_TRACEIO_DISABLED ));
    return;
  }

#ifdef USE_DYNAMIC_MESSAGE_BUFFER
  sqlcreatemutex( &eo15_BufferMapMutex );

  if ( NO_ERROR_EO001 != ALLOC_MEM_EO57 ((void **)&eo15_BufferMap, 
                    sizeof(teo15MessageBufferSpace *) * MAP_INCREASE_AMOUNT_EO15 ) )
  {
    eo15_BufferMap = NULL;
    eo15_BufferMapSize = 0;
    eo15_NumberOfBuffers = 0;
    MSGD(( INFO_TRACEIO_DISABLED_NOMEM ));
    return;
  }
  eo15_BufferMapSize = MAP_INCREASE_AMOUNT_EO15;
  eo15_NumberOfBuffers = NumberOfBuffers;
  memset(eo15_BufferMap, 0, sizeof(teo15MessageBufferSpace *) * MAP_INCREASE_AMOUNT_EO15);
#else

  {
    int i;

    eo15_BufferArray     = BufferArray;
    eo15_NumberOfBuffers = NumberOfBuffers;
  
    for ( i = 0; i < eo15_NumberOfBuffers; i++ )
    {
      eo15_BufferArray[i].m_Used = 0;
      eo15_BufferArray[i].m_Terminating = 0;
      eo15_BufferArray[i].m_BusyCount = 0;
      sqlcreatemutex( &(eo15_BufferArray[i].m_Mutex) );
    }
  }

#endif
  
  tsp00_ErrTextc  ThreadErrText;
  teo07_ThreadErr ThreadOk = THR_OK_EO07;

  sqlbeginthread( WORKER_STACK_EO15,
    eo15DumpMessageFile,
    NULL,
    THR_CREATE_SUSPENDED_EO07,
    &eo15_WorkerThread,
    ThreadErrText,
    &ThreadOk);
  if ( ThreadOk != THR_OK_EO07 )
  {
	MSGD(( ERR_CREATING_THREAD, &ThreadErrText[0], ThreadOk));	  /* PTS 1112679 */
    ABORT();
    return;
  }

  sqlresumethread( eo15_WorkerThread, ThreadErrText, &ThreadOk );
  if ( ThreadOk != THR_OK_EO07 )
  {
	MSGD(( ERR_RESUMING_THREAD, &ThreadErrText[0], ThreadOk ));	  /* PTS 1112679 */
    ABORT();
    return;
  }
    
  MSGD(( INFO_TRACEIO_ENABLED ));
}
/*---------------------------------------------------------------------------*/
externC void eo15CompleteMessageFiles( const tsp00_Int4, teo15MessageBufferSpace *)
{
  ROUTINE_DBG_MEO60 (_T("eo15CompleteMessageFiles"));

  if ( eo15_NumberOfBuffers <= 0 )
  {
    return;
  }

  //
  // Called after crash of kernel! Ignores any locking!!!!
  //
  tsp00_VfReturn_Param ReturnStatus; 

#ifdef USE_DYNAMIC_MESSAGE_BUFFER
  ceo15_LockMutex Lock( eo15_BufferMapMutex );
#endif

  teo15MessageBufferSpace *ActiveBuffer;
  for ( int i = 0; (i < NUMBER_OF_MESSAGE_BUFFERS) 
               && ((ActiveBuffer = MESSAGE_BUFFER_PTR(i)) != NULL) ; i++ )
  {
    ceo15_MessageFile TheMessageFile( *ActiveBuffer );
    
    TheMessageFile.eo15SetTerminating();
#if !defined(_WIN32)
    // For WIN32 reopen is never needed...
    TheMessageFile.eo15Reopen(ReturnStatus);
#endif
    TheMessageFile.eo15Close(ReturnStatus); // This automatically flushes
  }
}
/*---------------------------------------------------------------------------*/
externC void eo15FlushMessageFiles( void )
{
  ROUTINE_DBG_MEO60 (_T("eo15FlushMessageFiles"));
  if ( eo15_NumberOfBuffers <= 0 )
  {
    return;
  }

  //
  // Called after a successfull SHUTDOWN. Resets any locking and flushes by doing forgotten close.
  //
  tsp00_VfReturn_Param ReturnStatus; 
  
#ifdef USE_DYNAMIC_MESSAGE_BUFFER
  ceo15_LockMutex Lock( eo15_BufferMapMutex );
#endif

  teo15MessageBufferSpace *ActiveBuffer;
  for ( int i = 0; (i < NUMBER_OF_MESSAGE_BUFFERS) 
               && ((ActiveBuffer = MESSAGE_BUFFER_PTR(i)) != NULL) ; i++ )
  {
    ActiveBuffer->m_Terminating = 0;
    ActiveBuffer->m_BusyCount = 0;

    sqlcreatemutex( &(ActiveBuffer->m_Mutex) );

    ceo15_MessageFile TheMessageFile( *ActiveBuffer );
    
    if ( TheMessageFile.eo15IsUsed() )
    {
      TheMessageFile.eo15Close(ReturnStatus);
    }
  }
}
/*---------------------------------------------------------------------------*/
// PTS 1105164
externC tsp00_Int4 eo15OpenToAppendToMessageFile( const tsp00_Char *    Filename,
                                                  tsp00_VfReturn_Param &ReturnStatus )
{
  ROUTINE_DBG_MEO60 (_T("eo15AppendToMessageFile"));

  if ( eo15_NumberOfBuffers <= 0 )
  {
    ReturnStatus = vf_notok;
    return INVALID_BUFFER_EO15;
  }

  if ( !eo15_CheckFilename(Filename) )
  {
    MSGD((ERR_TRACEIO_APPEND_OPEN_FAILED, Filename == NULL ? "<NULL>" : Filename, "invalid file name"));
    ReturnStatus = vf_notok;
    return INVALID_BUFFER_EO15;
  }

#ifdef USE_DYNAMIC_MESSAGE_BUFFER
  ceo15_LockMutex Lock( eo15_BufferMapMutex );
#endif
  // Inspect all buffers to find first free and verify against doublicated filename
  tsp00_Int4 FileIndex;
  tsp00_Int4 NewFileIndex;

  for ( FileIndex = 1, NewFileIndex = INVALID_BUFFER_EO15; 
        FileIndex <= NUMBER_OF_MESSAGE_BUFFERS; 
        FileIndex++ )
  {
#ifdef USE_DYNAMIC_MESSAGE_BUFFER
    if ( MESSAGE_BUFFER_PTR(FileIndex-1) == NULL )
    {
      if ( INVALID_BUFFER_EO15 == NewFileIndex )
      {
        NewFileIndex = FileIndex; // First free buffer index is memorized
      }
      continue;
    }
#endif

    ceo15_MessageFile TheMessageFile( *MESSAGE_BUFFER_PTR(FileIndex-1) );

    if ( !TheMessageFile.eo15Lock(ReturnStatus) )
    {
      MSGD((ERR_TRACEIO_APPEND_OPEN_FAILED, Filename, "buffer lock failed"));
TEST_PRINT("Lock failed\n")
      return INVALID_BUFFER_EO15;
    }
    
    if ( !TheMessageFile.eo15IsUsed() 
      && (INVALID_BUFFER_EO15 == NewFileIndex) )
    {
      NewFileIndex = FileIndex; // First free buffer index is memorized
    }
    
    // Leave loop if filename is duplicated (but return valid index!)
    if ( TheMessageFile.eo15CompareFilename(Filename) )
    {
      break;
    }
    
    // Keep only first free buffer locked....
    if ( NewFileIndex != FileIndex )
    {
      TheMessageFile.eo15Unlock();
    }
  }
  
  // If Filename is already in use, clear lock on first free buffer and return old index
  if ( FileIndex != (NUMBER_OF_MESSAGE_BUFFERS+1) )
  {
    // release not needed but locked buffer
    if ( INVALID_BUFFER_EO15 != NewFileIndex )
    {
      ceo15_MessageFile( *MESSAGE_BUFFER_PTR(NewFileIndex-1) ).eo15Unlock();
    }
    
    ceo15_MessageFile TheMessageFile( *MESSAGE_BUFFER_PTR(FileIndex-1) );
    TheMessageFile.eo15AddRef();
    TheMessageFile.eo15Unlock();
    ReturnStatus = vf_ok;
TEST_PRINT("Ok. Added Ref\n")
    return FileIndex;
  }
  
#ifdef USE_DYNAMIC_MESSAGE_BUFFER
  if ( INVALID_BUFFER_EO15 == NewFileIndex )
  {
    teo15MessageBufferSpace **NewMap;

    // Reallocate Dynamic Buffer Map
    if ( NO_ERROR_EO001 != ALLOC_MEM_EO57( (void **)&NewMap, sizeof(teo15MessageBufferSpace **) 
                      * ( NUMBER_OF_MESSAGE_BUFFERS + MAP_INCREASE_AMOUNT_EO15 ) ) )
    {
      MSGD((ERR_TRACEIO_APPEND_OPEN_FAILED, Filename, "Extending BufferMap failed"));
      return INVALID_BUFFER_EO15;
    }

    memcpy(NewMap, eo15_BufferMap, sizeof(teo15MessageBufferSpace **) * NUMBER_OF_MESSAGE_BUFFERS );
    memset(NewMap + NUMBER_OF_MESSAGE_BUFFERS, 0, sizeof(teo15MessageBufferSpace **) * MAP_INCREASE_AMOUNT_EO15 );
    NewFileIndex = NUMBER_OF_MESSAGE_BUFFERS + 1;
    teo15MessageBufferSpace **OldMap = eo15_BufferMap;
    eo15_BufferMap = NewMap;
    NUMBER_OF_MESSAGE_BUFFERS = NUMBER_OF_MESSAGE_BUFFERS + MAP_INCREASE_AMOUNT_EO15;
    FREE_MEM_EO57(OldMap);
  }
#endif

  if ( INVALID_BUFFER_EO15 != NewFileIndex )
  {
    /* Got a new buffer place */
#ifdef USE_DYNAMIC_MESSAGE_BUFFER
    if ( MESSAGE_BUFFER_PTR(NewFileIndex-1) == NULL )
    {
      if ( NO_ERROR_EO001 != ALLOC_MEM_EO57( (void **)&MESSAGE_BUFFER_PTR(NewFileIndex-1), 
                                              sizeof(teo15MessageBufferSpace)) )
      {
        MSGD((ERR_TRACEIO_APPEND_OPEN_FAILED, Filename, "Out of memory"));
        return INVALID_BUFFER_EO15;
      }

      if ( NewFileIndex > eo15_NumberOfBuffers )
      {
        MSGD(( WRN_TRACEIO_MAXIMUM_EXCEEDED, NewFileIndex, eo15_NumberOfBuffers ));
      }

      sqlcreatemutex( &(MESSAGE_BUFFER_PTR(NewFileIndex-1)->m_Mutex) );
      MESSAGE_BUFFER_PTR(NewFileIndex-1)->m_Used = 0;
      MESSAGE_BUFFER_PTR(NewFileIndex-1)->m_Terminating = 0;
      MESSAGE_BUFFER_PTR(NewFileIndex-1)->m_BusyCount = 0;
      ceo15_MessageFile( *MESSAGE_BUFFER_PTR(NewFileIndex-1) ).eo15Lock(ReturnStatus);
    }
#endif
    ceo15_MessageFile TheMessageFile( *MESSAGE_BUFFER_PTR(NewFileIndex-1) );
    
    TheMessageFile.eo15Append(Filename, ReturnStatus);
    TheMessageFile.eo15Unlock();
    if ( vf_ok == ReturnStatus )
    {
TEST_PRINT("Ok. Newfile\n")
      return NewFileIndex;
    }
    else
    {
TEST_PRINT("eo15Append failed\n")
      MSGD((ERR_TRACEIO_APPEND_OPEN_FAILED, Filename, "Check OS error code"));
      return INVALID_BUFFER_EO15;
    }
  }
  else
  {
    MSGD((ERR_TRACEIO_APPEND_OPEN_FAILED, Filename, "No more internal buffer"));
  }

  ReturnStatus = vf_notok;
TEST_PRINT("Strange... eo15Append failed\n")
  return INVALID_BUFFER_EO15;
}
/*---------------------------------------------------------------------------*/
externC tsp00_Int4 eo15OpenMessageFile( const tsp00_Char *    Filename,
                                       const teo15_Position  MaxFileSize,
                                       tsp00_VfReturn_Param &ReturnStatus )
{
  ROUTINE_DBG_MEO60 (_T("eo15OpenMessageFile"));

  if ( eo15_NumberOfBuffers <= 0 )
  {
    ReturnStatus = vf_notok;
    return INVALID_BUFFER_EO15;
  }

  if ( !eo15_CheckFilename(Filename) )
  {
    MSGD((ERR_TRACEIO_OPEN_FAILED, Filename == NULL ? "<NULL>" : Filename, "invalid file name"));
    ReturnStatus = vf_notok;
    return INVALID_BUFFER_EO15;
  }
  
#ifdef USE_DYNAMIC_MESSAGE_BUFFER
  ceo15_LockMutex Lock( eo15_BufferMapMutex );
#endif
  // Inspect all buffers to find first free and verify against doublicated filename
  tsp00_Int4 FileIndex;
  tsp00_Int4 NewFileIndex;

  for ( FileIndex = 1, NewFileIndex = INVALID_BUFFER_EO15; 
        FileIndex <= NUMBER_OF_MESSAGE_BUFFERS; 
        FileIndex++ )
  {
#ifdef USE_DYNAMIC_MESSAGE_BUFFER
    if ( MESSAGE_BUFFER_PTR(FileIndex-1) == NULL )
    {
      if ( INVALID_BUFFER_EO15 == NewFileIndex )
      {
        NewFileIndex = FileIndex; // First free buffer index is memorized
      }
      continue;
    }
#endif

    ceo15_MessageFile TheMessageFile( *MESSAGE_BUFFER_PTR(FileIndex-1) );
    
    if ( !TheMessageFile.eo15Lock(ReturnStatus) )
    {
      MSGD((ERR_TRACEIO_OPEN_FAILED, Filename, "buffer lock failed"));
      return INVALID_BUFFER_EO15;
    }
    
    if ( !TheMessageFile.eo15IsUsed() 
      && (INVALID_BUFFER_EO15 == NewFileIndex) )
    {
      NewFileIndex = FileIndex; // First free buffer index is memorized
    }
    
    // Leave loop if filename is duplicated (but return valid index!)
    if ( TheMessageFile.eo15CompareFilename(Filename) )
    {
      break;
    }
    
    // Keep only first free buffer locked....
    if ( NewFileIndex != FileIndex )
    {
      TheMessageFile.eo15Unlock();
    }
  }
  
  // If Filename is already in use, clear lock on first free buffer and return old index
  if ( FileIndex != (NUMBER_OF_MESSAGE_BUFFERS+1) )
  {
    // release not needed but locked buffer
    if ( INVALID_BUFFER_EO15 != NewFileIndex )
    {
      ceo15_MessageFile( *MESSAGE_BUFFER_PTR(NewFileIndex-1) ).eo15Unlock();
    }
    
    ceo15_MessageFile TheMessageFile( *MESSAGE_BUFFER_PTR(FileIndex-1) );
    TheMessageFile.eo15AddRef();
    TheMessageFile.eo15Unlock();
    ReturnStatus = vf_ok;
TEST_PRINT("Ok. Added Ref\n")
    return FileIndex;
  }
  
#ifdef USE_DYNAMIC_MESSAGE_BUFFER
  if ( INVALID_BUFFER_EO15 == NewFileIndex )
  {
    teo15MessageBufferSpace **NewMap;

    // Reallocate Dynamic Buffer Map
    if ( NO_ERROR_EO001 != ALLOC_MEM_EO57( (void **)&NewMap, sizeof(teo15MessageBufferSpace **) 
                      * ( NUMBER_OF_MESSAGE_BUFFERS + MAP_INCREASE_AMOUNT_EO15 ) ) )
    {
      MSGD((ERR_TRACEIO_OPEN_FAILED, Filename, "Extending BufferMap failed"));
      return INVALID_BUFFER_EO15;
    }

    memcpy(NewMap, eo15_BufferMap, sizeof(teo15MessageBufferSpace **) * NUMBER_OF_MESSAGE_BUFFERS );
    memset(NewMap + NUMBER_OF_MESSAGE_BUFFERS, 0, sizeof(teo15MessageBufferSpace **) * MAP_INCREASE_AMOUNT_EO15 );
    NewFileIndex = NUMBER_OF_MESSAGE_BUFFERS + 1;
    teo15MessageBufferSpace **OldMap = eo15_BufferMap;
    eo15_BufferMap = NewMap;
    NUMBER_OF_MESSAGE_BUFFERS = NUMBER_OF_MESSAGE_BUFFERS + MAP_INCREASE_AMOUNT_EO15;
    FREE_MEM_EO57(OldMap);
  }
#endif

  if ( INVALID_BUFFER_EO15 != NewFileIndex )
  {
    /* Got a new buffer place */
#ifdef USE_DYNAMIC_MESSAGE_BUFFER
    if ( MESSAGE_BUFFER_PTR(NewFileIndex-1) == NULL )
    {
      if ( NO_ERROR_EO001 != ALLOC_MEM_EO57( (void **)&MESSAGE_BUFFER_PTR(NewFileIndex-1), 
                                              sizeof(teo15MessageBufferSpace)) )
      {
        MSGD((ERR_TRACEIO_OPEN_FAILED, Filename, "Out of memory"));
        return INVALID_BUFFER_EO15;
      }

      if ( NewFileIndex > eo15_NumberOfBuffers )
      {
        MSGD(( WRN_TRACEIO_MAXIMUM_EXCEEDED, NewFileIndex, eo15_NumberOfBuffers ));
      }

      sqlcreatemutex( &(MESSAGE_BUFFER_PTR(NewFileIndex-1)->m_Mutex) );
      MESSAGE_BUFFER_PTR(NewFileIndex-1)->m_Used = 0;
      MESSAGE_BUFFER_PTR(NewFileIndex-1)->m_Terminating = 0;
      MESSAGE_BUFFER_PTR(NewFileIndex-1)->m_BusyCount = 0;
      ceo15_MessageFile( *MESSAGE_BUFFER_PTR(NewFileIndex-1) ).eo15Lock(ReturnStatus);
    }
#endif
    ceo15_MessageFile TheMessageFile( *MESSAGE_BUFFER_PTR(NewFileIndex-1) );
    
    TheMessageFile.eo15Open(Filename, MaxFileSize, ReturnStatus);
    TheMessageFile.eo15Unlock();
    if ( vf_ok == ReturnStatus )
    {
      return NewFileIndex;
    }
    else
    {
      MSGD((ERR_TRACEIO_OPEN_FAILED, Filename, "Check OS error code"));
      return INVALID_BUFFER_EO15;
    }
  }
  else
  {
    MSGD((ERR_TRACEIO_OPEN_FAILED, Filename, "No more internal buffer"));
  }
  ReturnStatus = vf_notok;
  return INVALID_BUFFER_EO15;
}
/*---------------------------------------------------------------------------*/
// PTS 1105164
externC tsp00_Int4 eo15GetMessageFileIndex( const tsp00_Char *    Filename )
{
  ROUTINE_DBG_MEO60 (_T("eo15GetMessageFileIndex"));

  tsp00_Int4 FileIndex = INVALID_BUFFER_EO15;

  if ( !eo15_CheckFilename(Filename) )
  {
    MSGD((ERR_TRACEIO_GET_INDEX_FAILED, Filename == NULL ? "<NULL>" : Filename, "invalid file name"));
  }
  else
  {
    tsp00_VfReturn_Param ReturnStatus;  
    tsp00_Int4 SearchIndex;
    // Inspect all buffers to find first free and verify against doublicated filename
    for ( SearchIndex = 1; FileIndex == INVALID_BUFFER_EO15 
      && SearchIndex <= NUMBER_OF_MESSAGE_BUFFERS; SearchIndex++ )
    {
#ifdef USE_DYNAMIC_MESSAGE_BUFFER
      if ( MESSAGE_BUFFER_PTR(SearchIndex) == NULL )
        continue;
#endif

      ceo15_MessageFile TheMessageFile( *MESSAGE_BUFFER_PTR(SearchIndex) );
  
      if ( !TheMessageFile.eo15Lock(ReturnStatus) )
      {
        break;
      }
  
      // Leave loop if filename is found (and return valid index!)
      if ( TheMessageFile.eo15IsUsed() 
        && TheMessageFile.eo15CompareFilename(Filename) )
      {
        FileIndex = SearchIndex;
      }
  
      TheMessageFile.eo15Unlock();
    }
  }
  return FileIndex;
}

/*---------------------------------------------------------------------------*/

// PTS 1105164
externC void eo15CheckExist( const tsp00_Char * Filename,
                            bool  &DoesExist,
                            bool  &IsDirectory,
                            tsp00_VfReturn_Param VAR_VALUE_REF ReturnStatus )
{
  ROUTINE_DBG_MEO60 (_T("eo15CheckExist"));

#if defined(_WIN32)
  DWORD Attributes = GetFileAttributes(Filename);
  
  DoesExist = ( Attributes != 0xFFFFFFFF );
  
  if ( DoesExist )
  {
    ReturnStatus = vf_ok;
    IsDirectory = ( (Attributes & FILE_ATTRIBUTE_DIRECTORY) == FILE_ATTRIBUTE_DIRECTORY );
  }
  else
  {
    ReturnStatus = (GetLastError() == ERROR_FILE_NOT_FOUND ? vf_ok : vf_notok);
    IsDirectory = false;
  }
#else
  struct stat testBuf;
  
  // stat follows symbolic links...
  DoesExist = ( stat( Filename, &testBuf ) == 0 );
  
  if ( DoesExist )
  {
    ReturnStatus = vf_ok;
    IsDirectory = ( testBuf.st_mode == S_IFDIR );
  }
  else
  {
    ReturnStatus = ( errno == ENOENT ? vf_ok : vf_notok );
    IsDirectory = false;
  }
#endif
  
  return;
}

/*---------------------------------------------------------------------------*/

externC void eo15WriteMessageFile( const tsp00_Int4  FileIndex,
                                  const tsp00_Char *Message,
                                  tsp00_VfReturn_Param &ReturnStatus )
{
  ROUTINE_DBG_MEO60 (_T("eo15WriteMessageFile"));

  if ( !ceo15_MessageFile::eo15CheckIndex(FileIndex) )
  {
    MSGD((ERR_TRACEIO_BAD_FILEINDEX, "Write", FileIndex ));
    ReturnStatus = vf_notok;
    return;
  }

  ceo15_MessageFile TheMessageFile( *MESSAGE_BUFFER_PTR(FileIndex-1) );
  
  if ( !TheMessageFile.eo15Lock(ReturnStatus) )
  {
    return;
  }

  if ( !TheMessageFile.eo15IsUsed() )
  {
    MSGD((WRN_TRACEIO_BAD_FILEINDEX, "Write(UnUsed)", FileIndex ));
  }
  else
  {
    TheMessageFile.eo15Write(Message, ReturnStatus);
  }

  TheMessageFile.eo15Unlock();
}
/*---------------------------------------------------------------------------*/
// PTS 1105164
externC void eo15SyncMessageFile( const tsp00_Int4 FileIndex,
                                 tsp00_VfReturn_Param &ReturnStatus )
{
  ROUTINE_DBG_MEO60 (_T("eo15SyncMessageFile"));

  if ( !ceo15_MessageFile::eo15CheckIndex(FileIndex) )
  {
    MSGD((ERR_TRACEIO_BAD_FILEINDEX, "Sync", FileIndex ));
    ReturnStatus = vf_notok;
    return;
  }

  ceo15_MessageFile TheMessageFile( *MESSAGE_BUFFER_PTR(FileIndex-1) );
    
  if ( !TheMessageFile.eo15Lock(ReturnStatus) )
  {
    return;
  }

  if ( !TheMessageFile.eo15IsUsed() )
  {
    MSGD((WRN_TRACEIO_BAD_FILEINDEX, "Sync(UnUsed)", FileIndex ));
  }
  else
  {
    TheMessageFile.eo15Sync(ReturnStatus);
  }

  TheMessageFile.eo15Unlock();
}
/*---------------------------------------------------------------------------*/
// Alternative Interface using Filename instead of FileIndex
// Locking prevents the index from becoming invalid in between, but make reuse of
// other interface function impossible.
externC void eo15SyncNamedMessageFile( const tsp00_Char *Filename,
                                      tsp00_VfReturn_Param &ReturnStatus )
{
  ROUTINE_DBG_MEO60 (_T("eo15SyncNamedMessageFile"));

  tsp00_Int4 FileIndex;
  
  eo15_MapAndLockMessageFile( Filename, FileIndex, ReturnStatus );
 
  if ( vf_ok != ReturnStatus )
  {
    MSGD((ERR_TRACEIO_CHECK_FILENAME, Filename == NULL ? "<NULL>" : Filename, "not opened"));
    return;
  }

  if ( !ceo15_MessageFile::eo15CheckIndex(FileIndex) )
  {
    MSGD((ERR_TRACEIO_BAD_FILEINDEX, "SyncNamed", FileIndex ));
    ReturnStatus = vf_notok;
    return;
  }

  ceo15_MessageFile TheMessageFile( *MESSAGE_BUFFER_PTR(FileIndex-1) );
    
  if ( !TheMessageFile.eo15IsUsed() )
  {
    MSGD((WRN_TRACEIO_BAD_FILEINDEX, "SyncNamed(UnUsed)", FileIndex ));
  }
  else
  {
    TheMessageFile.eo15Sync(ReturnStatus);
  }

  TheMessageFile.eo15Unlock();
}
/*---------------------------------------------------------------------------*/
externC void eo15CloseMessageFile( const tsp00_Int4 FileIndex,
                                  tsp00_VfReturn_Param &ReturnStatus )
{
  ROUTINE_DBG_MEO60 (_T("eo15CloseMessageFile"));

  if ( !ceo15_MessageFile::eo15CheckIndex(FileIndex) )
  {
    MSGD((WRN_TRACEIO_BAD_FILEINDEX, "Close", FileIndex ));
    ReturnStatus = vf_notok;
    return;
  }

  ceo15_MessageFile TheMessageFile( *MESSAGE_BUFFER_PTR(FileIndex-1) );
  
  if ( !TheMessageFile.eo15Lock(ReturnStatus) )
  {
    return;
  }

  if ( !TheMessageFile.eo15IsUsed() )
  {
    MSGD((WRN_TRACEIO_BAD_FILEINDEX, "Close(UnUsed)", FileIndex ));
  }
  else
  {
    TheMessageFile.eo15Close(ReturnStatus);
  }

  TheMessageFile.eo15Unlock();
}
/*---------------------------------------------------------------------------*/
externC void eo15CopyMessageFile( const tsp00_Int4 FileIndex,
                                 const char *CopyFileName,
                                 tsp00_VfReturn_Param &ReturnStatus )
{
  ROUTINE_DBG_MEO60 (_T("eo15CopyMessageFile"));

  if ( !ceo15_MessageFile::eo15CheckIndex(FileIndex) )
  {
    MSGD((ERR_TRACEIO_BAD_FILEINDEX, "CopyFile", FileIndex ));
    ReturnStatus = vf_notok;
    return;
  }

  ceo15_MessageFile TheMessageFile( *MESSAGE_BUFFER_PTR(FileIndex-1) );
  
  if ( !TheMessageFile.eo15Lock(ReturnStatus) )
  {
    return;
  }

  if ( !TheMessageFile.eo15IsUsed() )
  {
    MSGD((ERR_TRACEIO_BAD_FILEINDEX, "CopyFile(UnUsed)", FileIndex ));
    ReturnStatus = vf_notok;
  }
  else
  {
    TheMessageFile.eo15Copy(CopyFileName, ReturnStatus);
  }

  TheMessageFile.eo15Unlock();
}
/*---------------------------------------------------------------------------*/
// PTS 1105164
externC void       eo15RemoveMessageFile( const tsp00_Char *RemoveFileName,
                                          tsp00_VfReturn_Param VAR_VALUE_REF ReturnStatus )
{
  ROUTINE_DBG_MEO60 (_T("eo15RemoveMessageFile"));

  tsp00_Int4 FileIndex;

  eo15_MapAndLockMessageFile( RemoveFileName, FileIndex, ReturnStatus );
  
  if ( ceo15_MessageFile::eo15CheckIndex(FileIndex) )
  {
    ceo15_MessageFile TheMessageFile( *MESSAGE_BUFFER_PTR(FileIndex-1) );
    
    TheMessageFile.eo15Unlock();

    ReturnStatus = vf_notok;
  }
  else
  {
    eo15_Unlink(RemoveFileName, ReturnStatus);
  }
}
/*---------------------------------------------------------------------------*/
// PTS 1105164
externC void eo15GetWorkingDirectory( tsp00_Pathc          &WorkingDirectory,
                                     tsp00_VfReturn_Param &ReturnStatus )
{
  ROUTINE_DBG_MEO60 (_T("eo15GetWorkingDirectory"));

  ReturnStatus = ( 
#if defined(_WIN32)
    GetCurrentDirectory(sizeof(tsp00_Pathc), &WorkingDirectory[0]) == 0xFFFFFFFF 
#else
    getcwd(&WorkingDirectory[0], sizeof(tsp00_Pathc)) == NULL 
#endif
    ? vf_notok : vf_ok );
}
/*===========================================================================*
*  LOCAL/PRIVATE/PROTECTED FUNCTIONS (CODE)                                 *
*===========================================================================*/
void ceo15_MessageFile::eo15Write(const char *Message, tsp00_VfReturn_Param &ReturnStatus )
{
  ROUTINE_DBG_MEO60 (_T("ceo15_MessageFile::eo15Write"));

  teo15_Position LengthInBytes = strlen(Message);
  
  if ( (m_Space.m_BytesStored + LengthInBytes) <= MESSAGE_FILE_PAGE_MXEO15 )
  {
TEST_DPRINT3("Copy Write to 0x%lx [%d] (0x%lx bytes)\n", 
           &(m_Space.m_Buffer[m_Space.m_CurrentBuffer][m_Space.m_BytesStored]), 
           m_Space.m_CurrentBuffer, (long)LengthInBytes )
    memcpy( &(m_Space.m_Buffer[m_Space.m_CurrentBuffer][m_Space.m_BytesStored]), Message, LengthInBytes );
    m_Space.m_BytesStored += LengthInBytes;
//TEST_PRINT1("Bytes stored now 0x%lx bytes\n", (long)m_Space.m_BytesStored) 
  }
  else
  {
    do {
      if ( m_Space.m_BytesToFlush > 0 )
      {
    /* Workerthread failed to do its job, so we do it...
      Otherwise we would be blocked anyhow */
        TEST_PRINT("Doing worker threads job...\n")
          eo15Flush( ReturnStatus );
      }
      
      tsp00_ErrTextc  ThreadErrText;
      teo07_ThreadErr ThreadOk;
      
      m_Space.m_BytesToFlush = m_Space.m_BytesStored;
      m_Space.m_CurrentBuffer = 1 - m_Space.m_CurrentBuffer;
      if ( LengthInBytes > MESSAGE_FILE_PAGE_MXEO15 )
      {
//TEST_PRINT3("Copy Write(2) to 0x%lx from 0x%lx (0x%lx bytes)\n", 
//           &(m_Space.m_Buffer[m_Space.m_CurrentBuffer][0]), 
//           Message, (long)MESSAGE_FILE_PAGE_MXEO15 )
        memcpy( &(m_Space.m_Buffer[m_Space.m_CurrentBuffer][0]), Message, MESSAGE_FILE_PAGE_MXEO15 );
        m_Space.m_BytesStored  = MESSAGE_FILE_PAGE_MXEO15;
        LengthInBytes -= MESSAGE_FILE_PAGE_MXEO15;
      }
      else
      {
//TEST_PRINT3("Copy Write(3) to 0x%lx from 0x%lx (0x%lx bytes)\n", 
//           &(m_Space.m_Buffer[m_Space.m_CurrentBuffer][0]), 
//           Message, (long)LengthInBytes )
        memcpy( &(m_Space.m_Buffer[m_Space.m_CurrentBuffer][0]), Message, LengthInBytes );
        m_Space.m_BytesStored  = LengthInBytes;
        LengthInBytes = 0;
      }
//TEST_PRINT("Resume Worker Thread\n")
      sqlresumethread( eo15_WorkerThread, ThreadErrText, &ThreadOk );
//TEST_PRINT("After Resume Worker Thread\n")
    } while ( LengthInBytes != 0 );
  }
}
/*---------------------------------------------------------------------------*/
void ceo15_MessageFile::eo15Flush( tsp00_VfReturn_Param &ReturnStatus )
{
  ROUTINE_DBG_MEO60 (_T("ceo15_MessageFile::eo15Flush"));

  if ( m_Space.m_BytesToFlush > 0 )
  {
//    TEST_PRINT1("Flushing %ld bytes...\n", m_Space.m_BytesToFlush)
      if ( m_Space.m_MaxFileSize != UNLIMITED_FILE_SIZE_EO15 )
      {
//    TEST_PRINT1("m_MaxFileSize 0x%lx\n", m_Space.m_MaxFileSize)
        char Cycle[16];
        teo15_Position BytesWritten; // write result
        teo15_Position NextPosition; // seek result
        
        if ( (teo15_Position)(m_Space.m_BytesToFlush
          + m_Space.m_NextWritePosition
          + strlen(END_LINE_EO15) + CYCLE_COUNT_SIZE_EO15)
          > m_Space.m_MaxFileSize )
        {
          eo15_Seek( m_Space.m_HostFileNo,
            m_Space.m_NextWritePosition,
            eo15_SeekSet,
            NextPosition,
            ReturnStatus );
          if ( ReturnStatus != vf_ok ) 
          {
            eo15_ShowLastOsError("Flush(1st Seek)", m_Space.m_Filename.asCharp() );
            return;
          }
          
          eo15_Write( m_Space.m_HostFileNo,
            CLR_LINE_EO15,
            strlen(CLR_LINE_EO15),
            BytesWritten,
            ReturnStatus );
          if ( ReturnStatus != vf_ok ) 
          {
            eo15_ShowLastOsError("Flush(Write CLR)", m_Space.m_Filename.asCharp() );
            return;
          }
          
          eo15_Seek( m_Space.m_HostFileNo,
            0,
            eo15_SeekSet,
            NextPosition,
            ReturnStatus );
          if ( ReturnStatus != vf_ok ) 
          {
            eo15_ShowLastOsError("Flush(2nd Seek)", m_Space.m_Filename.asCharp() );
            return;
          }
          
          m_Space.m_CycleCounter++;
          sprintf(Cycle, "%-12d\n", m_Space.m_CycleCounter);
          
          eo15_Write( m_Space.m_HostFileNo,
            FIRST_LINE_EO15,
            strlen(FIRST_LINE_EO15),
            BytesWritten,
            ReturnStatus );
          if ( ReturnStatus != vf_ok ) 
          {
            eo15_ShowLastOsError("Flush(Write FIRST)", m_Space.m_Filename.asCharp() );
            return;
          }

          eo15_Write( m_Space.m_HostFileNo,
            Cycle,
            strlen(Cycle),
            BytesWritten,
            ReturnStatus );
          if ( ReturnStatus != vf_ok ) 
          {
            eo15_ShowLastOsError("Flush(Write Cycle)", m_Space.m_Filename.asCharp() );
            return;
          }
          
          m_Space.m_NextWritePosition = strlen(FIRST_LINE_EO15) + strlen(Cycle);
        }
        else
        {
          eo15_Seek( m_Space.m_HostFileNo,
            m_Space.m_NextWritePosition,
            eo15_SeekSet,
            NextPosition,
            ReturnStatus );
          if ( ReturnStatus != vf_ok ) 
          {
            eo15_ShowLastOsError("Flush(1st Seek)", m_Space.m_Filename.asCharp() );
            return;
          }
        }
        
        if ( m_Space.m_Terminating == 1 
          && m_Space.m_BusyCount != 0 )
        {
          char EndOfLine[3];
          sprintf(EndOfLine, "\n");
          
          eo15_Write( m_Space.m_HostFileNo,
            CRASH_MARK_EO15,
            strlen(CRASH_MARK_EO15),
            BytesWritten,
            ReturnStatus );
          if ( ReturnStatus != vf_ok ) 
          {
            eo15_ShowLastOsError("Flush(Write Mark)", m_Space.m_Filename.asCharp() );
            return;
          }
          eo15_Write( m_Space.m_HostFileNo,
            EndOfLine,
            strlen(EndOfLine),
            BytesWritten,
            ReturnStatus );
          if ( ReturnStatus != vf_ok ) 
          {
            eo15_ShowLastOsError("Flush(Write EOL)", m_Space.m_Filename.asCharp() );
            return;
          }
          /* Crash mark only written once!!! */
          ++(m_Space.m_Terminating);
          m_Space.m_NextWritePosition += strlen(CRASH_MARK_EO15) + strlen(EndOfLine);
        }
        
        eo15_Write( m_Space.m_HostFileNo,
          m_Space.m_Buffer[1 - m_Space.m_CurrentBuffer],
          m_Space.m_BytesToFlush,
          BytesWritten,
          ReturnStatus );
        if ( ReturnStatus != vf_ok ) 
        {
          eo15_ShowLastOsError("Flush(Write Buffer)", m_Space.m_Filename.asCharp() );
          return;
        }
        
        m_Space.m_TotalBytesWritten += BytesWritten;
        m_Space.m_NextWritePosition += BytesWritten;
        
        eo15_Write( m_Space.m_HostFileNo,
          END_LINE_EO15,
          strlen(END_LINE_EO15),
          BytesWritten,
          ReturnStatus );
        if ( ReturnStatus != vf_ok ) 
        {
          eo15_ShowLastOsError("Flush(Write END)", m_Space.m_Filename.asCharp() );
          return;
        }
        
        sprintf(Cycle, "%-12d", m_Space.m_CycleCounter);
        
        eo15_Write( m_Space.m_HostFileNo,
          Cycle,
          strlen(Cycle),
          BytesWritten,
          ReturnStatus );
        if ( ReturnStatus != vf_ok ) 
        {
          eo15_ShowLastOsError("Flush(Write Cycle)", m_Space.m_Filename.asCharp() );
          return;
        }
      }
      else
      {
    TEST_PRINT("File unlimited\n")
          teo15_Position BytesWritten; // write result
          
          if ( m_Space.m_Terminating == 1 
            && m_Space.m_BusyCount != 0 )
          {
    TEST_PRINT("Terminating\n")
            char EndOfLine[3];
            sprintf(EndOfLine, "\n");
            
            eo15_Write( m_Space.m_HostFileNo,
              CRASH_MARK_EO15,
              strlen(CRASH_MARK_EO15),
              BytesWritten,
              ReturnStatus );
            if ( ReturnStatus != vf_ok ) 
            {
              eo15_ShowLastOsError("Flush(Write Mark)", m_Space.m_Filename.asCharp() );
              return;
            }

            eo15_Write( m_Space.m_HostFileNo,
              EndOfLine,
              strlen(EndOfLine),
              BytesWritten,
              ReturnStatus );
            if ( ReturnStatus != vf_ok ) 
            {
              eo15_ShowLastOsError("Flush(Write EOL)", m_Space.m_Filename.asCharp() );
              return;
            }
            /* Crash mark only written once!!! */
            ++(m_Space.m_Terminating);
          }
          
    TEST_PRINT1("Bytes to flush: 0x%lx\n", (long)m_Space.m_BytesToFlush)
    TEST_PRINT1("Write from 0x%lx\n", &(m_Space.m_Buffer[1- m_Space.m_CurrentBuffer]))
    TEST_PRINT1("Writing '%29.29s'\n", &(m_Space.m_Buffer[1- m_Space.m_CurrentBuffer]))
          eo15_Write( m_Space.m_HostFileNo,
            m_Space.m_Buffer[1 - m_Space.m_CurrentBuffer],
            m_Space.m_BytesToFlush,
            BytesWritten,
            ReturnStatus );
          if ( ReturnStatus != vf_ok ) 
          {
            eo15_ShowLastOsError("Flush(Write Buffer)", m_Space.m_Filename.asCharp() );
            return;
          }
          
          m_Space.m_TotalBytesWritten += BytesWritten;
          m_Space.m_NextWritePosition += BytesWritten;
        }
        
        m_Space.m_BytesToFlush = 0;
    }
    else
    {
//      TEST_PRINT("Flushing not needed...\n")
    }
}

/*---------------------------------------------------------------------------*/
bool ceo15_MessageFile::eo15Lock( tsp00_VfReturn_Param &ReturnStatus )
{
  ROUTINE_DBG_MEO60 (_T("ceo15_MessageFile::eo15Lock"));

  if ( m_Space.m_Terminating > 0 )
  {
    MSGD((INFO_TRACEIO_LOCK_TERMINATING, m_Space.m_Filename.asCharp() ));
    ReturnStatus = vf_notok;
    return false;
  }
    // PTS 11054990
  sqlbeginmutex( &(m_Space.m_Mutex) );
  ++(m_Space.m_BusyCount);

  ReturnStatus = vf_ok;
  return true;
}

/*---------------------------------------------------------------------------*/
void ceo15_MessageFile::eo15Unlock()
{
  ROUTINE_DBG_MEO60 (_T("ceo15_MessageFile::eo15Unlock"));

  --(m_Space.m_BusyCount);
  sqlendmutex( &(m_Space.m_Mutex) );
}
/*---------------------------------------------------------------------------*/
extern "C"
{
  // static       is not allowed for extern "C" functions...
  void *eo15DumpMessageFile( void * )
  {
#ifndef HPUX
    // aCC compiler complains:
    // Warning 495: "veo15.cpp", line 1032 # The linkage directive is ignored 
    // for an object or function declared static.
    ROUTINE_DBG_MEO60 (_T("eo15DumpMessageFile"));
#endif
    int FileIndex;
    tsp00_ErrTextc  ThreadErrText;
    teo07_ThreadErr ThreadOk;
    tsp00_VfReturn_Param ReturnStatus;
    
    while ( eo15_NumberOfBuffers > 0 )
    {
      for ( FileIndex = 1; FileIndex <= NUMBER_OF_MESSAGE_BUFFERS; FileIndex++ )
      {
#ifdef USE_DYNAMIC_MESSAGE_BUFFER
        ceo15_LockMutex Lock( eo15_BufferMapMutex );

        if ( MESSAGE_BUFFER_PTR(FileIndex-1) == NULL )
          continue;
#endif /* USE_DYNAMIC_MESSAGE_BUFFER */

        ceo15_MessageFile TheMessageFile( *MESSAGE_BUFFER_PTR(FileIndex-1) );
        
        if ( !TheMessageFile.eo15Lock(ReturnStatus) )
        {
          continue;
        }

        if ( TheMessageFile.eo15IsUsed() )
        {
//  TEST_PRINT1("Worker trys to flush File %d\n", FileIndex)
          TheMessageFile.eo15Flush( ReturnStatus );
        }

//  TEST_PRINT("Worker unlocks\n")
        TheMessageFile.eo15Unlock();
      }
//  TEST_PRINT("Worker suspends itself\n")
      sqlsuspendthread( eo15_WorkerThread, ThreadErrText, &ThreadOk );
    }
    
    return (void *)0;
  }
};
/*---------------------------------------------------------------------------*/
externC void eo15_Open ( teo15_Handle     &FileHandle,
                        const tsp00_Char     *Path,
                        const teo15_OpenMode  FileMode,
                        const bool            DoCreate,
                        teo15_SecurityAttrib  Security,
                        tsp00_VfReturn_Param &ReturnStatus )
{
  ROUTINE_DBG_MEO60 (_T("eo15_Open"));

#if defined(_WIN32)
  DWORD OpenMode;
  switch( FileMode )
  {
  case eo15_ReadWrite:
    OpenMode = GENERIC_READ|GENERIC_WRITE;
    break;
  case eo15_ReadOnly:
    OpenMode = GENERIC_READ;
    break;
  case eo15_WriteOnly:
    OpenMode = GENERIC_WRITE;
    break;
  default:
    OpenMode = GENERIC_READ|GENERIC_WRITE;
    break;
  }
  FileHandle = CreateFile( Path,
    OpenMode,
    FILE_SHARE_READ,
    Security,
    ( DoCreate ? CREATE_ALWAYS : OPEN_EXISTING ),
    FILE_ATTRIBUTE_NORMAL, // WRITE_THROUGH would only be needed to protect againt system crash
    INVALID_HANDLE_VALUE);
  if ( FileHandle == INVALID_HANDLE_VALUE )
#else
    int OpenMode;
  switch( FileMode )
  {
  case eo15_ReadWrite:
    OpenMode = O_RDWR;
    break;
  case eo15_ReadOnly:
    OpenMode = O_RDONLY;
    break;
  case eo15_WriteOnly:
    OpenMode = O_WRONLY;
    break;
  default:
    OpenMode = O_RDWR;
    break;
  }
  FileHandle = open( Path, OpenMode | ( DoCreate ? O_CREAT : 0 ), ( Security == 0 ? 0777 : Security ) );
  if ( FileHandle < 0 )
#endif
  {
    ReturnStatus = vf_notok;
  }
  else
  {
    ReturnStatus = vf_ok;
  }
}
/*---------------------------------------------------------------------------*/
externC void eo15_Close ( const teo15_Handle     FileHandle,
                         tsp00_VfReturn_Param      &ReturnStatus )
{
  ROUTINE_DBG_MEO60 (_T("eo15_Close"));

#if defined(_WIN32)
  BOOL Closed;

  try {
    Closed = CloseHandle( FileHandle );
  }
  catch(...)
  {
    Closed = false;
  }
  if ( !Closed )
#else
    if ( close( FileHandle ) == -1 )
#endif
    {
      ReturnStatus = vf_notok;
    }
    else
    {
      ReturnStatus = vf_ok;
    }
}
/*---------------------------------------------------------------------------*/
/* PTS 1106522 */
#if defined(_WIN32)
/* This function is used after each write, to simulate UNIX modification time behaviour */
static void  eo15_UpdateModificationTime(const teo15_Handle FileHandle)
{
  SYSTEMTIME SystemTime;
  FILETIME   LastModificationTime;

  GetSystemTime(&SystemTime);
  SystemTimeToFileTime(&SystemTime, &LastModificationTime);
  /* LastAccessTime must be updated on read,write,execute */
  /* LastWriteTime must be updated on write */
  SetFileTime(FileHandle, NULL, &LastModificationTime, &LastModificationTime);
}
/* This function is used after each write, to simulate UNIX access time behaviour */
static void  eo15_UpdateAccessTime(const teo15_Handle FileHandle)
{
  SYSTEMTIME SystemTime;
  FILETIME   LastAccessTime;

  GetSystemTime(&SystemTime);
  SystemTimeToFileTime(&SystemTime, &LastAccessTime);
  /* LastAccessTime must be updated on read,write,execute */
  SetFileTime(FileHandle, NULL, &LastAccessTime, NULL);
}
#endif /* _WIN32 */
/*---------------------------------------------------------------------------*/
externC void eo15_Write ( const teo15_Handle    FileHandle,
                         const void           *Buffer,
                         const teo15_Position  WantedSizeInBytes,
                         teo15_Position       &WrittenSizeInBytes,
                         tsp00_VfReturn_Param &ReturnStatus )
{
  ROUTINE_DBG_MEO60 (_T("eo15_Write"));
TEST_DPRINT3("Writing to %d 0x%lx (0x%lx bytes)\n", FileHandle, Buffer, WantedSizeInBytes )
  teo15_Position Wanted = WantedSizeInBytes;
  teo15_Position Written;
  
  WrittenSizeInBytes = 0;
  do {
#if defined(_WIN32)
    if ( !WriteFile( FileHandle,
      ((char *)Buffer)+WrittenSizeInBytes,
      Wanted,
      &Written,
      NULL) )
    {
      break;
    }
    
    if ( Written > 0 )
    {
      WrittenSizeInBytes += Written;
      Wanted -= Written;
    }
#else
    Written = write( FileHandle, ((char *)Buffer)+WrittenSizeInBytes, Wanted );
    if ( Written > 0 )
    {
      WrittenSizeInBytes += Written;
      Wanted -= Written;
    }
    else
    {
      break;
    }
#endif
  } while ( Wanted > 0 );
  
  if ( WrittenSizeInBytes != WantedSizeInBytes )
  {
    ReturnStatus = vf_notok;
  }
  else
  {
    ReturnStatus = vf_ok;
  }
#if defined(_WIN32)
  /* PTS 1106522 */
  eo15_UpdateModificationTime(FileHandle);
#endif /* _WIN32 */
}
/*---------------------------------------------------------------------------*/
externC void eo15_Read ( const teo15_Handle      FileHandle,
                        void                   *Buffer,
                        const teo15_Position    WantedSizeInBytes,
                        teo15_Position         &ReceivedSizeInBytes,
                        tsp00_VfReturn_Param   &ReturnStatus )
{
  ROUTINE_DBG_MEO60 (_T("eo15_Read"));

#if defined(_WIN32)
  if ( !ReadFile( FileHandle,
    Buffer,
    WantedSizeInBytes,
    &ReceivedSizeInBytes,
    NULL) )
#else
    ReceivedSizeInBytes = read( FileHandle, Buffer, WantedSizeInBytes );
  
  if ( ReceivedSizeInBytes < 0 )
#endif
  {
    ReturnStatus = vf_notok;
  }
  else
  {
    if ( ReceivedSizeInBytes == 0 )
    {
      ReturnStatus = vf_eof;
    }
    else
    {
      ReturnStatus = vf_ok;
    }
  }
#if defined(_WIN32)
  /* PTS 1106522 */
  eo15_UpdateAccessTime(FileHandle);
#endif /* _WIN32 */
}
/*---------------------------------------------------------------------------*/
externC void eo15_Seek ( const teo15_Handle      FileHandle,
                        const teo15_Position    Position,
                        const teo15_SeekMode    SeekMode,
                        teo15_Position         &NewPosition,
                        tsp00_VfReturn_Param   &ReturnStatus )
{
  ROUTINE_DBG_MEO60 (_T("eo15_Seek"));

#if defined(_WIN32)
  DWORD whence;
  switch ( SeekMode )
  {
  case eo15_SeekSet:
    whence = FILE_BEGIN;
    break;
  case eo15_SeekEnd:
    whence = FILE_END;
    break;
  case eo15_SeekCurrent:
    whence = FILE_CURRENT;
    break;
  default:
    whence = FILE_BEGIN;
    break;
  }
  NewPosition = SetFilePointer( FileHandle,
    Position,
    NULL,
    whence);
#else
  int whence;
  switch ( SeekMode )
  {
  case eo15_SeekSet:
    whence = SEEK_SET;
    break;
  case eo15_SeekEnd:
    whence = SEEK_END;
    break;
  case eo15_SeekCurrent:
    whence = SEEK_CUR;
    break;
  default:
    whence = SEEK_SET;
    break;
  }
  NewPosition = lseek( FileHandle,
    Position,
    whence);
#endif
  if ( NewPosition == -1 )
  {
    ReturnStatus = vf_noseek;
  }
  else
  {
    ReturnStatus = vf_ok;
  }
}
/*---------------------------------------------------------------------------*/
externC void eo15_CopyFile ( const tsp00_Char     *Original,
                            const tsp00_Char     *Destination,
                            tsp00_VfReturn_Param &ReturnStatus )
{
  ROUTINE_DBG_MEO60 (_T("eo15_CopyFile"));

#if defined(_WIN32)
  ReturnStatus = ( CopyFile( Original, Destination, false ) ? vf_ok : vf_notok );
#else
  teo15_Handle OriginalFd;
  teo15_Handle DestinationFd;
  tsp00_VfReturn_Param IgnoredError; // close errors of temporary files are ignored here
  
  eo15_Open( OriginalFd, Original, eo15_ReadOnly, false, DEFAULT_SECURITY_EO15, ReturnStatus );
  if ( vf_ok == ReturnStatus )
  {
    eo15_Open( DestinationFd, Destination, eo15_WriteOnly, true, DEFAULT_SECURITY_EO15, ReturnStatus );
    if ( vf_ok != ReturnStatus )
    {
      eo15_Open( DestinationFd, Destination, eo15_WriteOnly, false, DEFAULT_SECURITY_EO15, ReturnStatus );
      if ( vf_ok != ReturnStatus )
      {
        eo15_ShowLastOsError("Open(Destination)", Destination );
        eo15_Close( OriginalFd, IgnoredError );
        return;
      }
    }
    
    teo15_Position BytesWritten;
    teo15_Position BytesRead;
    long Buffer[COPY_SIZE_EO15];
    
    do {
      eo15_Read( OriginalFd, Buffer, COPY_SIZE_EO15, BytesRead, ReturnStatus );
      if ( vf_ok == ReturnStatus )
      {
        eo15_Write( DestinationFd, Buffer, BytesRead, BytesWritten, ReturnStatus );
        if ( vf_ok != ReturnStatus )
        {
          eo15_ShowLastOsError("Write(Destination)", Destination );
        }
      }
      else if ( vf_eof != ReturnStatus ) /* 1110222 */
      {
        eo15_ShowLastOsError("Read(Original)", Original );
      }
    } while ( vf_ok == ReturnStatus && 0 != BytesRead );
    eo15_Close( OriginalFd, IgnoredError );
    eo15_Close( DestinationFd, IgnoredError );
    if ( ReturnStatus == vf_eof && BytesRead == 0 )
    {
      ReturnStatus = vf_ok;
    }
  }
  else
  {
    eo15_ShowLastOsError("Open(Original)", Original );
  }
#endif
}

/*---------------------------------------------------------------------------*/

externC void eo15_Unlink ( const tsp00_Char                 *FileName,
                           tsp00_VfReturn_Param VAR_VALUE_REF ReturnStatus )
{
  ROUTINE_DBG_MEO60 (_T("eo15_Unlink"));

#if defined(_WIN32)
    ReturnStatus = ( DeleteFile(FileName) ? vf_ok : vf_notok );
#else
    ReturnStatus = ( unlink(FileName) == 0 ? vf_ok : vf_notok );
#endif /* WIN32 */
    if ( vf_ok != ReturnStatus )
    {
      eo15_ShowLastOsError("Unlink", FileName );
    }
}

/*---------------------------------------------------------------------------*/
static void eo15_ShowLastOsError( const tsp00_Char *    Routine,
                                  const tsp00_Char *    Filename )
{
 teo15_LastErrorMessage LastErrorMessage;
#ifdef WIN32
  // Error message on Win32 are sometimes much much too long so we only report Error code...
  sp77sprintf( &LastErrorMessage[0], sizeof(teo15_LastErrorMessage)-1, "[%d]", GetLastError() );
#else
  sp77sprintf( &LastErrorMessage[0], sizeof(teo15_LastErrorMessage)-1, "[%d] %s", errno, sqlerrs() );
#endif
  MSGD(( ERR_TRACEIO_OS_ERROR, 
         Routine, 
         Filename == NULL ? "" : Filename, 
         &LastErrorMessage[0] ));
}
/*---------------------------------------------------------------------------*/
static void eo15_MapAndLockMessageFile( const tsp00_Char *    Filename,
                                       tsp00_Int4           &FileIndex,
                                       tsp00_VfReturn_Param &ReturnStatus )
{
  ROUTINE_DBG_MEO60 (_T("eo15_MapAndLockMessageFile"));

  FileIndex = INVALID_BUFFER_EO15;
  
  if ( !eo15_CheckFilename(Filename) )
  {
    ReturnStatus = vf_notok;
  }
  else
  {

    tsp00_Int4 SearchIndex;
    // Inspect all buffers to find first free and verify against doublicated filename
    for ( SearchIndex = 1; SearchIndex <= NUMBER_OF_MESSAGE_BUFFERS; SearchIndex++ )
    {
#ifdef USE_DYNAMIC_MESSAGE_BUFFER
      ceo15_LockMutex Lock(eo15_BufferMapMutex);

      if ( MESSAGE_BUFFER_PTR(SearchIndex) == NULL )
        continue;
#endif

      ceo15_MessageFile TheMessageFile( *MESSAGE_BUFFER_PTR(SearchIndex-1) );
      
      if ( !TheMessageFile.eo15Lock(ReturnStatus) )
      {
        break;
      }
      
      // Leave loop if filename is found (and return valid index!)
      if ( TheMessageFile.eo15IsUsed() 
        && TheMessageFile.eo15CompareFilename(Filename) )
      {
        FileIndex = SearchIndex;
        break;
      }
      
      TheMessageFile.eo15Unlock();
    }
  }
}
/*---------------------------------------------------------------------------*/
static bool eo15_CheckFilename(const tsp00_Char *Filename)
{
  bool ReturnStatus = true;

  if ( (Filename == NULL) || (*Filename == '\0') || strlen(Filename) > sizeof(tsp00_Path) )
  {
    ReturnStatus = false;

    if ( Filename == NULL )
    {
      MSGD((ERR_TRACEIO_CHECK_FILENAME, "<NULL>", "invalid file name"));
    } else if ( *Filename == '\0' ) {
      MSGD((ERR_TRACEIO_CHECK_FILENAME, "<Empty>", "zero length file name"));
    } else {
      MSGD((ERR_TRACEIO_CHECK_FILENAME, Filename, "file name too long"));
    }
  }

  return ReturnStatus;
}
//
//
//
//
//=================================== TEST Program =============================
//
//
//
//
#ifdef TEST

extern "C" void sqlabort( void )
{
  printf("PANIC...Aborting program\n");
  exit(0);
}

extern "C" void vabort(bool doDump)
{
  printf("PANIC...Aborting program %s\n", doDump ? "with dump" : "without dump");
  exit(0);
}

#if !defined(_WIN32)
externC char *sqlerrs ()
{
	/* imported data */
	if ( errno == 0 )
	{
		return ( (char *)"NO ERROR(0)" );
	}
#if defined(SUN) && defined(SOLARIS7)
	{
		const char *ErrorMessage = strerror(errno);
		if ( ErrorMessage == NULL )
		{
			ErrorMessage = "ERRNO NOT IN sys_errlist[]";
		}
		return (char *)ErrorMessage;
	}
#else
    if ( (errno < 1) || (errno > sys_nerr) )
	{
		return ( (char *)"ERRNO NOT IN sys_errlist[]" );
	}
	return ( (char *)sys_errlist [ errno ] );
#endif
}

extern "C" void e76_alarm0(void)
{
  sqlyieldthread();
}

extern "C" void sql60c_msg_8 ( int                  msg_no ,
                               int                  msg_type ,
                               const char          *msg_label ,
                               const char          *msg_format ,
                               ...                           ) 
{
    va_list         ap;
    va_start (ap, msg_format);
    (void)vfprintf (stdout, msg_format, ap);
    va_end (ap);
}
#else
extern "C" void  sql60_msg_diag               ( tsp00_Longint    lMsgID,
                                             tsp00_Longuint      ulEventType,
                                             const char         *pszLabel,
                                             const char         *pszFormatStr,
                                             ... )
{
    (void)fprintf (stdout, "%s:", pszLabel);
    va_list         ap;
    va_start (ap, pszFormatStr);
    (void)vfprintf (stdout, pszFormatStr, ap);
    va_end (ap);
    (void)fprintf (stdout, "\n");
}
#endif /* !WIN32 */

teo15MessageBufferSpace theBuffer[3];

void ShowLine(tsp00_Int4 FileIndex)
{
  static long lines[3] = {1,1,1};
  static long total = 0;
  tsp00_VfReturn_Param ReturnStatus;
  char MessageBuffer[256];
  
  sprintf(MessageBuffer, "Line %ld File %d Total %ld\n", lines[FileIndex-1], FileIndex, total);
  eo15WriteMessageFile( FileIndex, MessageBuffer, ReturnStatus);
  if ( ReturnStatus != vf_ok )
  {
    printf("Write line %d to file %d failed!!!\n", lines[FileIndex-1], FileIndex);
  }
  else
  {
    lines[FileIndex-1]++;
    total++;
  }
}

extern "C" void *StressThread(void *arg)
{
  tsp00_Int4 MessageIndex = *(tsp00_Int4 *)arg;
  const char *Message = "Message\n";
  tsp00_VfReturn_Param ReturnStatus;

  for ( int k = 0; k < 1000; k++ )
  {
    if ( k % 10 == 0 ) printf("%d:Writing 10000 Messages\n", k/10);
    for ( int j = 0; j < 1000; j++ )
    {
      eo15WriteMessageFile( MessageIndex, Message, ReturnStatus);
    }
  }
  return NULL;
}

int m_StressTest()
{
  int                  i;
  tsp00_Int4           MessageIndex;
  tsp00_VfReturn_Param ReturnStatus;
  tsp00_ErrTextc       ThreadErrText;
  teo07_ThreadErr      ThreadOk = THR_OK_EO07;
  teo07_Thread         Threads[3];

  MessageIndex = eo15OpenMessageFile( "StressFile.txt", 1024*14, ReturnStatus );
  if ( vf_ok != ReturnStatus )
  {
    printf("Failed to OpenMessageFile 'StressFile.txt' \n");
    return 1;
  }
  
  for ( i = 0; i < 3; i++ )
  {
    sqlbeginthread( WORKER_STACK_EO15,
      StressThread,
      &MessageIndex,
      THR_CREATE_SUSPENDED_EO07,
      &Threads[i],
      ThreadErrText,
      &ThreadOk );
    if ( ThreadOk != THR_OK_EO07 )
    {
      ABORT();
      return 2;
    }
    sqlresumethread( Threads[i], ThreadErrText, &ThreadOk );
    if ( ThreadOk != THR_OK_EO07 )
    {
      ABORT();
      return 3;
    }
  }

  for ( i = 0; i < 3; i++ )
  {
    tsp00_Int4 Status;
    sqljointhread( Threads[i], &Status, ThreadErrText, &ThreadOk );
    if ( ThreadOk != THR_OK_EO07 )
    {
      ABORT();
      return 4;
    }
  }

  eo15CloseMessageFile( MessageIndex, ReturnStatus );
  return 0;
}

int main(int argc, char *argv[])
{
  teo15_Handle         FileHandle;
  tsp00_VfReturn_Param ReturnStatus;
  int i;
  teo07_Thread         MainThread;
  tsp00_ErrTextc       ThreadErrText;
  teo07_ThreadErr      ThreadErr;

  setbuf( stdout, NULL );
  setbuf( stderr, NULL );

  sqlinit_multithreading( &MainThread, ThreadErrText, &ThreadErr);

  if ( ThreadErr == THR_NOT_OK_EO07 )
  {
    printf("Init multithreadring failed\n");
    return 0;
  }

  printf("Testbuffer begin 0x%lx ends at 0x%lx\n", theBuffer, &theBuffer[3]);

  memset(theBuffer, 0, 3*sizeof(teo15MessageBufferSpace));
  
  eo15InitMessageBufferSpace(3, theBuffer );
  
  if ( argc > 1 && strcmp(argv[1], "m") == 0 )
  {
    return m_StressTest();
  }

  eo15_Open ( FileHandle,
    "testfile.txt",
    eo15_ReadWrite,
    true,
    DEFAULT_SECURITY_EO15,
    ReturnStatus );
  
  if ( vf_ok != ReturnStatus )
  {
    printf("Open failed\n");
    return 0;
  }
  printf("Opened file testfile.txt\n");
  
  teo15_Position       WrittenSizeInBytes;
  for ( i = 0; i < 256; i++ )
  {
    char buffer[1];
    
    buffer[0] = (i%10) + '0';
    eo15_Write( FileHandle,
      buffer,
      1,
      WrittenSizeInBytes,
      ReturnStatus );
    if ( vf_ok != ReturnStatus )
    {
      printf("Write failed\n");
      return 0;
    }
  }
  printf("Numbers written\n");
  
  eo15_CopyFile("testfile.txt",
    "otherfile.txt",
    ReturnStatus );
  if ( vf_ok != ReturnStatus )
  {
    printf("Copyfile failed! Exists otherfile.txt already?\n");
  }
  else
  {
    printf("Copied file from testfile.txt to otherfile.txt\n");
  }
  
  teo15_Position NewPosition;
  eo15_Seek( FileHandle,
    12,
    eo15_SeekSet,
    NewPosition,
    ReturnStatus );
  if ( vf_ok != ReturnStatus )
  {
    printf("Seek failed\n");
    return 0;
  }
  printf("New position %ld\n", (long)NewPosition);
  
  {
    char ReadBuffer[5];
    teo15_Position BytesRead;
    
    eo15_Read( FileHandle,
      ReadBuffer,
      4,
      BytesRead,
      ReturnStatus );
    if ( vf_ok != ReturnStatus )
    {
      printf("Read failed\n");
      return 0;
    }
    
    ReadBuffer[4] = 0;
    printf("Expected 2345 Got %s\n", ReadBuffer);
  }
  
  eo15_Seek( FileHandle,
    12,
    eo15_SeekSet,
    NewPosition,
    ReturnStatus );
  if ( vf_ok != ReturnStatus )
  {
    printf("Seek failed\n");
    return 0;
  }
  
  eo15_Write( FileHandle,
    "AHA",
    3,
    WrittenSizeInBytes,
    ReturnStatus );
  if ( vf_ok != ReturnStatus )
  {
    printf("Write failed\n");
    return 0;
  }
  
  eo15_Close( FileHandle,
    ReturnStatus );
  if ( vf_ok != ReturnStatus )
  {
    printf("Close failed\n");
    return 0;
  }

  eo15RemoveMessageFile("Messagefile.txt", ReturnStatus);
  eo15RemoveMessageFile("Messagefile2.txt", ReturnStatus);
  eo15RemoveMessageFile("Messagefile3.txt", ReturnStatus);

  tsp00_Int4 MessageIndex;
  
  MessageIndex = eo15OpenMessageFile( "Messagefile.txt",
    1024*20,
    ReturnStatus );
  if ( vf_ok != ReturnStatus )
  {
    printf("Failed to OpenMessageFile 'Messagefile.txt' \n");
    return 0;
  }
  eo15CloseMessageFile( MessageIndex, ReturnStatus );
  if ( vf_ok != ReturnStatus )
  {
    printf("Failed to CloseMessageFile 'Messagefile.txt' \n");
    return 0;
  }
  eo15CloseMessageFile( MessageIndex, ReturnStatus );
  if ( vf_ok != ReturnStatus )
  {
    printf("Failed to 2nd CloseMessageFile 'Messagefile.txt' \n");
    return 0;
  }
  MessageIndex = eo15OpenMessageFile( "Messagefile.txt",
    1024*20,
    ReturnStatus );
  if ( vf_ok != ReturnStatus )
  {
    printf("Failed to OpenMessageFile 'Messagefile.txt' \n");
    return 0;
  }
  
  tsp00_Int4 MessageIndex2;
  MessageIndex2 = eo15OpenMessageFile( "Messagefile2.txt",
    1024*14,
    ReturnStatus );
  if ( vf_ok != ReturnStatus )
  {
    printf("Failed to OpenMessageFile 'Messagefile2.txt' \n");
    return 0;
  }
  
  tsp00_Int4 MessageIndex3;
  MessageIndex3 = eo15OpenMessageFile( "Messagefile3.txt",
    UNLIMITED_FILE_SIZE_EO15,
    ReturnStatus );
  if ( vf_ok != ReturnStatus )
  {
    printf("Failed to OpenMessageFile 'Messagefile3.txt' \n");
    return 0;
  }
  
  tsp00_Int4 FailingMessageIndex;
  FailingMessageIndex = eo15OpenMessageFile( "Messagefile4.txt",
    1024,
    ReturnStatus );
  if ( vf_ok != ReturnStatus )
  {
    printf("Failed to OpenMessageFile 'Messagefile4.txt' (expected)\n");
  }
  
  char MessageBuffer[256];
  sprintf(MessageBuffer, "A good day today\n");
  eo15WriteMessageFile( MessageIndex, MessageBuffer, ReturnStatus);
  eo15WriteMessageFile( MessageIndex2, MessageBuffer, ReturnStatus);
  eo15WriteMessageFile( MessageIndex3, MessageBuffer, ReturnStatus);
  
  for ( i = 0; i < 12345; i++ )
  {
    ShowLine(MessageIndex);
    ShowLine(MessageIndex2);
  }
  
  printf("Some message are written\n");

  sqlyieldthread();
  
  eo15CopyMessageFile( MessageIndex, "MessagefileCopy.txt", ReturnStatus);
  
  sprintf(MessageBuffer, "Beginning with more info\n");
  eo15WriteMessageFile( MessageIndex, MessageBuffer, ReturnStatus);
  eo15WriteMessageFile( MessageIndex2, MessageBuffer, ReturnStatus);
  eo15WriteMessageFile( MessageIndex3, MessageBuffer, ReturnStatus);
  
  for ( i = 0; i < 12; i++ )
  {
    ShowLine(MessageIndex);
    ShowLine(MessageIndex2);
    ShowLine(MessageIndex3);
  }
  
  sprintf(MessageBuffer, "Last message to file\n");
  eo15WriteMessageFile( MessageIndex, MessageBuffer, ReturnStatus);
  eo15WriteMessageFile( MessageIndex2, MessageBuffer, ReturnStatus);
  eo15WriteMessageFile( MessageIndex3, MessageBuffer, ReturnStatus);
  
  eo15CloseMessageFile( MessageIndex, ReturnStatus );

  MessageIndex = eo15OpenToAppendToMessageFile( "Messagefile.txt", ReturnStatus );
  if ( vf_ok != ReturnStatus )
  {
    printf("Failed to eo15OpenToAppendToMessageFile 'Messagefile.txt' \n");
    return 0;
  }
  else
  {
    DbgPrint = 1;
    printf("Succeeded in OpenToAppendToMessageFile 'Messagefile.txt' : %d\n", MessageIndex);
    sprintf(MessageBuffer, "Appending to message to file\n");
    eo15WriteMessageFile( MessageIndex, MessageBuffer, ReturnStatus);
    eo15CloseMessageFile( MessageIndex, ReturnStatus );
  }

  eo15CloseMessageFile( MessageIndex2, ReturnStatus );
  
  eo15CompleteMessageFiles(3, theBuffer );
  
  tsp00_Pathc Working;

  eo15GetWorkingDirectory( Working, ReturnStatus );

  printf("Working directory: %s\n", &Working[0]);

  return 0;
}

#endif /* TEST */

/*===========================================================================*
*  END OF CODE                                                              *
*===========================================================================*/
