/****************************************************************************

  module      : veo381.cpp
  
  -------------------------------------------------------------------------
    
  responsible : ThomasA
      
  special area: Memory Management
  description : Heap
     
  last changed: 2000-01-10  9:34
            
  -------------------------------------------------------------------------
            
  copyright:    Copyright by SAP AG, 1999
              


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/


/*===========================================================================*
*  INCLUDES                                                                 *
*===========================================================================*/

#include "heo381.h"
#include <stdlib.h> // nocheck
#include <malloc.h> // nocheck

typedef ceo381_Chunk* teo381_ChunkPtr;

/* Macros */

#ifdef _DEBUG
#define CHECK_ALLOCATED_CHUNK(H,VICTIM,BLOCKSIZE) VICTIM->CheckAllocatedChunk(H,BLOCKSIZE)
#define CHECK_IN_USE_CHUNK(P) CheckInUseChunk(P)
#else
#define CHECK_ALLOCATED_CHUNK(H,VICTIM,BLOCKSIZE)
#define CHECK_IN_USE_CHUNK(P)
#endif

#define BINBLOCKWIDTH        4   /* bins per block */
#define MAX_SMALLBIN        63
#define MAX_SMALLBIN_SIZE  512
#define SMALLBIN_WIDTH       8

/* size field is or'ed with PREV_INUSE when previous adjacent chunk in use */

#define PREV_INUSE 0x1 

/* size field is or'ed with IS_MMAPPED if the chunk was obtained with mmap() */

#define SIZE_BITS (PREV_INUSE)

#define MALLOC_ALIGNMENT       (sizeof(ceo381_ChunkSizeInfo))
#define MALLOC_ALIGN_MASK      (MALLOC_ALIGNMENT - 1)
#define MINSIZE                (sizeof(ceo381_Chunk))


class ceo381_ChunkSizeInfo {
public :
  teo381_ChunkSize   prev_size;      /* Size of previous chunk (if free). */
  teo381_ChunkSize   size;           /* Size in bytes, including overhead. */
  ceo381_ChunkSizeInfo() : prev_size(0), size(0) {}
};

/*!
  class: ceo381_Chunk

  description: Defined the structure and methods of a chunk or memory of a ceo381_Heap

*/
class ceo381_Chunk : public ceo381_ChunkSizeInfo
{
public :
  teo381_ChunkPtr       next;      /* double links -- used only if free. */
  teo381_ChunkPtr       prev;

#ifdef _DEBUG
  INLINE void CheckAllocatedChunk(ceo381_Heap* heap, teo381_ChunkSize size);
#endif
/*! 
	function: ClearPrevInUseAndGetChunkSize 

	description: returns the size of a chunk
*/
  INLINE teo381_ChunkSize ClearPrevInUseAndGetChunkSize();
/*! 
	function: ChunkSize 

	description: returns the size of a chunk
*/
  INLINE teo381_ChunkSize ChunkSize();
/*! 
	function: ChunkAtOffset

	description: returns a pointer to the chunk starting at offset bytes from the current chunk
*/
  INLINE teo381_ChunkPtr ChunkAtOffset(teo381_ChunkSize offset);
/*! 
	function: ChunkAtNegativeOffset

	description: returns a pointer to the chunk starting at -offset bytes from the current chunk
*/
  INLINE teo381_ChunkPtr ChunkAtNegativeOffset(teo381_ChunkSize offset);
/*!
  function : First

  description : returns the first chunk of free list, used only if 'this' is a free list header
*/
  INLINE teo381_ChunkPtr First() const;
/*!
  function : InUse

  description : returns true, if current chunk is allocated; false if chunk is free
*/
  INLINE bool InUse();
/*!
  function : Last

  description : returns the last chunk of free list, used only if 'this' is a free list header
*/
  INLINE teo381_ChunkPtr Last() const;
/*! 
  function : NextChunk

  description : returns a pointer to the next chunk of current chunk
*/
  INLINE teo381_ChunkPtr NextChunk();
/*!
  function : NextFreeList

  description : returns a pointer to the next free list. Only if current chunk is a free list header.
*/
  INLINE teo381_ChunkPtr NextFreeList();
/*!
  function PrevChunk

  description : returns a pointer to the next chunk of current chunk
*/
  INLINE teo381_ChunkPtr PrevChunk();
/*!
  function : PrevFreeList

  description : returns a pointer to the prev free list. Only if current chunk is a free list header.
*/
  INLINE teo381_ChunkPtr PrevFreeList();
/*!
  function : PrevInUse

  description : returns true, if predecessor of current chunk is in use
*/
  INLINE bool PrevInUse();
/*!
  function : SetHead

  description : writes size of chunk into chunk header
*/
  INLINE void SetHead(teo381_ChunkSize sz);
  /*!
  function : SetHeadAndFoot

  description : writes size of chunk into chunk header
*/
  INLINE void SetHeadAndFoot(teo381_ChunkSize sz);

  INLINE bool InUseBitAtOffset(teo381_ChunkSize offset);

  INLINE void SetInUseAtOffset(teo381_ChunkSize offset);
/*!
  function : ChunkMem

  description : returns a pointer to the user part of current chunk 
*/
  INLINE void* ChunkMem();
/*!
  function : Unlink

  description : removes current chunk from free list
*/
  INLINE void Unlink();
/*
  function : Mem2Chunk

  description : for a given pointer to the user part of a chunk the pointer to the chunk is returned
*/
  INLINE static teo381_ChunkPtr Mem2Chunk(void *mem);
};

/*===========================================================================*
*  IMPLEMENTATION OF ceo381_Chunk methods                                                                *
*===========================================================================*/

INLINE teo381_ChunkSize ceo381_Chunk::ClearPrevInUseAndGetChunkSize() {
  size &= ~(SIZE_BITS);
  return size;
}

INLINE teo381_ChunkSize ceo381_Chunk::ChunkSize() {
  return (size & ~(SIZE_BITS));
}

INLINE teo381_ChunkPtr ceo381_Chunk::ChunkAtOffset(teo381_ChunkSize offset) {
  return ((teo381_ChunkPtr)(((char*)(this)) + (offset)));
}

INLINE teo381_ChunkPtr ceo381_Chunk::ChunkAtNegativeOffset(teo381_ChunkSize offset) {
  return ((teo381_ChunkPtr)(((char*)(this)) - (offset)));
}

INLINE teo381_ChunkPtr ceo381_Chunk::First() const {
  return next;
}

INLINE bool ceo381_Chunk::InUse() {
  return ((((teo381_ChunkPtr)(((char*)(this))+(ChunkSize())))->size) & PREV_INUSE);
}

INLINE teo381_ChunkPtr ceo381_Chunk::Last() const {
  return prev;
}

INLINE teo381_ChunkPtr ceo381_Chunk::NextChunk() {
  return ((teo381_ChunkPtr)( ((char*)(this)) + (ChunkSize()) ));
}

INLINE teo381_ChunkPtr ceo381_Chunk::NextFreeList() {
  return REINTERPRET_CAST (teo381_ChunkPtr,
    REINTERPRET_CAST(char*, this) + 2 * sizeof(teo381_ChunkPtr));
}

INLINE teo381_ChunkPtr ceo381_Chunk::PrevChunk() {
  return ((teo381_ChunkPtr)( ((char*)(this)) - (prev_size) ));
}

INLINE teo381_ChunkPtr ceo381_Chunk::PrevFreeList() {
  return REINTERPRET_CAST (teo381_ChunkPtr,
    REINTERPRET_CAST(char*, this) - 2 * sizeof(teo381_ChunkPtr));
}

INLINE bool ceo381_Chunk::PrevInUse() {
  return (this->size & PREV_INUSE);
}

INLINE void ceo381_Chunk::SetHead(teo381_ChunkSize sz) {
  //    assert(sz >= 8);
  size = sz;
}

INLINE void ceo381_Chunk::SetHeadAndFoot(teo381_ChunkSize sz) {
  this->size = sz | PREV_INUSE;
  ((teo381_ChunkPtr)((char*)(this) + (sz)))->prev_size = sz;
}

INLINE bool ceo381_Chunk::InUseBitAtOffset(teo381_ChunkSize offset) {
  return (((teo381_ChunkPtr)(((char*)(this)) + (offset)))->size & PREV_INUSE);
}

INLINE void ceo381_Chunk::SetInUseAtOffset(teo381_ChunkSize offset) {
  (((teo381_ChunkPtr)(((char*)(this)) + offset))->size |= PREV_INUSE);
}

INLINE void* ceo381_Chunk::ChunkMem() {
  return ((void*)((char*)(this) + sizeof(ceo381_ChunkSizeInfo)));
}

INLINE void ceo381_Chunk::Unlink() {
  teo381_ChunkPtr pPrev = prev;
  teo381_ChunkPtr pNext = next;  
  pPrev->next = pNext;
  pNext->prev = pPrev;
}

INLINE teo381_ChunkPtr ceo381_Chunk::Mem2Chunk(void *mem) {
  return ((teo381_ChunkPtr)((char*)(mem) - sizeof(ceo381_ChunkSizeInfo)));
}

#ifdef _DEBUG

INLINE void ceo381_Chunk::CheckAllocatedChunk(ceo381_Heap* heap, teo381_ChunkSize CheckSize) {
  teo381_ChunkSize sz = CheckSize & ~PREV_INUSE;
  long room = sz - CheckSize;

  heap->CheckInUseChunk(this);

  /* Legal size ... */
  assert((long)sz >= (long)MINSIZE);
  assert((sz & MALLOC_ALIGN_MASK) == 0);
  assert(room >= 0);
  assert(room < (long)MINSIZE);
  /* ... and alignment */
  //assert(aligned_OK(chunk2mem(p)));
  /* ... and was allocated at front of an available chunk */
  assert(this->PrevInUse());
}

#endif

struct TRawNextChunk {
  TRawNextChunk* m_next;
#if !defined(BIT64)
  void*          m_filler;
#endif
  TRawNextChunk(TRawNextChunk* next) : m_next(next) {}
  void* operator new(size_t sz, void* p) {
    return p;
  }
#if defined(USE_OPERATOR_DELETE_EO381)
  void operator delete(void* p, void*q) {}
#endif
  void operator delete(void* p); // cannot be deleted
};

struct TRawChunkFoot {
  TRawChunkFoot() {
    REINTERPRET_CAST(teo381_ChunkPtr, &m_firstChunk )->SetHead(sizeof(m_firstChunk ) | PREV_INUSE);
    REINTERPRET_CAST(teo381_ChunkPtr, &m_secondChunk)->SetHead(sizeof(m_secondChunk) | PREV_INUSE);
  }
  
  void* operator new(size_t sz, void* p) {
    return p;
  }
#if defined(USE_OPERATOR_DELETE_EO381)
  void operator delete(void* p, void*q) {}
#endif
  void operator delete(void* p); // cannot be deleted
  ceo381_ChunkSizeInfo m_firstChunk;
  ceo381_ChunkSizeInfo m_secondChunk;
};

CFreeChunkLists::CFreeChunkLists() {
  int ix;
  int jx = 0;
  
  for (ix = 0; ix < 128; ix++) {
    m_freeLists[jx].m_next = operator[](ix);
    m_freeLists[jx].m_prev = operator[](ix);
    ++jx;
  }
}

INLINE teo381_ChunkPtr CFreeChunkLists::operator[] (int ix) {
  return REINTERPRET_CAST(teo381_ChunkPtr,
    REINTERPRET_CAST(unsigned char*, &m_freeLists[ix]) - sizeof(ceo381_ChunkSizeInfo));
}

INLINE int CFreeChunkLists::GetFreeListIndex(teo381_ChunkSize sz) {
  return (((((unsigned long)(sz)) >> 9) ==    0) ? (((unsigned long)(sz)) >>  3):
  ((((unsigned long)(sz)) >> 9) <=    4) ?  56 + (((unsigned long)(sz)) >>  6):
  ((((unsigned long)(sz)) >> 9) <=   20) ?  91 + (((unsigned long)(sz)) >>  9):
  ((((unsigned long)(sz)) >> 9) <=   84) ? 110 + (((unsigned long)(sz)) >> 12):
  ((((unsigned long)(sz)) >> 9) <=  340) ? 119 + (((unsigned long)(sz)) >> 15):
  ((((unsigned long)(sz)) >> 9) <= 1364) ? 124 + (((unsigned long)(sz)) >> 18):
  126);
}

INLINE int CFreeChunkLists::GetSmallFreeListIndex(teo381_ChunkSize size) {
  return ((tsp00_Uint4) size) >> 3;
}


ceo381_Heap::ceo381_Heap(tsp00_Uint4 FirstMalloc_KB, tsp00_Uint4 Supplement_KB) :
             m_base(NULL),
             m_lastRemainder(NULL),
             m_first(NULL),
             m_firstSize(FirstMalloc_KB), 
             m_supplement(Supplement_KB), 
             m_maxHeap(MAX_INT4_SP00), 
             m_used(0), 
             m_mallocedSize(0),
             m_extends(0),
             m_freeLists(), 
             m_binBlocks(0)
{
  this->CheckConstraints();
}

/*-------------------------------------------------------------------*/

ceo381_Heap::ceo381_Heap(tsp00_Uint4 FirstMalloc_KB, tsp00_Uint4 Supplement_KB, tsp00_Uint4 MaxHeap_KB) :
             m_base(NULL),
             m_lastRemainder(NULL),
             m_first(NULL),
             m_firstSize(FirstMalloc_KB), 
             m_supplement(Supplement_KB), 
             m_maxHeap(MaxHeap_KB), 
             m_used(0), 
             m_mallocedSize(0),
             m_extends(0),
             m_freeLists(), 
             m_binBlocks(0)
{
  this->CheckConstraints();
}

/*-------------------------------------------------------------------*/

INLINE teo381_ChunkSize ceo381_Heap::AlignedSize(teo381_ChunkSize req) {
  return  (((req + (sizeof(teo381_ChunkSize) + MALLOC_ALIGN_MASK)) < 
    (long)(MINSIZE + MALLOC_ALIGN_MASK)) ? MINSIZE : 
  (((req) + (sizeof(teo381_ChunkSize) + MALLOC_ALIGN_MASK)) & ~(MALLOC_ALIGN_MASK)));
}

/*-------------------------------------------------------------------*/

INLINE void ceo381_Heap::FrontLink(teo381_ChunkPtr pChunk, teo381_ChunkSize chunkSize)                                          \
{              
  int             ix;
  teo381_ChunkPtr prv;
  teo381_ChunkPtr nxt;
  if (chunkSize < MAX_SMALLBIN_SIZE) {                                                  
    ix = m_freeLists.GetSmallFreeListIndex(chunkSize);                                                  
    MarkBinBlock(ix);                                                       
    prv = m_freeLists[ix];                                                         
    nxt = prv->next;                                                              
  }                                                                           
  else                                                                        
  {                                                                           
    ix  = m_freeLists.GetFreeListIndex(chunkSize);                                                       
    prv = m_freeLists[ix];                                                         
    nxt = prv->next;                                                              
    if (nxt == prv) {
      MarkBinBlock(ix);
    }
    else {                                                          
      while (nxt != prv && chunkSize < nxt->ChunkSize()) {
        nxt = nxt->next;
      }
      prv = nxt->prev;                                                            
    }                                                                         
  }  
  pChunk->prev = prv;                                                               
  pChunk->next = nxt;
  prv->next    = pChunk;
  nxt->prev    = pChunk;
}

/*-------------------------------------------------------------------*/

INLINE unsigned long ceo381_Heap::Idx2BinBlock(int ix) {
  return ((unsigned long)1 << (ix / BINBLOCKWIDTH));
}

/*-------------------------------------------------------------------*/

INLINE bool ceo381_Heap::IsSmallRequest (teo381_ChunkSize req) {
  return (req < MAX_SMALLBIN_SIZE - SMALLBIN_WIDTH);
}

/*-------------------------------------------------------------------*/

INLINE void ceo381_Heap::MarkBinBlock(int ix) {
  m_binBlocks |= Idx2BinBlock(ix);
}

/*-------------------------------------------------------------------*/
             
void ceo381_Heap::CheckConstraints() {
  if (m_firstSize > MAX_INT4_SP00) {
    m_firstSize = MAX_INT4_SP00;
  }
  if (m_supplement > MAX_INT4_SP00) {
    m_supplement = MAX_INT4_SP00;
  }
  if (m_maxHeap < m_firstSize) {
    m_maxHeap = m_firstSize;
  }
}

/*-------------------------------------------------------------------*/

ceo381_Heap::~ceo381_Heap() {
  freeHeap();
}

/*-------------------------------------------------------------------*/

void ceo381_Heap::FreeMem(teo381_ChunkPtr p) {
  free(p);
}

/*-------------------------------------------------------------------*/

void ceo381_Heap::getHeapInfo(THeapInfo& heapInfo) {
  heapInfo.heapUsed  = m_used;
  heapInfo.allocated = m_mallocedSize;
  heapInfo.extends   = m_extends;
}

/*-------------------------------------------------------------------*/

teo381_ChunkPtr ceo381_Heap::GetMem(size_t sz) {
  return REINTERPRET_CAST(teo381_ChunkPtr, malloc(sz));
}

/*-------------------------------------------------------------------*/

void* ceo381_Heap::allocate (teo381_ChunkSize bytes) {
  teo381_ChunkPtr  victim;          /* inspected/selected chunk */
  teo381_ChunkSize victim_size;     /* its size */
  int              idx;             /* index for free list traversal */
  teo381_ChunkPtr  bin;             /* associated bin */
  teo381_ChunkSize remainder_size;  /* its size */
  unsigned long    block;           /* block traverser bit */
  int              startidx;        /* first bin of a traversed block */
  int              extends = 0;
  teo381_ChunkPtr  q;               /* misc temp */
 
  teo381_ChunkSize nb  = AlignedSize(bytes);  

  /* Check for exact match in a free list */
  do {
    if (IsSmallRequest(nb)) { /* Faster version for small requests */
      /* No traversal or size check necessary for small free lists.  */
      idx    = m_freeLists.GetSmallFreeListIndex(nb);
      q      = m_freeLists[idx];
      victim = q->Last();
      
      /* Also scan the next one, since it would have a remainder < MINSIZE */
      if (victim == q)
      {
        q      = q->NextFreeList();
        victim = q->Last();
      }
      if (victim != q) {
        victim_size   = victim->ChunkSize();
        this->m_used += victim_size;
        victim->Unlink();
        victim->SetInUseAtOffset(victim_size);
        CHECK_ALLOCATED_CHUNK(this, victim, nb);
        return victim->ChunkMem();
      }
      idx += 2; /* Set for bin scan below. We've already scanned 2 free lists. */
    }
    else
    {
      idx    = m_freeLists.GetFreeListIndex(nb);
      q      = m_freeLists[idx];
      
      for (victim = q->Last(); victim != q; victim = victim->prev)
      {
        victim_size = victim->ChunkSize();
        if (victim_size >= nb) {
          if (victim_size - nb >= MINSIZE) {
            if ((m_lastRemainder != NULL) && (m_lastRemainder->ChunkSize() > nb)) {
              /* leave, use last remainder */
              victim      = m_lastRemainder;
              victim_size = victim->ChunkSize();
              goto use_remainder;
            }
          }
          victim->Unlink();
          victim->SetInUseAtOffset(victim_size);
          CHECK_ALLOCATED_CHUNK(this, victim, nb);
          this->m_used += victim_size;
          return victim->ChunkMem();
        }
      }
      ++idx; 
    }
    
    /* Try to use the last split-off remainder */
    
    if (m_lastRemainder != NULL)
    {
      victim      = m_lastRemainder;
      victim_size = victim->ChunkSize();
      if (victim_size >= nb) {
use_remainder :
        remainder_size = victim_size - nb;
        if (remainder_size >= MINSIZE) {
          m_lastRemainder = m_lastRemainder->ChunkAtOffset(nb);
          m_lastRemainder->SetHeadAndFoot(remainder_size);
          victim->SetHead(nb | PREV_INUSE);
          CHECK_ALLOCATED_CHUNK(this, victim, nb);
          this->m_used += nb;
          return victim->ChunkMem();
        }
        else {
          m_lastRemainder = NULL;
          victim->SetInUseAtOffset(victim_size);
          CHECK_ALLOCATED_CHUNK(this, victim, nb);
          this->m_used += victim_size;
          return victim->ChunkMem();
        }
      }
      else {
        m_lastRemainder = NULL;
        FrontLink(victim, victim_size);
      }
    }
    
    /* 
    If there are any possibly nonempty big-enough blocks, 
    search for best fitting chunk by scanning free lists in blockwidth units.
    */

    if ( (block = Idx2BinBlock(idx)) <= m_binBlocks)  
    {
      
      /* Get to the first marked block */
      
      if ( (block & m_binBlocks) == 0) 
      {
        /* force to an even block boundary */
        idx = (idx & ~(BINBLOCKWIDTH - 1)) + BINBLOCKWIDTH;
        block <<= 1;
        while ((block & m_binBlocks) == 0)
        {
          idx += BINBLOCKWIDTH;
          block <<= 1;
        }
      }
      
      /* For each possibly nonempty block ... */
      for (;;)  
      {
        startidx = idx;          /* (track incomplete blocks) */
        q = bin = m_freeLists[idx];
        
        /* For each bin in this block ... */
        do
        {
          /* Find and use first big enough chunk ... */
          
          for (victim = bin->Last(); victim != bin; victim = victim->prev)
          {
            victim_size = victim->ChunkSize();
            if (victim_size >= nb) {
              remainder_size = victim_size - nb;
              if (remainder_size >= MINSIZE) /* split */
              {
                m_lastRemainder = victim->ChunkAtOffset(nb);
                victim->SetHead(nb | PREV_INUSE);
                victim->Unlink();
                m_lastRemainder->SetHeadAndFoot(remainder_size);
                CHECK_ALLOCATED_CHUNK(this, victim, nb);
                this->m_used += nb;
                return victim->ChunkMem(); 
              }
              else {
                victim->SetInUseAtOffset(victim_size);
                victim->Unlink();
                CHECK_ALLOCATED_CHUNK(this, victim, nb);
                this->m_used += victim_size;
                return victim->ChunkMem();
              }
            }
          }
          bin = bin->NextFreeList();
        }
        while ((++idx & (BINBLOCKWIDTH - 1)) != 0);
        
        /* Clear out the block bit. */
        
        do   /* Possibly backtrack to try to clear a partial block */
        {
          if ((startidx & (BINBLOCKWIDTH - 1)) == 0)
          {
            m_binBlocks &= ~block;
            break;
          }
          --startidx;
          q = q->PrevFreeList();
        } while (q->First() == q);
        
        /* Get to the next possibly nonempty block */
        
        if ( (block <<= 1) <= m_binBlocks && (block != 0) ) 
        {
          while ((block & m_binBlocks) == 0)
          {
            idx += BINBLOCKWIDTH;
            block <<= 1;
          }
        }
        else
          break;
      }
    }
    /* Try to extend */
    if (0 == extends) {
      ++extends;
      if (MallocExtend(nb)) {
        continue;
      }
    }
    return NULL;  
  }
  while(1);
}

/*-------------------------------------------------------------------*/

bool ceo381_Heap::MallocExtend(teo381_ChunkSize required) {
  teo381_ChunkPtr  p;
  teo381_ChunkSize sz = (NULL == m_base) ? m_firstSize : m_supplement;
  sz <<= 10;
  if (required + sizeof(TRawNextChunk) + 2 * sizeof(teo381_ChunkSize) + sizeof(TRawChunkFoot) > sz) {
    sz = required + sizeof(TRawNextChunk) + 2 * sizeof(teo381_ChunkSize) + sizeof(TRawChunkFoot);
  }
  if (m_mallocedSize < (m_maxHeap << 10)  - sz) {  
    p = this->GetMem(sz);
  }
  else {
    return false;
  }
  if (p) {
    ++m_extends;
    m_mallocedSize += sz;
    m_first = new(p) TRawNextChunk(m_first);
    sz -= sizeof(TRawNextChunk);
    p = p->ChunkAtOffset(sizeof(TRawNextChunk));
    if (NULL == m_base) {
      m_base  = (char*) m_first;
    }
    else {
      if ((char*) m_first < (char*) m_base) {
        m_base = (char*) m_first;
      }
    }
    char* q = (char*) p + sz - sizeof(TRawChunkFoot);
    new(q) TRawChunkFoot();
    sz -= sizeof(TRawChunkFoot);
    p->SetHead(sz | PREV_INUSE);
    if (sz >= MINSIZE) {
      m_used += sz;
      deallocate(p->ChunkMem());
    }
  }
  return true;
}

/*-------------------------------------------------------------------*/

bool ceo381_Heap::checkHeap() {
  teo381_Size heapUsed = 0;
  for (HeapIterator iter = first(); iter; ++iter) {
    teo381_ChunkPtr p = iter();
    if (p->InUse()) {
      heapUsed += p->ChunkSize();
#ifdef _DEBUG
      CheckInUseChunk(p);
#endif
    }
    else {
#ifdef _DEBUG
      CheckFreeChunk(p);
#endif
    }
  }
#ifdef _DEBUG
  assert(heapUsed == m_used);
#endif
  if (heapUsed != m_used) {
    return false;
  }
  return true;
}

/*-------------------------------------------------------------------*/

teo381_ChunkSize ceo381_Heap::chunkSize(void* p) {
  return ceo381_Chunk::Mem2Chunk(p)->ChunkSize();
}

/*-------------------------------------------------------------------*/

void ceo381_Heap::deallocate(void* mem)
{
  teo381_ChunkPtr  p;               /* chunk corresponding to mem */
  teo381_ChunkSize sz;              /* its size */
  teo381_ChunkPtr  next;            /* next contiguous chunk */
  teo381_ChunkSize nextsz;          /* its size */
  teo381_ChunkSize prevsz;          /* size of previous contiguous chunk */

  if (NULL == mem)  /* free(0) has no effect */
    return;

  p  = ceo381_Chunk::Mem2Chunk(mem);
  
  CHECK_IN_USE_CHUNK(p);
    
  sz      = p->ChunkSize();
  m_used -= sz;
  next    = p->NextChunk();
  nextsz  = next->ClearPrevInUseAndGetChunkSize();
    
  if (!(p->PrevInUse()))                    /* consolidate backward */
  {
    prevsz = p->prev_size;
    p = p->ChunkAtNegativeOffset(prevsz);
    sz += prevsz;
    
    if (p == m_lastRemainder) {             /* keep as last_remainder */
      if (!(next->InUseBitAtOffset(nextsz))) {
        sz += nextsz;
        next->Unlink();
      }
      p->SetHeadAndFoot(sz);
      return;
    }
    else
      p->Unlink();
  }
  
  if (!(next->InUseBitAtOffset(nextsz)))   /* consolidate forward */
  {
    sz += nextsz;
    
    if (next == m_lastRemainder)  /* re-insert last_remainder */
    {
      m_lastRemainder = p;
      p->SetHeadAndFoot(sz);
      return;
    }
    else
      next->Unlink();
  }

  p->SetHeadAndFoot(sz);
  FrontLink(p, sz);  
}

/*-------------------------------------------------------------------*/

/*
  These routines make a number of assertions about the states
  of data structures that should be true at all times. If any
  are not true, it's very likely that a user program has somehow
  trashed memory. (It's also possible that there is a coding error
  in malloc. In which case, please report it!)
*/

#ifdef _DEBUG

void ceo381_Heap::CheckChunk(teo381_ChunkPtr p) 
{ 
  teo381_ChunkSize sz = p->ChunkSize() & ~PREV_INUSE;

  /* Check for legal address ... */
  assert((char*)p >= m_base);
  //if (p != Top()) 
  //  assert((char*)p + sz <= (char*)Top());
  //else
  //  assert((char*)p + sz <= m_base + m_size);
}

/*-------------------------------------------------------------------*/

void ceo381_Heap::CheckFreeChunk(teo381_ChunkPtr p)
{ 
  teo381_ChunkSize sz = p->ChunkSize() & ~PREV_INUSE;
  teo381_ChunkPtr next   = p->ChunkAtOffset(sz);

  CheckChunk(p);

  /* Check whether it claims to be free ... */
  assert(!p->InUse());

  /* Unless a special marker, must have OK fields */
  if ((long)sz >= (long)MINSIZE)
  {
    assert((sz & MALLOC_ALIGN_MASK) == 0);
//    assert(aligned_OK(chunk2mem(p)));
    /* ... matching footer field */
    assert(next->prev_size == sz);
    /* ... and is fully consolidated */
    assert(p->PrevInUse());
    assert (next->InUse());
    
    /* ... and has minimally sane links */
    if (p != m_lastRemainder) {
      assert(p->next->prev == p);
      assert(p->prev->next == p);
    }
  }
  else /* markers are always of size SIZE_SZ */
    assert(sz == sizeof(teo381_ChunkSize)); 
}

void ceo381_Heap::CheckInUseChunk(teo381_ChunkPtr p) 
{ 
  teo381_ChunkPtr next = p->NextChunk();
  CheckChunk(p);

  /* Check whether it claims to be in use ... */
  assert(p->InUse());

  /* ... and is surrounded by OK chunks.
    Since more things can be checked with free chunks than inuse ones,
    if an inuse chunk borders them and debug is on, it's worth doing them.
  */
  if (!p->PrevInUse()) 
  {
    teo381_ChunkPtr prv = p->PrevChunk();
    assert(prv->NextChunk() == p);
    CheckFreeChunk(prv);
  }
  if (!next->InUse()) {
    CheckFreeChunk(next);
  }
}

/*-------------------------------------------------------------------*/

void ceo381_Heap::CheckMallocedChunk(teo381_ChunkPtr p, teo381_ChunkSize s) 
{
  teo381_ChunkSize sz = p->size & ~PREV_INUSE;
  long room = sz - s;

  CheckInUseChunk(p);

  /* Legal size ... */
  assert((long)sz >= (long)MINSIZE);
  assert((sz & MALLOC_ALIGN_MASK) == 0);
  assert(room >= 0);
  assert(room < (long)MINSIZE);

  /* ... and alignment */
  //assert(aligned_OK(p->Chunk2Mem()));


  /* ... and was allocated at front of an available chunk */
  assert(p->PrevInUse());

}

#endif
/*-------------------------------------------------------------------*/

ceo381_Heap::HeapIterator ceo381_Heap::first() {
  HeapIterator iter;
  TRawNextChunk* curr = m_first;
  TRawNextChunk* min  = curr;
  while (NULL != curr) {
    if (curr < min) {
      min = curr;
    }
    curr = curr->m_next;
  }
  iter.m_pHeap     = this;
  if (NULL != min) {
    iter.m_currBlock = min;
    iter.m_currChunk = (teo381_ChunkPtr) ((char*) min + sizeof(*min));
  }
  return iter;
}

/*-------------------------------------------------------------------*/

void ceo381_Heap::freeHeap() {
  TRawNextChunk* toFree;
  while (NULL != m_first) {
    toFree  = m_first;
    m_first = m_first->m_next;
    FreeMem(REINTERPRET_CAST(teo381_ChunkPtr, toFree)); /* PTS 1109469 */
  }
  new(this) ceo381_Heap(m_firstSize, m_supplement, m_maxHeap);
}

/*-------------------------------------------------------------------*/

void ceo381_Heap::setHeapLimit(teo381_Size limit_KB) {
  if ((limit_KB << 10) > m_mallocedSize) {
    m_maxHeap = limit_KB;
  }
}

/*-------------------------------------------------------------------*/

ceo381_Heap::HeapIterator::HeapIterator() : m_pHeap(NULL), m_currBlock(NULL), m_currChunk(NULL)
{
}

/*-------------------------------------------------------------------*/

ceo381_Heap::HeapIterator::operator bool() const {
  return (m_currChunk != NULL);
}

/*-------------------------------------------------------------------*/

void ceo381_Heap::HeapIterator::operator++() {
  m_currChunk = m_currChunk->NextChunk();
  if (m_currChunk->ChunkSize() == sizeof(ceo381_ChunkSizeInfo)) {
    TRawNextChunk* curr = m_pHeap->m_first;
    TRawNextChunk* min = NULL;
    while (NULL != curr) {
      if (curr > m_currBlock) {
        min  = curr;
        curr = curr->m_next;
        while (NULL != curr) {
          if ((curr < min) && (curr > m_currBlock)) {
            min = curr;
          }
          curr = curr->m_next;
        }
          
      }
      else {
        curr = curr->m_next;
      }
    }
    m_currBlock = min;
    if (m_currBlock != NULL) {
      m_currChunk = (teo381_ChunkPtr) ((char*) min + sizeof(*min));
    }
    else {
      m_currChunk = NULL;
    }
  }
}

/*-------------------------------------------------------------------*/

const teo381_ChunkPtr ceo381_Heap::HeapIterator::operator()() const {
  return m_currChunk;
}