/*!***************************************************************************

  module      : veo861.c

  -------------------------------------------------------------------------

  responsible : StefanP

  special area: Diagnostic files
  description : Saving of diagnostic files at restart if the database
                was shutting down abnormally the last time.
                Functions for the client and server site.


  last changed: 2001-02-08  10:58
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin  GPL
  Copyright (C) 2001 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*****************************************************************************/



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include  "gsp00.h"
#include  "geo007_1.h"
#include  "geo001.h"
#include  "heo15.h"
#include  "hsp77.h"
#include  "geo50_0.h"
#include  "geo57.h"
#include  "heo15.h"
#include  "heo861.h"

#if !(defined(_WIN32))     
#include <sys/types.h>
#include <dirent.h>
#include "hen40.h"      /* nocheck */
#endif


/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/
#if defined(_WIN32)
/*--- NT    */
#define     SEPERATOR                   "\\"
#else
/*--- UNIX  */
#define     SEPERATOR                   "/"
#endif

#define     CURRENT_DIRECTORY           "."
#define     PARENT_DIRECTORY            ".."

#define     ARBITRARY_FILE              "*"

#define     LAST_ERROR_MESSAGE_SIZE     64

#define     ERROR_TEXT_LEN              80

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/

/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/
/*--- Structure of the array which has to be generated to clean up  ---*/
/*--- the history entries on disk                                   ---*/
typedef struct
{
    tsp00_Pathc     path;
    time_t          gentime;
}   HISTORY; 

/*---------------------------------------------------------------------------*/


/*===========================================================================*
 *  STATIC/INLINE FUNCTIONS (PROTOTYPES)                                     *
 *===========================================================================*/
static  void                RTEDiag_DeleteDirectory         (tsp00_Char        *path);
  

static  int                 RTEDiag_CompareHistDate         (const void        *history1,
                                                             const void        *history2);

static  tsp00_Bool          RTEDiag_GetDateTimeFromHistory  (tsp00_DbNamec      dbname,
                                                             tsp00_Char        *history,
                                                             time_t            *p_time);

/*===========================================================================*
 *  EXTERN FUNCTIONS                                                         *
 *===========================================================================*/

/***************************************************************/
#if defined(_WIN32)
/***************************************************************/

/* For the documentation see header file */
extern void eo861_OsError    (tsp00_Char        *routine,
                              tsp00_Char        *callFunction,
                              tsp00_Char        *zusText,
                              teo001_SystemRc    rc,
                              tsp00_Uint4        type)  
{
   /*===========================================================================*
    *  Locals                                                                   *
    *===========================================================================*/
    tsp00_Char    lastErrorMessage[LAST_ERROR_MESSAGE_SIZE];
    tsp00_Char    buf[ERROR_TEXT_LEN];


   /*===========================================================================*
    *  Instructions                                                             *
    *===========================================================================*/
    /* Error message on Win32 are sometimes much much too long so we only report Error code... */
    sp77sprintf(lastErrorMessage, sizeof(lastErrorMessage) - 1, "%s, [%d]", zusText == NULL ? "" : zusText,
                rc == UNDEF_ERROR ? GetLastError() : rc);

    switch (type)
    {
        case INFO_TYPE:
            MSGD((INFO_DIAG_HIST_OS_ERROR, routine, callFunction == NULL ? "" : callFunction, lastErrorMessage));
            break;

        case ERR_TYPE:
            MSGD((ERR_DIAG_HIST_OS_ERROR, routine, callFunction == NULL ? "" : callFunction, lastErrorMessage));
            break;

        default:
            sp77sprintf (buf, sizeof(buf) - 1, "eo861_OsError: Unknown error type %u", type);
            MSGD((ERR_ERROR_TEXT, buf));
    }

    return;
}


/* For the documentation see header file */
extern  void  eo861_DeleteOldHistories   (tsp00_Uint4       numRemainHist,
                                          tsp00_DbNamec     dbname,
                                          tsp00_Pathc       crashHistoryPath)
{
/*===========================================================================*
 *  Locals                                                                   *
 *===========================================================================*/

    static  tsp00_Char          routine[] = "eo861_DeleteOldHistories";
            tsp00_Pathc         searchFile;
            tsp00_Pathc         tmpPath;
            HISTORY            *pHistory = NULL; 
            tsp00_Uint4         ixHistory = 0;
            teo001_SystemRc     rc = NO_ERROR_EO001;
            BOOL                nextFile = true;
            WIN32_FIND_DATA     fileData;
            tsp00_Uint4         numHist;
            teo15_Handle        hFile;


/*===========================================================================*
 *  Instructions                                                             *
 *===========================================================================*/
    strcpy (tmpPath, crashHistoryPath);
    strcat (tmpPath, SEPERATOR);
    
    strcpy (searchFile, tmpPath);
    strcat (searchFile, dbname);
    strcat (searchFile, ARBITRARY_FILE);
  
    hFile = FindFirstFile (searchFile, &fileData);
    if (hFile == INVALID_HANDLE_EO15)
    {
        eo861_OsError (routine, "FindFirstFile", searchFile, UNDEF_ERROR, INFO_TYPE);
        return;
    }

    numHist = numRemainHist + 1;

    /*--- Allocate Memory for the history list 'pHistory' ---*/
    if (ALLOC_MEM_EO57 (&pHistory, numHist * sizeof(HISTORY)) != NO_ERROR_EO001)
    {
        MSGD ((ERR_MALLOC_FAILED_FOR ,"Diagnose history list"));
    }
    
    
    /* Collecting the histories which should remain on disk */
    if (numRemainHist > 0)  /* Of course, this is only to be done if at least one history should remain */
    {
        do
        {
            if (0 != (fileData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY))
            {
                if (RTEDiag_GetDateTimeFromHistory (dbname, fileData.cFileName, &pHistory[ixHistory].gentime))
                {
                    strcpy (pHistory[ixHistory].path, tmpPath);
                    strcat (pHistory[ixHistory].path, fileData.cFileName);
                    ++ixHistory;
                }
            }
        }
        while ((nextFile = FindNextFile (hFile, &fileData)) && ixHistory < numRemainHist);
    }

    if (!nextFile)
    {
        rc = GetLastError ();
        if (rc != ERROR_SUCCESS && rc != ERROR_NO_MORE_FILES)
        {
            eo861_OsError (routine, "FindNextFile", searchFile, rc, ERR_TYPE);
        }

        if (!FindClose (hFile))
        {
            eo861_OsError (routine, "FindClose", searchFile, UNDEF_ERROR, ERR_TYPE);
        }

        if ((rc = FREE_MEM_EO57 (pHistory)) != NO_ERROR_EO001)
        {
            MSGD ((ERR_CANT_FREE_MEMORY ,rc));
        }
        return;
    }

    /* ixHistory == numRemainHist and next file found */
    do
    {
        if (0 != (fileData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY))
        {
            /* The following conversion of the date and time makes even sense for numHist = 1 because  */
            /* there also takes place an check of the rigth format of the directory name. Therewith we */
            /* can be sure to really delete an history directory                                       */
            if (RTEDiag_GetDateTimeFromHistory (dbname, fileData.cFileName, &pHistory[numRemainHist].gentime))
            {
                strcpy (pHistory[numRemainHist].path, tmpPath);
                strcat (pHistory[numRemainHist].path, fileData.cFileName);
                if (numHist > 1) /* we need no sorting for one element */
                {
                    /* Since the first numHist-1 elements are already sorted it may be not the      */
                    /* quickest way of sorting here. But at this point it surely does not matter.   */
                    qsort (pHistory, numHist, sizeof(HISTORY), RTEDiag_CompareHistDate);
                }

                /* delete the oldest history in the list */
                RTEDiag_DeleteDirectory (pHistory[numRemainHist].path);
            }
        }
    }
    while (FindNextFile (hFile, &fileData));

    rc = GetLastError ();
    if (rc != ERROR_SUCCESS && rc != ERROR_NO_MORE_FILES)
    {
        eo861_OsError (routine, "FindNextFile", searchFile, rc, ERR_TYPE);
    }

    if (!FindClose (hFile))
    {
        eo861_OsError (routine, "FindClose", searchFile, UNDEF_ERROR, ERR_TYPE);
    }

    if ((rc = FREE_MEM_EO57 (pHistory)) != NO_ERROR_EO001)
    {
        MSGD ((ERR_CANT_FREE_MEMORY, rc));
    }
    return;
}

/***************************************************************/
#else /* UNIX, LINUX */
/***************************************************************/
/* For the documentation see header file */
extern void eo861_OsError    (tsp00_Char        *routine, 
                                tsp00_Char        *callFunction,
                                tsp00_Char        *zusText,
                                teo001_SystemRc    dummy,
                                tsp00_Uint4        type) 
{
/*===========================================================================*
 *  Locals                                                                   *
 *===========================================================================*/
    tsp00_Char        lastErrorMessage[LAST_ERROR_MESSAGE_SIZE];
    tsp00_Char        buf[ERROR_TEXT_LEN];

/*===========================================================================*
 *  Instructions                                                             *
 *===========================================================================*/
    dummy = dummy;

    sp77sprintf (lastErrorMessage, sizeof(lastErrorMessage) - 1, "%s, [%d] %s", 
                 zusText == NULL ? "" : zusText, errno, sqlerrs());

    switch (type)
    {
        case INFO_TYPE:
            MSGD((INFO_DIAG_HIST_OS_ERROR, routine, callFunction == NULL ? "" : callFunction, lastErrorMessage));
            break;

        case ERR_TYPE:
            MSGD((ERR_DIAG_HIST_OS_ERROR, routine, callFunction == NULL ? "" : callFunction, lastErrorMessage));
            break;

        default:
            sp77sprintf (buf, sizeof(buf) - 1, "eo861_OsError: Unknown error type %u", type);
            MSGD((ERR_ERROR_TEXT, buf));
    }

    return;
}

/* For the documentation see header file */
extern  void  eo861_DeleteOldHistories (tsp00_Uint4       numRemainHist,
                                          tsp00_DbNamec     dbname,
                                          tsp00_Pathc       crashHistoryPath)
{
/*===========================================================================*
 *  Locals                                                                   *
 *===========================================================================*/

    static  tsp00_Char          routine[] = "eo861_DeleteOldHistories";
            tsp00_Pathc         tmpPath;
            tsp00_Uint4         numHist;
            DIR                *hFile = NULL;
            struct dirent      *pfileData = NULL;
            size_t              len_dbname;
            HISTORY            *pHistory = NULL;
            tsp00_Uint4         ixHistory = 0;
            teo001_SystemRc     rc = NO_ERROR_EO001;
            struct stat         fileInfo;
            tsp00_Pathc         histDirectory;

/*===========================================================================*
 *  Instructions                                                             *
 *===========================================================================*/
    strcpy (tmpPath, crashHistoryPath);
    strcat (tmpPath, SEPERATOR);

    len_dbname = strlen (dbname);

    hFile = opendir(tmpPath);

    if (NULL == hFile)
    {
        eo861_OsError (routine, "opendir", crashHistoryPath, UNDEF_ERROR, INFO_TYPE);
        return;
    }

    numHist = numRemainHist + 1;

    /*--- Allocate Memory for the history list 'pHistory' ---*/
    if (ALLOC_MEM_EO57 ((void**)(&pHistory), numHist * sizeof(HISTORY)) != NO_ERROR_EO001)
    {                   
        MSGD ((ERR_MALLOC_FAILED_FOR ,"Diagnose history list"));
    }

    errno = NO_ERROR_EO001;
    /* Collecting the histories which should remain on disk */
    if (numRemainHist > 0)  /* Of course, this is only to be done if at least one history should remain */
    {
        do
        {
            errno = NO_ERROR_EO001;
            pfileData = readdir(hFile);
            if (pfileData != NULL && strlen (pfileData->d_name) >= len_dbname 
                && !strncmp (pfileData->d_name, dbname, len_dbname)) 
            {
                strcpy (histDirectory, tmpPath);
                strcat (histDirectory, pfileData->d_name);
                if (lstat(histDirectory, &fileInfo) != 0)
                {
                    eo861_OsError (routine, "lstat", histDirectory, UNDEF_ERROR, ERR_TYPE);
                    return;
                }

                if (S_ISDIR(fileInfo.st_mode))
                {
                    if (RTEDiag_GetDateTimeFromHistory (dbname, pfileData->d_name, &pHistory[ixHistory].gentime))
                    {
                        strcpy (pHistory[ixHistory].path, histDirectory);
                        ++ixHistory;
                    }
                }
            }
        }
        while (pfileData != NULL && ixHistory < numRemainHist);
    }

    if (NO_ERROR_EO001 != errno)
    {
        eo861_OsError (routine, "readdir", "", UNDEF_ERROR, ERR_TYPE);
        if (closedir(hFile) != 0)
        {
            eo861_OsError (routine, "closedir", crashHistoryPath, UNDEF_ERROR, INFO_TYPE);
        }

        if ((rc = FREE_MEM_EO57 (pHistory)) != NO_ERROR_EO001)
        {
            MSGD ((ERR_CANT_FREE_MEMORY, rc));
        }
        return;
    }

    if (numRemainHist > 0 && pfileData == NULL)
    {
        if (closedir(hFile) != 0)
        {
            eo861_OsError (routine, "closedir", crashHistoryPath, UNDEF_ERROR, INFO_TYPE);
        }

        if ((rc = FREE_MEM_EO57 (pHistory)) != NO_ERROR_EO001)
        {
            MSGD ((ERR_CANT_FREE_MEMORY, rc));
        }
        return;

    }


    /* ixHistory == numRemainHist and next file found */
    do
    {
        errno = NO_ERROR_EO001;
        pfileData = readdir(hFile);
        if (pfileData != NULL && strlen (pfileData->d_name) >= len_dbname 
            && !strncmp (pfileData->d_name, dbname, len_dbname)) 
        {
            strcpy (histDirectory, tmpPath);
            strcat (histDirectory, pfileData->d_name);
            if (lstat(histDirectory, &fileInfo) != 0)
            {
                eo861_OsError (routine, "lstat", histDirectory, UNDEF_ERROR, ERR_TYPE);
                return;
            }

            if (S_ISDIR(fileInfo.st_mode))
            {
                /* The following conversion of the date and time makes even sense for numHist = 1 because  */
                /* there also takes place an check of the rigth format of the directory name. Therewith we */
                /* can be sure to really delete an history directory                                       */
                if (RTEDiag_GetDateTimeFromHistory (dbname, pfileData->d_name, &pHistory[numRemainHist].gentime))
                {
                    strcpy (pHistory[numRemainHist].path, histDirectory);
                    if (numHist > 1) /* we need no sorting for one element */
                    {
                        /* Since the first numHist-1 elements are already sorted it may be not the              */
                        /* quickest way of sorting at this point. But at this point it surely does not matter.  */
                        qsort (pHistory, numHist, sizeof(HISTORY), RTEDiag_CompareHistDate);
                    }

                    /* delete the oldest history in the list */
                    RTEDiag_DeleteDirectory (pHistory[numRemainHist].path);
                }
            }
        }
    }
    while (pfileData != NULL);

    if (NO_ERROR_EO001 != errno)
    {
        eo861_OsError (routine, "readdir", "", UNDEF_ERROR, ERR_TYPE);
        if (closedir(hFile) != 0)
        {
            eo861_OsError (routine, "closedir", crashHistoryPath, UNDEF_ERROR, INFO_TYPE);
        }
    }

    if ((rc = FREE_MEM_EO57 (pHistory)) != NO_ERROR_EO001)
    {
        MSGD ((ERR_CANT_FREE_MEMORY, rc));
    }
    return;
}


/*!
  EndChapter: UNIX only
 */ 

/***************************************************************/
#endif
/***************************************************************/

/*===========================================================================*
 *  STATIC FUNCTIONS                                                         *
 *===========================================================================*/

/*!
  Chapter: RTE Diagnostic (Common, NT and UNIX)
 */
/*!
  Function:     RTEDiag_CompareHistDate (static)
  description:  Input function for qsort. Compares the time of two elements of type HISTORY.
                .
  arguments:
    history1    [in]    Pointer to a HISTORY-Element 
    history2    [in]    Pointer to a HISTORY-Element 


  return value: -1      history1 > history2
                 0      history1 = history2
                 1      history1 < history2
*/
static  int  RTEDiag_CompareHistDate    (const void  *history1,
                                         const void  *history2) 
{
/*===========================================================================*
 *  Locals                                                                   *
 *===========================================================================*/

/*===========================================================================*
 *  Instructions                                                             *
 *===========================================================================*/
    if (((HISTORY *)history1)->gentime > ((HISTORY *)history2)->gentime)
    {
        return (-1);
    }
    else if (((HISTORY *)history1)->gentime < ((HISTORY *)history2)->gentime)
    {
        return (1);
    }
    
    return (0);
}


/*!
  Function:     RTEDiag_GetDateTimeFromHistory (static)
  description:  The function converts the date and time of a history name of the following format
                    
                      <dbname>_YYYYMMDD_HH-MM-SS

                into a value of type time_t.
                .
  arguments:
    dbname      [in]    Name of the database
    history     [in]    Pointer to the name of the history directory
    p_time      [out]   Pointer to converted date and time

  return value:     true    Format of the history is correct and conversion succeeded
                    false   Conversion failed
*/
static  tsp00_Bool  RTEDiag_GetDateTimeFromHistory    (tsp00_DbNamec        dbname,
                                                       tsp00_Char          *history,
                                                       time_t              *p_time) 
{
/*===========================================================================*
 *  Locals                                                                   *
 *===========================================================================*/
    struct tm       timestruct;
    tsp00_Char     *phistory;
    tsp00_Char      buf[HIST_LEN_YEAR + 1];

#define     OFFSET_YEAR                 1900    /* Offset is needed to determinate local time */
#define     OFFSET_MONTH                1

#define     PUTDATETIMEVALUE(dt_elem, dt_len, dt_value, dt_off)         \  
            {                                                           \
                    size_t          offset = 0;                             \
                    size_t          length = 0;                             \
                                                                        \
                                                                        \
                offset = strcspn (HIST_NAME_SKEL, (dt_elem));           \
                length = strspn (HIST_NAME_SKEL + offset, (dt_elem));   \
                if  (length != (dt_len))                                \
                {                                                       \
                    return (false);                                     \
                }                                                       \   
                                                                        \
                if (offset <= strlen (HIST_NAME_SKEL) - (dt_len))       \
                {                                                       \
                    memcpy (buf, phistory + offset, (dt_len));          \
                    buf[(dt_len)] = '\0';                               \
                    (dt_value) = atoi(buf) - (dt_off);                  \
                }                                                       \
                else                                                    \
                {                                                       \
                   return (false);                                      \
                }                                                       \     
       	                                                                \
            }                                                           \

        

/*===========================================================================*
 *  Instructions                                                             *
 *===========================================================================*/
    memset (buf, '\0', sizeof (buf));
    memset (&timestruct, 0, sizeof (timestruct));

    phistory = history + strlen (dbname);

    /* Checking the length of the History name */
    if (strlen (phistory) != strlen (HIST_NAME_SKEL))
    {
        return (false);
    }

    PUTDATETIMEVALUE (HIST_YEAR, HIST_LEN_YEAR,     timestruct.tm_year,     OFFSET_YEAR);
    PUTDATETIMEVALUE (HIST_MONTH, HIST_LEN_MONTH,   timestruct.tm_mon,      OFFSET_MONTH);
    PUTDATETIMEVALUE (HIST_DAY, HIST_LEN_DAY,       timestruct.tm_mday,     0);
    PUTDATETIMEVALUE (HIST_HOUR, HIST_LEN_HOUR,     timestruct.tm_hour,     0);
    PUTDATETIMEVALUE (HIST_MINUTE, HIST_LEN_MINUTE, timestruct.tm_min,      0);
    PUTDATETIMEVALUE (HIST_SECOND, HIST_LEN_SECOND, timestruct.tm_sec,      0);

    *p_time = mktime (&timestruct);

    if ((time_t) -1 == *p_time)
    {
        MSGD ((ERR_TIME_CONV, history));
        return (false);
    }

    return (true);
}

/*!
  EndChapter: RTE Diagnostic
 */

/***************************************************************/
#if defined(_WIN32)
/***************************************************************/
/*!
  Chapter: NT only 
 */

/*!
  Function:     RTEDiag_DeleteDirectory (static)
  description:  The specified directory is deleted. All files and subdirectories
                are beeing deleted also.
                .
  arguments:
    path    [in]    Pathname of the directory to be deleted


  return value: none 
  
*/
static  void  RTEDiag_DeleteDirectory (tsp00_Char *path)
{
/*===========================================================================*
 *  Locals                                                                   *
 *===========================================================================*/
    static  tsp00_Char          routine[] = "RTEDiag_DeleteDirectory";
            teo15_Handle        hFile;
            WIN32_FIND_DATA     fileData;
            tsp00_Pathc         searchFile;
            tsp00_Pathc         tmpPath;
            tsp00_Pathc         delFile;
            teo001_SystemRc     rc = NO_ERROR_EO001;

/*===========================================================================*
 *  Instructions                                                             *
 *===========================================================================*/
  strcpy (tmpPath, path);
  strcat (tmpPath, SEPERATOR);


  strcpy (searchFile, tmpPath);
  strcat (searchFile, ARBITRARY_FILE);
  hFile = FindFirstFile (searchFile, &fileData);
  if (hFile == INVALID_HANDLE_EO15)
  {
    eo861_OsError (routine, "FindFirstFile", path, UNDEF_ERROR, INFO_TYPE);
    return;
  }
  
  do
  {
    strcpy (delFile, tmpPath);
    strcat (delFile, fileData.cFileName);
    if (0 != (fileData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY))
    {
      /* Ignore '.' and '..' */
      if (strcmp (fileData.cFileName, CURRENT_DIRECTORY) &&  
          strcmp (fileData.cFileName, PARENT_DIRECTORY))
      {
        RTEDiag_DeleteDirectory (delFile);
      }
    }
    else
    {
      if(!DeleteFile(delFile))
      {
          eo861_OsError (routine, "DeleteFile", fileData.cFileName, UNDEF_ERROR, ERR_TYPE);
      }
    }
  }
  while (FindNextFile (hFile, &fileData));

  rc = GetLastError ();
  if (rc != ERROR_SUCCESS && rc != ERROR_NO_MORE_FILES)
  {
      eo861_OsError (routine, "FindNextFile", path, rc, ERR_TYPE);
  }

  if (!FindClose (hFile))
  {
      eo861_OsError (routine, "FindClose", path, UNDEF_ERROR, ERR_TYPE);
  }

  if(!RemoveDirectory (path))
  {
      eo861_OsError (routine, "RemoveDirectory", path, UNDEF_ERROR, ERR_TYPE);
  }

  return;
}


/*!
  EndChapter: NT only
 */

/***************************************************************/
#else /* UNIX, LINUX */
/***************************************************************/
/*!
  Chapter: UNIX only
 */

/*!
  Function:     RTEDiag_DeleteDirectory (static)
  description:  The specified directory is deleted. All files and subdirectories
                are beeing deleted also (symbolic links are beeing ignored).
                .
  arguments:
    path    [in]    Pathname of the directory to be deleted


  return value: none 
  
*/
static  void  RTEDiag_DeleteDirectory (tsp00_Char *path)
{
/*===========================================================================*
 *  Locals                                                                   *
 *===========================================================================*/
    static  tsp00_Char      routine[] = "RTEDiag_DeleteDirectory";
            DIR            *hFile = NULL;
            struct dirent  *pfileData = NULL;
            tsp00_Pathc     delFile;
            tsp00_Pathc     tmpPath;
            struct stat     fileInfo;

/*===========================================================================*
 *  Instructions                                                             *
 *===========================================================================*/
    strcpy (tmpPath, path);
    strcat (tmpPath, SEPERATOR);

    hFile = opendir(path);

    if (NULL == hFile)
    {
        eo861_OsError (routine, "opendir", path, UNDEF_ERROR, INFO_TYPE);
        return;
    }

    do
    {
        errno = NO_ERROR_EO001;
        pfileData = readdir(hFile);
        if (pfileData != NULL) 
        {
            strcpy (delFile, tmpPath);
            strcat (delFile, pfileData->d_name);
            if (lstat(delFile, &fileInfo) != 0)
            {
                eo861_OsError (routine, "lstat", delFile, UNDEF_ERROR, ERR_TYPE);
                
                if (closedir(hFile) != 0)
                {
                    eo861_OsError (routine, "closedir", path, UNDEF_ERROR, INFO_TYPE);
                }
                return;
            }

            if (S_ISDIR(fileInfo.st_mode))
            {
                /* Ignore '.' and '..' */
                if (strcmp (pfileData->d_name, CURRENT_DIRECTORY) &&  
                    strcmp (pfileData->d_name, PARENT_DIRECTORY))
                {
                    RTEDiag_DeleteDirectory (delFile);
                }
            }
            else if (S_ISLNK(fileInfo.st_mode))
            {   /* Ignore symbolic links */
                ; 
            }
            else
            {
                if (unlink (delFile) != 0)
                {
                    eo861_OsError (routine, "unlink", delFile, UNDEF_ERROR, ERR_TYPE);
                    if (closedir(hFile) != 0)
                    {
                        eo861_OsError (routine, "closedir", path, UNDEF_ERROR, INFO_TYPE);
                    }
                    return;                
                }
            }
        }
    }
    while (pfileData != NULL);

    if (NO_ERROR_EO001 != errno)
    {
        eo861_OsError (routine, "readdir", "", UNDEF_ERROR, ERR_TYPE);
        if (closedir(hFile) != 0)
        {
            eo861_OsError (routine, "closedir", path, UNDEF_ERROR, INFO_TYPE);
        }
        return;
    }

    if (rmdir(path) != 0)
    {
        eo861_OsError (routine, "rmdir", path, UNDEF_ERROR, ERR_TYPE);
    }

    if (closedir(hFile) != 0)
    {
        eo861_OsError (routine, "closedir", path, UNDEF_ERROR, INFO_TYPE);
    }

    return;
}

/*!
  EndChapter: UNIX only
 */ 

/***************************************************************/
#endif
/***************************************************************/

/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/
