/*
  -----------------------------------------------------------------------------
 
  module: vls07.cpp
 
  -----------------------------------------------------------------------------
 
  responsible:  SteffenS

  special area: Replication Manager
 
  description:  Numerical and code conversion functions
                
  version:      7.3.

  -----------------------------------------------------------------------------
 
  copyright:    SAP AG, 1999-2003
 
  -----------------------------------------------------------------------------


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*/

#include <assert.h>

#include "gls00.h"
#include "hls07.h"
#include "hls98msg.h"


/*
  -----------------------------------------------------------------------------
 
  function:     ls07ConvertHexToByte

  -----------------------------------------------------------------------------

  description:  Converts a user given hexadecimal value into an byte string.

                The function allocates memory for the value to return which must
                be freed by the caller.

  arguments:    pszHexValue     [IN]    - pointer, hexadecimal value
                lHexValueLen    [IN]    - length of hex value
                pszByteValue    [OUT]   - pointer to pointer, converted value NOT
                                          null terminated
                lByteValueLen   [OUT]   - Length of converted value

  returnvalue:  errOK_els00             - NO errors
                errInvalidHex_els98     - invalid hex value (contains other characters
                                          than 0-9 and a-f)
  -----------------------------------------------------------------------------
*/
tsp00_Int4
ls07ConvertHexToByte(tsp00_Addr  pszHexValue,
                     tsp00_Int4  lHexValueLen,
                     tsp00_Addr  *pszByteValue,
                     tsp00_Int4& lByteValueLen)
{
    ROUTINE_DBG_MEO00 ("ls07ConvertHexToByte");

    //assert(NULL != pszHexValue);

    tsp00_Int4 RetValue = errOK_els00;

    if (0 != lHexValueLen)
    {
        bool        bInvalidValue = false;

        tsp00_Addr  pszMyValue = new char[lByteValueLen = 2*lHexValueLen];      // No terminating 0!!!

        s41pbyte(pszMyValue, 1, lByteValueLen, pszHexValue, 1, lHexValueLen, bInvalidValue);
        if (true == bInvalidValue)
        {
            delete [] pszMyValue;
            RetValue = errInvalidHex_els98;
        }
        else
        {
            *pszByteValue = pszMyValue;
        }
    }

    return RetValue;
}
//  ls07ConvertHexToByte()



/*
  -----------------------------------------------------------------------------
 
  function:     ls07ConvertByteToHex

  -----------------------------------------------------------------------------

  description:  Converts a user given byte stream to a hexadecimal stream

                The function allocates memory for the value to return which must
                be freed by the caller.

  arguments:    pszByteValue    [IN]   - pointer, value NOT null terminated
                lByteValueLen   [IN]   - Length of value
                pszHexValue     [OUT]  - pointer to pointer, hexadecimal (converted)
                                         value. The value is NULL TERMINATED!!!
                lHexValueLen    [OUT]  - reference, length of hex value
                

  returnvalue:  errOK_els00             - NO errors
                errInvalidHex_els98     - invalid hex value (contains other characters
                                          than 0-9 and a-f)
  -----------------------------------------------------------------------------
*/
tsp00_Int4
ls07ConvertByteToHex(tsp00_Addr   pszByteValue,
                     tsp00_Int4   lByteLen,
                     tsp00_Addr   *pszHexValue,
                     tsp00_Int4 & lHexLen)
{
    ROUTINE_DBG_MEO00 ("ls07ConvertByteToHex");

    tsp00_Int4 rc         = errOK_els00;
    bool       bTruncated = false;

    tsp00_Int4 lMyLen = 2*lByteLen;
    tsp00_Addr pszMyValue = new char[2*lByteLen + 1];

    
    s40gbyte(pszByteValue, 1, lByteLen, pszMyValue, 1, lMyLen, bTruncated);

    pszMyValue[lMyLen] = '\0';

    *pszHexValue = pszMyValue;
    lHexLen      = lMyLen;

    return rc;
}
// ls07ConvertByteToHex()


/*
  -----------------------------------------------------------------------------
 
  function:     ls07ConvertByteToHexEx

  -----------------------------------------------------------------------------

  description:  Converts a user given byte stream to a hexadecimal stream

                Die funktion erwartet einen string, der den hexwert aufnehmen
                kann in der richtigen groesse ->
                2 * (laenge des zu konvertierenden wertes)

  arguments:    pszByteValue  [IN]   - pointer, value NOT null terminated
                lByteValueLen [IN]   - Length of value
                pszHexValue   [OUT]  - pointer, hexadecimal (converted)
                                       value. The value is NOT NULL TERMINATED!!!

  returnvalue:  No
  -----------------------------------------------------------------------------
*/
void
ls07ConvertByteToHexEx(tsp00_Addr pszByteValue,
                       tsp00_Int4 lByteLen,
                       tsp00_Addr pszHexValue)
{
    ROUTINE_DBG_MEO00 ("ls07ConvertByteToHexEx");

    bool       bTruncated = false;
    tsp00_Int4 lMyLen     = lByteLen + lByteLen;
    
    s40gbyte(pszByteValue, 1, lByteLen, pszHexValue, 1, lMyLen, bTruncated);

    return;
}
// ls07ConvertByteToHexEx()


/*
  -----------------------------------------------------------------------------
 
  function:     ls07Swap2BIntToChar

  -----------------------------------------------------------------------------

  description:  Vertauscht die reihenfolge der ziffern eines 2 byte integers (swap)
                und speichert den wert als 2 byte char.

  arguments:    nValue     [IN]    - value to swap
                SSwap      [IN]    - source swap behavior (little endian, big endian)
                DSwap      [OUT]   - destinatin swap behavior
                pszBuffer  [OUT]   - pointer, buffer for swapped character value

  returnvalue:  No
  -----------------------------------------------------------------------------
*/
void
ls07Swap2BIntToChar(tsp00_Int2          nValue,
                    tsp00_SwapKind_Enum SSwap,
                    tsp00_SwapKind_Enum DSwap,
                    tsp00_Addr          pszBuffer)
{
    ROUTINE_DBG_MEO00 ("ls07Swap2BIntToChar");

    assert(2 == sizeof(nValue));

    if (SSwap == DSwap)
    {
        memcpy( pszBuffer, &nValue, sizeof(nValue) );
    }
    else
    {
        pszBuffer[0] = (REINTERPRET_CAST(char*,&nValue))[1];
        pszBuffer[1] = (REINTERPRET_CAST(char*,&nValue))[0];

    }
}
//  ls07Swap2BIntToChar()


/*
  -----------------------------------------------------------------------------
 
  function:     ls07Swap4BIntToChar

  -----------------------------------------------------------------------------
*/
void
ls07Swap4BIntToChar(tsp00_Uint4         nValue,
                    tsp00_SwapKind_Enum SSwap,
                    tsp00_SwapKind_Enum DSwap,
                    tsp00_Addr          pszBuffer)
{
    ROUTINE_DBG_MEO00 ("ls07Swap4BIntToChar");

    if (SSwap == DSwap)
    {
        memcpy( pszBuffer, &nValue, sizeof(nValue) );
    }
    else if ( ((sw_normal == SSwap) && (sw_full_swapped == DSwap)) ||
              ((sw_normal == DSwap) && (sw_full_swapped == SSwap)) )
    {
        pszBuffer[0] = (REINTERPRET_CAST(char*,&nValue))[3];
        pszBuffer[1] = (REINTERPRET_CAST(char*,&nValue))[2];
        pszBuffer[2] = (REINTERPRET_CAST(char*,&nValue))[1];
        pszBuffer[3] = (REINTERPRET_CAST(char*,&nValue))[0];
    }
    else if ( ((sw_normal == SSwap) && (sw_part_swapped == DSwap)) ||
              ((sw_normal == DSwap) && (sw_part_swapped == SSwap)) )
    {
        pszBuffer[0] = (REINTERPRET_CAST(char*,&nValue))[1];
        pszBuffer[1] = (REINTERPRET_CAST(char*,&nValue))[0];
        pszBuffer[2] = (REINTERPRET_CAST(char*,&nValue))[3];
        pszBuffer[3] = (REINTERPRET_CAST(char*,&nValue))[2];
    }
    else if ( ((sw_full_swapped == SSwap) && (sw_part_swapped == DSwap)) ||
              ((sw_full_swapped == DSwap) && (sw_part_swapped == SSwap)) )
    {
        pszBuffer[0] = (REINTERPRET_CAST(char*,&nValue))[2];
        pszBuffer[1] = (REINTERPRET_CAST(char*,&nValue))[3];
        pszBuffer[2] = (REINTERPRET_CAST(char*,&nValue))[0];
        pszBuffer[3] = (REINTERPRET_CAST(char*,&nValue))[1];
    }
}
//  ls07Swap4BIntToChar()

/*
  -----------------------------------------------------------------------------
 
  function:     ls07Swap4BHexToLong

  -----------------------------------------------------------------------------
*/
tsp00_Int4 ls07Swap4BHexToLong(tsp00_Addr           pszValue,
                               tsp00_SwapKind_Enum  SSwap,
                               tsp00_SwapKind_Enum  DSwap)
{
    ROUTINE_DBG_MEO00 ("ls07Swap4BHexToLong");

    union IntegerUnion
    {
        tsp00_Int4 lIntValue;
        char       szIntValue[4];
    } u;

    
    if (SSwap == DSwap)
    {
        memcpy( u.szIntValue, pszValue, sizeof(u.lIntValue) );
    }
    else if ( ((sw_normal == SSwap) && (sw_full_swapped == DSwap)) ||
              ((sw_normal == DSwap) && (sw_full_swapped == SSwap)) )
    {
        u.szIntValue[0] = pszValue[3];
        u.szIntValue[1] = pszValue[2];
        u.szIntValue[2] = pszValue[1];
        u.szIntValue[3] = pszValue[0];
    }
    else if ( ((sw_normal == SSwap) && (sw_part_swapped == DSwap)) ||
              ((sw_normal == DSwap) && (sw_part_swapped == SSwap)) )
    {
        u.szIntValue[0] = pszValue[1];
        u.szIntValue[1] = pszValue[0];
        u.szIntValue[2] = pszValue[3];
        u.szIntValue[3] = pszValue[2];
    }
    else if ( ((sw_full_swapped == SSwap) && (sw_part_swapped == DSwap)) ||
              ((sw_full_swapped == DSwap) && (sw_part_swapped == SSwap)) )
    {
        u.szIntValue[0] = pszValue[2];
        u.szIntValue[1] = pszValue[3];
        u.szIntValue[2] = pszValue[0];
        u.szIntValue[3] = pszValue[1];
    }

    return u.lIntValue;
}
//  ls07Swap4BHexToLong()


/*
  -----------------------------------------------------------------------------
 
  function:     ls07Swap2BHexToShort

  -----------------------------------------------------------------------------
*/
short ls07Swap2BHexToShort(tsp00_Addr           pszValue,
                           tsp00_SwapKind_Enum  SSwap,
                           tsp00_SwapKind_Enum  DSwap)
{
    ROUTINE_DBG_MEO00 ("ls07Swap2BHexToShort");

    union IntegerUnion
    {
        short lIntValue;
        char  szIntValue[2];
    } u;

    if (SSwap == DSwap)
    {
        memcpy( u.szIntValue, pszValue, sizeof(u.lIntValue) );
    }
    else
    {
        u.szIntValue[0] = pszValue[1];
        u.szIntValue[1] = pszValue[0];
    }

    return u.lIntValue;
}
//  ls07Swap2BHexToShort()



/*
  -----------------------------------------------------------------------------
 
  function:     ls07Swap2BInt

  -----------------------------------------------------------------------------
*/
tsp00_Int2
ls07Swap2BInt(tsp00_Int2 lSource,
              tsp00_SwapKind_Enum  SSwap,
              tsp00_SwapKind_Enum  DSwap)
{
    ROUTINE_DBG_MEO00 ("ls07Swap2BInt");

    union IntegerUnion
    {
        tsp00_Int2 lIntValue;
        char       szIntValue[2];
    } u;

    

    if (SSwap == DSwap)
    {
        u.lIntValue = lSource;
        //memcpy( u.szIntValue, pszValue, sizeof(u.lIntValue) );
    }
    else
    {
        u.szIntValue[0] = REINTERPRET_CAST(char*, &lSource)[1];
        u.szIntValue[1] = REINTERPRET_CAST(char*, &lSource)[0];
    }

    return u.lIntValue;
}
//  ls07Swap2BInt()


/*
  -----------------------------------------------------------------------------
  function:     ls07Swap4BInt
  -----------------------------------------------------------------------------
*/
tsp00_Int4
ls07Swap4BInt(tsp00_Int4 lSource,
              tsp00_SwapKind_Enum  SSwap,
              tsp00_SwapKind_Enum  DSwap)
{
    ROUTINE_DBG_MEO00 ("ls07Swap4BInt");

    tsp00_Int4 lTarget = 0;
    
    if (SSwap == DSwap)
    {
        lTarget = lSource;
    }
    else if ( ((sw_normal == SSwap) && (sw_full_swapped == DSwap)) ||
              ((sw_normal == DSwap) && (sw_full_swapped == SSwap)) )
    {
        REINTERPRET_CAST(char*, &lTarget)[0] = REINTERPRET_CAST(char*, &lSource)[3];
        REINTERPRET_CAST(char*, &lTarget)[1] = REINTERPRET_CAST(char*, &lSource)[2];
        REINTERPRET_CAST(char*, &lTarget)[2] = REINTERPRET_CAST(char*, &lSource)[1];
        REINTERPRET_CAST(char*, &lTarget)[3] = REINTERPRET_CAST(char*, &lSource)[0];
    }
    else if ( ((sw_normal == SSwap) && (sw_part_swapped == DSwap)) ||
              ((sw_normal == DSwap) && (sw_part_swapped == SSwap)) )
    {
        REINTERPRET_CAST(char*, &lTarget)[0] = REINTERPRET_CAST(char*, &lSource)[1];
        REINTERPRET_CAST(char*, &lTarget)[1] = REINTERPRET_CAST(char*, &lSource)[0];
        REINTERPRET_CAST(char*, &lTarget)[2] = REINTERPRET_CAST(char*, &lSource)[3];
        REINTERPRET_CAST(char*, &lTarget)[3] = REINTERPRET_CAST(char*, &lSource)[2];
    }
    else if ( ((sw_full_swapped == SSwap) && (sw_part_swapped == DSwap)) ||
              ((sw_full_swapped == DSwap) && (sw_part_swapped == SSwap)) )
    {
        REINTERPRET_CAST(char*, &lTarget)[0] = REINTERPRET_CAST(char*, &lSource)[2];
        REINTERPRET_CAST(char*, &lTarget)[1] = REINTERPRET_CAST(char*, &lSource)[3];
        REINTERPRET_CAST(char*, &lTarget)[2] = REINTERPRET_CAST(char*, &lSource)[0];
        REINTERPRET_CAST(char*, &lTarget)[3] = REINTERPRET_CAST(char*, &lSource)[1];
    }

    return lTarget;
}
//  ls07Swap4BInt()


/*
  -----------------------------------------------------------------------------
  function:     ls07Swap4BCharInPlace
  -----------------------------------------------------------------------------
*/
void
ls07Swap4BCharInPlace(tsp00_Addr          pszSource,
                      tsp00_SwapKind_Enum SSwap,
                      tsp00_SwapKind_Enum DSwap)
{
    ROUTINE_DBG_MEO00 ("ls07Swap4BCharInPlace");

    char c0 = 0;
    char c1 = 0;

    if (SSwap == DSwap)
    {
        return;
    }
    else if ( ((sw_normal == SSwap) && (sw_full_swapped == DSwap)) ||
              ((sw_normal == DSwap) && (sw_full_swapped == SSwap)) )
    {
        c0 = pszSource[0];
        c1 = pszSource[1];

        pszSource[0] = pszSource[3];
        pszSource[1] = pszSource[2];
        pszSource[2] = c1;
        pszSource[3] = c0;
    }
    else if ( ((sw_normal == SSwap) && (sw_part_swapped == DSwap)) ||
              ((sw_normal == DSwap) && (sw_part_swapped == SSwap)) )
    {
        c0 = pszSource[0];
        c1 = pszSource[2];

        pszSource[0] = pszSource[1];
        pszSource[1] = c0;
        pszSource[2] = pszSource[3];
        pszSource[3] = c1;
    }
    else if ( ((sw_full_swapped == SSwap) && (sw_part_swapped == DSwap)) ||
              ((sw_full_swapped == DSwap) && (sw_part_swapped == SSwap)) )
    {
        c0 = pszSource[0];
        c1 = pszSource[1];

        pszSource[0] = pszSource[2];
        pszSource[1] = pszSource[3];
        pszSource[2] = c0;
        pszSource[3] = c1;
    }

    return;
}
// ls07Swap4BCharInPlace


/*
  -----------------------------------------------------------------------------
 
  function:     ls07FourByteToSReal

  -----------------------------------------------------------------------------
*/
tsp00_Shortreal ls07FourByteToSReal(tsp00_Addr  pszValue)
{
    ROUTINE_DBG_MEO00 ("ls07FourByteToSReal");

    union ShortRealUnion
    {
        tsp00_Shortreal fShort;
        char            szShortChar[4];
    } u;

    memcpy( u.szShortChar, pszValue, sizeof(tsp00_Shortreal) );

    return  u.fShort;
}
//  ls07FourByteToSReal()


/*
  -----------------------------------------------------------------------------
 
  function:     ls07EightByteToLReal

  -----------------------------------------------------------------------------

  description:  Konvertiert einen 8 byte hex-wert zu double (long real).

  arguments:    pszValue   [IN]    - pointer, value to convert

  returnvalue:  converted value as long real
  -----------------------------------------------------------------------------
*/
tsp00_Longreal ls07EightByteToLReal(tsp00_Addr  pszValue)
{
    ROUTINE_DBG_MEO00 ("ls07EightByteToLReal");

    union LongRealUnion
    {
        tsp00_Longreal fLong;
        char           szLongChar[8];
    } u;

    memcpy( u.szLongChar, pszValue, sizeof (u.fLong) );

    return  u.fLong;
}
//  ls07EightByteToLReal()



/*
  -----------------------------------------------------------------------------
 
  function:     ls07NumberToInt

  -----------------------------------------------------------------------------

  description:  Konvertiert einen number wert der laenge mxsp_resnum (=7) zu
                einem 4 byte int.

  arguments:    pszNumberString  [IN]   - pointer, value to convert
                Dest             [OUT]  - reference, converted 4 byte integer

  returnvalue:  errOK_els00              - everything went fine
                errNumberTruncated_els98 -
                errNumberOverflow_els98  -
                errInvalidNumber_els98   -
  -----------------------------------------------------------------------------
*/
tsp00_Int4 ls07NumberToInt(tsp00_Addr pszNumberString, tsp00_Int4& Dest)
{
    ROUTINE_DBG_MEO00 ("ls07NumberToInt");
    
    tsp00_NumError Result;
    tsp00_Int4     rc     = errOK_els00;

    Result.becomes(num_ok);

    s40g4int(pszNumberString, 1, Dest, Result);
    switch (Result)
    {
        case num_ok:
        {
            break;
        }
        case num_trunc:
        {   // Eigentlich nur eine warnung. Individuell von jedem aufrufer zu behandeln
            rc = errNumberTruncated_els98;
            break;
        }
        case num_overflow:
        {
            rc = errNumberOverflow_els98;
            break;
        }
        default:        // entspricht num_invalid
        {
            rc = errInvalidNumber_els98;
        }
    }

    return rc;
}
//  ls07NumberToInt()



/*
  -----------------------------------------------------------------------------
 
  function:     ls07IntToNumber

  -----------------------------------------------------------------------------

  description:  Konvertiert einen 4 byte int in einen number-wert der laenge
                mxsp_resnum (max = 7).

  arguments:    Source           [IN]    - 4 byte integer to convert
                pszNumberString  [OUT]   - pointer, buffer to hold converted
                                           value (fixed(10,0))
                
  returnvalue:  errOK_els00              - everything went fine
                errNumberTruncated_els98 -
                errNumberOverflow_els98  -
                errInvalidNumber_els98   -
  -----------------------------------------------------------------------------
*/
tsp00_Int4
ls07IntToNumber(tsp00_Int4 Source, tsp00_Addr pszNumberString)
{
    ROUTINE_DBG_MEO00 ("ls07IntToNumber");
    
    tsp00_NumError Result;
    tsp00_Int4     rc     = errOK_els00;

    Result.becomes(num_ok);

    s41p4int(pszNumberString, 1, Source, Result);
    switch (Result)
    {
        case num_ok:
        {
            break;
        }
        case num_trunc:
        {   // Eigentlich nur eine warnung. Individuell von jedem aufrufer zu behandeln
            rc = errNumberTruncated_els98;
            break;
        }
        case num_overflow:
        {
            rc = errNumberOverflow_els98;
            break;
        }
        default:        // entspricht num_invalid
        {
            rc = errInvalidNumber_els98;
        }
    }

    return rc;
}
// ls07IntToNumber()



// Conversion table muss durch den characterset ersetzt werden, der gerade gewuenscht ist,!!!!!!
// das muss mit dem kern abgestimmt sein
/*
  -----------------------------------------------------------------------------
 
  function:     ls07CharsetToStandard

  -----------------------------------------------------------------------------

  description:  Konvertiert einen string von einem charset zu standard ascii.

                Die aufrufende funktion muss im falle, dass hier kein speicherplatz
                zur verfuegung gestellt werden soll (fAlloc = false),
                sicherstellen, dass der uebergebenen string gross genug ist,
                um den konvertierten string aufzunehmen.    

  arguments:    pszDestination  [OUT]   - pointer to pointer, buffer for converted
                                          string (null terminated)
                pszSource       [IN]    - pointer, source string
                lSourceSize     [IN]    - length of source string
                ConvTable       [IN]    - reference, conversion tables
                fAlloc          [IN]    - flag indicating if the function has to
                                          allocate memory for the buffer

  returnvalue:  errOK_els00  - everything went fine
                -1           - invalid length specification
  -----------------------------------------------------------------------------
*/
int ls07CharsetToStandard(char**            pszDestination,
                          const char*       pszSource,
                          tsp00_Int4        lSourceSize,
                          tsp00_Ctable &    ConvTable,
                          bool              fAlloc)
{
    ROUTINE_DBG_MEO00 ("ls07CharsetToStandard");

    char*   szNewString;

    if ( (lSourceSize <= 0) || (STATIC_CAST(tsp00_Uint4,lSourceSize) > strlen(pszSource)) )
    {
        return (-1);
    }

    if (true == fAlloc)
    {
        szNewString = new char[lSourceSize + 1];        // ein byte mehr fuer abschliessende null
        memset(szNewString, '\0', lSourceSize + 1);

        *pszDestination = szNewString;
    }
    else
    {
        szNewString = *pszDestination;
    }

    while (lSourceSize)
    {
        *szNewString++ = ConvTable[(unsigned char)*pszSource++];
        lSourceSize--;
    }

    return errOK_els00;
}
//  ls07CharsetToStandard()


/*
  -----------------------------------------------------------------------------
  function:     ls07CheckIntValueRange
  -----------------------------------------------------------------------------
*/
int
ls07CheckIntValueRange(_TCHAR *pszValue)
{
    int rc = 0;     // initialize to error
    
    size_t lLen = _tcslen(pszValue);

    if (lLen <= 10)
    {
        size_t i = 0;
        for (i; ( (i < lLen) && (0 != isdigit(pszValue[i]) ) ); ++i);
        if (lLen == i)      // no not digit found
        {
            rc = 1;
            if (10 == lLen)
            {
                if (0 > memcmp("2147483647", pszValue, lLen) )
                {
                    rc = 0;
                }
            }
        }
    }

    return rc;
}
// ls07CheckValueRange()


/*
  -----------------------------------------------------------------------------
  function:     ls07IntCompare
  -----------------------------------------------------------------------------
*/
tsp00_Int4
ls07IntCompare(const void *arg1, const void *arg2)
{
    if (*((tsp00_Int4*)arg1) < *((tsp00_Int4*)arg2))
    {
        return -1;
    }
    else if (*((tsp00_Int4*)arg1) > *((tsp00_Int4*)arg2))
    {
        return 1;
    }

    return 0;
}
// ls07IntCompare()
