/*
  -----------------------------------------------------------------------------

  module: vls15.cpp

  -----------------------------------------------------------------------------

  responsible:  SteffenS

  special area: Replication Manager

  description:  user management

  version:      7.3.

  -----------------------------------------------------------------------------

  copyright:    SAP AG, 1997-2002

  -----------------------------------------------------------------------------


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*/
#include <assert.h>

#include "gls00.h"
#include "hls98msg.h"

#include "hls03.h"
#include "hls04.h"
#include "hls05.h"
#include "hls07.h"      // NumberToInt conversion
#include "hls13.h"
#include "hls15.h"
#include "hls99.h"

externPascal tin_global_in_vars i01g;

#define _INTERNAL_FILE_ID_    _T("15")

/* Lokale Funktionen: ******************************************************************/

/*
  -----------------------------------------------------------------------------
  function:     ls15_CheckConnectPacket - local function
  -----------------------------------------------------------------------------

  description:  Gets the following information from the reply packet of the
                logon request:
                 - page size
                 - conversion tables
                 - kernel version
                 - user group of connected user

  arguments:    pDBInfo  [IN]    - structure with all info on the running session
                                    and the db connected to.
                PacketNo [IN]    - Number of packet to be filled.
                pszErrText [OUT] - error text

  returnvalue:  errOK_els00        - everything went fine
                errInternal_els98  - error getting information; logs the 'real' error
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls15_CheckConnectPacket(tls00_DBInfo *pDBInfo, tsp00_Int4 PacketNo, tsp00_Addr pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls15_CheckConnectPacket");

    tsp00_Int4       rc         = errOK_els00;
    RepServMessages *pMsgObj    = RepServMessages::Instance();
    tsp1_packet     *pSqlPacket = pDBInfo->dbiPktRcvList_ls00[PacketNo];
    tsp1_part       *pSqlPart   = NULL;

    //*
    //*     Get page size, code type and kernel version
    //*
    if ( true == ls04GetPart(pSqlPacket, sp1pk_session_info_returned, pSqlPart) )
    {
        // Code type (check for UNICODE kernel)
        _TCHAR fUnicodeFlag;
        ls04BufFromPartEx(pSqlPart,0, 1, &fUnicodeFlag, 1);
        if (1 == fUnicodeFlag)
        {
            pDBInfo->dbiEncoding_ls00 = ctUCS2_els00;
        }

        // Page size
        // Convert size from internal format to readable format. In case the number
        // is truncated that's not an error - we proceed working.
        _TCHAR szPageSize[RESNUM_MXSP00] = {0};

        ls04BufFromPartEx(pSqlPart, 2193, 5, szPageSize, mxsp_resnum);

        rc = ls07NumberToInt(szPageSize, pDBInfo->dbiPageSize_ls00);
        if (errOK_els00 != rc)
        {
            if (errNumberTruncated_els98 != rc)
            {
                pMsgObj->ls98MsgToString(layBasicConnect_els98, pszErrText, errInternal_els98,
                                                                            _INTERNAL_FILE_ID_,
                                                                            __LINE__, errConvertPageSize_els98);
                TRACE_BUFF_MLS99("ls15_CheckConnectPacket", "Page size(number format):",
                                                            RESNUM_MXSP00, szPageSize);
                rc = errInternal_els98;
            }
            else
            {
                rc = errOK_els00;
            }
        }
        if (errOK_els00 == rc)
        {
            TRACE_PRNF_MLS99(("ls15_CheckConnectPacket", "Page size converted: %d\n", pDBInfo->dbiPageSize_ls00));
        }

        // Kernel version
        if (errOK_els00 == rc)
        {
            unsigned char szVersion[5] = {0};
            ls04BufFromPartEx(pSqlPart, 2200, 5, REINTERPRET_CAST(char*, szVersion), 5);

            pDBInfo->dbiDBVersion_ls00.MajorVersion_sp100 = szVersion[0];
            pDBInfo->dbiDBVersion_ls00.MinorVersion_sp100 = szVersion[2];
            pDBInfo->dbiDBVersion_ls00.CorrLevel_sp100    = szVersion[4];
        }
    }
    else
    {
        pMsgObj->ls98MsgToString(layBasicConnect_els98, pszErrText, errInternal_els98,
                                                                    _INTERNAL_FILE_ID_,
                                                                    __LINE__, errUnknownPart_els98);
        rc = errInternal_els98;
    }

    //*
    //*     Get the conversion tables and put them into GLOBAL (!!!) structure i01g
    //*     Put the code type in i01g, too.
    //*
    if (errOK_els00 == rc)
    {
        pSqlPart = NULL;
        if ( true == ls04GetPart(pSqlPacket, sp1pk_conv_tables_returned, pSqlPart) )
        {
            tsp00_Int4 Len = ls04GetPartLength(pSqlPart);
            ls04BufFromPartEx(pSqlPart,
                              0,
                              Len,
                              REINTERPRET_CAST(char*, pDBInfo->dbiSession_ls00.siConversionTables_ls00),
                              sizeof(pDBInfo->dbiSession_ls00.siConversionTables_ls00));

            //  This is only for the conversion of date, time and timestamp values using the old
            //  vin - methods.
            for (int i = 0; i < (Len/STATIC_CAST(int, sizeof(tsp00_Ctable))); ++i)
            {
                i01g->i32.transtables[i] = pDBInfo->dbiSession_ls00.siConversionTables_ls00[i];
            }

            i01g->i20.code_type = pSqlPacket->sp1_header.sp1h_mess_code;
        }
        else
        {
            pMsgObj->ls98MsgToString(layBasicConnect_els98, pszErrText, errInternal_els98,
                                                                        _INTERNAL_FILE_ID_,
                                                                        __LINE__, errUnknownPart_els98);
            rc = errInternal_els98;
        }
    }
    
    //*
    //*     Get some user related information: UserGroup name
    //*
    if (errOK_els00 == rc)
    {
        pSqlPart = NULL;
        if ( true == ls04GetPart(pSqlPacket, sp1pk_user_info_returned, pSqlPart) )
        {
            _TCHAR szGroup[512];

            tsp00_Int4 Len = ls04GetPartLength(pSqlPart);
            ls04BufFromPartEx(pSqlPart, 0, Len, szGroup, sizeof(szGroup));

            //  Separate usergroup out of the returned value. The value returned contains the
            //  following information (each info is preceeded by a byte given the length of info):
            //  - sysdba of serverdb to which current user is connected (max. 96 bytes)
            //  - sysdba of serverdb where the current user was created (max. 96 bytes)
            //  - user type (max. 16 bytes)
            //  - user group user belongs to (max. 96 bytes)
            short count = 0;
            for (int j = 0; j < 3; ++j)     // only 3 loops because the first count is the init
            {
                count += szGroup[count] + 1;
            }
            if (0 != szGroup[count])
            {
                memcpy(pDBInfo->dbiSession_ls00.siUserGroup_ls00, &szGroup[count] + sizeof(_TCHAR), szGroup[count]);
            }
            else
            {
                memcpy(pDBInfo->dbiSession_ls00.siUserGroup_ls00.asCharp(),
                       pDBInfo->dbiSession_ls00.siUserName_ls00.asCharp(),
                       pDBInfo->dbiSession_ls00.siUserName_ls00.length() );
            }
        }
        else
        {
            pMsgObj->ls98MsgToString(layBasicConnect_els98, pszErrText, errInternal_els98,
                                                                        _INTERNAL_FILE_ID_,
                                                                        __LINE__, errUnknownPart_els98);
            rc = errInternal_els98;
        }
    }

    return rc;
}
// ls15_CheckConnectPacket()


/* Globale Funktionen: *****************************************************************/


/*
  -----------------------------------------------------------------------------
  function:     ls15CommitUser
  -----------------------------------------------------------------------------
*/
int ls15CommitUser(tls00_DBInfo* pDBInfo, tsp00_Addr pszErrText, bool bInternal)
{
    ROUTINE_DBG_MEO00 ("ls15CommitUser");

    RepServMessages*  pMsgObj = RepServMessages::Instance();    // error reporting object

    tsp00_Int4  rc = errConnectNoServerDB_els98;                    // Initialize to error

    //*
    //*     Check for db connection and user connection to db
    //*     Notify caller if no connection to db established. Otherwise release connection.
    //*
    if (true == pDBInfo->dbiSession_ls00.siSqlStat_ls00)
    {
        if (false == bInternal)     // If internally the connection is changed from OrderInterface
        {                           // to ODBC don't notify user
            if (true == pDBInfo->dbiSession_ls00.siAutoCommit_ls00)
            {
                pMsgObj->ls98PrintMessage(layBasicConnect_els98, msgReleaseUserComm_els98,
                                                                 pDBInfo->dbiSession_ls00.siUserName_ls00.asCharp());
            }
            else
            {
                pMsgObj->ls98PrintMessage(layBasicConnect_els98, msgReleaseUserRollback_els98,
                                                                 pDBInfo->dbiSession_ls00.siUserName_ls00.asCharp());
            }
        }
        else
        {
            TRACE_PRNF_MLS99(("ls15CommitUser", "OI connection internally closed.\n"));
        }

        rc = ls15Rollback(pDBInfo, pszErrText, true);       // rollback work release
        if (errOK_els00 != rc)
        {
            TRACE_PRNF_MLS99(("ls15CommitUser", "Rollback failed with rc = %d.\n", rc));
        }

        // Even if rollback failes we'll return success
        rc = errOK_els00;
    }

    if (true == pDBInfo->dbiSession_ls00.siDBConnection_ls00)
    {
        ls03SQLRelease (pDBInfo);    // release db connection
        rc = errOK_els00;
    }

    return rc;
}
// ls15CommitUser()


/*
  -----------------------------------------------------------------------------
  function:     ls15SQLGetUser
  -----------------------------------------------------------------------------
*/
int ls15SQLGetUser(tls00_SessionInfo* Session, tsp00_Addr pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls15SQLGetUser");

    tsp00_Int4         rc         = errSqlConnectUserkey_els98; // Initialize to error
    tsp00_Bool         ErrFlag;
    tsp00_ErrText      pszMyText;
    tsp00_Addr         pszUK      = Session->siUserKey_ls00;
    tsp4_xuser_record *UsrRecord  = new tsp4_xuser_record;

    size_t             lUKLength  = strlen(pszUK);

    // Initialize UsrRecord
    if (lUKLength >= sizeof (tsp4_xuserkey))
    {
        memcpy(UsrRecord->xu_key, pszUK, sizeof (tsp4_xuserkey));
    }
    else
    {
        memset(UsrRecord->xu_key, BLANK_LS00, sizeof (tsp4_xuserkey));
        memcpy(UsrRecord->xu_key, pszUK, lUKLength);
    }

    // Get user information and store it in session structure.
    sqlgetuser(UsrRecord, pszMyText, &ErrFlag);
    if (0 != ErrFlag)
    {
        Session->siDBName_ls00.Init();
        Session->siDBName_ls00.p2c(UsrRecord->xu_serverdb);

        Session->siDBNode_ls00.rawAssign(UsrRecord->xu_servernode.asCharp());

        Session->siUserName_ls00.Init();
        Session->siUserName_ls00.rawAssign(UsrRecord->xu_user_61);

        memcpy(Session->siUserCPwd_ls00, UsrRecord->xu_password, sizeof (tsp00_CryptPw));

        rc = errOK_els00;
    }
    else
    {   // make error text null terminated string
        strncpy(pszErrText, pszMyText, sizeof(tsp00_ErrText));
        pszErrText[sizeof(tsp00_ErrText)] = '\0';
        ls05StrTrimRightEx(pszErrText);
    }

    delete UsrRecord;

    return rc;
}
//  ls15SQLGetUser()


/*
  -----------------------------------------------------------------------------
  function:     ls15RSCommit
  -----------------------------------------------------------------------------
*/
tsp00_Int4 ls15RSCommit(tls00_DBInfo *DBInfo, tsp00_Addr ErrText)
{
    ROUTINE_DBG_MEO00 ("ls15RSCommit");

    tsp00_Int4  rc = errOK_els00;

    if (true == DBInfo->dbiSession_ls00.siSqlStat_ls00)
    {
        if (true == DBInfo->dbiSession_ls00.siAutoCommit_ls00)
        {
            rc = ls15Commit(DBInfo, ErrText);
        }
    }
    // No db connection no commit

    return rc;
}
//  ls15RSCommit()



/*
  -----------------------------------------------------------------------------
  function:     ls15Commit
  -----------------------------------------------------------------------------
*/
tsp00_Int4 ls15Commit(tls00_DBInfo *DBInfo, tsp00_Addr ErrText)
{
    ROUTINE_DBG_MEO00 ("ls15Commit");

    tsp00_Int4  rc        = errOK_els00;
    tsp00_Int4  ErrPos    = 0;
    char        szCmd[64] = "COMMIT WORK";

    rc = ls13ExecIntSQLStmt(DBInfo, 0, szCmd, ErrText, ErrPos);
    if (errOK_els00 != rc)
    {
        // Here we get already an error text that is only to be printed. We'll use
        // the errSimpleSQL_els98-message to do it.
        RepServMessages::Instance()->ls98PrintMessage(layBasicConnect_els98,
                                                      errSimpleSQL_els98,
                                                      ErrText);
    }

    return rc;
}
//  ls15Commit()


/*
  -----------------------------------------------------------------------------
  function:     ls15Rollback
  -----------------------------------------------------------------------------
*/
tsp00_Int4 ls15Rollback(tls00_DBInfo *DBInfo, tsp00_Addr ErrText, bool bRelease)
{
    ROUTINE_DBG_MEO00 ("ls15Rollback");

    tsp00_Int4  rc      = errOK_els00;
    tsp00_Int4  ErrPos  = 0;
    char        szCmd[64] = "ROLLBACK WORK";

    if (true == bRelease)
    {
        szCmd[0] = '\0';
        strcpy(szCmd, "ROLLBACK WORK RELEASE");
    }

    rc = ls13ExecIntSQLStmt(DBInfo, 0, szCmd, ErrText, ErrPos);
    if (errOK_els00 != rc)
    {
        // Here we get already an error text that is only to be printed. We'll use
        // the errSimpleSQL_els98-message to do it.
        RepServMessages::Instance()->ls98PrintMessage(layBasicConnect_els98,
                                                      errSimpleSQL_els98,
                                                      ErrText);
    }

    return rc;
}
//  ls15Rollback()


/*
  -----------------------------------------------------------------------------
  function:     ls15SwitchUser
  -----------------------------------------------------------------------------
*/
tsp00_Int4
ls15SwitchUser(tls00_DBInfo            *pDBInfo,
               SAPDB_DynamicUTF8String  UserName,
               tsp00_Addr               pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls15SwitchUser");

    tsp00_Int4 rc = errOK_els00;

    _TCHAR szCmd[256];
    szCmd[0] = 0;
    sp77sprintf(szCmd, 256, "CURRENT USER \"%s\"", UserName.StrPtr());

     tsp00_Int4 lErrPos;
    _TCHAR      szErr[MAX_REPLY_LENGTH_LS00];
    rc = ls13ExecIntSQLStmtUTF8(pDBInfo, 0, szCmd, szErr, lErrPos);
    if (errOK_els00 != rc)
    {
        RepServMessages::Instance()->ls98Msg(layTableExtract_els98, pszErrText,
                                                                    errCurrentUserCmd_els98,
                                                                    UserName.StrPtr(), szErr);
        rc = errCurrentUserCmd_els98;
    }

    return rc;
}
// ls15SwitchUser()


/*
  -----------------------------------------------------------------------------
  function: ls15CheckUserMode
  -----------------------------------------------------------------------------
*/
tsp00_Int4
ls15CheckUserMode(tls00_DBInfo            *pDBInfo,
                  SAPDB_DynamicUTF8String  UserName,
                  tls00_DBUserMode         UserMode,
                  tsp00_Addr               pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls15CheckUserMode");

    tsp00_Int4 rc = errOK_els00;

    assert(dbuUnknown_els00 != UserMode);

    _TCHAR szCmd[256];
    _TCHAR szUserMode[32];
    szUserMode[0] = 0;
    szCmd[0]      = 0;

    switch (UserMode)
    {
        case dbuSysdba_els00:
        {
            _tcscpy(szUserMode, _T("SYSDBA"));
            break;
        }
        case dbuDba_els00:
        {
            _tcscpy(szUserMode, _T("DBA"));
            break;
        }
        case dbuResource_els00:
        {
            _tcscpy(szUserMode, _T("RESOURCE"));
            break;
        }
        default:    // dbuColdUser_els00
        {
            _tcscpy(szUserMode, _T("COLDUSER"));
        }
    }
    sp77sprintf(szCmd, 256, "SELECT * FROM DOMAIN.USERS WHERE USERNAME = '%s' AND USERMODE = '%s'",
                            UserName.StrPtr(), szUserMode);

    tsp00_Int4 lErrPos = 0;
    _TCHAR     szErr[MAX_REPLY_LENGTH_LS00];

    rc = ls13ExecIntSQLStmtUTF8(pDBInfo, 0, szCmd, szErr, lErrPos);
    if (errOK_els00 != rc)
    {
        if (dberrRowNotFound_els98 != rc)
        {
            // some other error executing command
            RepServMessages::Instance()->ls98Msg(layTableExtract_els98, pszErrText,
                                                                        errInternSQL_els98,
                                                                        szCmd, szErr);
            rc = errInternSQL_els98;
        }
    }

    return rc;
}
// ls15CheckUserMode()


/*
  -----------------------------------------------------------------------------
  function:     ls15ConnectUser
  -----------------------------------------------------------------------------
*/
tsp00_Int4
ls15ConnectUser(tls00_DBInfo* DBInfo, tsp00_Addr ErrText)
{
    ROUTINE_DBG_MEO00 ("ls15ConnectUser");

    tsp00_Int4 rc = errOK_els00;

    RepServMessages   *pMsgObj  = RepServMessages::Instance();    // error reporting object
    tls00_SessionInfo *pSession = &DBInfo->dbiSession_ls00;



    if (0 == pSession->siUserName_ls00.length())
    {
        return rc;
    }


    TRACE_PRNF_MLS99(("ls15ConnectUser", "Connecting user: %s at %s.\n",
                                         pSession->siUserName_ls00.asCharp(),
                                         pSession->siDBName_ls00.asCharp()));
    //*
    //*     Build connect command
    //*
    _TCHAR szCmd[1024];
    _stprintf(szCmd, CONNECT_STMT_LS15, pSession->siUserName_ls00.asCharp());

    if (true == pSession->siNolog_ls00)       // NOLOG - session ?
    {
        strcat(szCmd, " NOLOG");
    }

    if ( (false == pSession->siIgnoreTCS_ls00) && (' ' != pSession->siTCharSet_ls00[0]) )
    {
        strcat(szCmd, " TERMCHAR SET ");
        strcat(szCmd, pSession->siTCharSet_ls00.asCharp());
    }

    //*
    //*     Build data string depending on encoding of db
    //*
    //* Data string looks like:
    //* ------------------------------------------------------------------------------
    //* def-byte |      crypted pw     | def-byte |            term-id               |
    //*-------------------------------------------------------------------------------
    //*
    tsp00_Int4    CPwdLen  = sizeof(tsp00_CryptPw);            // length of crypted pw stays the same for ascii and unicode
    tsp00_Longint lDataLen = CPwdLen + 2*TERMID_MXSP00 + 2;    // term id for unicode twice as long as for ascii
    tsp00_Addr    pszData  = new _TCHAR[lDataLen];

    if ( false == ls00IsDBUnicode(DBInfo) )
    {
        // def-byte in front of pw and termid
        pszData[0] = pszData[sizeof(DEF_BYTE_ASCII_LS00) + CPwdLen] = DEF_BYTE_ASCII_LS00;
        lDataLen = sizeof(DEF_BYTE_ASCII_LS00);

        // put pw in
        memcpy(&pszData[lDataLen], pSession->siUserCPwd_ls00.asCharp(), CPwdLen);
        lDataLen += sizeof(DEF_BYTE_ASCII_LS00) + CPwdLen;
        
        // put termid in
        memcpy(&pszData[lDataLen], DBInfo->dbiSqlTerm_ls00, TERMID_MXSP00);
        lDataLen += TERMID_MXSP00;
    }
    else
    {
        // def-byte in front of pw and termid
        pszData[0] = pszData[sizeof(DEF_BYTE_UNICODE_LS00) + CPwdLen] = DEF_BYTE_UNICODE_LS00;
        lDataLen   = sizeof(DEF_BYTE_UNICODE_LS00);

        // put pw in
        // in case of unicode pw needs to be stored as pascal string (unused chars filled with blank);
        // pascal string must be stored as UTF8, then converted to UCS2 and the crypted (uuuh)
        tsp00_Pw pszPwdP;
        pszPwdP.SetBlank();
        pszPwdP.c2p(pSession->siUserPwd_ls00.asCharp());
        
        tsp00_Pw       szUCS2Pwd;
        tsp00_Longuint lPWLen    = PW_MXSP00;

        rc = ls05ConvertToUCS2Simple(REINTERPRET_CAST(unsigned char*, szUCS2Pwd.asCharp()),
                                     lPWLen,
                                     REINTERPRET_CAST(unsigned char*, pszPwdP.asCharp()),
                                     PW_MXSP00/2,
                                     ctAscii_els00,
                                     ErrText);
        if (errOK_els00 == rc)
        {
            ls15ApplEncrypt(szUCS2Pwd, pSession->siUserCPwd_ls00);

            memcpy(&pszData[lDataLen], pSession->siUserCPwd_ls00.asCharp(), CPwdLen);
            lDataLen += sizeof(DEF_BYTE_UNICODE_LS00) + CPwdLen;

            // put termid in
            tsp00_Longuint lTermIDLen = 2*TERMID_MXSP00;

            rc = ls05ConvertToUCS2Simple(REINTERPRET_CAST(unsigned char*, &pszData[lDataLen]),
                                         lTermIDLen,
                                         REINTERPRET_CAST(unsigned char*, DBInfo->dbiSqlTerm_ls00.asCharp()),
                                         TERMID_MXSP00,
                                         ctAscii_els00,
                                         ErrText);
            if (errOK_els00 == rc)      // function returns length on success in lTermIDLen
            {
                lDataLen += lTermIDLen;
            }
        }
        // convert function ls05ConvertToUCS2Simple returns length on success; in case of failure
        // the function logs the message and returns an error code

    }   // end else branch of if ( false == ls00IsDBUnicode(DBInfo) )

    if (errOK_els00 == rc)
    {
        tsp00_Int4 lErrorPos = 0;

        rc = ls13ExecIntSQLStmtEx(DBInfo, 0, szCmd, pszData, STATIC_CAST(tsp00_Int4, lDataLen), ErrText, lErrorPos);
        delete [] pszData;
        pszData = NULL;
        if (errOK_els00 == rc)  // in case of error the message is already logged to return string
        {
            // function ls15_CheckConnectPacket may only return errInternal_els98; the text is
            // logged to return string and log file
            rc = ls15_CheckConnectPacket(DBInfo, 0, ErrText);

            //*
            //*     Flag a successful connection
            //*
            if (errOK_els00 == rc)
            {
                DBInfo->dbiSession_ls00.siSqlStat_ls00 = true;
            }
        }
    }

    if (NULL != pszData)
    {
        delete [] pszData;
        pszData = NULL;
    }

    TRACE_USER_MLS99("ls15ConnectUser", DBInfo);        // Trace im bedarfsfall alle userinformationen

    return rc;
}
// ls15ConnectUser()


/*
  -----------------------------------------------------------------------------
  function:     ls15Connect
  -----------------------------------------------------------------------------
*/
tsp00_Int4
ls15Connect(tls00_DBInfo *pDBInfo, tsp00_Addr pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls15Connect");

    tsp00_Int4  rc = errOK_els00;

    rc = ls03SQLConnect(pDBInfo, pszErrText);           // db connection
    if (errOK_els00 == rc)
    {
        _TCHAR aErrText[MAX_REPLY_LENGTH_LS00];         // ls15ConnectUser delivers error only in return text;
                                                        // so aErrText is necessary to build the right text then

        rc = ls15ConnectUser(pDBInfo, aErrText);        // user connection
        if (errOK_els00 != rc)
        {
            // If not yet happened try connecting using UNICODE;
            // to do so we need some manipulation
            if ( false == ls00IsDBUnicode(pDBInfo) )
            {
                rc = errSQLConnectUser_els98;

                // Releasing db connection is possible only after check because the
                // connection is checked via assertion in ls00IsDBUnicode
                ls03SQLRelease(pDBInfo);                 // Error: release db connnection

                pDBInfo->dbiEncoding_ls00 = ctUCS2_els00;

                tsp00_Int4 rc2 = errOK_els00;
                rc2 = ls15Connect(pDBInfo, pszErrText);
                if (errOK_els00 != rc2)
                {
                    pDBInfo->dbiEncoding_ls00 = ctAscii_els00;

                    if (dbNotLoadedCharset_els00 != rc2)
                    {
                        rc = rc2;
                    }
                    else
                    {
                        RepServMessages::Instance()->ls98Msg(layBasicConnect_els98,
                                                             pszErrText, errSQLConnectUser_els98,
                                                                         pDBInfo->dbiSession_ls00.siUserName_ls00.asCharp(),
                                                                         pDBInfo->dbiSession_ls00.siDBName_ls00.asCharp(),
                                                                         aErrText);
                    }
                }
            }
            else
            {
                if (dbNotLoadedCharset_els00 != rc)
                {
                    RepServMessages::Instance()->ls98Msg(layBasicConnect_els98,
                                                         pszErrText, errSQLConnectUser_els98,
                                                                     pDBInfo->dbiSession_ls00.siUserName_ls00.asCharp(),
                                                                     pDBInfo->dbiSession_ls00.siDBName_ls00.asCharp(),
                                                                     aErrText);
                    rc = errSQLConnectUser_els98;
                }

                ls03SQLRelease(pDBInfo);                 // Error: release db connnection
            }
        }   // end if (errOK_els00 != rc)
    }

    return rc;
}
// ls15Connect()
