/* 
  -----------------------------------------------------------------------------
 
  module: vls20.cpp
 
  -----------------------------------------------------------------------------
 
  responsible:  SteffenS and Hans-GeorgB

  special area: Replication Manager
 
  description:  Implementation of the DATALOAD functionality
                
  version:      7.3.

  -----------------------------------------------------------------------------
 
  copyright:    SAP AG, 1997-2003
 
  -----------------------------------------------------------------------------


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*/

#include "gls00.h"
#include "hls01.h"
#include "hls03.h"
#include "hls04.h"
#include "hls041.h"
#include "hls05.h"
#include "hls07.h"
#include "hls08_long.h"         // long value handling
#include "hls13.h"
#include "hls15.h"
#include "hls16_condition.h"
#include "hls18.h"
#include "hls20.h"
#include "hls24.h"
#include "hls25.h"
#include "hls30.h"
#include "hls99.h"
#include "hls98msg.h"


/*!  
  -----------------------------------------------------------------------------
  Chapter: Local functions
  -----------------------------------------------------------------------------
 */


/*!
  -----------------------------------------------------------------------------
  function:     ls20_CreateDloadRec
  -----------------------------------------------------------------------------

  description:  Creates and initializes data load record structure

  arguments:    pDloadRec [IN] - structure with information on processing the
                                  dataload command
                DBInfo    [IN] - structure with information on the running
                                  session and the database connected to.                
  returnvalue:  No

  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls20_CreateDloadRec(tls00_DloadRec *& pDloadRec, tls00_DBInfo *DBInfo);


/*!
  -----------------------------------------------------------------------------
  function:     ls20_InitStructures
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls20_InitStructures(tls00_DBInfo     *pDBInfo,
                    tls00_DatLoadCmd *pDloadCmd,
                    tls00_DloadRec   *pDloadRec,
                    tsp00_Addr        pszErrText);


/*!
  -----------------------------------------------------------------------------
  function:     ls20_InitAndOpenFiles
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls20_InitAndOpenFiles(tls00_DatLoadCmd *pDloadCmd,
                      tls00_DloadRec   *pDLRec,
                      tsp00_Addr        pszErrText);


/*
  -----------------------------------------------------------------------------
  function:     ls20_Destroy
  -----------------------------------------------------------------------------

  description:  Pseudo destructor

  arguments:    DloadRec  [IN] - structure with information on processing the
                                  dataload command

                pDloadCmd [IN] - structure with user information for
                                  processing dataload command.
  returnvalue:  No

  -----------------------------------------------------------------------------
*/
static void ls20_Destroy(tls00_DloadRec *& DloadRec, tls00_DatLoadCmd *pDLoadCmd);


/*!
  -----------------------------------------------------------------------------
  function:     ls20_GetRecord
  -----------------------------------------------------------------------------

  description:  Build a database record
  
                This function builds one database record and writes it to a buffer.

  arguments:    DloadCmd    [IN]    - structure with user information for
                                       processing dataload command.
                DloadRec    [IN]    - structure with information on
                                       processing the dataload command
                Errtext     [OUT]   - error text

  returnvalue:  errOK_els00         - in case of NO errors
                some other errors   - in case of errors
  -----------------------------------------------------------------------------
*/
static int
ls20_GetRecord(tls00_DatLoadCmd  *DloadCmd,
               tls00_DloadRec    *DloadRec,
               tsp1_sqlmode_Enum  SQLMode,
               bool               bUnicodeOI,
               tsp00_Addr         ErrText);


/*!
  -----------------------------------------------------------------------------
  function:     ls20_Insert
  -----------------------------------------------------------------------------

  description:  Build as many database records as fit into one packet

                This function builds one database record and writes it to a buffer.

  arguments:    DBInfo      [IN]    - structure with info on running session and
                                       database connected to
                DloadCmd    [IN]    - structure with user information for
                                       processing dataload command
                DloadRec    [IN]    - structure with information on
                                       processing the dataload command
                ErrText     [OUT]   - error text
                RowCount    [OUT]   - Count of records stored in pszRecord
                bPutVal     [OUT]   - flags the need of using putval - oi-packets
                                       to process remaining long values
                
  returnvalue:  errOK_els00         - in case of NO errors
                error reading file
                error in processing data uebersteigt maxerrorcount
                  some other errors - in case of errors
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls20_Insert(tls00_DBInfo     *DBInfo,
            tls00_DatLoadCmd *DloadCmd,
            tls00_DloadRec   *DloadRec,
            bool              bRecoverLong,
            tsp00_Addr        ErrText,
            tsp00_Int4       &RowCount,
            bool             &bPutVal);


/*!
  -----------------------------------------------------------------------------
  function:     ls20_Send
  -----------------------------------------------------------------------------

  description:  Loop for building mass insert requests, sending them to the kernel
                checking the kernel answer, sending putvals, ...


  arguments:    DBInfo      [IN]    - structure with info on running session and
                                       database connected to
                DloadCmd    [IN]    - structure with user information for
                                       processing dataload command
                DloadRec    [IN]    - structure with information on
                                       processing the dataload command
                ErrText     [OUT]   - error text

  returnvalue:  errOK_els00         - in case of NO errors
                some other errors   - in case of errors
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls20_Send(tls00_DBInfo     *DBInfo,
          tls00_DatLoadCmd *DloadCmd,
          tls00_DloadRec   *DloadRec,
          tls00_String     &pOI_InitBlock,
          tsp00_Addr        ErrText);


/*!
  -----------------------------------------------------------------------------
  function:     ls20_ParseInsertStmt
  -----------------------------------------------------------------------------

  description:  Prepare INSERT Statement as mass command.
                Extraxt table information out of the answer packet

  arguments:    DBInfo      [IN]    - structure with info on running session and
                                       database connected to
                DloadCmd    [IN]    - structure with user information for
                                       processing dataload command
                DloadRec    [IN]    - structure with information on
                                       processing the dataload command
                ErrText     [OUT]   - error text

  returnvalue:  errOK_els00         - in case of NO errors
                some other errors   - in case of errors
  -----------------------------------------------------------------------------
*/
static int
ls20_ParseInsertStmt(tls00_DBInfo*     pDBInfo,
                     tls00_DatLoadCmd* pDloadCmd,
                     tls00_DloadRec*   pDloadRec,
                     tsp00_Addr        pszErrText);


/*!
  -----------------------------------------------------------------------------
  function:     ls20_ValueToOIBuffer
  -----------------------------------------------------------------------------

  description:  Writes single data to data part buffer in the right format.
                Data has be given in the right database format.

  arguments:    pInputData          [IN]     - Data to write to buffer
                ParameterInfo       [IN]     - Info describing data
                pColumn             [IN]     - Structure describing column
                PredefinedDefaults  [IN]     - Predefined defaults like USER
                DataFileEncoding    [IN]     - Data file encoding
                bUnicodeOI          [IN]     - Database UNICODE or ASCII
                pOIValueBuffer      [IN/OUT] - Buffer to put the data in
                pszErrText          [OUT]    - Error text
  
  returnvalue:  errOK_els00                 - No error
                errIncompatibleConst_els98  - Error inserting assigned value
                errUnknownAssignment_els98  - Error inserting assigned value
                errInternal_els98           - Error converting between encodings
                errConvertingEncoding_els98 - Error converting between encodings
  -----------------------------------------------------------------------------
*/
tsp00_Int4
ls20_ValueToOIBuffer(tls00_String        &pInputData,
                     tls00_ParamInfo     &ParameterInfo,
                     tls00_Column        *pColumn,
//                    tls00_PredefDefault &PredefinedDefaults,
                     tls00_CodeType       DataFileEncoding,
                     bool                 bUnicodeOI,
                     unsigned char       *pOIValueBuffer,
                     tsp00_Addr           pszErrText);

/*!  
  -----------------------------------------------------------------------------
  EndChapter: Local functions
  -----------------------------------------------------------------------------
 */



/*  
  -----------------------------------------------------------------------------
  Global functions: implementation
  -----------------------------------------------------------------------------
*/

/*
  -----------------------------------------------------------------------------
  function:     ls20ExecDataload
  -----------------------------------------------------------------------------
*/
int
ls20ExecDataload(tls00_DatLoadCmd *DloadCmd, tls00_DBInfo *DBInfo, tsp00_Addr ErrText)
{
    ROUTINE_DBG_MEO00 ("ls20ExecDataload");

    tsp00_Int4 rc = errOK_els00;

    RepServMessages *pMsgObj = RepServMessages::Instance(); // error reporting obj


    //*
    //*     Before doing anything else we check external encoding and
    //*     database encoding.
    //*     Not supported is loading UNICODE data into ASCII databases.
    //*
    rc = ls05AreEncodingsCompatible(DloadCmd->dloInfile_ls00.fsFormat_ls00.ffoCodeType_ls00,
                                    ls00IsDBUnicode(DBInfo) ? ctUCS2_els00 : ctAscii_els00);
    if (errOK_els00 != rc)
    {
        pMsgObj->ls98Msg(layDataLoad_els98, ErrText, errIncompatibleEncodings_els98);
    }

    //*
    //*     create dataload record structure
    //*
    tls00_DloadRec *DloadRec = NULL;
    if (errOK_els00 == rc)
    {
        rc = ls20_CreateDloadRec(DloadRec, DBInfo);
    }

    //*
    //*     Start load with prepare of insert command and getting parameter info
    //*
    if (rc == errOK_els00)
    {
        rc = ls20_ParseInsertStmt(DBInfo, DloadCmd, DloadRec, ErrText);
    }

    if (errOK_els00 == rc)
    {
        rc = ls20_InitAndOpenFiles(DloadCmd, DloadRec, ErrText);
    }

    if (errOK_els00 == rc)
    {
        rc = ls20_InitStructures(DBInfo, DloadCmd, DloadRec, ErrText);
    }


    //*
    //*     Initialize packet used sending data to the kernel
    //*
    tls00_Buffer *pDataBuffer = &DloadRec->dlrDataPartHandling_ls00.dphBuffer_ls00;
    tls00_String pOIPacketInitBlock;
    pOIPacketInitBlock.strLeng_ls00 = 0;
    pOIPacketInitBlock.strAddr_ls00 = NULL;
    
    if (errOK_els00 == rc)
    {
        rc = ls20InitializeOIPacket(DBInfo,
                                    DloadRec->dlrParseId_ls00,
                                    DloadRec->dlrResultCount_ls00,
                                    DloadRec->dlrPartAttr_ls00,
                                    pOIPacketInitBlock,
                                    pDataBuffer);
        if (errOK_els00 != rc)
        {
            pMsgObj->ls98Msg(layDataLoad_els98, ErrText, STATIC_CAST(ErrorNumbers, rc));
        }
    }

    //*
    //*     Run dataload
    //*
    if (errOK_els00 == rc)
    {
        rc = ls20_Send(DBInfo, DloadCmd, DloadRec, pOIPacketInitBlock, ErrText);
        if (errOK_els00 == rc)
        {
            if (0 == DloadRec->dlrRecordCount_ls00)
            {
                // Nothing sent so far: we treat it as an error
                 pMsgObj->ls98Msg(layDataLoad_els98, ErrText, errEmptyDataFile_els98,
                                                              DloadRec->dlrInfile_ls00->vfFileName_ls00);
                rc = errEmptyDataFile_els98;
            }
            else
            {
//                rc = ls15RSCommit(DBInfo, ErrText);
            }
        }
    }

    // some cleaning work
    if (NULL != pOIPacketInitBlock.strAddr_ls00)
    {
        delete [] pOIPacketInitBlock.strAddr_ls00;
        pOIPacketInitBlock.strAddr_ls00 = NULL;
    }


    if (NULL != DloadRec)
    {
        // Adapt dlrLastCommLine_ls00 for the case of a defined start position
        if (UNDEFINED_LS00 != DloadCmd->dloInfile_ls00.fsExtract_ls00.feStart_ls00)
        {
            DloadRec->dlrLastCommLine_ls00 += DloadCmd->dloInfile_ls00.fsExtract_ls00.feStart_ls00;
        }


        if (true == DBInfo->dbiSession_ls00.siAutoCommit_ls00)
        {
            pMsgObj->ls98PrintMessage(layDataLoad_els98, msgLastCommInputLine_els98,
                                                         DloadRec->dlrLastCommLine_ls00);
        }
        else
        {
            pMsgObj->ls98PrintMessage(layDataLoad_els98, msgLastInsInputLine_els98,
                                                         DloadRec->dlrLastCommLine_ls00);
        }

        pMsgObj->ls98PrintMessage(layDataLoad_els98, msgInsertedRejectedRows_els98,
                                                     DloadRec->dlrInsertedRows_ls00,
                                                     DloadRec->dlrLRejected_ls00);
    }

    ls20_Destroy(DloadRec, DloadCmd);

    return rc;
}
//  ls20ExecDataload()


/*
  -----------------------------------------------------------------------------
  function:     ls20InitDataSourceStruct
  -----------------------------------------------------------------------------
*/
void
ls20InitDataSourceStruct(tls00_DataSource  *pDataSource,
                         tsp00_Int4         lFieldCount,        // tabledescription->fieldcount_ls00
                         tls00_MultColSpec *pMultColSpec,
                         tsp00_Int4         lSetColumnIndex)    // applies only to DATAUPDATE command; defaults to 0
{
    ROUTINE_DBG_MEO00 ("ls20InitDataSourceStruct");
    
    tls00_Column *pCol = NULL;

    pDataSource->dasInputBuffer.strAddr_ls00 = new char[LINELENGTH_MXSP00];
    pDataSource->dasInputBuffer.strLeng_ls00 = 0;

    ls18RecordConstructor(lFieldCount, pDataSource->dasConvertedData);

    pDataSource->lReadFieldsCnt_ls00 = lFieldCount;

    // We need only as much read positions as we have data fields to read
    pDataSource->lReadPos_ls00           = new tsp00_Uint4[pDataSource->lReadFieldsCnt_ls00];
    pDataSource->plDataToColIndex_ls00   = new tsp00_Uint4[lFieldCount];
    pDataSource->plDataToParamIndex_ls00 = new tsp00_Uint4[lFieldCount];
        
    tsp00_Int4 lSetColCount = lFieldCount - lSetColumnIndex;

    for (int j = 0, k=0; j < pMultColSpec->mcsCount_ls00; ++j)
    {
        pCol = pMultColSpec->mcsColumn_ls00[j];
        if (asUnknown_els00 == pCol->colAssign_ls00.fasAsgnTyp_ls00)
        {
            pDataSource->plDataToColIndex_ls00[k] = j;

            pDataSource->lReadPos_ls00[k] = (0 == pCol->colFPos_ls00.fpoStart_ls00) ?
                                            j + 1 :
                                            pCol->colFPos_ls00.fpoStart_ls00;
            if (0 != lSetColumnIndex)
            {
                if (k < lSetColumnIndex)
                {
                    pDataSource->plDataToParamIndex_ls00[k] = k + lSetColCount;
                }
                else
                {
                    pDataSource->plDataToParamIndex_ls00[k] = k - lSetColumnIndex;
                }
            }
            else
            {
                pDataSource->plDataToParamIndex_ls00[k] = k;
            }
            ++k;
        }   // end if (asUnknown_els00 == pCol->colAssign_ls00.fasAsgnTyp_ls00)
    }   // end for (int j = 0, k=0; j < pMultColSpec->mcsCount_ls00; ++j)

    return;
}
// ls20InitDataSourceStruct()


/*
  -----------------------------------------------------------------------------
  function:     ls20InitializeOIPacket
  -----------------------------------------------------------------------------
*/
tsp00_Int4
ls20InitializeOIPacket(tls00_DBInfo         *pDBInfo,
                       char                 *pszParseID,
                       unsigned char        *pszResultCount,
                       tsp1_part_attributes  PartAttribute,
                       tls00_String         &pOIPacketInitBlock,
                       tls00_Buffer         *pDataPartBuffer)
{
    ROUTINE_DBG_MEO00 ("ls20InitializeOIPacket");

    tsp00_Int4 rc = errOK_els00;

    tsp1_packet  *pSqlPacket = pDBInfo->dbiPktSndList_ls00[0];
    tsp1_segment *pSqlSegm   = NULL;
    tsp1_part    *pSqlPart   = NULL;

    //*
    //*     Generate Execute Command
    //*
    //* 1.part: parse id
    //* 2.part: daten
    //* 3.part: result count
    //*
    rc = ls04FastInit(pSqlPacket, &pSqlSegm, &pSqlPart, pDBInfo);
    ls04SetMsgType(pSqlSegm, sp1m_execute);                      // Message Type EXECUTE'
    ls04SetMCFlag(pSqlSegm, true);                               // Mass insert 
    pSqlPart->sp1p_part_kind().becomes(sp1pk_parsid);


    // Insert parse id part
    rc = ls04BufToPart(pSqlPart, pszParseID, MAX_PARSEID_LENGTH_LS00);    // do not write trailing NULL
    if (errOK_els00 == rc)
    {
        rc = ls04FinishPart(pSqlPacket, pSqlPart);
    }

    // insert resultcount part
    if (errOK_els00 == rc)
    {
        rc = ls04InitPart(pSqlPacket, pSqlSegm, &pSqlPart, sp1pk_resultcount);
        if (errOK_els00 == rc)
        {
            rc = ls04BufToPart(pSqlPart,REINTERPRET_CAST(tsp00_Addr, pszResultCount), sizeof(csp_rescnt_unknown));
            if (errOK_els00 == rc)
            {
                rc = ls04FinishPart(pSqlPacket, pSqlPart);
            }
        }
    }
    
    // insert data part
    if (errOK_els00 == rc)
    {
        pSqlPart = NULL;
        rc = ls04InitPart(pSqlPacket, pSqlSegm, &pSqlPart, sp1pk_data);
        if (errOK_els00 != rc)
        {
            // Set part attribute for the first time (always: partattr = firstpacket)
            ls04SetPartAttribute(pSqlPart, PartAttribute);
        }
    }
    
    // static cast:IA64
    pOIPacketInitBlock.strLeng_ls00 = PartHSize_ls00 + 
                                        STATIC_CAST(tsp00_Int4, REINTERPRET_CAST(char*, pSqlPart) - REINTERPRET_CAST(char*, pSqlPacket));

    //*
    //*     Allocate space for initial block in order interface packet that may be reused
    //*     for every single data packet sent to the kernel while inserting data for one single table
    //*
    pOIPacketInitBlock.strAddr_ls00 = new char[pOIPacketInitBlock.strLeng_ls00];
    memcpy(pOIPacketInitBlock.strAddr_ls00, (char*) &pSqlPacket->sp1_header, pOIPacketInitBlock.strLeng_ls00);

    // Initialize working buffer with the size of the data part
    pDataPartBuffer->lBufSize_ls00 = pSqlPart->sp1p_buf_size();
    pDataPartBuffer->lBufLen_ls00  = 0;
    pDataPartBuffer->pBufPtr_ls00  = new unsigned char[pDataPartBuffer->lBufSize_ls00];

    return rc;
}
// ls20InitializeOIPacket()


/*
  -----------------------------------------------------------------------------
  function:     ls20ReadData
  -----------------------------------------------------------------------------
*/
tsp00_Int4
ls20ReadData(tls00_VFile         *pInfile,
             tls00_MetaData      *pMeta,
             tls00_FileFormat    *pFileFormat,
             bool                 bFirstTimeRead,
             tsp00_SwapKind_Enum  DSwap,
             tls00_String        *pBuffer,
             tsp00_Addr           pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls20ReadData");

    tsp00_Int4 rc = errOK_els00;

    tsp00_Int4 rlen = 0;

    if (true == pFileFormat->ffoCompact_ls00)
    {
        // In case of file format COMPACT the data file has the following structure where
        // the long data is only optional if the table has long columns
        //|------------------------------------------------------------------------------------ ...
        //| meta data1 | data1 | [long data1] | meta data2 | data2 | [long data2] | meta data3 | ...
        //|------------------------------------------------------------------------------------...
        //
        // To minimize the file access we use the following strategy for tables with no
        // long columns:
        // During first read of the data (indicated by recordcount = 0) we read the first
        // meta data1 and compute the length of the corresponding data1. Then we read
        // immediately afterwards this data1 because we know the length and the next meta data2 at once.
        // The length of the meta data is fix for the whole file. The meta data2 is kept
        // then until this function is entered next time. Then we compute the length of data2
        // and read data2 and the next meta data3 at once and so on.
        //
        // For tables with long columns we first read the meta data1 and compute the length
        // of the data1. Then we read only data1 and process it. Thus in this case meta data
        // and data is read separately.
        rlen = pMeta->mdLengthArrayLen_ls00 * sizeof(tsp00_Uint2);

        if ( (true == bFirstTimeRead) || (pMeta->mdLongFieldCount_ls00 > 0) )
        {
            rc = ls30VFReadLnEx(*pMeta->pmdMetaFile_ls00,
                                REINTERPRET_CAST(_TCHAR*, pMeta->pmdLengthArray_ls00),
                                pszErrText,
                                rlen);  //pMeta->mdLengthArrayLen_ls00 * sizeof(tsp00_Uint2));
            if (rc >= 0)
            {
                rc = errOK_els00;

                ls20ReadDataLength(pMeta, DSwap, pszErrText);
                if (0 == pMeta->mdLongFieldCount_ls00)
                {
                    rlen += pMeta->mdRecordLength_ls00;
                }
                else
                {
                    rlen = pMeta->mdRecordLength_ls00;
                }
            }
            else
            {
                if (EOF_LS00 == rc)
                {
                    rc = STOP_NOMORE_DATA_LS00;
                }
            }
        }
        else // if ( (true == bFirstTimeRead) || (pMeta->mdLongFieldCount_ls00 > 0) )
        {
            // save meta information
            memcpy(pMeta->pmdLengthArray_ls00, &pBuffer->strAddr_ls00[pMeta->mdRecordLength_ls00], rlen);

            // Analyse meta information. THIS CALL CHANGES mdRecordLength!!
            ls20ReadDataLength(pMeta, DSwap, pszErrText);
            rlen += pMeta->mdRecordLength_ls00;
        }
    }
    else
    {
        rlen = pFileFormat->ffoBytesPerLine_ls00;
    }


    // Read the real data. In case of compact format we read the meta information of
    // the next record at this time, too.
    if (errOK_els00 == rc)
    {
        pBuffer->strLeng_ls00 = ls30VFReadLnEx(*pInfile,       // read one row in data file
                                               pBuffer->strAddr_ls00,
                                               pszErrText,
                                               rlen);

        if (0 > pBuffer->strLeng_ls00)      // InputLen = EOF_LS00 (= -1)
        {                                   // InputLen = errFileRead_els98 in case of read error
            // Coming here the rc may have the values EOF_LS00 (- end reached and no data
            // read) or some error code
            if (EOF_LS00 == pBuffer->strLeng_ls00)
            {
                rc = STOP_NOMORE_DATA_LS00;
            }
            else
            {
                rc = pBuffer->strLeng_ls00;
            }
        }
        else
        {
            if ( (true == pFileFormat->ffoCompact_ls00) && (pBuffer->strLeng_ls00 < rlen) )
            {
                // Coming here we could not read the data and the next meta data as
                // requested. This can only mean that the file is fully read. We format
                // the length array buffer area to 0 to force a EOF_LS00 answer when
                // executing the read the next time.
                memset(&pBuffer->strAddr_ls00[pBuffer->strLeng_ls00],
                       0,
                       pMeta->mdLengthArrayLen_ls00 * sizeof(tsp00_Uint2));

                rc = STOP_LAST_DATA_LS00;
            }
        }
    }   // end if (errOK_els00 == rc)

    return rc;
}
// ls20ReadData()


/*
  -----------------------------------------------------------------------------
  function:     ls20ReadDataLength
  -----------------------------------------------------------------------------
*/
void
ls20ReadDataLength(tls00_MetaData *pMeta, tsp00_SwapKind_Enum DSwap, tsp00_Addr pszErrText)
{

    //  This function is used to read the length of data fields in case of
    //  file format COMPACT. The data is structured as follows
    //|------------------------------------------------------------------------------------ ...
    //| meta data1 | data1 | [long data1] | meta data2 | data2 | [long data2] | meta data3 | ...
    //|------------------------------------------------------------------------------------...
    //
    // meta data1 contains the length of all data1 and long data1 fields. Every data field
    // is characterized by a 2 byte integer length whereas every long data field is
    // characterized by a 4 byte integer. NULL values are indicated by a length of -12.
    // All length fields are signed integers.
    // The length values are swapped in the current machine swap.
    //
    // This function computes the record length, too. Because of negative length of NULL values
    // this length is not added to the record length. Additionally the length of long
    // values is also not added to the record length because longs are handled separately.

    
    ROUTINE_DBG_MEO00 ("ls20ReadDataLength");

    pMeta->mdRecordLength_ls00 = 0;

    tsp00_Uint4 i = 0;
    tsp00_Uint4 j = 0;
    bool        bLong = false;
    for (i,j; i < pMeta->mdFieldCount_ls00; ++i)
    {
        if (j < pMeta->mdLongFieldCount_ls00)
        {
            if ((i+1) == pMeta->pmdLongFieldPos_ls00[j])
            {
                if (DSwap != pMeta->mdSwapKind_ls00)
                {
                    ls07Swap4BCharInPlace(REINTERPRET_CAST(_TCHAR*, REINTERPRET_CAST(tsp00_Int4*, &pMeta->pmdLengthArray_ls00[i+j])),
                                          pMeta->mdSwapKind_ls00, DSwap);
                }
                bLong = true;
            }
        }

        if (false == bLong)
        {
            if (DSwap != pMeta->mdSwapKind_ls00)
            {
                pMeta->pmdLengthArray_ls00[i+j] = ls07Swap2BInt(pMeta->pmdLengthArray_ls00[i+j],
                                                                pMeta->mdSwapKind_ls00, DSwap);
            }
            if (INSERT_NULL != pMeta->pmdLengthArray_ls00[i+j])     // NULL-value; do not add length to record length
            {
                pMeta->mdRecordLength_ls00 += pMeta->pmdLengthArray_ls00[i+j];
            }
        }
        else
        {
            bLong = false;
            ++j;
        }
    } //while (i < pMeta->mdLengthArrayLen_ls00);

    return;
}
// ls20ReadDataLength()


/*
  -----------------------------------------------------------------------------
  function:     ls20BuildRecord
  -----------------------------------------------------------------------------
*/
tsp00_Longint
ls20BuildRecord(tls00_DataSource       *pDataSource,
                tls00_Buffer           *pDataPartBuffer,
                tls00_TableDescription *pTableDesc,
                tls00_MultColSpec      *pMultCol,
                tls00_LongExecInfo     *pLongExecInfo,
                tls00_CodeType          CodeType,
                bool                    bUnicodeOI,
                tsp00_Int4              RecordCount,
                tsp00_Addr              pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls20BuildRecord");

    tsp00_Longint     rc         = errOK_els00;
    tsp00_Int4        lResult    = 0;
    unsigned char*    pColValue  = NULL;
    tls00_Column*     pColumn    = NULL;
    tls00_ParamInfo*  pPI        = NULL;
    tls00_String*     RawData    = pDataSource->dasConvertedData;
    unsigned char*    pszBuffer  = &(STATIC_CAST(unsigned char*, pDataPartBuffer->pBufPtr_ls00)[pDataPartBuffer->lBufLen_ls00]);
    

    for(tsp00_Int4 i=0; (i < pTableDesc->tdFieldCount_ls00) && (rc >= 0); ++i)
    {
        pColumn = pMultCol->mcsColumn_ls00[pDataSource->plDataToColIndex_ls00[i]];

        // Even if loop index i corresponds exactly plDataToParamIndex_ls00[i] for DATALOAD
        // commands we need it here as we use this function for DATAUPDATE also; for DATAUPDATE
        // both indexes are not the same
        pPI       = pTableDesc->tdParamArray_ls00[pDataSource->plDataToParamIndex_ls00[i]];

        // Points to undef-byte before bufpos for value to build
        pColValue = pszBuffer + (pPI->piBufpos_ls00 - 1);

        lResult = ls20_ValueToOIBuffer(RawData[i], *pPI, pColumn, CodeType, bUnicodeOI, pColValue, pszErrText);
        if (lResult >= 0)
        {
            // Additional check/work for LONGs: if NOT NULL increase counter
            if ( (1           == szMap_LongDataType_ls00[pPI->piBasicInfo_ls00.fbDataType_ls00]) &&
                 (INSERT_NULL != RawData[i].strLeng_ls00) )
            {
                pLongExecInfo->dliNNLongs_ls00[pLongExecInfo->dliCntNNLongs_ls00] = i;
                ++pLongExecInfo->dliCntNNLongs_ls00;
            }
        }
        else    // OOOPs, error --> adapt return value, too!
        {
            RepServMessages *pMsgObj = RepServMessages::Instance(); // error msg obj

            pMsgObj->ls98Msg(layDataLoad_els98, pszErrText, errAtRow_els98, RecordCount);
            pMsgObj->ls98Msg(layDataLoad_els98, pszErrText, STATIC_CAST(ErrorNumbers, rc),
                                                            pTableDesc->tdColName_ls00[i]->asCharp());
            rc = lResult;
        }
    }   // end for(i=0, j=0; (i < TableInfo->tdFieldCount_ls00) && (lResult >= 0); i++)

    return rc;
}
// ls20BuildRecord()


/*!
  -----------------------------------------------------------------------------
  function:     ls20_ValueToOIBuffer
  -----------------------------------------------------------------------------
*/
tsp00_Int4
ls20_ValueToOIBuffer(tls00_String     &pInputData,            // = RawData
                     tls00_ParamInfo  &ParameterInfo,
                     tls00_Column*     pColumn,
                     tls00_CodeType    DataFileEncoding,
                     bool              bUnicodeOI,
                     unsigned char*    pOIValueBuffer,
                     tsp00_Addr        pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls20_ValueToOIBuffer");

    tsp00_Longint  rc       = 0;
    tsp00_Int4     lResult  = 0;
    tsp00_Longuint lDestLen = 0;

    tls00_FieldBasics   BasicInfo = ParameterInfo.piBasicInfo_ls00;
    tsp00_SwapKind_Enum CurrSwap  = ls07DefineSwap(NULL);

    switch(pInputData.strLeng_ls00)
    {
        case INSERT_DEFAULT:
        {
            *pOIValueBuffer = csp_default_byte;
            break;
        }
        case INSERT_NULL: 
        {
            *pOIValueBuffer = csp_undef_byte;
            break;
        }
        default:
        {
            *pOIValueBuffer = ls07GetDefByte(BasicInfo.fbDataType_ls00);     // Insert defined-byte

            if (1 == szMap_LongDataType_ls00[BasicInfo.fbDataType_ls00])
            {
                // Initialize Longdescriptor to vm_nodata
                ls08InitLONGDescriptor(REINTERPRET_CAST(tsp00_LongDescriptor*, (pOIValueBuffer + 1)));
            }
            else
            {
                if ( (true == bUnicodeOI) && (true == bUnicodeOverOI_ls00[BasicInfo.fbDataType_ls00]) )
                {
                    // coming here only char data types, blobs or date, time and timestamp arrive
                    // char must not be converted if the DataFileEncoding is already UCS2
                    // date, time and timestamp must ALWAYS be converted because the internal routines
                    // to convert the representation to ISO, ... work only with ascii data
                    if (1 == szMap_SpecConstDataType_ls00[BasicInfo.fbDataType_ls00])
                    {
                        lDestLen = BasicInfo.fbInOutLength_ls00;

                        rc = ls05ConvertToUCS2Simple(pOIValueBuffer + 1,
                                                     lDestLen,
                                                     REINTERPRET_CAST(unsigned char*, pInputData.strAddr_ls00),
                                                     pInputData.strLeng_ls00,
                                                     ctAscii_els00,
                                                     pszErrText);
                        if (errOK_els00 == rc)
                        {
                            lResult = STATIC_CAST(tsp00_Int4, lDestLen);
                        }
                    }
                    else
                    {
                        // char data already in UCS2 must not be converted again, but may be swapped;
                        // char data UTF8 or ASCII encoded must be converted here
                        if ( (ctUCS2_els00 != DataFileEncoding) && (ctUCS2Swapped_els00 != DataFileEncoding) )
                        {
                            lDestLen = BasicInfo.fbInOutLength_ls00;
                            
                            rc = ls05ConvertToUCS2(pOIValueBuffer + 1,
                                                   lDestLen,
                                                   REINTERPRET_CAST(unsigned char*, pInputData.strAddr_ls00),
                                                   pInputData.strLeng_ls00,
                                                   DataFileEncoding,
                                                   BasicInfo.fbDataType_ls00,
                                                   pszErrText);
                            if (errOK_els00 == rc)
                            {
                                lResult = STATIC_CAST(tsp00_Int4, lDestLen);
                            }
                        }
                        else
                        {
                            memcpy(pOIValueBuffer + 1, pInputData.strAddr_ls00, pInputData.strLeng_ls00);
                            lResult = pInputData.strLeng_ls00;
                        }
                    }
                
                }   // end if ( (true == bUnicodeOI) && (true == bUnicodeOverOI_ls00[BasicInfo.fbDataType_ls00]) )
                else
                {
                    memcpy(pOIValueBuffer + 1, pInputData.strAddr_ls00, pInputData.strLeng_ls00);
                    lResult = pInputData.strLeng_ls00;
                }

                if (lResult >= 0)
                {
                    // length parameter counts the defined byte
                    ls05InsertFillBytes(pOIValueBuffer,
                                        lResult + sizeof(char),
                                        BasicInfo.fbDataType_ls00,
                                        BasicInfo.fbInOutLength_ls00,
                                        bUnicodeOI,
                                        CurrSwap);
                }
            }
            break;
        }
    } // End OF SWITCH RAWDATA TYPE

    return lResult;
}
// ls20_ValueToOIBuffer()


/*!
  -----------------------------------------------------------------------------
  function:     ls20ConvertValue2InternEncoding
  -----------------------------------------------------------------------------
*/
tsp00_Longint
ls20ConvertValue2InternEncoding(tls00_String        &RawData,
                                tsp00_DataType_Enum  DataType,
                                tls00_CodeType       CodeType,
                                tsp00_Addr           pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls20ConvertValue2InternEncoding");

    tsp00_Longint rc = 0;

    //*
    //*     FIRST TRANSFORMATION of delivered value: bring it into 'internal'
    //*     encoding (ascii so far) to process it if the external encoding
    //*     is UNICODE (i.e. UCS2 or UTF8).
    //*
    //* Transformation to 'internal' encoding for values of type
    //*  - date, time, timestamp (value must be brought into the right format and
    //*    transformation routines may only handle ascii data)
    //*  - boolean - delivered as external user defined representation
    //*  - numeric - delivered as readable value
    //*  - blobs   - delivered in readable hex notation
    //*  - long    - data file contains meta information on real long values located
    //*              in long files; this meta information has to be converted; the real
    //*              longs are gotten later when processing the descriptor
    //* No transformation for values of all other types because
    //*  - char/varchar - transformed later with regards to the external representation
    bool bConversionNecessary = true;
    if (1 == szMap_CharDataType_ls00[DataType])
    {
         if (0 == szMap_BlobDataType_ls00[DataType])
         {
             bConversionNecessary = false;
         }
         else
         {
             if (ctUTF8_els00 == CodeType)
             {
                 // In case of UTF8 BLOBs (i.e. readable hex notation) does not need to
                 // be converted because it is in any case US7ASCII
                 bConversionNecessary = false;
             }
         }
    }   // end if (1 == szMap_LongDataType_ls00[DataType])


    if (true == bConversionNecessary)
    {
        if (ctUTF8_els00 == CodeType)
        {
            tsp00_Longuint len = RawData.strLeng_ls00;
            
            rc = ls05ConvertFromUTF8(REINTERPRET_CAST(unsigned char*, RawData.strAddr_ls00),
                                     len,
                                     REINTERPRET_CAST(unsigned char*, RawData.strAddr_ls00),
                                     RawData.strLeng_ls00,
                                     ctAscii_els00,
                                     pszErrText);
            if (errOK_els00 == rc)
            {
                RawData.strLeng_ls00 = STATIC_CAST(tsp00_Int4, len);
            }
        }
        else if ( (ctUCS2_els00 == CodeType) || (ctUCS2Swapped_els00 == CodeType) )
        {
            _TCHAR        szConversionBuffer[8192];
            tsp00_Longint lConversionBufLen        = 8192;

            memcpy(szConversionBuffer, RawData.strAddr_ls00, RawData.strLeng_ls00);

            rc = ls05ConvertFromUCS2Simple(REINTERPRET_CAST(unsigned char*, RawData.strAddr_ls00),
                                           REINTERPRET_CAST(unsigned char*, szConversionBuffer),
                                           lConversionBufLen,
                                           RawData.strLeng_ls00,
                                           ctAscii_els00,
                                           pszErrText);
            if (errOK_els00 == rc)
            {
                RawData.strLeng_ls00 = STATIC_CAST(tsp00_Int4, lConversionBufLen);
            }

        }
    }   // end if (true == bConversionNecessary)

    return rc;
}
// ls20ConvertValue2InternEncoding()


/*
  -----------------------------------------------------------------------------
  function:     ls20TransformColumnValue
  -----------------------------------------------------------------------------
*/
tsp00_Longint
ls20TransformColumnValue(tls00_String        *pRawValue,
                         tls00_Column        *pColumn,
                         tls00_String        *pInputBuffer,
                         tls00_ParamInfo     *pParamInfo,
                         tls00_FileSpec      *pFileSpec,
                         tsp00_SwapKind_Enum  SwapKind,
                         tsp1_sqlmode_Enum    SQLMode,
                         tsp00_Addr           pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls20TransformColumnValue");
    
    RepServMessages *pMsgObj = RepServMessages::Instance();    // error reporting obj
    tsp00_Longint    rc      = errOK_els00;

    tls00_FileFormat *pFileFormat = &pFileSpec->fsFormat_ls00;

    // last (INSERT_NULL) is valid only for fileformat COMPACT
    if ( (INSERT_EMPTY_VALUE != pRawValue->strLeng_ls00) && (INSERT_NULL != pRawValue->strLeng_ls00) )
    {
        //*
        //*     Trim values right - remove trailing blanks
        //* Done for LONG data that is of external data type CHAR in case of format COMPRESSED and FORMATTED
        //* these values in the data file are simply positions specs and can be trimmed
        //*
        if ( (false == pFileFormat->ffoCompact_ls00) && (ioChar_els00 == pColumn->colFormat_ls00.ffoFormat_ls00) )
        {
            if ( (true == pFileFormat->ffoFormatted_ls00) ||  //http://pts:1080/webpts?wptsdetail=yes&ErrorType=0&ErrorID=1118955
                 ( (true == pFileFormat->ffoCompress_ls00)   &&
                   (1    == szMap_LongDataType_ls00[pParamInfo->piBasicInfo_ls00.fbDataType_ls00]) ) )
            {
                tsp00_Longint len  = pRawValue->strLeng_ls00;
                // Handle byte values in files with FORMATTED BINARY differently from not binary
                // formatted files
                if ( (true == pFileFormat->ffoFormatted_ls00) && (true == pFileFormat->ffoBinary_ls00) &&
                     (1    == szMap_BlobDataType_ls00[pParamInfo->piBasicInfo_ls00.fbDataType_ls00]) &&
                     (0    == szMap_LongDataType_ls00[pParamInfo->piBasicInfo_ls00.fbDataType_ls00]) &&
                     (false == pColumn->colFormat_ls00.ffoIsHex_ls00) )
                {
                    ls05StrTrimRight(pRawValue->strAddr_ls00, len, '\0', pFileFormat->ffoCodeType_ls00);
                }
                else
                {
                    ls05StrTrimRight(pRawValue->strAddr_ls00, len, BLANK_LS00, pFileFormat->ffoCodeType_ls00);
                }
                // In case of format FORMATTED trimming prevents from errors where the data field exceeds
                // the allowed column length but only because it is padded with blanks;
                // but if the data consists entirely of blanks we'll retain it here
                if (len > 0)        // http://pts:1080/webpts?wptsdetail=yes&ErrorType=0&ErrorID=1124605
                {
                    pRawValue->strLeng_ls00 = STATIC_CAST(tsp00_Int4, len);
                }
            }
        }

        //*
        //*     Convert values to internal encoding (ASCII) for further processing
        //*
        if ( (ctAscii_els00 != pFileFormat->ffoCodeType_ls00)                                      &&
             (ioChar_els00  == pColumn->colFormat_ls00.ffoFormat_ls00)                             &&
             ( (true != pFileFormat->ffoCompact_ls00)                                             ||
               ( (true == pFileFormat->ffoCompact_ls00)                                         &&
                 (0    == szMap_LongDataType_ls00[pParamInfo->piBasicInfo_ls00.fbDataType_ls00]) ) ) )
        {
            rc = ls20ConvertValue2InternEncoding(*pRawValue,
                                                 pParamInfo->piBasicInfo_ls00.fbDataType_ls00,
                                                 pFileFormat->ffoCodeType_ls00,
                                                 pszErrText);
        }

        //* Check for:
        //*  - hex representation of data -> convert it
        //*  - NULL value representation
        //*  - LONGs
        //*  - specials like DATE, TIME, TIMESTAMP, BOOLEAN
        //*  - NUMBER (convert to internal db format)
    
        // (data in HEX form) or (BLOB data type and file format is not binary)?

        if (errOK_els00 == rc)
        {
            if ( (true == pColumn->colFormat_ls00.ffoIsHex_ls00) ||
                 ( (1 == szMap_BlobDataType_ls00[pParamInfo->piBasicInfo_ls00.fbDataType_ls00]) &&
                   (0 == szMap_LongDataType_ls00[pParamInfo->piBasicInfo_ls00.fbDataType_ls00]) &&
                   (false == pFileFormat->ffoBinary_ls00) ) )
            {                                                       // -->  convert to byte
                tsp00_Addr  pNewValue = NULL;
                tsp00_Int4  lNewValueLen = 0;
            
                rc = ls07ConvertHexToByte(pRawValue->strAddr_ls00, pRawValue->strLeng_ls00,
                                          &pNewValue,              lNewValueLen);
                if (errOK_els00 != rc)
                {
                    pMsgObj->ls98MsgToString(layDataLoad_els98, pszErrText, STATIC_CAST(ErrorNumbers,rc),
                                                                            pRawValue->strLeng_ls00,
                                                                            pRawValue->strAddr_ls00);
                }
                else
                {
                    pRawValue->strAddr_ls00 = pNewValue;
                    pRawValue->strLeng_ls00 = lNewValueLen;
                }
            }
        }

        // Check for NULL DEFINITION
        // NULL condition may be checked only for not COMPACT formats. In case of COMPACT format
        // the NULL value is already given with the read data - here the length has the
        // value -12 = INSERT_NULL
        // A NULL condition is specified in the load command if the null condition value is not NULL itself
        if ( (NULL != pColumn->colNullCond_ls00) && (false == pFileFormat->ffoCompact_ls00) )
        {
            rc = ls24GetNILConstants(*pRawValue,
                                     *pInputBuffer,
                                     pParamInfo->piMode_ls00,
                                     *pColumn->colNullCond_ls00,
                                     *pFileFormat);
            if (errNullNotAllowed_els98 == rc)
            {
                pMsgObj->ls98MsgToString(layDataLoad_els98, pszErrText,
                                                            STATIC_CAST(ErrorNumbers,rc),
                                                            pColumn->colName_ls00.asCharp());
            }
            // In case the NULL-check went without problems we have 2 cases to distinguish:
            // 1. there is a null-value --> leave the function (pRawValue->strLeng_ls00 is set to INSERT_NULL)
            // 2. there is a real value --> continue processing value
        }

        if ( (errOK_els00 == rc) && (INSERT_NULL != pRawValue->strLeng_ls00) )
        {
            // Process LONG value?
            if (1 == szMap_LongDataType_ls00[pParamInfo->piBasicInfo_ls00.fbDataType_ls00])
            {
                // There need to be data
                if ( (pRawValue->strLeng_ls00 >= 0) && (false == pFileFormat->ffoCompact_ls00) )
                {
                    rc = ls08ReadLongFileInfo(pColumn, pRawValue, pszErrText);
                    if (errOK_els00 != rc)
                    {
                        if( (errFileClose_els98 != rc) && (errFileOpen_els98 != rc) && (errFileRead_els98 != rc) )
                        {
                            pMsgObj->ls98MsgToString(layDataLoad_els98, pszErrText, errLongfileSpecMissing_els98);
                        }
                    }
                }
            }
            else
            {
                if (1 == szMap_NumberDataType_ls00[pParamInfo->piBasicInfo_ls00.fbDataType_ls00])
                {                                                       // --> convert to internal db format
                    rc = ls25PrepareNumbers(&pRawValue->strAddr_ls00,
                                            pRawValue->strLeng_ls00,
                                            *pColumn,
                                            pParamInfo->piBasicInfo_ls00,
                                            SwapKind,
                                            *pFileFormat);
                    if (errOK_els00 != rc)
                    {
                        if (errNumberTruncated_els98 != rc)
                        {
                            pMsgObj->ls98MsgToString(layDataLoad_els98, pszErrText, STATIC_CAST(ErrorNumbers, rc),
                                                                                    pRawValue->strLeng_ls00,
                                                                                    pRawValue->strAddr_ls00);
                        }
                    }
                }
                else if (1 == szMap_SpecConstDataType_ls00[pParamInfo->piBasicInfo_ls00.fbDataType_ls00])
                {
             
                    rc = ls24SpecialFormat(*pRawValue, pParamInfo->piBasicInfo_ls00, *pFileFormat);
                    if (errOK_els00 != rc)
                    {
                        pMsgObj->ls98MsgToString(layDataLoad_els98, pszErrText, STATIC_CAST(ErrorNumbers, rc),
                                                                                pRawValue->strLeng_ls00,
                                                                                pRawValue->strAddr_ls00);
                    }
                }
            }   // end else branch if (1 == szMap_LongDataType_ls00[pParamInfo->piBasicInfo_ls00.fbDataType_ls00])
        
        }   // end if ( (errOK_els00 == rc) && (INSERT_NULL != pRawValue->strLeng_ls00) )
    
    }   // end if ( (INSERT_EMPTY_VALUE != pRawValue->strLeng_ls00) && ...
    else if (INSERT_EMPTY_VALUE == pRawValue->strLeng_ls00)
    {
        // Coming here there is no data found in data file for specified column;
        // check if column has null condition (even an empty string may be defined
        // to represent the NULL value)
        if (NULL != pColumn->colNullCond_ls00)
        {
            rc = ls24GetNILConstants(*pRawValue,
                                     *pInputBuffer,
                                     pParamInfo->piMode_ls00,
                                     *pColumn->colNullCond_ls00,
                                     *pFileFormat);
        }

        // NULL-value check may reveal that the data is not the NULL value;
        // in this case the length remains unchanged (INSERT_EMPTY_VALUE)
        if (INSERT_EMPTY_VALUE == pRawValue->strLeng_ls00)
        {
            // nothing read but no null condition specified; check if data is of data type char
            if ( (1 == szMap_CharDataType_ls00[pParamInfo->piBasicInfo_ls00.fbDataType_ls00]) ||
                 (1 == szMap_BlobDataType_ls00[pParamInfo->piBasicInfo_ls00.fbDataType_ls00]) )
            {
                if (sp1sm_oracle == SQLMode)
                {
                    // if sqlmode = oracle an empty field means to insert NULL; check if it is defined
                    if ( true == pParamInfo->piMode_ls00.includes(sp1ot_optional) )
                    {
                        pRawValue->strLeng_ls00 = INSERT_NULL;
                    }
                    else
                    {
                        rc = errNullNotAllowed_els98;
                    }
                }
                else
                {
                    // distinguish between  byte and text char columns
                    pRawValue->strLeng_ls00 = 1;
                    pRawValue->strAddr_ls00 = new char[pRawValue->strLeng_ls00];
                    // ls07GetDefByte delivers right value depending on data type
                    pRawValue->strAddr_ls00[0] = ls07GetDefByte(pParamInfo->piBasicInfo_ls00.fbDataType_ls00);
                }
            }
            else
            {
                rc = errMissingData_els98;
            }
        }

        if (errOK_els00 != rc)
        {
            pMsgObj->ls98MsgToString(layDataLoad_els98, pszErrText, STATIC_CAST(ErrorNumbers, rc),
                                                                    pColumn->colName_ls00.asCharp());
        }
    }   // end else of if (INSERT_DEFAULT != pRawValue->strLeng_ls00)


    //*
    //* Check value length
    //* Length of number values will be checked twice because it is first time check at conversion
    //* --> doesn't matter
    //*
    if ( (errOK_els00 == rc) && (0 == szMap_LongDataType_ls00[pParamInfo->piBasicInfo_ls00.fbDataType_ls00]) )
    {
        rc = ls18CheckValueLength(pRawValue->strLeng_ls00, pParamInfo->piBasicInfo_ls00);
        if (rc != errOK_els00)
        {
            pMsgObj->ls98MsgToString(layDataLoad_els98, pszErrText, STATIC_CAST(ErrorNumbers, rc),
                                                                    pColumn->colName_ls00.asCharp());
        }
    }

    return rc;
}
// ls20TransformColumnValue()


/*
  -----------------------------------------------------------------------------
  function:     ls20Recover
  -----------------------------------------------------------------------------
*/
void
ls20Recover(tls00_DataPartHandling  *pHandling,
            tsp00_Int4               lLongColCnt,
            tls00_TableDescription  *pTD,
            tsp00_Int4              &RowCount,
            tsp00_Int4               lErrorPos,
            bool                    &bPutVal)
{
    ROUTINE_DBG_MEO00 ("ls20Recover");

    tsp00_Int4 rc = errOK_els00;
    
    tls00_Buffer   *pBuffer   = &pHandling->dphBuffer_ls00;
    unsigned char  *pszBufPtr = STATIC_CAST(unsigned char*, pBuffer->pBufPtr_ls00);

    if (RowCount > lErrorPos)
    {
        RowCount = RowCount - lErrorPos;

        tsp00_Int4 i = 0;
        tsp00_Int4 j = 0;

        //*
        //*     Move valid records in buffer to the beginning of the buffer overwriting
        //*     the already inserted and the erroneous one
        //*
        tsp00_Int4 lMoveDistance = pHandling->dphRecordSpace_ls00[lErrorPos - 1];

        memmove(&pszBufPtr[0],
                &pszBufPtr[lMoveDistance],
                (pHandling->dphRecordSpace_ls00[RowCount] - pHandling->dphRecordSpace_ls00[lErrorPos - 1]) );
        
        pBuffer->lBufSize_ls00 += lMoveDistance;                // adapt free
        pBuffer->lBufLen_ls00  -= lMoveDistance;                // and used space in packet buffer  


        //*
        //*     Special treatment for long columns: adapt positions
        //*
        tsp00_Int4 *pLongIndex = NULL;
        if (0 != lLongColCnt)
        {
            //*
            //*     Get the indexes of longs in param array
            //*
            pLongIndex = new tsp00_Int4[lLongColCnt];
            
            for (i = 0, j = 0; i < pTD->tdFieldCount_ls00; ++i)
            {   //*     LONG column ?
                if (1 == szMap_LongDataType_ls00[pTD->tdParamArray_ls00[i]->piBasicInfo_ls00.fbDataType_ls00])
                {
                    pLongIndex[j] = i;
                    ++j;
                }
            }
        }

        //*
        //*     Adapt long descriptors in every single record and
        //*     adapt values for dlrRecordSpace_ls00 and dlrRecordNumbers_ls00
        //*
        //* Recalculate valpos (absolute value) and use the length of the LONG for computing the
        //* beginning of the next record in buffer. If the LONGs are all NULL the record length is
        //* the length to move the pointer to match the next record.
        //* In case the long descriptor indicates that there is no LONG in the buffer
        //* (ld_valmode() = vm_nodata) treat them as if they were NULL (the long length is not set).
        //*
        tsp00_Int4            lLongLen  = 0;
        tsp00_LongDescriptor  LongDesc;
        unsigned char         cLongColDefByte;
        unsigned char        *pszRecord = &pszBufPtr[0];   // initializing pszRecord to first record in packet buffer
                                                           // this ptr will be wandering through the buffer
        for (i = 0; i < RowCount; ++i)                     // RowCount with new value here: RowCount - lErrorPos
        {
            for (j = 0, lLongLen = 0; j < lLongColCnt; ++j)
            {
                // We need to subtract 1 from piBufPos_ls00 because Bufpos starts counting at 1!
                cLongColDefByte = *(pszRecord + pTD->tdParamArray_ls00[pLongIndex[j]]->piBufpos_ls00 - 1);
                
                // We need to copy the long descriptor before manipulating it to be
                // sure not to cause a crash when casting it to an odd memory address
                memcpy(&LongDesc,
                       pszRecord + pTD->tdParamArray_ls00[pLongIndex[j]]->piBufpos_ls00,
                       sizeof(tsp00_LongDescriptor));

                if ( (csp_undef_byte != cLongColDefByte) &&
                     (vm_nodata      != LongDesc.ld_valmode()) )
                {
                    LongDesc.ld_valpos() -= lMoveDistance;          // adapt valpos - absolute value
                    lLongLen             += LongDesc.ld_vallen();   // add len of THIS longvalue to lLongLen

                    // We need to move the descriptor back to its place
                    memcpy(pszRecord + pTD->tdParamArray_ls00[pLongIndex[j]]->piBufpos_ls00,
                           &LongDesc,
                           sizeof(tsp00_LongDescriptor));
                }
            }

            pszRecord += lLongLen + pHandling->dphRecordLength_ls00;

            pHandling->dphRecordNumbers_ls00[i] = pHandling->dphRecordNumbers_ls00[i + lErrorPos];
            pHandling->dphRecordSpace_ls00[i]   = pHandling->dphRecordSpace_ls00[i + lErrorPos] - lMoveDistance;
        }

        if (NULL != pLongIndex)
        {
            delete [] pLongIndex;
            pLongIndex = NULL;
        }
        
    }
    else        //if (RowCount > lErrorPos)
    {
        // Ok, all rows of a packet are inserted -> reset row counter to 0
        pBuffer->lBufSize_ls00 = pBuffer->lBufLen_ls00 + pBuffer->lBufSize_ls00;  // adapt free space in packet buffer
        pBuffer->lBufLen_ls00  = 0;
        bPutVal                = false;                                           // reset bPutVal
        RowCount               = 0;
    }

    return;
}
// ls20Recover()

/*  
  -----------------------------------------------------------------------------
  Local functions: implementation
  -----------------------------------------------------------------------------
*/

/*
  -----------------------------------------------------------------------------
  function:     ls20_CreateDloadRec
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls20_CreateDloadRec(tls00_DloadRec *& pDLRec, tls00_DBInfo *DBInfo)
{
    ROUTINE_DBG_MEO00 ("ls20_CreateDloadRec");

    tsp00_Int4 rc = errOK_els00;

    pDLRec = new tls00_DloadRec;
    if (NULL == pDLRec)
    {
        rc = errNoMoreMemory_els98;
    }
    else
    {
        memset(pDLRec, 0, sizeof(tls00_DloadRec));

        pDLRec->dlrResultCount_ls00[0] = 0xff;
        
        pDLRec->dlrPartAttr_ls00.becomes(sp1pa_first_packet);
        
        pDLRec->dlrpLongDescriptorVector_ls00 = NULL;

        pDLRec->dlrSwapKind_ls00              = DBInfo->dbiOwnSwap_ls00;

        // Infile structure
        pDLRec->dlrInfile_ls00                = new tls00_VFile;
        memset(pDLRec->dlrInfile_ls00, 0, sizeof(tls00_VFile));

        // Meta file structure
        memset(&pDLRec->dlrMetaData_ls00, 0, sizeof(tls00_MetaData));

        memset(&pDLRec->dlrDataSource_ls00, 0, sizeof(tls00_DataSource));

        // Data buffer
        pDLRec->dlrDataPartHandling_ls00.dphBuffer_ls00.pBufPtr_ls00  = NULL;
        pDLRec->dlrDataPartHandling_ls00.dphBuffer_ls00.lBufSize_ls00 = 0;
        pDLRec->dlrDataPartHandling_ls00.dphBuffer_ls00.lBufLen_ls00  = 0;

        // Pattern matching structure
        memset(&pDLRec->dlrPattern_ls00, 0, sizeof(tls00_Pattern));
    }

    return rc;
}
//  ls20_CreateDloadRec()


/*
  -----------------------------------------------------------------------------
  function:     ls20_Destroy
  -----------------------------------------------------------------------------
*/
static void
ls20_Destroy(tls00_DloadRec *& pDLRec, tls00_DatLoadCmd *pDLCmd)
{
    ROUTINE_DBG_MEO00 ("ls20_Destroy");

    if (NULL != pDLRec)
    {
        tls00_TableDescription  &T  = pDLRec->dlrTableInfo_ls00;
        int                     i   = 0;

        // We have to delete only as much values as we have allocated.
        for (i=0; i < T.tdFieldCount_ls00; ++i)
        {
            if (NULL != T.tdParamArray_ls00[i])
            {
                delete [] T.tdParamArray_ls00[i];
                T.tdParamArray_ls00[i] = NULL;
            }
        }

        if(pDLRec->dlrDataPartHandling_ls00.dphRecordNumbers_ls00 != NULL)
        {
            delete [] pDLRec->dlrDataPartHandling_ls00.dphRecordNumbers_ls00;
            pDLRec->dlrDataPartHandling_ls00.dphRecordNumbers_ls00 = NULL;
        }

        if (NULL != pDLRec->dlrDataPartHandling_ls00.dphRecordSpace_ls00)
        {
            delete [] pDLRec->dlrDataPartHandling_ls00.dphRecordSpace_ls00;
            pDLRec->dlrDataPartHandling_ls00.dphRecordSpace_ls00 = NULL;
        }

        if (NULL != pDLRec->dlrDInfo_ls00.dliNNLongs_ls00)
        {
            delete [] pDLRec->dlrDInfo_ls00.dliNNLongs_ls00;
            pDLRec->dlrDInfo_ls00.dliNNLongs_ls00 = NULL;
        }

        if (NULL != pDLRec->dlrpLongDescriptorVector_ls00)
        {
            delete [] pDLRec->dlrpLongDescriptorVector_ls00;
            pDLRec->dlrpLongDescriptorVector_ls00 = NULL;
        }

        char szErr[256];
        szErr[0] = 0;

        // close infile
        ls30VFClose(*pDLRec->dlrInfile_ls00, szErr);

        // delete infile structure
        if (NULL != pDLRec->dlrInfile_ls00)
        {
            delete pDLRec->dlrInfile_ls00;
            pDLRec->dlrInfile_ls00 = NULL;
        }

        // Delete input buffer
        if (NULL != pDLRec->dlrDataSource_ls00.dasInputBuffer.strAddr_ls00)
        {
            delete [] pDLRec->dlrDataSource_ls00.dasInputBuffer.strAddr_ls00;
            pDLRec->dlrDataSource_ls00.dasInputBuffer.strAddr_ls00 = NULL;
        }

        // Delete converted data buffer
        ls18RecordDestructor(pDLRec->dlrTableInfo_ls00.tdFieldCount_ls00, pDLRec->dlrDataSource_ls00.dasConvertedData);

        // Delete remaining members of struct tls00_DataSource
        if (NULL != pDLRec->dlrDataSource_ls00.lReadPos_ls00)
        {
            delete [] pDLRec->dlrDataSource_ls00.lReadPos_ls00;
            pDLRec->dlrDataSource_ls00.lReadPos_ls00 = NULL;
        }
        if (NULL != pDLRec->dlrDataSource_ls00.plDataToColIndex_ls00)
        {
            delete [] pDLRec->dlrDataSource_ls00.plDataToColIndex_ls00;
            pDLRec->dlrDataSource_ls00.plDataToColIndex_ls00 = NULL;
        }
        if (NULL != pDLRec->dlrDataSource_ls00.plDataToParamIndex_ls00)
        {
            delete [] pDLRec->dlrDataSource_ls00.plDataToParamIndex_ls00;
            pDLRec->dlrDataSource_ls00.plDataToParamIndex_ls00 = NULL;
        }

        // delete pattern matching structure
        ls18DestructPattern(&pDLRec->dlrPattern_ls00);

        // clear meta data
        if (NULL != pDLRec->dlrMetaData_ls00.pmdMetaFile_ls00)
        {
            if (NULL != pDLRec->dlrMetaData_ls00.pmdLengthArray_ls00)
            {
                delete [] pDLRec->dlrMetaData_ls00.pmdLengthArray_ls00;
                pDLRec->dlrMetaData_ls00.pmdLengthArray_ls00 = NULL;
            }
        }

        // close LONG data files still open only if the file format is not COMPACT
        tls00_MultColSpec &MultCol = pDLCmd->dloColSpec_ls00;

        if ( (0 != pDLRec->dlrLongColCnt_ls00) && (false == pDLCmd->dloInfile_ls00.fsFormat_ls00.ffoCompact_ls00) )
        {        
            tls00_Column *pCol = NULL;

            for (i = 0; i < MultCol.mcsCount_ls00; ++i)
            {
                pCol = MultCol.mcsColumn_ls00[i];
                if ( (NULL != pCol->colLongInfo.loiFileStruct_ls00) &&
                     (true == pCol->colLongInfo.loiIsLong_ls00) )
                {
                    ls30VFClose(*pCol->colLongInfo.loiFileStruct_ls00, szErr);
                }
            }
        }

        // Allocated buffer for data
        if (NULL != pDLRec->dlrDataPartHandling_ls00.dphBuffer_ls00.pBufPtr_ls00)
        {
            delete [] (STATIC_CAST(unsigned char*, pDLRec->dlrDataPartHandling_ls00.dphBuffer_ls00.pBufPtr_ls00));
            pDLRec->dlrDataPartHandling_ls00.dphBuffer_ls00.pBufPtr_ls00 = NULL;
        }

        delete pDLRec;
        pDLRec = NULL;
    }   // end if (NULL != pDLRec)
}
//  ls20_Destroy()


/*!
  -----------------------------------------------------------------------------
  function:     ls20_GetRecord
  -----------------------------------------------------------------------------
*/
static int
ls20_GetRecord(tls00_DatLoadCmd  *pDLCmd,
               tls00_DloadRec    *pDLRec,
               tsp1_sqlmode_Enum  SQLMode,
               bool               bUnicodeOI,
               tsp00_Addr         ErrText)
{
    ROUTINE_DBG_MEO00 ("ls20_GetRecord");

    RepServMessages *pMsgObj = RepServMessages::Instance();    // error msg obj
    tsp00_Int4       rc      = errOK_els00;

    tls00_ParamInfo  *pPI     = NULL;
    tls00_FileFormat  Format  = pDLCmd->dloInfile_ls00.fsFormat_ls00;
    tls00_Column     *pCol    = NULL;
    tls00_String     *RawData = pDLRec->dlrDataSource_ls00.dasConvertedData;

    tsp00_Addr pszInputLine = pDLRec->dlrDataSource_ls00.dasInputBuffer.strAddr_ls00;
    tsp00_Int4 InputLen     = pDLRec->dlrDataSource_ls00.dasInputBuffer.strLeng_ls00;

    //*
    //*     Read data
    //*
    tsp00_Int4 lFailIndex = 0;
    if (true == Format.ffoFormatted_ls00)           // read formatted data
    {
        rc = ls18GetFormattedValueEx(pszInputLine,
                                     InputLen,
                                     &pDLCmd->dloColSpec_ls00,
                                     pDLRec->dlrDataSource_ls00.lReadFieldsCnt_ls00,
                                     pDLRec->dlrDataSource_ls00.plDataToColIndex_ls00,
                                     RawData,
                                     lFailIndex);
    }
    else if (true == Format.ffoCompress_ls00)
    {
        rc = ls18GetUnformattedValueEx(pszInputLine,
                                       InputLen,
                                       &pDLRec->dlrPattern_ls00,
                                       pDLRec->dlrDataSource_ls00.lReadFieldsCnt_ls00,
                                       pDLRec->dlrDataSource_ls00.lReadPos_ls00,
                                       pDLCmd->dloInfile_ls00.fsFormat_ls00.ffoCodeType_ls00,
                                       RawData,
                                       lFailIndex);
    }
    else if (true == Format.ffoCompact_ls00)
    {
        rc = ls18GetCompactValue(pszInputLine,
                                 &pDLRec->dlrMetaData_ls00,
                                 pDLRec->dlrDataSource_ls00.lReadFieldsCnt_ls00,
                                 pDLRec->dlrDataSource_ls00.lReadPos_ls00,
                                 RawData,
                                 &pDLCmd->dloColSpec_ls00,
                                 lFailIndex);
    }
    if (errOK_els00 != rc)
    {
        pMsgObj->ls98Msg(layDataLoad_els98, ErrText, errAtRow_els98, pDLRec->dlrRecordCount_ls00);

        if (errMissingData_els98 == rc)
        {
            pMsgObj->ls98Msg(layDataLoad_els98, ErrText, errMissingData_els98,
                                                         pDLRec->dlrTableInfo_ls00.tdColName_ls00[lFailIndex]->asCharp());
        }
    }

    tsp00_Longint len = 0;
    tsp00_Uint4   j   = 0;

    for (j=0; ( (errOK_els00 == rc) && (j < pDLRec->dlrDataSource_ls00.lReadFieldsCnt_ls00) ) ; ++j)
    {
        pPI  = pDLRec->dlrTableInfo_ls00.tdParamArray_ls00[j];
        pCol = pDLCmd->dloColSpec_ls00.mcsColumn_ls00[pDLRec->dlrDataSource_ls00.plDataToColIndex_ls00[j]];

        rc = ls20TransformColumnValue(&RawData[j],
                                      pCol,
                                      &pDLRec->dlrDataSource_ls00.dasInputBuffer,
                                      pPI,
                                      &pDLCmd->dloInfile_ls00,
                                      pDLRec->dlrSwapKind_ls00,
                                      SQLMode,
                                      ErrText);
        if (errOK_els00 != rc)
        {
            pMsgObj->ls98PrintMessage(layDataLoad_els98, errAtRow_els98, pDLRec->dlrRecordCount_ls00);
            pMsgObj->ls98EFromRcToFile(layDataLoad_els98, ErrText, STATIC_CAST(ErrorNumbers, rc));
        }

    }   // end for (j=0; ( (errOK_els00 == rc) && (j < pDLRec->dlrDataSource_ls00.lReadFieldsCnt_ls00) ) ; ++j)

    return rc;
}
//  ls20_GetRecord()


/*
  -----------------------------------------------------------------------------
  function:     ls20_Insert
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls20_Insert(tls00_DBInfo     *DBInfo,
            tls00_DatLoadCmd *DloadCmd,
            tls00_DloadRec   *DloadRec,
            bool              bRecoverLong,
            tsp00_Addr        ErrText,
            tsp00_Int4       &RowCount,
            bool             &bPutVal)
{
    ROUTINE_DBG_MEO00 ("ls20_Insert");

    tsp00_Int4 rc          = errOK_els00;
    tsp00_Int4 rcREAD      = errOK_els00;

    DloadRec->dlrPartAttr_ls00.becomes(sp1pa_next_packet);

    tsp00_Int4    CountOfRows     = RowCount;
    tls00_String &LineStruct      = DloadRec->dlrDataSource_ls00.dasInputBuffer;        // source buffer (char buffer)
    tls00_String *RawData         = DloadRec->dlrDataSource_ls00.dasConvertedData;
    tls00_Buffer *pDataPartBuffer = &DloadRec->dlrDataPartHandling_ls00.dphBuffer_ls00; // target buffer

    bool bFirst     = false;
    bool bUnicodeOI = ls00IsDBUnicode(DBInfo);


    // Check if there are some long values to insert before continuing (may happen if some
    // records could not be inserted and bPutVal was set to true
    if (true == bRecoverLong)
    {
        // LONGs? Insert them here. Last inserted 'normal' record is the one for
        // which we have to import the LONG
        rc = ls08InsertLongs(DloadRec->dlrDataPartHandling_ls00,
                             DloadRec->dlrDInfo_ls00,
                             DloadCmd->dloColSpec_ls00,
                             DloadRec->dlrTableInfo_ls00,
                             &DloadRec->dlrDataSource_ls00,
                             bPutVal,
                             bUnicodeOI,
                             ErrText);
        if (errOK_els00 == rc)
        {
            // save used space value for THIS record;
            // (iRow - 1) because iRow gives the number of rows handled up to this point; this must
            // be subtracted by 1 to come to the right index
            DloadRec->dlrDataPartHandling_ls00.dphRecordSpace_ls00[CountOfRows-1] = STATIC_CAST(tsp00_Int4, pDataPartBuffer->lBufLen_ls00);
        }
        else
        {   // Error -undo inserted record by restoring used space of previous record

            // At this point we might have added already some part of the LONG restoring to the packet;
            // We need to remove all: the record and the long data
            pDataPartBuffer->lBufSize_ls00 += (DloadRec->dlrDataPartHandling_ls00.dphRecordSpace_ls00[CountOfRows-1] -
                                                DloadRec->dlrDataPartHandling_ls00.dphRecordSpace_ls00[CountOfRows-2]);
            pDataPartBuffer->lBufLen_ls00  -= (DloadRec->dlrDataPartHandling_ls00.dphRecordSpace_ls00[CountOfRows-1] -
                                                DloadRec->dlrDataPartHandling_ls00.dphRecordSpace_ls00[CountOfRows-2]);
        }

        bRecoverLong = false;
    }



    // Get the records for one request
    for (CountOfRows;
         ( (CountOfRows < DloadRec->dlrMaxRowsPPckt_ls00) &&
           (pDataPartBuffer->lBufSize_ls00 >= DloadRec->dlrDataPartHandling_ls00.dphRecordLength_ls00) &&
           (errOK_els00 == rcREAD) && (errOK_els00 == rc) &&
           (false == bPutVal) );
         ++CountOfRows)
    {
        bFirst = (0 == DloadRec->dlrRecordCount_ls00) ? true : false;
        rcREAD = ls20ReadData(DloadRec->dlrInfile_ls00,
                              &DloadRec->dlrMetaData_ls00,
                              &DloadCmd->dloInfile_ls00.fsFormat_ls00,
                              bFirst,
                              DloadRec->dlrSwapKind_ls00,
                              &LineStruct,
                              ErrText);
        if (rcREAD < 0)
        {
            // following values may happen:
            // - STOP_LAST_DATA_LS00    --> process read record and then stop processing
            // - STOP_NOMORE_DATA_LS00  --> nothing more to process - get outa here
            // - error code             --> nothing more to process - get outa here
            if ( (STOP_NOMORE_DATA_LS00 != rcREAD) && (STOP_LAST_DATA_LS00 != rcREAD) )
            {
                rc = rcREAD;
            }
            else if (STOP_NOMORE_DATA_LS00 == rcREAD)
            {
                --CountOfRows;
            }
            
            DloadRec->dlrPartAttr_ls00.becomes(sp1pa_last_packet);  // Last data in any case
        }


        //*
        //*     Check for load condition for table
        //*
        if ( (errOK_els00 == rc) && (STOP_NOMORE_DATA_LS00 != rcREAD) )
        {
            if (NULL != DloadCmd->dloTable_ls00.tsCondSpec_ls00)
            {
                if ( false == ls16CheckCondition(*DloadCmd->dloTable_ls00.tsCondSpec_ls00,
                                                 LineStruct,
                                                 DloadCmd->dloInfile_ls00.fsFormat_ls00) )
                {
                    --CountOfRows;
                    continue;
                }
            }
        }

        if ( (errOK_els00 == rc) && (STOP_NOMORE_DATA_LS00 != rcREAD) )
        {

            DloadRec->dlrRecordCount_ls00++;

            rc = ls20_GetRecord(DloadCmd,     //  insert record into RawData structure
                                DloadRec,
                                DBInfo->dbiSession_ls00.siSQLMode_ls00,
                                bUnicodeOI,
                                ErrText);
        }

        
        if ( (errOK_els00 == rc) && (STOP_NOMORE_DATA_LS00 != rcREAD) )
        {
            //*
            //* Save build record number for later error messages
            //*
            DloadRec->dlrDataPartHandling_ls00.dphRecordNumbers_ls00[CountOfRows] = DloadRec->dlrRecordCount_ls00;
            
            //*
            //* set LONG value descriptions
            //*
            DloadRec->dlrDInfo_ls00.dliCntNNLongs_ls00      = 0;
            DloadRec->dlrDInfo_ls00.dliFinishedNNLongs_ls00 = 0;

            rc = ls20BuildRecord(&DloadRec->dlrDataSource_ls00,
                                 pDataPartBuffer,
                                 &DloadRec->dlrTableInfo_ls00,
                                 &DloadCmd->dloColSpec_ls00,
                                 &DloadRec->dlrDInfo_ls00,
                                 DloadCmd->dloInfile_ls00.fsFormat_ls00.ffoCodeType_ls00,
                                 bUnicodeOI,
                                 DloadRec->dlrRecordCount_ls00,
                                 ErrText);
            if (errOK_els00 == rc)
            {
                // adapt free and used space in buffer
                pDataPartBuffer->lBufSize_ls00 -= DloadRec->dlrDataPartHandling_ls00.dphRecordLength_ls00;
                pDataPartBuffer->lBufLen_ls00  += DloadRec->dlrDataPartHandling_ls00.dphRecordLength_ls00;

                if (0 != DloadRec->dlrDInfo_ls00.dliCntNNLongs_ls00)
                {
                    // LONGs? Insert them here. Last inserted 'normal' record is the one for
                    // which we have to import the LONG
                    rc = ls08InsertLongs(DloadRec->dlrDataPartHandling_ls00,
                                         DloadRec->dlrDInfo_ls00,
                                         DloadCmd->dloColSpec_ls00,
                                         DloadRec->dlrTableInfo_ls00,
                                         &DloadRec->dlrDataSource_ls00,
                                         bPutVal,
                                         ls00IsDBUnicode(DBInfo),
                                         ErrText);
                    if (errOK_els00 == rc)
                    {
                        // save used space value for THIS record
                        DloadRec->dlrDataPartHandling_ls00.dphRecordSpace_ls00[CountOfRows] = pDataPartBuffer->lBufLen_ls00;
                    }
                    else
                    {   // Error --> undo inserted record by restoring used space of previous record
                        pDataPartBuffer->lBufSize_ls00 += DloadRec->dlrDataPartHandling_ls00.dphRecordLength_ls00;
                        pDataPartBuffer->lBufLen_ls00  -= DloadRec->dlrDataPartHandling_ls00.dphRecordLength_ls00;
                    }
                }
                else
                {
                    // save used space value for THIS record
                    DloadRec->dlrDataPartHandling_ls00.dphRecordSpace_ls00[CountOfRows] = pDataPartBuffer->lBufLen_ls00;
                }
            }
        }

        // Following if-branch handles errors from GetRecord and BuildRecord. Both functions log the error
        // so we have to adapt the error count only here.
        if (errOK_els00 != rc)
        {
            ++DloadRec->dlrLRejected_ls00;
            if(DloadRec->dlrLRejected_ls00 >= DBInfo->dbiSession_ls00.siMaxAdmissibleErrors)
            {
                rc = msgTableNotImport_els98;
            }
            else
            {
                rc = errOK_els00;       // reset the error code to read more data;
            }

            DloadRec->dlrLastCommLine_ls00 = DloadRec->dlrRecordCount_ls00 - 1;
            --CountOfRows;
        }

        ls18RecordInit(DloadRec->dlrDataSource_ls00.lReadFieldsCnt_ls00, RawData, LineStruct);

    } // END OF FOR count_of_rows

    RowCount = CountOfRows;

    if ( (errOK_els00 == rc) && (STOP_NOMORE_DATA_LS00 == rcREAD) )
    {
        rc = rcREAD;
    }
            
    return rc;  
}
//  ls20_Insert()


/*!
  -----------------------------------------------------------------------------
  function:     ls20_Send
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls20_Send(tls00_DBInfo     *DBInfo,
          tls00_DatLoadCmd *DloadCmd,
          tls00_DloadRec   *DloadRec,
          tls00_String     &pOI_InitBlock,
          tsp00_Addr        ErrText)
{
    ROUTINE_DBG_MEO00 ("ls20_Send");
    
    RepServMessages *pMsgObj = RepServMessages::Instance();    // error reporting obj

    tsp00_Int4 rc       = errOK_els00;
    tsp00_Int4 rcINSERT = errOK_els00;

    tsp1_packet   *SqlPacket    = NULL;
    tsp1_part     *DataPart     = NULL;
    tsp00_Longint  lCommitCount = 0;

    // Define max count of rows to send before sending commit
    tsp00_Int4 lRowCntBeforeCommit = (0 == DBInfo->dbiSession_ls00.siTASize_ls00) ?
                                      DloadRec->dlrMaxRowsPPckt_ls00 :
                                      DBInfo->dbiSession_ls00.siTASize_ls00;

    // We now use DloadRec->dlrMaxRowsPPckt_ls00 as a variable to ls20_Insert
    if ( (0 != DBInfo->dbiSession_ls00.siTASize_ls00) && (DBInfo->dbiSession_ls00.siTASize_ls00 < DloadRec->dlrMaxRowsPPckt_ls00) )
    {
        DloadRec->dlrMaxRowsPPckt_ls00 = DBInfo->dbiSession_ls00.siTASize_ls00;
    }
    
    tsp00_Int4 MaxRowsInPacket     = DloadRec->dlrMaxRowsPPckt_ls00;        // to memorize the value for later user

    tls00_DataPartHandling *pHandling    = &DloadRec->dlrDataPartHandling_ls00;
    tls00_Buffer           *pDataBuffer  = &pHandling->dphBuffer_ls00;
    bool                    bPutVal      = false;
    bool                    bRecoverLong = false;
    tsp00_Int4              RowCount     = 0;
    tsp00_Int4              tmpRowCnt    = 0;
    char                    szErr[MAX_REPLY_LENGTH_LS00];
    szErr[0] = 0;

    do
    {
        //*
        //*     Check RowCount from processing data:
        //* RowCount is either MaxRowsInPacket (initialized/reinitialized) or
        //* has a lower value (when called ls20Recover)
        //*
        //* The count of rows used in call to ls20_Insert depends on 2 values:
        //* 1. overall count of rows to insert set by customer using START option in DATALOAD command
        //* 2. count of rows that can be send using 1 order interface packet
        //*
        if ( (UNDEFINED_LS00 != DloadCmd->dloInfile_ls00.fsExtract_ls00.feEnd_ls00) && (0 == RowCount) )
        {
            tsp00_Int4 lDifference = STATIC_CAST(tsp00_Int4, (DloadCmd->dloInfile_ls00.fsExtract_ls00.feEnd_ls00 - 
                                      (DloadRec->dlrLRejected_ls00 + DloadRec->dlrInsertedRows_ls00)) );
            // Check if count of rows to insert is already reached
            if (lDifference > 0)
            {
                if (lDifference < MaxRowsInPacket)
                {
                    DloadRec->dlrMaxRowsPPckt_ls00 = lDifference;
                }
                else
                {
                    DloadRec->dlrMaxRowsPPckt_ls00 = MaxRowsInPacket;
                }
            }
            else // if (lDifference >= 0)       lt should not happen
            {
                rcINSERT = STOP_NOMORE_DATA_LS00;
            }
        }


        // Insert data into buffer
        if ( (errOK_els00 == rcINSERT) && (errOK_els00 == rc) )
        {
            rcINSERT = ls20_Insert(DBInfo, DloadCmd, DloadRec, bRecoverLong, ErrText, RowCount, bPutVal);
            if ( (errOK_els00 != rcINSERT) && (STOP_NOMORE_DATA_LS00 != rcINSERT) )
                // At this point we know:
            {                           // - either a critical error happened (error reading file ...)
                rc = rcINSERT;          // - or count of errors exceeds maxerrorcount
            }
        }

        if ( (0 == RowCount) && (errOK_els00 == rc) )  // Mmmh, no row read, but no error --> nothing to read -> exit
        {
            rcINSERT = STOP_NO_DATA_LS00;
        }

        //*
        //* Everything went fine: Insert data into packet
        //*

        if ( (errOK_els00 == rcINSERT) || (STOP_NOMORE_DATA_LS00 == rcINSERT) )
        {
            //*
            //*     Before sending the stuff to the kernel we check for cancel request
            //*
            if (errOK_els00 == rc)
            {
                rc = ls01CheckForCancel(DBInfo->dbiSession_ls00.sipCancelByte_ls00);
                if (errOK_els00 != rc)
                {
                    RepServMessages::Instance()->ls98Msg(layDataLoad_els98, ErrText, errCancelRequest_els98);
                }
            }

            do
            {
            // First we copy initialized packet block to right place in packet
            SqlPacket = DBInfo->dbiPktSndList_ls00[0];
            memcpy((char*) &SqlPacket->sp1_header, pOI_InitBlock.strAddr_ls00, pOI_InitBlock.strLeng_ls00);

            // Get pointer to data part to work with
            ls04GetPart(SqlPacket, sp1pk_data, DataPart);

            // Set part attribute to current value (either nextpacket or lastpacket)
            ls04SetPartAttribute(DataPart, DloadRec->dlrPartAttr_ls00);
            DataPart->sp1p_arg_count() = RowCount;                          // insert count of records in packet
            ls04AddPartAttribute(DataPart, DloadRec->dlrPartAttr_ls00);     // mark last packet

            rc = ls04BufToPart(DataPart, REINTERPRET_CAST(char*, pDataBuffer->pBufPtr_ls00), pDataBuffer->lBufLen_ls00);
            if (errOK_els00 == rc)
            {
                rc = ls04FinishPart (SqlPacket, DataPart);          // finish packet
            }


            //*
            //*     Send single insert packet to kernel and process kernels answer
            //*
            if (errOK_els00 == rc)
            {
                tmpRowCnt = RowCount;       // temp var needed for error handling when inserting longs with putval

                //*
                //*     Send and receive and handle error in SQL packet
                //*
                szErr[0] = 0;
                tsp00_Int4 lErrorPos = -1;                  // initialize to impossible error position
                rc = ls03ProcessSQLPacket(DBInfo, 0, lErrorPos, szErr);
                if (errOK_els00 == rc)
                {
                    // reporting
                    DloadRec->dlrInsertedRows_ls00 += RowCount;
                    DloadRec->dlrLastCommLine_ls00  = DloadRec->dlrDataPartHandling_ls00.dphRecordNumbers_ls00[RowCount - 1];
        
                    // Reinitializing
                    pDataBuffer->lBufSize_ls00 = pDataBuffer->lBufLen_ls00 + pDataBuffer->lBufSize_ls00;
                    pDataBuffer->lBufLen_ls00  = 0;
                    RowCount                   = 0;
                }
                else
                {
                    // In case the error position (lErrRecordNo) is not adapted but an error reported
                    // the error happened sending and/or receiving the packet and we stop at this point
                    if (-1 != lErrorPos)
                    {
                        pMsgObj->ls98PrintMessage(layDataLoad_els98, errAtRow_els98, DloadRec->dlrDataPartHandling_ls00.dphRecordNumbers_ls00[lErrorPos - 1]);
                        pMsgObj->ls98Msg(layDataLoad_els98, ErrText, errSimpleSQL_els98, szErr);

                        // Set error code
                        rc = errSimpleSQL_els98;

                        ++DloadRec->dlrLRejected_ls00;                  // increase error counter
                        DloadRec->dlrInsertedRows_ls00 += lErrorPos - 1;
        
                        if (1 != lErrorPos)                             // adapt only if not the first
                        {                                               // record was erroneous
                            DloadRec->dlrLastCommLine_ls00 = DloadRec->dlrDataPartHandling_ls00.dphRecordNumbers_ls00[lErrorPos - 2];
                        }

                        //*
                        //*     Check if user specified number of errors allowed and if we already exceed this number
                        //*
                        if (DloadRec->dlrLRejected_ls00 < DBInfo->dbiSession_ls00.siMaxAdmissibleErrors)
                        {
                            ls20Recover(&DloadRec->dlrDataPartHandling_ls00,
                                        DloadRec->dlrLongColCnt_ls00,
                                        &DloadRec->dlrTableInfo_ls00,
                                        RowCount,
                                        lErrorPos,
                                        bPutVal);

                            rc = errOK_els00;       // reset error text
                            ErrText[0] = '\0';
                            
                            // If we have some place in the packet we must recover longs first
                            if (true == bPutVal)
                            {
                                bRecoverLong = true;
                                bPutVal      = false;
                            }

                        }
                    }
                }
            }   // end if (errOK_els00 == rc)
            }
            while ( (STOP_NOMORE_DATA_LS00 == rcINSERT) && (RowCount > 0) && (errOK_els00 == rc) );


            //*
            //*     Process remaining LONGs using putval
            //*
            if (errOK_els00 == rc)
            {
//TODOTODO     MUSS DER ROWCOUNT HIER WIRKLICH NOCH UEBERPRUEFT WERDEN ODER GEHT ES AUCH SO???????????
                if (true == bPutVal)     // RowCount is set to 0 if there was no error in the data
                {
                    rc = ls08ExecPutVal(*pDataBuffer,
                                        DBInfo,
                                        DloadRec->dlrDInfo_ls00,
                                        DloadRec->dlrpLongDescriptorVector_ls00,
                                        DloadCmd->dloColSpec_ls00,
                                        DloadRec->dlrPartAttr_ls00,
                                        DloadRec->dlrTableInfo_ls00,
                                        ErrText);
                    if (errOK_els00 != rc)
                    {
                        ++DloadRec->dlrLRejected_ls00;
                        --DloadRec->dlrInsertedRows_ls00;
                        DloadRec->dlrLastCommLine_ls00 = DloadRec->dlrDataPartHandling_ls00.dphRecordNumbers_ls00[tmpRowCnt - 2];

                        pMsgObj->ls98Msg(layDataLoad_els98, ErrText, errAtRow_els98,
                                                                     pHandling->dphRecordNumbers_ls00[tmpRowCnt - 1]);
                        if(DloadRec->dlrLRejected_ls00 >= DBInfo->dbiSession_ls00.siMaxAdmissibleErrors)
                        {
                            rc = msgTableNotImport_els98;
                        }
                        else
                        {
                            ErrText[0] = '\0';
                        }
                    }
                    bPutVal = false;    // reinitialize
                }
            }

            // Commit in any case
            tsp00_Longint tmpCnt = DloadRec->dlrInsertedRows_ls00 / lRowCntBeforeCommit;
            if ( tmpCnt > lCommitCount)
            {
                lCommitCount = tmpCnt;
                rc = ls15RSCommit(DBInfo, ErrText);

                TRACE_PRNF_MLS99(("ls20_Send", "Commit after %d rows\n", DloadRec->dlrInsertedRows_ls00) );
            }

        }   // end if ( (errOK_els00 == rcINSERT) || (STOP_NOMORE_DATA_LS00 == rcINSERT) )

    }   // end while (1)
    while ( (errOK_els00 == rcINSERT) && (errOK_els00 == rc));

    // Send again an commit (even if errors occurred) if at least one row was successfully inserted.
    if (0 != DloadRec->dlrInsertedRows_ls00)
    {
        ls15RSCommit(DBInfo, ErrText);
    }

    return rc;
}
//  ls20_Send()


/*
  -----------------------------------------------------------------------------
  function:     ls20_ParseInsertStmt
  -----------------------------------------------------------------------------
*/
static int
ls20_ParseInsertStmt(tls00_DBInfo*     pDBInfo,
                     tls00_DatLoadCmd* pDloadCmd,
                     tls00_DloadRec*   pDloadRec,
                     tsp00_Addr        pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls20_ParseInsertStmt");

    RepServMessages *pMsgObj = RepServMessages::Instance();    // error reporting obj
    tsp00_Int4       rc      = errOK_els00;

    tls00_MultColSpec      &MultCol       = pDloadCmd->dloColSpec_ls00;
    tls00_TableSpec        &TableSpec     = pDloadCmd->dloTable_ls00;
    tls00_TableDescription *pTI           = &pDloadRec->dlrTableInfo_ls00;
    tls00_Column*           pColumn       = NULL;

    tsp00_Int4             _lCmdLen = 0;
    bool                   _bAssignments  = false;
    tsp00_Int4             i              = 0;
    tsp00_Uint4            _lPos          = 0;


    //*
    //*     - Check if user specified column names in command; if not get them here
    //*     - Check for assignments in command
    //*
    if (0 == MultCol.mcsCount_ls00)
    {
        // No columns specified in command: get'em from db
        rc = ls13GetColumnNames(pDBInfo, TableSpec.tsFullTabName_ls00, pTI, pszErrText);
        if (errOK_els00 != rc)
        {   
            return rc;
        }
        else
        {
            MultCol.mcsCount_ls00 = pTI->tdFieldCount_ls00;

            // Create structure MultCol
            for (int i=0; i < pTI->tdFieldCount_ls00; i++)
            {
                if (NULL == MultCol.mcsColumn_ls00[i])
                {
                    MultCol.mcsColumn_ls00[i] = new tls00_Column;
                    memset(MultCol.mcsColumn_ls00[i], DEF_BYTE_LS00, sizeof(tls00_Column));
                }
                _tcscpy(MultCol.mcsColumn_ls00[i]->colName_ls00.asCharp(), pTI->tdColName_ls00[i]->asCharp());
            }
        }
    }
    else
    {
        // User specified columns in DATALOAD command 
        // --> assignments are possible; we check for assignments and add their length here already
        for (int j=0, k=0; j < MultCol.mcsCount_ls00; ++j)
        {
            pColumn = MultCol.mcsColumn_ls00[j];

            if (asUnknown_els00 != pColumn->colAssign_ls00.fasAsgnTyp_ls00)
            {
                if (0 != pColumn->colAssign_ls00.fasAsgnLen_ls00)
                {
                    _lCmdLen += pColumn->colAssign_ls00.fasAsgnLen_ls00;
                }
                else
                {
                    _lCmdLen += _tcslen(szDBFunctionNames_ls20[pColumn->colAssign_ls00.fasAsgnTyp_ls00-DBFUNC_NAME_OFFSET_LS20]);
                }
            }
            else
            {
                // Column names are of interest only if those columns must be filled with
                // data from data file; that doesn't apply for assignments
                pTI->tdColName_ls00[k] = new tsp00_KnlIdentifierc;
                _tcscpy(pTI->tdColName_ls00[k]->asCharp(), pColumn->colName_ls00.asCharp());
                ++k;
            }
        }
        if (_lCmdLen > 0)
        {
            _bAssignments = true;
        }
    }


    //*
    //*     Build INSERT statement and parse it
    //* First time parsing is done as if the user did not specify assignments. This
    //* way the kernel returns information on the columns to handle. This information
    //* is necessary for second time parsing in case of assignments: in case of
    //* literal assignment we have to decide whether to include the data in single
    //* quotes or not (depending on whether the columns are number or char columns).
    //* 
    _lCmdLen += (MultCol.mcsCount_ls00 + 12) * sizeof(tsp00_KnlIdentifier);

    char *pszCmd = new char[_lCmdLen];
    pszCmd[0] = '\0';

    if (0 != TableSpec.tsTabOwnr_ls00.length())
    {
        _lPos = sp77sprintf(pszCmd, _lCmdLen, "INSERT INTO \"%s\".\"%s\" (\"%s",
                                              TableSpec.tsTabOwnr_ls00.asCharp(),
                                              TableSpec.tsTabName_ls00.asCharp(),
                                              MultCol.mcsColumn_ls00[0]->colName_ls00.asCharp());
    }
    else
    {
        _lPos = sp77sprintf(pszCmd, _lCmdLen, "INSERT INTO \"%s\" (\"%s", 
                                              TableSpec.tsTabName_ls00.asCharp(),
                                              MultCol.mcsColumn_ls00[0]->colName_ls00.asCharp());
    }
    for (i=1; i < MultCol.mcsCount_ls00; i++)
    {
        _lPos += sp77sprintf(pszCmd+_lPos, _lCmdLen-_lPos, "\",\"%s", MultCol.mcsColumn_ls00[i]->colName_ls00.asCharp());
    }
    _lPos += sp77sprintf(pszCmd+_lPos, _lCmdLen-_lPos, "\") VALUES(?");

    // Define pos in statement where value specification starts for later use in case
    // the user specified assignments in command
    size_t _lValueStartPos = _lPos - sizeof(char);

    for (i=1; i < MultCol.mcsCount_ls00; ++i)
    {
        _lPos += sp77sprintf(pszCmd+_lPos, _lCmdLen-_lPos, ",?");
    }
    _lPos += sp77sprintf(pszCmd+_lPos, _lCmdLen-_lPos, ") %s", szDuplicateString[pDloadCmd->dloDplMode_ls00]);

    TRACE_PRNF_MLS99(("ls20_ParseInsertStmt", "Stmt to parse: %s\n", pszCmd) );

    // Let's parse
    rc = ls13Parse(pszCmd, pDBInfo, pDloadRec->dlrParseId_ls00, pTI, pszErrText, false, true);
    if ( (errOK_els00 == rc) && (true == _bAssignments) )
    {
        TRACE_PRNF_MLS99(("ls20_ParseInsertStmt", "Assignments in load command; parsing 2nd time\n") );

        //*
        //*     Build INSERT statement a second time and parse again if the user
        //*     specified assignments in command
        //*

        // We have filled
        //                   pTI->tdParamArray_ls00
        // with the first execution of ls13Parse;
        // we'll use it here for literal assignments to decide wether the value has to
        // be included in single quotes or not (data type);

        tls00_FieldAsgn* pAssignment = NULL;
        
        pszCmd[_lValueStartPos] = 0;    // reinitialize pszCmd string to the pos where the value spec begins
        _lPos                   = _lValueStartPos;

        for (i=0; i < MultCol.mcsCount_ls00; ++i)
        {
            pAssignment = &MultCol.mcsColumn_ls00[i]->colAssign_ls00;
            
            if (asUnknown_els00 == MultCol.mcsColumn_ls00[i]->colAssign_ls00.fasAsgnTyp_ls00)
            {
                _lPos += sp77sprintf(pszCmd+_lPos, _lCmdLen-_lPos, "?,");
            }
            else
            {
                switch (pAssignment->fasAsgnTyp_ls00)
                {
                    case asLiteral_els00:
                    {
                        if (1 == szMap_NumberDataType_ls00[pTI->tdParamArray_ls00[i]->piBasicInfo_ls00.fbDataType_ls00])
                        {
                            _lPos += sp77sprintf(pszCmd+_lPos, _lCmdLen-_lPos, "%s,", pAssignment->fasAsgnVal_ls00);
                        }
                        else
                        {
                            _lPos += sp77sprintf(pszCmd+_lPos, _lCmdLen-_lPos, "'%s',", pAssignment->fasAsgnVal_ls00);
                        }
                        break;
                    }
                    default:
                    {
                        _lPos += sp77sprintf(pszCmd+_lPos, _lCmdLen-_lPos, 
                                                           "%s,", szDBFunctionNames_ls20[pAssignment->fasAsgnTyp_ls00-DBFUNC_NAME_OFFSET_LS20]);
                    }
                }   // end switch (MultCol.mcsColumn_ls00[i]->colAssign_ls00.fasAsgnTyp_ls00)
            }
        
        }
        _lPos -= sizeof(char);
        _lPos += sp77sprintf(pszCmd+_lPos, _lCmdLen-_lPos, ") %s", szDuplicateString[pDloadCmd->dloDplMode_ls00]);


        // Before running ls13Parse the 2nd time we have to free
        //                   pTI->tdParamArray_ls00
        // in order to avoid memory leaks
        for (tsp00_Int4 j=0; j < pTI->tdFieldCount_ls00; ++j)
        {
            if (NULL != pTI->tdParamArray_ls00[j])
            {
                delete [] pTI->tdParamArray_ls00[j];
                pTI->tdParamArray_ls00[j] = NULL;
            }
        }
        pTI->tdFieldCount_ls00 = 0;

        TRACE_PRNF_MLS99(("ls20_ParseInsertStmt", "Stmt to parse (2nd): %s\n", pszCmd) );

        rc = ls13Parse(pszCmd, pDBInfo, pDloadRec->dlrParseId_ls00, pTI, pszErrText, false, true);
    }   // end if ( (errOK_els00 == rc) && (true == _bAssignments) )
    
    delete [] pszCmd;
    
    if (errOK_els00 == rc)
    {
        //*
        //*     Define record length
        pDloadRec->dlrDataPartHandling_ls00.dphRecordLength_ls00 = ls18GetRecordLength(*pTI, true);
    
        //*
        //*     Compute max records to send in one packet
        //* Parts resultcount und parseid need to fit into packet too!
        //*
        pDloadRec->dlrMaxRowsPPckt_ls00 = ( pDBInfo->dbiPktSize_ls00 -
                                           (BlockHSize_ls00 + PartHSize_ls00 + PartHSize_ls00 +
                                            sizeof(pDloadRec->dlrResultCount_ls00) + MAX_PARSEID_LENGTH_LS00 + 16) ) /
                                           pDloadRec->dlrDataPartHandling_ls00.dphRecordLength_ls00;
        if(0 == pDloadRec->dlrMaxRowsPPckt_ls00)
        {
            pDloadRec->dlrMaxRowsPPckt_ls00 = 1;
        }

        TRACE_PRNF_MLS99(("ls20_ParseInsertStmt", "Max count of rows per packet: %d \n", pDloadRec->dlrMaxRowsPPckt_ls00) );


        pDloadRec->dlrDataPartHandling_ls00.dphRecordNumbers_ls00 = new tsp00_Int4[pDloadRec->dlrMaxRowsPPckt_ls00];
        pDloadRec->dlrDataPartHandling_ls00.dphRecordSpace_ls00   = new tsp00_Int4[pDloadRec->dlrMaxRowsPPckt_ls00];

        pDloadRec->dlrDInfo_ls00.dliNNLongs_ls00 = new tsp00_Int2[pTI->tdFieldCount_ls00];
    }

    return rc;
}
//  ls20_ParseInsertStmt()


/*
  -----------------------------------------------------------------------------
  function:     ls20_InitStructures
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls20_InitStructures(tls00_DBInfo     *pDBInfo,
                    tls00_DatLoadCmd *pDloadCmd,
                    tls00_DloadRec   *pDloadRec,
                    tsp00_Addr        pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls20_InitStructures");
    
    RepServMessages *pMsgObj = RepServMessages::Instance();    // error reporting obj
    tsp00_Int4       rc      = errOK_els00;

    tls00_MultColSpec      &MultCol   = pDloadCmd->dloColSpec_ls00;
    tls00_TableDescription &TableInfo = pDloadRec->dlrTableInfo_ls00;
    tls00_Column           *pCol      = NULL;

    //*
    //*     Assign table name
    //*
    TableInfo.tdTablename_ls00.Init();
    TableInfo.tdTablename_ls00.rawAssign(pDloadCmd->dloTable_ls00.tsTabName_ls00.asCharp());

    //*
    //*     NULL value representation
    //* Check and convert. We assume the repmserver runs on ascii.
    //*
    if (NULL != pDloadCmd->dloTable_ls00.tsCondSpec_ls00)
    {
        rc = ls16ConvertCondition(*pDloadCmd->dloTable_ls00.tsCondSpec_ls00,
                                  pDBInfo->dbiOwnSwap_ls00,
                                  pDloadCmd->dloInfile_ls00.fsFormat_ls00,
                                  pszErrText);
    }




    // Now allocate space and initialize data source structure
    if (errOK_els00 == rc)
    {
        ls20InitDataSourceStruct(&pDloadRec->dlrDataSource_ls00, TableInfo.tdFieldCount_ls00, &MultCol);

        // Fill pattern structure only if the file format is COMPRESSED
        if (true == pDloadCmd->dloInfile_ls00.fsFormat_ls00.ffoCompress_ls00)
        {
            rc = ls18ConstructPattern(&pDloadRec->dlrPattern_ls00,
                                      pDloadCmd->dloInfile_ls00.fsFormat_ls00.ffoSeperat_ls00,
                                      pDloadCmd->dloInfile_ls00.fsFormat_ls00.ffoDelimit_ls00,
                                      pDloadCmd->dloInfile_ls00.fsFormat_ls00.ffoCodeType_ls00,
                                      pszErrText);
        }
    }   // end if (errOK_els00 == rc)






    //*
    //*     Check for column conditions      and
    //*     Set special abstract types       and
    //*     Adapt LONG value specifications  and
    //*     Adapt fraction value: if UNDEFINED_LS00 we take the db value.
    //*
    tsp00_Int4         i   = 0;
    tls00_FieldBasics *pBI = NULL;

    for (i; ( (errOK_els00 == rc) && (i < TableInfo.tdFieldCount_ls00) ); ++i)
    {
        pCol = MultCol.mcsColumn_ls00[pDloadRec->dlrDataSource_ls00.plDataToColIndex_ls00[i]];
        pBI  = &TableInfo.tdParamArray_ls00[i]->piBasicInfo_ls00;
        
        if (NULL != pCol->colNullCond_ls00)
        {
            rc = ls16ConvertCondition(*pCol->colNullCond_ls00,
                                      pDBInfo->dbiOwnSwap_ls00,
                                      pDloadCmd->dloInfile_ls00.fsFormat_ls00,
                                      pszErrText);
        }

        if ( (errOK_els00 == rc) && (UNDEFINED_LS00 == pCol->colFormat_ls00.ffoFrac_ls00) )
        {
            pCol->colFormat_ls00.ffoFrac_ls00 = pBI->fbFraction_ls00;
        }

        //*
        //*     LONG columns
        //*
        if ( (errOK_els00 == rc) && (1 == szMap_LongDataType_ls00[pBI->fbDataType_ls00]) )
        {
            // First of all check for assignments to long fields and - reject processing
            if ( (0 == pCol->colFPos_ls00.fpoStart_ls00) &&
                 (asUnknown_els00 != pCol->colAssign_ls00.fasAsgnTyp_ls00) )
            {
                pMsgObj->ls98Msg(layDataLoad_els98, pszErrText, errLongColIllegal_els98);
                rc = errLongColIllegal_els98;
            }

            // After that fill long column structure
            if (errOK_els00 == rc)
            {
                pCol->colLongInfo.loiIsLong_ls00     = true;

                if (false == pDloadCmd->dloInfile_ls00.fsFormat_ls00.ffoCompact_ls00)
                {
                    pCol->colLongInfo.loiFileStruct_ls00 = new tls00_VFile;
                    memset(pCol->colLongInfo.loiFileStruct_ls00, 0, sizeof(tls00_VFile));
                }
                else
                {
                    pCol->colLongInfo.loiFileStruct_ls00 = pDloadRec->dlrInfile_ls00;
                }

                // Check compatibility of external code type to db data type
                rc = ls08CheckLONG_DT_Compatibility(pCol->colLongInfo.loiFileEncoding_ls00, pBI->fbDataType_ls00);
                if (errOK_els00 != rc)
                {
                    // error: data types do not match (function returns only ok or errIncompatibleDataTypes_els98)
                    pMsgObj->ls98Msg(layDataLoad_els98, pszErrText, errIncompatibleDataTypes_els98,
                                                                    TableInfo.tdColName_ls00[i]->asCharp() );
                }
                ++pDloadRec->dlrLongColCnt_ls00;
            }
        }
    } // end for (i; i < MultCol.mcsCount_ls00; ++i)

    // Initialize vector of long descriptors; used for handling LONGs using putval packets
    if ( (errOK_els00 == rc) && (0 != pDloadRec->dlrLongColCnt_ls00) )
    {
        pDloadRec->dlrpLongDescriptorVector_ls00 = new tsp00_LongDescriptor[pDloadRec->dlrLongColCnt_ls00];
    }

    return rc;
}
// ls20_InitStructures()


/*
  -----------------------------------------------------------------------------
  function:     ls20_InitAndOpenFiles
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls20_InitAndOpenFiles(tls00_DatLoadCmd *pDloadCmd,
                      tls00_DloadRec   *pDLRec,
                      tsp00_Addr        pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls20_InitAndOpenFiles");
    
    RepServMessages *pMsgObj     = RepServMessages::Instance();    // error reporting obj

    tsp00_Int4       rc          = errOK_els00;

    tls00_FileSpec  *pInFileSpec = &pDloadCmd->dloInfile_ls00;      // info from user
    tls00_VFile     *pVFile      = pDLRec->dlrInfile_ls00;       // file struct we work with

    //*
    //*     Init colLongInfo
    //* Assigning LONFFILE Info's from command string
    //*
    ls08InitLongFileInfo(&pDloadCmd->dloColSpec_ls00, &pDloadCmd->dloLongSpec_ls00);

    
    //*
    //* Compute length of record in data file
    //*
    tsp00_Int4 lByteCount = ls18GetRowLength(pDloadCmd->dloColSpec_ls00,
                                             pDLRec->dlrTableInfo_ls00,
                                             pInFileSpec->fsFormat_ls00.ffoBinary_ls00);

    //*
    //*     Initialize file structure of data source we are later on working with
    //*
    tls00_FileFormat *pFFormat = &pInFileSpec->fsFormat_ls00;

    if ( (true == pFFormat->ffoBinary_ls00) || (true == pFFormat->ffoCompact_ls00) )
    {
        ls30VFInitialize(*pVFile, pInFileSpec->fsFileName_ls00);
    }
    else
    {
        ls30VFInitialize(*pVFile, pInFileSpec->fsFileName_ls00, SP5VF_TEXT);
    }

    rc = ls30VFOpen(*pVFile, pszErrText);    // Open file (data source)
    if (errOK_els00 == rc)
    {
        //*
        //*     Check for loading a part of the file only and move to start position
        //*
        if (UNDEFINED_LS00 != pInFileSpec->fsExtract_ls00.feStart_ls00)
        {
            tls00_LargeNumber lFilePos;
            lFilePos.lnu2GB_Blocks_ls00 = 0;

            if ( (false == pFFormat->ffoBinary_ls00) ||     // text file
                 (true  == pFFormat->ffoCompact_ls00) )     // compact file format
            {
                lFilePos.lnuRemainder_ls00 = pInFileSpec->fsExtract_ls00.feStart_ls00;
            }
            else
            {
                lFilePos.lnuRemainder_ls00 = pFFormat->ffoBytesPerLine_ls00 * (pInFileSpec->fsExtract_ls00.feStart_ls00 - 1);
            }
            rc = ls30VFMoveToStartPosition(*pVFile, &lFilePos, pszErrText);
        }



        if ( (errOK_els00 == rc) && (true == pFFormat->ffoCompact_ls00) )
        {
            // In case of file format COMPACT the data file contains a header with
            // meta information.

            tls30_FileMetaDataHeader Header;
            tls00_MetaData           *pMeta = &pDLRec->dlrMetaData_ls00;

            tsp00_Addr *pszCmdString = NULL;
            rc = ls30VFReadMetaHeader(*pVFile, Header, pMeta->pmdLongFieldPos_ls00, pszCmdString, &pMeta->mdSwapKind_ls00, pszErrText);
            if (errOK_els00 == rc)
            {
                pMeta->mdFieldCount_ls00     = Header.mdFieldCount_ls30;
                pMeta->mdLongFieldCount_ls00 = Header.mdLongCount_ls30;

                pMeta->mdLengthArrayLen_ls00 = pMeta->mdFieldCount_ls00 + pMeta->mdLongFieldCount_ls00;
                pMeta->pmdLengthArray_ls00   = new tsp00_Int2[pMeta->mdLengthArrayLen_ls00];
                memset(pMeta->pmdLengthArray_ls00, 0, (sizeof(tsp00_Int2) * pMeta->mdLengthArrayLen_ls00));

                // Assign file pointer
                pMeta->pmdMetaFile_ls00 = pVFile;
            }
        // FOLGENDES ABFANGEN: mehr spalten angegeben als fieldcounts in datei eingetragen
        }
        else if (true == pFFormat->ffoBinary_ls00)
        {
            // Try to read a header; it contains information about the record length
            // of one single record in the file;
            // if the header does not exist we use the earlier calculated record length
            if (0 == pFFormat->ffoBytesPerLine_ls00)       // NOHEADER not found in command
            {
                rc = ls30VFReadHeader(*pVFile, pFFormat->ffoBytesPerLine_ls00, pszErrText);
                if (errOK_els00 != rc)
                {
                    if (errNoDatafileHeader_ls98 == rc)
                    {
                        pFFormat->ffoBytesPerLine_ls00 = lByteCount;
                        pMsgObj->ls98PrintMessage(layDataLoad_els98, errNoDatafileHeader_ls98,
                                                                     pFFormat->ffoBytesPerLine_ls00);
                        rc = errOK_els00;
                    }
                    // if file read error, stop execution;
                    // if only no header log it and compute length
                }
            }
            // else-zweig is done because the right value is set
        }
        else
        {
            pFFormat->ffoBytesPerLine_ls00 = lByteCount;
        }
    }

    return rc;
}
// ls20_InitAndOpenFiles()
