/*! 
  -----------------------------------------------------------------------------
 
  module: vls26.cpp
 
  -----------------------------------------------------------------------------
 
  responsible:  SteffenS

  special area: Replication Server
 
  description:  Implementation of TABLEEXTRACT
                
  version:      7.3.

  -----------------------------------------------------------------------------
 
  copyright:    SAP AG, 1999-2002
 
  -----------------------------------------------------------------------------


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*/

#include "gls00.h"
#include "hls04.h"
#include "hls041.h"
#include "hls05.h"
#include "hls07.h"
#include "hls12.h"
#include "hls13.h"
#include "hls15.h"          // user switch for 'TE all/user'
#include "hls26.h"
#include "hls26_TETable.h"
#include "hls30.h"
#include "hls98msg.h"
#include "hls99.h"

#include "hwd07dbc.h"           // odbc connection
#include "hls31_odbc_Result.h"  // TableExtractDefODBC class

/*  
  -----------------------------------------------------------------------------
  Chapter: Private functions declaration
  -----------------------------------------------------------------------------
*/

/*
  -----------------------------------------------------------------------------
 
  function:     ls26_Constructor
  -----------------------------------------------------------------------------

  description:  Initialize structure used for execution of table extract

  arguments:    DBInfo  [IN]   - structure with all information on the running
                                  session and the db connected to.
                pTECmd  [IN]   - structure with information for te supplied by the user
                pTERec  [OUT]  - tableextract record structure that must be initialized.

  returnvalue:  errOK_els00             - no error
                errNoMoreMemory_els00   - no more memory to allocate structure
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls26_Constructor(tls00_DBInfo *DBInfo, tls00_TabExtrCmd *pTECmd, tls00_TextractRec *&pTERec);


/*
  -----------------------------------------------------------------------------

  function: ls26_Destructor
  -----------------------------------------------------------------------------

  description:  destroys the tableextract structure built in ls26_Constructor

  arguments:    pTECmd  [IN]   - structure with information for te supplied by the user
                pTERec  [IN]   - tableextract record structure that must be destroyed.

  returnvalue:  No
  -----------------------------------------------------------------------------
*/
static void
ls26_Destructor(tls00_TextractRec *pTERec, tls00_TabExtrCmd *pTECmd);


/*
  -----------------------------------------------------------------------------

  function: ls26_FileLTMediumSize
  -----------------------------------------------------------------------------

  description:  Checks if the file length is lower than the user given medium size.
                At the moment the given medium size is defined by the server itself
                and has a value of 1 GB.
                Handles files greater than 2 GB.

  arguments:    lFileLen         [IN] - file length as structure of type tls00_LargeNumber
                lMediumSize      [IN] - medium size as structure of type tls00_LargeNumber
                lAddToFileLen    [IN] - addend that shculd be added to file size

  returnvalue:  true   - file length is lower than medium size
                false  - the opposite
  -----------------------------------------------------------------------------
*/
bool
ls26_FileLenLTMediumSize(tls00_LargeNumber lFileLen,
                         tls00_LargeNumber lMediumSize,
                         tsp00_Longint     lAddToFileLen);     // in Byte ?! assuming that not greater than 2 GB


/*
  -----------------------------------------------------------------------------

  function: ls26_GetTableSize
  -----------------------------------------------------------------------------

  description:  Gets the size of a table from db in blocks of 8kB.
  
  arguments:    pDBInfo   [IN] - structure with all information on the running
                                  session and the db connected to.
                UserName  [IN] - name of owner of table (UTF8)
                TableName [IN] - table name (UTF8)

  returnvalue:  > 0 - size of table in kB
                0   - error while executing
                In case of error the error is only logged to the trace because
                it has no impact on the extraction of this table.
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls26_GetTableSize(tls00_DBInfo            *pDBInfo,
                  SAPDB_DynamicUTF8String &UserName,
                  SAPDB_DynamicUTF8String &TableName);


/*
  -----------------------------------------------------------------------------

  function: ls26_OpenNewOutfile
  -----------------------------------------------------------------------------

  description:  Opens new outfile (data file). By doing that it initializes the
                number of pages in a file to 0 and the number of tables in a file
                to 0. It writes the file header and it writes the file name of the
                data file to the mapping file in case of TE USER or TE ALL.
                
                Behavior:
                If no file is open (beginning of extraction, the first file is
                initialized in ls26_Constructor) the first file will be opened.
                If a file is already open function checks the size of this function
                (gt medium size?) and
                  - file size is gt or et medium size
                    function writes the file trailer and closes the open file
                    function opens a new file (defines the name of the file by use
                    of the file counter that was increased by 1) as described above
                  - file size is lt medium size
                    function does nothing
  

  arguments:    pDBInfo          [IN] - structure with all information on the running
                                         session and the db connected to.
                pTERec           [IN] - tableextract record structure
                bTableIncomplete [IN] - indicates if data file contains incompletely
                                        extracted table data which is logged in the
                                        file trailer
                pszErrText       [IN] - error text

  returnvalue:  errOK_els00                    - no error
                errTEFileCounterExceeded_els98 - file counter exceeded limit of 9999
                err_FileWrite_els98            - error writing to files
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls26_OpenNewOutfile(tls00_DBInfo      *pDBInfo,
                    tls00_TextractRec *pTERec,
                    bool               bTabIncomplete,
                    tsp00_Addr         pszErrText);


/*
  -----------------------------------------------------------------------------

  function: ls26_ExtractLoop
  -----------------------------------------------------------------------------

  description:  Main loop that extracts single tables.

  arguments:    pDBInfo    [IN] - structure with all information on the running
                                   session and the db connected to.
                pTECmd     [IN] - structure with information for te supplied by the user
                pTERec     [IN] - tableextract record structure
                pszErrText [IN] - error text

  returnvalue:  errOK_els00                    - no error
                errors from extracting a single table
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls26_ExtractLoop(tls00_DBInfo      *pDBInfo,
                 tls00_TabExtrCmd  *pTECmd,
                 tls00_TextractRec *pTERec,
                 sapdbwa_DBC       *pODBCConnection,
                 tsp00_Addr         pszErrText);



/*
  -----------------------------------------------------------------------------

  function: ls26_ExtractMultipleTables
  -----------------------------------------------------------------------------

  description:  Main function for extracting tables.

  arguments:    pDBInfo         [IN] - structure with all information on the running
                                        session and the db connected to.
                pTECmd          [IN] - structure with information for te supplied by the user
                pTERec          [IN] - tableextract record structure
                pODBCConnection [IN] - ODBC connection
                pszErrText      [IN] - error text

  returnvalue:  errOK_els00                    - no error
                errors extracting tables
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls26_ExtractMultipleTables(tls00_DBInfo      *pDBInfo,
                           tls00_TabExtrCmd  *pTECmd,
                           tls00_TextractRec *pTERec,
                           sapdbwa_DBC       *pODBCConnection,
                           tsp00_Addr         pszErrText);


/*  
  -----------------------------------------------------------------------------
  EndChapter: Private functions declaration
  -----------------------------------------------------------------------------
*/




/*
  -----------------------------------------------------------------------------
  Chapter: Public functions implementation
  -----------------------------------------------------------------------------
*/

/*
  -----------------------------------------------------------------------------
  function:     ls26ExecTableExtract
  -----------------------------------------------------------------------------
*/
tsp00_Int4
ls26ExecTableExtract(tls00_TabExtrCmd *TECmd, tls00_DBInfo *DBInfo, tsp00_Addr  ErrText)
{
    ROUTINE_DBG_MEO00 ("ls26ExecTableExtract");

    tsp00_Int4 rc  = errOK_els00;

    //*
    //* Additionally to the already existing connection (order interface) we establish
    //* an ODBC connection to get the table names to extract.
    //*
    SAPDB_Bool      rcConnect          = SAPDB_FALSE;
    sapdbwa_DBCPool ODBCConnectionPool;
    sapdbwa_DBC     ODBCConnection;

    if (ODBCConnection.AddAttr( SQL_PRODUCER , sp1pr_internal_cmd ) &&
        ODBCConnection.AddAttr( SQL_COMPONENT, sql_csp_comp_load  )  )
    {
        tsp00_NodeIdc   siDBNodec;
        siDBNodec.p2c(DBInfo->dbiSession_ls00.siDBNode_ls00);

        rcConnect = ODBCConnectionPool.Connect(ODBCConnection,
                                               NULL,
                                               NULL,
                                               siDBNodec,
                                               DBInfo->dbiSession_ls00.siDBName_ls00,
                                               DBInfo->dbiSession_ls00.siUserName_ls00,
                                               DBInfo->dbiSession_ls00.siUserPwd_ls00 ,
                                               NULL,
                                               0,
                                               0,
                                               sapdbwa_DBC_EXCL);
        if (SAPDB_FALSE == rcConnect)
        {
            // Error connecting. Get the error and the error text and print it.
            RepServMessages::Instance()->ls98Msg(layTableExtract_els98, ErrText, errODBCErrors_els98,
                                                                                 ODBCConnection.GetErr().GetErrMsg());
            rc = errODBCErrors_els98;
        }
    }

    //*
    //*     Initialize tableextract record structure
    //*
    tls00_TextractRec *pTERec = NULL;
    if (errOK_els00 == rc)
    {
        rc = ls26_Constructor(DBInfo, TECmd, pTERec);
    }

    if (errOK_els00 == rc)
    {
        rc = ls26_ExtractMultipleTables(DBInfo, TECmd, pTERec, &ODBCConnection, ErrText);
    }

    if (NULL != pTERec)
    {
        ls26_Destructor(pTERec, TECmd);
    }

    return rc;
}
// ls26ExecTableExtract()


/*
  -----------------------------------------------------------------------------
  function: ls26SwitchUser
  -----------------------------------------------------------------------------
*/
tsp00_Int4
ls26SwitchUser(tls00_DBInfo             *pDBInfo,
               SAPDB_DynamicUTF8String   CurrentUser,
               SAPDB_DynamicUTF8String   NewUser,
               tsp00_Addr                pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls26SwitchUser");

    tsp00_Int4 rc = errOK_els00;

    if (CurrentUser.Size() != NewUser.Size())
    {
        rc = ls15SwitchUser(pDBInfo, NewUser, pszErrText);
    }
    else
    {
        if (CurrentUser != NewUser)
        {
            rc = ls15SwitchUser(pDBInfo, NewUser, pszErrText);
        }
    }
    return rc;
}
// ls26SwitchUser()

/*
  -----------------------------------------------------------------------------
  EndChapter: Public functions implementation
  -----------------------------------------------------------------------------
*/





/*
  -----------------------------------------------------------------------------
  Chapter: Private functions implementation
  -----------------------------------------------------------------------------
*/

/*
  -----------------------------------------------------------------------------
  function:     ls26_Constructor  - local function
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls26_Constructor(tls00_DBInfo *DBInfo, tls00_TabExtrCmd *pTECmd, tls00_TextractRec *& pTERec)
{
    ROUTINE_DBG_MEO00 ("ls26_Constructor");

    tsp00_Int4 rc = errOK_els00;

    pTERec = new tls00_TextractRec;
    if (NULL == pTERec)
    {
        rc = errNoMoreMemory_els98;
    }
    else
    {
        memset(pTERec, 0, sizeof(tls00_TextractRec));


        pTERec->terPgPerPckt_ls00                     = ALIGN_EO00(DBInfo->dbiPktSize_ls00, 1024) / DBInfo->dbiPageSize_ls00; 
        pTERec->terPageSize_ls00                      = DBInfo->dbiPageSize_ls00;
        pTERec->terSwapKind_ls00                      = DBInfo->dbiOwnSwap_ls00;          //  Swap behavior
        pTERec->terpPageBuffer_ls00                   = new _TCHAR[DBInfo->dbiPageSize_ls00];
        pTERec->terMediumSize_ls00.lnu2GB_Blocks_ls00 = 0;
        pTERec->terMediumSize_ls00.lnuRemainder_ls00  = MAX_MEDIUM_SIZE_IN_BYTES_LS26;    // 1 GB in Byte so far !
        pTERec->terPageTypeExtracting_ls00            = tptUnknown_els00;


        //create and initialize mapping file and outfile
        if ( (true == pTECmd->texPartSpec_ls00.psSelAll_ls00) ||
             (true == pTECmd->texPartSpec_ls00.psSelUser_ls00) )
        {
            // Here the filenames must newly be created
            // mapping file
            pTERec->terMappingFile_ls00 = new tls00_VFile;
            
            pTERec->terMappingFile_ls00->vfFileName_ls00 = new _TCHAR[PATH_MXSP00];
            pTERec->terMappingFile_ls00->vfFileName_ls00[0] = 0;
            sprintf(pTERec->terMappingFile_ls00->vfFileName_ls00, "%s%0*d",
                                                                  pTECmd->texOutfile_ls00.fsFileName_ls00,
                                                                  FILE_EXTENSION_SIZE_LS26,
                                                                  pTERec->terFileCounter_ls00);
            ls30VFInitialize(*pTERec->terMappingFile_ls00,
                             pTERec->terMappingFile_ls00->vfFileName_ls00,
                             SP5VF_TEXT,
                             SP5VF_WRITE);
        
            ++pTERec->terFileCounter_ls00;

            // outfile
            pTERec->terOutFile_ls00 = new tls00_VFile;

            pTERec->terOutFile_ls00->vfFileName_ls00 = new _TCHAR[PATH_MXSP00];
            pTERec->terOutFile_ls00->vfFileName_ls00[0] = 0;
            sprintf(pTERec->terOutFile_ls00->vfFileName_ls00, "%s%0*d",
                                                              pTECmd->texOutfile_ls00.fsFileName_ls00,
                                                              FILE_EXTENSION_SIZE_LS26,
                                                              pTERec->terFileCounter_ls00);
            ls30VFInitialize(*pTERec->terOutFile_ls00,
                             pTERec->terOutFile_ls00->vfFileName_ls00,
                             SP5VF_BINARY,
                             SP5VF_WRITE);
        }
        else
        {
            // Here the filename is simply a pointer to the user given file name in TE cmd

            // outfile
            pTERec->terOutFile_ls00 = new tls00_VFile;

            ls30VFInitialize(*pTERec->terOutFile_ls00,
                             pTECmd->texOutfile_ls00.fsFileName_ls00,
                             SP5VF_BINARY,
                             SP5VF_WRITE);
        }
    }   // end if (NULL == pTERec)
    
    return rc;
}
//  ls26_Constructor()


/*
  -----------------------------------------------------------------------------
  function: ls26_Destructor
  -----------------------------------------------------------------------------
*/
static void
ls26_Destructor(tls00_TextractRec *pTERec, tls00_TabExtrCmd *pTECmd)
{
    ROUTINE_DBG_MEO00 ("ls26_Destructor");

    if (NULL != pTERec->terpPageBuffer_ls00)
    {
        delete [] pTERec->terpPageBuffer_ls00;
    }
    
    if (NULL != pTERec->terMappingFile_ls00)
    {
        if (0 != pTERec->terMappingFile_ls00->vfFileHndl_ls00)      // close last data file if still open
        {
            _TCHAR szErr[256];
            ls30VFClose(*pTERec->terMappingFile_ls00, szErr);
        }

        if (NULL != pTERec->terMappingFile_ls00->vfFileName_ls00)
        {
            delete pTERec->terMappingFile_ls00->vfFileName_ls00;
        }

        delete pTERec->terMappingFile_ls00;
    }


    if (NULL != pTERec->terOutFile_ls00)
    {
        if (0 != pTERec->terOutFile_ls00->vfFileHndl_ls00)      // close last data file if still open
        {
            _TCHAR szErr[256];
            ls30VFClose(*pTERec->terOutFile_ls00, szErr);
        }

        if (pTERec->terOutFile_ls00->vfFileName_ls00 != pTECmd->texOutfile_ls00.fsFileName_ls00)
        {
            delete [] pTERec->terOutFile_ls00->vfFileName_ls00;
            pTERec->terOutFile_ls00->vfFileName_ls00 = NULL;
        }

        delete pTERec->terOutFile_ls00;
    }

    delete pTERec;     // at last delete the structure itself

    return;
}
// ls26_Destructor()


/*
  -----------------------------------------------------------------------------
  function: ls26_FileLTMediumSize
  -----------------------------------------------------------------------------
*/
bool
ls26_FileLenLTMediumSize(tls00_LargeNumber lFileLen,
                         tls00_LargeNumber lMediumSize,
                         tsp00_Longint     lAddToFileLen)      // in Byte ?! assuming that not greater than 2 GB
{
    bool brc = true;

    if (lFileLen.lnuRemainder_ls00 > (MAX_INT4_SP00 - lAddToFileLen) )    // do we exceed the limit of 2 GB when adding the value to len?
    {
        ++lFileLen.lnu2GB_Blocks_ls00;
        lFileLen.lnuRemainder_ls00 = -(MAX_INT4_SP00 - lFileLen.lnuRemainder_ls00 - lAddToFileLen);
    }

    if (lFileLen.lnu2GB_Blocks_ls00 == lMediumSize.lnu2GB_Blocks_ls00)
    {
        if ( (lFileLen.lnuRemainder_ls00 + lAddToFileLen) >= lMediumSize.lnuRemainder_ls00)
        {
            brc = false;
        }
    }
    else if (lFileLen.lnu2GB_Blocks_ls00 > lMediumSize.lnu2GB_Blocks_ls00)
    {
        brc = false;
    }

    return brc;
}
// ls26_FileLTMediumSize()


/*
  -----------------------------------------------------------------------------
  function: ls26_GetTableSize
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls26_GetTableSize(tls00_DBInfo            *pDBInfo,
                  SAPDB_DynamicUTF8String &UserName,
                  SAPDB_DynamicUTF8String &TableName)
{
    ROUTINE_DBG_MEO00 ("ls26_GetTableSize");

    tsp00_Int4 rc      = errOK_els00;
    tsp00_Int4 lResult = 0;

    _TCHAR szCmd[1024];
    szCmd[0] = 0;

    // this is a view wich selects rights for everyone
    sp77sprintf(szCmd, 1024, GET_TABLE_SIZE_STMT_LS26, UserName.StrPtr(), TableName.StrPtr());

    tsp00_Int4 lErrPos;
    _TCHAR     szErr[MAX_REPLY_LENGTH_LS00];
    rc = ls13ExecIntSQLStmtUTF8(pDBInfo, 0, szCmd, szErr, lErrPos);
    if (errOK_els00 != rc)
    {
        TRACE_PRNF_MLS99(("ls26_GetTableSize", "Error getting table size for %s (%s)\n",
                                               TableName.StrPtr(), szErr));
    }
    else
    {
        tsp00_Int4 tmpRowCnt = 1;
        rc = ls12Fetch(*pDBInfo, NULL, tmpRowCnt, false, szErr);
        if (errOK_els00 == rc)
        {
            tsp1_packet *pSqlPacket = pDBInfo->dbiPktRcvList_ls00[0];
            tsp1_part   *pSqlPart = NULL;

            bool Res = ls04GetPart(pSqlPacket, sp1pk_data, pSqlPart);
            if (false == Res)
            {
                TRACE_PRNF_MLS99(("ls26_GetTableSize", "Error getting table size for %s (error no: %d)\n",
                                                       TableName.StrPtr(), rc));
            }
            else
            {
                // we expect the result with a length of 8 byte (fixed number (10,0))
                _TCHAR szResult[8];

                tsp00_Int4 PartLength = ls04GetPartLength(pSqlPart);
            
                ls04BufFromPartEx(pSqlPart, 0, PartLength, szResult, PartLength);

                rc = ls07NumberToInt(szResult + 1, lResult);                        // szResult: jump over def byte
                if (errOK_els00 == rc)
                {
                    lResult *= (pDBInfo->dbiPageSize_ls00/1024);
                }
            }
        }
    }

    return lResult;
}
// ls26_GetTableSize()


/*
  -----------------------------------------------------------------------------
  function: ls26_OpenNewOutfile
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls26_OpenNewOutfile(tls00_DBInfo      *pDBInfo,
                    tls00_TextractRec *pTERec,
                    bool               bTabIncomplete,
                    tsp00_Addr         pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls26_OpenNewOutfile");

    tsp00_Int4 rc = errOK_els00;
    tls00_VFile *pOutFile = pTERec->terOutFile_ls00;

    if (0 != pOutFile->vfFileHndl_ls00)     // is file already open?
    {
        // WRITE DATA FILE TRAILER FOR OLD FILE
        _TCHAR szFileName[PATH_MXSP00];
        szFileName[0] = 0;
        sprintf(szFileName, "%s%0*d",
                            &pOutFile->vfFileName_ls00[_tcslen(pOutFile->vfFileName_ls00) - FILE_EXTENSION_SIZE_LS26],
                            FILE_EXTENSION_SIZE_LS26,
                            pTERec->terFileCounter_ls00);
        ls26WriteFileTrailer(pDBInfo, pTERec, bTabIncomplete, szFileName, pszErrText);

        ls30VFClose(*pTERec->terOutFile_ls00, pszErrText);    // close it; return value is of no interest here

        if (pTERec->terFileCounter_ls00 < FILECOUNT_MXLS26)
        {
            // generate new file
            ++pTERec->terFileCounter_ls00;
            sprintf(&pOutFile->vfFileName_ls00[_tcslen(pOutFile->vfFileName_ls00) - FILE_EXTENSION_SIZE_LS26],
                    "%0*d",
                    FILE_EXTENSION_SIZE_LS26,
                    pTERec->terFileCounter_ls00);
            ls30VFInitialize(*pOutFile, pOutFile->vfFileName_ls00, SP5VF_BINARY, SP5VF_WRITE);
        }
        else
        {
            RepServMessages::Instance()->ls98Msg(layTableExtract_els98, pszErrText, errTEFileCounterExceeded_els98,
                                                                                    FILECOUNT_MXLS26);
            rc = errTEFileCounterExceeded_els98;
        }
    }

    if ( (errOK_els00 == rc) && (0 == pOutFile->vfFileHndl_ls00) )      // check again if file is already open
    {
        rc = ls30VFOpen(*pOutFile, pszErrText);    // open it

        // Write file name to mapping file only if not extracting a single table 
        if (NULL != pTERec->terMappingFile_ls00)
        {
            _TCHAR szFileNameEntry[PATH_MXSP00 + PATH_MXSP00];
            szFileNameEntry[0] = 0;

            _TCHAR* pszDest = strrchr(pOutFile->vfFileName_ls00, SLASH_LS00);
            if (NULL == pszDest)
            {
                pszDest = strrchr(pOutFile->vfFileName_ls00, BACKSLASH_LS00);
                if (NULL == pszDest)
                {
                    pszDest = pOutFile->vfFileName_ls00;
                }
            }
            // if pszDest does NOT point to the file name itself but to a part of it
            // we need to add 1 to the pointer
            pszDest = (pszDest == pOutFile->vfFileName_ls00) ? pszDest : pszDest + 1;

            if (1 == pTERec->terFileCounter_ls00)
            {
                _stprintf(szFileNameEntry, "[%s]", pszDest);
            }
            else
            {
                _stprintf(szFileNameEntry, NEW_LINE"[%s]", pszDest);
            }

            rc = ls30VFWriteLn(*pTERec->terMappingFile_ls00, szFileNameEntry, _tcslen(szFileNameEntry), pszErrText);
        }

        // WRITE DATA FILE HEADER PAGE!
        bool bMoreThan1Table = false;
        if (NULL != pTERec->terMappingFile_ls00)        // not a single table but TE ALL | USER
        {
            bMoreThan1Table = true;
        }
        rc = ls26WriteHeaderPage(pDBInfo, pTERec, bMoreThan1Table, pszErrText);

        // Reinitialize page count and number of tables in file
        pTERec->terPageCount_ls00    = 0;
        pTERec->terNumTabInFile_ls00 = 0;
            
    }

    return rc;
}
// ls26_OpenNewOutfile()


/*
  -----------------------------------------------------------------------------
  function: ls26_ExtractLoop
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls26_ExtractLoop(tls00_DBInfo      *pDBInfo,
                 tls00_TabExtrCmd  *pTECmd,
                 tls00_TextractRec *pTERec,
                 sapdbwa_DBC       *pODBCConnection,
                 tsp00_Addr         pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls26_TableExtractLoop");

    tsp00_Int4 rc = errOK_els00;

    tsp00_Int4 lTabSize = 0;

    // Position used to mark begin and end position of data in data file
    tls00_LargeNumber lBeginPos;
    lBeginPos.lnu2GB_Blocks_ls00 = 0;
    lBeginPos.lnuRemainder_ls00  = 0;

    // Output data file
    tls00_VFile *pOutFile = pTERec->terOutFile_ls00;

    //*
    //*     Define current connected user for extracting data of single tables
    //*
    SAPDB_DynamicUTF8String CurrentUser;
    
    SAPDB_DynamicUTF8String SessionUser;
    rc = ls05ConvertToUTF8(SessionUser,
                           REINTERPRET_CAST(unsigned char*, pDBInfo->dbiSession_ls00.siUserName_ls00.asCharp()),
                           _tcslen(pDBInfo->dbiSession_ls00.siUserName_ls00.asCharp()),
                           ctAscii_els00, pszErrText);
    CurrentUser = SessionUser;

    // Define user and table names as UTF8 strings
    SAPDB_DynamicUTF8String UserName;
    SAPDB_DynamicUTF8String TableName;

    // Return value used with the TableExtractDefODBC class
    SAPDB_Bool rcBOOL = SAPDB_FALSE;


    //*
    //*     Construct TableExtractDefODBC object
    //*
    TableExtractDefODBC *pTEObject = NULL;
    pTEObject = new TableExtractDefODBC(*pODBCConnection, &pTECmd->texPartSpec_ls00);

    //*
    //*     Initialize TableExtractDefODBC object
    //*
    rcBOOL = pTEObject->initResultSet(*pODBCConnection);

    //*
    //*     Main extract loop
    //*
    while ( (SAPDB_TRUE == rcBOOL) && (errOK_els00 == rc) )
    {
//TOTOTODO Was passiert, wenn ein fehler beim extrahieren einer tabelle auftritt? Mit den anderen weitermachen
//         und eine meldung ausgeben? Oder abbrechen?
//TODO AUF FEHLER NICHT REAGIEREN, ABER EINEN SCHWELLENWERT EINBAUEN (ZB 30), DH BEI AUFEINDERFOLGEN VON
//     (SCHWELLENWERT) FEHLERN WIRD ABGEBROCHEN, DA VON EINEM SCHWERWIEGENDEN FEHLER AUSGEGANGEN WERDEN KANN

        //*
        //*     Get user name and table name for table to extract.
        //*
        rcBOOL = pTEObject->GetOwnerTablePair(*pODBCConnection, &UserName, &TableName);
        if (SAPDB_TRUE == rcBOOL)
        {
            //*
            //*     Extract single table
            //*
            rc       = ls26SwitchUser(pDBInfo, CurrentUser, UserName, pszErrText);
            lTabSize = ls26_GetTableSize(pDBInfo, UserName, TableName);

            if (false == ls26_FileLenLTMediumSize(pOutFile->vfFileLen_ls00,
                                                  pTERec->terMediumSize_ls00,   // terMediumSize_ls00 comes in Bytes
                                                  lTabSize * 1024))             // lTabSize comes in kB
            {
                rc = ls26_OpenNewOutfile(pDBInfo, pTERec, false, pszErrText);
            }

            if (errOK_els00 == rc)
            {
                lBeginPos.lnuRemainder_ls00  = pOutFile->vfFileLen_ls00.lnuRemainder_ls00;
                lBeginPos.lnu2GB_Blocks_ls00 = pOutFile->vfFileLen_ls00.lnu2GB_Blocks_ls00;

                rc = ls26ExtractSingleTable(pDBInfo, TableName, pTERec, pszErrText);
                if (errOK_els00 == rc)
                {
                    //*
                    //*     Write entry to mapping file in UTF8 encoding
                    //*
                    if (NULL != pTERec->terMappingFile_ls00)
                    {
                        _TCHAR fileentry[256];
                        fileentry[0] = 0;
                        sp77sprintf(fileentry, 256, "\"%s\".\"%s\" %d:%d-%d:%d",
                                                    UserName.StrPtr(), TableName.StrPtr(),
                                                    lBeginPos.lnu2GB_Blocks_ls00, lBeginPos.lnuRemainder_ls00,
                                                    pOutFile->vfFileLen_ls00.lnu2GB_Blocks_ls00,
                                                    pOutFile->vfFileLen_ls00.lnuRemainder_ls00);

                        rc = ls30VFWriteLn(*pTERec->terMappingFile_ls00, fileentry, _tcslen(fileentry), pszErrText);
                    }

                    ++pTERec->terNumTabExtracted_ls00;
                    ++pTERec->terNumTabInFile_ls00;
                }
                else
                {
                    ++pTERec->terNumTabExtractFailed_ls00;

                    // In case of TE USER|ALL we may continue with extracting tables,
                    // in case of TE TABLE we'll stop on error
                    if ( (true == pTECmd->texPartSpec_ls00.psSelAll_ls00) || (true == pTECmd->texPartSpec_ls00.psSelUser_ls00) )
                    {
                        pszErrText[0] = 0;
                        rc = errOK_els00;
                    }
                }

    //TODO bei einem fehler muss hier eine besondere nachricht ausgegeben werden, aber wir brechen noch nicht ab.
    //     die anzahl der nicht extrahierten tabellen sollte am ende zusammengefasst ausgegeben werden
    // was, wenn beim extrahieren bei den longs abgebrochen wird aber schon die anderen daten (davor) dieser tabelle
    // extrahiert wurden: dann muss auf die letzte position vor dieser tabelle in der datei zurueckgesetzt werden
            }   // end if (errOK_els00 == rc)

            // Assign new user name
            CurrentUser = UserName;

            pTERec->terConstraintCnt_ls00 = 0;
            pTERec->terDefaultCnt_ls00    = 0;
            pTERec->terLongColCnt_ls00    = 0;

        }   // end if (SAPDB_TRUE == rcBOOL)
    }   // end while ( (SAPDB_TRUE == rcBOOL) && (errOK_els00 == rc) )

    // switch user back to session user
    char szErr[MAX_REPLY_LENGTH_LS00];
    tsp00_Int4 rc2 = ls26SwitchUser(pDBInfo, CurrentUser, SessionUser, szErr);

    return rc;
}
// ls26_TableExtractLoop()


/*
  -----------------------------------------------------------------------------
  function: ls26_ExtractMultipleTables
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls26_ExtractMultipleTables(tls00_DBInfo      *pDBInfo,
                           tls00_TabExtrCmd  *pTECmd,
                           tls00_TextractRec *pTERec,
                           sapdbwa_DBC       *pODBCConnection,
                           tsp00_Addr         pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls26_ExtractMultipleTables");

    RepServMessages *pMsgObj = RepServMessages::Instance();     // error msg object
    tsp00_Int4       rc      = errOK_els00;

    //*
    //*     Check if user is sysdba (in case of TABLEEXTRACT ALL)
    //*
    if (true == pTECmd->texPartSpec_ls00.psSelAll_ls00)
    {
        rc = ls15CheckUserMode(pDBInfo,
                               pDBInfo->dbiSession_ls00.siUserName_ls00.asCharp(),
                               dbuSysdba_els00,
                               pszErrText);
        if ( (errOK_els00 != rc) && (dbRowNotFound_els00 == rc) )
        {
            pMsgObj->ls98Msg(layTableExtract_els98, pszErrText, errRestrictedStmt_els98);
        }
    }

    //*
    //*     Open mapping file (has been initialized in initialization routine)
    //* Only necessary for TABLEXTRACT ALL/USER
    //*
    if ( (errOK_els00 == rc) &&
         ( (true == pTECmd->texPartSpec_ls00.psSelAll_ls00) || (true == pTECmd->texPartSpec_ls00.psSelUser_ls00) ) )
    {
        tsp00_Int4 rc2 = ls30VFOpen(*pTERec->terMappingFile_ls00, pszErrText);
//TODOTODO  fehler ist an dieser stelle nicht schlimm: es sollte lediglich eine nachricht ausgegeben werden
//          dass diese datei nicht angelegt werden konnte
    }


    //*
    //*     Main loop to extract tables
    //*
    if (errOK_els00 == rc)
    {
        rc = ls26_OpenNewOutfile(pDBInfo, pTERec, false, pszErrText);
        if (errOK_els00 == rc)
        {
            rc = ls26_ExtractLoop(pDBInfo, pTECmd, pTERec, pODBCConnection, pszErrText);

            //*
            //*     Write trailer to last file only if no error occurred
            //* This may only happen in case of TABLEEXTRACT TABLE; TE ALL|USER will
            //* always return errOK_els00; in the latter case however the value
            //* pTERec->terNumTabExtractFailed_ls00 may have been set indicating errors
            //* at one or more table while extracting
            //*
            if (errOK_els00 == rc)
            {
                if (0 != pTERec->terOutFile_ls00->vfFileHndl_ls00)
                {
                    ls26WriteFileTrailer(pDBInfo, pTERec, false, NULL, pszErrText);
                }
            }
            else
            {
                // In case we have no tables extracted but have a file opened - we remove it here
                if (0 == pTERec->terNumTabExtracted_ls00)
                {
                    char szText[MAX_REPLY_LENGTH_LS00];
                    szText[0] = 0;
                    ls30VFErase(*pTERec->terOutFile_ls00, szText);
                }
            }

            if (0 != pTERec->terNumTabExtractFailed_ls00)
            {
                char szFileName[MAX_REPLY_LENGTH_LS00];
                szFileName[0] = 0;

                pMsgObj->ls98GetLogFileName(szFileName);
                pMsgObj->ls98MsgToString(layTableExtract_els98, pszErrText, errCmdUnsuccessful_els98,
                                                                            szFileName);
                rc = errCmdUnsuccessful_els98;
            }

            pMsgObj->ls98PrintMessage(layTableExtract_els98, msgExtractedFailedTables_els98,
                                                             pTERec->terNumTabExtracted_ls00,
                                                             pTERec->terNumTabExtractFailed_ls00);
            pMsgObj->ls98PrintMessage(layTableExtract_els98, msgTESuccessful_els98);
        }
    }

    return rc;
}
// ls26_ExtractMultipleTables()

/*
  -----------------------------------------------------------------------------
  EndChapter: Private functions implementation
  -----------------------------------------------------------------------------
*/