/*! 
  -----------------------------------------------------------------------------
 
  module: vls26_TETable.cpp
 
  -----------------------------------------------------------------------------
 
  responsible:  SteffenS

  special area: Replication Server
 
  description:  Implementation of TABLEEXTRACT for a single table
                
  version:      7.3.

  -----------------------------------------------------------------------------
 
  copyright:    SAP AG, 2000-2002
 
  -----------------------------------------------------------------------------


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*/
#include <time.h>

#include "gls00.h"
#include "hls26_TETable.h"
#include "hls04.h"
#include "hls03.h"
#include "hls05.h"
#include "hls07.h"
#include "hls13.h"
#include "hls15.h"
#include "hls30.h"
#include "hls98msg.h"
#include "hls99.h"
#include "gbd00.h"

#include "hsp100.h"     // versionsbestimmung


/*  
  -----------------------------------------------------------------------------
  Chapter: Private functions declaration
  -----------------------------------------------------------------------------
*/

/*
  -----------------------------------------------------------------------------
 
  function:     ls26_WriteConstraintInfo
  -----------------------------------------------------------------------------

  description:  Gets constraint descriptions from kernel (describe constraint <table name>)
                and logs them on pages into data file.

  attention:    Prerequisite is that single constraint descriptions may not exceed a
                length of 8kB.

  arguments:    pDBInfo     [IN]   - contains all information on the running
                                       session and the database connected to
                TableName   [IN]   - table name UTF8 encoded
                pTERec      [IN]   - structure holding information for
                                      executing TABLEEXTRACT for a single table
                pszErrText  [OUT]  - error text

  returnvalue:  errOK_els00        - no error
                errFileWrite_els98 - error writing data
                errInternal_els98  - Unknown part data when retrieving constraints from
                                      kernel
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls26_WriteConstraintInfo(tls00_DBInfo            *pDBInfo,
                         SAPDB_DynamicUTF8String  TableName,
                         tls00_TextractRec       *pTERec,
                         tsp00_Addr               pszErrText);


/*
  -----------------------------------------------------------------------------
 
  function:     ls26_WriteDefaultInfo
  -----------------------------------------------------------------------------

  description:  Gets default descriptions from kernel (describe default <table name>)
                and logs them on pages into data file

  arguments:    DBInfo     [IN]   - contains all information on the running
                                     session and the database connected to
                TableName  [IN]   - table name UTF8 encoded
                TERec      [IN]   - structure holding information for
                                     executing TABLEEXTRACT for a single table
                pszErrText [OUT]  - error text

  returnvalue:  errOK_els00        - no error
                errFileWrite_els98 - error writing data
                errInternal_els98  - Unknown part data when retrieving defaults from
                                      kernel
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls26_WriteDefaultInfo(tls00_DBInfo            *DBInfo,
                      SAPDB_DynamicUTF8String  TableName,
                      tls00_TextractRec       *TERec,
                      tsp00_Addr               pszErrText);


/*
  -----------------------------------------------------------------------------
 
  function:     ls26_PageExtractLoop
  -----------------------------------------------------------------------------

  description:  Gets data pages from kernel and saves them in data file.
                Main routine to extract data.

  arguments:    pDBInfo     [IN]   - Contains all information on the running
                                      session and the database connected to
                pTERec      [IN]   - Structure holding information for
                                      executing TABLEEXTRACT for a single table
                pszErrText  [OUT]  - Error text

  returnvalue:  errOK_els00        - No error
                Errors of called functions are forwarded
                --> errInternal_els98, errFileWrite_els98, errFinishPart_els98,
                    errDbComm_els98
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls26_PageExtractLoop(tls00_DBInfo      *pDBInfo,
                     tls00_TextractRec *pTERec,
                     tsp00_Addr         pszErrText);

/*
  -----------------------------------------------------------------------------
 
  function:     ls26_Analyse
  -----------------------------------------------------------------------------

  description:  Extrahiert die datenseiten aus einem part des paketes. Passt dabei
                die Seitennummer und den seitentyp an.
                Bestimmt beim ersten aufruf fuer einen seitentyp die maximal
                moegliche anzahl von seiten im paket. Wird diese dann spaeter bei
                der bearbeitung der daten fuer diesen seitentyp unterschritten, ist
                das das signal, dass das letzte paket bearbeitet wird.

  arguments:    tsp1_packet_ptr [IN]   - zeiger; zu bearbeitendes paket
                pTERec          [IN]   - zeiger; Struktur mit daten fuer tableextract
                pszErrText      [OUT]  - Fehlertext

  returnvalue:  errOK_els00            - everything went fine
                STOP_ANALYSE_LS26      - ende der daten eines seitentyps erreicht
                errInternal_els98      - interner fehler: enthaelt eigentlichen fehlercode
                                         und zeile sowie dateinamen
                errFileWrite_els98     - fehler beim schreiben der daten in extraktionsdatei

                Die funktion vermerkt den fehler bereits im logfile und im
                rueckgabestring.
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls26_Analyse(tsp1_packet_ptr    pSqlPacket,
             tls00_TextractRec *pTERec,
             tsp00_Addr         pszErrText);


/*
  -----------------------------------------------------------------------------
 
  function:     ls26_SendBdinfoPckt
  -----------------------------------------------------------------------------

  description:  Sendet ein bdinfo an den kern

  arguments:    pDBInfo         [IN]   - zeiger; contains all 
                                          information on the running session and the
                                          database connected to.
                lPcktNo         [IN]   - Nummer des zu nutzenden paketes (0 oder 1)
                pszRawChars     [IN]   - zeiger; zeigt auf zu sendende daten (z.b.
                                          bdinfo)
                lRawCharsLength [IN]   - laenge der zu sendenden daten
                pszErrText      [OUT]  - Fehlertext

  returnvalue:  errOK_els00            - everything went fine
                errFinishPart_els98    - der zu sendende part konnte nicht erstellt werden
                errDbComm_els98        - fehler beim senden des paketes

                Die funktion vermerkt den fehler bereits im logfile und im
                rueckgabestring.
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls26_SendBdinfoPckt(tls00_DBInfo *pDBInfo,
                    tsp00_Int4    lPcktNo,
                    tsp00_Addr    pszRawChars,
                    tsp00_Int4    lRawCharsLength,
                    tsp00_Addr    pszErrText);

/*
  -----------------------------------------------------------------------------
 
  function:     ls26_WriteShortStringPages
  -----------------------------------------------------------------------------

  description:  Extracts short string data pages.
                Runs command 'BEGIN SAVE LONG <table name>' and receives
                initial bdinfo. This bdinfo is a parameter to function
                ls26_PageExtractLoop.
                Last mentioned function finally gets the pages.

  arguments:    pDBInfo     [IN]   - contains all information on the running
                                      session and the database connected to
                TableName   [IN]   - table name UTF8 encoded
                pTERec      [IN]   - structure holding information for
                                      executing TABLEEXTRACT for a single table
                pszErrText  [OUT]  - error text

  returnvalue:  errOK_els00          - no error
                errInternSQL_els98   - command 'BEGIN...' failed
                errInternal_els98    - unknown part
                
                error of called functions are forwarded
                --> errInternal_els98, errFileWrite_els98, errFinishPart_els98,
                    errDbComm_els98
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls26_WriteShortStringPages(tls00_DBInfo            *pDBInfo,
                           SAPDB_DynamicUTF8String  TableName,
                           tls00_TextractRec       *pTERec,
                           tsp00_Addr               pszErrText);

/*
  -----------------------------------------------------------------------------
 
  function:     ls26_WriteLongStringPages
  -----------------------------------------------------------------------------

  description:  Extracts long string data pages.
                Runs command 'BEGIN SAVE COLUMN' and receives
                initial bdinfo. This bdinfo is a parameter to function
                ls26_PageExtractLoop.
                Last mentioned function finally gets the pages.
  
  arguments:    pDBInfo     [IN]   - contains all information on the running
                                      session and the database connected to
                pTERec      [IN]   - structure holding information for
                                      executing TABLEEXTRACT for a single table
                pszErrText  [OUT]  - error text

  returnvalue:  errOK_els00          - no error
                errInternSQL_els98   - command 'BEGIN...' failed
                errInternal_els98    - unknown part
                
                error of called functions are forwarded
                --> errInternal_els98, errFileWrite_els98, errFinishPart_els98,
                    errDbComm_els98
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls26_WriteLongStringPages(tls00_DBInfo      *pDBInfo,
                          tls00_TextractRec *pTERec,
                          tsp00_Addr         pszErrText);


/*
  -----------------------------------------------------------------------------
 
  function:     ls26_WritePrimaryDataPages
  -----------------------------------------------------------------------------

  description:  Extracts primary data pages.
                Runs command 'BEGIN SAVE TABLE <table name>' and receives
                initial bdinfo. This bdinfo is a parameter to function
                ls26_PageExtractLoop.
                Last mentioned function finally gets the pages.

  arguments:    pDBInfo     [IN]   - contains all information on the running
                                      session and the database connected to
                TableName   [IN]   - table name UTF8 encoded
                pTERec      [IN]   - structure holding information for
                                      executing TABLEEXTRACT for a single table
                pszErrText  [OUT]  - error text

  returnvalue:  errOK_els00          - no error
                errInternSQL_els98   - command 'BEGIN...' failed
                errInternal_els98    - unknown part
                
                error of called functions are forwarded
                --> errInternal_els98, errFileWrite_els98, errFinishPart_els98,
                    errDbComm_els98
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls26_WritePrimaryDataPages(tls00_DBInfo            *pDBInfo,
                           SAPDB_DynamicUTF8String  TableName,
                           tls00_TextractRec       *pTERec,
                           tsp00_Addr               pszErrText);



/*
  -----------------------------------------------------------------------------
 
  function:     ls26_WDataHeaderPage
  -----------------------------------------------------------------------------

  description:  Writes special header page for data pages of type short string
                and primary data.
                The page contains the structure tls26_DescPageHeader at the
                beginning and at the end.
                Header structure tls26_DescPageHeader at those header pages
                contains only values for
                dphPageNo_ls26, dphType_ls26 and dphPagType_ls26.

  arguments:    pTERec        [IN]   - pointer; structure containing info for
                                        tableextract
                pszErrText    [OUT]  - error text

  returnvalue:  errOK_els00          - No error
                errFileWrite_els98   - error writing page to file
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls26_WDataHeaderPage(tls00_TextractRec *pTERec, tsp00_Addr pszErrText);



/*
  -----------------------------------------------------------------------------
 
  function:     ls26_WLongDataHeaderPage
  -----------------------------------------------------------------------------

  description:  Writes special header page for data pages of type long string
  .
                The page contains the structure tls26_DescPageHeader at the
                beginning and at the end.
                Header structure tls26_DescPageHeader contains only values for
                dphPageNo_ls26, dphType_ls26, dphPagType_ls26 and dphSpecialID_ls26.
                The latter contains the key length.

                Additionally to the header the header page contains the key of the
                record for which the long column was extracted for.

  arguments:    pTERec        [IN]   - pointer; structure containing info for
                                        tableextract
                lColumnNo     [IN]   - 2 byte integer, number of column in single
                                        record extracted
                lRecordKeyLen [IN]   - length of key of single record
                pszRecordKey  [IN]   - key of single record
                pszErrText    [OUT]  - error text

  returnvalue:  errOK_els00          - No error
                errFileWrite_els98   - error writing page to file
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls26_WLongDataHeaderPage(tls00_TextractRec *pTERec,
                         tsp00_Int2         lColumnNo,
                         tsp00_Int4         lRecordKeyLen,         // nur fuer long-werte von interesse, fuer alle anderen aufrufer auf 0 setzen
                         tsp00_Addr         pszRecordKey,
                         tsp00_Addr         pszErrText);


/*
  -----------------------------------------------------------------------------
  function:     ls26_SaveIndexDescription
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls26_SaveIndexDescription(tls00_DBInfo            *pDBInfo,
                          tls00_TextractRec       *pTERec,
                          SAPDB_DynamicUTF8String  TableName,
                          tsp00_Addr               pszErrText);


/*
  -----------------------------------------------------------------------------
  function:     ls26_SaveTableDescription
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls26_SaveTableDescription(tls00_DBInfo            *pDBInfo,
                          tls00_TextractRec       *pTERec,
                          SAPDB_DynamicUTF8String  TableName,
                          tsp00_Addr              &pszTableDesc,
                          tsp00_Addr               pszErrText);

/*  
  -----------------------------------------------------------------------------
  EndChapter: Private functions declaration
  -----------------------------------------------------------------------------
*/





/*
  -----------------------------------------------------------------------------
  Chapter: Public functions implementation
  -----------------------------------------------------------------------------
*/


/*
  -----------------------------------------------------------------------------
  function:     ls26ExtractSingleTable
  -----------------------------------------------------------------------------
*/
tsp00_Int4
ls26ExtractSingleTable(tls00_DBInfo            *DBInfo,
                       SAPDB_DynamicUTF8String &TableName,
                       tls00_TextractRec       *TERec,
                       tsp00_Addr               pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls26ExtractSingleTable");

    tsp00_Int4 rc = errOK_els00;
    
    bool bShortColumnFileAvailable = false;

    //*
    //*     Get table description
    //*
    tsp00_Addr pszTabDesc = NULL;
    if (errOK_els00 == rc)
    {
        rc = ls26_SaveTableDescription(DBInfo, TERec, TableName, pszTabDesc, pszErrText);
    }
    if (errOK_els00 == rc)
    {
        tsp7_table_description *TabDescStruct_sp07 = REINTERPRET_CAST(tsp7_table_description*, pszTabDesc);

        TERec->terConstraintCnt_ls00 = TabDescStruct_sp07->sp7t_constraint_cnt();

        int i = 0;
        int j = 0;

        tsp7_ffield_info *pFieldInfo = NULL;
        TERec->terDefaultArray_ls00 = new tsp00_Int2[TabDescStruct_sp07->sp7t_field_cnt()];

        for (i, j; i < TabDescStruct_sp07->sp7t_field_cnt(); ++i)
        {
            pFieldInfo = &TabDescStruct_sp07->sp7t_fields()[i];
            // Long defaults are typed as stamp even if the correct type would be other_default;
            // get those defaults, too
            if ( (sp7_other_default == pFieldInfo->sp7fi_default)               ||
                 ( (sp7_default_stamp == pFieldInfo->sp7fi_default)           &&
                   (1 == szMap_LongDataType_ls00[pFieldInfo->sp7fi_data_type]) ) )
            {
                TERec->terDefaultArray_ls00[j] = pFieldInfo->sp7fi_def_col_no;
                ++TERec->terDefaultCnt_ls00;
                ++j;
            }

            if (1 == szMap_LongDataType_ls00[pFieldInfo->sp7fi_data_type])
            {
                // Long columns in table;
                ++TERec->terLongColCnt_ls00;

                // Check for short column file in db -> available
                // only if at least one long column is of data type dstra, dstre, dstrb, dstrdb or dstruni
                if ((pFieldInfo->sp7fi_data_type < dlonga) || (pFieldInfo->sp7fi_data_type == dstruni) )
                {
                    bShortColumnFileAvailable = true;
                }
            }
        }
    }

    //*
    //*     Index description
    //*
    if (errOK_els00 == rc)
    {
        rc = ls26_SaveIndexDescription(DBInfo, TERec, TableName, pszErrText);
    }


    //*
    //*     Constraint descriptions
    //*
    if ( (errOK_els00 == rc) && (0 != TERec->terConstraintCnt_ls00) )
    {
        TERec->terPageTypeExtracting_ls00 = tptConstraint_els00;
        rc = ls26_WriteConstraintInfo(DBInfo, TableName, TERec, pszErrText);
    }

    //*
    //*     Default descriptions
    //*
    if ( (errOK_els00 == rc) && (0 != TERec->terDefaultCnt_ls00) )
    {
        TERec->terPageTypeExtracting_ls00 = tptDefault_els00;
        rc = ls26_WriteDefaultInfo(DBInfo, TableName, TERec, pszErrText);

        // Cleaning
        delete [] TERec->terDefaultArray_ls00;
        TERec->terDefaultArray_ls00 = NULL;
    }


    //*
    //*     Short long values
    //*
    if ( (errOK_els00 == rc ) && (true == bShortColumnFileAvailable))
    {
        TERec->terPageTypeExtracting_ls00 = tptShortString_els00;

        rc = ls26_WriteShortStringPages(DBInfo, TableName, TERec, pszErrText);
    }


    if (errOK_els00 == rc)
    {
        TERec->terPageTypeExtracting_ls00 = tptData_els00;

        rc = ls26_WritePrimaryDataPages(DBInfo, TableName, TERec, pszErrText);
    }


    if ( (errOK_els00 == rc ) && (0 != TERec->terLongColCnt_ls00) )
    {
        TERec->terPageTypeExtracting_ls00 = tptLongString_els00;

        rc = ls26_WriteLongStringPages(DBInfo, TERec, pszErrText);
    }

    // Write EndOfTable page to data file
    if (errOK_els00 == rc)
    {
        TERec->terPageTypeExtracting_ls00 = tptEOT_els00;
        rc = ls26_WDataHeaderPage(TERec, pszErrText);
    }


    // If the WHOLE table is extracted without errors release the table lock
    if (errOK_els00 == rc)
    {
        rc = ls15Commit(DBInfo, pszErrText);
        if (errOK_els00 != rc)      // error is of no interest here
        {
            rc = errOK_els00;
            pszErrText[0] = 0;
        }
    }

    return rc;
}
// ls26ExtractSingleTable()


/*
  -----------------------------------------------------------------------------
  function:     ls26WriteHeaderPage
  -----------------------------------------------------------------------------
*/
tsp00_Int4
ls26WriteHeaderPage(tls00_DBInfo      *DBInfo,
                    tls00_TextractRec *TERec,
                    bool               bMultipleTables,     // multiple table data in one data file ?
                    tsp00_Addr         pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls26WriteHeaderPage");

    tsp00_Int4 rc = errOK_els00;

    tsp00_Int4 lBufPos = 0;

    _TCHAR pszBuffer[TE_FILE_HEADER_LENGTH_LS00];
    memset(pszBuffer, 0, TE_FILE_HEADER_LENGTH_LS00);

    pszBuffer[lBufPos] = STATIC_CAST(unsigned char, ls07DefineSwap(NULL));
    lBufPos += sizeof(unsigned char);

    pszBuffer[lBufPos] = STATIC_CAST(unsigned char, DBInfo->dbiEncoding_ls00);
    lBufPos += sizeof(unsigned char);

    tsp00_C64c pszId;
    pszId.Init();
    pszId.rawAssign(HEADER_IDENTIFIER_LS00);
    memcpy(&pszBuffer[lBufPos], pszId.asCharp(), pszId.length());
    lBufPos += pszId.size();

    tsp00_Versionc pszVersion;
    sp100_GetVersionString(COMP_NAME_REPL_MAN_SP100, s100buildnumber, pszVersion);
    memcpy(&pszBuffer[lBufPos], pszVersion.asCharp(), pszVersion.length());
    lBufPos += pszVersion.size();

    tsp00_Date tmpDate;
    tsp00_Time tmpTime;
    sqldattime(tmpDate, tmpTime);
    tsp00_Datec pszDate;
    tsp00_Timec pszTime;
    pszDate.p2c(tmpDate);
    pszTime.p2c(tmpTime);
    memcpy(&pszBuffer[lBufPos], pszDate.asCharp(), pszDate.length());
    lBufPos += pszDate.size();
    memcpy(&pszBuffer[lBufPos], pszTime.asCharp(), pszTime.length());
    lBufPos += pszTime.size();

    memcpy(&pszBuffer[lBufPos], REINTERPRET_CAST(char*, &DBInfo->dbiPageSize_ls00), sizeof(DBInfo->dbiPageSize_ls00));
    lBufPos += sizeof(DBInfo->dbiPageSize_ls00);

    memcpy(&pszBuffer[lBufPos], REINTERPRET_CAST(char*, &TERec->terCount_ls00), sizeof(TERec->terCount_ls00));
    lBufPos += sizeof(TERec->terCount_ls00);
    
    memcpy(&pszBuffer[lBufPos], REINTERPRET_CAST(char*, &bMultipleTables), sizeof(bMultipleTables));

    
    rc = ls30VFWriteLn(*TERec->terOutFile_ls00, pszBuffer, TE_FILE_HEADER_LENGTH_LS00, pszErrText);

    return rc;
}
// ls26WriteHeaderPage() 


/*
  -----------------------------------------------------------------------------
  function:     ls26WriteFileTrailer
  -----------------------------------------------------------------------------
*/
tsp00_Int4
ls26WriteFileTrailer(tls00_DBInfo      *DBInfo,
                    tls00_TextractRec *pTERec,
                    bool               bTableIncomplete,
                    tsp00_Addr         pszFileName,         // NULL indicates no next or last file --> no entry for file
                    tsp00_Addr         pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls26WriteFileTrailer");

    tsp00_Int4 rc = errOK_els00;

    _TCHAR *pszBuffer = pTERec->terpPageBuffer_ls00;        // has length of one db page

    ++pTERec->terPageCount_ls00;

    tls26_TEFileTrailer Trailer;
    memset(&Trailer, 0, sizeof(tls26_TEFileTrailer));
    
    Trailer.tftPageNo_ls26           = pTERec->terPageCount_ls00;
    Trailer.tftType_ls26             = MY_PAGE_LS00;
    Trailer.tftPageType_ls26         = tptEOF_els00;
    Trailer.tftNumOfTabInFile_ls26   = pTERec->terNumTabInFile_ls00;
    Trailer.tftNumOfPagesInFile_ls26 = pTERec->terPageCount_ls00 - 1;
    Trailer.tftTableIncomplete_ls26  = bTableIncomplete;
    if (true == bTableIncomplete)
    {
        // type of last extracted data
        Trailer.tftDataTypeAtInterrupt_ls26 = pTERec->terPageTypeExtracting_ls00;
    }
    
    size_t lNextFileLen = 0;
    if (NULL != pszFileName)        // indicating no next or last file
    {
        lNextFileLen = _tcslen(pszFileName);
    }

    // Write trailer to beginning of page
    memcpy(pszBuffer, REINTERPRET_CAST(_TCHAR*, &Trailer), sizeof(tls26_TEFileTrailer));
    
    // Write file name to beginning of page
    memcpy(&pszBuffer[sizeof(tls26_TEFileTrailer)], pszFileName, lNextFileLen);
    pszBuffer[sizeof(tls26_TEFileTrailer) + lNextFileLen] = 0;


    // Write same trailer again to the end of the page (include 0 of filename)
    memcpy(&pszBuffer[pTERec->terPageSize_ls00 - STATIC_CAST(tsp00_Int4, sizeof(tls26_TEFileTrailer)) - lNextFileLen - 1],
           REINTERPRET_CAST(_TCHAR*, &Trailer),
           sizeof(tls26_TEFileTrailer));

    // Write same file name to end of page
    memcpy(&pszBuffer[pTERec->terPageSize_ls00 - lNextFileLen - 1], pszFileName, lNextFileLen);
    pszBuffer[pTERec->terPageSize_ls00 - 1] = 0;

    // Write trailer page to file
    rc = ls30VFWriteLn(*pTERec->terOutFile_ls00, pszBuffer, pTERec->terPageSize_ls00, pszErrText);

    return rc;
}
// ls26WriteFileTrailer()

/*
  -----------------------------------------------------------------------------
  EndChapter: Public functions implementation
  -----------------------------------------------------------------------------
*/



/*
  -----------------------------------------------------------------------------
  Chapter: Private functions implementation
  -----------------------------------------------------------------------------
*/


/*
  -----------------------------------------------------------------------------
   function:     ls26_WLongDataHeaderPage
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls26_WLongDataHeaderPage(tls00_TextractRec *pTERec,
                         tsp00_Int2         lColumnNo,
                         tsp00_Int4         lRecordKeyLen,         // nur fuer long-werte von interesse, fuer alle anderen aufrufer auf 0 setzen
                         tsp00_Addr         pszRecordKey,
                         tsp00_Addr         pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls26_WLongDataHeaderPage");

    tsp00_Int4 rc = errOK_els00;
    
    _TCHAR *pszBuffer = pTERec->terpPageBuffer_ls00;        // der enthaelt eigentlich schon die laenge einer datenseite

    ++pTERec->terPageCount_ls00;
    
    tls26_DescPageHeader Header;
    Header.dphPageNo_ls26    = pTERec->terPageCount_ls00;
    Header.dphType_ls26      = MY_PAGE_LS00;
    Header.dphPageType_ls26  = pTERec->terPageTypeExtracting_ls00;;
    Header.dphSpecialID_ls26 = lRecordKeyLen;
    Header.dphMore_ls26      = 0;                           // actually of no use here
    Header.dphFillerInt_ls26 = lColumnNo;

    // write header to the beginning of page
    memcpy(pszBuffer, REINTERPRET_CAST(_TCHAR*, &Header), sizeof(tls26_DescPageHeader));

    // write record key to page (located behind header)
    memcpy(&pszBuffer[sizeof(tls26_DescPageHeader)], pszRecordKey, lRecordKeyLen);

    // write same header again to the END of page
    memcpy(&pszBuffer[pTERec->terPageSize_ls00 - sizeof(tls26_DescPageHeader)],
           REINTERPRET_CAST(_TCHAR*, &Header),
           sizeof(tls26_DescPageHeader));

    // write page to file
    rc = ls30VFWriteLn(*pTERec->terOutFile_ls00, pszBuffer, pTERec->terPageSize_ls00, pszErrText);

    return rc;
}
// ls26_WLongDataHeaderPage() 


/*
  -----------------------------------------------------------------------------
   function:     ls26_WDataHeaderPage
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls26_WDataHeaderPage(tls00_TextractRec *pTERec, tsp00_Addr pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls26_WDataHeaderPage");

    tsp00_Int4 rc = errOK_els00;
    
    _TCHAR *pszBuffer = pTERec->terpPageBuffer_ls00;        // der enthaelt eigentlich schon die laenge einer datenseite

    ++pTERec->terPageCount_ls00;
    
    tls26_DescPageHeader Header;
    Header.dphPageNo_ls26    = pTERec->terPageCount_ls00;
    Header.dphType_ls26      = MY_PAGE_LS00;
    Header.dphPageType_ls26  = pTERec->terPageTypeExtracting_ls00;
    Header.dphSpecialID_ls26 = 0;                           // actually of no use here
    Header.dphMore_ls26      = 0;                           // actually of no use heredphFillerInt_ls26
    Header.dphFillerInt_ls26 = 0;                           // actually of no use here

    // Write header to beginning of page
    memcpy(pszBuffer, REINTERPRET_CAST(_TCHAR*, &Header), sizeof(tls26_DescPageHeader));

    TRACE_PRNF_MLS99(("ls26_WDataHeaderPage", "page size: %d.\n", pTERec->terPageSize_ls00));
    TRACE_BUFF_MLS99("ls26_WDataHeaderPage", "HEADER", sizeof(tls26_DescPageHeader), REINTERPRET_CAST(_TCHAR*, &Header));

    // Write same header again to the end of the page
    memcpy(&pszBuffer[pTERec->terPageSize_ls00 - sizeof(tls26_DescPageHeader)],
           REINTERPRET_CAST(_TCHAR*, &Header),
           sizeof(tls26_DescPageHeader));

    TRACE_BUFF_MLS99("ls26_WDataHeaderPage", "FOOTER", sizeof(tls26_DescPageHeader), REINTERPRET_CAST(_TCHAR*, &Header));

    // Write header page to file
    rc = ls30VFWriteLn(*pTERec->terOutFile_ls00, pszBuffer, pTERec->terPageSize_ls00, pszErrText);

    return rc;
}
// ls26_WDataHeaderPage() 


/*
  -----------------------------------------------------------------------------
  function:     ls26_WriteTableIndexDescToFile
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls26_WriteTableIndexDescToFile(tls00_TextractRec *TERec,
                               tls00_TEPageType   DescriptionType,
                               tsp00_Addr         pDesc,   // wird als raw char-pointer uebergeben
                               tsp00_Int4         lDescLength,
                               tsp00_Addr         pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls26_WriteTableIndexDescToFile");

    tsp00_Int4 rc = errOK_els00;

    _TCHAR *pszBuffer = TERec->terpPageBuffer_ls00;

    tls26_DescPageHeader Header;
    Header.dphPageNo_ls26   = TERec->terPageCount_ls00;
    Header.dphType_ls26     = MY_PAGE_LS00;
    Header.dphPageType_ls26 = DescriptionType;

    // Halte die gesamtlaenge der beschreibung fest
    Header.dphFillerInt_ls26 = lDescLength;

    tsp00_Int4 lHeaderLength  = sizeof(tls26_DescPageHeader);
    tsp00_Int4 lMaxDataLength = TERec->terPageSize_ls00 - lHeaderLength;
    tsp00_Int4 lWriteLength   = lDescLength;

    while (lWriteLength > 0)
    {
        Header.dphPageNo_ls26 += 1;
        if (lWriteLength <= lMaxDataLength)
        {
            Header.dphSpecialID_ls26 = lWriteLength;
            Header.dphMore_ls26      = 0;
        }
        else
        {
            Header.dphSpecialID_ls26 = lMaxDataLength;
            Header.dphMore_ls26      = 1;
        }

        //*     Puffer fuellen
        memcpy(pszBuffer, REINTERPRET_CAST(_TCHAR*, &Header), lHeaderLength);
        memcpy(&pszBuffer[lHeaderLength], &pDesc[lDescLength - lWriteLength], Header.dphSpecialID_ls26);

        //*     Puffer in die datei schreiben
        rc = ls30VFWriteLn(*TERec->terOutFile_ls00, pszBuffer, TERec->terPageSize_ls00, pszErrText);
        if (errOK_els00 != rc)
        {
            break;
        }

        lWriteLength -= Header.dphSpecialID_ls26;
    }

    if (errOK_els00 == rc)
    {
        TERec->terPageCount_ls00 = Header.dphPageNo_ls26;
    }
    return rc;
}
// ls26_WriteTableIndexDescToFile()


/*
  -----------------------------------------------------------------------------
  function:     ls26_WriteConstraintInfo
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls26_WriteConstraintInfo(tls00_DBInfo            *pDBInfo,
                         SAPDB_DynamicUTF8String  TableName,
                         tls00_TextractRec       *pTERec,
                         tsp00_Addr               pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls26_WriteConstraintInfo");

    tsp00_Int4 rc = errOK_els00;

    _TCHAR *pszBuffer = pTERec->terpPageBuffer_ls00;

    tls26_DescPageHeader Header;
    Header.dphPageNo_ls26   = pTERec->terPageCount_ls00;
    Header.dphType_ls26     = MY_PAGE_LS00;
    Header.dphPageType_ls26 = pTERec->terPageTypeExtracting_ls00;

    tsp00_Int4 lHeaderLength  = sizeof(tls26_DescPageHeader);
    tsp00_Int4 lMaxDataLength = pTERec->terPageSize_ls00 - lHeaderLength;   // max. length of data on single page
    tsp00_Int4 lBufLength     = lMaxDataLength;
    tsp00_Int4 lmyOffset      = lHeaderLength;                              // offset im puffer

    tsp00_Int2           lPartLen           = 0;
    tsp00_Int2           myConstraintOnPage = 0;
    tls26_Co_De_LenType  lConstraintLen     = 0;
    tsp00_Int2           lLengthLen         = sizeof(tls26_Co_De_LenType);       // length of the constraint length

    tsp1_packet_ptr      SqlRcvPacket       = NULL;
    tsp1_part           *SqlDataPart        = NULL;


    TRACE_PRNF_MLS99(("ls26_WriteConstraintInfo", "Overall number of constraints: %d.\n", pTERec->terConstraintCnt_ls00));

    _TCHAR     szErr[MAX_REPLY_LENGTH_LS00];
    tsp00_Int4 lErrorPos = 0;

    _TCHAR szCmd[lStmtLength_ls26];
    szCmd[0] = 0;

    sp77sprintf(szCmd, lStmtLength_ls26, DESCRIBE_C_STMT_LS26, TableName.StrPtr());

    tsp00_Int2 i = 0;
    for (i = 1, myConstraintOnPage; i <= pTERec->terConstraintCnt_ls00; ++i, ++myConstraintOnPage)
    {
        //*
        //*     Get single constraints of table from kernel
        //* We supply the kernel additionally to the command 'describe constraint <tablename>' with
        //* the constraint number of each single constraint. This number is simply the running index
        //* up to the overall count of constraints on the current table (pTERec->terConstraintCnt_ls00).
        rc = ls13ExecIntSQLStmtExUTF8(pDBInfo,
                                      0,            // packet no
                                      szCmd,
                                      REINTERPRET_CAST(char*, &i),
                                      sizeof(tsp00_Int2),
                                      pszErrText,
                                      lErrorPos);
        if (errOK_els00 != rc)
        {
            RepServMessages::Instance()->ls98Msg(layTableExtract_els98, pszErrText, errInternSQL_els98,
                                                                                    szCmd, szErr);
            rc = errInternSQL_els98;
            break;
        }

        //*     Ergebnis auswerten
        SqlRcvPacket = pDBInfo->dbiPktRcvList_ls00[0];
        if ( false == ls04GetPart(SqlRcvPacket, sp1pk_data, SqlDataPart) )
        {
            RepServMessages::Instance()->ls98Msg(layTableExtract_els98, pszErrText, errInternal_els98,
                                                                                    __FILE__,
                                                                                    __LINE__,
                                                                                    errUnknownPart_els98);
            rc = errInternal_els98;
            break;
        }

        lPartLen       = ls04GetPartLength(SqlDataPart);
        lConstraintLen = lPartLen  + lLengthLen;    // include the length of the column number of the constraint 

        if ( (lConstraintLen  + lLengthLen) > lBufLength )              // include the length of the contraint length 
        {
            // before writing the new value we must save the buffer to the file
            Header.dphPageNo_ls26   += 1;
            Header.dphSpecialID_ls26 =  myConstraintOnPage;
            if (i < (pTERec->terConstraintCnt_ls00 - 1) )
            {
                Header.dphMore_ls26 = 1;
            }
            else
            {
                Header.dphMore_ls26 = 0;
            }
            memcpy(&pszBuffer[0], REINTERPRET_CAST(_TCHAR*, &Header), lHeaderLength);

            TRACE_PRNF_MLS99(("ls26_WriteConstraintInfo", "page no: %d.\n", Header.dphPageNo_ls26));
            TRACE_PRNF_MLS99(("ls26_WriteConstraintInfo", "constraint cnt on page: %d.\n", Header.dphSpecialID_ls26));

            rc = ls30VFWriteLn(*pTERec->terOutFile_ls00, pszBuffer, pTERec->terPageSize_ls00, pszErrText);
            if (errOK_els00 != rc)
            {
                break;
            }

            // Reinitialize
            lmyOffset          = lHeaderLength;
            lBufLength         = lMaxDataLength;
            myConstraintOnPage = 0;

        }   // end if ( (lConstraintLen  + lLengthLen) > lBufLength )
        
        // Put length of constraints into page
        memcpy(&pszBuffer[lmyOffset], REINTERPRET_CAST(char*, &lConstraintLen), lLengthLen);
        lBufLength -= lLengthLen;
        lmyOffset  += lLengthLen;
    
        TRACE_PRNF_MLS99(("ls26_WriteConstraintInfo", "constraint length: %d.\n", lConstraintLen));

        // write column number into page as part of constraint
        memcpy(&pszBuffer[lmyOffset], REINTERPRET_CAST(char*, &i), lLengthLen);
        lBufLength -= lLengthLen;
        lmyOffset  += lLengthLen;
    
        // Put constraint into page
        ls04BufFromPartEx(SqlDataPart, 0, lPartLen, &pszBuffer[lmyOffset], lBufLength);
        lBufLength -= lPartLen;
        lmyOffset  += lPartLen;
    }   // end for (i = 1, myConstraintOnPage; i <= pTERec->terConstraintCnt_ls00; ++i, ++myConstraintOnPage)

    //*     Last writing of buffer to file
    //* only if there is still one that was not saved so far
    if ( (lmyOffset > lHeaderLength) && (errOK_els00 == rc) )
    {
        Header.dphPageNo_ls26   += 1;
        Header.dphSpecialID_ls26 =  myConstraintOnPage;
        Header.dphMore_ls26      = 0;                       // definitely last contraint page

        memcpy(&pszBuffer[0], REINTERPRET_CAST(_TCHAR*, &Header), lHeaderLength);

        TRACE_PRNF_MLS99(("ls26_WriteConstraintInfo", "page no: %d.\n", Header.dphPageNo_ls26));
        TRACE_PRNF_MLS99(("ls26_WriteConstraintInfo", "constraint cnt on page: %d.\n", Header.dphSpecialID_ls26));
        
        rc = ls30VFWriteLn(*pTERec->terOutFile_ls00, pszBuffer, pTERec->terPageSize_ls00, pszErrText);
    }

    if (errOK_els00 == rc)
    {
        pTERec->terPageCount_ls00 = Header.dphPageNo_ls26;
    }

    return rc;
}
// ls26_WriteConstraintInfo()


/*
  -----------------------------------------------------------------------------
  function:     ls26_WriteDefaultInfo
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls26_WriteDefaultInfo(tls00_DBInfo            *DBInfo,
                      SAPDB_DynamicUTF8String  TableName,
                      tls00_TextractRec       *TERec,
                      tsp00_Addr               pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls26_WriteDefaultInfo");

    tsp00_Int4 rc = errOK_els00;

    _TCHAR *pszBuffer = TERec->terpPageBuffer_ls00;

    tls26_DescPageHeader Header;
    Header.dphPageNo_ls26   = TERec->terPageCount_ls00;
    Header.dphType_ls26     = MY_PAGE_LS00;
    Header.dphPageType_ls26 = TERec->terPageTypeExtracting_ls00;

    tsp00_Int4       lHeaderLength   = sizeof(tls26_DescPageHeader);
    tsp00_Int4       lMaxDataLength  = TERec->terPageSize_ls00 - lHeaderLength;
    tsp00_Int4       lBufLength      = lMaxDataLength;
    tsp00_Int4       lmyOffset       = lHeaderLength;
    
    tsp00_Int2       myDefaultOnPage = 0;
    tsp00_Int4       lPartLen        = 0;
    tsp00_Int2       lColNoLength    = sizeof(tls26_Co_De_LenType);

    tsp1_packet_ptr  SqlRcvPacket    = NULL;
    tsp1_part       *SqlDataPart     = NULL;

    TRACE_PRNF_MLS99(("ls26_WriteDefaultInfo", "Overall number of defaults: %d.\n", TERec->terDefaultCnt_ls00));

    tsp00_Addr pDefault = NULL;
    tsp00_Int2 lDefLen  = 0;

    tsp00_Int2 i = 0;
    for (i, myDefaultOnPage; i < TERec->terDefaultCnt_ls00; ++i, ++myDefaultOnPage)
    {
        //*     Get single default values
        rc = ls13GetDefaultValue(DBInfo, TableName, TERec->terDefaultArray_ls00[i], pDefault, lDefLen, pszErrText);
        if (errOK_els00 == rc)
        {
            if ( (lDefLen + lColNoLength) > lBufLength )
            {
                // write buffer to file before saving the next value
                Header.dphPageNo_ls26   += 1;
                Header.dphSpecialID_ls26 =  myDefaultOnPage;
                if (i < (TERec->terDefaultCnt_ls00 - 1) )
                {
                    Header.dphMore_ls26 = 1;
                }
                else
                {
                    Header.dphMore_ls26 = 0;
                }
                memcpy(&pszBuffer[0], REINTERPRET_CAST(_TCHAR*, &Header), lHeaderLength);

                TRACE_PRNF_MLS99(("ls26_WriteDefaultInfo", "page no: %d.\n", Header.dphPageNo_ls26));
                TRACE_PRNF_MLS99(("ls26_WriteDefaultInfo", "default cnt on page: %d.\n", Header.dphSpecialID_ls26));


                rc = ls30VFWriteLn(*TERec->terOutFile_ls00, pszBuffer, TERec->terPageSize_ls00, pszErrText);
                if (errOK_els00 != rc)
                {
                    break;
                }

                //* reinitialize
                lmyOffset       = lHeaderLength;
                lBufLength      = lMaxDataLength;
                myDefaultOnPage = 1;        // reinit to 1 because the next one is already found

            }

            // what does the saved value look like
            //      ----------------------------------------------------------------------------
            //      | column number | value length | default value with leading defined byte   |
            //      ----------------------------------------------------------------------------
            //         2 byte           2 byte           length byte

            // Save column number of default
            memcpy(&pszBuffer[lmyOffset], REINTERPRET_CAST(char*, &TERec->terDefaultArray_ls00[i]), lColNoLength);
            lBufLength -= lColNoLength;
            lmyOffset  += lColNoLength;

            // Save length of default value
            memcpy(&pszBuffer[lmyOffset], REINTERPRET_CAST(char*, &lDefLen), sizeof(lDefLen));
            lBufLength -= sizeof(lDefLen);
            lmyOffset  += sizeof(lDefLen);

            // Save default value
            memcpy(&pszBuffer[lmyOffset], pDefault, lDefLen);

            lBufLength -= lDefLen;
            lmyOffset  += lDefLen;
        }

    }   // for (i, myDefaultOnPage; i < TERec->terDefaultCnt_ls00; ++i, ++myDefaultOnPage)

    //*     Last time writing the buffer
    //* only if there is still one thats not yet saved
    if ( (lmyOffset > lHeaderLength) && (errOK_els00 == rc) )
    {
        Header.dphPageNo_ls26   += 1;
        Header.dphSpecialID_ls26 =  myDefaultOnPage;
        Header.dphMore_ls26      = 0;                       // surely the last default page

        memcpy(&pszBuffer[0], REINTERPRET_CAST(_TCHAR*, &Header), lHeaderLength);
        
        TRACE_PRNF_MLS99(("ls26_WriteDefaultInfo", "page no: %d.\n", Header.dphPageNo_ls26));
        TRACE_PRNF_MLS99(("ls26_WriteDefaultInfo", "default cnt on page: %d.\n", Header.dphSpecialID_ls26));

        rc = ls30VFWriteLn(*TERec->terOutFile_ls00, pszBuffer, TERec->terPageSize_ls00, pszErrText);
    }

    if (errOK_els00 == rc)
    {
        TERec->terPageCount_ls00 = Header.dphPageNo_ls26;
    }

    return rc;
}
// ls26_WriteDefaultInfo()


/*
  -----------------------------------------------------------------------------
  function:     ls26_WriteShortStringPages
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls26_WriteShortStringPages(tls00_DBInfo            *pDBInfo,
                           SAPDB_DynamicUTF8String  TableName,
                           tls00_TextractRec       *pTERec,
                           tsp00_Addr               pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls26_WriteShortStringPages");

    tsp00_Int4 rc = errOK_els00;


    _TCHAR szCmd[lStmtLength_ls26];
    szCmd[0] = 0;
    sp77sprintf(szCmd, lStmtLength_ls26, SAVE_LONG_STMT_LS26, TableName.StrPtr());

    tsp00_Int4 lErrPos;
    _TCHAR     szErr[MAX_REPLY_LENGTH_LS00];

    rc = ls13ExecIntSQLStmtUTF8(pDBInfo, 0, szCmd, szErr, lErrPos);
    if (errOK_els00 != rc)
    {
        RepServMessages::Instance()->ls98Msg(layTableExtract_els98, pszErrText, errInternSQL_els98,
                                                                                szCmd, szErr);
        rc = errInternSQL_els98;
    }
    else
    {
        tsp1_packet *SqlPacket = pDBInfo->dbiPktRcvList_ls00[0];
        tsp1_part   *SqlPart   = NULL;

        if ( false == ls04GetPart(SqlPacket, sp1pk_bdinfo, SqlPart) )
        {
            RepServMessages::Instance()->ls98Msg(layTableExtract_els98, pszErrText, errInternal_els98,
                                                                                    __FILE__,
                                                                                    __LINE__,
                                                                                    errUnknownPart_els98);
            rc = errInternal_els98;
        }
        else
        {
            ls04BufFromPartEx(SqlPart, 0, BDINFO_SIZE_LS00, pTERec->terBdInfo_ls00, BDINFO_SIZE_LS00);

            //*
            //*     Write intro page for short string data to data file
            //*
            rc = ls26_WDataHeaderPage(pTERec, pszErrText);
            if (errOK_els00 == rc)
            {
                //*
                //*     Get short string data pages
                //*
                rc = ls26_PageExtractLoop(pDBInfo, pTERec, pszErrText);
            }
        }
    }
    return rc;
}
// ls26_WriteShortStringPages()


/*
  -----------------------------------------------------------------------------
  function:     ls26_WriteLongStringPages
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls26_WriteLongStringPages(tls00_DBInfo      *pDBInfo,
                          tls00_TextractRec *pTERec,
                          tsp00_Addr         pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls26_WriteLongStringPages");

    tsp00_Int4 rc = errOK_els00;

    tsp00_Int4 lErrorPos = 0;
    _TCHAR     szErr[MAX_REPLY_LENGTH_LS00];
    _TCHAR     szCmd[lStmtLength_ls26];
    szCmd[0] = 0;
    
    // Create command (because it doesn't contain identifiers it is already UTF8 encoded)
    _tcscpy(szCmd, SAVE_COLUMN_STMT_LS26);

    tls00_Parts4InternalSql myStruct;
    
    myStruct.pfiCount_ls00 = 3;
    myStruct.pfiPartKind_ls00    = new tsp1_part_kind_Enum[myStruct.pfiCount_ls00];
    myStruct.pfiPartKind_ls00[0] = sp1pk_command;
    myStruct.pfiPartKind_ls00[1] = sp1pk_bdinfo;
    myStruct.pfiPartKind_ls00[2] = sp1pk_data;
    
    
    myStruct.pfiData_ls00        = new tls00_String[myStruct.pfiCount_ls00];

    // cmd part
    myStruct.pfiData_ls00[0].strAddr_ls00 = szCmd;
    myStruct.pfiData_ls00[0].strLeng_ls00 = STATIC_CAST(tsp00_Int4, _tcslen(szCmd));

    // bdinfo part
    myStruct.pfiData_ls00[1].strAddr_ls00 = new _TCHAR[BDINFO_SIZE_LS00];
    myStruct.pfiData_ls00[1].strLeng_ls00 = BDINFO_SIZE_LS00;
    memcpy(myStruct.pfiData_ls00[1].strAddr_ls00, pTERec->terLongBdInfo_ls00, BDINFO_SIZE_LS00);

    // data part
    // this part needs to be reinitialized for every long column to be extracted; to simplify
    // work we'll put a pointer to this part

    // folgendes aussehen
    //--------------------------------------------------------------------------------------
    //          |          |              |            |
    // FilePos  | long col | current long | key length | key 
    //          |   cnt    |  col number  |            |
    //--------------------------------------------------------------------------------------
    //  8 byte      2 byte       2 byte      2 byte       (key length) byte
    
    tsp00_Int4    lNumPartLen = FILE_POS_MXGG00 + 3 * INT2_MXSP00;      // this length is used later on
    
    tls00_String *pMyData     = &myStruct.pfiData_ls00[2];
    pMyData->strLeng_ls00     = lNumPartLen + MAX_KEYLEN_GG00;
    pMyData->strAddr_ls00     = new _TCHAR[myStruct.pfiData_ls00[2].strLeng_ls00];

    // first 8 byte are initially of no interest and are skipped therefore
    memcpy(&pMyData->strAddr_ls00[FILE_POS_MXGG00],
           REINTERPRET_CAST(char*, &pTERec->terLongColCnt_ls00),
           INT2_MXSP00);

    // next 10 bytes are initialized with 0
    // - first 2 bytes represent long_col_no (must initially be 0)
    // - next 8 byte are length representations in a record (record header) (and must initially also be 0)
    memset(&pMyData->strAddr_ls00[FILE_POS_MXGG00 + INT2_MXSP00], 0, (cgg_rec_key_offset + 2*INT2_MXSP00));

    while (1)
    {
        rc = ls13ExecIntSQLMultPartsUTF8(pDBInfo, 0, myStruct, szErr, lErrorPos);
        if (errOK_els00 != rc)
        {
            if (100 == rc)      // no more LONGs to extract found by kernel
            {
                rc = errOK_els00;
            }
            else
            {
                RepServMessages::Instance()->ls98Msg(layTableExtract_els98, pszErrText, errInternSQL_els98,
                                                                                        szCmd, szErr);
                rc = errInternSQL_els98;
            }
            break;
        }

        tsp1_packet *SqlPacket = pDBInfo->dbiPktRcvList_ls00[0];
        tsp1_part   *SqlPart   = NULL;

        if ( false == ls04GetPart(SqlPacket, sp1pk_bdinfo, SqlPart) )
        {
            RepServMessages::Instance()->ls98Msg(layTableExtract_els98, pszErrText, errInternal_els98,
                                                                                    __FILE__,
                                                                                    __LINE__,
                                                                                    errUnknownPart_els98);
            rc = errInternal_els98;
            break;
        }
        else
        {
            ls04BufFromPartEx(SqlPart, 0, BDINFO_SIZE_LS00, pTERec->terBdInfo_ls00, BDINFO_SIZE_LS00);
        }

        if ( false == ls04GetPart(SqlPacket, sp1pk_key, SqlPart) )
        {
            RepServMessages::Instance()->ls98Msg(layTableExtract_els98, pszErrText, errInternal_els98,
                                                                                    __FILE__,
                                                                                    __LINE__,
                                                                                    errUnknownPart_els98);
            rc = errInternal_els98;
            break;
        }
        else
        {
            // right data here to run next ls13ExecIntSQLMultParts
            myStruct.pfiData_ls00[2].strLeng_ls00 = ls04GetPartLength(SqlPart);
            ls04BufFromPartEx(SqlPart, 0, pMyData->strLeng_ls00, pMyData->strAddr_ls00, pMyData->strLeng_ls00);
        }


        //*
        //*     Write intro page for long string data to data file
        //*
        //* Make colno and key length usable
        //*
        tsp00_Int2 lColNo = *REINTERPRET_CAST(tsp00_Int2*, &pMyData->strAddr_ls00[FILE_POS_MXGG00 + INT2_MXSP00]);
        tsp00_Int4 lKeyLen = *REINTERPRET_CAST(tsp00_Int2*, &pMyData->strAddr_ls00[FILE_POS_MXGG00 + INT2_MXSP00 + INT2_MXSP00]);

        rc = ls26_WLongDataHeaderPage(pTERec, lColNo, lKeyLen, &pMyData->strAddr_ls00[lNumPartLen], pszErrText);
        if (errOK_els00 != rc)
        {
            break;
        }
    
        //*
        //*     Get long string data pages
        //*
        rc = ls26_PageExtractLoop(pDBInfo, pTERec, pszErrText);
        if (errOK_els00 != rc)
        {
            break;
        }
    }   // end while (1)


    if (NULL != myStruct.pfiPartKind_ls00)
    {
        delete [] myStruct.pfiPartKind_ls00;
        myStruct.pfiPartKind_ls00 = NULL;
    }

    if (NULL != myStruct.pfiData_ls00[1].strAddr_ls00)
    {
        delete [] myStruct.pfiData_ls00[1].strAddr_ls00;
        myStruct.pfiData_ls00[1].strAddr_ls00 = NULL;
    }
    if (NULL != myStruct.pfiData_ls00[2].strAddr_ls00)
    {
        delete [] myStruct.pfiData_ls00[2].strAddr_ls00;
        myStruct.pfiData_ls00[2].strAddr_ls00 = NULL;
    }

    if (NULL != myStruct.pfiData_ls00)
    {
        delete [] myStruct.pfiData_ls00;
        myStruct.pfiData_ls00 = NULL;
    }

    return rc;
}
// ls26_WriteLongStringPages()


/*
  -----------------------------------------------------------------------------
  function:     ls26_WritePrimaryDataPages
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls26_WritePrimaryDataPages(tls00_DBInfo            *pDBInfo,
                           SAPDB_DynamicUTF8String  TableName,
                           tls00_TextractRec       *pTERec,
                           tsp00_Addr               pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls26_WritePrimaryDataPages");

    tsp00_Int4 rc = errOK_els00;


    _TCHAR szCmd[lStmtLength_ls26];
    szCmd[0] = 0;
    
    //* Kommando erzeugen
    sp77sprintf(szCmd, lStmtLength_ls26, SAVE_TABLE_STMT_LS26, TableName.StrPtr());

    tsp00_Int4 lErrPos;
    _TCHAR     szErr[MAX_REPLY_LENGTH_LS00];

    rc = ls13ExecIntSQLStmtUTF8(pDBInfo, 0, szCmd, szErr, lErrPos);
    if (errOK_els00 != rc)
    {
        RepServMessages::Instance()->ls98Msg(layTableExtract_els98, pszErrText, errInternSQL_els98,
                                                                                szCmd, szErr);
        rc = errInternSQL_els98;
    }
    else
    {
        tsp1_packet *SqlPacket = pDBInfo->dbiPktRcvList_ls00[0];
        tsp1_part   *SqlPart   = NULL;

        if ( false == ls04GetPart(SqlPacket, sp1pk_bdinfo, SqlPart) )
        {
            RepServMessages::Instance()->ls98Msg(layTableExtract_els98, pszErrText, errInternal_els98,
                                                                                    __FILE__,
                                                                                    __LINE__,
                                                                                    errUnknownPart_els98);
            rc = errInternal_els98;
        }
        else
        {
            ls04BufFromPartEx(SqlPart, 0, BDINFO_SIZE_LS00, pTERec->terBdInfo_ls00, BDINFO_SIZE_LS00);

            // Hier muessen wir den bdinfo noch fuer die spaetere extraktion der long-werte speichern
            memcpy(pTERec->terLongBdInfo_ls00, pTERec->terBdInfo_ls00, BDINFO_SIZE_LS00);

            //*
            //*     Write intro page for primary data to data file
            //*
            rc = ls26_WDataHeaderPage(pTERec, pszErrText);
            if (errOK_els00 == rc)
            {
                //*
                //*     Get primary data pages
                //*
                rc = ls26_PageExtractLoop(pDBInfo, pTERec, pszErrText);
            }
        }
    }

    return rc;
}
// ls26_WritePrimaryDataPages()


/*
  -----------------------------------------------------------------------------
  function:     ls26_SendBdinfoPckt
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls26_SendBdinfoPckt(tls00_DBInfo *pDBInfo,
                    tsp00_Int4    lPcktNo,
                    tsp00_Addr    pszRawChars,      // z.b. BdInfo,
                    tsp00_Int4    lRawCharsLength,
                    tsp00_Addr    pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls26_SendBdinfoPckt");
    
    tsp00_Int4 rc = errOK_els00;

    tsp1_packet  *SqlPacket = pDBInfo->dbiPktSndList_ls00[lPcktNo];
    tsp1_segment *SqlSegm   = NULL;
    tsp1_part    *SqlPart   = NULL;

    //*
    //*     Initialize/fill packet for statement
    //* Setzt autocommit immer auf false fuer interne kommandos.
    //* Der sql mode wird in FastInit IMMER auf INTERNAL fuer interne kommandos gesetzt
    //*
    ls04FastInit(SqlPacket, &SqlSegm, &SqlPart, pDBInfo);    // sollte keine fehler produzieren

    ls04SetProducer(SqlSegm, sp1pr_internal_cmd);
    
    ls04SetMsgType(SqlSegm, sp1m_unload);

    // Eigentlich muesste fuer die folgende funktion der fehlercode abgefragt werden. Der kann
    // allerdings nur beinhalten, dass der puffer zu klein ist. Da das hier aber das einzige
    // ist, was wir eintragen, reicht der puffer in jedem fall aus.
    ls04BufToPart(SqlPart, pszRawChars, lRawCharsLength);

    rc = ls04FinishPart (SqlPacket, SqlPart);
    if (errOK_els00 == rc)
    {
        // Sende das paket
        rc = ls03SQLRequest (pDBInfo, lPcktNo, pszErrText);
        if (errOK_els00 == rc)
        {
            TRACE_PRNF_MLS99(("ls26_SendBdinfoPckt", "Sent PacketNo: %i.\n------------------------\n", lPcktNo));
        }
    }    
    else
    {
        RepServMessages::Instance()->ls98Msg(layTableExtract_els98, pszErrText, STATIC_CAST(ErrorNumbers, rc));
        rc = errFinishPart_els98;
    }
 
    return rc;
}
//  ls26_SendBdinfoPckt()


/*
  -----------------------------------------------------------------------------
   function:   ls26_PageExtractLoop - local
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls26_PageExtractLoop(tls00_DBInfo      *pDBInfo,
                     tls00_TextractRec *pTERec,
                     tsp00_Addr         pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls26_PageExtractLoop");
    
    tsp00_Int4  rc = errOK_els00;
    
    tsp00_Int4 lSndPcktNo = 0;
    tsp00_Int4 lRcvPcktNo = 0;

    tsp00_Int4 nSndPcktCnt = 0;

    //*     HAUPTTEIL DES tableextracts
    rc = ls26_SendBdinfoPckt(pDBInfo, lSndPcktNo, pTERec->terBdInfo_ls00, BDINFO_SIZE_LS00, pszErrText);
    if (errOK_els00 == rc)
    {
        nSndPcktCnt += 1;
        lSndPcktNo   = (lSndPcktNo + 1) % 2;

        tsp1_packet *SqlAuswertPacket = NULL;
        tsp1_part   *SqlAuswertPart   = NULL;

        tsp00_Int4 lLoopCount = 0;
    
        while (1)
        {   
            if (lSndPcktNo != lRcvPcktNo)   // Gegebenenfalls anderes SQL-Packet empfangen
            {
                //*     SQL-Packet empfangen
                //  Wenn erfolgreich passe die nummer des zu empfangenden pakets an
                rc = ls13ReceiveBdinfoPckt(pDBInfo, lRcvPcktNo, pTERec->terBdInfo_ls00, pszErrText);
                if (errOK_els00 != rc)
                {
                    // die funktion schreibt schon die fehlermeldung
                    break;
                }
            }
        
            //*     SQL-Packet senden
            //  Wenn erfolgreich passe die nummer des zu sendenden pakets an
            rc = ls26_SendBdinfoPckt(pDBInfo, lSndPcktNo, pTERec->terBdInfo_ls00, BDINFO_SIZE_LS00, pszErrText);
            if (errOK_els00 != rc)
            {
                // die funktion schreibt schon die fehlermeldung
                break;
            }
            else
            {
                nSndPcktCnt += 1;
                lSndPcktNo   = (lSndPcktNo + 1) % 2;
            }

            //  Werte ein SQL-Packet aus
            SqlAuswertPacket = pDBInfo->dbiPktRcvList_ls00[lRcvPcktNo];
            SqlAuswertPart   = NULL;

            lRcvPcktNo = (lRcvPcktNo + 1) % 2;      // Setzen der naechsten empfangspaketnummer !

            rc = ls26_Analyse(SqlAuswertPacket, pTERec, pszErrText);
            if (errOK_els00 != rc)
            {
                if (STOP_ANALYSE_LS26 == rc)
                {
                    rc = errOK_els00;
                }
                break;
            }

            ++lLoopCount;
            TRACE_PRNF_MLS99(("ls26_PageExtractLoop", "LoopCount: %d\n", lLoopCount));
        }   // while loop
    }

    //*     letztes SQL-Packet
    //* bei diesem interessiert uns das resultat nicht mehr, da es sich um ein ueberbleibsel handelt
    if (errOK_els00 == rc)
    {
        tsp00_Int4 rc2 = ls13ReceiveBdinfoPckt(pDBInfo, lRcvPcktNo, pTERec->terBdInfo_ls00, pszErrText);
    }

    return rc;
}
//  ls26_PageExtractLoop()


/*
  -----------------------------------------------------------------------------
  function:     ls26_Analyse
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls26_Analyse(tsp1_packet_ptr    pSqlPacket,
             tls00_TextractRec *pTERec,
             tsp00_Addr         pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls26_Analyse");
    
    tsp00_Int4  rc = errOK_els00;

    tsp00_Int4 lPageCnt = 0;
    tsp1_part *pSqlPart = NULL;
    if ( false == ls04GetPart(pSqlPacket, sp1pk_page, pSqlPart) )
    {
        RepServMessages::Instance()->ls98Msg(layTableExtract_els98, pszErrText, errInternal_els98,
                                                                                __FILE__,
                                                                                __LINE__,
                                                                                errUnknownPart_els98);
        rc = errInternal_els98;
    }
    else
    {
        tsp00_Int4 len = ls04GetPartLength(pSqlPart);
        if (0 == len)
        {
            rc = STOP_ANALYSE_LS26;
        }
        else
        {
            lPageCnt = len/pTERec->terPageSize_ls00;

            tsp00_Int4 lPageOffset  = 0;
            tsp00_Int4 lPageNo      = pTERec->terPageCount_ls00;
            tsp00_Addr pszBuffer    = NULL;
            tsp00_Addr pBufStartPos = NULL;
            
            int i = 0;
            for (i; i < lPageCnt; ++i)
            {
                lPageNo += 1;
                lPageOffset = i*pTERec->terPageSize_ls00;
                ls04BufPtrFromPart(pSqlPart, lPageOffset, pszBuffer);

                if (0 == i)
                {
                    pBufStartPos = pszBuffer;
                }

                // Mache den puffer kurzzeitig zu einer db-page und trage die erforderlichen werte ein
                // um zu kennzeichnen, dass es eine db-page ist, wird der page type auf 0 gesetzt
                (REINTERPRET_CAST(tbd_node*, pszBuffer))->nd_id()  = lPageNo;
                (REINTERPRET_CAST(tbd_node*, pszBuffer))->nd_pt().becomes(ptNil_egg00);
            }

            rc = ls30VFWriteLn(*pTERec->terOutFile_ls00, pBufStartPos, len, pszErrText);
            if (errOK_els00 == rc)
            {
                pTERec->terPageCount_ls00 = lPageNo;
                if (lPageCnt < pTERec->terPgPerPckt_ls00)
                {
                    rc = STOP_ANALYSE_LS26;
                }
            }
        }
    }

    return rc;
}
// ls26_Analyse()


/*
  -----------------------------------------------------------------------------
  function:     ls26_SaveIndexDescription
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls26_SaveIndexDescription(tls00_DBInfo            *pDBInfo,
                          tls00_TextractRec       *pTERec,
                          SAPDB_DynamicUTF8String  TableName,
                          tsp00_Addr               pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls26_SaveIndexDescription");
    
    tsp00_Int4  rc = errOK_els00;

    tsp00_Addr pszDesc = NULL;
    tsp00_Int4 lDescLen = 0;
    _TCHAR     szErr[MAX_REPLY_LENGTH_LS00];
    szErr[0] = 0;

    rc = ls13GetIndexDescription(pDBInfo, 0, TableName, lDescLen, pszDesc, szErr);
    if (errOK_els00 == rc)
    {
        rc = ls26_WriteTableIndexDescToFile(pTERec, tptIndexDescription_els00, pszDesc, lDescLen, pszErrText);
    }

    return rc;
}
// ls26_SaveIndexDescription()


//returns pointer to tabledescription
/*
  -----------------------------------------------------------------------------
  function:     ls26_SaveTableDescription
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls26_SaveTableDescription(tls00_DBInfo            *pDBInfo,
                          tls00_TextractRec       *pTERec,
                          SAPDB_DynamicUTF8String  TableName,
                          tsp00_Addr              &pszTableDesc,
                          tsp00_Addr               pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls26_SaveTableDescription");
    
    tsp00_Int4  rc = errOK_els00;

    tsp00_Int4 lDescLen = 0;
    
    rc = ls13GetTableDescription(pDBInfo, 0, TableName, lDescLen, pszTableDesc, pszErrText);
    if (errOK_els00 == rc)
    {
        rc = ls26_WriteTableIndexDescToFile(pTERec, tptTableDescription_els00, pszTableDesc, lDescLen, pszErrText);
    }
    return rc;
}
// ls26_SaveTableDescription()

/*
  -----------------------------------------------------------------------------
  EndChapter: Private functions implementation
  -----------------------------------------------------------------------------
*/
