/*!
  -----------------------------------------------------------------------------

  module: vls27_DEExec.cpp

  -----------------------------------------------------------------------------

  responsible:  SteffenS and Hans-GeorgB

  special area: Replication Manager

  description:  Implementation of functions for dataextract execution

  version:      7.3.

  -----------------------------------------------------------------------------

  copyright:    SAP AG, 2000-2003

  -----------------------------------------------------------------------------


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*/
#include "gls00.h"

#include "hls01.h"
#include "hls03.h"
#include "hls04.h"
#include "hls041.h"
#include "hls05.h"      // spezielle string routinen
#include "hls07.h"
#include "hls08_long.h"     // long value support
#include "hls12.h"
#include "hls15.h"
#include "hls18.h"
#include "hls24.h"
#include "hls25.h"
#include "hls27_DEExec.h"
#include "hls30.h"
#include "hls31_odbc_Result.h"
#include "hls98msg.h"
#include "hls99.h"

_TCHAR const *pszLoadCmdBeginString[6] = {_T("NOT USED"),
                                          _T("DATALOAD TABLE "),
                                          _T("FASTLOAD WITH 80%% USAGE TABLE "),
                                          _T("DATAUPDATE TABLE "),
                                          _T("DATALOAD TABLE "),
                                          _T("FASTLOAD WITH 80%% USAGE TABLE ")};

_TCHAR const *pszDateTimeType[6] = {_T("FREE MASK"),
                                    _T("ISO"),
                                    _T("USA"),
                                    _T("EUR"),
                                    _T("JIS"),
                                    _T("INTERNAL")};

_TCHAR const *pszEncoding[8] = {_T("NOT USED"),
                                _T("ASCII"),
                                _T("EBCDIC"),
                                _T("UTF8"),
                                _T("BINARY"),
                                _T("UCS2"),
                                _T("UCS2"),
                                _T("UCS2")};


/*  
  -----------------------------------------------------------------------------
  Chapter: Private functions declaration
  -----------------------------------------------------------------------------
*/
/*
  -----------------------------------------------------------------------------
  function:     ls27_TransformData
  -----------------------------------------------------------------------------

  description:  

  arguments:    DBInfo        [IN]     - pointer; to db session structure
                DextRec       [IN]     - pointer; to dataextract record structure
                pColumn       [IN]     - pointer; to column structure
                pszMsgBuffer  [IN]     - zeiger; datenpuffer im fetch-paket
                pResultString [OUT]    - reference to pointer; enthaelt die parameter
                                         der longdatendatei fuer die zieldatendatei;
                                         muss als NULL-pointer uebergeben werden, da
                                         der wert neu angelegt wird
                pszErrText    [OUT]    - Fehlertext


  returnvalue:  errOK_els00          - wenn alles ohne fehler abgelaufen ist
                errLongFileNumGTMax_els98 - fehler beim auslesen der deskriptoren im

                Fehler von ls27_ExecGETVAL

                Liefert den fehlercode und auch gleich den fehlertext im fehlerstring.
  -----------------------------------------------------------------------------
*/
static tsp00_Longint
ls27_TransformData(tls00_DBInfo      *DBInfo,
                   tls00_DextractRec *DextRec,
                   tls00_DatExtrCmd  *DextCmd,
                   tls27_DataSink    *pDataSink,
                   tls00_Buffer2     *pConvertBuffer,
                   tsp00_Addr         pszErrText);

/*
  -----------------------------------------------------------------------------
  function:     ls27_EditLongOffsetForOutput
  -----------------------------------------------------------------------------

  description:  

  arguments:    pColumnData     [IN] - column data structure
                pLongExecInfo   [IN] - info on number of extracted longs
                pMultCol        [IN] - column descriptions supplied by user

  returnvalue:  errOK_els00               - everything ok
                errLongFileNumGTMax_els98 - counter for numbered files exceeds max
  -----------------------------------------------------------------------------
*/
static tsp00_Longint
ls27_EditLongOffsetForOutput(tls00_ColumnData     *pColumnData,     //destination
                             tls00_LongExecInfo   *pLongExecInfo,   //source
                             tls00_MultColSpec    *pMultCol,
                             bool                  bUnicodeOI,
                             tls00_Buffer2        *pConvertBuffer,
                             tsp00_Addr            pszErrText);

/*
  -----------------------------------------------------------------------------
  function:     ls27_GetLongLength
  -----------------------------------------------------------------------------

  description:  Gets length of all not null long values in one record

  attention:    

  arguments:    


  returnvalue:  
  -----------------------------------------------------------------------------
*/
static tsp00_Longint
ls27_GetLongLength(tls00_DBInfo      *pDBInfo,
                   tls00_DextractRec *pDextRec,
                   tls00_DatExtrCmd  *pDextCmd,
                   tsp00_Addr         pszErrText);


/*
  -----------------------------------------------------------------------------

  function:     ls27_ConvertTableData

  -----------------------------------------------------------------------------

  description:  

  attention:    

  arguments:    


  returnvalue:  
  -----------------------------------------------------------------------------
*/
static tsp00_Longint
ls27_ConvertTableData(tls00_DBInfo       *DBInfo,
                      tsp00_Int4          RowCount,
                      tls00_DextractRec  *DextRec,
                      tls00_DatExtrCmd   *DextCmd,
                      tls27_DataSink     *pDataSink,
                      tls00_Buffer2      *pConvertBuffer,   // convert buffer used to convert data between encodings
                      tsp00_Addr          pszErrText);

/*
  -----------------------------------------------------------------------------
  function:     ls27_SpecialConversion
  -----------------------------------------------------------------------------
*/
static tsp00_Longint
ls27_SpecialConversion(tls00_ColumnData    *pSingleColData,
                       bool                 bUnicodeOI,
                       tls00_Buffer2       *pConvertBuffer,
                       tsp00_Addr           pszErrText);

/*
  -----------------------------------------------------------------------------
  function:     ls27_AssignNILValue
  -----------------------------------------------------------------------------
*/
inline tsp00_Int4
ls27_AssignNILValue(tsp00_Addr       &pszDestData,      // Der pointer wird 'umgehaengt'
                    tls00_FieldAsgn  *pAssignment,
                    _TCHAR          *&pNILRepresentation,
                    tsp00_Int2        lNILRepresentationLen)
{
    tsp00_Int4 len = 0;

    if(0 == pAssignment->fasAsgnLen_ls00)
    {
        pszDestData  = pNILRepresentation;
        len          = lNILRepresentationLen;
    }
    else
    {
        pszDestData  = pAssignment->fasAsgnVal_ls00;
        len          = pAssignment->fasAsgnLen_ls00;
    }

    return len;
}
// ls27_AssignNILValue


/*
  -----------------------------------------------------------------------------
  EndChapter: Private functions declaration
  -----------------------------------------------------------------------------
*/



/*
  -----------------------------------------------------------------------------
  Chapter: Public functions implementation
  -----------------------------------------------------------------------------
*/

/*
  -----------------------------------------------------------------------------
  function:     ls27ExtractTableData
  -----------------------------------------------------------------------------
*/
tsp00_Int4
ls27ExtractTableData(tls00_DBInfo      *pDBInfo,
                     tls00_DextractRec *pDextRec,
                     tls00_DatExtrCmd  *pDextCmd,
                     tsp00_Addr        pszErrText)
{

    ROUTINE_DBG_MEO00 ("ls27ExtractTableData");

    tsp00_Longint rc = errOK_els00;

    //*
    //*     Define the max count of rows to fetch in a single packet
    //*
    // In case that a table contains longs we limit the count of rows to fetch to the
    // half of the maximum
    if (0 != pDextRec->derLongColCnt_ls00)
    {
        pDextRec->derMaxRowsPPckt_ls00 = pDextRec->derMaxRowsPPckt_ls00/2;
        if (0 == pDextRec->derMaxRowsPPckt_ls00)
        {
            pDextRec->derMaxRowsPPckt_ls00 = 1;
        }
    }
    tsp00_Int4 RowsFetched = pDextRec->derMaxRowsPPckt_ls00;

    tsp1_packet *SqlPacket = NULL;
    tsp1_part   *SqlPart   = NULL;

    tls00_DataSource *pDataSrc = &pDextRec->derDataSource_ls00;
    
    
    // Initialisiere die Datenstruktur. Der permanent gehaltene platz im puffer ist das
    // doppelte einer datenseite. Mehr daten koennen z.z. nicht in einer spalte der tabelle
    // gespeichert werden - dabei ist die hex-darstellung schon beachtet.
    tls27_DataSink *pDataSink = new tls27_DataSink;

    memset(pDataSink, 0, sizeof(tls27_DataSink));
    pDataSink->dskStruct_ls27.dasAllocMemPtr_ls27 = new char[pDBInfo->dbiPageSize_ls00 + pDBInfo->dbiPageSize_ls00];

    //*
    //*     Create buffer for line to write
    //*
    pDataSink->dskOutLine_ls27.lBufLen_ls00 = 4 * pDextCmd->dexDatFile_ls00.fsFormat_ls00.ffoBytesPerLine_ls00;

    if (true == pDextCmd->dexDatFile_ls00.fsFormat_ls00.ffoCompact_ls00)
    {
        pDataSink->dskMetaLine_ls27.lBufLen_ls00 = 
            sizeof(tsp00_Longint)*pDextRec->derLongColCnt_ls00 +
             sizeof(tsp00_Int2)*(pDextRec->derTableInfo_ls00.tdFieldCount_ls00 - pDextRec->derLongColCnt_ls00);

        pDataSink->dskOutLine_ls27.lBufLen_ls00 += pDataSink->dskMetaLine_ls27.lBufLen_ls00;
    }

    pDataSink->dskOutLine_ls27.pBufPtr_ls00 = new _TCHAR[pDataSink->dskOutLine_ls27.lBufLen_ls00];

    //*
    //*     Initialize file structure and write headers
    //*
    tls00_FileSpec *pDataFile = &pDextCmd->dexDatFile_ls00;
    if (true == pDataFile->fsFormat_ls00.ffoCompact_ls00)
    {
        ls30VFWriteMetaHeader(*pDextRec->derDataFile_ls00, pDextRec, pDextCmd, pszErrText );
    }
    else if (true == pDataFile->fsFormat_ls00.ffoBinary_ls00)
    {
        ls30VFWriteHeader(*pDextRec->derDataFile_ls00,
                          pDataFile->fsFormat_ls00.ffoBytesPerLine_ls00,
                          pDextCmd->dexFullUserInput_ls00,
                          pszErrText);
    }


    tls00_Buffer2 *pConvertBuffer = new tls00_Buffer2;
    pConvertBuffer->lBufLen_ls00 = 6*pDBInfo->dbiPageSize_ls00;
    pConvertBuffer->pBufPtr_ls00 = new _TCHAR[pConvertBuffer->lBufLen_ls00];
    pConvertBuffer->lBufPos_ls00 = 0;

    //*
    //*     Main loop to extract data
    //*
    while (RowsFetched == pDextRec->derMaxRowsPPckt_ls00)
    {
        rc = ls12Fetch(*pDBInfo, pDextRec->derRTableName_ls00, RowsFetched, true, pszErrText);
        if (errOK_els00 != rc)
        {
            if (dbRowNotFound_els00 == rc)
            {
                rc = errOK_els00;       // No more rows to process, but no error
            }
            break;
        }

        SqlPacket = pDBInfo->dbiPktRcvList_ls00[0];

        bool Res = ls04GetPart(SqlPacket, sp1pk_data, SqlPart);
        if (false == Res)
        {
            rc = errUnknownPart_els98;
            break;
        }


//TODO: partlength durch rowsFetched MUSS derRecordLen_ls00 ergeben, das koennte hier geprueft werden
        pDextRec->derDataSource_ls00.dasInputBuffer.strLeng_ls00 = ls04GetPartLength(SqlPart);
        ls04BufPtrFromPart(SqlPart, 0, pDextRec->derDataSource_ls00.dasInputBuffer.strAddr_ls00);


        rc = ls27_ConvertTableData(pDBInfo, RowsFetched, pDextRec, pDextCmd, pDataSink, pConvertBuffer, pszErrText);
        if (errOK_els00 != rc)
        {
            break;
        }

        // Check for cancel request
        rc = ls01CheckForCancel(pDBInfo->dbiSession_ls00.sipCancelByte_ls00);
        if (errOK_els00 != rc)
        {
            RepServMessages::Instance()->ls98Msg(layDataExtract_els98, pszErrText, errCancelRequest_els98);
            break;
        }
    }   // end while (RowsFetched == pDextRec->derMaxRowsPPckt_ls00)


    if (NULL != pDataSink->dskStruct_ls27.dasAllocMemPtr_ls27)
    {
        delete [] pDataSink->dskStruct_ls27.dasAllocMemPtr_ls27;
        pDataSink->dskStruct_ls27.dasAllocMemPtr_ls27 = NULL;
    }

    if (NULL != pDataSink->dskMetaLine_ls27.pBufPtr_ls00)
    {
        // simply mark the pointer as unused
        pDataSink->dskMetaLine_ls27.pBufPtr_ls00 = NULL;
    }

    if (NULL != pDataSink->dskOutLine_ls27.pBufPtr_ls00)
    {
        delete [] (STATIC_CAST(_TCHAR*, pDataSink->dskOutLine_ls27.pBufPtr_ls00));
        pDataSink->dskOutLine_ls27.pBufPtr_ls00 = NULL;
    }


    if (NULL != pDataSink)
    {
        delete pDataSink;
    }

    //*
    //*     Delete conversion buffer
    //*
    if (NULL != pConvertBuffer)
    {
        if (NULL != pConvertBuffer->pBufPtr_ls00)
        {
            delete [] STATIC_CAST(_TCHAR*, pConvertBuffer->pBufPtr_ls00);
            pConvertBuffer->pBufPtr_ls00 = NULL;
        }

        delete pConvertBuffer;
        pConvertBuffer = NULL;
    }


    return STATIC_CAST(tsp00_Int4, rc);
}
// ls27ExtractTableData()


/*
  -----------------------------------------------------------------------------
  function:     ls27BuildLoadCtrlFile
  -----------------------------------------------------------------------------
*/
tsp00_Int4
ls27BuildLoadCtrlFile(tls00_DBInfo      * pDBInfo,
                      tls00_DextractRec * pDextRec,
                      tls00_DatExtrCmd  * pDextCmd,
                      tsp00_Addr          pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls27BuildLoadCtrlFile");

    RepServMessages *pMsgObj = RepServMessages::Instance();    // error reporting obj
    tsp00_Int4       rc      = errOK_els00;
    tsp00_Int4       rcDDL   = errOK_els00;

    // If not processing DE for DATALOAD/FASTLOAD/DATAUPDATE we're done here
    if ( (resUnknown_els00 == pDextCmd->dexRestore_ls00) ||
         ( (true == pDextRec->derTabEmpty_ls00) && (pDextCmd->dexRestore_ls00 > resDataUpdt_els00) ) )
    {
        return rc;
    }   

    //*
    //*     Init and open command file
    //*
    tls00_VFile vCmdFile = *pDextRec->derCmdFile_ls00;

    //*
    //*     Generate 'create table' - stmt
    //*
    //* To generate the 'create table' and the 'create index' stmt we use an odbc-connection;
    //* therefore we have to release the already existing connection first and to reconnect
    //* using odbc; after work is done we connect the other way around
    //*
// HGB
    const SAPDB_Char *pszODBC_Conn_Err = NULL;
    sapdbwa_Err Err;

    CatalogDefODBC   *pCatalogDefODBC  = NULL;
    SAPDB_Bool        isDBUnicode      = SAPDB_TRUE ;

    StudioOAL_String  TabOwnr;
    StudioOAL_String  TabName;
    
    if (pDextCmd->dexRestore_ls00 < resDataUpdt_els00)
    {
        pCatalogDefODBC  = new CatalogDefODBC();  

        if ( ls00IsDBUnicode( pDBInfo ) == SAPDB_FALSE )            // odbc supports only unicode yet ( 22/03/01 )
        {
            isDBUnicode = SAPDB_FALSE;

            // close oi database connection
            ls15CommitUser(pDBInfo, pszErrText, true);

            // Build part spec
            tls00_PartSpec aPartSpec;
            memset(&aPartSpec, 0, sizeof(aPartSpec));
            aPartSpec.psSelUser_ls00 = false;
            aPartSpec.psSelAll_ls00  = false;

            aPartSpec.psTabName_ls00.rawAssign(pDextCmd->dexTable_ls00.tsTabName_ls00.asCharp());
            if (0 != pDextCmd->dexTable_ls00.tsTabOwnr_ls00.length ())
            {
                aPartSpec.psTabOwnr_ls00.rawAssign(pDextCmd->dexTable_ls00.tsTabOwnr_ls00.asCharp());
            }

            if (0 != pDextCmd->dexTable_ls00.tsTabOwnr_ls00.length())
            {
                TabOwnr.ConvertFromASCII_Latin1(pDextCmd->dexTable_ls00.tsTabOwnr_ls00.asCharp(),
                                                pDextCmd->dexTable_ls00.tsTabOwnr_ls00.asCharp() +
                                                 pDextCmd->dexTable_ls00.tsTabOwnr_ls00.length () );
            }

            TabName.ConvertFromASCII_Latin1(pDextCmd->dexTable_ls00.tsTabName_ls00.asCharp(),
                                            pDextCmd->dexTable_ls00.tsTabName_ls00.asCharp() +
                                             pDextCmd->dexTable_ls00.tsTabName_ls00.length () );

            pCatalogDefODBC->setPartSpec(aPartSpec);
            pCatalogDefODBC->setExtrMode(emExtractAsDDL_els31 );

            if (NULL != pCatalogDefODBC)
            {
                // establish odbc database connection
                pCatalogDefODBC->Connect(pDBInfo);
            }

            //Err  = pCatalogDefODBC->getODBCConnection().GetErr(); //!!ToDo Bug
        
            if (NULL == pszODBC_Conn_Err)
            {
                // reuse established file connection
                tls00_FileInfo CmdFileInfo;
                CmdFileInfo.vFile    = vCmdFile;
                CmdFileInfo.fileSpec = pDextCmd->dexCmdFile_ls00;

                pCatalogDefODBC->setFileConnection(CmdFileInfo);

                // execute command
                rcDDL = pCatalogDefODBC->tableExtrCmdExecDirect(TabOwnr, TabName);
            }
            else
            {
                pszODBC_Conn_Err = Err.GetErrMsg();
                pMsgObj->ls98PrintMessage(layDataExtract_els98, err_DE_GenerateCreateCmd_els98, pszODBC_Conn_Err);

                // This is an error but not heavy enough to stop executing; so we simply
                // print a msg to the log file
                rcDDL = err_DE_GenerateCreateCmd_els98;
            }

            vCmdFile = pCatalogDefODBC->getFileConnection().vFile;
        }   // end if ( ls00IsDBUnicode( pDBInfo ) == SAPDB_FALSE )
// HGB
    }   // end if (pDextCmd->dexRestore_ls00 < resDataUpdt_els00)


    //*
    //*     Generate dataload command - only if table is not empty
    //*
    if (true == pDextRec->derTabEmpty_ls00)
    {
        pMsgObj->ls98PrintMessage(layDataExtract_els98, msgNoCtrlCmdAtDataextract_els98);
    }
    else
    {
        //*
        //*     Do we have to write as first the separator line?
        //*
        //* In case the file is not created but contains already commands
        //* we need to write at first a separator line
        //*
        if ( (0 != vCmdFile.vfFileLen_ls00.lnuRemainder_ls00) ||
             (0 != vCmdFile.vfFileLen_ls00.lnu2GB_Blocks_ls00) )
        {
            char p[] = CMD_SEPARATOR_TOKEN_LS00;
            rc = ls30VFWriteLn(vCmdFile, p, CMD_SEPARATOR_TOKEN_LEN_LS00, pszErrText);
        }

        //*
        //*     Write dataload command
        //*
        tls00_MultColSpec *MultColumns    = &pDextCmd->dexOutSpec_ls00;  // Spaltenangaben zur datendatei
        tls00_FileFormat  *DataFileFormat = &pDextCmd->dexDatFile_ls00.fsFormat_ls00;

        tsp00_Int4 lPos       = 0;
        tsp00_Int4 lCmdLength = ((MultColumns->mcsCount_ls00 + 2) * (KNL_IDENTIFIER_SIZE + 50)) + 256;

        tsp00_Addr pszLoadCmd = new _TCHAR[lCmdLength];
        pszLoadCmd[0] = '\0';

        //*
        //*     FASTLOAD, DATALOAD or DATAUPDATE ?
        //*
        if ( (resFastload_els00 == pDextCmd->dexRestore_ls00) &&
             ( (0    != pDextRec->derLongColCnt_ls00)  ||
               (true == pDextCmd->dexDatFile_ls00.fsFormat_ls00.ffoCompact_ls00) ) )
        {
            // If in case of DATAEXTRACT FOR FASTLOAD the extracted table has LONGs or
            // the extract format is records FASTLOAD is replaced by DATALOAD
            pMsgObj->ls98PrintMessage(layDataExtract_els98, wrnDEforFLCreatesDLStmt_els98);

            lPos = sp77sprintf(pszLoadCmd, lCmdLength, pszLoadCmdBeginString[resDataload_els00]);
        }
        else
        {
            lPos = sp77sprintf(pszLoadCmd, lCmdLength, pszLoadCmdBeginString[STATIC_CAST(int, pDextCmd->dexRestore_ls00)]);
        }

        //*
        //*     Table name
        //*
        lPos += sp77sprintf(pszLoadCmd + lPos, lCmdLength - lPos, "\"%s\""NEW_LINE_LS00"", pDextCmd->dexTable_ls00.tsTabName_ls00.asCharp());

        //*
        //*     Column descriptions
        //*
        tsp00_Int4 i = 0;
        
        tls00_ParamInfo        *pParamInfo = NULL;
        tls00_Column           *pMyColumn  = NULL;
        tsp00_Addr              pszColName = NULL;
        tls00_TableDescription *pTInfo     = &pDextRec->derTableInfo_ls00;
    
        for (i; i < MultColumns->mcsCount_ls00; ++i)
        {
            pszColName = (*pTInfo->tdColName_ls00[i]).asCharp();
            pParamInfo = pTInfo->tdParamArray_ls00[i];
            pMyColumn  = MultColumns->mcsColumn_ls00[i];

            if (true == pMyColumn->colIsLit_ls00)       // Special processing of literals
            {
                continue;
            }

            if (true == DataFileFormat->ffoFormatted_ls00)
            {
                lPos += sp77sprintf(pszLoadCmd + lPos, lCmdLength - lPos, _T("\"%s\" %d-%d CHAR"NEW_LINE_LS00""),
                                                                          pszColName,
                                                                          pMyColumn->colFPos_ls00.fpoStart_ls00,
                                                                          pMyColumn->colFPos_ls00.fpoEnd_ls00);
            }
            else
            {
                lPos += sp77sprintf(pszLoadCmd + lPos, lCmdLength - lPos, _T("\"%s\" %d CHAR"NEW_LINE_LS00""),
                                                                          pszColName,
                                                                          pMyColumn->colFPos_ls00.fpoStart_ls00);
            }

            if ( (false == pParamInfo->piMode_ls00.includes(sp1ot_mandatory)) ||
                 (true == pParamInfo->piMode_ls00.includes(sp1ot_default)) )
            {
                lPos += sp77sprintf(pszLoadCmd + (lPos - NEW_LINE_LEN_LS00), lCmdLength - lPos + NEW_LINE_LEN_LS00,
                                                                             _T(" DEFAULT NULL"NEW_LINE_LS00"")) - NEW_LINE_LEN_LS00;
            }
        }

        //*
        //*     Infile name
        //*
        lPos += sp77sprintf(pszLoadCmd + lPos, lCmdLength - lPos, _T("INFILE '%s' "),
                                                                  pDextCmd->dexDatFile_ls00.fsFileName_ls00);

        //*
        //*     Infile format and encoding type
        //*
        if (true == DataFileFormat->ffoCompact_ls00)
        {
            lPos += sp77sprintf(pszLoadCmd + lPos, lCmdLength - lPos, _T(" COMPACT "));
        }
        else if (true == DataFileFormat->ffoFormatted_ls00)
        {
            if (false == DataFileFormat->ffoBinary_ls00)
            {
                lPos += sp77sprintf(pszLoadCmd + lPos, lCmdLength - lPos, _T(" FORMATTED "));
            }
            else
            {
                lPos += sp77sprintf(pszLoadCmd + lPos, lCmdLength - lPos, _T(" FORMATTED BINARY"NEW_LINE_LS00""));
            }
        }
        else
        {
            lPos += sp77sprintf(pszLoadCmd + lPos, lCmdLength - lPos, _T(" COMPRESSED "));
        }

        //*     encoding (not if FORMATTED BINARY)
        if ( (false == DataFileFormat->ffoFormatted_ls00) ||
             ( (true == DataFileFormat->ffoFormatted_ls00) && (false == DataFileFormat->ffoBinary_ls00) ) )
        {
            lPos += sp77sprintf(pszLoadCmd + lPos, lCmdLength - lPos, _T("%s"NEW_LINE_LS00""),
                                                                      pszEncoding[DataFileFormat->ffoCodeType_ls00]);
        }


        //*
        //*     Decimal symbol
        //*
        if (NO_THSD_GROUPS_LS00 == DataFileFormat->ffoThsdChar_ls00)
        {
            lPos += sp77sprintf(pszLoadCmd + lPos, lCmdLength - lPos, _T("DECIMAL '//%c/'"NEW_LINE_LS00""),
                                                                      DataFileFormat->ffoDecPoint_ls00);
        }
        else
        {
            lPos += sp77sprintf(pszLoadCmd + lPos, lCmdLength - lPos, _T("DECIMAL '/%c/%c/'"NEW_LINE_LS00""),
                                                                      DataFileFormat->ffoThsdChar_ls00,
                                                                      DataFileFormat->ffoDecPoint_ls00);
        }

        //*
        //*     String separator (if it is not the default=comma)
        //*
        if (COMMA_LS00 != DataFileFormat->ffoSeperat_ls00[0])
        {
            lPos += sp77sprintf(pszLoadCmd + lPos, lCmdLength - lPos, _T("SEPARATOR '%c'"NEW_LINE_LS00""),
                                                                      DataFileFormat->ffoSeperat_ls00[0]);
        }

        //*
        //*     String delimiter (if it is not the default=double quotes)
        //*
        if (QUOTE_LS00 != DataFileFormat->ffoDelimit_ls00[0])
        {
            lPos += sp77sprintf(pszLoadCmd + lPos, lCmdLength - lPos, _T("DELIMITER '%c'"NEW_LINE_LS00""),
                                                                      DataFileFormat->ffoDelimit_ls00[0]);
        }

        //*
        //*     DATE format
        //*
        if (DataFileFormat->ffoDateType_ls00 == dtFreeMask_els00)
        {
            lPos += sp77sprintf(pszLoadCmd + lPos, lCmdLength - lPos, _T("DATE '%s'"NEW_LINE_LS00""),
                                                                      DataFileFormat->ffoDateMask_ls00);
        }
        else
        {
            lPos += sp77sprintf(pszLoadCmd + lPos, lCmdLength - lPos,
                                                   _T("DATE %s"NEW_LINE_LS00""),
                                                   pszDateTimeType[STATIC_CAST(int, DataFileFormat->ffoDateType_ls00)]);
        }

        //*
        //*     TIME format
        //*
        if (DataFileFormat->ffoTimeType_ls00 == dtFreeMask_els00)
        {
            lPos += sp77sprintf(pszLoadCmd + lPos, lCmdLength - lPos, _T("TIME '%s'"NEW_LINE_LS00""),
                                                                      DataFileFormat->ffoTimeMask_ls00);
        }
        else
        {
            lPos += sp77sprintf(pszLoadCmd + lPos, lCmdLength - lPos,
                                                   _T("TIME %s"NEW_LINE_LS00""),
                                                   pszDateTimeType[STATIC_CAST(int, DataFileFormat->ffoTimeType_ls00)]);
        }


        //*
        //*     TIMESTAMP format
        //*
        if (DataFileFormat->ffoTimeStampType_ls00 == dtFreeMask_els00)
        {
            lPos += sp77sprintf(pszLoadCmd + lPos, lCmdLength - lPos, _T("TIMESTAMP '%s'"NEW_LINE_LS00""),
                                                                      DataFileFormat->ffoTimeStampMask_ls00);
        }
        else
        {
            lPos += sp77sprintf(pszLoadCmd + lPos, lCmdLength - lPos,
                                                   _T("TIMESTAMP %s"NEW_LINE_LS00""),
                                                   pszDateTimeType[STATIC_CAST(int, DataFileFormat->ffoTimeStampType_ls00)]);
        }


        //*
        //*     NULL value representation
        //*
        lPos += sp77sprintf(pszLoadCmd + lPos, lCmdLength - lPos, _T("NULL '%s'"NEW_LINE_LS00""), DataFileFormat->ffoNullSpec_ls00);

        //*
        //*     boolean value representation
        //*
        lPos += sp77sprintf(pszLoadCmd + lPos, lCmdLength - lPos, _T("BOOLEAN '%s/%s'"NEW_LINE_LS00""),
                                                                  DataFileFormat->ffoTrueVal_ls00,
                                                                  DataFileFormat->ffoFalseVal_ls00);
        //*
        //*     long  columns?
        //*
        if ( (0 != pDextRec->derLongColCnt_ls00) && (false == pDextCmd->dexDatFile_ls00.fsFormat_ls00.ffoCompact_ls00) )
        {
            for (i = 0; i < MultColumns->mcsCount_ls00; ++i)
            {
                pMyColumn = MultColumns->mcsColumn_ls00[i];

                if (1 == szMap_LongDataType_ls00[pDextRec->pderColumnData_ls00[i].cdSQLDataType_ls00])
                {
                    if (false == pMyColumn->colLongInfo.loibFileNumbered_ls00)
                    {   // long file not numbered
                        lPos += sp77sprintf(pszLoadCmd + lPos, lCmdLength - lPos,
                                                               _T("LONGFILE \"%s\" '%s' %s"NEW_LINE_LS00""),
                                                               pMyColumn->colName_ls00.asCharp(),
                                                               pMyColumn->colLongInfo.loiFileName_ls00,
                                                               pszEncoding[pMyColumn->colLongInfo.loiFileEncoding_ls00]);
                    }
                }
            }
        }

        // Remove last carriage return-line feed
        lPos -= 1;
        while (iscntrl(pszLoadCmd[lPos]))
        {
            pszLoadCmd[lPos] = 0;
            --lPos;
        }
        lPos += 1;

        rc = ls30VFWriteLn(vCmdFile, pszLoadCmd, lPos, pszErrText);

        if (NULL != pszLoadCmd)
        {
            delete [] pszLoadCmd;
            pszLoadCmd = NULL;
        }
    }
        
    //*
    //*     Generate 'create index' - stmts (only if connected via odbc)
    //*
// HGB
    if (pDextCmd->dexRestore_ls00 < resDataUpdt_els00)
    {
        if ( isDBUnicode == SAPDB_FALSE )
        {
            if (errOK_els00 == rcDDL)
            {
                // TabName and TabOwner are defined when extracting table information
                pCatalogDefODBC->indexExtrCmdExecDirect(TabOwnr, TabName);
            }
        
            //*
            //*     Close ODBC database connection (opened when extracting table information)
            //*
            if (NULL == pszODBC_Conn_Err)
            {
                pCatalogDefODBC->Disconnect();
            }

            //*
            //*     Reestablish OrderInterface database connection
            //*
            rc = ls03SQLConnect(pDBInfo, pszErrText);
            rc = ls15ConnectUser(pDBInfo, pszErrText);
        }
    }
    if (NULL != pCatalogDefODBC)
    {
         delete pCatalogDefODBC;
         pCatalogDefODBC = NULL;
    }

// HGB

    if ( (errOK_els00 == rc) && (errOK_els00 != rcDDL) )
    {
        char szFileName[MAX_REPLY_LENGTH_LS00];
        szFileName[0] = 0;
        pMsgObj->ls98GetLogFileName(szFileName);

        pMsgObj->ls98MsgToString(layDataExtract_els98, pszErrText, errCmdUnsuccessful_els98, szFileName);
        rc = errCmdUnsuccessful_els98;
    }

    return rc;
}
// ls27BuildLoadCtrlFile()

/*
  -----------------------------------------------------------------------------
  EndChapter: Public functions implementation
  -----------------------------------------------------------------------------
*/

//////////////////////////////////////////////////////////////////////////////////////////////
/*
  -----------------------------------------------------------------------------
  Chapter: Private functions implementation
  -----------------------------------------------------------------------------
*/

/*
  -----------------------------------------------------------------------------
  function:     ls27_GetLongLength
  -----------------------------------------------------------------------------
*/
//Hier tls00_ColumnData mit dazunehmen und die laenge eintragen fuer compact
//und wert eintragen fuer compressed/formatted
static tsp00_Longint
ls27_GetLongLength(tls00_DBInfo      *pDBInfo,
                   tls00_DextractRec *pDextRec,
                   tls00_DatExtrCmd  *pDextCmd,
                   tsp00_Addr         pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls27_GetLongLength");

    //*
    //*     This module is only called for data stream format RECORDS (COMPACT)
    //*     Coming to this point no long value of the extracted database record
    //*     has been processed yet.
    //*
    RepServMessages *pMsgObj = RepServMessages::Instance();    // Error reporting object
    tsp00_Longint    rc      = errOK_els00;


    tls00_LongExecInfo   *pLongInfoStruct  = &pDextRec->derLongInfo_ls00;
    tsp00_Int4            lLongCnt         = pLongInfoStruct->dliCntNNLongs_ls00; // Count of longs that must be processed
    tsp00_LongDescriptor *pLongD           = NULL;
    tls00_Column         *pCol             = NULL;
    tls00_ColumnData     *pColData         = NULL;  // Contains data and meta data for a certain column in record

    tsp00_Int4            lCompleteLongCnt = 0;     // Holds the number of longs completely included in the packet
    tsp00_Int4            j                = 0;

    //*
    //*     Check wich LONGs did fit completely
    //* for those the descriptor shows values vm_datapart, vm_alldata or vm_lastdata for ld_valmode
    //*
    for (lCompleteLongCnt = 0;
         ( (lCompleteLongCnt < lLongCnt) &&
           (pDextRec->derLongDescArray_ls00[lCompleteLongCnt].ld_valmode() < vm_nodata) );
         ++lCompleteLongCnt)
             ;
    
    //*
    //*     If there are incomplete delivered LONGs we get their length here separately
    //*
    if (lCompleteLongCnt < lLongCnt)
    {
        tsp1_packet  *SqlSentPckt     = pDBInfo->dbiPktSndList_ls00[1];;
        tsp1_packet  *SqlReceivedPckt = NULL;
        tsp1_segment *SqlSegm         = NULL;
        tsp1_part    *SqlPart         = NULL;

        lLongCnt -= lCompleteLongCnt;

        //*     Prepare getval-packet to get the LONG lengths
        rc = ls04FastInit(SqlSentPckt, &SqlSegm, &SqlPart, pDBInfo); 
        ls04SetMsgType(SqlSegm, sp1m_getval);                        
        SqlPart->sp1p_part_kind().becomes(sp1pk_longdata);

        // Compute maximum number of descriptors fitting into one packet
        tsp00_Int4 lMaxLongDescInPckt = SqlPart->sp1p_buf_size()/(LongDescSize_ls00 + 1);

        // Limit number of LONG descriptors if necessary
        lLongCnt = ( lLongCnt >= lMaxLongDescInPckt) ? lMaxLongDescInPckt : lLongCnt;

        // Write LONG descriptors into packet to send to kernel
        for (j = lCompleteLongCnt; ( (j < (lLongCnt + lCompleteLongCnt)) && (errOK_els00 == rc) ); ++j)
        {
            pLongD = &pDextRec->derLongDescArray_ls00[j];
            pLongD->ld_vallen() = 2;
            pLongD->ld_infoset().addElement(ld_no_close);

            rc = ls04CharToPart(SqlPart, 0);                                    // undef byte
            rc = ls04BufToPart(SqlPart, REINTERPRET_CAST(char*, pLongD), LongDescSize_ls00);
        }

        SqlPart->sp1p_arg_count() = lLongCnt;
        if (errOK_els00 == rc)
        {
            rc = ls04FinishPart(SqlSentPckt, SqlPart);
        }

        //*
        //*     Send descriptors to kernel and get new descriptors which contain the lengths
        //*
        if (errOK_els00 == rc)
        {
            tsp00_Int4 lErrPos = 0;
            rc = ls03ProcessSQLPacket(pDBInfo, 1, lErrPos, pszErrText);        
            if (errOK_els00 == rc)
            {
                //  Get descriptors
                rc = ls04GetLONGDescriptor(pDBInfo->dbiPktRcvList_ls00[1],
                                           lLongCnt,
                                           pDextRec->derLongDescArray_ls00,
                                           lCompleteLongCnt);
                if (errOK_els00 != rc)
                {
                    RepServMessages::Instance()->ls98Msg(layDataExtract_els98, pszErrText,
                                                                               STATIC_CAST(ErrorNumbers, rc));
                }
            }
        }

        //* next code extract is only because the kernel had a bug and did not deliver
        //* right valmode if only 2 bytes were requested
        if (errOK_els00 == rc)
        {
            for (j=lCompleteLongCnt; j < pLongInfoStruct->dliCntNNLongs_ls00; ++j)
            {
                // This is because of a kernel bug
                if (vm_data_trunc != pDextRec->derLongDescArray_ls00[j].ld_valmode())
                {
                    pDextRec->derLongDescArray_ls00[j].ld_valmode().becomes(vm_data_trunc);
                }
            }
        }
    }   // end if (lCompleteLongCnt < lLongCntInPckt)


    //*
    //*     Now define the LONG lengths of all LONGs of the record extracted
    //*
    for (tsp00_Int4 k = 0; k < pLongInfoStruct->dliCntNNLongs_ls00; ++k)
    {
        pLongD = &pDextRec->derLongDescArray_ls00[k];

        pCol     = pDextCmd->dexOutSpec_ls00.mcsColumn_ls00[pLongInfoStruct->dliNNLongs_ls00[k]];
        pColData = &pDextRec->pderColumnData_ls00[pLongInfoStruct->dliNNLongs_ls00[k]];
        
        pCol->colLongInfo.loiLength_ls00 = pLongD->ld_maxlen();

        // Suppose the database encoding type is UNICODE; then all requests of the Loader to the
        // database are send as UCS2 requests and all data is delivered in UCS2
        // Suppose further the LONGs processed have code type ascii (which is allowed! -
        // for instance: create table <table name> (col1 long ascii) ); then
        // the kernel delivers the ascii length of the LONGs but the LONG data in UCS2 -->
        // thus we have to double the length in this case
        if ( ( (dstra == pColData->cdSQLDataType_ls00) || (dlonga == pColData->cdSQLDataType_ls00) ) &&
             ( pDextCmd->dexDatFile_ls00.fsFormat_ls00.ffoCodeType_ls00 > ctUCS2_els00)               )
        {
            pCol->colLongInfo.loiLength_ls00 *= 2;
        }

        // For all incomplete sent LONGs we have to reinitialize intern position and value length
        if (vm_data_trunc == pLongD->ld_valmode())
        {
            pLongD->ld_intern_pos() = 1;
            pLongD->ld_vallen()     = 0;
        }
    }

    return rc;
}
// ls27_GetLongLength()



/*
  -----------------------------------------------------------------------------
  function:     ls27_EditLongOffsetForOutput
  -----------------------------------------------------------------------------
*/
static tsp00_Longint
ls27_EditLongOffsetForOutput(tls00_ColumnData     *pColumnData,     //destination
                             tls00_LongExecInfo   *pLongExecInfo,   //source
                             tls00_MultColSpec    *pMultCol,
                             bool                  bUnicodeOI,
                             tls00_Buffer2        *pConvertBuffer,
                             tsp00_Addr            pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls27_EditLongOffsetForOutput");

    tsp00_Longint rc = errOK_els00;

    
    tls00_ParamInfo_LONG *pColLongInfo = NULL;

    for (tsp00_Int4 i = 0, j = 0; ( (errOK_els00 == rc) && (i < pLongExecInfo->dliFinishedNNLongs_ls00) ); ++i)
    {
        j = pLongExecInfo->dliNNLongs_ls00[i];
        pColLongInfo = &pMultCol->mcsColumn_ls00[j]->colLongInfo;
        
        if (true == pColLongInfo->loibFileNumbered_ls00)
        {
            pColumnData[j].cdPtrToData_ls00 = pColLongInfo->loiFileName_ls00;
            pColumnData[j].cdDataLen_ls00   = STATIC_CAST(tsp00_Int4, _tcslen(pColLongInfo->loiFileName_ls00));
        }
        else
        {
            tsp00_Longint spos = pColLongInfo->loiFileStruct_ls00->vfFilePos_ls00.lnuRemainder_ls00 -
                                    pColumnData[j].cdLONGLen_ls00 + 1;

            // At this point we 'initialize' the data buffer every time we come here to be
            // sure the buffer contains afterwards the things we really want
            pColumnData[j].cdDataBuffer_ls00[0] = 0;

            sp77sprintf(pColumnData[j].cdDataBuffer_ls00, STATIC_CAST(tsp00_Int4, pColumnData[j].cdDataBufferLen_ls00),
                        "%ld-%ld", spos, (pColLongInfo->loiFileStruct_ls00->vfFilePos_ls00.lnuRemainder_ls00));
            pColumnData[j].cdDataLen_ls00 = STATIC_CAST(tsp00_Int4, _tcslen(pColumnData[j].cdDataBuffer_ls00) );

            pColumnData[j].cdPtrUsed_ls00 = false;
        }


        rc = ls27_SpecialConversion(&pColumnData[j], bUnicodeOI, pConvertBuffer, pszErrText);

    }   // end for (tsp00_Int4 i = 0, j = 0; i < pLongExecInfo->dliFinishedNNLongs_ls00; ++i)


    return rc;
}
// ls27_EditLongOffsetForOutput()


/*
  -----------------------------------------------------------------------------
  function:     ls27_SpecialConversion
  -----------------------------------------------------------------------------
*/
static tsp00_Longint
ls27_SpecialConversion(tls00_ColumnData    *pSingleColData,
                       bool                 bUnicodeOI,
                       tls00_Buffer2       *pConvertBuffer,
                       tsp00_Addr           pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls27_SpecialConversion");

    tsp00_Longint rc = errOK_els00;

    tsp00_DataType DataType = pSingleColData->cdSQLDataType_ls00;
    //*
    //*     No special conversion for binary data
    //*
    if (ioChar_els00 != pSingleColData->cdExtDataType_ls00)
    {
        return rc;
    }


    tsp00_Addr pDataPtr = pSingleColData->cdPtrToData_ls00;
    if (false == pSingleColData->cdPtrUsed_ls00)
    {
        pDataPtr = pSingleColData->cdDataBuffer_ls00;
    }

    // in case of unicode order interface (oi) (we use unicode oi only if the database
    // is unicode defined itself) char data is stored as unicode data. If the data will be
    // written to file also as unicode we don't transform it here to any other encoding. Therefor
    // the trim function must handle unicode, too.
    
    //*
    //*     Trim character string right; function adapts only the length
    //*
    //* Define if it is necessary to trim unicode characters or only ascii characters
    tls00_CodeType CodeType = ctAscii_els00;
    if ( (true == bUnicodeOI) && (1 == szMap_CharDataType_ls00[DataType]) )
    {
        CodeType = ls07Swapped(NULL) ? ctUCS2Swapped_els00 : ctUCS2_els00;
    }
    char szTrim = ls07GetFillByte(pSingleColData->cdSQLDataType_ls00);
    ls05StrTrimRight(pDataPtr, pSingleColData->cdDataLen_ls00, szTrim, CodeType);

    //*
    //*     Trim numerical strings left, too
    //*
    //* In case of numerical values is always the buffer of tls00_ColumnData structure used
    if (1 == szMap_NumberDataType_ls00[DataType])
    {
        ls05StrTrimLeftEx(pDataPtr, pSingleColData->cdDataLen_ls00);
    }


    // fuer blobs ist immer das flag der hexdarstellung gesetzt! deshalb hier keine extrabehandlung

    // jetzt muss noch eine umwandlung in die hexadezimale und/oder unicode darstellung erfolgen,
    // da das noch nicht fuer alle datentypen gemacht wurde: speziell nicht fuer zeitstempel..., zahlen
    // und boolsche werte sowie fuer literale
    
    //*
    //*     Special transformation to right external encoding
    //*
    //* The used conversion buffer is large enough to hold the data
    //*
    tsp00_Longint  destlen   = 0;
    tsp00_Longuint lDestlenU = 0;

    // Conversion is ONLY needed (in any case) if data is not of blob type;
    // some blob type handling is done in the next transformation step
    if (0 == szMap_BlobDataType_ls00[DataType])
    {
        switch (pSingleColData->cdExtEncoding_ls00)
        {
            case ctAscii_els00:
            {
                // Conversion needed if
                // - kernel is UCS2 and
                //   - data type is of type char and
                //   - data is not NULL (NULL value representation is in ASCII)
                if ( (true == bUnicodeOI) && 
                     (1 == szMap_CharDataType_ls00[DataType]) && (false == pSingleColData->cdIsNILValue_ls00))
                {
                    destlen = pSingleColData->cdDataBufferLen_ls00;

                    memcpy(pConvertBuffer->pBufPtr_ls00, pDataPtr, pSingleColData->cdDataLen_ls00);

                    rc  = ls05ConvertFromUCS2Simple(REINTERPRET_CAST(unsigned char*, pSingleColData->cdDataBuffer_ls00),
                                                    STATIC_CAST(unsigned char*, pConvertBuffer->pBufPtr_ls00),
                                                    destlen,
                                                    pSingleColData->cdDataLen_ls00,
                                                    ctAscii_els00,
                                                    pszErrText);
                    if (errOK_els00 == rc)
                    {
                        pSingleColData->cdDataLen_ls00 = STATIC_CAST(tsp00_Longint, destlen);

                        pDataPtr                       = pSingleColData->cdDataBuffer_ls00;
                        pSingleColData->cdPtrUsed_ls00 = false;
                    }
                }
                break;

            }
            case ctUCS2_els00:              // ATTENTION: double fall through
            case ctUCS2Swapped_els00:
            {
                // Conversion needed if 
                // - the database is ASCII or
                // - the database is UCS2 and 
                //    - the data is not of type char or
                //    - the data is NULL (NULL value representation is in ASCII)
                bool bConversionNecessary = true;
                if (true == bUnicodeOI)
                {
                    if ( (1 == szMap_CharDataType_ls00[DataType]) && (false == pSingleColData->cdIsNILValue_ls00) )
                    {
                        bConversionNecessary = false;
                    }
                }

                if (true == bConversionNecessary)
                {
                    // transform
                    memcpy(pConvertBuffer->pBufPtr_ls00, pDataPtr, pSingleColData->cdDataLen_ls00);

                    lDestlenU = pSingleColData->cdDataBufferLen_ls00;

                    rc = ls05ConvertToUCS2Simple(REINTERPRET_CAST(unsigned char*, pSingleColData->cdDataBuffer_ls00),
                                                 lDestlenU,
                                                 STATIC_CAST(unsigned char*, pConvertBuffer->pBufPtr_ls00),
                                                 pSingleColData->cdDataLen_ls00,
                                                 ctAscii_els00,
                                                 pszErrText);
                    if (errOK_els00 == rc)
                    {
                        pSingleColData->cdDataLen_ls00 = STATIC_CAST(tsp00_Longint, lDestlenU);

                        pDataPtr                       = pSingleColData->cdDataBuffer_ls00;
                        pSingleColData->cdPtrUsed_ls00 = false;
                    }
                }
                break;
            }
            case ctUTF8_els00:
            {
                tls00_CodeType CodeType = ctAscii_els00;

                // Char data is held up to this point in the code type of the database order interface;
                // All other data (date, time, timestamp, numerical, boolean) has to be converted from ASCII
                //  because it is processed by the repman as ASCII data;
                // Null value representation is still handled as ASCII data and must be converted here
                if (true == bUnicodeOI)
                {
                    CodeType = ctUCS2Native_els00;
                    if ( (true == pSingleColData->cdIsNILValue_ls00) || (0 == szMap_CharDataType_ls00[DataType]) )
                    {
                        CodeType = ctAscii_els00;
                    }
                }

                SAPDB_DynamicUTF8String ConvertedData;
                memcpy(pConvertBuffer->pBufPtr_ls00, pDataPtr, pSingleColData->cdDataLen_ls00);

                rc = ls05ConvertToUTF8(ConvertedData,
                                       STATIC_CAST(unsigned char*, pConvertBuffer->pBufPtr_ls00),
                                       pSingleColData->cdDataLen_ls00,
                                       CodeType,
                                       pszErrText);
                if (errOK_els00 == rc)
                {
                    pSingleColData->cdDataLen_ls00 = STATIC_CAST(tsp00_Longint, ConvertedData.BasisSize());
                    memcpy(pSingleColData->cdDataBuffer_ls00, ConvertedData.StrPtr(), pSingleColData->cdDataLen_ls00);

                    pDataPtr                       = pSingleColData->cdDataBuffer_ls00;
                    pSingleColData->cdPtrUsed_ls00 = false;
                }
                break;
            }
        }   // end switch (pSingleColData->cdExtEncoding_ls00)
    }   // end if (0 == szMap_BlobDataType_ls00[DataType])

    //*
    //*     Conversion to hex representation 
    //*
    if ( (errOK_els00 == rc) && (true == pSingleColData->cdExtHexRepresentation_ls00) )
    {
        //*
        //*     First conversion from byte value to hex representation
        //*
        memcpy(pConvertBuffer->pBufPtr_ls00, pDataPtr, pSingleColData->cdDataLen_ls00);

        ls07ConvertByteToHexEx(STATIC_CAST(char*, pConvertBuffer->pBufPtr_ls00),
                               STATIC_CAST(tsp00_Int4, pSingleColData->cdDataLen_ls00),
                               pSingleColData->cdDataBuffer_ls00);

        pSingleColData->cdDataLen_ls00 = 2*pSingleColData->cdDataLen_ls00;

        //*
        //*     Second conversion
        //* In case of external encoding type UCS2 further conversion of the hex
        //* representation to UCS2 hex is necessary;
        //* In case of external encoding type UTF8 no further conversion needed,
        //*  readable hex representation may only be US7ASCII which is already UTF8
        //*
        if ( (ctUCS2_els00        == pSingleColData->cdExtEncoding_ls00) ||
             (ctUCS2Swapped_els00 == pSingleColData->cdExtEncoding_ls00)  )
        {
            memcpy(pConvertBuffer->pBufPtr_ls00, pSingleColData->cdDataBuffer_ls00, pSingleColData->cdDataLen_ls00);
            
            lDestlenU = pSingleColData->cdDataBufferLen_ls00;
            
            rc = ls05ConvertToUCS2Simple(REINTERPRET_CAST(unsigned char*, pSingleColData->cdDataBuffer_ls00),
                                         lDestlenU,
                                         STATIC_CAST(unsigned char*, pConvertBuffer->pBufPtr_ls00),
                                         pSingleColData->cdDataLen_ls00,
                                         ctAscii_els00,
                                         pszErrText);
            if (errOK_els00 == rc)
            {
                pSingleColData->cdDataLen_ls00 = STATIC_CAST(tsp00_Longint, lDestlenU);
            }
        }

        pSingleColData->cdPtrUsed_ls00 = false;     // Reset flag of pointer use
    }

    return rc;
}
// ls27_SpecialConversion()


// liefert z.z. IMMER errOK_els00 zurueck. Aufgetretene fehler sieht man an einer aenderung des
// wertes DextRec->derErrNumInRecord_ls00. Der letzte aufgetretene Fehler ist noch im fehlertext
// pszErrText enthalten!!!
/*
  -----------------------------------------------------------------------------
  function:     ls27_TransformData
  -----------------------------------------------------------------------------
*/
static tsp00_Longint
ls27_TransformData(tls00_DBInfo      *DBInfo,
                   tls00_DextractRec *DextRec,
                   tls00_DatExtrCmd  *DextCmd,
                   tls27_DataSink    *pDataSink,
                   tls00_Buffer2     *pConvertBuffer,
                   tsp00_Addr         pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls27_TransformData");

    tsp00_Longint rc = errOK_els00;

    //*
    //*     Initialize LONG descriptions for a single record
    //*
    tsp00_LongDescriptor *pLongDescArray = NULL;
    if (0 != DextRec->derLongColCnt_ls00)
    {
        DextRec->derLongInfo_ls00.dliCntNNLongs_ls00      = 0;
        DextRec->derLongInfo_ls00.dliFinishedNNLongs_ls00 = 0;
    }

    tls00_MultColSpec *MultColumns    = &DextCmd->dexOutSpec_ls00;  // Spaltenangaben zur datendatei
    tls00_FileFormat  *DataFileFormat = &DextCmd->dexDatFile_ls00.fsFormat_ls00;
    tls00_ColumnData  *pColumnData    = DextRec->pderColumnData_ls00;

    tls00_ParamInfo *pParamDesc  = NULL;        // parameter description structure (from db)
    tls00_Column    *pColumnDesc = NULL;        // column description structure    (from user)
    bool             bFirstLong  = true;
    bool             bUnicodeOI  = ls00IsDBUnicode(DBInfo);

    // Bei dieser schleife ist mehreres zu beachten. Zum einen existiert fuer jeden index in
    // multcol ein entsprechender index in DextRec->derMapIndex_ls00, der den multcol-index auf
    // den entsprechenden index von tableinfo abbildet. D.h. wir arbeiten bei TableInfo nicht
    // mit dem laufindex der schleife, sondern mit dem mapindex[von laufindex].
    // Ausserdem haben wir nur soviele pRawData-variablen angelegt, wie tableinfo member (tdFieldCount)
    // hat. Da aber in der spaltenausgabeliste auch literale eingestreut sein koennen, die
    // nicht in Tableinfo aufgenommen werden, muessen wir den index von TableInfo immer um die
    // bereits gefundene anzahl der Literale bereinigen (deshalb pRawData[i-j]).

    // Als loopcount kann der zaehler in MultColumns gewaehlt werden, da diese struktur
    // IMMER gefuellt wird, auch wenn keine spaltenliste angegeben wurde im kommando.

    // l ist der laufindex fuer longdescriptoren

    for (tsp00_Int4 i = 0, l = 0; i < MultColumns->mcsCount_ls00; ++i)
    {
        pColumnDesc  = MultColumns->mcsColumn_ls00[i];
        pParamDesc = DextRec->derTableInfo_ls00.tdParamArray_ls00[pColumnData[i].cdPosInStmt_ls00];


        // first check for literal: if it is a literal simple assign the pointer to
        // the column name which is in this case the literal
        if (true == pColumnDesc->colIsLit_ls00)
        {
            pColumnData[i].cdPtrToData_ls00 = pColumnDesc->colName_ls00.asCharp();
            pColumnData[i].cdDataLen_ls00   = pColumnDesc->colName_ls00.length();
        }
        else
        {
            if (INSERT_NULL == pColumnData[i].cdDataLen_ls00)
            {
                if(0 == pColumnDesc->colAssign_ls00.fasAsgnLen_ls00)
                {
                    pColumnData[i].cdPtrToData_ls00 = DataFileFormat->ffoNullSpec_ls00;
                    pColumnData[i].cdDataLen_ls00   = DataFileFormat->ffoNSpecLen_ls00;
                }
                else
                {
                    pColumnData[i].cdPtrToData_ls00  = pColumnDesc->colAssign_ls00.fasAsgnVal_ls00;
                    pColumnData[i].cdDataLen_ls00    = pColumnDesc->colAssign_ls00.fasAsgnLen_ls00;
                }
                pColumnData[i].cdIsNILValue_ls00 = true;
                pColumnData[i].cdPtrUsed_ls00    = true;
            }
            else
            {
                if (1 == szMap_NumberDataType_ls00[pColumnData[i].cdSQLDataType_ls00])
                {
                    //*
                    //*     Numerical data type
                    //*
                    rc = ls25ExtractNumbers(pColumnData[i].cdPtrToData_ls00,
                                            pColumnData[i].cdDataBuffer_ls00,
                                            pColumnData[i].cdDataLen_ls00,
                                            pColumnDesc,
                                            &pParamDesc->piBasicInfo_ls00,
                                            DataFileFormat,
                                            DextRec->derSwapKind_ls00,
                                            DextRec->derExtLenFrac_ls00[i]);
                    if (errOK_els00 != rc)
                    {
                        RepServMessages::Instance()->ls98Msg(layDataExtract_els98,
                                                             pszErrText,
                                                             errNumberTruncAtExtract_els98,
                                                             pColumnDesc->colName_ls00.asCharp());
                        ++DextRec->derErrNumInRecord_ls00;
                        rc = errOK_els00;
                    }
                    else
                    {
                        pColumnData[i].cdPtrUsed_ls00 = false;
                    }
                }
                else if (1 == szMap_SpecConstDataType_ls00[pColumnData[i].cdSQLDataType_ls00])
                {
                    //*
                    //*     Special data type like DATE, TIME, TIMESTAMP and BOOLEAN
                    //*
                    if (dboolean == pColumnData[i].cdSQLDataType_ls00)
                    {
                        if (0 == (pColumnData[i].cdPtrToData_ls00)[0])
                        {
                            pColumnData[i].cdPtrToData_ls00  = DataFileFormat->ffoFalseVal_ls00;
                            pColumnData[i].cdDataLen_ls00    = DataFileFormat->ffoFValLen_ls00;
                        }
                        else
                        {
                            pColumnData[i].cdPtrToData_ls00  = DataFileFormat->ffoTrueVal_ls00;
                            pColumnData[i].cdDataLen_ls00    = DataFileFormat->ffoTValLen_ls00;
                        }
                    }
                    else
                    {
                        if (false == pColumnData[i].cdPtrUsed_ls00)
                        {
                            // let the pointer point to the buffer of it's own element
                            pColumnData[i].cdPtrToData_ls00 = pColumnData[i].cdDataBuffer_ls00;
                        }
                        ls24ExtractDateTime(pColumnData[i].cdPtrToData_ls00,
                                            pColumnData[i].cdDataBuffer_ls00,
                                            pColumnData[i].cdDataLen_ls00,
                                            pColumnData[i].cdSQLDataType_ls00,
                                            DataFileFormat);

                        pColumnData[i].cdPtrUsed_ls00 = false;
                    }
                }
                else if (1 == szMap_LongDataType_ls00[pColumnData[i].cdSQLDataType_ls00])
                {
                    //*
                    //*     LONG data type
                    //*
                    // We simply copy the longdescriptor to a long descriptor array
                    if (INSERT_NULL  != pColumnData[i].cdDataLen_ls00)
                    {
                        memcpy(&DextRec->derLongDescArray_ls00[DextRec->derLongInfo_ls00.dliCntNNLongs_ls00],
                               pColumnData[i].cdPtrToData_ls00,
                               LongDescSize_ls00);
                        ++DextRec->derLongInfo_ls00.dliCntNNLongs_ls00;
                        DextRec->derLongInfo_ls00.dliNNLongs_ls00[l] = i;
                        ++l;
                    }
                }

                // This handles blobs, too. In this case simply the output format is ALWAYS set
                // to hex for a blob column and the write functions treat them right.

            }   // end if (INSERT_NULL == pColumnData[i].cdDataLen_ls00)

        }   // endif (true == pColumnDesc->colIsLit_ls00)


        //*
        //*     Special conversion in case of char output if not a LONG value
        //*
        if ( (ioChar_els00 == pColumnData[i].cdExtDataType_ls00) &&
             (0            == szMap_LongDataType_ls00[pColumnData[i].cdSQLDataType_ls00]) )
        {
            rc = ls27_SpecialConversion(&pColumnData[i],
                                        bUnicodeOI,
                                        pConvertBuffer,
                                        pszErrText);
        }

    }   // end for (i = 0; i < MultColumns->mcsCount_ls00; ++i)

    return rc;
}
// ls27_TransformData()


/*
  -----------------------------------------------------------------------------
  function:     ls27_ConvertTableData
  -----------------------------------------------------------------------------
*/
static tsp00_Longint
ls27_ConvertTableData(tls00_DBInfo       *DBInfo,
                      tsp00_Int4          RowCount,
                      tls00_DextractRec  *DextRec,
                      tls00_DatExtrCmd   *DextCmd,
                      tls27_DataSink     *pDataSink,
                      tls00_Buffer2      *pConvertBuffer,
                      tsp00_Addr          pszErrText)
{

    ROUTINE_DBG_MEO00 ("ls27_ConvertTableData");

    RepServMessages        *pMsgObj        = RepServMessages::Instance();         // error reporting obj
    tsp00_Longint           rc             = errOK_els00;

    bool                    bUnicodeOI     = ls00IsDBUnicode(DBInfo);

    tls00_TableDescription &TableInfo      = DextRec->derTableInfo_ls00;
    tls00_FileFormat       &DataFileFormat = DextCmd->dexDatFile_ls00.fsFormat_ls00;
    tls00_MultColSpec      &MultCol        = DextCmd->dexOutSpec_ls00;
    tls00_ColumnData       *pColumnData    = DextRec->pderColumnData_ls00;
    tls00_Column           *pColumn        = NULL;

    tsp00_Int4              lErrNum        = DextRec->derErrNumInRecord_ls00;
    tsp00_Int4              BufOffset      = 0;

    DextRec->derUseGetvalPckt_ls00 = false;         // must be reinitialized after fetching rows

    //*
    //*     Define output and input buffers
    //*
    //* In case of output format COMPACT we need to adjust the buffer pointer especially
    tsp00_Addr pszO_Buffer = STATIC_CAST(_TCHAR*, pDataSink->dskOutLine_ls27.pBufPtr_ls00);
    tsp00_Addr pszI_Buffer = DextRec->derDataSource_ls00.dasInputBuffer.strAddr_ls00;

    if (true == DataFileFormat.ffoCompact_ls00)     // Adjust for COMPACT format
    {
        // First part of the buffer is dedicated to the meta information; this
        // is immediately followed by the real data
        //
        //<---                 buffer                                  --->
        //-----------------------------------------------------------------
        //meta info| values ...                                           |
        //-----------------------------------------------------------------
        pDataSink->dskMetaLine_ls27.pBufPtr_ls00  = pszO_Buffer;
        pszO_Buffer                              += pDataSink->dskMetaLine_ls27.lBufLen_ls00;
    }


    for (tsp00_Int4 j = 0; ( (j < RowCount) && (errOK_els00 == rc) ); ++j)
    {
        //*
        //*     Process single record
        //*

        // Get data from order interface communication packet
        rc = ls18BindDataToCol(&pszI_Buffer[BufOffset],
                               DextRec->derRecordLen_ls00,
                               &TableInfo,
                               pColumnData,
                               DextCmd->dexOutSpec_ls00.mcsCount_ls00,
                               bUnicodeOI,
                               pszErrText);

        // Reinitialize buffer pointer
        pDataSink->dskOutLine_ls27.lBufPos_ls00 = 0;
        if (true == DataFileFormat.ffoCompact_ls00)
        {
            pDataSink->dskMetaLine_ls27.lBufPos_ls00 = 0;
        }

        // Convert those data
        if (errOK_els00 == rc)
        {
            rc = ls27_TransformData(DBInfo, DextRec, DextCmd, pDataSink, pConvertBuffer, pszErrText);
        }
  
        // hier angekommen, sind alle daten so wie sie ausgegeben werden koennen in der
        // tls00_columndata-struktur gespeichert, ausgenommen die longs. Diese muessen hier 
        // extra behandelt werden, da bei COMPACT und nicht COMPACT eine unterschiedliche
        // behandlung notwendig.
        



        if (true == DataFileFormat.ffoCompact_ls00)
        {
            rc = ls27_GetLongLength(DBInfo, DextRec, DextCmd, pszErrText);

            if (errOK_els00 == rc)
            {
                if (true == DataFileFormat.ffoCompact_ls00)
                {
                    pDataSink->dskOutLine_ls27.lBufPos_ls00 =
                        ls18PutCompactValue(&pDataSink->dskMetaLine_ls27,
                                            pszO_Buffer,        // zeigt auf pszI_Buffer[StartPos]
                                            pDataSink->dskOutLine_ls27.lBufLen_ls00,
                                            pColumnData,
                                            &DextCmd->dexOutSpec_ls00);
                }
            }

            
            //*
            //*     Write data to outfile and process longs if needed
            //*
            if (errOK_els00 == rc)      // Die funktion liefert immer errOK_els00 zurueck, nur die fehleranzahl
            {                           // hat sich u.u. veraendert
                // Write Meta Data (only in case of file format = COMPACT)
                // Write converted record to file
                rc = ls30VFWriteLn(*DextRec->derDataFile_ls00,
                                   STATIC_CAST(_TCHAR*, pDataSink->dskOutLine_ls27.pBufPtr_ls00),
                                   pDataSink->dskOutLine_ls27.lBufPos_ls00 + pDataSink->dskMetaLine_ls27.lBufPos_ls00,
                                   pszErrText);
                if (errOK_els00 != rc)
                {
                    break;  // fatal error; stop it
                }
            }
            
            //*
            //*     Extract long values after data has been successfully extracted
            //*
            if (errOK_els00 == rc)
            {
                if (0 != DextRec->derLongInfo_ls00.dliCntNNLongs_ls00)
                {
                    rc = ls08ExtractLongs(DBInfo,
                                          pszI_Buffer,
                                          &DextCmd->dexOutSpec_ls00,
                                          pColumnData,
                                          &DextRec->derLongInfo_ls00,
                                          DextRec->derLongDescArray_ls00,
                                          DextRec->derUseGetvalPckt_ls00,
                                          pConvertBuffer,
                                          pszErrText);
                    if (errOK_els00 != rc)      // Die fkt schreibt schon die fehlermeldung
                    {
                        ++DextRec->derErrNumInRecord_ls00;
                        rc = errOK_els00;
                    }
                }
            }
        }
        else
        {
            //*
            //*     Extract in external format COMPRESSED and FORMATTED
            //*

            //*
            //*     Extract long values and edit offset info for data file
            //*
            if (0 != DextRec->derLongInfo_ls00.dliCntNNLongs_ls00)
            {
                rc = ls08ExtractLongs(DBInfo,
                                      pszI_Buffer,
                                      &DextCmd->dexOutSpec_ls00,
                                      pColumnData,
                                      &DextRec->derLongInfo_ls00,
                                      DextRec->derLongDescArray_ls00,
                                      DextRec->derUseGetvalPckt_ls00,
                                      pConvertBuffer,
                                      pszErrText);

                if (errOK_els00 == rc)
                {
                    rc = ls27_EditLongOffsetForOutput(pColumnData,
                                                      &DextRec->derLongInfo_ls00,
                                                      &MultCol,
                                                      bUnicodeOI,
                                                      pConvertBuffer,
                                                      pszErrText);
                    if (errOK_els00 != rc)
                    {   // We rely on the fact that the only error rc is errLongFileNumGTMax_els98
                        pMsgObj->ls98Msg(layDataExtract_els98, pszErrText, STATIC_CAST(ErrorNumbers, rc),
                                                                           pColumn->colName_ls00.asCharp());
                    }
                }
            }   // end if (0 != DextRec->derLongInfo_ls00.dliCntNNLongs_ls00)

            if (errOK_els00 != rc)      // Die fkt schreibt schon die fehlermeldung
            {
                ++DextRec->derErrNumInRecord_ls00;
                rc = errOK_els00;
            }


            if (errOK_els00 == rc)
            {
                if (true == DataFileFormat.ffoFormatted_ls00)       // FORMATTED
                {
                    pDataSink->dskOutLine_ls27.lBufPos_ls00 =
                        ls18PutFormattedValue(pszO_Buffer,
                                              pDataSink->dskOutLine_ls27.lBufLen_ls00,
                                              pColumnData,
                                              &DextCmd->dexOutSpec_ls00);
                }
                else if (true == DataFileFormat.ffoCompress_ls00)   // COMPRESSED is the default
                {
                    pDataSink->dskOutLine_ls27.lBufPos_ls00 =
                        ls18PutUnformattedValue(pszO_Buffer,
                                                pDataSink->dskOutLine_ls27.lBufLen_ls00,
                                                DextCmd->dexOutSpec_ls00.mcsCount_ls00,
                                                pColumnData,
                                                DextRec->pderPattern_ls00);
                }
            }

            //*
            //*     Write data to outfile and process longs if needed
            //*
            if (errOK_els00 == rc)      // Die funktion liefert immer errOK_els00 zurueck, nur die fehleranzahl
            {                           // hat sich u.u. veraendert
                // Write Meta Data (only in case of file format = COMPACT)
                // Write converted record to file
                rc = ls30VFWriteLn(*DextRec->derDataFile_ls00,
                                   STATIC_CAST(_TCHAR*, pDataSink->dskOutLine_ls27.pBufPtr_ls00),
                                   pDataSink->dskOutLine_ls27.lBufPos_ls00 + pDataSink->dskMetaLine_ls27.lBufPos_ls00,
                                   pszErrText);
                if (errOK_els00 != rc)
                {
                    break;  // fatal error; stop it
                }
            }

        }   // end if (true == DataFileFormat.ffoCompact_ls00)

        ++DextRec->derRecordCount_ls00;     // Zaehle die anzahl geschriebener Zeilen hoch, auch wenn fehler
                                            // aufgetreten sind !!!

        //*
        //*     Write data to outfile and process longs if needed
        //*
        if (errOK_els00 == rc)      // Die funktion liefert immer errOK_els00 zurueck, nur die fehleranzahl
        {                           // hat sich u.u. veraendert
            // werte den moeglicherweiser aufgetretenen fehler beim zusammenstellen des records aus!!
            if (lErrNum != DextRec->derErrNumInRecord_ls00)
            {
                tsp00_Int4 lPosition = 1;
                if (true == DataFileFormat.ffoFormatted_ls00)
                {
                    lPosition = STATIC_CAST(tsp00_Int4,
                                            (pDataSink->dskOutLine_ls27.lBufPos_ls00 * (DextRec->derRecordCount_ls00 - 1)) );
                }
                RepServMessages::Instance()->ls98Msg(layDataExtract_els98, pszErrText,
                                                                           errAtExtractedRec_els98,
                                                                           DextRec->derRecordCount_ls00,
                                                                           lPosition);
                --DextRec->derLRejected_ls00;
                if (DextRec->derLRejected_ls00 <= 0)
                {
                    // der fehler wurde schon ausgegeben, der fehlercode ist gesetzt, also einfach
                    // zurueckkehren
                    break;
                }
                else
                {
                    lErrNum = DextRec->derErrNumInRecord_ls00;
                }
            }
        }

        BufOffset += DextRec->derRecordLen_ls00;
        lErrNum    = DextRec->derErrNumInRecord_ls00;

        // ReInitialize structure pColumnData
        for (tsp00_Longint i = 0; i <  MultCol.mcsCount_ls00; ++i)
        {
            pColumnData[i].cdDataLen_ls00    = 0;
            pColumnData[i].cdPtrUsed_ls00    = true;
            pColumnData[i].cdPtrToData_ls00  = NULL;
            pColumnData[i].cdLONGLen_ls00    = 0;
            pColumnData[i].cdIsNILValue_ls00 = false;
        }
    }

    // Bring pointer to its initial value for later remove
    if (true == DataFileFormat.ffoCompact_ls00)
    {
        pDataSink->dskOutLine_ls27.pBufPtr_ls00 = pDataSink->dskMetaLine_ls27.pBufPtr_ls00;
    }

    return rc;
}
// ls27_ConvertTableData()

/*
  -----------------------------------------------------------------------------
  EndChapter: Private functions implementation
  -----------------------------------------------------------------------------
*/
