/*! 
  -----------------------------------------------------------------------------
 
  module: vls27_dataextract.cpp
 
  -----------------------------------------------------------------------------
 
  responsible:  SteffenS and d033893

  special area: Replication Manager
 
  description:  Implementation of functions for dataextract
                
  version:      7.3.

  -----------------------------------------------------------------------------
 
  copyright:    SAP AG, 1999-2003
 
  -----------------------------------------------------------------------------


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*/

/* 
  -----------------------------------------------------------------------------
  includes
  -----------------------------------------------------------------------------
*/

#include "gls00.h"

#include "hls041.h"
#include "hls05.h"
#include "hls07.h"
#include "hls08_long.h"
#include "hls27_dataextract.h"
#include "hls27_DEExec.h"
#include "hls12.h"
#include "hls18.h"
#include "hls30.h"
#include "hls98msg.h"
#include "hls99.h"
#include "hls19_check.h"


/*  
  -----------------------------------------------------------------------------
  Chapter: Local functions declaration
  -----------------------------------------------------------------------------
*/

/*!
  -----------------------------------------------------------------------------
 
  function:     ls27_Constructor - local function
  -----------------------------------------------------------------------------

  description:  Creates and initializes the data extract record.

  arguments:    pDERec  [IN/OUT]  - reference to pointer; structure to create
                                     and initialize
                pDECmd  [IN]      - extract command structure, which
                                     contains the structure of type tls00_MultColSpec
                DBInfo  [IN]      - db session structure

  returnvalue:  errOK_els00            - successful
                errNoMoreMemory_els98  - error allocating memory for structure
                errInternal_els98      - error constructing pattern for read of
                                          files in format COMPRESSED
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls27_Constructor(tls00_DextractRec *&pDERec,
                 tls00_DatExtrCmd   *pDECmd,
                 tls00_DBInfo       *pDBInfo);


/*!
  -----------------------------------------------------------------------------
 
  function:     ls27_Destructor - local function
  -----------------------------------------------------------------------------

  description:  Destroyes the data extract record.

  arguments:    DextRec  [IN/OUT]  - pointer; structure to initialize

  returnvalue:  No
  -----------------------------------------------------------------------------
*/
static void
ls27_Destructor(tls00_DextractRec *DextRec, tls00_DatExtrCmd *pDextCmd);


/*!
  -----------------------------------------------------------------------------
 
  function:     ls27_DeclareCursor
  -----------------------------------------------------------------------------

  description:  Declares the cursor for extracting data from one table and
                retrieves all relevant information for processing the extract.

  arguments:    pDBInfo       [IN]   - DB session structure
                pDextCmd      [IN]   - User delivered data for command processing
                pTableDesc    [OUT]  - Stores the information on the table
                pszCursorName [OUT]  - Cursor name created here
                lResultCount  [OUT]  - Result count
                bTableEmpty   [OUT]  - Table empty or not
                pszErrText    [OUT]  - Error text

  returnvalue:  errOK_els00          - cursor successfully created
                errInternSQL_els98   - error creating cursor
                errUnknownPart_els98 - error reading information after successful
                                        creation of cursor
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls27_DeclareCursor(tls00_DBInfo           *pDBInfo,
                   tls00_DatExtrCmd       *pDextCmd,
                   tls00_TableDescription *pTableDesc,
                   tsp00_Addr             &pszCursorName,
                   tsp00_Int4             &lResultCount,
                   bool                   &bTableEmpty,
                   tsp00_Addr              pszErrText);


/*
  -----------------------------------------------------------------------------
  function:     ls27_InitAndOpenFiles
  -----------------------------------------------------------------------------

  description:  Initializes VFile structures for data and command file and
                opens those files.

  arguments:    pDECmd      [IN]   - User delivered data for command processing
                pDERec      [OUT]  - internal struct for command processing
                pszErrText  [OUT]  - Error text

  returnvalue:  errOK_els00         - cursor successfully created
                errFileOpen_els98   - error opening file
                The function logs the error message.
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls27_InitAndOpenFiles(tls00_DatExtrCmd  *pDECmd,
                      tls00_DextractRec *pDERec,
                      tsp00_Addr         pszErrText);


/*
  -----------------------------------------------------------------------------
 
  function:     ls27_PrepareDataextract
  -----------------------------------------------------------------------------

  description:  Checks length values for output columns.
                Initializes and fills tls00_ColumnData structure.
                Fills tls00_Column structure if no output columns specified in
                dataextract command.
                
  arguments:    DextCmd  [IN]  - User delivered data on output columns
                DextRec  [IN]  - Internal structure for de command processing
                ErrText  [OUT] - Error text

  returnvalue:  errOK_els00               - Success
                errNotInSelList_els98     - Cols in column list but not in
                                            select list
                errWrongFieldLength_els98 - Wrong length specs in column list
  -----------------------------------------------------------------------------
*/
static int
ls27_PrepareDataextract(tls00_DatExtrCmd *DextCmd,
                        tls00_DextractRec *DextRec,
                        tsp00_Addr ErrText);


/*
  -----------------------------------------------------------------------------
  function:     ls27_InitAndOpenLongFiles
  -----------------------------------------------------------------------------

  description:  Initializes VFile structures for long files and opens them.
                
  arguments:    pMultCol        [IN]  - User delivered data on output columns;
                                         stores the long data file struct, too
                pColumnData     [IN]  - Internal struct for column data description
                pDataFile       [IN]  - Data file struct
                bCompactFormat  [IN]  - Output in compact format or not
                ErrText         [OUT] - Error text

  returnvalue:  errOK_els00        - Success
                errFileOpen_els98  - error opening file
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls27_InitAndOpenLongFiles(tls00_MultColSpec  *pMultCol,
                          tls00_ColumnData   *pColumnData,
                          tls00_VFile       *&pDataFile,
                          bool                bCompactFormat,
                          tsp00_Addr          pszErrText);

/*
  -----------------------------------------------------------------------------
  function:     ls27_CloseFiles
  -----------------------------------------------------------------------------

  description:  Closes long and data and command files.
                
  arguments:    pMultCol        [IN]  - User delivered data on output columns;
                                         stores the long data file struct, too
                pDERec          [IN]  - Internal structure for de command processing
                bCompactFormat  [IN]  - Output in compact format or not
                ErrText         [OUT] - Error text

  returnvalue:  errOK_els00         - Success
                errFileClose_els98  - error opening file
  -----------------------------------------------------------------------------
*/
static void
ls27_CloseFiles(tls00_MultColSpec *pMultCol,
                tls00_DextractRec *pDERec,
                bool               bCompactFormat,
                tsp00_Addr         pszErrText);

/*
  -----------------------------------------------------------------------------
 
  function:     ls27_FillColStruct - local function

  -----------------------------------------------------------------------------

  description:  Fuellt eine struktur vom typ tls00_MultColSpec mit allen fuer
                ein dataextract notwendigen informationen zu den spalten. Dazu
                wird die struktur TableInfo ausgewertet, die zuvor gefuellt worden
                sein muss.
                Bei der Laengenbestimmung der einzelnen spalten im formatted-fall
                werden die laengen fuer die ausgabewerte fuer boolsche spalten
                und nullwerte des replication servers beruecksichtigt.
                Das in der dataextract-anweisung keine spalten in der spaltenliste
                angegeben wurden, werden alle spalten mit dem ausgabeformat
                ioChar_ls00 beschrieben. Eine ausgabe im hex-format ist in diesem
                fall ebenfalls nicht moeglich.
                Die folgenden werte werden in der tls00_Column-struktur innerhalb
                struktur vom typ tls00_MultColSpec nicht oder auf den default gesetzt
                colNumber_ls00       --> not used in this case
                colIsLit_ls00        --> automatically set to false
                colIsKey_ls00        --> automatically set to false
                colIsSet_ls00        --> automatically set to false
                colAbstractType_ls00 --> automatically set to unknown!!
                colSpecialIndex      --> not used here
                colOption_ls00       --> not used here
                colAssign_ls00       --> not used here
                colNullCond_ls00     --> no condition in this case
                colLongInfo          --> must be defined later in any case

  arguments:    DextCmd    [IN/OUT]  - pointer; to extract command structure, which
                                       contains the structure of type tls00_MultColSpec
                TableInfo  [IN]      - pointer; contains all necessary information
                                       on the columns

  returnvalue:  No
  -----------------------------------------------------------------------------
*/
static void
ls27_FillColStruct(tls00_DatExtrCmd *DextCmd, tls00_TableDescription *TableInfo);


/*
  -----------------------------------------------------------------------------
 
  function:     ls27_MakeMapIndex - local function

  -----------------------------------------------------------------------------

  description:  Builds 'mapping index' from column descriptions in structure
                MultCol and TableInfo.
                Checks for columns given in the output column list but not
                selected.
                Enthaelt das kommando literale, ist der vergebene index immer 0.
                Soll in weiteren funktionen mit diesem erstellten index gearbeitet
                werden, muessen die literale extra behandelt werden, da sonst fehler
                auftreten koennen.
                Es werden erst alle spalten ueberprueft und dann eine
                'gesammelte' fehlermeldung erzeugt, wenn notwendig.

                ATTENTION: Columns in MultCol must be sorted ascending by positions

  arguments:    MultCol     [IN]    - pointer; to column structure
                TableInfo   [IN]    - pointer; to table description
                pColumnData [OUT]   - Internal struct for column data description;
                                       stores the 'mapping index'
                pszErrText  [OUT]   - Error text

  returnvalue:  errNotInSelList_els98 - spalten in spaltenliste aber nicht in
                                        select liste
                errOK_els00           - keine fehler
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls27_MakeMapIndex(tls00_MultColSpec      *MultCol,
                  tls00_TableDescription *TableInfo,
                  tls00_ColumnData       *pColumnData,
                  tsp00_Addr              pszErrText);


/*
  -----------------------------------------------------------------------------
 
  function:     ls27_CheckLength - local function

  -----------------------------------------------------------------------------

  description:  Prueft die laengenangaben fuer die spalten in der spaltenliste.
                Beruecksichtigt die darstellungen fuer boolsche werte und fuer
                die moegliche hexadezimale ausgabe der werte.
                Laengen der Nullwertdarstellungen werden ebenfalls ueberprueft,
                wenn die spalte nullwerte zulaesst. Ist die laenge fuer die
                ausgabe der nullwerte zu klein, werden aber nur warnungen 
                erzeugt.
                Laengend der angegebenen Literale im Kommando werden ebenfalls
                geprueft, aber nur warnungen ins server log file ausgegeben.
                Es werden erst alle spalten ueberprueft und dann eine
                'gesammelte' fehlermeldung erzeugt, wenn notwendig.

  arguments:    MultCol     [IN]  - pointer to column structure
                TableInfo   [IN]  - pointer to table description
                FileFormat  [IN]  - pointer to data file format structure
                pColumnData [IN]  - Internal struct for column data description
                IntFracPart [IN]  - 2-dim array which holds calculated
                                   integral and fractional parts of numeric
                                   columns
                pszErrText  [OUT] - Error text

  returnvalue:  errWrongFieldLength_els98 - spalten in spaltenliste mit falschen
                                            laengenangaben
                errOK_els00               - keine fehler
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls27_CheckLength(tls00_MultColSpec      *MultCol,
                 tls00_TableDescription *TableInfo,
                 tls00_FileFormat       *FileFormat,
                 tls00_ColumnData       *pColumnData,
                 tsp00_Int4             (*IntFracPart)[2],
                 tsp00_Addr             pszErrText);


/*
  -----------------------------------------------------------------------------
 
  function:     ls27_CheckCompatibility - local function

  -----------------------------------------------------------------------------

  description:  Prueft die kompatibilitaet von db und externen datentypen sowie
                der datentypen der gewaehlten nullwertdarstellung einer spalte und
                deren externen datentyp.

                Es werden erst alle spalten ueberprueft und dann eine
                'gesammelte' fehlermeldung erzeugt, wenn notwendig.

  arguments:    MultCol    [IN]  - pointer to column structure
                TableInfo  [IN]  - pointer to table description
                pColumnData [IN]  - Internal struct for column data description
                pszErrText [OUT] - Error text

  return type:  tsp00_Int4

  returnvalue:  errIncompatibleDataTypes_els98 - datentypen sind nicht kompatibel
                errCompatibilityCheckSum_els98 - mehr als eine spalte mit nicht
                                                 kompatiblen datentypen
                errIncompatibleNullValue_els98 - datentypen von nullwertdarstellung
                                                 und spalte sind inkompatibel
                errOK_els00                    - keine fehler
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls27_CheckCompatibility(tls00_MultColSpec      *MultCol,
                        tls00_TableDescription *TableInfo,
                        tls00_ColumnData       *pColumnData,
                        tls00_FileFormat       *FileFormat,
                        tsp00_Addr             pszErrText);


/*
  -----------------------------------------------------------------------------
 
  function:     ls27_RecalculateLength - local function

  -----------------------------------------------------------------------------

  description:  Berechnet fuer numerische spalten die anzahl der vorkomma- und
                nachkommastellen neu und bestimmt daraus die neue gesamtlaenge
                der zu extrahierenden werte und den anteil an nachkommastellen.
                Eingang in die berechnung finden die daten ueber die spalten
                von der db und eventuell vom kunden angegebene
                skalierungs- und rundungsfaktoren.

  arguments:    MultCol    [IN]  - pointer to column structure
                TableInfo  [IN]  - pointer to table description
                pColumnData [IN]  - Internal struct for column data description
                IntFracPart[OUT] - 2-dim array which holds newly to calculate
                                   integral and fractional parts of numeric
                                   columns

  returnvalue:  No
  -----------------------------------------------------------------------------
*/
static void
ls27_RecalculateLength(tls00_MultColSpec      *MultCol,
                       tls00_TableDescription *TableInfo,
                       tls00_ColumnData       *pColumnData,
                       tsp00_Int4             (*LenFrac)[2]);   // 2-dim array


/*!
  -----------------------------------------------------------------------------
 
  function:     ls27_AnalyzeLongInfo - local function

  -----------------------------------------------------------------------------

  description:  Analysiert die informationen, die fuer das entladen von
                LONG-Spalten benoetigt werden. Dafuer werden die informationen
                vom Kern ueber die spalten und die informationen vom benutzer
                in der struktur tls00_ParamInfo_LONG (innerhalb von tls00_Column
                gespeichert und angepasst.
                Passt die folgenden Variablen der Struktur an fuer jede zu entladende
                LONG-Spalte an:
                    loibFileNumbered_ls00
                    loiNumberSize_ls00
                    loiGreatestNum_ls00
                    loiFileName_ls00
                    loiFileOffset_ls00 ==> 1
                Alle anderen werte sind beim parsen bereits auf 0 gesetzt worden.
                Die funktion reserviert fuer das array DextRec->derLongDescArray_ls00
                speicher, wenn die anzahl der longspalten ungleich 0 ist

                Die folgenden szenarios sind beim entladen von LONG-Spalten moeglich
                1. es sollen longspalten entladen werden, aber es sind keine
                   longfileinfos im kommando
                   --> fehler               ODER
                   es sollen longspalten in normale datentypen entladen werden
                   --> fehler
                2. die LONG-spalte, die im kommando angegeben wurde, gibt es nicht
                   --> die longfileinfo wird nicht beachtet
                3. es werden keine long spalten entladen
                   --> sehr schoen, nichts zu tun
                4. es sollen long spalten entladen werden und alles ist richtig
                5. mehrere unterschiedliche/gleiche longfileinfos fuer eine (gleiche) spalte
                   --> die erste longfileinfo wird beachtet, die andere verworfen und
                       eine warnung ausgegeben, dass es ueberfluessige longfileinfos gibt
    
  arguments:    DextCmd  [IN/OUT]  - pointer; extract command structure
                DextRec  [IN/OUT]  - pointer; extract record structure
                ErrText  [OUT]     - Error text

  returnvalue:  errOK_els00               - alles fehlerfrei gelaufen
                errSurplusLongInfo_els98  - warnung: zusaetzliche, ueberzaehlige long infos
                                            im kommando, die keiner der angegebenen spalten
                                            zugeordnet werden koennen
                errNoLongFileInfo_els98   - long infos fuer angegebene long spalten
                                            fehlen
                errAnalyzeLongSum_els98   - mehr als ein fehler aufgetreten; die fehler sind
                                            in der protokolldatei vermerkt
                errWrongFieldLength_els98 - fuer eine durchnumerierte datei, die die longdaten
                                            aufnehmen soll, wurde eine zu kleine laenge
                                            in der zieldatendatei angegeben
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls27_AnalyzeLongInfo(tls00_DatExtrCmd  *DextCmd,
                     tls00_DextractRec *DextRec,
                     tsp00_Addr         pszErrText);




/*
  -----------------------------------------------------------------------------
  EndChapter: Local functions declaration
  -----------------------------------------------------------------------------
*/




/*
  -----------------------------------------------------------------------------
  Chapter: Global functions implementation
  -----------------------------------------------------------------------------
*/

/*
  -----------------------------------------------------------------------------
   function:     ls27ExecDataExtract
  -----------------------------------------------------------------------------
*/
int
ls27ExecDataExtract(tls00_DatExtrCmd *DextCmd, tls00_DBInfo *DBInfo, tsp00_Addr ErrText)
{
    ROUTINE_DBG_MEO00 ("ls27ExecDataExtract");

    RepServMessages *pMsgObj = RepServMessages::Instance();    // Fr error reports
    tsp00_Int4       rc      = errOK_els00;
    
    //*
    //*     Before doing anything else we check external encoding and
    //*     database encoding.
    //*     Not supported is loading UNICODE data into ASCII databases.
    //*
    rc = ls05AreEncodingsCompatible(DextCmd->dexDatFile_ls00.fsFormat_ls00.ffoCodeType_ls00,
                                    ls00IsDBUnicode(DBInfo) ? ctUCS2_els00 : ctAscii_els00,
                                    false);
    if (errOK_els00 != rc)
    {
        pMsgObj->ls98Msg(layDataExtract_els98, ErrText, errIncompatibleEncodings_els98);
    }

    
    //*
    //*    Create dataextract record structure
    //*
    tls00_DextractRec *pDERec = NULL;
    if (errOK_els00 == rc)
    {
        rc = ls27_Constructor(pDERec, DextCmd, DBInfo);
    }


    tls00_FileFormat        *pFileFormat = &DextCmd->dexDatFile_ls00.fsFormat_ls00;
    tls00_MultColSpec       *pMC         = &DextCmd->dexOutSpec_ls00;
    tls00_TableDescription  *pTI         = &pDERec->derTableInfo_ls00;   

    //*
    //*     Declare cursor for extracting values
    //*
    //* Function delivers
    //*  - detailed info on the table for processing the extract
    //*  - if the table is empty
    if (errOK_els00 == rc)
    {
        rc = ls27_DeclareCursor(DBInfo,
                                DextCmd,
                                &pDERec->derTableInfo_ls00,
                                pDERec->derRTableName_ls00,
                                pDERec->derResultCount_ls00,
                                pDERec->derTabEmpty_ls00,
                                ErrText);
    }


    //*
    //*     Create and open files
    //*
    //* takes empty tables into account; in this case only a command file
    //* is created
    if (errOK_els00 == rc)
    {
        rc = ls27_InitAndOpenFiles(DextCmd, pDERec, ErrText);
    }

    if ( (errOK_els00 == rc) && (false == pDERec->derTabEmpty_ls00) )
    {
        //*
        //*     Prepare dataextract
        //*
        if (errOK_els00 == rc)
        {
            rc = ls27_PrepareDataextract(DextCmd, pDERec, ErrText);
        }

        //*
        //*     Compute maximum number of records fitting in one order
        //*     interface packet for this special command
        //*
        if (errOK_els00 == rc)
        {
            pDERec->derMaxRowsPPckt_ls00 = ( DBInfo->dbiPktSize_ls00 -
                                             (BlockHSize_ls00 + PartHSize_ls00 + mxsp_resnum + 16) ) /
                                              pDERec->derRecordLen_ls00;
            if(0 == pDERec->derMaxRowsPPckt_ls00)
            {
                pDERec->derMaxRowsPPckt_ls00 = 1;
            }
        }


        //*
        //*     Analyze user given long file infos
        //*
        if (errOK_els00 == rc)
        {
            rc = ls27_InitAndOpenLongFiles(pMC,
                                           pDERec->pderColumnData_ls00,
                                           pDERec->derDataFile_ls00,
                                           DextCmd->dexDatFile_ls00.fsFormat_ls00.ffoCompact_ls00,
                                           ErrText);
        }

        // MAIN FUNCTION for dataextract
        if (errOK_els00 == rc)
        {
            rc = ls27ExtractTableData(DBInfo, pDERec, DextCmd, ErrText);
            if (errOK_els00 != rc)
            {
                char szFileName[MAX_REPLY_LENGTH_LS00];
                szFileName[0] = 0;
                pMsgObj->ls98GetLogFileName(szFileName);

                pMsgObj->ls98MsgToString(layDataExtract_els98, ErrText, errCmdUnsuccessful_els98, szFileName);
                rc = errCmdUnsuccessful_els98;
            }

            pMsgObj->ls98PrintMessage(layDataExtract_els98, msgNumExtractedRecords_els98,
                                                            pDERec->derRecordCount_ls00,
                                                            (DBInfo->dbiSession_ls00.siMaxAdmissibleErrors - 
                                                            pDERec->derLRejected_ls00) );
        }

        // Closing cursor is only necessary if the table is not empty
        ls12CloseCursor(*DBInfo, pDERec->derRTableName_ls00, ErrText);
    }
    

    if (errOK_els00 == rc)
    {
        //*
        //*     Build command file for 'de for dataload/fastload/dataupdate'
        //*
        //* If the table is empty no data has been extracted and no dataload command
        //* is generated.
        rc = ls27BuildLoadCtrlFile(DBInfo, pDERec, DextCmd, ErrText);
    }

    if (NULL != pDERec)
    {
        ls27_CloseFiles(&DextCmd->dexOutSpec_ls00, pDERec, DextCmd->dexDatFile_ls00.fsFormat_ls00.ffoCompact_ls00, ErrText);
    }


    ls27_Destructor(pDERec, DextCmd);
    
    return rc;
}
//  ls27ExecDataExtract()


/*
  -----------------------------------------------------------------------------
  EndChapter: Global functions implementation
  -----------------------------------------------------------------------------
*/

//////////////////////////////////////////////////////////////////////////////////////////////
/*
  -----------------------------------------------------------------------------
  Chapter: Local functions implementation
  -----------------------------------------------------------------------------
*/



/*
  -----------------------------------------------------------------------------
  function: ls27_Constructor
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls27_Constructor(tls00_DextractRec *&pDERec,
                 tls00_DatExtrCmd   *pDECmd,
                 tls00_DBInfo       *pDBInfo)
{
    ROUTINE_DBG_MEO00 ("ls27_Constructor");

    tsp00_Int4 rc = errOK_els00;

    pDERec = new tls00_DextractRec;
    if (NULL == pDERec)
    {
        rc = errNoMoreMemory_els98;
    }
    else
    {
        memset(pDERec, 0, sizeof(tls00_DextractRec));

        pDERec->derRawResultCount_ls00 = new char[sizeof(csp_rescnt_unknown) + 1];
        pDERec->derRawResultCount_ls00[0] = '\0';

        pDERec->derFillValue_ls00 = ls07GetFillByte(dcha);

        pDERec->derTabEmpty_ls00 = false;

        memset(&pDERec->derDataSource_ls00, 0, sizeof(tls00_DataSource));
        
        // Achtung! Wir initialisieren hier die anzahl der abgelehnten zeilen auf den
        // maximal moeglichen wert. Spaeter bei der verarbeitung wird dieser wert dann
        // immer um eins verringert, wenn ein fehler auftritt.
        // Fuer die ausgabe der anzahl abgelehnter zeilen muss dann die subtraktion
        // siMaxAdmissibleErrors - derLRejected_ls00 ausgefuehrt werden.
        pDERec->derLRejected_ls00 = pDBInfo->dbiSession_ls00.siMaxAdmissibleErrors;

        pDERec->derSwapKind_ls00 = pDBInfo->dbiOwnSwap_ls00;    //  Swap verhalten

        pDERec->derTableInfo_ls00.tdKernelSwap_ls00.becomes(sw_dummy);

        // Pattern structure in case of external data file format COMPRESSED
        if (true == pDECmd->dexDatFile_ls00.fsFormat_ls00.ffoCompress_ls00)
        {
            char szErrText[MAX_REPLY_LENGTH_LS00];

            pDERec->pderPattern_ls00 = new tls00_Pattern;
            memset(pDERec->pderPattern_ls00, 0, sizeof(tls00_Pattern));

            rc = ls18ConstructPattern(pDERec->pderPattern_ls00,
                                      pDECmd->dexDatFile_ls00.fsFormat_ls00.ffoSeperat_ls00,
                                      pDECmd->dexDatFile_ls00.fsFormat_ls00.ffoDelimit_ls00,
                                      pDECmd->dexDatFile_ls00.fsFormat_ls00.ffoCodeType_ls00,
                                      szErrText);
        }
    }
    
    return rc;
}
//  ls27_Constructor()


/*
  -----------------------------------------------------------------------------
  function: ls27_Destructor
  -----------------------------------------------------------------------------
*/
static void
ls27_Destructor(tls00_DextractRec *pDERec, tls00_DatExtrCmd *pDECmd)
{
    ROUTINE_DBG_MEO00 ("ls27_Destructor");

    if (NULL != pDERec)
    {
        delete [] pDERec->derRawResultCount_ls00;
        pDERec->derRawResultCount_ls00 = NULL;

        tls00_TableDescription &T = pDERec->derTableInfo_ls00;
    
        int i = 0;
        for (i; i < csp_max_fields; ++i)
        {
            if (NULL != T.tdParamArray_ls00[i])
            {
                delete [] T.tdParamArray_ls00[i];
                T.tdParamArray_ls00[i] = NULL;
            }

            if (NULL != T.tdColName_ls00[i])
            {
                delete T.tdColName_ls00[i];
                T.tdColName_ls00[i] = NULL;
            }
        }
    
        //*
        //*     Delete column data structure
        //*
        if (NULL != pDERec->pderColumnData_ls00)
        {
            for (int k = 0; k < pDECmd->dexOutSpec_ls00.mcsCount_ls00; ++k)
            {
                pDERec->pderColumnData_ls00[k].cdPtrToData_ls00 = NULL; // remove references

                if (NULL != pDERec->pderColumnData_ls00[k].cdDataBuffer_ls00)
                {
                    delete [] pDERec->pderColumnData_ls00[k].cdDataBuffer_ls00;
                    pDERec->pderColumnData_ls00[k].cdDataBuffer_ls00 = NULL;
                }
            }

            delete [] pDERec->pderColumnData_ls00;
            pDERec->pderColumnData_ls00 = NULL;
        }

        if (NULL != pDERec->derExtLenFrac_ls00)
        {
            delete [] pDERec->derExtLenFrac_ls00;
        }

        if (NULL != pDERec->derLongInfo_ls00.dliNNLongs_ls00)
        {
            delete [] pDERec->derLongInfo_ls00.dliNNLongs_ls00;
            pDERec->derLongInfo_ls00.dliNNLongs_ls00 = NULL;
        }

        if (NULL != pDERec->derLongDescArray_ls00)
        {
            delete [] pDERec->derLongDescArray_ls00;
            pDERec->derLongDescArray_ls00 = NULL;
        }

        // delete pattern
        ls18DestructPattern(pDERec->pderPattern_ls00);

        delete pDERec;     // at last delete the structure itself
    
    }   // end if (NULL != pDERec)

    return;
}
// ls27_Destructor()


/*
  -----------------------------------------------------------------------------
  function:     ls27_DeclareCursor
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls27_DeclareCursor(tls00_DBInfo           *pDBInfo,
                   tls00_DatExtrCmd       *pDextCmd,
                   tls00_TableDescription *pTableDesc,
                   tsp00_Addr             &pszCursorName,
                   tsp00_Int4             &lResultCount,
                   bool                   &bTableEmpty,
                   tsp00_Addr              pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls27_DeclareCursor");

    RepServMessages *pMsgObj = RepServMessages::Instance();    // Fr error reports
    tsp00_Int4       rc      = errOK_els00;

    //* Create cursor name: tempLOA_<DATUM><ZEIT>
    tsp00_Date Date;
    tsp00_Time Time;
    sqldattime(Date, Time);

    tsp00_Datec szDate;
    tsp00_Timec szTime;
    szDate.p2c(Date);
    szTime.p2c(Time);

    pszCursorName = new char[KNL_IDENTIFIER_SIZE];
    ls12CreateResultTabName(pszCursorName, KNL_IDENTIFIER_SIZE);

    tsp00_Addr pszStmt        = NULL;
    size_t     lSelectStmtLen = _tcslen(pDextCmd->dexSelOrdStr_ls00);
    size_t     lCursorStmtLen = KNL_IDENTIFIER_SIZE + lSelectStmtLen + 64;
    
    if (resUnknown_els00 == pDextCmd->dexRestore_ls00)
    {
        pszStmt = new _TCHAR[lCursorStmtLen];
        
        sp77sprintf(pszStmt, STATIC_CAST(tsp00_Int4, lCursorStmtLen), CURSOR_FOR_SEL_LS27,
                                                                      pszCursorName,
                                                                      pDextCmd->dexSelOrdStr_ls00);
    }
    else
    {
        lCursorStmtLen +=  _tcslen(pDextCmd->dexTable_ls00.tsFullTabName_ls00);
        pszStmt = new _TCHAR[lCursorStmtLen];

        if (lSelectStmtLen > 0)
        {
            sp77sprintf(pszStmt, STATIC_CAST(tsp00_Int4, lCursorStmtLen),
                                 CURSOR_FOR_SEL_STAR_ORDER_LS27, pszCursorName,
                                                                 pDextCmd->dexTable_ls00.tsFullTabName_ls00,
                                                                 pDextCmd->dexSelOrdStr_ls00);
        }
        else
        {
            sp77sprintf(pszStmt, STATIC_CAST(tsp00_Int4, lCursorStmtLen),
                                 CURSOR_FOR_SEL_STAR_LS27,
                                 pszCursorName, pDextCmd->dexTable_ls00.tsFullTabName_ls00);
        }
    }

    _TCHAR szErr[MAX_REPLY_LENGTH_LS00];
    szErr[0] = 0;
    rc = ls12DeclareCursor(*pDBInfo, NULL, pszStmt, szErr);
    if (errOK_els00 != rc)
    {
        if (dbRowNotFound_els00 == rc)             // empty table is no error
        {
            bTableEmpty = true;
            rc = errOK_els00;
        }
        else
        {
            pMsgObj->ls98Msg(layCursor_els98, pszErrText, errInternSQL_els98, pszStmt, szErr);
            rc = errInternSQL_els98;
        }
    }
    else
    {
        //*
        //*     Get info from different parts
        //*
        //* The following 3 funcs return errUnknownPart in case of error
        tsp1_packet_ptr pRcvPacket = pDBInfo->dbiPktRcvList_ls00[0];

        // Short info part
        rc = ls04GetShortInfo(pRcvPacket, csp_max_fields, pTableDesc->tdParamArray_ls00, pTableDesc->tdFieldCount_ls00);
        if (errOK_els00 == rc)
        {
            tsp00_Int2 fc = 0;

            // Column names part
            rc = ls04GetColumnNames(pRcvPacket, csp_max_fields, pTableDesc->tdColName_ls00, fc);

            // Hier haben wir den ersten boesen fehler, wenn fc und TableInfo.tdFieldCount_ls00
            // nicht uebereinstimmen sollten. Das muss abgefangen werden, auch wenn es wenig
            // wahrscheinlich ist.
            if (fc != pTableDesc->tdFieldCount_ls00)
            {
                rc = errUnknownPart_els98;
            }
        }
        if (errOK_els00 == rc)
        {
            // result count part
            // Get result count as integer value
            rc = ls04GetResultCount(pRcvPacket, NULL, lResultCount); // result count part
        }

        if (errOK_els00 != rc)
        {
            pMsgObj->ls98Msg(layDataExtract_els98, pszErrText, STATIC_CAST(ErrorNumbers, rc) );
        }
    }


    if (NULL != pszStmt)
    {
        delete [] pszStmt;
        pszStmt = NULL;
    }

    return rc;
}
// ls27_DeclareCursor()


/*
  -----------------------------------------------------------------------------
  function: ls27_FillColStruct
  -----------------------------------------------------------------------------
*/
static void
ls27_FillColStruct(tls00_DatExtrCmd *DextCmd, tls00_TableDescription *TableInfo)
{
    ROUTINE_DBG_MEO00 ("ls27_FillColStruct");

    tls00_MultColSpec &MultCol        = DextCmd->dexOutSpec_ls00;
    tls00_FileFormat  &DataFileFormat = DextCmd->dexDatFile_ls00.fsFormat_ls00;
    
    MultCol.mcsCount_ls00 = TableInfo->tdFieldCount_ls00;

    tls00_FieldBasics *pBI              = NULL;
    tsp00_Int4         ValLen           = 0;
    tsp00_Int4         NILValLen        = 0;
    tsp00_Int4         _long_file_index = 0;            // running loop variable for long file structure

    for (int i = 0; i < TableInfo->tdFieldCount_ls00; ++i)
    {
        pBI = &TableInfo->tdParamArray_ls00[i]->piBasicInfo_ls00;

        tls00_Column *pCol =  new tls00_Column;
        memset(pCol, DEF_BYTE_LS00, sizeof(tls00_Column));

        pCol->colName_ls00.rawAssign(TableInfo->tdColName_ls00[i]->asCharp());
        pCol->colNumber_ls00 = i + 1;           // necessary if long column and for 'DATAEXTRACT FOR DATALOAD/...'

        // Because we do define the output length here we need to define if the 
        // column is to be dumped as hexadecimal
        if ( (1 == szMap_BlobDataType_ls00[pBI->fbDataType_ls00]) &&
             (0 == szMap_LongDataType_ls00[pBI->fbDataType_ls00])  )
        {
            if (false == DataFileFormat.ffoBinary_ls00)
            {
                pCol->colFormat_ls00.ffoIsHex_ls00 = true;
            }
        }   


        if (true == DataFileFormat.ffoFormatted_ls00)       // formatierte ausgabe, d.h. FORMATTED
        {
            if (0 == i)
            {
                pCol->colFPos_ls00.fpoStart_ls00 = 1;
            }
            else
            {
                pCol->colFPos_ls00.fpoStart_ls00 = MultCol.mcsColumn_ls00[i-1]->colFPos_ls00.fpoEnd_ls00 + 1;
            }
            
            ValLen    = 0;
            NILValLen = 0;
        
            if (1 == szMap_LongDataType_ls00[pBI->fbDataType_ls00])
            {
                SAPDB_Int4 longlen =
                    STATIC_CAST(tsp00_Int4, _tcslen(DextCmd->dexLongSpec_ls00->mfsFiles_ls00[_long_file_index]->fsFileName_ls00));
                ls19CheckLengthForExtract(longlen, pCol, pBI, &DataFileFormat, ValLen, NULL);
                ++_long_file_index;
            }
            else
            {
                // We get the needed length by specifying a length of 0
                ls19CheckLengthForExtract(0, pCol, pBI, &DataFileFormat, ValLen, NULL);
            }

            // Check for allowed null values
            ls19CheckNILValLength(0, pCol, TableInfo->tdParamArray_ls00[i]->piMode_ls00, &DataFileFormat, NILValLen);
            pCol->colFPos_ls00.fpoEnd_ls00 = (NILValLen > ValLen) ?
                                                 pCol->colFPos_ls00.fpoStart_ls00 + NILValLen - 1 :
                                                 pCol->colFPos_ls00.fpoStart_ls00 + ValLen - 1;
        }
        else
        {
            pCol->colFPos_ls00.fpoStart_ls00 = i + 1;   // schleifenvariable startet bei 0, die
            pCol->colFPos_ls00.fpoEnd_ls00 = 0;         // position aber bei 1 !!!
        }
        
        pCol->colFormat_ls00.ffoFormat_ls00 = ioChar_els00;
        MultCol.mcsColumn_ls00[i] = pCol;
    }

    return;
}
// ls27_FillColStruct()


/*
  -----------------------------------------------------------------------------
  function: ls27_MakeMapIndex
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls27_MakeMapIndex(tls00_MultColSpec      *MultCol,
                  tls00_TableDescription *TableInfo,
                  tls00_ColumnData       *pColumnData,
                  tsp00_Addr             pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls27_MakeMapIndex");
    
    tsp00_Int4 rc = errOK_els00;

    // Die daten in tableinfo muessen gegebenenfalls umsortiert und an die ausgabe angepasst werden.
    // Wir erstellen dafuer einen index, mit dem der index von MultCol auf den von TableInfo
    // gemapt wird.
    tsp00_Int2 StrIndex  = 0;
    _TCHAR     *ErrArray = NULL;

    tls00_Column *pCol = NULL;

    for (tsp00_Int4 i=0, j = 0; i < MultCol->mcsCount_ls00; ++i)
    {
        pCol = MultCol->mcsColumn_ls00[i];
        
        bool bFound = false;
        if (0 != pCol->colNumber_ls00)
        {
            if (pCol->colNumber_ls00 <= TableInfo->tdFieldCount_ls00)
            {
                pColumnData[i].cdPosInStmt_ls00 = pCol->colNumber_ls00 - 1;
                bFound = true;
            }
        }
        else
        {
            for (j=0; j < TableInfo->tdFieldCount_ls00; ++j)
            {
                if ( 0 == _tcsicmp(pCol->colName_ls00.asCharp(), (*TableInfo->tdColName_ls00[j]).asCharp()))
                {
                    bFound = true;
                    pColumnData[i].cdPosInStmt_ls00 = j;
                }
            }
        }

        if (false == bFound)
        {
            // Nur wenn es sich nicht um eine zuweisung handelt, einen fehler generieren.
            // Eine zuweisung wird durch colIsLit_ls00 = true angezeigt.
            if (false == pCol->colIsLit_ls00)
            {
                if (NULL == ErrArray)
                {
                    ErrArray = new _TCHAR[(KNL_IDENTIFIER_SIZE+4) * MultCol->mcsCount_ls00];
                }

                if (0 != pCol->colNumber_ls00)
                {
                    StrIndex += sprintf(ErrArray + StrIndex, _T("%d, "), pCol->colNumber_ls00);
                }
                else
                {
                    StrIndex += sprintf(ErrArray + StrIndex, _T("%s, "), pCol->colName_ls00.asCharp());
                }
            }
        }
    }   // end for (i=0; i < MultCol->mcsCount_ls00; ++i)

    if (NULL != ErrArray)
    {
        ErrArray[strlen(ErrArray) - 2] = '\0';  // letztes ueberzaehliges komma entfernen

        RepServMessages::Instance()->ls98Msg(layDataExtract_els98, pszErrText, errNotInSelList_els98, ErrArray);
        delete [] ErrArray;
    
        rc = errNotInSelList_els98;
    }

    return rc;
}
// ls27_MakeMapIndex()

/*
  -----------------------------------------------------------------------------
  function: ls27_RecalculateLength
  -----------------------------------------------------------------------------
*/
static void
ls27_RecalculateLength(tls00_MultColSpec    *MultCol,
                       tls00_TableDescription *TableInfo,
                       tls00_ColumnData       *pColumnData,
                       tsp00_Int4             (*LenFrac)[2])
{
    ROUTINE_DBG_MEO00 ("ls27_RecalculateLength");

    for(tsp00_Int4 i = 0; i < MultCol->mcsCount_ls00; ++i)
    {
        tls00_Column      *pCol = MultCol->mcsColumn_ls00[i];
        tls00_FieldBasics  FB   = TableInfo->tdParamArray_ls00[pColumnData[i].cdPosInStmt_ls00]->piBasicInfo_ls00;

        if (true == pCol->colIsLit_ls00)
        {
            continue;
        }

        if (1 == szMap_NumberDataType_ls00[pColumnData[i].cdSQLDataType_ls00])
        {
            ls19CalcExtLenFrac(FB, pCol->colOption_ls00, LenFrac[i]);
        }
    }

    return;
}
// ls27_RecalculateLength()


/*
  -----------------------------------------------------------------------------
  function: ls27_CheckLength
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls27_CheckLength(tls00_MultColSpec      *MultCol,
                 tls00_TableDescription *TableInfo,
                 tls00_FileFormat       *FileFormat,
                 tls00_ColumnData       *pColumnData,
                 tsp00_Int4             (*IntFracPart)[2],
                 tsp00_Addr             pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls27_CheckLength");

    RepServMessages *pMsgObj = RepServMessages::Instance();    // error reporting object
    
    tsp00_Int4 rc  = errOK_els00;
    tsp00_Int4 rc2 = errOK_els00;

    if (true == FileFormat->ffoFormatted_ls00)
    {
        tsp00_Int4       lErrNum = 0;

        tls00_Column    *pCol    = NULL;
        tls00_ParamInfo *pPI     = NULL;
        tsp00_Int4       ValLen  = 0;
        tsp00_Int4       len     = 0;

        for(tsp00_Int4 i = 0; i < MultCol->mcsCount_ls00; ++i)
        {
            pCol   = MultCol->mcsColumn_ls00[i];
            pPI    = TableInfo->tdParamArray_ls00[pColumnData[i].cdPosInStmt_ls00];
            ValLen = pCol->colFPos_ls00.fpoEnd_ls00 - pCol->colFPos_ls00.fpoStart_ls00 + 1;
            len    = 0;

            if (true == pCol->colIsLit_ls00)
            {   // Literals
                // in case the length is to little we create a warning; no further checks necessary

                if ( pCol->colName_ls00.length() > ValLen )
                {
                    pMsgObj->ls98PrintMessage(layDataExtract_els98, errLiteralTruncated_els98,
                                                                    pCol->colName_ls00.asCharp());
                }
            }
            else
            {
                rc = ls19CheckLengthForExtract(ValLen, pCol, &pPI->piBasicInfo_ls00, FileFormat, len, IntFracPart[i]);
                if ( errOK_els00 != rc )
                {
                    if (1 == szMap_LongDataType_ls00[pColumnData[i].cdSQLDataType_ls00])
                    {
                        rc = errOK_els00;
                    }
                    else
                    {
                        // generate a message to the error string and to the log file
                        if ( (errFractionTooSmall_els98 == rc) || (errWrongFieldLength_els98 == rc) )
                        {
                            pMsgObj->ls98Msg(layDataExtract_els98, pszErrText, STATIC_CAST(ErrorNumbers,rc),
                                                                               pCol->colName_ls00.asCharp(),
                                                                               len);
                            ++lErrNum;
                        }
                        else if (errIntValueOverflow_els98 == rc)   // only a warning
                        {
                            pMsgObj->ls98Msg(layDataExtract_els98, pszErrText, STATIC_CAST(ErrorNumbers,rc),
                                                                               pCol->colName_ls00.asCharp());
                            rc = errOK_els00;
                        }
                        else    // special error messages
                        {
                            pMsgObj->ls98Msg(layDataExtract_els98, pszErrText, STATIC_CAST(ErrorNumbers,rc),
                                                                               pCol->colName_ls00.asCharp());
                            ++lErrNum;
                        }
                    }
                }

                // Check length of null values. The used function generates only warnings.
                // REUSING variable len!
                rc2 = ls19CheckNILValLength(ValLen, pCol, pPI->piMode_ls00, FileFormat, len=0);
                if (errOK_els00 != rc2)
                {
                    pMsgObj->ls98PrintMessage(layDataExtract_els98, STATIC_CAST(ErrorNumbers,rc2),
                                                                     pCol->colName_ls00.asCharp());
                }
            }
        }   // end for(i = 0; i < MultCol->mcsCount_ls00; ++i)

        if (0 != lErrNum)
        {
            rc = errLengthCheckSum_els98;
            pMsgObj->ls98MsgToString(layDataExtract_els98, pszErrText, errLengthCheckSum_els98);
        }
    }   // end if (true == FileFormat->ffoFormatted_ls00)

    return rc;
}
// ls27_CheckLength()


/*
  -----------------------------------------------------------------------------
  function:     ls27_CheckCompatibility
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls27_CheckCompatibility(tls00_MultColSpec      *MultCol,
                        tls00_TableDescription *TableInfo,
                        tls00_ColumnData       *pColumnData,
                        tls00_FileFormat       *FileFormat,
                        tsp00_Addr             pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls27_CheckCompatibility");

    tsp00_Int4       rc      = errOK_els00;
    RepServMessages *pMsgObj = RepServMessages::Instance();    // message object

    tsp00_Int4       lErrNum = 0;
    tls00_Column    *pCol    = NULL;
    tls00_ParamInfo *pPI     = NULL;


    for(tsp00_Int4 i = 0; i < MultCol->mcsCount_ls00; ++i)
    {
        pCol = MultCol->mcsColumn_ls00[i];
        pPI  = TableInfo->tdParamArray_ls00[pColumnData[i].cdPosInStmt_ls00];

        // Literal? Nothing to do.
        if (true == pCol->colIsLit_ls00)
        {
            continue;
        }

        rc = ls19CheckDataTypeCompatibility(pCol, &pPI->piBasicInfo_ls00);
        if (errOK_els00 != rc)
        {
            // Es wird immer eine meldung sowohl ins logfile als auch in den fehlerstring
            // ausgegeben. Wenn mehr als ein fehler auftreten, dann wird die letzte fehlermeldung
            // mit einer summenfehlermeldung weiter unten ueberschrieben.
            pMsgObj->ls98PrintMessage(layDataExtract_els98, STATIC_CAST(ErrorNumbers,rc),
                                                            pCol->colName_ls00.asCharp());
            ++lErrNum;
            rc = errOK_els00;
        }

        if (1 == szMap_LongDataType_ls00[pColumnData[i].cdSQLDataType_ls00])
        {
            rc = ls08CheckLONG_DT_Compatibility(pCol->colLongInfo.loiFileEncoding_ls00,
                                                pColumnData[i].cdSQLDataType_ls00,
                                                false);
            if (errOK_els00 != rc)
            {
                pMsgObj->ls98PrintMessage(layDataExtract_els98, STATIC_CAST(ErrorNumbers,rc),
                                                                pCol->colName_ls00.asCharp());
                ++lErrNum;
                rc = errOK_els00;
            }
        }

        rc = ls19CheckNILValCompatibility(pCol, pPI->piMode_ls00, FileFormat, pszErrText);
        if (errOK_els00 != rc)
        {
            // Es wird immer eine meldung sowohl ins logfile als auch in den fehlerstring
            // ausgegeben. Wenn mehr als ein fehler auftreten, dann wird die letzte fehlermeldung
            // mit einer summenfehlermeldung weiter unten ueberschrieben.
            pMsgObj->ls98PrintMessage(layDataExtract_els98, STATIC_CAST(ErrorNumbers,rc),
                                                            pCol->colName_ls00.asCharp());
            ++lErrNum;
            rc = errOK_els00;
        }

    }

    if (0 != lErrNum)
    {
        rc = errCompatibilityCheckSum_els98;
        pMsgObj->ls98MsgToString(layDataExtract_els98, pszErrText, errCompatibilityCheckSum_els98);
    }

    return rc;
}
// ls27_CheckCompatibility()


/*
  -----------------------------------------------------------------------------
  function: ls27_PrepareDataextract
  -----------------------------------------------------------------------------
*/
static int
ls27_PrepareDataextract(tls00_DatExtrCmd *DextCmd, tls00_DextractRec *DextRec, tsp00_Addr ErrText)
{
    ROUTINE_DBG_MEO00 ("ls27_PrepareDataextract");

    RepServMessages *MsgInstance = RepServMessages::Instance(); // Error reporting object

    tsp00_Int4 rc    = errOK_els00;
    bool       bSelf = false;

    tls00_TableDescription  *pTI         = &DextRec->derTableInfo_ls00;
    tls00_MultColSpec       &MultCol     = DextCmd->dexOutSpec_ls00;
    tls00_ColumnData       *&pColumnData = DextRec->pderColumnData_ls00;
    tls00_FileFormat        *pFFormat    = &DextCmd->dexDatFile_ls00.fsFormat_ls00;
    tls00_Column            *pCol        = NULL;
    tls00_FieldBasics       *pBI         = NULL;
            

    if (0 == MultCol.mcsCount_ls00)
    {
        ls27_FillColStruct(DextCmd, pTI);
        bSelf = true;
    }


    //*
    //*     1. Initialize column data structure with the count of columns to extract
    //*     2. Define fraction and hex representation
    //*
    //* This can't be done in the constructor because in case of DATAEXTRACT FOR DATALOAD the
    //* count of columns in the command is 0!
    //*
    pColumnData = new tls00_ColumnData[MultCol.mcsCount_ls00];
    memset(pColumnData, 0, (sizeof(tls00_ColumnData) * MultCol.mcsCount_ls00) );

    if (errOK_els00 == rc)
    {
        rc = ls27_MakeMapIndex(&MultCol, pTI, pColumnData, ErrText);
    }


//TODOTODO hex representation muss noch mal ueberarbeitet werden und nur in column data uebernommen werden
    tsp00_Int4        lValLen  = 0;
    tls00_ColumnData *pColData = NULL;
    tls00_ParamInfo  *pParamInfo = NULL;

    for (int k = 0, l = 0; k < MultCol.mcsCount_ls00; ++k)
    {
        pColData   = &pColumnData[k];
        pCol       = MultCol.mcsColumn_ls00[k];
        pParamInfo = pTI->tdParamArray_ls00[pColData->cdPosInStmt_ls00];
        pBI        = &pParamInfo->piBasicInfo_ls00;
        
        pColData->cdSQLDataType_ls00 = pBI->fbDataType_ls00;
        pColData->cdSQLLen_ls00      = pBI->fbInOutLength_ls00;
        pColData->cdExtEncoding_ls00 = pFFormat->ffoCodeType_ls00;

        
        // Blob data type -> make the output format hex if it is not long byte;
        // in case of long byte the output value to data file is the offset or file
        // name of the long file, the long value output must remain byte
        if ( (1 == szMap_BlobDataType_ls00[pColData->cdSQLDataType_ls00]) &&
             (0 == szMap_LongDataType_ls00[pColData->cdSQLDataType_ls00])  )
        {
            if ( (false == pFFormat->ffoBinary_ls00) && (false == pCol->colFormat_ls00.ffoIsHex_ls00) )
            {
                pCol->colFormat_ls00.ffoIsHex_ls00 = true;
            }
        }   

        pColData->cdExtHexRepresentation_ls00 = pCol->colFormat_ls00.ffoIsHex_ls00;
        
        if (false == pFFormat->ffoFormatted_ls00)
        {
            if (1 == szMap_NumberDataType_ls00[pColData->cdSQLDataType_ls00])
            {
                // We get the needed length by specifying a length of 0
                lValLen = 0;
                ls19CheckLengthForExtract(0, pCol, pBI, pFFormat, lValLen, NULL);
                pColData->cdDataBufferLen_ls00 = lValLen;
            }
            else if (dboolean == pColData->cdSQLDataType_ls00)
            {
                // boolean values are delivered from kernel with a length of 1 byte; output
                // length is derived from the external boolean representation
                pColData->cdDataBufferLen_ls00 = MAX_BOOLVALUE_LENGTH_LS00;
            }
            else
            {
                pColData->cdDataBufferLen_ls00 = pBI->fbInOutLength_ls00;
            }

            // Check for writing (possible) NULL values and lengthen the buffer accordingly
            if ( false == pParamInfo->piMode_ls00.includes(sp1ot_mandatory) )
            {
                // column is not mandatory and thus MAY allow NULL values -> be prepared
                if (0 != pCol->colAssign_ls00.fasAsgnLen_ls00)
                {
                    // Customer specified special NULL value for this column
                    if (pColData->cdDataBufferLen_ls00 < pCol->colAssign_ls00.fasAsgnLen_ls00)
                    {
                        pColData->cdDataBufferLen_ls00 = pCol->colAssign_ls00.fasAsgnLen_ls00;
                    }
                }
                else
                {
                    // Customer specified no special NULL value for this column
                    if (pColData->cdDataBufferLen_ls00 < pFFormat->ffoNSpecLen_ls00)
                    {
                        pColData->cdDataBufferLen_ls00 = pFFormat->ffoNSpecLen_ls00;
                    }
                }
            }

            // be prepared for unicode and hex representation: make the buffer large enough;
            // for long values only the descriptor is saved in the data buffer of tls00_Column,
            // this length does not depend on the external encoding
            if (0 == szMap_LongDataType_ls00[pColData->cdSQLDataType_ls00])
            {
                 if ( (ctUCS2_els00        == pColData->cdExtEncoding_ls00) ||
                      (ctUCS2Swapped_els00 == pColData->cdExtEncoding_ls00)  )
                {
                    if (true == pColData->cdExtHexRepresentation_ls00)
                    {
                        pColData->cdDataBufferLen_ls00 *= 8;        //2 (hex) *2 (unicode) *2 (hex)
                    }
                    else
                    {
                        pColData->cdDataBufferLen_ls00 *= 2;
                    }
                }
                else if (ctUTF8_els00 == pColData->cdExtEncoding_ls00)
                {
                    if (true == pColData->cdExtHexRepresentation_ls00)
                    {
                        pColData->cdDataBufferLen_ls00 *= 16;
                    }
                    else
                    {
                        pColData->cdDataBufferLen_ls00 *= 4;
                    }
                }
                else
                {
                    if (true == pColData->cdExtHexRepresentation_ls00)
                    {
                        pColData->cdDataBufferLen_ls00 *= 2;
                    }
                }
            }   // end if (0 == szMap_LongDataType_ls00[pColData->cdSQLDataType_ls00])
        }
        else
        {
            pColData->cdDataBufferLen_ls00 = pCol->colFPos_ls00.fpoEnd_ls00 - pCol->colFPos_ls00.fpoStart_ls00 + 1;
        }

        pColData->cdDataBuffer_ls00  = new _TCHAR[pColData->cdDataBufferLen_ls00];
        pColData->cdExtDataType_ls00 = pCol->colFormat_ls00.ffoFormat_ls00;
        pColData->cdIsNILValue_ls00  = false;
        pColData->cdPtrToData_ls00   = NULL;
        pColData->cdPtrUsed_ls00     = true;

        
        //*
        //*     fraction value and hex representation
        //*
        //* If fraction is still undefined make it have the db value
        if ( (false == pCol->colIsLit_ls00) &&
             (UNDEFINED_LS00 == pCol->colFormat_ls00.ffoFrac_ls00) )
        {
            pCol->colFormat_ls00.ffoFrac_ls00 = pBI->fbFraction_ls00;
        }
    }   // end for (int k = 0, l = 0; k < DextCmd->dexOutSpec_ls00.mcsCount_ls00; ++k)


    if (errOK_els00 == rc)
    {
        rc = ls27_AnalyzeLongInfo(DextCmd, DextRec, ErrText);
    }


    //*
    //*     Initialize Long descriptor array
    //*
    //* This can't be done in the constructor because in case of DATAEXTRACT FOR DATALOAD with
    //* format COMPACT the count of long files in the command is 0!
    //*
    DextRec->derLongDescArray_ls00 = new tsp00_LongDescriptor[DextRec->derLongColCnt_ls00];
    DextRec->derLongInfo_ls00.dliNNLongs_ls00 = new tsp00_Int2[DextRec->derLongColCnt_ls00];
    memset(DextRec->derLongInfo_ls00.dliNNLongs_ls00, 0, sizeof(tsp00_Int2) * DextRec->derLongColCnt_ls00);


    //*
    //*     Compatibility and length checks
    //*
    if (errOK_els00 == rc)
    {
        rc = ls27_CheckCompatibility(&MultCol, pTI, pColumnData, pFFormat, ErrText);
    }

    if (errOK_els00 == rc)
    {
        DextRec->derExtLenFrac_ls00 = new tsp00_Int4[MultCol.mcsCount_ls00][2];

        ls27_RecalculateLength(&MultCol, pTI, pColumnData, DextRec->derExtLenFrac_ls00);
    }

    if ( (errOK_els00 == rc) && (false == bSelf) )
    {        
        rc = ls27_CheckLength(&MultCol, pTI, pFFormat, pColumnData, DextRec->derExtLenFrac_ls00, ErrText);
    }


    //*
    //*     Set long file encoding in column data structure
    //*     Compute record length, bytes per line
    //*
    if (errOK_els00 == rc)
    {
        pColData = NULL;
        for (int i = 0, j = 0; i < DextCmd->dexOutSpec_ls00.mcsCount_ls00; ++i)
        {
            pColData = &pColumnData[i];
            pCol     = MultCol.mcsColumn_ls00[i];
            pBI      = &pTI->tdParamArray_ls00[pColData->cdPosInStmt_ls00]->piBasicInfo_ls00;

            // long file encoding
            if (1 == szMap_LongDataType_ls00[pColData->cdSQLDataType_ls00])
            {
                pColData->cdExtLONGEncoding_ls00 = pCol->colLongInfo.loiFileEncoding_ls00;
                ++j;
            }
        }

        DextRec->derRecordLen_ls00 = ls18GetRecordLength(*pTI, true);
    
        pFFormat->ffoBytesPerLine_ls00 = ls18GetRowLengthForExtract(&MultCol, pColumnData, pTI, pFFormat);
    }


    return rc;
}
// ls27_PrepareDataextract()


/*
  -----------------------------------------------------------------------------
  function:     ls27_AnalyzeLongInfo - local function
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls27_AnalyzeLongInfo(tls00_DatExtrCmd *DextCmd, tls00_DextractRec *DextRec, tsp00_Addr pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls27_AnalyzeLongInfo");

    RepServMessages *pMsgObj = RepServMessages::Instance();    // Fr error reports
    
    tsp00_Int4 rc = errOK_els00;

    tsp00_Int4 i       = 0;
    tsp00_Int4 j       = 0;
    tsp00_Int4 lErrNum = 0;
    bool       bFound  = false;

    tls00_MultColSpec    &MultCol        = DextCmd->dexOutSpec_ls00;
    tls00_FileSpec       *myLongFileSpec = NULL;
    tls00_Column         *pColumn        = NULL;
    tls00_ParamInfo_LONG *pLongInfo      = NULL;

    tsp00_Int4           lLongFileCnt   = DextCmd->dexLongSpec_ls00->mfsCount_ls00;
    bool               * bLongIsDone    = NULL;     // Array fuer bereits bearbeitete longinfos des kommandos
    
    if (0 != lLongFileCnt)
    {
        if (NULL == bLongIsDone)
        {
            bLongIsDone = new bool[lLongFileCnt];
            memset(bLongIsDone, 0, (sizeof(bool) * lLongFileCnt) );
        }
    }


    for (i; i < MultCol.mcsCount_ls00; ++i)
    {
        pColumn = MultCol.mcsColumn_ls00[i];
        bFound  = false;

        // Literals may not be of type LONG
        if ( (false == pColumn->colIsLit_ls00) &&
             (1     == szMap_LongDataType_ls00[DextRec->pderColumnData_ls00[i].cdSQLDataType_ls00]) )
        {
            if (true == DextCmd->dexDatFile_ls00.fsFormat_ls00.ffoCompact_ls00)
            {   // in case of file format compact
                bFound = true;
            }
            else
            {
                // In der select-liste hat der benutzer eine spalte angegeben, die vom typ long ist.
                // Suche diese spalte auch in der spaltenliste.
                // Einmal abgearbeitete longs werden dabei ausgeschlossen.
                for (j = 0; ( (j < lLongFileCnt) && (false == bFound) ); ++j)
                {
                    if (false == bLongIsDone[j])
                    {
                        myLongFileSpec = DextCmd->dexLongSpec_ls00->mfsFiles_ls00[j];

                        if (NULL != myLongFileSpec->fsLongColName_ls00)     // Spaltennamen vergleichen
                        {
                            // Ist der vergleichsname auch wirklich ungleich NULL oder hat der benutzer in
                            // der spaltenliste keine namen sondern nummern angegeben? In diesem fall
                            // ist der string pColumn->colName_ls00 NULL und ein vergleich wuerde zu einem
                            // fehler fuehren
                            if (NULL == pColumn->colName_ls00.asCharp())    
                            {
                                // Spaltenname ist NULL; versuche es noch mit der spalten-ID
                                if (pColumn->colNumber_ls00 == myLongFileSpec->fsLongColID_ls00)
                                {
                                    bFound = bLongIsDone[j] = true;
                                }
                            }
                            else
                            {
                                if (0 == _tcsicmp(pColumn->colName_ls00.asCharp(), myLongFileSpec->fsLongColName_ls00))
                                {
                                    bFound = bLongIsDone[j] = true;
                                }
                            }
                        }
                        else                                            // Spalten-Id's vergleichen
                        {
                            if (pColumn->colNumber_ls00 == myLongFileSpec->fsLongColID_ls00)
                            {
                                bFound = bLongIsDone[j] = true;
                            }
                        }
                    }
                }   // end for (j = 0; ( (j < lLongFileCnt) && (false == bFound) ); ++j)
            }   // end if (true == DextCmd->dexDatFile_ls00.fsFormat_ls00.ffoCompact_ls00)

            if (true == bFound)
            {
                // Initialisierung der benoetigten werte. Wir verlassen uns an dieser stelle darauf,
                // dass im parser diese struktur bereits auf 0 gesetzt wurde fuer alle variablen.
                pLongInfo = &pColumn->colLongInfo;

                pLongInfo->loiFileOffset_ls00.lnuRemainder_ls00  = 1L;
                pLongInfo->loiFileOffset_ls00.lnu2GB_Blocks_ls00 = 0L;
                pLongInfo->loiGreatestNum_ls00 = 1;
                if (true == DextCmd->dexDatFile_ls00.fsFormat_ls00.ffoCompact_ls00)
                {
                    pLongInfo->loiFileName_ls00     = DextCmd->dexDatFile_ls00.fsFileName_ls00;
                    pLongInfo->loiFileEncoding_ls00 = DextCmd->dexDatFile_ls00.fsFormat_ls00.ffoCodeType_ls00;

                }
                else
                {
                    pLongInfo->loiFileEncoding_ls00 = myLongFileSpec->fsFormat_ls00.ffoCodeType_ls00;
                
                    pLongInfo->loiFileName_ls00 = myLongFileSpec->fsFileName_ls00;

                    //*
                    //*     Datei durchnumeriert oder mit offset ?
                    //*
                    tsp00_Int4 tmpFileNameLen = STATIC_CAST(tsp00_Int4, _tcslen(pLongInfo->loiFileName_ls00) );

                    while ( ('#' == pLongInfo->loiFileName_ls00[tmpFileNameLen - 1]) &&
                            (tmpFileNameLen > 0) )
                    {
                        ++pLongInfo->loiNumberSize_ls00;
                        --tmpFileNameLen;
                    }
                    if (0 != pLongInfo->loiNumberSize_ls00)
                    {
                        // Durchnumerierte Datei. Bestimme die groesste moegliche nummer!
                        for (tsp00_Int4 l = 0; l < pLongInfo->loiNumberSize_ls00; ++l)
                        {
                            pLongInfo->loiGreatestNum_ls00 *= 10;
                        }
                        pLongInfo->loiGreatestNum_ls00     -= 1;

                        pLongInfo->loibFileNumbered_ls00   = true;

                        // Ist bei datenformat FORMATTED die angegebene laenge kleiner als der
                        // dateiname, wird hier ein fehler generiert (hex wird auch beachtet).
                        if (true == DextCmd->dexDatFile_ls00.fsFormat_ls00.ffoFormatted_ls00)
                        {
                            tsp00_Int4 lValLen = pColumn->colFPos_ls00.fpoEnd_ls00 - 
                                                 (pColumn->colFPos_ls00.fpoStart_ls00 - 1);
                            tmpFileNameLen += pLongInfo->loiNumberSize_ls00;

                            if (true == pColumn->colFormat_ls00.ffoIsHex_ls00)
                            {
                                tmpFileNameLen += tmpFileNameLen;
                            }

                            if (tmpFileNameLen > lValLen)
                            {
                                pMsgObj->ls98PrintMessage(layDataExtract_els98, errWrongFieldLength_els98,
                                                                                pColumn->colName_ls00.asCharp(),
                                                                                tmpFileNameLen);
                                ++lErrNum;
                            }
                        }
                    }
                    // else-zweig ist mit initialisierung bereits fertig
                }

                ++DextRec->derLongColCnt_ls00;
            }
            else
            {
                // szenario 1
                pMsgObj->ls98PrintMessage(layDataExtract_els98, errNoLongFileInfo_els98,
                                                                pColumn->colName_ls00.asCharp());
                ++lErrNum;
            }   // end if (true == bFound)
        }   // end if ( (false == pColumn->colIsLit_ls00) && ...
    }       // end for (i; i < MultCol.mcsCount_ls00; ++i)


    // szenario 2: Warnungen in protokolldatei und ok zurueckgeben.
    if ( ( (0 == DextRec->derLongColCnt_ls00) && (0 != lLongFileCnt) ) ||
         (DextRec->derLongColCnt_ls00 < lLongFileCnt) )
    {
        for (j = 0; j < lLongFileCnt; ++j)
        {
            if (false == bLongIsDone[j])
            {
                // long file infos specified in command but no column assigned -> generate warning
                myLongFileSpec = DextCmd->dexLongSpec_ls00->mfsFiles_ls00[j];
                if (NULL != myLongFileSpec->fsLongColName_ls00)
                {
                    pMsgObj->ls98PrintMessage(layDataExtract_els98, errSurplusLongInfo_els98,
                                                                    myLongFileSpec->fsLongColName_ls00);
                }
                else
                {
                    char sztmpStr[10];  // nur notwendig, um die fehlerausgabe hinzubekommen
                    sztmpStr[0] = '\0';
                    _itot(myLongFileSpec->fsLongColID_ls00, sztmpStr, 10);
                    pMsgObj->ls98PrintMessage(layDataExtract_els98, errSurplusLongInfo_els98,
                                                                    sztmpStr);
                }
            }
        }   // end for (j = 0; j < lLongFileCnt; ++j)
    }

    if (0 != lErrNum)
    {
        rc = errAnalyzeLongSum_els98;
        pMsgObj->ls98MsgToString(layDataExtract_els98, pszErrText, errAnalyzeLongSum_els98);
    }

    if (NULL != bLongIsDone)
    {
        delete [] bLongIsDone;
        bLongIsDone = NULL;
    }

    return rc;
}
// ls27_AnalyzeLongInfo


/*
  -----------------------------------------------------------------------------
  function:     ls27_CloseFiles
  -----------------------------------------------------------------------------
*/
static void
ls27_CloseFiles(tls00_MultColSpec *pMultCol,
                tls00_DextractRec *pDERec,
                bool               bCompactFormat,
                tsp00_Addr         pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls27_CloseFiles");

    tsp00_Int4 rc = errOK_els00;

    if (false == pDERec->derTabEmpty_ls00)
    {
        if (NULL != pDERec->derDataFile_ls00)
        {
            ls30VFClose(*pDERec->derDataFile_ls00, pszErrText);
            delete pDERec->derDataFile_ls00;
            pDERec->derDataFile_ls00 = NULL;
        }
    
        // Close long files
        if (0 != pDERec->derLongColCnt_ls00)
        {
            tls00_Column *pCol = NULL;
            
            for (int i = 0; i < pMultCol->mcsCount_ls00; ++i)
            {
                pCol = pMultCol->mcsColumn_ls00[i];

                // Literals may not be of type LONG
                if ( (false == pCol->colIsLit_ls00) &&
                     (1 == szMap_LongDataType_ls00[pDERec->pderColumnData_ls00[i].cdSQLDataType_ls00]) )
                {
                    // In case of compact format we use the data file for LONG
                    // which has already been closed.
                    if (false == bCompactFormat)
                    {
                        if (NULL != pCol->colLongInfo.loiFileStruct_ls00)
                        {
                            ls30VFClose(*pCol->colLongInfo.loiFileStruct_ls00, pszErrText);
                            delete pCol->colLongInfo.loiFileStruct_ls00;
                            pCol->colLongInfo.loiFileStruct_ls00 = NULL;
                    }                   }
                }
            }   // end for (i; i < pDECmd->dexColSpec_ls00.mcsCount_ls00; ++i)
        }
    }

    // At the end close cmd file
    if (NULL != pDERec->derCmdFile_ls00)
    {
        ls30VFClose(*pDERec->derCmdFile_ls00, pszErrText);
        delete pDERec->derCmdFile_ls00;
        pDERec->derCmdFile_ls00 = NULL;
    }
        
    pszErrText[0] = 0;
    return;
}
// ls27_CloseFiles()


/*
  -----------------------------------------------------------------------------
  function:     ls27_InitAndOpenFiles
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls27_InitAndOpenFiles(tls00_DatExtrCmd  *pDECmd,
                      tls00_DextractRec *pDERec,
                      tsp00_Addr         pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls27_InitAndOpenFiles");

    tsp00_Int4 rc = errOK_els00;

    if (false == pDERec->derTabEmpty_ls00)
    {
        tls00_FileSpec *pDataFileSpec = &pDECmd->dexDatFile_ls00;

        //*
        //*     Initialize and open data file
        //*
        //* Initialize as binary file if output format is COMPACT or FORMATTED BINARY
        //*
        pDERec->derDataFile_ls00 = new tls00_VFile;
        memset(pDERec->derDataFile_ls00, 0, sizeof(tls00_VFile));

        if ( (true == pDataFileSpec->fsFormat_ls00.ffoBinary_ls00) ||
             (true == pDataFileSpec->fsFormat_ls00.ffoCompact_ls00) )
        {   // Binary file
            ls30VFInitialize(*pDERec->derDataFile_ls00, pDataFileSpec->fsFileName_ls00, SP5VF_BINARY, SP5VF_WRITE);
        }
        else
        {   // Text file
            ls30VFInitialize(*pDERec->derDataFile_ls00, pDataFileSpec->fsFileName_ls00, SP5VF_TEXT, SP5VF_WRITE);
        }

        rc = ls30VFOpen(*pDERec->derDataFile_ls00, pszErrText);
    }

    //*
    //*     at the end open file that will contain the commands
    //* only if table is not empty or NODDL is not true
    //*
    if ( (errOK_els00 == rc)                           &&
         (resUnknown_els00 != pDECmd->dexRestore_ls00) &&
         ( (false == pDERec->derTabEmpty_ls00) || (pDECmd->dexRestore_ls00 < resDataUpdt_els00) ) )
    {
        pDERec->derCmdFile_ls00 = new tls00_VFile;
        if (true == pDECmd->dexCmdFile_ls00.fsAppend_ls00)
        {
            ls30VFInitialize(*pDERec->derCmdFile_ls00, pDECmd->dexCmdFile_ls00.fsFileName_ls00, SP5VF_TEXT, SP5VF_APPEND);
        }
        else
        {
            ls30VFInitialize(*pDERec->derCmdFile_ls00, pDECmd->dexCmdFile_ls00.fsFileName_ls00, SP5VF_TEXT, SP5VF_WRITE);
        }

        rc = ls30VFOpen(*pDERec->derCmdFile_ls00, pszErrText);
    }

    return rc;
}
// ls27_InitAndOpenFiles()


/*
  -----------------------------------------------------------------------------
  function:     ls27_InitAndOpenLongFiles
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls27_InitAndOpenLongFiles(tls00_MultColSpec  *pMultCol,
                          tls00_ColumnData   *pColumnData,
                          tls00_VFile       *&pDataFile,
                          bool                bCompactFormat,
                          tsp00_Addr          pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls27_InitAndOpenLongFiles");

    tsp00_Int4 rc = errOK_els00;
    
    tls00_Column *pCol = NULL;

    for (int i = 0; i < pMultCol->mcsCount_ls00; ++i)
    {
        pCol = pMultCol->mcsColumn_ls00[i];

        // Literals may not be of type LONG
        if ( (false == pCol->colIsLit_ls00) &&
             (1 == szMap_LongDataType_ls00[pColumnData[i].cdSQLDataType_ls00]) )
        {
            if (true == bCompactFormat)
            {
                // In case of compact format we use the data file for LONG, too.
                pCol->colLongInfo.loiFileStruct_ls00 = pDataFile;
            }
            else
            {
                pCol->colLongInfo.loiFileStruct_ls00 = new tls00_VFile;
                memset(pCol->colLongInfo.loiFileStruct_ls00, 0, sizeof(tls00_VFile));

                // Open the long value file at this point only in case it is not numbered;
                // numbered means every long value of a column is saved in its own file
                if ( (false == pCol->colLongInfo.loibFileNumbered_ls00) &&
                     (NULL  != pCol->colLongInfo.loiFileStruct_ls00)    &&
                     (0     == pCol->colLongInfo.loiFileStruct_ls00->vfFileHndl_ls00) )
                {
                    // Init as binary file
                    ls30VFInitialize(*pCol->colLongInfo.loiFileStruct_ls00, pCol->colLongInfo.loiFileName_ls00,
                                     SP5VF_BINARY, SP5VF_WRITE);

                    rc = ls30VFOpen(*pCol->colLongInfo.loiFileStruct_ls00, pszErrText);
                }
            }
        }
    }   // end for (i; i < pDECmd->dexColSpec_ls00.mcsCount_ls00; ++i)

    return rc;
}
// ls27_InitAndOpenLongFiles()


/*
  -----------------------------------------------------------------------------
  EndChapter: Local functions implementation
  -----------------------------------------------------------------------------
*/
