/*! 
  -----------------------------------------------------------------------------
 
  module: vls28_TLTable.cpp
 
  -----------------------------------------------------------------------------
 
  responsible:  SteffenS and d033893

  special area: Replication Server
 
  description:  Implementation of TABLELOAD for single table
                
  version:      7.3.

  -----------------------------------------------------------------------------
 
  copyright:    SAP AG, 2000-2003
 
  -----------------------------------------------------------------------------


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*/

#include "gls00.h"
#include "hls01.h"
#include "hls28_TLTable.h"
#include "hls26_TETable.h"
#include "hls28_TLSwap.h"
#include "hls03.h"
#include "hls04.h"
#include "hls05.h"      // UCS2 - ASCII conversion routines
#include "hls041.h"
#include "hls07.h"
#include "hls13.h"
#include "hls15.h"
#include "hls30.h"
#include "hls51.h"
#include "hls98msg.h"
#include "hls99.h"
#include "gbd00.h"


#if defined(_WIN32)
#define NEW_LINE "\r\n"
#else
#define NEW_LINE "\n"
#endif

#define _INTERNAL_FILE_ID_    _T("28_1")


char const *pszDataType_ls28[39] = {"fixed",    "float",       "char asc", "char ebc",     "char byte", "rowid",
                                    "long asc", "long ebc",    "long byt", "9",            "date",      "time",
                                    "12",       "timestmp",    "14",       "15",           "16",        "17",
                                    "18",       "19",          "20",       "21",           "22",        "boolean",
                                    "char uni", "25",          "26",       "27",           "28",        "smallint",
                                    "integer",  "varchar asc", "vchr ebc", "varchar byte", "long uni",  "35",
                                    "vchr uni", "37",          "38"};

char const *pszLengthType_ls28[3] = {"fix length", "variable length", "variable length long"};

char const *pszColumnMode_ls28[5] = {"syskey", "key", "mandatory", "optional", "dropped"};

char const *pszDefaultType_ls28[12] = {"no default",    "user",     "usergroup", "date", "time",
                                       "stamp",         "timestamp","uid",       "true", "false",
                                       "other default", "serial"};

/*  
  -----------------------------------------------------------------------------
  Chapter: Private functions declaration
  -----------------------------------------------------------------------------
*/


/*
  -----------------------------------------------------------------------------
 
  function:     ls28_GetSourceTableDesc
  -----------------------------------------------------------------------------

  description:  Reads table description (saved in format of tsp7_table_description)
                from data file and swaps the integers in the description.
                Sets the values tlrSwapFromDb_ls00 in pTLRec and (in case TL ALL
                and TL USER) psFullTabName_ls00, psTabName_ls00 in 
                TLCmd->tloPartSpec_ls00.
                Replaces table name and table owner in table description.

  arguments:    pTLRec       [IN]     - pointer; structure, holding information for
                                         executing TABLELOAD for a single table
                pRange       [IN|OUT] - pointer; structure with information about
                                        type of TL (ALL, USER, TABLE) and values
                                        for table name and table owner
                myReadBuffer [IN]     - reference; read file buffer object
                pszErrText   [OUT]    - pointer; error text

  returnvalue:  errOK_els00              - no error
                errWrongTEFileType_els98 - corrupted data file (no table desc found)
                some error from reading the file

                The function logs the error message to the log file and to
                the error text return buffer.
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_GetSourceTableDesc(tls00_DBInfo            *pDBInfo,
                        tls00_TloadRec          *pTLRec,
                        SAPDB_DynamicUTF8String  &UserNameUTF8,
                        SAPDB_DynamicUTF8String  &TableNameUTF8,
                        tls00_PartSpec          *pRange,
                        ReadBuffer              &myReadBuffer,
                        tsp00_Addr               pszErrText);


/*
  -----------------------------------------------------------------------------
 
  function:     ls28_GetSourceIndexDesc
  -----------------------------------------------------------------------------

  description:  Reads index description (saved in format of tsp7_index_description)
                from data file and swaps the integers in the description.

  arguments:    pTLRec       [IN]  - pointer; structure, holding information for
                                      executing TABLELOAD for a single table
                myReadBuffer [IN]  - reference; read file buffer object
                pszErrText   [OUT] - pointer; error text

  returnvalue:  errOK_els00              - no error
                some error from reading the file

                The function logs the error message to the log file and to
                the error text return buffer.
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_GetSourceIndexDesc(tls00_TloadRec *pTLRec,
                        ReadBuffer     &myReadBuffer,
                        tsp00_Addr      pszErrText);


/*
  -----------------------------------------------------------------------------
 
  function:     ls28_CheckTableExistence
  -----------------------------------------------------------------------------

  description:  Checks the existence of table <owner name>.<table name> in
                target db.

  arguments:    pDBInfo       [IN]  - contains all information on the running
                                       session and the database connected to.
                TableOwner    [IN]  - table owner UTF8 encoded
                TableName     [IN]  - table name UTF8 encoded
                bTableExists  [OUT] - indicates existence of table
                pszErrText    [OUT] - error text

  returnvalue:  errOK_els00        - no error
                errInternSQL_els98 - error executing command that checks the
                                      existence
                The function logs the error message to the log file and to
                the error text return buffer.
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_CheckTableExistence(tls00_DBInfo            *pDBInfo,
                         SAPDB_DynamicUTF8String &TableOwner,
                         SAPDB_DynamicUTF8String &TableName,
                         bool                    &bTableExists,
                         tsp00_Addr               pszErrText);


/*
  -----------------------------------------------------------------------------
 
  function:     ls28_CreateTable
  -----------------------------------------------------------------------------

  description:  Creates a table on db (including constraints and defaults)

  arguments:    pDBInfo      [IN]  - contains all information on the running
                                      session and the database connected to.
                pTLRec       [IN]  - structure, holding information for
                                      executing TABLELOAD for a single table
                TableName    [IN]  - table name UTF8 encoded
                myReadBuffer [OUT] - read file buffer object
                pszErrText   [OUT] - error text

  returnvalue:  errOK_els00        - no error
                errInternSQL_els98 - error executing commands to create the 
                                      different table objects
                The function logs the error message to the log file and to
                the error text return buffer.
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_CreateTable(tls00_DBInfo            *pDBInfo,
                 tls00_TloadRec          *pTLRec,
                 SAPDB_DynamicUTF8String &OwnerName,
                 SAPDB_DynamicUTF8String &TableName,
                 ReadBuffer              &myReadBuffer,
                 tsp00_Addr               pszErrText);


/*
  -----------------------------------------------------------------------------
 
  function:     ls28_ReplaceTable
  -----------------------------------------------------------------------------

  description:  Drops the table an recreates it on db (including constraints
                and defaults)

  arguments:    pDBInfo      [IN]  - contains all information on the running
                                      session and the database connected to.
                pTLRec       [IN]  - structure, holding information for
                                      executing TABLELOAD for a single table
                TableName    [IN]  - table name UTF8 encoded
                myReadBuffer [OUT] - read file buffer object
                pszErrText   [OUT] - error text

  returnvalue:  errOK_els00        - no error
                errInternSQL_els98 - error executing commands to create the 
                                      different table objects
                The function logs the error message to the log file and to
                the error text return buffer.
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_ReplaceTable(tls00_DBInfo            *pDBInfo,
                  tls00_TloadRec          *pTLRec,
                  SAPDB_DynamicUTF8String &OwnerName,
                  SAPDB_DynamicUTF8String &TableName,
                  ReadBuffer              &myReadBuffer,
                  tsp00_Addr               pszErrText);


/*
  -----------------------------------------------------------------------------
 
  function:     ls28_RestoreTable
  -----------------------------------------------------------------------------

  description:  Not yet implemented. See body for some explanations


  arguments:    pDBInfo      [IN]   - contains all information on the running
                                       session and the database connected to
                pTLRec       [IN]   - structure, holding information for
                                       executing TABLELOAD for a single table
                OwnerName    [IN]   - owner name UTF8 encoded
                TableName    [IN]   - table name UTF8 encoded
                myReadBuffer [IN]   - read file buffer object
                pszErrText   [OUT]  - error text

  returnvalue:  errOK_els00     No error
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_RestoreTable(tls00_DBInfo            *pDBInfo,
                  tls00_TloadRec          *pTLRec,
                  SAPDB_DynamicUTF8String &OwnerName,
                  SAPDB_DynamicUTF8String &TableName,                 
                  ReadBuffer              &myReadBuffer,
                  tsp00_Addr               pszErrText);


/*
  -----------------------------------------------------------------------------
 
  function:     ls28_CompareTableSchemes
  -----------------------------------------------------------------------------

  description:  Compare table description of source and target table.


  arguments:    pDBInfo      [IN]   - contains all information on the running
                                       session and the database connected to
                pTLRec       [IN]   - structure, holding information for
                                       executing TABLELOAD for a single table
                TableName    [IN]   - table name UTF8 encoded
                AdaptMode    [OUT]  - returns adapting mode for table
                pszErrText   [OUT]  - error text

  returnvalue:  errOK_els00     No error
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_CompareTableSchemes(tls00_DBInfo               *pDBInfo,
                         tls00_TloadRec             *pTLRec,
                         SAPDB_DynamicUTF8String    &OwnerName,
                         SAPDB_DynamicUTF8String    &TableName,
                         tls28_TableSchemaAdaptMode &AdaptMode,
                         tsp00_Addr                 pszErrText);

/*
  -----------------------------------------------------------------------------
 
  function:     ls28_RestoreShortStrings
  -----------------------------------------------------------------------------

  description:  Restores short strings in table to be loaded.

  arguments:    pDBInfo      [IN]  - contains all information on the running
                                      session and the database connected to.
                pTLRec       [IN]  - structure, holding information for
                                      executing TABLELOAD for a single table
                TableName    [IN]  - table name UTF8 encoded
                PageBuffer   [IN]  - read file buffer object
                pszErrText   [OUT] - error text

  returnvalue:  errOK_els00              - no error
                errWrongTEFileType_els98 - manipulated short string header page ?
                errInternSQL_els98       - error executing commands to load the shorts
                errInternal_els98        - error loading the pages

                The function logs the error message to the log file and to
                the error text return buffer.
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_RestoreShortStrings(tls00_DBInfo            *pDBInfo,
                         tls00_TloadRec          *pTLRec,
                         SAPDB_DynamicUTF8String &OwnerName,
                         SAPDB_DynamicUTF8String &TableName,
                         ReadBuffer              &PageBuffer,
                         tsp00_Addr               pszErrText);

/*
  -----------------------------------------------------------------------------
 
  function:     ls28_RestorePrimaryData
  -----------------------------------------------------------------------------

  description:  Restores primary data in table to be loaded.

  arguments:    pDBInfo      [IN]  - contains all information on the running
                                      session and the database connected to.
                pTLRec       [IN]  - structure, holding information for
                                      executing TABLELOAD for a single table
                TableName    [IN]  - table name UTF8 encoded
                PageBuffer   [IN]  - read file buffer object
                pszErrText   [OUT] - error text

  returnvalue:  errOK_els00              - no error
                errWrongTEFileType_els98 - manipulated primary data header page ?
                errInternSQL_els98       - error executing commands to load the shorts
                errInternal_els98        - error loading the pages

                The function logs the error message to the log file and to
                the error text return buffer.
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_RestorePrimaryData(tls00_DBInfo             *pDBInfo,
                        tls00_TloadRec           *pTLRec,
                        SAPDB_DynamicUTF8String  &OwnerName,
                        SAPDB_DynamicUTF8String  &TableName,
                        ReadBuffer               &PageBuffer,
                        tsp00_Addr                pszErrText);


/*
  -----------------------------------------------------------------------------
 
  function:     ls28_RestoreLongStrings
  -----------------------------------------------------------------------------

  description:  Restores primary data in table to be loaded.

  arguments:    pDBInfo      [IN]  - contains all information on the running
                                      session and the database connected to.
                pTLRec       [IN]  - structure, holding information for
                                      executing TABLELOAD for a single table
                PageBuffer   [IN]  - read file buffer object
                pszErrText   [OUT] - error text

  returnvalue:  errOK_els00              - no error
                errWrongTEFileType_els98 - manipulated primary data header page ?
                errInternSQL_els98       - error executing commands to load the shorts
                errInternal_els98        - error loading the pages

                The function logs the error message to the log file and to
                the error text return buffer.
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_RestoreLongStrings(tls00_DBInfo    *pDBInfo,
                        tls00_TloadRec  *pTLRec,
                        ReadBuffer      &PageBuffer,
                        tsp00_Addr       pszErrText);


//*****************************************************************************
//*
//*         local support functions
//*
//*****************************************************************************

/*
  -----------------------------------------------------------------------------
 
  function:     ls28_ReadDescFromFile
  -----------------------------------------------------------------------------

  description:  Extracts table|index description from data file.

  arguments:    myReadBuffer    [IN]  - read file buffer object
                DescriptionType [IN]  - enumeration; type of description to extract
                SourceSwap      [IN]  - swap of source system; is used to read the header
                                         of the page containing the description
                lPageSize       [IN]  - source system page size; used to read the header of
                                         the page containing the description
                pDescription    [OUT] - refrence to pointer; takes the description and
                                         its length
                pszErrText      [OUT] - pointer; error text

  returnvalue:  errOK_els00
                WRONG_TE_HEADER_LS28   - description type page not found (caller must
                                            handle it as error or warning)
                EOF_LS00                 - eof reached reading the description (highly unlikely)
                
                some error from reading the file - only these errors will be logged
                                                   to the log file and return text
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_ReadDescFromFile(ReadBuffer       &myReadBuffer,
                      tls00_TEPageType  DescriptionType,
                      tsp00_SwapKind    SourceSwap,
                      tsp00_Int4        lPageSize,      // source page size, but source and target size may not differ!
                      tls00_String     &pDescription,
                      tsp00_Addr        pszErrText);


/*
  -----------------------------------------------------------------------------
 
  function:     ls28_CheckTypeHeaderPage
  -----------------------------------------------------------------------------

  description:  Checks the headers on pages for every extracted type of data.
                The page size is necessary because for some types of data a header
                and a footer are written and both are identically so they can
                be compared.
                Swaps the integers in the header in place.

  arguments:    pszBuffer       [IN] - pointer; buffer of size page size which
                                        contains the header
                lPageSize       [IN] - source system page size; size of pszBuffer
                PageType        [IN] - type of data we're looking for
                SourceSwapKind  [IN] - swap kind used to swap integers

  returnvalue:  errOK_els00
                UNKNOWN_TE_HEADER_LS28 - not a known header - only for DATA of type
                                            primary data (tptData_els00), short string
                                            (tptShortString_els00) and long
                                            string (tptLongString_els00)
                WRONG_TE_HEADER_LS28   - for descriptions it is an error (w/o table
                                            description from source restore is not
                                            possible; for data this can be a message
                                            that no data of this type was extracted
                                            during TABLEEXTRACT
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_CheckTypeHeaderPage(tsp00_Addr       pszBuffer,
                         tsp00_Int4       lPageSize,
                         tls00_TEPageType PageType,
                         tsp00_SwapKind   SourceSwapKind);

/*
  -----------------------------------------------------------------------------
 
  function:     ls28_ReplaceTableSchema
  -----------------------------------------------------------------------------

  description:  Replaces table schema if schemas don't match or creates table
                if it does not exist.


  arguments:    pDBInfo      [IN]   - 
                pTLRec       [IN]   - 
                pszTableName [IN]   - 
                myReadBuffer [IN]   - 
                pszErrText   [OUT]  - Fehlertext

  returnvalue:  errOK_els00     No error
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_ReplaceTableSchema(tls00_DBInfo   *pDBInfo,
                        tls00_TloadRec *pTLRec,
                        tsp00_Addr      pszTableName,
                        ReadBuffer      myReadBuffer,
                        tsp00_Addr      pszErrText);


/*
  -----------------------------------------------------------------------------
 
  function:     ls28_Default_ConstraintFromFile
  -----------------------------------------------------------------------------

  description:  Reads defaults/constraints from extract file.
                Holt den wert raw raus, ohne ihn zu interpretieren



  arguments:    pDBInfo     [IN]
                pTLRec      [IN]   - 
                PageBuffer  [IN|OUT]
                ReadInfo    [IN|OUT]   - 
                PageType    [IN|OUT]  - 
                pValue      [OUT]
                pszErrText  [OUT]  - Fehlertext

  returnvalue:  errOK_els00     No error
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_Default_ConstraintFromFile(tls00_DBInfo       *pDBInfo,
                                tls00_TloadRec     *pTLRec,
                                ReadBuffer         &PageBuffer,
                                tls28_PageReadInfo *ReadInfo,
                                tls00_TEPageType    PageType,
                                tls00_String       *pValue,
                                tsp00_Addr          pszErrText);


/*
  -----------------------------------------------------------------------------
 
  function:     ls28_CreateDefault
  -----------------------------------------------------------------------------

  description:  Creates defaults using 'INTO DEFAULT <table name>...'.

  arguments:    

  returnvalue:  errOK_els00
                other
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_CreateDefault(tls00_DBInfo            *pDBInfo,
                   SAPDB_DynamicUTF8String &TableName,
                   tls00_String            *pValue,
                   tsp00_Addr               pszErrText);


/*
  -----------------------------------------------------------------------------
 
  function:     ls28_CreateConstraint
  -----------------------------------------------------------------------------

  description:  Creates constraints using 'INTO CONSTRAINT <table name>...'.

  arguments:    

  returnvalue:  errOK_els00
                other
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_CreateConstraint(tls00_DBInfo            *pDBInfo,
                      SAPDB_DynamicUTF8String &TableName,
                      tls00_String            *pValue,
                      tsp00_Addr               pszErrText);


/*
  -----------------------------------------------------------------------------
 
  function:     ls28_BeginRestore
  -----------------------------------------------------------------------------

  description:  Executes the command "restore long|table". Stores the bdinfo
                coming from kernel as answer in pszBdinfo.

  arguments:    pDBInfo    [IN]  - contains all information on the running
                                      session and the database connected to.
                pszCmd     [IN]  - command (restore long or restore table);
                                    UTF8 encoded
                pszBdinfo  [OUT] - buffer holding bdinfo coming from kernel
                pszErrText [OUT] - error text

  returnvalue:  errOK_els00              - no error
                errInternSQL_els98       - error executing command
                errInternal_els98        - error getting bdinfo

                The function logs the error message to the log file and to
                the error text return buffer.
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_BeginRestore(tls00_DBInfo *pDBInfo,
                  tsp00_Addr    pszCmd,
                  tsp7_bdinfo  &pszBdinfo,
                  tsp00_Addr    pszErrText);


/*
  -----------------------------------------------------------------------------
 
  function:     ls28_DataLoadLoop
  -----------------------------------------------------------------------------

  description:  Main module to execute the page load. Used for all kind of data.

  arguments:    pDBInfo    [IN]     - contains all information on the running
                                       session and the database connected to.
                pTLRec     [IN]     - structure, holding information for
                                       executing TABLELOAD for a single table
                ReadBuffer [IN]     - read file buffer object
                pszBdinfo  [IN|OUT] - buffer holding bdinfo for first packet to
                                       send and of last answer packet; the latter
                                       is used to execute "end restore" command
                pszErrText [OUT]    - error text

  returnvalue:  errOK_els00              - no error
                errInternal_els98        - error sending/getting page to/from kernel
                error reading the data file

                The function logs the error message to the log file and to
                the error text return buffer.
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_DataLoadLoop(tls00_DBInfo   *pDBInfo,
                  tls00_TloadRec *pTLRec,
                  ReadBuffer     &PageBuffer,
                  tsp7_bdinfo    &pszBdInfo,
                  tsp00_Addr      pszErrText);


/*
  -----------------------------------------------------------------------------
 
  function:     ls28_FillPacketBuffer
  -----------------------------------------------------------------------------

  description:  Fills a (request) packet with data pages. Initializes the packet to take
                the data AND the bdinfo.
                Used by ls28_DataLoadLoop.

  arguments:    pDBInfo    [IN] - contains all information on the running
                                      session and the database connected to.
                pTLRec     [IN] - structure, holding information for
                                   executing TABLELOAD for a single table
                lPcktNo    [IN] - number of packet to fill
                ReadBuffer [IN] - read file buffer object
                pszErrText [OUT] - error text


  returnvalue:  errOK_els00 - no error
                STOP_LS28   - no more data found; process last packet
                errors from reading file
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_FillPacketBuffer(tls00_DBInfo   *pDBInfo,
                      tls00_TloadRec *pTLRec,
                      tsp00_Int4      lPcktNo,
                      ReadBuffer     &PageBuffer,
                      tsp00_Addr      pszErrText);


/*
  -----------------------------------------------------------------------------
 
  function:     ls28_EndRestore
  -----------------------------------------------------------------------------

  description:  Executes the command "end restore"

  arguments:    pDBInfo    [IN]  - contains all information on the running
                                    session and the database connected to.
                pszBdinfo  [IN]  - buffer holding bdinfo needed to execute command
                pszErrText [OUT] - error text

  returnvalue:  errOK_els00
                errors from executing the command

                The function logs the error message to the log file and to
                the error text return buffer.
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_EndRestore(tls00_DBInfo *pDBInfo, tsp7_bdinfo pszBdinfo, tsp00_Addr pszErrText);

/*
  -----------------------------------------------------------------------------
 
  function:     ls28_BeginRestoreLong
  -----------------------------------------------------------------------------

  description:  Executes command "restore column". Is called for every single
                row in table that contains long values.

  arguments:    pDBInfo         [IN]  - contains all information on the running
                                         session and the database connected to.
                PartStruct      [IN]  - structure, holding data for parts to send
                                         to kernel (necessary for executing the cmd)
                pszColumnBdinfo [OUT] - bdinfo for row that has long values
                pszErrText      [OUT] - error text

  returnvalue:  errOK_els00       - no error
                errInternal_els98 - error executing command

                The function logs the error message to the log file and to
                the error text return buffer.
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_BeginRestoreLong(tls00_DBInfo            *pDBInfo,
                      tls00_Parts4InternalSql &PartStruct,
                      tsp7_bdinfo             &pszColumnBdinfo,
                      tsp00_Addr               pszErrText);


/*  
  -----------------------------------------------------------------------------
  EndChapter: Private functions declaration
  -----------------------------------------------------------------------------
*/


/*
  -----------------------------------------------------------------------------
  Chapter: Public functions implementation
  -----------------------------------------------------------------------------
*/

/*
  -----------------------------------------------------------------------------
  function:     ls28LoadSingleTable
  -----------------------------------------------------------------------------
*/
tsp00_Int4
ls28LoadSingleTable(tls00_DBInfo            *pDBInfo,
                    tls00_TabLoadCmd        *TLCmd,
                    tls00_TloadRec          *TLRec,
                    SAPDB_DynamicUTF8String &OwnerName,
                    SAPDB_DynamicUTF8String &TableName,
                    tsp00_Addr               pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls28LoadSingleTable");

    tsp00_Int4 rc = errOK_els00;
    
    //*
    //*     initialize read buffer class
    //*
    ReadBuffer myReadBuffer(TLRec->tlrInFile_ls00, TLRec->tlrPageSize_ls00, TLRec->tlrPgPerPckt_ls00);

    //*
    //*     Get table description
    //*
    if (errOK_els00 == rc)
    {
        rc = ls28_GetSourceTableDesc(pDBInfo,   TLRec,                    OwnerName,
                                     TableName, &TLCmd->tloPartSpec_ls00, myReadBuffer, pszErrText);
    }


    //*
    //*     Get index description
    //*
    if (errOK_els00 == rc)
    {
        rc = ls28_GetSourceIndexDesc(TLRec, myReadBuffer, pszErrText);
    }


    //*
    //*     Check existence of table in target system
    //* In case of TL USER or TL ALL we need the table description of the source
    //* table do execute the check.
    //*
    bool bTableExists = false;
    if (errOK_els00 == rc)
    {
        rc = ls28_CheckTableExistence(pDBInfo, OwnerName, TableName, bTableExists, pszErrText);
        if (errOK_els00 == rc)
        {
            if (false == bTableExists)
            {
                rc = ls28_CreateTable(pDBInfo, TLRec, OwnerName, TableName, myReadBuffer, pszErrText);
            }
            else
            {
                tls28_TableSchemaAdaptMode AdaptMode = tsaIncopyNone_ls28;
                rc = ls28_CompareTableSchemes(pDBInfo, TLRec, OwnerName, TableName, AdaptMode, pszErrText);
                if (errOK_els00 == rc)
                {
                    switch (AdaptMode)
                    {
                        case tsaIncopyAll_ls28:
                        {
                            // drops the table and calls ls28_CreateTable
                            rc = ls28_ReplaceTable(pDBInfo, TLRec, OwnerName, TableName, myReadBuffer, pszErrText);
                            break;
                        }
                        case tsaIncopySome_ls28:
                        {
                            // adapts the fields and defaults, constraints
                            rc = ls28_RestoreTable(pDBInfo, TLRec, OwnerName, TableName, myReadBuffer, pszErrText);
                            // at first we believe that coming here all schema definitions of source and
                            // target match; so we do nothing

                            break;
                        }
                        default:
                        {
                            // do nothing
                        }
                    }
                }
            }
        }   // end if (errOK_els00 == rc) for ls28_CheckTableExistence
    }

    // Check for cancel request
    if (errOK_els00 == rc)
    {
        rc = ls01CheckForCancel(pDBInfo->dbiSession_ls00.sipCancelByte_ls00);
    }

    
    // In case the table schema has been successfully restored we save that
    if (errOK_els00 == rc)
    {
        rc = ls15Commit(pDBInfo, pszErrText);
    }

    //*
    //*     Restore short string data
    //*
    if (errOK_els00 == rc)
    {
        TLRec->tlrPageTypeExtracting_ls00 = tptShortString_els00;
        
        rc = ls28_RestoreShortStrings(pDBInfo, TLRec, OwnerName, TableName, myReadBuffer, pszErrText);
    }

    if (errOK_els00 == rc)
    {
        rc = ls01CheckForCancel(pDBInfo->dbiSession_ls00.sipCancelByte_ls00);   // Check for cancel request
    }

    //*
    //*     Restore primary data
    //*
    if (errOK_els00 == rc)
    {
        TLRec->tlrPageTypeExtracting_ls00 = tptData_els00;
        
        rc = ls28_RestorePrimaryData(pDBInfo, TLRec, OwnerName, TableName, myReadBuffer, pszErrText);
    }

    
    if (errOK_els00 == rc)
    {
        rc = ls01CheckForCancel(pDBInfo->dbiSession_ls00.sipCancelByte_ls00);   // Check for cancel request
    }
    
    //*
    //*     Restore long string data
    //*
    if (errOK_els00 == rc)
    {
        TLRec->tlrPageTypeExtracting_ls00 = tptLongString_els00;

        rc = ls28_RestoreLongStrings(pDBInfo, TLRec, myReadBuffer, pszErrText);
    }

    if (errOK_els00 == rc)
    {
        rc = ls15Commit(pDBInfo, pszErrText);
    }

    if (errOK_els00 == rc)
    {
        rc = ls01CheckForCancel(pDBInfo->dbiSession_ls00.sipCancelByte_ls00);   // Check for cancel request
    }

    //*
    //*     Create indexes
    //*
    if (errOK_els00 == rc)
    {
        rc = ls13CreateIndexes(pDBInfo,
                               REINTERPRET_CAST(tsp7_table_description*, TLRec->tlrExtrTableInfo_ls00.strAddr_ls00),
                               REINTERPRET_CAST(tsp7_index_description*, TLRec->tlrExtrIndexInfo_ls00.strAddr_ls00),
                               TableName,
                               pszErrText);
    }

    if (errOK_els00 == rc)
    {
        rc = ls15Commit(pDBInfo, pszErrText);
    }

    // Write message in case of cancel request
    if (errCancelRequest_els98 == rc)
    {
        RepServMessages::Instance()->ls98Msg(layTableLoad_els98, pszErrText, errCancelRequest_els98);
    }

    return rc;
}
// ls28LoadSingleTable()


/*
  -----------------------------------------------------------------------------
  function:     ls28ReadDataFileHeader
  -----------------------------------------------------------------------------
*/
tsp00_Int4
ls28ReadDataFileHeader(tls00_TloadRec *pTLRec, tsp00_Addr pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls28ReadDataFileHeader");

    RepServMessages  *MsgInstance = RepServMessages::Instance();    // Fr error reports
    tsp00_Int4        rc          = errOK_els00;
    _TCHAR           *pszBuffer   = NULL;

    // Header at extract
    //----------------------------------------------------------------------------------------------------
    // repserver | code |            |         |      |      | page size    | count of        | more than
    // swap at   | type | identifier | version | date | time | of extracted | pages in        | one table
    // extract   |      |            |         |      |      | db           | file (blockung) | in file ?
    //----------------------------------------------------------------------------------------------------
    // 1 byte    |  1   |      64    | 40      | 8    | 8    |    4         |  4              |   1


    int lReadLen = 0;
    lReadLen = ls30VFReadLn(*pTLRec->tlrInFile_ls00, &pszBuffer, pszErrText, TE_FILE_HEADER_LENGTH_LS00);
    if (lReadLen >= 0)
    {
        size_t lBufPos = 0;
        
        pTLRec->tlrSwapFromTE_ls00.becomes(tsp00_SwapKind::fromInt(STATIC_CAST(int, pszBuffer[lBufPos])));

        lBufPos += sizeof(unsigned char);


        // Get encoding type of source database
        pTLRec->tlrSourceEncoding_ls00 = STATIC_CAST(tls00_CodeType, pszBuffer[lBufPos]);
        lBufPos += sizeof(unsigned char);

        // Check if file created by Load Server
        int result = memcmp(&pszBuffer[lBufPos], HEADER_IDENTIFIER_LS00, sizeof(HEADER_IDENTIFIER_LS00));
        if (0 != result)
        {
            MsgInstance->ls98Msg(layTableLoad_els98, pszErrText, errWrongTEFileType_els98,
                                                                 pTLRec->tlrInFile_ls00->vfFileName_ls00,
                                                                 "wrong identifier");
            rc = errWrongTEFileType_els98;
        }
        else
        {
            lBufPos += 64 + 40 + 8 + 8;     // Skip version, date and time
            tsp00_Int4 lExtractedPageSize = ls07Swap4BHexToLong(&pszBuffer[lBufPos],
                                                                pTLRec->tlrSwapFromTE_ls00,
                                                                ls07DefineSwap(NULL));

            if (lExtractedPageSize != pTLRec->tlrPageSize_ls00)
            {
                MsgInstance->ls98Msg(layTableLoad_els98, pszErrText, errWrongTEPageSize_els98,
                                                                     lExtractedPageSize,
                                                                     pTLRec->tlrPageSize_ls00);
                rc = errWrongTEPageSize_els98;
            }
            else
            {
                lBufPos += sizeof(tsp00_Int4);
                pTLRec->tlrCount_ls00 = ls07Swap4BHexToLong(&pszBuffer[lBufPos],
                                                            pTLRec->tlrSwapFromTE_ls00,
                                                            ls07DefineSwap(NULL));

                lBufPos += sizeof(tsp00_Int4);
                pTLRec->tlrMultTabInFile_ls00 = (pszBuffer[lBufPos] != 0);
            }
        }
    }
    else
    {
        rc = lReadLen;
    }   // end if (lReadLen >= 0)

    return rc;
}
// ls28ReadDataFileHeader()

/*
  -----------------------------------------------------------------------------
  EndChapter: Public functions implementation
  -----------------------------------------------------------------------------
*/

/*
  -----------------------------------------------------------------------------
  Chapter: Private functions implementation
  -----------------------------------------------------------------------------
*/
/*
  -----------------------------------------------------------------------------
  function:     ls28_GetSourceTableDesc
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_GetSourceTableDesc(tls00_DBInfo            *pDBInfo,
                        tls00_TloadRec          *pTLRec,
                        SAPDB_DynamicUTF8String  &UserNameUTF8,
                        SAPDB_DynamicUTF8String  &TableNameUTF8,
                        tls00_PartSpec          *pRange,
                        ReadBuffer              &myReadBuffer,
                        tsp00_Addr               pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls28_GetSourceTableDesc");

    RepServMessages *pMsgInstance = RepServMessages::Instance();    // for error reports

    tsp00_Int4 rc = errOK_els00;

    rc = ls28_ReadDescFromFile(myReadBuffer,
                               tptTableDescription_els00,
                               pTLRec->tlrSwapFromTE_ls00,
                               pTLRec->tlrPageSize_ls00,   
                               pTLRec->tlrExtrTableInfo_ls00,
                               pszErrText);
    if (errOK_els00 == rc)
    {
        tsp7_table_description *pTD = REINTERPRET_CAST(tsp7_table_description*, pTLRec->tlrExtrTableInfo_ls00.strAddr_ls00);

        // get (source db-)kernel swap
        pTLRec->tlrSwapFromDb_ls00 = pTD->sp7t_kernel_swap();

        // now swap the integers in description if necessary
        ls28SwapTableDescription(*pTD, pTLRec->tlrSwapFromTE_ls00);

        TRACE_PRNF_MLS99(("ls28_GetSourceTableDesc", "Source DB kernel swap = %d\n", STATIC_CAST(int, pTD->sp7t_kernel_swap())));

        //*
        //*     Replace user name and table name in table description and fill
        //*     psFullTabName_ls00
        //*
        //* TL ALL or TL USER:
        //*     - table name in table description has the right value -> check it and report error in case
        //*     - user name in table description needs to be replaced (we do not compare it before replacing)
        //*
        //* TL TABLE <table name>:
        //*     - table name in table description must be replaced (no comparing beforehand)
        //*     - user name in table description must be replaced (no comparing beforehand)
        //*     - psFullTabName_ls00 and psTabName_ls00 have already the right values


        //*
        //*     Replace user name in any case
        //*
        tsp00_KnlIdentifierc szUserName;
        tsp00_Longuint       lUserNameLen  = sizeof(tsp00_KnlIdentifierc);
        tsp00_KnlIdentifierc szTableName;
        tsp00_Longuint       lTableNameLen = sizeof(tsp00_KnlIdentifierc);
        tls00_CodeType       DestCodeType  = ctAscii_els00;
        tsp00_DataType       DT;
        DT.becomes(dcha);

        // Define if DB is UCS2
        bool bDBIsUnicode = ls00IsDBUnicode(pDBInfo);
        if (true == bDBIsUnicode)
        {
            DestCodeType = ctUCS2Native_els00;
        }
        
        // Fill szTableName with blanks for comparing later on
        ls05InsertFillBytes(REINTERPRET_CAST(unsigned char*,szTableName.asCharp()),
                            0,
                            DT,
                            sizeof(tsp00_KnlIdentifierc),
                            bDBIsUnicode,
                            pDBInfo->dbiOwnSwap_ls00);

        rc = ls05ConvertFromUTF8(REINTERPRET_CAST(unsigned char*, szUserName.asCharp()), lUserNameLen,
                                 UserNameUTF8, DestCodeType, pszErrText);
        if (errOK_els00 == rc)
        {
            rc = ls05ConvertFromUTF8(REINTERPRET_CAST(unsigned char*, szTableName.asCharp()), lTableNameLen,
                                     TableNameUTF8, DestCodeType, pszErrText);;
        }

        if (errOK_els00 == rc)
        {
            //*
            //*     Replace user name
            //*
            tsp00_KnlIdentifier &UserName = pTD->sp7t_authid;

            ls05InsertFillBytes(REINTERPRET_CAST(unsigned char*,UserName.asCharp()),
                                0,
                                DT,
                                sizeof(tsp00_KnlIdentifier),
                                bDBIsUnicode,
                                pDBInfo->dbiOwnSwap_ls00);
            memcpy(UserName.asCharp(), szUserName.asCharp(), lUserNameLen);


            //*
            //*     Replace/check table name
            //*
            tsp00_KnlIdentifier &TabName = pTD->sp7t_tablename;

            if ( (true == pRange->psSelAll_ls00) || (true == pRange->psSelUser_ls00) )
            {   // check table name
                if (0 != memcmp(TabName.asCharp(),szTableName.asCharp(), sizeof(tsp00_KnlIdentifier)) )
                {
                    // error, table not at this position in file report this and quit
                    pMsgInstance->ls98Msg(layTableLoad_els98, pszErrText, errTLTableDataNotFound_els98,
                                                                          pRange->psTabName_ls00.asCharp(),
                                                                          pTLRec->tlrInFile_ls00->vfFileName_ls00);
                    rc = errTLTableDataNotFound_els98;
                }
            }
            else
            {   // replace table name
                ls05InsertFillBytes(REINTERPRET_CAST(unsigned char*,TabName.asCharp()),
                                    0,
                                    DT,
                                    sizeof(tsp00_KnlIdentifier),
                                    bDBIsUnicode,
                                    pDBInfo->dbiOwnSwap_ls00);
                memcpy(TabName.asCharp(), szTableName.asCharp(), lTableNameLen);
            }
        }
    }
    else
    {
        if ( (WRONG_TE_HEADER_LS28 == rc) || (EOF_LS00 == rc) )
        {
            pMsgInstance->ls98Msg(layTableLoad_els98, pszErrText, errWrongTEFileType_els98,
                                                                  pTLRec->tlrInFile_ls00->vfFileName_ls00,
                                                                  "missing table description");
            rc = errWrongTEFileType_els98;
        }
    }

    return rc;
}
// ls28_GetSourceTableDesc()


/*
  -----------------------------------------------------------------------------
  function:     ls28_GetSourceIndexDesc
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_GetSourceIndexDesc(tls00_TloadRec *pTLRec,
                        ReadBuffer     &myReadBuffer,
                        tsp00_Addr      pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls28_GetSourceIndexDesc");

    tsp00_Int4 rc = errOK_els00;

    rc = ls28_ReadDescFromFile(myReadBuffer,
                               tptIndexDescription_els00,
                               pTLRec->tlrSwapFromTE_ls00,
                               pTLRec->tlrPageSize_ls00,
                               pTLRec->tlrExtrIndexInfo_ls00,
                               pszErrText);
    if (errOK_els00 == rc)              // now swap the integers in description if necessary
    {
        ls28SwapIndexDescription(*(REINTERPRET_CAST(tsp7_index_description*, pTLRec->tlrExtrIndexInfo_ls00.strAddr_ls00)),
                                 pTLRec->tlrSwapFromTE_ls00);
    }
    else
    {
        if ( (WRONG_TE_HEADER_LS28 == rc) || (EOF_LS00 == rc) )
        {
            rc = errOK_els00;
        }
    }

    return rc;
}
// ls28_GetSourceIndexDesc()


/*
  -----------------------------------------------------------------------------
  function:     ls28_CheckTableExistence
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_CheckTableExistence(tls00_DBInfo            *pDBInfo,
                         SAPDB_DynamicUTF8String &TableOwner,
                         SAPDB_DynamicUTF8String &TableName,
                         bool                    &bTableExists,
                         tsp00_Addr               pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls28_CheckTableExistence");

    tsp00_Int4 rc = errOK_els00;

    //*
    //*     Check existence
    //* In case that table does not exist we get the error number -4004.
    //*
    bTableExists = false;

    _TCHAR szCmd[lStmtLength_ls26];
    szCmd[0] = 0;

    sp77sprintf(szCmd, lStmtLength_ls26, EXISTS_TABLE_STMT_LS28, TableOwner.StrPtr(), TableName.StrPtr());

    tsp00_Int4 lErrPos;
    _TCHAR     szErr[MAX_REPLY_LENGTH_LS00];

    rc = ls13ExecIntSQLStmtUTF8(pDBInfo, 0, szCmd, szErr, lErrPos);
    if (errOK_els00 != rc)
    {
        if (dbUnknownTableName_els00 == rc)
        {
            rc = errOK_els00;
        }
        else
        {
            RepServMessages::Instance()->ls98Msg(layTableLoad_els98, pszErrText, errInternSQL_els98,
                                                                                 szCmd, szErr);
            rc = errInternSQL_els98;
        }
    }
    else
    {
        bTableExists = true;
    }

    return rc;
}
// ls28_CheckTableExistence()


/*
  -----------------------------------------------------------------------------
  function:     ls28_CreateTable
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_CreateTable(tls00_DBInfo            *pDBInfo,
                 tls00_TloadRec          *pTLRec,
                 SAPDB_DynamicUTF8String &OwnerName,
                 SAPDB_DynamicUTF8String &TableName,
                 ReadBuffer              &myReadBuffer,
                 tsp00_Addr               pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls28_CreateTable");

    tsp00_Int4 rc = errOK_els00;

    // tabellenname in tabellenbeschreibung ist bereits angepasst
    //*
    //*     Create table using binary table description
    //*
    _TCHAR szCmd[lStmtLength_ls26];
    szCmd[0] = 0;
    sp77sprintf(szCmd, lStmtLength_ls26, INTO_TABLE_STMT_LS28, OwnerName.StrPtr(), TableName.StrPtr());

    tsp00_Int4 lErrorPos = 0;
    _TCHAR     szErr[MAX_REPLY_LENGTH_LS00];
    rc = ls13ExecIntSQLStmtExUTF8(pDBInfo,
                                  0,
                                  szCmd,
                                  pTLRec->tlrExtrTableInfo_ls00.strAddr_ls00,
                                  pTLRec->tlrExtrTableInfo_ls00.strLeng_ls00,
                                  szErr,
                                  lErrorPos);
    if (errOK_els00 != rc)
    {
        RepServMessages::Instance()->ls98Msg(layTableLoad_els98, pszErrText, errInternSQL_els98,
                                                                             szCmd, szErr);
        rc = errInternSQL_els98;
    }
    else
    {
        //*
        //*     create constraints and defaults
        //*
        tls28_PageReadInfo myReadInfo;
        tls00_String Value;

        tsp00_Int4 rc2 = errOK_els00;
    
        myReadInfo.priCount_ls28  = 0;
        myReadInfo.priMore_ls28   = 0;
        myReadInfo.priBuffer_ls28 = NULL;
    
        //*
        //*     Create constraints
        //*
        do
        {
            rc2 = ls28_Default_ConstraintFromFile(pDBInfo,
                                                  pTLRec,
                                                  myReadBuffer,
                                                  &myReadInfo,
                                                  tptConstraint_els00,
                                                  &Value,
                                                  pszErrText);
            if ( (errOK_els00 == rc2) || (STOP_LS28 == rc2) )
            {
                rc = ls28_CreateConstraint(pDBInfo, TableName, &Value, pszErrText);
            }
        }
        while ( (rc >= errOK_els00) && (rc2 >= errOK_els00) );
    
        if ( (errOK_els00 != rc2) && (STOP_LS28 != rc2) && (STOP_NODATA_LS28 != rc2) )
        {
            rc = rc2;
        }


        //*
        //*     Create defaults
        //*
        if (errOK_els00 == rc)
        {
            myReadInfo.priCount_ls28  = 0;
            myReadInfo.priMore_ls28   = 0;
            myReadInfo.priBuffer_ls28 = NULL;

            rc2 = errOK_els00;
            do
            {
                rc2 = ls28_Default_ConstraintFromFile(pDBInfo,
                                                      pTLRec,
                                                      myReadBuffer,
                                                      &myReadInfo,
                                                      tptDefault_els00,
                                                      &Value,
                                                      pszErrText);
                if ( (errOK_els00 == rc2) || (STOP_LS28 == rc2) )
                {
                    rc = ls28_CreateDefault(pDBInfo, TableName, &Value, pszErrText);
                }
            }
            while ( (rc >= errOK_els00) && (rc2 >= errOK_els00) );
        
            if ( (errOK_els00 != rc2) && (STOP_LS28 != rc2) && (STOP_NODATA_LS28 != rc2) )
            {
                rc = rc2;
            }
        }       // end if (errOK_els00 == rc) - create defaults
    }   // end if (errOK_els00 != rc) - ls13ExecIntSQLStmtEx

    //*
    //*     Close transaction of creating table with the END LOAD statement
    //*
    if (errOK_els00 == rc)
    {
        szCmd[0]  = 0;
        lErrorPos = 0;

        sp77sprintf(szCmd, lStmtLength_ls26, END_LOAD_TABLE_STMT_LS28, TableName.StrPtr());
    
        rc = ls13ExecIntSQLStmtUTF8(pDBInfo, 0, szCmd, szErr, lErrorPos);
        if (errOK_els00 != rc)
        {
            RepServMessages::Instance()->ls98Msg(layTableLoad_els98, pszErrText, errInternSQL_els98,
                                                                                 szCmd, szErr);
            rc = errInternSQL_els98;
        }
    }
    
    return rc;
}
// ls28_CreateTable()


/*
  -----------------------------------------------------------------------------
   function:     ls28_ReplaceTable
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_ReplaceTable(tls00_DBInfo            *pDBInfo,
                  tls00_TloadRec          *pTLRec,
                  SAPDB_DynamicUTF8String &OwnerName,
                  SAPDB_DynamicUTF8String &TableName,
                  ReadBuffer              &myReadBuffer,
                  tsp00_Addr               pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls28_ReplaceTable");

    tsp00_Int4 rc = errOK_els00;

/*
    if ( ('7' == pDBInfo->dbiDBVersion_ls00.MajorVersion_sp100) &&
         ('2' != pDBInfo->dbiDBVersion_ls00.MinorVersion_sp100) )
    {
        _TCHAR szCmd[lStmtLength_ls26];
        szCmd[0] = 0;
        _tcscpy(szCmd, DROP_TABLE_STMT_LS28);
        strcat(szCmd, pszTableName);

        tsp00_Int4 lErrPos;
        _TCHAR     szErr[MAX_REPLY_LENGTH_LS00];
        rc = ls13ExecIntSQLStmt(pDBInfo, 0, szCmd, szErr, lErrPos);
        if (errOK_els00 != rc)
        {
            RepServMessages::Instance()->ls98Msg(layTableLoad_els98, pszErrText, errInternSQL_els98,
                                                                                 szCmd, szErr);
            rc = errInternSQL_els98;
        }
    }
*/
    RepServMessages::Instance()->ls98Msg(layTableLoad_els98, pszErrText, errTLTableNotExisting_els98,
                                                                         TableName.StrPtr());
    rc = errTLTableNotExisting_els98;

    
    if (errOK_els00 == rc)
    {
        rc = ls28_CreateTable(pDBInfo, pTLRec, OwnerName, TableName, myReadBuffer, pszErrText);
    }
    
    return rc;
}
// ls28_ReplaceTable()


/*
  -----------------------------------------------------------------------------
   function:     ls28_RestoreTable
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_RestoreTable(tls00_DBInfo            *pDBInfo,
                  tls00_TloadRec          *pTLRec,
                  SAPDB_DynamicUTF8String &OwnerName,
                  SAPDB_DynamicUTF8String &TableName,
                  ReadBuffer              &myReadBuffer,
                  tsp00_Addr               pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls28_RestoreTable");

    tsp00_Int4 rc = errOK_els00;

    // first delete content of table if there is some
    _TCHAR szCmd[lStmtLength_ls26];
    szCmd[0] = 0;
    sp77sprintf(szCmd, lStmtLength_ls26, EMPTY_TABLE_STMT_LS28, OwnerName.StrPtr(), TableName.StrPtr());

    tsp00_Int4 lErrPos;
    _TCHAR     szErr[MAX_REPLY_LENGTH_LS00];
    rc = ls13ExecIntSQLStmtUTF8(pDBInfo, 0, szCmd, szErr, lErrPos);
    if (errOK_els00 != rc)
    {
        if (dbRowNotFound_els00 == rc)      // table is empty
        {
            rc = errOK_els00;
        }
        else
        {
            RepServMessages::Instance()->ls98Msg(layTableLoad_els98, pszErrText, errInternSQL_els98,
                                                                                 szCmd, szErr);
            rc = errInternSQL_els98;
        }
    }
    else        // there is some content - delete it
    {
        szCmd[0] = 0;
        sp77sprintf(szCmd, lStmtLength_ls26, DELETE_TABLE_STMT_LS28, OwnerName.StrPtr(), TableName.StrPtr());

        rc = ls13ExecIntSQLStmtUTF8(pDBInfo, 0, szCmd, szErr, lErrPos);
        if (errOK_els00 != rc)
        {
            RepServMessages::Instance()->ls98Msg(layTableLoad_els98, pszErrText, errInternSQL_els98,
                                                                                 szCmd, szErr);
            rc = errInternSQL_els98;
        }   
    }
    
    // commit changes
    if (errOK_els00 == rc)
    {
        rc = ls15Commit(pDBInfo, pszErrText);
    }


    // compare defaults and constraints and replace them if necessary: aim is the structure of the source table

    // FIRST ATTEMPT IS TO IGNORE DEFAULTS AND CONSTRAINTS. THEREFORE WE NEED TO SKIP EXTRACTED PAGES WHICH
    // CONTAIN DEFAULT AND/OR CONSTRAINT INFORMATION

    if (errOK_els00 == rc)
    {
        //* constraints
        tsp00_Int4         rc2        = errOK_els00;
        tls28_PageReadInfo myReadInfo;
        tls00_String       Value;
        
        myReadInfo.priCount_ls28  = 0;
        myReadInfo.priMore_ls28   = 0;
        myReadInfo.priBuffer_ls28 = NULL;

        do
        {
            rc2 = ls28_Default_ConstraintFromFile(pDBInfo, pTLRec, myReadBuffer, &myReadInfo,
                                                  tptConstraint_els00, &Value, pszErrText);
        }
        while (rc2 >= errOK_els00);
        
        if ( (errOK_els00 != rc2) && (STOP_LS28 != rc2) && (STOP_NODATA_LS28 != rc2) )
        {
            rc = rc2;
        }
        
        //* defaults
        if (errOK_els00 == rc)
        {
            myReadInfo.priCount_ls28  = 0;
            myReadInfo.priMore_ls28   = 0;
            myReadInfo.priBuffer_ls28 = NULL;

            rc2 = errOK_els00;
            do
            {
                rc2 = ls28_Default_ConstraintFromFile(pDBInfo, pTLRec, myReadBuffer, &myReadInfo,
                                                      tptDefault_els00, &Value, pszErrText);
            }
            while (rc2 >= errOK_els00);
            if ( (errOK_els00 != rc2) && (STOP_LS28 != rc2) && (STOP_NODATA_LS28 != rc2) )
            {
                rc = rc2;
            }
        }
    }

    //*
    //*     Compare index descriptions and mark the one that needs to be replaced
    //*
    if (errOK_els00 == rc)
    {
        //*
        //* Get target table description, and description of indexes on target table, too
        //*
        tsp00_Int4 lTabDescLen   = 0;
        tsp00_Addr pszTabDesc    = NULL;
        tsp00_Int4 lIndexDescLen = 0;
        tsp00_Addr pszIndexDesc  = NULL;

        rc = ls13GetTableDescription(pDBInfo, 0, TableName, lTabDescLen, pszTabDesc, pszErrText);
        if (errOK_els00 == rc)
        {
            rc = ls13GetIndexDescription(pDBInfo, 1, TableName, lIndexDescLen, pszIndexDesc, pszErrText);
        }

        if (errOK_els00 == rc)
        {
            //* Compare index descriptions (to be implemented)

            // drop existing indexes
            rc = ls13DropAllIndexesOnTable(pDBInfo,
                                           REINTERPRET_CAST(tsp7_table_description*, pszTabDesc),
                                           REINTERPRET_CAST(tsp7_index_description*, pszIndexDesc),
                                           TableName,
                                           pszErrText);
            // commit changes
            if (errOK_els00 == rc)
            {
                rc = ls15Commit(pDBInfo, pszErrText);
            }
        }
    }

    return rc;
}
// ls28_RestoreTable()


/*
  -----------------------------------------------------------------------------
  function:     ls28_CompareTableSchemes
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_CompareTableSchemes(tls00_DBInfo               *pDBInfo,
                         tls00_TloadRec             *pTLRec,
                         SAPDB_DynamicUTF8String    &OwnerName,
                         SAPDB_DynamicUTF8String    &TableName,
                         tls28_TableSchemaAdaptMode &AdaptMode,
                         tsp00_Addr                 pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls28_CompareTableSchemes");

    RepServMessages *pMsgObj = RepServMessages::Instance();
    tsp00_Int4       rc      = errOK_els00;

    // variables used to hold binary description of target table
    tsp00_Int4 lDescLen = 0;
    tsp00_Addr pszDesc  = NULL;

    //*
    //*     Get binary table description of target table
    //*
    rc = ls13GetTableDescription(pDBInfo, 0, OwnerName, TableName, lDescLen, pszDesc, pszErrText);
    if (errOK_els00 == rc)
    {
        char txt[1024];
        txt[0] = 0;

        //*
        //*     Compare table schemes
        //*
        tsp7_table_description *SourceTabDesc = REINTERPRET_CAST(tsp7_table_description*, pTLRec->tlrExtrTableInfo_ls00.strAddr_ls00);
        tsp7_table_description *TargetTabDesc = REINTERPRET_CAST(tsp7_table_description*, pszDesc);

        //*
        //*     First comparison: column count (dropped columns are counted??)
        //*
        if (SourceTabDesc->sp7t_field_cnt() != TargetTabDesc->sp7t_field_cnt())
        {
            sp77sprintf(txt, 1024, "different field count: target = %d, source = %d",
                                   TargetTabDesc->sp7t_field_cnt(), SourceTabDesc->sp7t_field_cnt());
            pMsgObj->ls98PrintMessage(layTableLoad_els98, errTLSchemeDifferences_els98, TableName.StrPtr(), txt);
            txt[0] = 0;

            AdaptMode = tsaIncopyAll_ls28;
            
            // That's all we need to know to stop
        }
        else
        {
            //*
            //*     Column count of target and source table are equal; now compare
            //*     the length of the binary descriptions
            //*
            if (pTLRec->tlrExtrTableInfo_ls00.strLeng_ls00 == lDescLen)
            {
                //*
                //*     Length of binary descriptions is equal; now compare
                //*     the table description structures
                //* Doing that log the differences in definition of table in source and target system
                //*
                if (0 != memcmp(SourceTabDesc, TargetTabDesc, lDescLen))
                {
                    AdaptMode = tsaIncopyAll_ls28;

                    tsp00_Longint lError    = 0;
                    tsp00_Longint lNoError  = 0;

                    //*
                    //*     Kernel swap (sp7t_kernel_swap()) - must not be equal
                    //*
                    //*  different swap is handled by load server internally

                    //*
                    //*     Table kind (syskey or not that's the question) - must be equal
                    //*
                    if (SourceTabDesc->sp7t_table_kind() != TargetTabDesc->sp7t_table_kind())
                    {
                        sp77sprintf(txt, 1024, "table kind: target = %d, source = %d",
                                               TargetTabDesc->sp7t_table_kind(), SourceTabDesc->sp7t_table_kind());
                        pMsgObj->ls98PrintMessage(layTableLoad_els98, errTLSchemeDifferences_els98, TableName.StrPtr(), txt);
                        txt[0] = 0;

                        lError++;
                    }

                    // sp7t_kernel_loc: not relevant - leftover from times of distributed database

                    //*
                    //*     Constraint count - must be equal
                    //*
                    if (SourceTabDesc->sp7t_constraint_cnt() != TargetTabDesc->sp7t_constraint_cnt())
                    {
                        sp77sprintf(txt, 1024, "constraint count: target = %d, source = %d",
                                               TargetTabDesc->sp7t_table_kind(), SourceTabDesc->sp7t_table_kind());
                        pMsgObj->ls98PrintMessage(layTableLoad_els98, errTLSchemeDifferences_els98, TableName.StrPtr(), txt);
                        txt[0] = 0;

                        lError++;
                    }

                    
                    tsp7_ffield_info         *pSrcField           = NULL;
                    tsp7_ffield_info         *pTgtField           = NULL;
                    SAPDB_DynamicUTF8String  *pTgtColumnName_UTF8 = new SAPDB_DynamicUTF8String[TargetTabDesc->sp7t_field_cnt() + 1];

                    //*
                    //*     Define encoding type for names in table descriptions;
                    //*     because we allow only extracting/loading between databases of the same
                    //*     encoding type we can define the encoding type here using the target database
                    //*
                    tls00_CodeType SrcEncoding = ctAscii_els00;
                    if (true == ls00IsDBUnicode(pDBInfo))
                    {
                        SrcEncoding = ctUCS2_els00;     // not swapped!
                    }

                    //*
                    //*     Get target table column names; they're needed in any case for all subsequent comparisons
                    //*
                    rc = ls13GetColumnNamesInRecordOrder(TargetTabDesc, pTgtColumnName_UTF8, SrcEncoding, pszErrText);

                    //*
                    //*     Compare column sorting in the record structure in src and tgt table
                    //* We assume that columns in src and tgt table are equally named.
                    //* If sorting is not equal we can stop here.
                    //*
                    tsp00_Int4  lColNameBufLen   = lDescLen - sizeof(tsp7_ffield_info) * TargetTabDesc->sp7t_field_cnt() -
                                                    sizeof(tsp7_head_buf) - 2*sizeof(tsp_knl_identifier);
                    _TCHAR     *pszTgtColNameBuf = TargetTabDesc->sp7t_colnames().asCharp() + sizeof(tsp7_ffield_info) * TargetTabDesc->sp7t_field_cnt();
                    _TCHAR     *pszSrcColNameBuf = SourceTabDesc->sp7t_colnames().asCharp() + sizeof(tsp7_ffield_info) * SourceTabDesc->sp7t_field_cnt();
                    if (0 != memcmp(pszTgtColNameBuf, pszSrcColNameBuf, lColNameBufLen))
                    {
                        // first increase error counter
                        lError++;

                        // create error message
                        tsp00_Int4 ErrTxtLen = 2*TargetTabDesc->sp7t_field_cnt()*sizeof(tsp00_KnlIdentifierc) + 1024;
                        tsp00_Int4 lBufCnt   = 0;
                        _TCHAR *pszColSortErrTxt = new _TCHAR[ErrTxtLen];
                        pszColSortErrTxt[0] = 0;

                        lBufCnt = sp77sprintf(pszColSortErrTxt, 1024, "different column sorting of source and target table");

                        //*
                        //*     Get source table column names
                        //*
                        SAPDB_DynamicUTF8String *pSrcColumnName_UTF8 = new SAPDB_DynamicUTF8String[SourceTabDesc->sp7t_field_cnt() + 1];
                        rc = ls13GetColumnNamesInRecordOrder(SourceTabDesc, pSrcColumnName_UTF8, SrcEncoding, pszErrText);


                        lBufCnt += sp77sprintf(pszColSortErrTxt + lBufCnt, ErrTxtLen - lBufCnt,
                                                                           ""NEW_LINE"/ %-32s%-32s",
                                                                           "source table columns", "target table columns");
                        for (int k = 0; k < TargetTabDesc->sp7t_field_cnt(); ++k)
                        {
                            lBufCnt += sp77sprintf(pszColSortErrTxt + lBufCnt, ErrTxtLen - lBufCnt,
                                                    ""NEW_LINE"/ %-32s%-32s", pSrcColumnName_UTF8[k].StrPtr(),
                                                                              pTgtColumnName_UTF8[k].StrPtr());
                        }

                        
                        pMsgObj->ls98PrintMessage(layTableLoad_els98, errTLSchemeDifferences_els98,
                                                                      TableName.StrPtr(), pszColSortErrTxt);
                        txt[0] = 0;

                        delete [] pSrcColumnName_UTF8;
                        delete [] pszColSortErrTxt;
                    }

                    //*
                    //*     Compare single column information
                    //*
                    if (0 == lError)
                    {
                        for (int i = 0; i < TargetTabDesc->sp7t_field_cnt(); ++i)
                        {
                            pSrcField = &(SourceTabDesc->sp7t_fields()[i]);
                            pTgtField = &(TargetTabDesc->sp7t_fields()[i]);

                            //*
                            //*     Data type - must be equal
                            //*
                            if ( tsp00_DataType_Enum(pSrcField->sp7fi_data_type) !=
                                 tsp00_DataType_Enum(pTgtField->sp7fi_data_type)  )
                            {
                                sp77sprintf(txt, 1024, "data type of %s: target = %s, source = %s",
                                                       pTgtColumnName_UTF8[i].StrPtr(),
                                                       pszDataType_ls28[pTgtField->sp7fi_data_type],
                                                       pszDataType_ls28[pSrcField->sp7fi_data_type]);
                                pMsgObj->ls98PrintMessage(layTableLoad_els98, errTLSchemeDifferences_els98, 
                                                                              TableName.StrPtr(), txt);
                                txt[0] = 0;

                                lError++;
                            }

                            //*
                            //*     Fraction - must be equal
                            //*
                            if (pSrcField->sp7fi_frac != pTgtField->sp7fi_frac)
                            {
                                sp77sprintf(txt, 1024, "fraction of %s: target = %d, source = %d",
                                                       pTgtColumnName_UTF8[i].StrPtr(),
                                                       pTgtField->sp7fi_frac, pSrcField->sp7fi_frac);
                                pMsgObj->ls98PrintMessage(layTableLoad_els98, errTLSchemeDifferences_els98,
                                                                              TableName.StrPtr(), txt);
                                txt[0] = 0;

                                lError++;
                            }

                            //*
                            //*     Column definition length - must be equal
                            //*
                            if (pSrcField->sp7fi_len != pTgtField->sp7fi_len)
                            {
                                sp77sprintf(txt, 1024, "length of %s: target = %d, source = %d",
                                                       pTgtColumnName_UTF8[i].StrPtr(),
                                                       pTgtField->sp7fi_len, pSrcField->sp7fi_len);
                                pMsgObj->ls98PrintMessage(layTableLoad_els98, errTLSchemeDifferences_els98,
                                                                              TableName.StrPtr(), txt);
                                txt[0] = 0;

                                lError++;
                            }

                            //*
                            //*     Deliver length using order interface (sp7fi_in_out_l) - must be equal
                            //*
                            //* But actually of no interest because the definition length says it all


                            //*
                            //*     Column mode - target = mandatory and source = optional is
                            //*                   the only allowed combination
                            //*
                            //* Possible values: mod_syskey
                            //*                  mod_key
                            //*                  mod_mandatory
                            //*                  mod_optional
                            //*                  mod_dropped
                            //*
                            if ( tsp00_ColMode_Enum(pSrcField->sp7fi_col_mode) != 
                                 tsp00_ColMode_Enum(pTgtField->sp7fi_col_mode)  )
                            {
                                if ( (mod_mandatory == pSrcField->sp7fi_col_mode) &&
                                     (mod_optional == pTgtField->sp7fi_col_mode)   )
                                {
                                    lNoError++;
                                }
                                else
                                {
                                    sp77sprintf(txt, 1024, "column mode of %s: target = %s, source = %s",
                                                           pTgtColumnName_UTF8[i].StrPtr(),
                                                           pszColumnMode_ls28[pTgtField->sp7fi_col_mode],
                                                           pszColumnMode_ls28[pSrcField->sp7fi_col_mode]);
                                    pMsgObj->ls98PrintMessage(layTableLoad_els98, errTLSchemeDifferences_els98,
                                                                                  TableName.StrPtr(), txt);
                                    txt[0] = 0;

                                    lError++;
                                }
                            }

                            //*
                            //*     Length type - must be equal
                            //*
                            //* Possible values: sp7_fix_len
                            //*                  sp7_var_len
                            //*                  sp7_var_len_long
                            //*
                            if ( tsp7_len_type_Enum(pSrcField->sp7fi_len_type) !=
                                 tsp7_len_type_Enum(pTgtField->sp7fi_len_type)  )
                            {
                                sp77sprintf(txt, 1024, "length type of %s: target = %s, source = %s",
                                                       pTgtColumnName_UTF8[i].StrPtr(),
                                                       pszLengthType_ls28[pTgtField->sp7fi_len_type],
                                                       pszLengthType_ls28[pSrcField->sp7fi_len_type]);
                                pMsgObj->ls98PrintMessage(layTableLoad_els98, errTLSchemeDifferences_els98,
                                                                              TableName.StrPtr(), txt);
                                txt[0] = 0;

                                lError++;
                            }
                            
                            //*
                            //*     Kind of default column has - must be equal
                            //*
                            //* Possible values: sp7_no_default
                            //*                  sp7_default_user
                            //*                  sp7_default_usergroup
                            //*                  sp7_default_date
                            //*                  sp7_default_time
                            //*                  sp7_default_stamp
                            //*                  sp7_default_timestamp
                            //*                  sp7_default_uid
                            //*                  sp7_default_true
                            //*                  sp7_default_false
                            //*                  sp7_other_default
                            //*                  sp7_default_serial
                            //*
                            if ( tsp7_default_Enum(pSrcField->sp7fi_default) !=
                                 tsp7_default_Enum(pTgtField->sp7fi_default)  )
                            {
                                sp77sprintf(txt, 1024, "default type of %s: target = %s, source = %s",
                                                       pTgtColumnName_UTF8[i].StrPtr(),
                                                       pszDefaultType_ls28[pTgtField->sp7fi_default],
                                                       pszDefaultType_ls28[pSrcField->sp7fi_default]);
                                pMsgObj->ls98PrintMessage(layTableLoad_els98, errTLSchemeDifferences_els98,
                                                                              TableName.StrPtr(), txt);
                                txt[0] = 0;

                                lError++;
                            }
                        }   // end for (int i = 0; i < TargetTabDesc->sp7t_field_cnt(); ++i)

                        // There is either lError != 0 or lNoError != 0 or both
                        if (0 != lError)
                        {
                            // Adapt mode has not been changed and has the right value: the defs are not equal
                            //AdaptMode = tsaIncopyAll_ls28;
                        }
                        else
                        {
                            if (0 != lNoError)
                            {
                                AdaptMode = tsaIncopySome_ls28;
                            }
                            else
                            {
                                // Adapt mode has not been changed and has the right value: the defs are not equal
                                //AdaptMode = tsaIncopyAll_ls28;
                                AdaptMode = tsaIncopySome_ls28;
                            }
                        }
                        

                    }   // end if (0 == lError)

                    delete [] pTgtColumnName_UTF8;
                }
                else
                {
//TODOTODO HIER MUESSTE EIGENTLICH tsaIncopyNone_ls28 ZURUECKGEGEBEN WERDEN ABER AUCH IN DIESEM FALL
//         MUESSEN ALLE INDIZES GELOESCHT WERDEN, WAS ABER NOCH NICHT PASSIERT
                    AdaptMode = tsaIncopySome_ls28;
                }   // end if (0 != memcmp(SourceTabDesc, TargetTabDesc, lDescLen))
            
            }   // end if (pTLRec->tlrExtrTableInfo_ls00.strLeng_ls00 == lDescLen)
            else
            {
                AdaptMode = tsaIncopyAll_ls28;
            }   // end if (pTLRec->tlrExtrTableInfo_ls00.strLeng_ls00 == lDescLen)
        
        }   // end else branch of if (SourceTabDesc->sp7t_field_cnt() != TargetTabDesc->sp7t_field_cnt())
        
        // Wenn jetzt auch noch die defaults und constraints uebereinstimmen
        //  INCOPYNONE;
    
    }   // end if (errOK_els00 == rc)

    return rc;
}
// ls28_CompareTableSchemes()


/*
  -----------------------------------------------------------------------------
  function:     ls28_RestoreShortStrings
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_RestoreShortStrings(tls00_DBInfo            *pDBInfo,
                         tls00_TloadRec          *pTLRec,
                         SAPDB_DynamicUTF8String &OwnerName,
                         SAPDB_DynamicUTF8String &TableName,
                         ReadBuffer              &PageBuffer,
                         tsp00_Addr               pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls28_RestoreShortStrings");

    tsp00_Int4  rc = errOK_els00;
    
    _TCHAR     *pszBuffer = PageBuffer.ls51GetNext();       // get next page from data file
    if (NULL == pszBuffer)                  // error or eof
    {
        rc = PageBuffer.ls51GetError(pszErrText);
        if (EOF_LS00 == rc)
        {
            rc = errOK_els00;
        }
    }
    else
    {
        rc = ls28_CheckTypeHeaderPage(pszBuffer, pDBInfo->dbiPageSize_ls00, tptShortString_els00, pTLRec->tlrSwapFromTE_ls00);
        if (errOK_els00 != rc)
        {
            // error numbers can be
            //  - WRONG_TE_HEADER_LS28   --> no shorts to restore
            //  - UNKNOWN_TE_HEADER_LS28 --> no shorts to restore
            //  - STOP_NODATA_LS28         --> no more data
            if (WRONG_TE_HEADER_LS28 == rc)
            {
                TRACE_PRNF_MLS99(("ls28_RestoreShortStrings", "No short strings.\n"));

                PageBuffer.ls51NotUsed();                       // mark page as 'not used'
                rc = errOK_els00;
            }
            else if (UNKNOWN_TE_HEADER_LS28 == rc)
            {
                RepServMessages::Instance()->ls98Msg(layTableLoad_els98, pszErrText, errWrongTEFileType_els98,
                                                                                     pTLRec->tlrInFile_ls00->vfFileName_ls00,
                                                                                     "manipulated short string header page");
                PageBuffer.ls51NotUsed();                       // mark page as 'not used'
                rc = errWrongTEFileType_els98;
            }
            // else if (STOP_NODATA_LS28 == rc)
            // --> no more data to process for this table; this branch must not be handled, return code has
            //     already the right value. It is highly unlikely that the return code becomes STOP_NODATA_LS28
            //     in this function because for every extracted table exists at least ONE page of primary data
            //     and this comes AFTER short string pages.
            //     But we cannot release the 'unused' header page here.

        }
        else
        {
            _TCHAR szCmd[lStmtLength_ls26];
            szCmd[0] = 0;

            sp77sprintf(szCmd, lStmtLength_ls26, RESTORE_STRING_STMT_LS28, OwnerName.StrPtr(), TableName.StrPtr());
//            sp77sprintf(szCmd, lStmtLength_ls26, RESTORE_STRING_STMT_LS28, TableName.StrPtr());

            rc = ls28_BeginRestore(pDBInfo, szCmd, pTLRec->tlrBdInfo_ls00, pszErrText);
            if (errOK_els00 == rc)
            {
                rc = ls28_DataLoadLoop(pDBInfo, pTLRec, PageBuffer, pTLRec->tlrBdInfo_ls00, pszErrText);
                if (errInternal_els98 == rc)
                {
                    RepServMessages::Instance()->ls98Msg(layTableExtract_els98, pszErrText, errInternal_els98,
                                                                                            _INTERNAL_FILE_ID_,
                                                                                            __LINE__,
                                                                                            STATIC_CAST(ErrorNumbers, rc));
                }
            }

            if (errOK_els00 == rc)
            {
                rc = ls28_EndRestore(pDBInfo, pTLRec->tlrBdInfo_ls00, pszErrText);
            }
        }
    }   // end else branch of if (NULL == pszBuffer)

    return rc;
}
// ls28_RestoreShortStrings()


/*
  -----------------------------------------------------------------------------
  function:     ls28_RestorePrimaryData
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_RestorePrimaryData(tls00_DBInfo            *pDBInfo,
                        tls00_TloadRec          *pTLRec,
                        SAPDB_DynamicUTF8String &OwnerName,
                        SAPDB_DynamicUTF8String &TableName,
                        ReadBuffer              &PageBuffer,
                        tsp00_Addr               pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls28_RestorePrimaryData");

    tsp00_Int4 rc = errOK_els00;

    _TCHAR *pszBuffer = PageBuffer.ls51GetNext();       // get next page from data file
    if (NULL == pszBuffer)                  // error or eof
    {
        rc = PageBuffer.ls51GetError(pszErrText);
        if (EOF_LS00 == rc)
        {
            rc = errOK_els00;
        }
    }
    else
    {
        rc = ls28_CheckTypeHeaderPage(pszBuffer, pDBInfo->dbiPageSize_ls00, tptData_els00, pTLRec->tlrSwapFromTE_ls00);
        if (errOK_els00 != rc)                          // There is no primary data to restore
        {
            // error numbers can be
            //  - WRONG_TE_HEADER_LS28   --> no primary data to restore
            //  - UNKNOWN_TE_HEADER_LS28 --> no primary data to restore
            //  - STOP_NODATA_LS28         --> no more data
            if (WRONG_TE_HEADER_LS28 == rc)
            {
                TRACE_PRNF_MLS99(("ls28_RestorePrimaryData", "No primary data.\n"));

                PageBuffer.ls51NotUsed();                   // mark page as 'not used'
                rc = errOK_els00;
            }
            else if (UNKNOWN_TE_HEADER_LS28 == rc)
            {
                RepServMessages::Instance()->ls98Msg(layTableLoad_els98, pszErrText, errWrongTEFileType_els98,
                                                                                     pTLRec->tlrInFile_ls00->vfFileName_ls00,
                                                                                     "manipulated primary data header page");
                PageBuffer.ls51NotUsed();                   // mark page as 'not used'
                rc = errWrongTEFileType_els98;
            }
            // else if (STOP_NODATA_LS28 == rc)
            // --> no more data to process for this table; this branch must not be handled, return code has
            //     already the right value. It is highly unlikely that the return code becomes STOP_NODATA_LS28
            //     in this function because for every extracted table exists at least ONE page of primary data.
            //     But we cannot release the 'unused' header page here.
        }
        else
        {
            _TCHAR szCmd[lStmtLength_ls26];
            szCmd[0] = 0;
            sp77sprintf(szCmd, lStmtLength_ls26, RESTORE_DATA_STMT_LS28, OwnerName.StrPtr(), TableName.StrPtr());

            rc = ls28_BeginRestore(pDBInfo, szCmd, pTLRec->tlrBdInfo_ls00, pszErrText);
            if (errOK_els00 == rc)
            {
                rc = ls28_DataLoadLoop(pDBInfo, pTLRec, PageBuffer, pTLRec->tlrBdInfo_ls00, pszErrText);
                if (errInternal_els98 == rc)
                {
                    RepServMessages::Instance()->ls98Msg(layTableExtract_els98, pszErrText, errInternal_els98,
                                                                                            _INTERNAL_FILE_ID_,
                                                                                            __LINE__,
                                                                                            STATIC_CAST(ErrorNumbers, rc));
                }
            }

            if (errOK_els00 == rc)
            {
                rc = ls28_EndRestore(pDBInfo, pTLRec->tlrBdInfo_ls00, pszErrText);
            }
        }
    }   // end else branch of if (NULL == pszBuffer)

    return rc;
}
// ls28_RestorePrimaryData()


/*
  -----------------------------------------------------------------------------
  function:     ls28_RestoreLongStrings
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_RestoreLongStrings(tls00_DBInfo    *pDBInfo,
                        tls00_TloadRec  *pTLRec,
                        ReadBuffer      &PageBuffer,
                        tsp00_Addr       pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls28_RestoreLongStrings");

    tsp00_Int4 rc = errOK_els00;

    tls26_DescPageHeader *pHeader  = NULL;
    tsp00_SwapKind_Enum   CurrSwap = ls07DefineSwap(NULL);

    tls00_Parts4InternalSql PartStruct;
    PartStruct.pfiPartKind_ls00 = NULL;
    PartStruct.pfiData_ls00     = NULL;

    tls00_String *pMyData = NULL;
    
    
    _TCHAR *pszBuffer = PageBuffer.ls51GetNext();
    if (NULL == pszBuffer)
    {
        rc = PageBuffer.ls51GetError(pszErrText);
    }

    if (errOK_els00 == rc)
    {
        rc = ls28_CheckTypeHeaderPage(pszBuffer, pDBInfo->dbiPageSize_ls00, tptLongString_els00, pTLRec->tlrSwapFromTE_ls00);
        if (errOK_els00 != rc)                          // There are no long strings to restore
        {
            // error numbers can be
            //  - WRONG_TE_HEADER_LS28   --> no longs to restore
            //  - UNKNOWN_TE_HEADER_LS28 --> no longs to restore
            //  - STOP_NODATA_LS28         --> no more data
            if (WRONG_TE_HEADER_LS28 == rc)
            {
                TRACE_PRNF_MLS99(("ls28_RestoreLongStrings", "No long strings.\n"));

                PageBuffer.ls51NotUsed();                   // mark page as 'not used'
            }
            else if (UNKNOWN_TE_HEADER_LS28 == rc)
            {
                RepServMessages::Instance()->ls98Msg(layTableLoad_els98, pszErrText, errWrongTEFileType_els98,
                                                                                     pTLRec->tlrInFile_ls00->vfFileName_ls00,
                                                                                     "manipulated long string header page");
                PageBuffer.ls51NotUsed();                   // mark page as 'not used'
                rc = errWrongTEFileType_els98;
            }
            // else if (STOP_NODATA_LS28 == rc)
            // --> no more data to process for this table; this branch must not be handled, return code has
            //     already the right value.
            //     We cannot release the 'unused' header page here.
        }
        else
        {
            PartStruct.pfiCount_ls00 = 3;               // 3 parts to deliver to kernel
            PartStruct.pfiPartKind_ls00 = new tsp1_part_kind_Enum[PartStruct.pfiCount_ls00];
            PartStruct.pfiPartKind_ls00[0] = sp1pk_command;
            PartStruct.pfiPartKind_ls00[1] = sp1pk_bdinfo;
            PartStruct.pfiPartKind_ls00[2] = sp1pk_data;

            PartStruct.pfiData_ls00 = new tls00_String[PartStruct.pfiCount_ls00];

            //*
            //*     Build statement
            //*
            PartStruct.pfiData_ls00[0].strAddr_ls00 = new _TCHAR[lStmtLength_ls26];
            _tcscpy(&PartStruct.pfiData_ls00[0].strAddr_ls00[0], RESTORE_LONG_STMT_LS28);
            PartStruct.pfiData_ls00[0].strLeng_ls00 = STATIC_CAST(tsp00_Int4, _tcslen(PartStruct.pfiData_ls00[0].strAddr_ls00));

            //*
            //*     Build bdinfo part
            //*
            PartStruct.pfiData_ls00[1].strAddr_ls00 = new _TCHAR[BDINFO_SIZE_LS00];
            PartStruct.pfiData_ls00[1].strLeng_ls00 = BDINFO_SIZE_LS00;
            memcpy(PartStruct.pfiData_ls00[1].strAddr_ls00, pTLRec->tlrBdInfo_ls00, BDINFO_SIZE_LS00);

            //*
            //*     Build data part
            //*
            tsp00_Int4 lDataPartLen = MAX_KEYLEN_GG00 + INT2_MXSP00 + INT2_MXSP00;
            PartStruct.pfiData_ls00[2].strAddr_ls00 = new _TCHAR[lDataPartLen];

            // Because this is the part we have to renew for every new long value to send to kernel we
            // assign a pointer to this data
            pMyData = &PartStruct.pfiData_ls00[2];

            pHeader = REINTERPRET_CAST(tls26_DescPageHeader*, pszBuffer);
        }
    }    

    
    while (errOK_els00 == rc)
    {
        // to restore long string we need to deliver the key and the column number; they're saved at
        // the header page

        // insert key length
        pMyData->strLeng_ls00 = INT2_MXSP00 + pHeader->dphSpecialID_ls26 + INT2_MXSP00;
        ls07Swap2BIntToChar(STATIC_CAST(tsp00_Int2, pHeader->dphSpecialID_ls26),
                            pTLRec->tlrSwapFromTE_ls00,
                            CurrSwap,
                            &pMyData->strAddr_ls00[0]);
        
        // insert key
        memcpy(&pMyData->strAddr_ls00[INT2_MXSP00], &pszBuffer[sizeof(tls26_DescPageHeader)], pHeader->dphSpecialID_ls26);

        ls07Swap2BIntToChar(STATIC_CAST(tsp00_Int2, pHeader->dphFillerInt_ls26),
                            pTLRec->tlrSwapFromTE_ls00,
                            CurrSwap,
                            &pMyData->strAddr_ls00[pHeader->dphSpecialID_ls26 + INT2_MXSP00]);
            
        rc = ls28_BeginRestoreLong(pDBInfo, PartStruct, pTLRec->tlrLongBdInfo_ls00, pszErrText);
        if (errOK_els00 == rc)
        {
            rc = ls28_DataLoadLoop(pDBInfo, pTLRec, PageBuffer, pTLRec->tlrLongBdInfo_ls00, pszErrText);
            if (errInternal_els98 == rc)
            {
                RepServMessages::Instance()->ls98Msg(layTableExtract_els98, pszErrText, errInternal_els98,
                                                                                        _INTERNAL_FILE_ID_,
                                                                                        __LINE__,
                                                                                        STATIC_CAST(ErrorNumbers, rc));
            }
        }

        if (errOK_els00 == rc)
        {
            rc = ls28_EndRestore(pDBInfo, pTLRec->tlrLongBdInfo_ls00, pszErrText);
        }

        // coming here w/o errors the load of the n-th long value has finished; we're now looking for the next
        // long value (next long type header page) in data file
        if (errOK_els00 == rc)
        {
            pszBuffer = PageBuffer.ls51GetNext();
            if (NULL == pszBuffer)
            {
                rc = PageBuffer.ls51GetError(pszErrText);
            }
            if (errOK_els00 == rc)
            {
                rc = ls28_CheckTypeHeaderPage(pszBuffer, pDBInfo->dbiPageSize_ls00, tptLongString_els00, pTLRec->tlrSwapFromTE_ls00);
                if (errOK_els00 != rc)                          // There are no more long strings to restore
                {
                    // error numbers can only be WRONG_TE_HEADER_LS28 and UNKNOWN_TE_HEADER_LS28;
                    // both indicate that there are no short strings to restore
                    if (WRONG_TE_HEADER_LS28 == rc)
                    {
                        TRACE_PRNF_MLS99(("ls28_RestoreLongStrings", "No long strings.\n"));

                        PageBuffer.ls51NotUsed();                   // mark page as 'not used'
                    }
                    else if (UNKNOWN_TE_HEADER_LS28 == rc)
                    {
                        RepServMessages::Instance()->ls98Msg(layTableLoad_els98, pszErrText, errWrongTEFileType_els98,
                                                                                             pTLRec->tlrInFile_ls00->vfFileName_ls00,
                                                                                             "manipulated long string header page");
                        PageBuffer.ls51NotUsed();                   // mark page as 'not used'
                        rc = errWrongTEFileType_els98;
                    }
                    // else if (STOP_NODATA_LS28 == rc)
                    // --> no more data to process for this table; this branch must not be handled, return code has
                    //     already the right value.
                    //     We cannot release the 'unused' header page here.
                }
                else
                {
                    pHeader = REINTERPRET_CAST(tls26_DescPageHeader*, pszBuffer);
                }
            }
        }
    }   // end while


    // some cleaning work
    if (NULL != PartStruct.pfiData_ls00)
    {
        delete [] PartStruct.pfiData_ls00[0].strAddr_ls00;
        PartStruct.pfiData_ls00[0].strAddr_ls00 = NULL;
        
        delete [] PartStruct.pfiData_ls00[1].strAddr_ls00;
        PartStruct.pfiData_ls00[1].strAddr_ls00 = NULL;
        
        delete [] PartStruct.pfiData_ls00[2].strAddr_ls00;
        PartStruct.pfiData_ls00[2].strAddr_ls00 = NULL;
        
        delete [] PartStruct.pfiData_ls00;
    }

    if (NULL != PartStruct.pfiPartKind_ls00)
    {
        delete [] PartStruct.pfiPartKind_ls00;
        PartStruct.pfiPartKind_ls00 = NULL;
    }

    if ( (WRONG_TE_HEADER_LS28 == rc) || (EOF_LS00 == rc) || (STOP_NODATA_LS28 == rc) )
    {
        // This is the last data processing function for one table. So we can set
        // the return code to ok in these cases.
        rc = errOK_els00;
    }

    return rc;
}
// ls28_RestoreLongStrings()


/*
  -----------------------------------------------------------------------------
  function:     ls28_ReadDescFromFile
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_ReadDescFromFile(ReadBuffer       &myReadBuffer,
                      tls00_TEPageType  DescriptionType,
                      tsp00_SwapKind    SourceSwap,
                      tsp00_Int4        lPageSize,
                      tls00_String     &pDescription,
                      tsp00_Addr        pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls28_ReadDescFromFile");

    tsp00_Int4 rc = errOK_els00;

    _TCHAR               *pszBuffer = NULL;
    tls26_DescPageHeader *pHeader   = NULL;
    tsp00_Int4            lOffset   = 0;

    while (errOK_els00 == rc)
    {
        pszBuffer = myReadBuffer.ls51GetNext();
        if (NULL == pszBuffer)                      // error or eof
        {
            rc = myReadBuffer.ls51GetError(pszErrText);
            myReadBuffer.ls51NotUsed();             // mark page as 'not used'
        }
        else
        {
            rc = ls28_CheckTypeHeaderPage(pszBuffer, lPageSize, DescriptionType, SourceSwap);
            if (errOK_els00 != rc)                  // There is no description to read. We should never reach this point
            {                                       // or the file is corrupted. At this moment we cannot recover from this.
                myReadBuffer.ls51NotUsed();         // mark page as 'not used'
            }
            else
            {
                pHeader = REINTERPRET_CAST(tls26_DescPageHeader*, pszBuffer);

                if (0 == lOffset)                   // 1st time we come here: allocate memory for description
                {
                    pDescription.strLeng_ls00 = pHeader->dphFillerInt_ls26;
                    pDescription.strAddr_ls00 = new _TCHAR[pDescription.strLeng_ls00];
                }    

                //*     copy n-th part of table|index description
                //* dphSpecialID_ls26 contains the length of written description info on THIS buffer page
                memcpy(&pDescription.strAddr_ls00[lOffset],
                       &pszBuffer[sizeof(tls26_DescPageHeader)],
                       pHeader->dphSpecialID_ls26);
                lOffset += pHeader->dphSpecialID_ls26;
                if (false == pHeader->dphMore_ls26)
                {
                    break;
                }
            }
        }
    }       // while (errOK_els00 == rc)

    return rc;
}
// ls28_ReadDescFromFile()


/*
  -----------------------------------------------------------------------------
  function:     ls28_CheckTypeHeaderPage
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_CheckTypeHeaderPage(tsp00_Addr       pszBuffer,
                         tsp00_Int4       lPageSize,
                         tls00_TEPageType PageType,
                         tsp00_SwapKind   SourceSwapKind)
{
    ROUTINE_DBG_MEO00 ("ls28_CheckTypeHeaderPage");

    tsp00_Int4 rc = errOK_els00;

    tls26_DescPageHeader *pHeader = REINTERPRET_CAST(tls26_DescPageHeader*, pszBuffer);
    tls26_DescPageHeader  Footer;

    if ( (tptData_els00 == PageType)        ||
         (tptShortString_els00 == PageType) ||
         (tptLongString_els00 == PageType)  )
    {
        memcpy(&Footer, &pszBuffer[lPageSize - sizeof(tls26_DescPageHeader)], sizeof(tls26_DescPageHeader));
        if (0 != memcmp(pHeader, &Footer, sizeof(tls26_DescPageHeader)))
        {
            rc = UNKNOWN_TE_HEADER_LS28;
        }
    }
    else
    {
        memcpy(&Footer, pHeader, sizeof(tls26_DescPageHeader));
    }

    if (errOK_els00 == rc)
    {
        ls28SwapHeader(SourceSwapKind, pHeader);

        if ( (MY_PAGE_LS00 != pHeader->dphType_ls26) || (PageType != pHeader->dphPageType_ls26) )
        {
            if (tptEOT_els00 == pHeader->dphPageType_ls26)      // no more table data in file (end of table)
            {
                rc = STOP_NODATA_LS28;
            }
            else
            {
                rc = WRONG_TE_HEADER_LS28;                    // actually not an error: only a message

                // there is a small thing to do - undo header swap to enable others to use the page freshly
                // ATTENTION: REUSING FOOTER HERE FOR DATA PAGES
                // we can use the footer or the temp saved header here to restore the header;
                // header and footer MUST originally have been the same only for data pages
                memcpy(pszBuffer, &Footer, sizeof(tls26_DescPageHeader));
            }
        }
    }

    return rc;
}
// ls28_CheckTypeHeaderPage()


/*
  -----------------------------------------------------------------------------
  function:     ls28_Default_ConstraintFromFile
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_Default_ConstraintFromFile(tls00_DBInfo       *pDBInfo,
                                tls00_TloadRec     *pTLRec,
                                ReadBuffer         &PageBuffer,
                                tls28_PageReadInfo *ReadInfo,        // wird hier erstmals gefuellt und dann auch hier immer ausgewertet
                                tls00_TEPageType    PageType,
                                tls00_String       *pValue,
                                tsp00_Addr          pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls28_Default_ConstraintFromFile");

    tsp00_Int4 rc = errOK_els00;
    
    tls26_DescPageHeader *pHeader  = NULL;

    if (0 == ReadInfo->priCount_ls28) // initial or (get next page) state
    {
        ReadInfo->priBuffer_ls28 = PageBuffer.ls51GetNext();
        if (NULL == ReadInfo->priBuffer_ls28)               // error reading from data file
        {
            rc = PageBuffer.ls51GetError(pszErrText);
        }

        if (errOK_els00 == rc)
        {
            rc = ls28_CheckTypeHeaderPage(ReadInfo->priBuffer_ls28,
                                          pDBInfo->dbiPageSize_ls00,
                                          PageType,
                                          pTLRec->tlrSwapFromTE_ls00);
            if (errOK_els00 != rc)                         // No defaults/constraints to restore
            {
                if (WRONG_TE_HEADER_LS28 == rc)
                {
                    rc = STOP_NODATA_LS28;
                }
                PageBuffer.ls51NotUsed();                       // mark page as 'not used'
            }
            else
            {
                pHeader = REINTERPRET_CAST(tls26_DescPageHeader*, ReadInfo->priBuffer_ls28);

                ReadInfo->priCount_ls28 = pHeader->dphSpecialID_ls26;
                ReadInfo->priMore_ls28  = pHeader->dphMore_ls26;
        
                ReadInfo->priBuffer_ls28 += sizeof(*pHeader);       // Initialize buffer
                TRACE_PRNF_MLS99(("ls28_Default_ConstraintFromFile", "%d values on this page\n", ReadInfo->priCount_ls28));
            }
        }
    }

    // Read value
    if (errOK_els00 == rc)
    {
        tsp00_Int2 lTmpLength =0;

        if (tptDefault_els00 == PageType)                       // default value
        {
            memcpy(&lTmpLength, &ReadInfo->priBuffer_ls28[sizeof(tls26_Co_De_LenType)], sizeof(tsp00_Int2));

            // Default value is always saved with a size of the default length of 2 byte!
            pValue->strLeng_ls00 = sizeof(tls26_Co_De_LenType) + sizeof(tsp00_Int2) + lTmpLength;
        }
        else if (tptConstraint_els00 == PageType)               // constraint value
        {
            memcpy(&lTmpLength, ReadInfo->priBuffer_ls28, sizeof(tsp00_Int2));

            pValue->strLeng_ls00 = lTmpLength;

            // Here the buf ptr needs to be moved by the size of the length info because this info is NOT
            // included in the length
            ReadInfo->priBuffer_ls28 += sizeof(tls26_Co_De_LenType);
        }
        pValue->strAddr_ls00 = ReadInfo->priBuffer_ls28;


        --ReadInfo->priCount_ls28;      // increase count of read values

        if (0 != ReadInfo->priCount_ls28)
        {
            ReadInfo->priBuffer_ls28 += pValue->strLeng_ls00;   // buffer moved to the beginning of the next value
        }
        else
        {
            if (0 == ReadInfo->priMore_ls28)
            {
                rc = STOP_LS28;
            }
        }
    }

    return rc;
}
// ls28_Default_ConstraintFromFile()


/*
  -----------------------------------------------------------------------------
  function:     ls28_CreateConstraint
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_CreateConstraint(tls00_DBInfo            *pDBInfo,
                      SAPDB_DynamicUTF8String &TableName,
                      tls00_String            *pValue,
                      tsp00_Addr               pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls28_CreateConstraint");

    tsp00_Int4 rc = errOK_els00;

    char szCmd[lStmtLength_ls26];
    szCmd[0] = 0;
    sp77sprintf(szCmd, lStmtLength_ls26, INTO_CONSTRAINT_STMT_LS28, TableName.StrPtr());

    tsp00_Int4 lErrorPos = 0;
    char       szErr[MAX_REPLY_LENGTH_LS00];

    rc = ls13ExecIntSQLStmtExUTF8(pDBInfo, 0, szCmd, pValue->strAddr_ls00, pValue->strLeng_ls00, szErr, lErrorPos);
    if (errOK_els00 != rc)
    {
        RepServMessages::Instance()->ls98Msg(layTableLoad_els98, pszErrText, errInternSQL_els98,
                                                                             szCmd, szErr);
        rc = errInternSQL_els98;
    }

    return rc;
}
// ls28_CreateConstraint()


/*
  -----------------------------------------------------------------------------
  function:     ls28_CreateDefault
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_CreateDefault(tls00_DBInfo            *pDBInfo,
                   SAPDB_DynamicUTF8String &TableName,
                   tls00_String            *pValue,
                   tsp00_Addr               pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls28_CreateDefault");

    tsp00_Int4 rc = errOK_els00;
    
    // The read default value comes from data file in the following form
    //
    //  ----------------------------------------------------------------------------
    //  | column number | value length | default value with leading defined byte   |
    //  ----------------------------------------------------------------------------
    //       2 byte         2 byte           (value length) byte
    //
    // The db kernel expects the delivered default value in this form for version 7.3
    // For version 7.2 the value length MUST be delivered with a size of 1 byte.
    

    // Build the command 'INTO DEFAULT <table name>'
    _TCHAR szCmd[lStmtLength_ls26];
    szCmd[0] = 0;
    sp77sprintf(szCmd, lStmtLength_ls26, INTO_DEFAULT_STMT_LS28, TableName.StrPtr());

    // Bring the default value into the right form
    tsp00_Addr pszDefault = pValue->strAddr_ls00;

    if ( ('7' == pDBInfo->dbiDBVersion_ls00.MajorVersion_sp100) &&
         ('2' == pDBInfo->dbiDBVersion_ls00.MinorVersion_sp100) )
    {
        size_t lDefValOffset = 2*sizeof(tsp00_Int2);

        pszDefault = new _TCHAR[pValue->strLeng_ls00 - sizeof(tsp00_Uint1)];

        memcpy(pszDefault, pValue->strAddr_ls00, sizeof(tsp00_Int2));                       // copy column number
        
        tsp00_Int2 l72DefaultLen = *(REINTERPRET_CAST(tsp00_Int2*, &pValue->strAddr_ls00[sizeof(tsp00_Int2)]));
        pszDefault[sizeof(tsp00_Int2)] = STATIC_CAST(tsp00_Uint1, l72DefaultLen);       // copy value length
        
        memcpy(&pszDefault[sizeof(tsp00_Int2) + sizeof(tsp00_Uint1)],                       // copy value
               &pValue->strAddr_ls00[lDefValOffset],
               (pValue->strLeng_ls00 - lDefValOffset) );
        
        pValue->strLeng_ls00 -= sizeof(tsp00_Uint1);
    }

    // Execute the command
    tsp00_Int4 lErrorPos = 0;
    _TCHAR     szErr[MAX_REPLY_LENGTH_LS00];

    rc = ls13ExecIntSQLStmtExUTF8(pDBInfo, 0, szCmd, pszDefault, pValue->strLeng_ls00, szErr, lErrorPos);
    if (errOK_els00 != rc)
    {
        RepServMessages::Instance()->ls98Msg(layTableLoad_els98, pszErrText, errInternSQL_els98,
                                                                             szCmd, szErr);
        rc = errInternSQL_els98;
    }

    if ( (NULL != pszDefault) && (pValue->strAddr_ls00 != pszDefault) )
    {
        delete [] pszDefault;
        pszDefault = NULL;
    }

    return rc;
}
// ls28_CreateDefault()


/*
  -----------------------------------------------------------------------------
  function:     ls28_BeginRestore
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_BeginRestore(tls00_DBInfo *pDBInfo,
                  tsp00_Addr    pszCmd,
                  tsp7_bdinfo  &pszBdinfo,
                  tsp00_Addr    pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls28_BeginRestore");
    tsp00_Int4 rc  = errOK_els00;

    _TCHAR     szErr[MAX_REPLY_LENGTH_LS00];
    tsp00_Int4 lErrPos = 0;

    rc = ls13ExecIntSQLStmtUTF8(pDBInfo, 0, pszCmd, szErr, lErrPos);
    if (0 != rc)
    {
        RepServMessages::Instance()->ls98Msg(layTableLoad_els98, pszErrText, errInternSQL_els98,
                                                                             pszCmd, szErr);
        rc = errInternSQL_els98;
    }
    else
    {
        rc = ls04GetBdinfo(pDBInfo->dbiPktRcvList_ls00[0], pszBdinfo);
        if (errOK_els00 != rc)
        {
            RepServMessages::Instance()->ls98Msg(layTableLoad_els98, pszErrText, errInternal_els98,
                                                                                 _INTERNAL_FILE_ID_,
                                                                                 __LINE__,
                                                                                 rc);
            rc = errInternal_els98;
        }
    }

    return rc;
}
// ls28_BeginRestore()


/*
  -----------------------------------------------------------------------------
  function:     ls28_DataLoadLoop
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_DataLoadLoop(tls00_DBInfo   *pDBInfo,
                  tls00_TloadRec *pTLRec,
                  ReadBuffer     &PageBuffer,
                  tsp7_bdinfo    &pszBdInfo,
                  tsp00_Addr      pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls28_DataLoadLoop");

    tsp00_Int4 rc     = errOK_els00;
    tsp00_Int4 rcFill = errOK_els00;

    tsp00_Int4 lSndPcktNo  = 0;
    tsp00_Int4 lRcvPcktNo  = 0;

    
    while ( (errOK_els00 == rcFill) && (errOK_els00 == rc) )
    {
        //  Fill packet
        rcFill = ls28_FillPacketBuffer(pDBInfo, pTLRec, lSndPcktNo, PageBuffer, pszErrText);
        
        if (lSndPcktNo != lRcvPcktNo)   // Gegebenenfalls anderes SQL-Packet empfangen
        {
            //*
            //*     Receive packet
            //* We adapt the number of received packet even if receiving fails.
            //* If it fails we do not need to receive it again.
            //*
            rc = ls13ReceiveBdinfoPckt(pDBInfo, lRcvPcktNo, pszBdInfo, pszErrText);
            if (errOK_els00 != rc)
            {
                RepServMessages::Instance()->ls98Msg(layTableExtract_els98, pszErrText, errInternal_els98,
                                                                                        _INTERNAL_FILE_ID_,
                                                                                        __LINE__,
                                                                                        STATIC_CAST(ErrorNumbers, rc));
                rc = errInternal_els98;
            }

            lRcvPcktNo = (lRcvPcktNo + 1) % 2;
        }
        
        //*     Send packet
        //* Only if filling the packet and receiving the answer ran without errors
        //*
        if ( ( (errOK_els00 == rcFill) || (STOP_LS28 == rcFill) ) && (errOK_els00 == rc) )
        {
            rc = ls13SendBdinfoPckt(pDBInfo, lSndPcktNo, pszBdInfo, pszErrText);
            if (errOK_els00 == rc)
            {
                lSndPcktNo = (lSndPcktNo + 1) % 2;  //  successful sent: recalculate packet number
            }
        }
    }
    

    if (lSndPcktNo != lRcvPcktNo)
    {
        //*     Receive last packet
        rc = ls13ReceiveBdinfoPckt(pDBInfo, lRcvPcktNo, pszBdInfo, pszErrText);
        if (errOK_els00 != rc)
        {
            RepServMessages::Instance()->ls98Msg(layTableExtract_els98, pszErrText, errInternal_els98,
                                                                                    _INTERNAL_FILE_ID_,
                                                                                    __LINE__,
                                                                                    STATIC_CAST(ErrorNumbers, rc));
            rc = errInternal_els98;
        }


//TODO        //  Bei der fehlerbehandlung haben wir ein problem. Wenn jetzt etwas schiefgeht beim empfangen
        //  des paketes aber vorher auch schon ein fehler bei der bearbeitung der pakete aufgetreten
        //  ist, welcher fehler wird dann gemeldet? Die fehler muessen nicht notwendigerweise voneinander
        //  abhaengen.
        //  Momentan wird der fehler beim empfang des letzten paketes nur dann nach oben gemeldet,
        //  wenn vorher kein fehler aufgetreten ist.
    }

    // some comparing
    if (errOK_els00 == rc)
    {
        if ( (errOK_els00 != rcFill) && (STOP_LS28 != rcFill) && (STOP_NODATA_LS28 != rcFill))
        {
            rc = rcFill;
        }
    }

    return rc;
}
// ls28_DataLoadLoop()


/*
  -----------------------------------------------------------------------------
  function:     ls28_FillPacketBuffer
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_FillPacketBuffer(tls00_DBInfo   *pDBInfo,
                      tls00_TloadRec *pTLRec,
                      tsp00_Int4      lPcktNo,
                      ReadBuffer     &PageBuffer,
                      tsp00_Addr      pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls28_FillPacketBuffer");
    
    tsp00_Int4 rc  = errOK_els00;
    tsp00_Int4 rc2 = errOK_els00;

    tsp1_packet *SqlPacket = NULL;
    tsp1_part   *SqlPart   = NULL;
    
    // prepare the packet to take the data -> puts a part of part kind page into packet
    ls13Prepare_FL_TL_Packet(pDBInfo, lPcktNo, SqlPacket, SqlPart);

    tsp00_Int2 i          = 0;
    _TCHAR     *pszBuffer = NULL;

    //*
    //*     Now fill the packet with data pages
    //*
    //* When computing tlrPgPerPckt_ls00 it is already taken into account, that a sent packet
    //* must contain a bdinfo AND the data.
    //*
    //* This function is ONLY called if we know exactly that there is some data for
    //* the requested type in the data file. Additionally the last page of every single
    //* table data file is a special header page. So we will never reach the end of
    //* file here if the tableextract went without any problems and the file is not
    //* corrupted afterwards (hopefully).
    //*
    for (i; ((i < pTLRec->tlrPgPerPckt_ls00) && (errOK_els00 == rc)); ++i)
    {
        pszBuffer = PageBuffer.ls51GetNext();
        if (NULL == pszBuffer)
        {
            rc = PageBuffer.ls51GetError(pszErrText);   // can never be eof, must be an error
        }
        
        if (errOK_els00 == rc)
        {
            //*
            //*     Check if page is the right one or if it is already the next data type
            //* In case this is the first page read here there is nothing to process.
            //*
            if (ptNil_egg00 != (REINTERPRET_CAST(tbd_node*, pszBuffer))->nd_pt())
            {
                PageBuffer.ls51NotUsed();
                if (0 == i)
                {
                    rc = STOP_NODATA_LS28;  // no page read from file here; nothing more to process
                }
                else
                {
                    rc = STOP_LS28;         // At least one page read that must be send to the kernel
                }
            }
            else
            {
                // swap length information on page
                ls28SwapPageValues(pszBuffer, pDBInfo->dbiPageSize_ls00, pTLRec);

                ls04BufToPart(SqlPart, pszBuffer, pTLRec->tlrPageSize_ls00);
            }
        }
    }   // end for (i; ((i < pTLRec->tlrPgPerPckt_ls00) && (errOK_els00 == rc)); ++i)

    // finish part
    if ( (errOK_els00 == rc) || (STOP_LS28 == rc) )
    {
        rc2 = ls04FinishPart(SqlPacket, SqlPart);
        if (errOK_els00 != rc2)
        {
            rc = rc2;
        }
    }
    
    return rc;
}
// ls28_FillPacketBuffer()


/*
  -----------------------------------------------------------------------------
  function:     ls28_EndRestore
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_EndRestore(tls00_DBInfo *pDBInfo, tsp7_bdinfo pszBdinfo, tsp00_Addr pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls28_EndRestore");
    tsp00_Int4 rc  = errOK_els00;

    _TCHAR szCmd[lStmtLength_ls26];
    szCmd[0] = 0;
    _tcscpy(szCmd, END_RESTORE_STMT_LS28);

    tsp00_Int4 lErrorPos = 0;
    _TCHAR     szErr[MAX_REPLY_LENGTH_LS00];

    rc = ls13ExecIntSQLStmtExUTF8(pDBInfo, 0, szCmd, pszBdinfo, BDINFO_SIZE_LS00, szErr, lErrorPos, sp1pk_bdinfo);
    if (errOK_els00 != rc)
    {
        RepServMessages::Instance()->ls98Msg(layTableLoad_els98, pszErrText, errInternSQL_els98,
                                                                             szCmd, szErr);
        rc = errInternSQL_els98;
    }

    return rc;
}
// ls28_EndRestore()


/*
  -----------------------------------------------------------------------------
  function:     ls28_BeginRestoreLong
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_BeginRestoreLong(tls00_DBInfo            *pDBInfo,
                      tls00_Parts4InternalSql &PartStruct,
                      tsp7_bdinfo             &pszColumnBdinfo,
                      tsp00_Addr               pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls28_BeginRestoreLong");
    
    tsp00_Int4 rc = errOK_els00;
    
    tsp00_Int4 lErrorPos = 0;
    _TCHAR     szErr[MAX_REPLY_LENGTH_LS00];
    szErr[0] = 0;
    
    rc = ls13ExecIntSQLMultPartsUTF8(pDBInfo, 0, PartStruct, szErr, lErrorPos);
    if (errOK_els00 != rc)
    {
        RepServMessages::Instance()->ls98Msg(layTableLoad_els98, pszErrText, errInternSQL_els98,
                                                                             PartStruct.pfiData_ls00[0].strAddr_ls00,
                                                                             szErr);
        rc = errInternSQL_els98;
    }
    else
    {
        rc = ls04GetBdinfo(pDBInfo->dbiPktRcvList_ls00[0], pszColumnBdinfo);
        if (errOK_els00 != rc)
        {
            RepServMessages::Instance()->ls98Msg(layTableLoad_els98, pszErrText, errInternal_els98,
                                                                                 _INTERNAL_FILE_ID_,
                                                                                 __LINE__,
                                                                                 rc);
            rc = errInternal_els98;
        }
    }

    return rc;
}
// ls28_BeginRestoreLong()

/*  
  -----------------------------------------------------------------------------
  EndChapter: Private functions implementation
  -----------------------------------------------------------------------------
*/
