/*! 
  -----------------------------------------------------------------------------
 
  module: vls28_TLoad.cpp
 
  -----------------------------------------------------------------------------
 
  responsible:  SteffenS and d033893

  special area: Replication Server
 
  description:  Implementation of TABLELOAD
                
  version:      7.3.

  -----------------------------------------------------------------------------
 
  copyright:    SAP AG, 2000-2003
 
  -----------------------------------------------------------------------------


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*/
#include "gls00.h"
#include "hls05.h"
#include "hls13.h"
#include "hls15.h"
#include "hls28_TLoad.h"
#include "hls26.h"
#include "hls26_TETable.h"
#include "hls28_TLTable.h"
#include "hls30.h"
#include "hls98msg.h"
#include "hls99.h"

/*  
  -----------------------------------------------------------------------------
  Chapter: Private functions declaration
  -----------------------------------------------------------------------------
*/

/*!
  -----------------------------------------------------------------------------
  function:     ls28_Constructor
  -----------------------------------------------------------------------------

  description:  Creates and initializes the structure used for
                execution of table load.
                Initializes and opens mapping file, data file and restart file.

  arguments:    DBInfo     [IN]     - structure that contains all info on the
                                       running session and the database connected to.
                pTLCmd     [IN]     - tableload command structure
                TLRec      [IN|OUT] - tableload structure that must be initialized.
                pszErrText [OUT]    - error text
                
  returnvalue:  errOK_els00, errNoMoreMemory_els00
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_Constructor(tls00_DBInfo      *DBInfo,
                 tls00_TabLoadCmd  *pTLCmd,
                 tls00_TloadRec   *&pTLRec,
                 tsp00_Addr         pszErrText);


/*!
  -----------------------------------------------------------------------------
  function:     ls28_Destructor
  -----------------------------------------------------------------------------

  description:  Deletes all allocated space and the tableload structure.
                Closes all open files.

  arguments:    TLRec      [IN|OUT] - tableload structure that must be initialized.
                pTLCmd     [IN]     - tableload command structure
                
  returnvalue:  No
  -----------------------------------------------------------------------------
*/
static void
ls28_Destructor(tls00_TloadRec *pTLRec, tls00_TabLoadCmd *pTLCmd);


/*!
  -----------------------------------------------------------------------------
  function:     ls28_GetTargetKernelSwap
  -----------------------------------------------------------------------------

  description:  Gets the swap of kernel from target db.

  arguments:    pDBInfo    [IN]  - structure that contains all info on the
                                    running session and the database connected to.
                pTLRec     [IN]  - structure, holding information for executing
                                    TABLELOAD for a single table
                pszErrText [OUT] - error text

  returnvalue:  errOK_els00              - no error

                The function logs the error message to the log file and to
                the error text return buffer.
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_GetTargetKernelSwap(tls00_DBInfo   *pDBInfo,
                         tls00_TloadRec *pTLRec,
                         tsp00_Addr      pszErrText);


/*!
  -----------------------------------------------------------------------------
  function:     ls28_GetTableNamesFromDataFile
  -----------------------------------------------------------------------------

  description:  Reads table names of tables to load from mapping file. Opens
                data files that are also logged in mapping file.
                Analyzes file header of newly opened data file.
                Used only for TL ALL/TL USER. Function works recursively.

  arguments:    pTLRec      [IN]  - structure, holding information for executing
                                     TABLELOAD for a single table
                pTLCmd      [IN]  - tableload command structure
                UserName    [OUT] - user name read from mapping file UTF8 encoded
                TableName   [OUT] - table name read from mapping file UTF8 encoded
                pszErrText  [OUT] - error text

  returnvalue:  errOK_els00                     - no error
                errTLSingleTabFileMultTab_els98 - data file for single table load
                STOP_NODATA_LS28                - no error; end of data reached

                The function logs the error message to the log file and to
                the error text return buffer.
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_GetTableNamesFromDataFile(tls00_DBInfo            *pDBInfo,
                               tls00_TloadRec          *pTLRec,
                               tls00_TabLoadCmd        *pTLCmd,
                               SAPDB_DynamicUTF8String &UserName,
                               SAPDB_DynamicUTF8String &TableName,
                               tsp00_Addr               pszErrText);


/*!
  -----------------------------------------------------------------------------
  function:     ls28_AnalyseDataFileHeader
  -----------------------------------------------------------------------------

  description:  Reads header of newly opened data file (infile) and checks if the
                data file was created by TABLEEXTRACT and checks some flags.
                Fills some values in pTLRec.

  arguments:    pDBInfo    [IN]      - structure that contains all info on the
                                        running session and the database connected to.
                pTLRec     [IN/OUT]  - structure, holding information for executing
                                       TABLELOAD for a single table
                pPartSpec  [IN]      - structure holding info on the kind of TL
                                       (USER/ALL)
                pszErrText [OUT]     - error text

  returnvalue:  errOK_els00                     - no error
                errTLMultTabFileSingleTab_els98 - data file with multiple table data
                                                  but TL command for single table
                errTLSingleTabFileMultTab_els98 - data file with data of one single
                                                  table but TL command for multiple
                                                  tables

                The function logs the error message to the log file and to
                the error text return buffer.
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_AnalyseDataFileHeader(tls00_DBInfo   *pDBInfo,
                           tls00_TloadRec *pTLRec,
                           tls00_PartSpec *pPartSpec,
                           tsp00_Addr      pszErrText);


/*!
  -----------------------------------------------------------------------------
  function:     ls28_HandleRestart
  -----------------------------------------------------------------------------

  description:  Restart handler. In case of restart the function reads the restart
                file and finds the name of table in mapping file restart begins
                with.
                
  arguments:    pRestartStruct  [IN/OUT] - structure with info for restart
                pMapFile        [IN]     - mapping file
                pszErrText      [OUT]    - error text

  returnvalue:  errOK_els00                     - no error
                errTLRestartFileOpen_els98      - failure open restart file
                errTLRestartEntryNotFound_els98 - restart entry not found in
                                                  mapping file

                The function logs the error message to the log file and to
                the error text return buffer.
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_HandleRestart(tls00_Restart *pRestartStruct,
                   tls00_VFile   *pMapFile,
                   tsp00_Addr     pszErrText);


/*!
  -----------------------------------------------------------------------------
  function:     ls28_ReopenDataFile
  -----------------------------------------------------------------------------

  description:  Closes already open data file and initializes and opens new one.
                Builds complete path of file to open from info in Tableload
                structure pTLRec and in pszSingleFileName.
                pszSingleFileName is filled by ls28_HandleRestart.
                
  arguments:    pDBInfo           [IN]  - structure that contains all info on the
                                           running session and the database connected to.
                pTLRec            [IN]  - structure, holding information for executing
                                           TABLELOAD for a single table
                pszSingleFileName [IN]  - file name of file to open
                pPartSpec         [IN]  - structure holding info on the kind of TL
                                           (USER/ALL)
                pszErrText        [OUT] - error text

  returnvalue:  errOK_els00                     - no error
                error opening the file
                errors from ls28_AnalyseDataFileHeader

                The function logs the error message to the log file and to
                the error text return buffer.
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_ReopenDataFile(tls00_DBInfo   *pDBInfo,
                    tls00_TloadRec *pTLRec,
                    tsp00_Addr      pszSingleFileName,
                    tls00_PartSpec *pPartSpec,
                    tsp00_Addr      pszErrText);


/*!
  -----------------------------------------------------------------------------
  function:     ls28_WriteRestartEntry
  -----------------------------------------------------------------------------

  description:  Writes restart entry to restart file.
                Used only for TL ALL/TL USER.

  arguments:    pRestartStruct  [IN]  - contains restart entry to write
                pszErrText [OUT] - error text

  returnvalue:  errOK_els00                     - no error
                errors from writing to file

                The function logs the error message to the log file and to
                the error text return buffer.
  -----------------------------------------------------------------------------
*/
inline tsp00_Int4
ls28_WriteRestartEntry(tls00_Restart *pRestartStruct,
                       tsp00_Addr        pszErrText);


/*!
  -----------------------------------------------------------------------------
  function:     ls28_LoadUserTables
  -----------------------------------------------------------------------------

  description:  Executes TABLELOAD USER command.

  arguments:    DBInfo     [IN]  - structure that contains all info on the
                                    running session and the database connected to.
                pTLCmd     [IN]  - tableload command structure
                pTLRec     [IN]  - structure, holding information for executing
                                    TABLELOAD for a single table
                pszErrText [OUT] - error text

  returnvalue:  errOK_els00                     - no error
                errors from loading single table

                The function logs the error message to the log file and to
                the error text return buffer.
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_LoadUserTables(tls00_DBInfo     *pDBInfo,
                    tls00_TabLoadCmd *pTLCmd,
                    tls00_TloadRec   *pTLRec,
                    tsp00_Addr        pszErrText);


/*!
  -----------------------------------------------------------------------------
  function:     ls28_LoadAllTables
  -----------------------------------------------------------------------------

  description:  Executes TABLELOAD ALL command.

  arguments:    DBInfo     [IN]  - structure that contains all info on the
                                    running session and the database connected to.
                pTLCmd     [IN]  - tableload command structure
                pTLRec     [IN]  - structure, holding information for executing
                                    TABLELOAD for a single table
                pszErrText [OUT] - error text

  returnvalue:  errOK_els00                     - no error
                errors from loading single table

                The function logs the error message to the log file and to
                the error text return buffer.
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_LoadAllTables(tls00_DBInfo     *pDBInfo,
                   tls00_TabLoadCmd *pTLCmd,
                   tls00_TloadRec   *pTLRec,
                   tsp00_Addr        pszErrText);


/*!
  -----------------------------------------------------------------------------
  function:     ls28_LoadOneTable
  -----------------------------------------------------------------------------

  description:  Executes TABLELOAD TABLE <table name> command.

  arguments:    pDBInfo    [IN]  - structure that contains all info on the
                                    running session and the database connected to.
                pTLCmd     [IN]  - tableload command structure
                pTLRec     [IN]  - structure, holding information for executing
                                    TABLELOAD for a single table
                pszErrText [OUT] - error text

  returnvalue:  errOK_els00                     - no error
                errors from loading single table

                The function logs the error message to the log file and to
                the error text return buffer.
  -----------------------------------------------------------------------------
*/
static tsp00_Longint
ls28_LoadOneTable(tls00_DBInfo     *pDBInfo,
                  tls00_TabLoadCmd *pTLCmd,
                  tls00_TloadRec   *pTLRec,
                  tsp00_Addr        pszErrText);


/*
  -----------------------------------------------------------------------------
  EndChapter: Private functions declaration
  -----------------------------------------------------------------------------
*/


/*
  -----------------------------------------------------------------------------
  Chapter: Public functions implementation
  -----------------------------------------------------------------------------
*/

/*
  -----------------------------------------------------------------------------
  function:     ls28ExecTableLoad
  -----------------------------------------------------------------------------
*/
tsp00_Int4
ls28ExecTableLoad(tls00_TabLoadCmd *pTLCmd, tls00_DBInfo *pDBInfo, tsp00_Addr pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls28ExecTableLoad");

    tsp00_Longint  rc  = errOK_els00;

    //*
    //*     Initialize tableload record structure
    //*
    tls00_TloadRec *pTLRec = NULL;
    if (errOK_els00 == rc)
    {
        rc = ls28_Constructor(pDBInfo, pTLCmd, pTLRec, pszErrText);
    }

    // Define target kernel swap
    if (errOK_els00 == rc)
    {
        rc = ls28_GetTargetKernelSwap(pDBInfo, pTLRec, pszErrText);
    }

    //*
    //*     load table(s)
    //*
    if (errOK_els00 == rc)
    {
        if (true == pTLCmd->tloPartSpec_ls00.psSelAll_ls00)
        {
            rc = ls28_LoadAllTables(pDBInfo, pTLCmd, pTLRec, pszErrText);
        }
        else if (true == pTLCmd->tloPartSpec_ls00.psSelUser_ls00)
        {
            rc = ls28_LoadUserTables(pDBInfo, pTLCmd, pTLRec, pszErrText);
        }
        else
        {
            rc = ls28_LoadOneTable(pDBInfo, pTLCmd, pTLRec, pszErrText);
        }
    }

    if (NULL != pTLRec)
    {
        ls28_Destructor(pTLRec, pTLCmd);
    }

    return STATIC_CAST(tsp00_Int4, rc);
}
// ls28ExecTableLoad()

/*

        // Warning for long columns at different swap of source and target
        if ( tsp00_SwapKind_Enum(pTLRec->tlrSwapFromDb_ls00) !=
             tsp00_SwapKind_Enum(pTLRec->tlrTargetSwap_ls00) )
        {
            pMsgObj->ls98PrintMessage(layTableLoad_els98, wrnTL_LongsButDifferentSwap_ls98);
        }


  */
/*
  -----------------------------------------------------------------------------
  EndChapter: Public functions implementation
  -----------------------------------------------------------------------------
*/



/*
  -----------------------------------------------------------------------------
  Chapter: Private functions implementation
  -----------------------------------------------------------------------------
*/

/*
  -----------------------------------------------------------------------------
  function:     ls28_Constructor
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_Constructor(tls00_DBInfo      *pDBInfo,
                   tls00_TabLoadCmd  *pTLCmd,
                   tls00_TloadRec   *&pTLRec,
                   tsp00_Addr         pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls28_Constructor");

    RepServMessages *pMsgObj = RepServMessages::Instance();
    tsp00_Int4       rc      = errOK_els00;

    pTLRec = new tls00_TloadRec;
    if (NULL == pTLRec)
    {
        rc = errNoMoreMemory_els98;
    }
    else
    {
        memset(pTLRec, 0, sizeof(tls00_TloadRec));

        //* Number of possible pages per packet to send
        //  tls00_PktBlock contains packet-, segment- and part header size. Because each packet
        //  consists of 2 parts the part header has to be substracted again.
        pTLRec->tlrPgPerPckt_ls00 = STATIC_CAST( tsp00_Int2, (pDBInfo->dbiPktSize_ls00 - sizeof(tls00_PktBlock)   
                                                                - sizeof(tsp1_part_header)
                                                                - sizeof(tsp7_bdinfo)) / pDBInfo->dbiPageSize_ls00 );
        
        // In case the page size allows only one page per packet we need to stop processing;
        // otherwise the kernel would core
        if (1 == pTLRec->tlrPgPerPckt_ls00)
        {
            RepServMessages::Instance()->ls98Msg(layTableLoad_els98, pszErrText, errTLPacketSizeTooSmall_els98, (pDBInfo->dbiPageSize_ls00/1024));
            rc = errTLPacketSizeTooSmall_els98;
        }
        else
        {
            pTLRec->tlrPageSize_ls00  = pDBInfo->dbiPageSize_ls00;

            pTLRec->tlrSwapFromDb_ls00.becomes(sw_dummy);
            pTLRec->tlrSwapFromTE_ls00.becomes(sw_dummy);

            pTLRec->tlrPageTypeExtracting_ls00 = tptUnknown_els00;
        
            pTLRec->tlrNumTabRejected_ls00 = pDBInfo->dbiSession_ls00.siMaxAdmissibleErrors;

            //*
            //*     create and initialize mapping file and infile
            //*
            if ( (true == pTLCmd->tloPartSpec_ls00.psSelAll_ls00) ||
                 (true == pTLCmd->tloPartSpec_ls00.psSelUser_ls00) )
            {
                //*
                //*     Mapping file
                //* Here the filenames must newly be created
                //*
                pTLRec->tlrMappingFile_ls00 = new tls00_VFile;
            
                pTLRec->tlrMappingFile_ls00->vfFileName_ls00 = new _TCHAR[PATH_MXSP00];
                pTLRec->tlrMappingFile_ls00->vfFileName_ls00[0] = 0;
                sprintf(pTLRec->tlrMappingFile_ls00->vfFileName_ls00, "%s%0*d",
                                                                      pTLCmd->tloInfile_ls00.fsFileName_ls00,
                                                                      FILE_EXTENSION_SIZE_LS26,
                                                                      0);
                ls30VFInitialize(*pTLRec->tlrMappingFile_ls00,
                                 pTLRec->tlrMappingFile_ls00->vfFileName_ls00,
                                 SP5VF_TEXT);

                rc = ls30VFOpen(*pTLRec->tlrMappingFile_ls00, pszErrText);
//TODOTODO make working without mapping file possible
                if (errOK_els00 == rc)
                {
                    //*
                    //*     Infile
                    //* simply initialize it; will be opened later depending on restart
                    //*
                    pTLRec->tlrInFile_ls00 = new tls00_VFile;

                    pTLRec->tlrInFile_ls00->vfFileName_ls00 = new _TCHAR[PATH_MXSP00];
                    pTLRec->tlrInFile_ls00->vfFileName_ls00[0] = 0;

                    _tcscpy(pTLRec->tlrInFile_ls00->vfFileName_ls00, pTLCmd->tloInfile_ls00.fsFileName_ls00);
                
                    ls30VFInitialize(*pTLRec->tlrInFile_ls00,
                                     pTLRec->tlrInFile_ls00->vfFileName_ls00,
                                     SP5VF_BINARY);


                    //*
                    //*     Restart file
                    //* Initialize and open it and initialize restart entries
                    //* In case of restart if the file does not exist we simply drop out a
                    //* msg to the log file saying that the file could not be found and
                    //* we start loading from the beginning.
                    //*
                    pTLRec->tlrRestart_ls00.rstRestartEntry_ls00 = new _TCHAR[RestartEntryLen_ls28];
                    pTLRec->tlrRestart_ls00.rstRestartDataFileName_ls00 = NULL;

                    pTLRec->tlrRestart_ls00.rstRestartFile_ls00 = new tls00_VFile;

                    if (true == pTLCmd->tloRestart_ls00)
                    {
                        // Check if file is available; this needs to be done before opening
                        // it because open it READWRITE will create it if it doesn't exist
                        tls00_LargeNumber l;
                        l.lnu2GB_Blocks_ls00 = 0;
                        l.lnuRemainder_ls00  = 0;
                        rc = ls30VFGetFileSize(pTLCmd->tloReport_ls00.fsFileName_ls00, &l);
                        if ( (errOK_els00 != rc) ||
                             ((0 == l.lnu2GB_Blocks_ls00) && (0 == l.lnuRemainder_ls00)) )
                        {
                            pMsgObj->ls98PrintMessage(layTableLoad_els98, wrnNoRestartFileFound_els98,
                                                                          pTLCmd->tloReport_ls00.fsFileName_ls00);
                            pTLCmd->tloRestart_ls00 = false;
                        }
                    }

                    // open for reading and writing; handle errors when open differently depending
                    // on whether restart requested or not
                    ls30VFInitialize(*pTLRec->tlrRestart_ls00.rstRestartFile_ls00,
                                     pTLCmd->tloReport_ls00.fsFileName_ls00,
                                     SP5VF_BINARY,
                                     SP5VF_READWRITE);
                    rc = ls30VFOpen(*pTLRec->tlrRestart_ls00.rstRestartFile_ls00, pszErrText);
                    if (errOK_els00 != rc)
                    {
                        if (true == pTLCmd->tloRestart_ls00)
                        {
                            pMsgObj->ls98PrintMessage(layTableLoad_els98, wrnNoRestartFileFound_els98,
                                                                          pTLCmd->tloReport_ls00.fsFileName_ls00);
                            pTLCmd->tloRestart_ls00 = false;
                        }
                        delete pTLRec->tlrRestart_ls00.rstRestartFile_ls00;
                        pTLRec->tlrRestart_ls00.rstRestartFile_ls00 = NULL;
                    }
                }
            }
            else        // branch of if ( (true == pTLCmd->tloPartSpec_ls00.psSelAll_ls00) || ...
            {
                // Here the filename is simply a pointer to the user given file name in TL cmd
                pTLRec->tlrInFile_ls00 = new tls00_VFile;       // infile

                ls30VFInitialize(*pTLRec->tlrInFile_ls00,
                                 pTLCmd->tloInfile_ls00.fsFileName_ls00,
                                 SP5VF_BINARY,
                                 SP5VF_READ);
            }   //end if TL ALL or TL USER IS chosen
        }   // end if (1 == pTLRec->tlrPgPerPckt_ls00)
    }   // end if (NULL == pTLRec)
    
    return rc;
}
// ls28_Constructor()


/*
  -----------------------------------------------------------------------------
  function: ls28_Destructor
  -----------------------------------------------------------------------------
*/
static void
ls28_Destructor(tls00_TloadRec *pTLRec, tls00_TabLoadCmd *pTLCmd)
{
    ROUTINE_DBG_MEO00 ("ls28_Destructor");

    if (NULL != pTLRec->tlrMappingFile_ls00)
    {
        if (0 != pTLRec->tlrMappingFile_ls00->vfFileHndl_ls00)      // close mapping file if open
        {
            _TCHAR szErr[256];
            ls30VFClose(*pTLRec->tlrMappingFile_ls00, szErr);
        }

        if (NULL != pTLRec->tlrMappingFile_ls00->vfFileName_ls00)
        {
            delete pTLRec->tlrMappingFile_ls00->vfFileName_ls00;
        }

        delete pTLRec->tlrMappingFile_ls00;
    }


    if (NULL != pTLRec->tlrInFile_ls00)
    {
        if (0 != pTLRec->tlrInFile_ls00->vfFileHndl_ls00)      // close last data file if still open
        {
            _TCHAR szErr[256];
            ls30VFClose(*pTLRec->tlrInFile_ls00, szErr);
        }

        if (pTLRec->tlrInFile_ls00->vfFileName_ls00 != pTLCmd->tloInfile_ls00.fsFileName_ls00)
        {
            delete [] pTLRec->tlrInFile_ls00->vfFileName_ls00;
            pTLRec->tlrInFile_ls00->vfFileName_ls00 = NULL;
        }

        delete pTLRec->tlrInFile_ls00;
    }

    if (NULL != pTLRec->tlrRestart_ls00.rstRestartEntry_ls00)
    {
        delete [] pTLRec->tlrRestart_ls00.rstRestartEntry_ls00;
        pTLRec->tlrRestart_ls00.rstRestartEntry_ls00 = NULL;
    }

    if (NULL != pTLRec->tlrRestart_ls00.rstRestartDataFileName_ls00)
    {
        delete [] pTLRec->tlrRestart_ls00.rstRestartDataFileName_ls00;
        pTLRec->tlrRestart_ls00.rstRestartDataFileName_ls00 = NULL;
    }


    if (NULL != pTLRec->tlrRestart_ls00.rstRestartFile_ls00)    // close report file
    {
        _TCHAR szErr[256];
        ls30VFClose(*pTLRec->tlrRestart_ls00.rstRestartFile_ls00, szErr);

        delete pTLRec->tlrRestart_ls00.rstRestartFile_ls00;
        pTLRec->tlrRestart_ls00.rstRestartFile_ls00 = NULL;
    }

    delete pTLRec;     // at last delete the structure itself

    return;
}
// ls28_Destructor()


/*
  -----------------------------------------------------------------------------
  function: ls28_AnalyseDataFileHeader
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_AnalyseDataFileHeader(tls00_DBInfo *pDBInfo, tls00_TloadRec *pTLRec, tls00_PartSpec *pPartSpec, tsp00_Addr pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls28_AnalyseDataFileHeader");

    RepServMessages *pMsgObj = RepServMessages::Instance();    // error report object
    tsp00_Int4       rc      = errOK_els00;

    rc = ls28ReadDataFileHeader(pTLRec, pszErrText);        // read file header (1024 byte)
    if (errOK_els00 == rc)
    {
        // Check for right encoding types - ascii may only be loaded to ascii db and
        // ucs2 may  only be loaded to ucs2 db
        if (true == ls00IsDBUnicode(pDBInfo))
        {
            if (ctAscii_els00 == pTLRec->tlrSourceEncoding_ls00)
            {
                pMsgObj->ls98Msg(layTableLoad_els98, pszErrText, errTLSrcAndTgtEncodingDifferent_els98,
                                                                 "ASCII", "UNICODE");
                rc = errTLSrcAndTgtEncodingDifferent_els98;
                
            }

            // Here we need to check for different swap of the lserver while
            // extracting and loading
            if ( tsp00_SwapKind_Enum(pDBInfo->dbiOwnSwap_ls00) != tsp00_SwapKind_Enum(pTLRec->tlrSwapFromTE_ls00) )
            {
                pMsgObj->ls98Msg(layTableLoad_els98, pszErrText, errTLSrcAndTgtSwapDifferent_els98);
                rc = errTLSrcAndTgtSwapDifferent_els98;
            }
        }
        else
        {
            if ( (ctUCS2_els00        == pTLRec->tlrSourceEncoding_ls00) ||
                 (ctUCS2Swapped_els00 == pTLRec->tlrSourceEncoding_ls00) ||
                 (ctUCS2Native_els00  == pTLRec->tlrSourceEncoding_ls00)  )
            {
                pMsgObj->ls98Msg(layTableLoad_els98, pszErrText, errTLSrcAndTgtEncodingDifferent_els98,
                                                                 "UNICODE", "ASCII");
                rc = errTLSrcAndTgtEncodingDifferent_els98;
            }
        }
        
        // Trace swap behavior of machine repman run on during TABLEEXTRACT
        TRACE_PRNF_MLS99(("ls28_AnalyseDataFileHeader", "Swap at TE = %d\n", STATIC_CAST(int, pTLRec->tlrSwapFromTE_ls00)));

        // Check in header data for multiple table flag and reject if not set because we cannot
        // process multiple table data files with a single table data file
        if (errOK_els00 == rc)
        {
            if ( (true == pPartSpec->psSelAll_ls00) || (true == pPartSpec->psSelUser_ls00) )
            {
                if (false == pTLRec->tlrMultTabInFile_ls00)
                {
                    pMsgObj->ls98Msg(layTableLoad_els98, pszErrText, errTLSingleTabFileMultTab_els98);
                    rc = errTLSingleTabFileMultTab_els98;
                }
            }
            else
            {
                if (true == pTLRec->tlrMultTabInFile_ls00)
                {
                    pMsgObj->ls98Msg(layTableLoad_els98, pszErrText, errTLMultTabFileSingleTab_els98,
                                                                     pPartSpec->psTabName_ls00.asCharp());
                    rc = errTLMultTabFileSingleTab_els98;
                }
            }
        }
    }
    return rc;
}
// ls28_AnalyseDataFileHeader()


/*
  -----------------------------------------------------------------------------
  function:     ls28_GetTargetKernelSwap
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_GetTargetKernelSwap(tls00_DBInfo *pDBInfo, tls00_TloadRec *pTLRec, tsp00_Addr pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls28_GetTargetKernelSwap");
    
    tsp00_Int4 rc = errOK_els00;

    _TCHAR szCmd[256];
    szCmd[0] = 0;
    _tcscpy(szCmd, "CREATE TABLE LOA_SWAP (COL1 INT)");

    tsp00_Int4 lErrPos;
    _TCHAR     szErr[MAX_REPLY_LENGTH_LS00];

    rc = ls13ExecIntSQLStmtUTF8(pDBInfo, 0, szCmd, szErr, lErrPos);
    if (errOK_els00 != rc)
    {
        RepServMessages::Instance()->ls98Msg(layTableLoad_els98, pszErrText, errInternSQL_els98,
                                                                             szCmd, szErr);
        rc = errInternSQL_els98;
    }
    else
    {
        _TCHAR szTabName[128] = {'L','O','A','_','S','W','A','P'};

        tsp00_Addr pszTD = NULL;
        tsp00_Int4 lLen  = 0;

        rc = ls13GetTableDescription(pDBInfo, 0, szTabName, lLen, pszTD, pszErrText);
        if (errOK_els00 == rc)
        {
            pTLRec->tlrTargetSwap_ls00 = (REINTERPRET_CAST(tsp7_table_description*, pszTD))->sp7t_kernel_swap();

            szCmd[0] = 0;
            _tcscpy(szCmd, "DROP TABLE LOA_SWAP");
            rc = ls13ExecIntSQLStmtUTF8(pDBInfo, 0, szCmd, szErr, lErrPos);
        }
        if (errOK_els00 != rc)
        {
            RepServMessages::Instance()->ls98Msg(layTableLoad_els98, pszErrText, errInternSQL_els98,
                                                                     szCmd, szErr);
            rc = errInternSQL_els98;
        }
    }

    TRACE_PRNF_MLS99(("ls28_GetTargetKernelSwap", "Target DB kernel swap = %d.\n", STATIC_CAST(int, pTLRec->tlrTargetSwap_ls00)));

    return rc;
}
// ls28_GetTargetKernelSwap()


/*
  -----------------------------------------------------------------------------
  function:     ls28_HandleRestart
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_HandleRestart(tls00_Restart *pRestartStruct,
                   tls00_VFile   *pMapFile,
                   tsp00_Addr     pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls28_HandleRestart");

    RepServMessages *pMsgObj = RepServMessages::Instance();    // error report object

    tsp00_Int4 rc = errOK_els00;

    bool bRestartEntryFound = false;

    // read restart entry from restart file
    tsp00_Int4 lRestartEntryLen = ls30VFReadLnEx(*pRestartStruct->rstRestartFile_ls00,
                                                 pRestartStruct->rstRestartEntry_ls00,
                                                 pszErrText,
                                                 RestartEntryLen_ls28);
    if (lRestartEntryLen > 0)
    {
        // Remove trailing new line characters from restart entry
        while ( iscntrl(pRestartStruct->rstRestartEntry_ls00[lRestartEntryLen - 1]) )
        {
            --lRestartEntryLen;
        }
        pRestartStruct->rstRestartEntry_ls00[lRestartEntryLen] = 0;     // set trailing 0


        tsp00_Longint  lRDFNameLen = 2 * PATH_MXSP00;
        _TCHAR         pszData[2 * PATH_MXSP00];                   // pointer holding read data
        _TCHAR        *pszPos      = NULL;                     // pointer for searching the string
        

        pRestartStruct->rstRestartDataFileName_ls00 = new _TCHAR[lRDFNameLen];

        // Read data for first time and then until restart entry is found
        tsp00_Int4 lByteCount = ls30VFReadLnEx(*pMapFile, pszData, pszErrText, 256);
        while ( (EOF_LS00 != lByteCount) && (lByteCount > 0) )
        {
            // check for either file name (line start with '[') or for the restart entry
            if ('[' == pszData[0])
            {
                pszPos = strchr(pszData, ']');

                tsp00_Longint len = (pszPos - &pszData[1]);
                if (len > lRDFNameLen)
                {
                    lRDFNameLen = len;
                    delete pRestartStruct->rstRestartDataFileName_ls00;
                    pRestartStruct->rstRestartDataFileName_ls00 = new _TCHAR[len + 1];
                }
                memcpy(pRestartStruct->rstRestartDataFileName_ls00, &pszData[1], len);
                pRestartStruct->rstRestartDataFileName_ls00[len] = 0;
            }
            else
            {
                if (0 == memcmp(pszData, pRestartStruct->rstRestartEntry_ls00, lRestartEntryLen))
                {
                    // When the restart entry is located in mapping file we need to get the end
                    // position because the restart entry always specifies the
                    // last successfully loaded table;
                    // position specification looks like this
                    // <BEGIN BLOCKS>:<BEGIN POS>-<END BLOCKS>:<END POS>
                    pszPos = strchr(pszData, '-');          // find backwards the position separating dash
                    if (NULL != pszPos)
                    {
                        pRestartStruct->rstRestartPos_ls00.lnu2GB_Blocks_ls00 = atoi(pszPos+1);

                        pszPos = strchr(pszPos, ':');       // find separating colon
                        if (NULL != pszPos)
                        {
                            pRestartStruct->rstRestartPos_ls00.lnuRemainder_ls00 = atoi(pszPos+1);
                            bRestartEntryFound = true;
                        }
                    }
                    break;
                }
            }

            lByteCount = ls30VFReadLnEx(*pMapFile, pszData, pszErrText, 256);
        }

        if ( (lByteCount < 0) && (lByteCount != EOF_LS00) )
        {
            // Error reading mapping file; the error is already logged and we return
            pMsgObj->ls98Msg(layTableLoad_els98, pszErrText, errTLRestartFileOpen_els98,
                                                             pMapFile->vfFileName_ls00);
            rc = errTLRestartFileOpen_els98;
        }
        else
        {
            if (false == bRestartEntryFound)
            {
                pMsgObj->ls98Msg(layTableLoad_els98, pszErrText, errTLRestartEntryNotFound_els98,
                                                                 pRestartStruct->rstRestartEntry_ls00,
                                                                 pMapFile->vfFileName_ls00);
                rc = errTLRestartEntryNotFound_els98;
            }
        }
    }
    else
    {
        if (EOF_LS00 != lRestartEntryLen)
        {
            pMsgObj->ls98Msg(layTableLoad_els98, pszErrText, errTLRestartFileOpen_els98,
                                                             pRestartStruct->rstRestartFile_ls00->vfFileName_ls00);
            rc = errTLRestartFileOpen_els98;
        }
    }   // end if (lReadLen > 0)

    return rc;
}
// ls28_HandleRestart()


/*
  -----------------------------------------------------------------------------
  function:     ls28_ReopenDataFile
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_ReopenDataFile(tls00_DBInfo   *pDBInfo,
                    tls00_TloadRec *pTLRec,
                    tsp00_Addr      pszSingleFileName,
                    tls00_PartSpec *pPartSpec,
                    tsp00_Addr      pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls28_ReopenDataFile");

    RepServMessages *pMsgObj = RepServMessages::Instance();    // error report object

    tsp00_Int4 rc = errOK_els00;

    tls00_VFile *pInFile = pTLRec->tlrInFile_ls00;

    // find position to concatenate new file name
    tsp00_Addr pPos = NULL;
    pPos = strrchr(pInFile->vfFileName_ls00, BACKSLASH_LS00);
    if (NULL == pPos)   // unix file?
    {
        pPos = strrchr(pInFile->vfFileName_ls00, SLASH_LS00);
    }

    if (NULL == pPos)       // no unix, no win file -> use the beginning of path name as starting point
    {
        pPos = pInFile->vfFileName_ls00;
    }
    else
    {
        pPos = &pInFile->vfFileName_ls00[pPos - pInFile->vfFileName_ls00 + 1];
    }

    // concatenate file name
    memcpy(pPos, pszSingleFileName, _tcslen(pszSingleFileName));
    pPos[_tcslen(pszSingleFileName)] = 0;

    //*
    //*     Open new infile
    //* - close old file if open; ls30VFClose doesn't pay any attention on the file name
    //*   but uses the file handle only; so the already assigned new file name is ok
    //* - open new file
    //* - analyze data file header
    //*
    ls30VFClose(*pInFile, pszErrText);

    rc = ls30VFOpen(*pInFile, pszErrText);
    if (errOK_els00 == rc)
    {
        pMsgObj->ls98PrintMessage(layTableLoad_els98, msgTL_InputFile_els98,
                                                      pInFile->vfFileName_ls00);

        rc = ls28_AnalyseDataFileHeader(pDBInfo, pTLRec, pPartSpec, pszErrText);
    }

    return rc;
}
// ls28_ReopenDataFile()


/*
  -----------------------------------------------------------------------------
  function:     ls28_GetTableNamesFromDataFile
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_GetTableNamesFromDataFile(tls00_DBInfo            *pDBInfo,
                               tls00_TloadRec          *pTLRec,
                               tls00_TabLoadCmd        *pTLCmd,
                               SAPDB_DynamicUTF8String &UserName,
                               SAPDB_DynamicUTF8String &TableName,
                               tsp00_Addr               pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls28_GetTableNamesFromDataFile");

    RepServMessages *pMsgObj = RepServMessages::Instance();    // error report object

    tsp00_Int4 rc = errOK_els00;

    tsp00_Addr pszBuffer = NULL;
    tsp00_Int4 anzahl = 0;

    if (true == pTLCmd->tloRestart_ls00)
    {
        // reset restart flag
        pTLCmd->tloRestart_ls00 = false;
        
        rc = ls28_HandleRestart(&pTLRec->tlrRestart_ls00, pTLRec->tlrMappingFile_ls00, pszErrText);
        if (errOK_els00 == rc)
        {

            rc = ls28_ReopenDataFile(pDBInfo,
                                     pTLRec,
                                     pTLRec->tlrRestart_ls00.rstRestartDataFileName_ls00,
                                     &pTLCmd->tloPartSpec_ls00,
                                     pszErrText);
            if (errOK_els00 == rc)
            {
                // seek to right restart position
                rc = ls30VFSeekLong(*pTLRec->tlrInFile_ls00,
                                    &pTLRec->tlrRestart_ls00.rstRestartPos_ls00,
                                    pszErrText,
                                    SP5VF_SEEK_BEGIN);
            }
        }
        else
        {
            if (EOF_LS00 == rc)     // in this case the restart file is simply empty and we start
            {                       // working as if no restart was specified
                rc = errOK_els00;
            }
        }

        if (errOK_els00 != rc)
        {
            return rc;
        }
    }


    anzahl = ls30VFReadLn(*pTLRec->tlrMappingFile_ls00, &pszBuffer, pszErrText, 0);
    if (EOF_LS00 != anzahl)
    {
        if ('[' == pszBuffer[0])
        {
            // build new file name for data file using read data
            pszBuffer[_tcslen(pszBuffer) - 1] = 0;      // remove closing ]

            // while calling remove opening [
            rc = ls28_ReopenDataFile(pDBInfo, pTLRec, &pszBuffer[1], &pTLCmd->tloPartSpec_ls00, pszErrText);
            if (errOK_els00 == rc)
            {
                // recursive call: get the first table name for new data file
                rc = ls28_GetTableNamesFromDataFile(pDBInfo, pTLRec, pTLCmd, UserName, TableName, pszErrText);
            }
        }
        else
        {
            // Variables used to ensure the string assigned is UTF8; otherwise it is converted
            unsigned int charLength   = 0;
            unsigned int byteLength   = 0;
            int          isTerminated = 0;
            int          isCorrupted  = 0;
            int          isExhausted  = 0;
            int          IsCorruptedUTF8  = 0;


            //*
            //*     Read and assign table name and user name and
            //*     move to right position in case not yet there
            //*

            //*
            //*     User name
            //*
            tsp00_Addr pQuotePos1 = NULL;
            tsp00_Addr pQuotePos2 = NULL;
            
            // im fall von TL USER ist der owner immer auf den angemeldeten benutzer gesetzt und damit
            // richtig und muss nicht weiter beachtet werden
            // im fall von TL ALL muss der owner immer auf den owner der entladenen tabelle gesetzt
            // werden, da hier alle entladenen user auch in der zieldatenbank vorhanden sein muessen
            pQuotePos1 = strchr(pszBuffer, QUOTE_LS00);           // there MUST be double quotes
            pQuotePos2 = strchr(pQuotePos1 + 1, QUOTE_LS00);      // there MUST be next double quotes

            const size_t lConstNameLen         = 3*sizeof(tsp00_KnlIdentifierc)+1;
            size_t       lNameLen              = 0;
            char         szName[lConstNameLen];
            
            if (true == pTLCmd->tloPartSpec_ls00.psSelAll_ls00)
            {
                lNameLen = pQuotePos2 - (pQuotePos1 + 1);
                memcpy(szName, pQuotePos1+1, lNameLen);
                szName[lNameLen] = 0;
            }
            else
            {
                lNameLen = _tcslen(pTLCmd->tloPartSpec_ls00.psTabOwnr_ls00.asCharp());
                memcpy(szName, pTLCmd->tloPartSpec_ls00.psTabOwnr_ls00.asCharp(), lNameLen);
                szName[lNameLen] = 0;
            }

            IsCorruptedUTF8 = sp77encodingUTF8->stringInfo(szName,       lNameLen,
                                                           1,            &charLength,
                                                           &byteLength,  &isTerminated,
                                                           &isCorrupted, &isExhausted);
            if (0 == IsCorruptedUTF8)
            {
                UserName.Assign(REINTERPRET_CAST(SAPDB_DynamicUTF8String::ConstPointer, szName));
            }
            else
            {
                rc = ls05ConvertToUTF8(UserName, REINTERPRET_CAST(unsigned char*, &szName[0]), lNameLen, ctAscii_els00, pszErrText);
            }

            //*
            //*     Table name
            //*
            pQuotePos1 = strchr(pQuotePos2+1, QUOTE_LS00);        // there MUST be double quotes
            pQuotePos2 = strchr(pQuotePos1+1, QUOTE_LS00);        // there MUST be next double quotes

            lNameLen = pQuotePos2 - (pQuotePos1 + 1);
            memcpy(szName, pQuotePos1+1, lNameLen);
            szName[lNameLen] = 0;

            IsCorruptedUTF8 = sp77encodingUTF8->stringInfo(szName,       lNameLen,
                                                           1,            &charLength,
                                                           &byteLength,  &isTerminated,
                                                           &isCorrupted, &isExhausted);
            if (0 == IsCorruptedUTF8)
            {
                TableName.Assign(REINTERPRET_CAST(SAPDB_DynamicUTF8String::ConstPointer, szName));
            }
            else
            {
                rc = ls05ConvertToUTF8(TableName, REINTERPRET_CAST(unsigned char*, &szName[0]), lNameLen, ctAscii_els00, pszErrText);
            }

            tsp00_Longint len = pQuotePos2 - pszBuffer + 1;
            memcpy(pTLRec->tlrRestart_ls00.rstRestartEntry_ls00, pszBuffer, len);
            pTLRec->tlrRestart_ls00.rstRestartEntry_ls00[len] = 0;

            // Get position of table data in data file
            char sznum[10];
            tls00_LargeNumber lDataPos;

            pQuotePos1 = pQuotePos2 + sizeof(char);
            pQuotePos2 = strchr(pQuotePos2, ':');

            len = pQuotePos2 - (pQuotePos1);
            memcpy(sznum, pQuotePos1, len);
            sznum[len] = 0;
            lDataPos.lnu2GB_Blocks_ls00 = atoi(sznum);

            pQuotePos1 = pQuotePos2 + sizeof(char);
            pQuotePos2 = strchr(pQuotePos2, '-');

            len = pQuotePos2 - (pQuotePos1);
            memcpy(sznum, pQuotePos1, len);
            sznum[len] = 0;
            lDataPos.lnuRemainder_ls00 = atoi(sznum);

            // Now seek to right position in data file (may be one table name was removed from mapping file)
            rc = ls30VFSeekLong(*pTLRec->tlrInFile_ls00, &lDataPos, pszErrText);

            delete pszBuffer;
        }
    }   // if ( anzahl != EOF_LS00 )

    if (EOF_LS00 == anzahl)
    {
        rc = STOP_NODATA_LS28;
    }

    return rc;
}
// ls28_GetTableNamesFromDataFile()


/*
  -----------------------------------------------------------------------------
  function:     ls28_WriteRestartEntry
  -----------------------------------------------------------------------------
*/
tsp00_Int4
ls28_WriteRestartEntry(tls00_Restart *pRestartStruct, tsp00_Addr pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls28_WriteRestartEntry");

    tsp00_Int4 rc = errOK_els00;
    
    if (NULL != pRestartStruct->rstRestartFile_ls00)
    {
        if (0 != pRestartStruct->rstRestartFile_ls00->vfFileHndl_ls00)      // file is open
        {
            rc = ls30VFSeek(*pRestartStruct->rstRestartFile_ls00, 0, pszErrText);       // before truncating we MUST seek

            rc = ls30VFTruncate(*pRestartStruct->rstRestartFile_ls00, pszErrText, 0);
            
            rc = ls30VFWriteLn(*pRestartStruct->rstRestartFile_ls00,    // write restart entry
                                pRestartStruct->rstRestartEntry_ls00,
                                _tcslen(pRestartStruct->rstRestartEntry_ls00),
                                pszErrText);
        }
    }

    return rc;
}
// ls28_WriteRestartEntry()


/*
  -----------------------------------------------------------------------------
  function:     ls28_LoadUserTables
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_LoadUserTables(tls00_DBInfo     *pDBInfo,
                    tls00_TabLoadCmd *pTLCmd,
                    tls00_TloadRec   *pTLRec,
                    tsp00_Addr        pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls28_LoadUserTables");

    RepServMessages *pMsgObj = RepServMessages::Instance();    // error report object

    tsp00_Int4 rc = errOK_els00;

    SAPDB_DynamicUTF8String UserName;
    SAPDB_DynamicUTF8String TableName;

    while (errOK_els00 == rc)
    {
        rc = ls28_GetTableNamesFromDataFile(pDBInfo, pTLRec, pTLCmd, UserName, TableName, pszErrText);
        if (errOK_els00 == rc)
        {
            rc = ls28LoadSingleTable(pDBInfo, pTLCmd, pTLRec, UserName, TableName, pszErrText);
            if (errOK_els00 == rc)
            {
                ++pTLRec->tlrNumTabLoaded_ls00;
                pszErrText[0] = 0;
                rc = ls28_WriteRestartEntry(&pTLRec->tlrRestart_ls00, pszErrText);
            }
        }
        
        if ( (errOK_els00 != rc) &&
            (STOP_NODATA_LS28 != rc) && (errInternal_els98 != rc) && (errCancelRequest_els98 != rc) )
        {
            --pTLRec->tlrNumTabRejected_ls00;
            if (pTLRec->tlrNumTabRejected_ls00 > 0)
            {
                rc = errOK_els00;
            }
        }
    }

    if (STOP_NODATA_LS28 == rc)     // no more data to process
    {
        rc = errOK_els00;
    }

    pMsgObj->ls98PrintMessage(layTableLoad_els98, msgLoadedRejectedTables_els98,
                                                  pTLRec->tlrNumTabLoaded_ls00,
                                                  (pDBInfo->dbiSession_ls00.siMaxAdmissibleErrors -
                                                   pTLRec->tlrNumTabRejected_ls00) );
    if (errOK_els00 == rc)
    {
        pMsgObj->ls98PrintMessage(layTableLoad_els98, msgTLSuccessful_els98);
    }
    
    return rc;
}
// ls28_LoadUserTables()


/*
  -----------------------------------------------------------------------------
  function:     ls28_LoadAllTables
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls28_LoadAllTables(tls00_DBInfo     *pDBInfo,
                   tls00_TabLoadCmd *pTLCmd,
                   tls00_TloadRec   *pTLRec,
                   tsp00_Addr        pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls28_LoadAllTables");

    RepServMessages *pMsgObj = RepServMessages::Instance();    // error report object

    tsp00_Int4 rc = errOK_els00;

    //*
    //*     define user name for switching users
    //*
    SAPDB_DynamicUTF8String CurrentUser;
    SAPDB_DynamicUTF8String NewUser;
    
    SAPDB_DynamicUTF8String SessionUser;
    rc = ls05ConvertToUTF8(SessionUser,
                           REINTERPRET_CAST(unsigned char*, pDBInfo->dbiSession_ls00.siUserName_ls00.asCharp()),
                           _tcslen(pDBInfo->dbiSession_ls00.siUserName_ls00.asCharp()),
                           ctAscii_els00, pszErrText);
    CurrentUser = SessionUser;

    rc = ls15CheckUserMode(pDBInfo, SessionUser, dbuSysdba_els00, pszErrText);
    if (errOK_els00 != rc)
    {
        if (dbRowNotFound_els00 == rc)
        {
            pMsgObj->ls98Msg(layTableExtract_els98, pszErrText, errRestrictedStmt_els98);
        }
    }
    else
    {
        //*
        //*     Main loop to load tables
        //*
        SAPDB_DynamicUTF8String TableName;

        while (errOK_els00 == rc)
        {
            // Gets the name already in UTF8 encoding
            rc = ls28_GetTableNamesFromDataFile(pDBInfo, pTLRec, pTLCmd, NewUser, TableName, pszErrText);
            if (errOK_els00 == rc)
            {
                rc = ls26SwitchUser(pDBInfo, CurrentUser, NewUser, pszErrText);   // user switch
                if (errOK_els00 == rc)
                {
                    CurrentUser = NewUser;

                    rc = ls28LoadSingleTable(pDBInfo, pTLCmd, pTLRec, NewUser, TableName, pszErrText);
                    if (errOK_els00 == rc)
                    {
                        ++pTLRec->tlrNumTabLoaded_ls00;
                        pszErrText[0] = 0;
                        rc = ls28_WriteRestartEntry(&pTLRec->tlrRestart_ls00, pszErrText);
                    }
                }
            }

            if ( (errOK_els00 != rc) &&
                 (STOP_NODATA_LS28 != rc) && (errInternal_els98 != rc) && (errCancelRequest_els98 != rc) )
            {
                --pTLRec->tlrNumTabRejected_ls00;
                if (pTLRec->tlrNumTabRejected_ls00 > 0)
                {
                    rc = errOK_els00;
                }
            }
        }   // end while (errOK_els00 == rc)

        if (STOP_NODATA_LS28 == rc)     // no more data to process
        {
            rc = errOK_els00;
        }

        pMsgObj->ls98PrintMessage(layTableLoad_els98, msgLoadedRejectedTables_els98,
                                                      pTLRec->tlrNumTabLoaded_ls00,
                                                      (pDBInfo->dbiSession_ls00.siMaxAdmissibleErrors -
                                                       pTLRec->tlrNumTabRejected_ls00) );

        if (errOK_els00 == rc)
        {
            pMsgObj->ls98PrintMessage(layTableLoad_els98, msgTLSuccessful_els98);
        }

        if (false == NewUser.Empty())     // last user switch to restore the acting user
        {
            tsp00_Int4 rc2 = ls26SwitchUser(pDBInfo, CurrentUser, SessionUser, pszErrText);
            if ( (errOK_els00 != rc2) && (errOK_els00 == rc) )
            {
                rc = rc2;
            }
        }

    }   // end if (errOK_els00 != rc)

    return rc;
}
// ls28_LoadAllTables()



/*
  -----------------------------------------------------------------------------
  function:     ls28_LoadOneTable
  -----------------------------------------------------------------------------
*/
static tsp00_Longint
ls28_LoadOneTable(tls00_DBInfo     *pDBInfo,
                  tls00_TabLoadCmd *pTLCmd,
                  tls00_TloadRec   *pTLRec,
                  tsp00_Addr        pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls28_LoadOneTable");
    
    RepServMessages *pMsgObj = RepServMessages::Instance();    // error report object
    tsp00_Longint    rc      = errOK_els00;

    rc = ls30VFOpen(*pTLRec->tlrInFile_ls00, pszErrText);       // open data file

    SAPDB_DynamicUTF8String UserName;
    SAPDB_DynamicUTF8String TableName;
    tsp00_Int4              lUserNameLen = _tcslen(pTLCmd->tloPartSpec_ls00.psTabOwnr_ls00.asCharp());

    if (lUserNameLen > 0)
    {
        //*
        //*     Check if user name and owner name are the same - if
        //*     not TABLELOAD is impossible in 7.3 release
        //*
        rc = errTLUserIsNotOwner_els98;
        if (lUserNameLen == pDBInfo->dbiSession_ls00.siUserName_ls00.length())
        {
            if (0 == memcmp(pTLCmd->tloPartSpec_ls00.psTabOwnr_ls00.asCharp(),
                            pDBInfo->dbiSession_ls00.siUserName_ls00.asCharp(),
                            lUserNameLen))
            {
                rc = errOK_els00;
                rc = ls05ConvertToUTF8(UserName,
                                       REINTERPRET_CAST(unsigned char *, pTLCmd->tloPartSpec_ls00.psTabOwnr_ls00.asCharp()),
                                       lUserNameLen,
                                       ctAscii_els00,
                                       pszErrText);
            }
        }

        // Print error message here only if user is not owner; ls05ConvertToUTF8 handles errors itself
        if (errTLUserIsNotOwner_els98 == rc)
        {
            pMsgObj->ls98Msg(layTableLoad_els98, pszErrText, errTLUserIsNotOwner_els98);
        }
    }
    else
    {
        // Make sure we deliver a user name to the next function
        rc = ls05ConvertToUTF8(UserName,
                               REINTERPRET_CAST(unsigned char *, pDBInfo->dbiSession_ls00.siUserName_ls00.asCharp()),
                               pDBInfo->dbiSession_ls00.siUserName_ls00.length(),
                               ctAscii_els00,
                               pszErrText);

    }

    if (errOK_els00 == rc)
    {
        rc = ls05ConvertToUTF8(TableName,
                               REINTERPRET_CAST(unsigned char *, pTLCmd->tloPartSpec_ls00.psTabName_ls00.asCharp()),
                               _tcslen(pTLCmd->tloPartSpec_ls00.psTabName_ls00.asCharp()),
                               ctAscii_els00,
                               pszErrText);
    
        if (errOK_els00 == rc)                                      // Read data file header
        {
            rc = ls28_AnalyseDataFileHeader(pDBInfo, pTLRec, &pTLCmd->tloPartSpec_ls00, pszErrText);
            if (errOK_els00 == rc)
            {
                rc = ls28LoadSingleTable(pDBInfo, pTLCmd, pTLRec, UserName, TableName, pszErrText);
            }
        }
    }

    if (errOK_els00 == rc)
    {
        pMsgObj->ls98PrintMessage(layTableLoad_els98, msgTLSuccessful_els98);
    }

    return rc;
}
// ls28_LoadOneTable()

/*
  -----------------------------------------------------------------------------
  Chapter: Private functions implementation
  -----------------------------------------------------------------------------
*/
