/*!
  -----------------------------------------------------------------------------

  module: vls30.cpp

  -----------------------------------------------------------------------------

  responsible:  SteffenS

  special area: Replication Manager

  description:  Implementation of functions for Virtual File access

  version:      7.3.

  -----------------------------------------------------------------------------

  copyright:    SAP AG, 1997-2002

  -----------------------------------------------------------------------------


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*/
#include <assert.h>

#include "gls00.h"
#include "hls30.h"
#include "hls07.h"
#include "hls99.h"
#include "hls98msg.h"

#include "hsp100.h"     // versionsbestimmung

/* Lokale Funktionen - DEFINITIONEN: ********************************************/


/*!
  -----------------------------------------------------------------------------

  function:     ls30_ReadLnTxt
  -----------------------------------------------------------------------------

  description:  Read line of a file opened in text mode. If the buffer max size
                is too small to hold a whole data row an error is generated.
                Skips empty lines.

  arguments:    pszBuffer   [OUT]   - data buffer
                Vfile       [IN]    - virtual file structure
                lBytes      [IN]    - Count of bytes to read.
                ErrText     [OUT]   - error text
                MaxReadLen  [IN]    - maximum bytes to read;
                                      defaults to LINELENGTH_MXSP00

  returnvalue:  Read Length     - number of bytes read
                EOF_LS00 = -1   - reached end of file
                errFileRead     - error reading data from file

                In case of error the function logs an error message in the error
                text string AND in the log file.
  -----------------------------------------------------------------------------
*/
static tsp00_Longint
ls30_ReadLnTxt(tsp00_Addr     pszBuffer,
               tls00_VFile   &VFile,
               tsp00_Longint  lBytes,
               tsp00_Addr     pszErrorText,
               tsp00_Longint  MaxReadLen = LINELENGTH_MXSP00);


/*!
  -----------------------------------------------------------------------------

  function:     ls30_ReadLnBin

  -----------------------------------------------------------------------------

  description:  Read requested count of bytes of a file opened in binary mode.


  arguments:    pszBuffer       [OUT]   - data buffer
                Vfile           [IN]    - virtual file structure
                lBytes          [IN]    - Count of bytes to read.
                ErrText         [OUT]   - error text

  returnvalue:  Read Length     - number of bytes read
                EOF_LS00 = -1   - reached end of file
                errFileRead     - error reading data from file

                In case of error the function logs an error message in the error
                text string AND in the log file.
  -----------------------------------------------------------------------------
*/
static tsp00_Longint
ls30_ReadLnBin(tsp00_Addr    &pszBuffer,
               tls00_VFile   &VFile,
               tsp00_Longint  lBytes,
               tsp00_Addr     pszErrorText);




/*
  -----------------------------------------------------------------------------
  Global functions: implementation
  -----------------------------------------------------------------------------
*/

/*
  -----------------------------------------------------------------------------
  function:     ls30VFInitialize
  -----------------------------------------------------------------------------
*/
void ls30VFInitialize(tls00_VFile&              NewFileStruct,
                      tsp00_Addr                pszFileName,
                      tsp5_rte_datakind         DataType,
                      tsp5_rte_filemode         FileMode,
                      tsp5_rte_buffering_kind   BufferingKind,
                      tsp5_rte_closeoption      CloseOption)
{
    ROUTINE_DBG_MEO00 ("ls30VFInitialize");

    NewFileStruct.vfFileHndl_ls00                   = 0;
    NewFileStruct.vfFileName_ls00                   = pszFileName;
    NewFileStruct.vfDataType_ls00                   = DataType;
    NewFileStruct.vfFileMode_ls00                   = FileMode;
    NewFileStruct.vfFileBuf_ls00                    = BufferingKind;
    NewFileStruct.vfFileCOpt_ls00                   = CloseOption;
    NewFileStruct.vfFilePos_ls00.lnu2GB_Blocks_ls00 = 0;
    NewFileStruct.vfFilePos_ls00.lnuRemainder_ls00  = 0;
    NewFileStruct.vfFileLen_ls00.lnu2GB_Blocks_ls00 = 0;
    NewFileStruct.vfFileLen_ls00.lnuRemainder_ls00  = 0;
    NewFileStruct.vfFileLineNum_ls00                = 0;

    return;
}
// ls30VFInitialize()


/*
  -----------------------------------------------------------------------------
  function:     ls30VFOpen
  -----------------------------------------------------------------------------
*/
int
ls30VFOpen(tls00_VFile &VFile, tsp00_Addr ErrText)
{
    ROUTINE_DBG_MEO00 ("ls30VFOpen");

    tsp00_Int4         rc = errOK_els00;
    tsp5_rte_fileerror RteError;
    tsp5_rte_fileinfo  RteInfo;

    VFile.vfFilePos_ls00.lnuRemainder_ls00 = EOF_LS00;                 // Initialize to error

    //*
    //*     Open file
    //*
    sqlfopenc(VFile.vfFileName_ls00, VFile.vfDataType_ls00, VFile.vfFileMode_ls00,
              VFile.vfFileBuf_ls00, &VFile.vfFileHndl_ls00, &RteError);
    if (vf_ok == RteError.sp5fe_result)             // success
    {
        VFile.vfFilePos_ls00.lnu2GB_Blocks_ls00 = 0;
        VFile.vfFilePos_ls00.lnuRemainder_ls00  = 0;

        sqlfinfoc (VFile.vfFileName_ls00, &RteInfo, &RteError); // Get some file information
        if ( ( vf_ok == RteError.sp5fe_result ) && (SP5VF_WRITE != VFile.vfFileMode_ls00) )
        {
            // in case of WRITE the initial values for file size are valid; so we get
            // in here only for read or append
            VFile.vfFileLen_ls00.lnu2GB_Blocks_ls00 = RteInfo.sp5fi_size_x;
            VFile.vfFileLen_ls00.lnuRemainder_ls00  = RteInfo.sp5fi_size;
        }
    }

    // Check for error; both functions sqlfopenc and sqlfinfoc deliver errors in RteError
    if (vf_ok != RteError.sp5fe_result)
    {
        // File handle is undefined in this case (-1); we set it to 0 to be later able to 
        // decide if the file is opened and has to be close (file handle == 0 --> file not open)
        VFile.vfFileHndl_ls00 = 0;

        tsp00_C40c pszTmpLZUText;
        pszTmpLZUText.p2c(RteError.sp5fe_text);
        RepServMessages::Instance()->ls98Msg(layVFile_els98, ErrText, errFileOpen_els98,
                                                                      VFile.vfFileName_ls00,
                                                                      pszTmpLZUText.asCharp());
        VFile.vfFilePos_ls00.lnuRemainder_ls00 = EOF_LS00;    // Indicates error
        rc = errFileOpen_els98;
    }

    return rc;
}
// ls30VFOpen()


/*
  -----------------------------------------------------------------------------
  function:     ls30VFReadLn
  -----------------------------------------------------------------------------
*/
// lcount muss das anzuhaengende null mit einschliessen
int ls30VFReadLn(tls00_VFile   &VFile,
                 tsp00_Addr    *pszLineBuffer,
                 tsp00_Addr     ErrText,
                 tsp00_Longint  lCount,
                 tsp00_Longint  MaxReadLen)
{
    ROUTINE_DBG_MEO00 ("ls30VFReadLn");

    RepServMessages *pMsgObj = RepServMessages::Instance(); // error reporting obj

    tsp00_Longint ReadLen   = 0;
    tsp00_Addr    pszBuffer = NULL;

    if (lCount > MaxReadLen)
    {
        pMsgObj->ls98Msg(layVFile_els98, ErrText,errFileRead_els98, VFile.vfFileName_ls00,
                                                                    "Maximum read length exceeded");
        return errFileRead_els98;
    }

    // adjust bytes to read
    tsp00_Longint lBytesToRead = (lCount==0) ? MaxReadLen : lCount;

    pszBuffer = new char[lBytesToRead + sizeof(_TCHAR)];    // allocate memory for data to read and initialize it
    pszBuffer[0] = 0;

    ReadLen = ls30VFReadLnEx(VFile, pszBuffer, ErrText, lBytesToRead);
    if (ReadLen > 0)
    {
        *pszLineBuffer = pszBuffer;
    }
    else
    {
        delete [] pszBuffer;
        pszBuffer = NULL;
    }

    return STATIC_CAST( tsp00_Int4, ReadLen);           // cast because incompatibility to IA64
}
//  ls30VFReadLn()


/*
  -----------------------------------------------------------------------------
  function:     ls30VFReadLnEx
  -----------------------------------------------------------------------------
*/
int
ls30VFReadLnEx(tls00_VFile   &VFile,
               tsp00_Addr     pszLineBuffer,
               tsp00_Addr     pszErrText,
               tsp00_Longint  lCount,
               tsp00_Longint  MaxReadLen)
{
    ROUTINE_DBG_MEO00 ("ls30VFReadLnEx");

    tsp00_Longint       ReadLen   = 0;
    tsp00_Addr          pszBuffer = NULL;

    if (lCount > MaxReadLen)
    {
        RepServMessages::Instance()->ls98Msg(layVFile_els98, pszErrText,
                                                             errFileRead_els98,
                                                             VFile.vfFileName_ls00,
                                                             "Maximum read length exceeded");
        return errFileRead_els98;
    }

    //*     Set count of bytes to read
    tsp00_Longint lBytesToRead = (0 == lCount) ? MaxReadLen : lCount;

    if (SP5VF_TEXT == VFile.vfDataType_ls00)                        // text mode
    {
        ReadLen = ls30_ReadLnTxt(pszLineBuffer, VFile, lBytesToRead, pszErrText);
    }
    else                                                            // binary mode
    {
        ReadLen = ls30_ReadLnBin(pszLineBuffer, VFile, lBytesToRead, pszErrText);
    }


    if (ReadLen > 0)
    {
        // Adapt VFile structure members
        tsp00_Longint lFilePosIncrement = ReadLen;
        ++VFile.vfFileLineNum_ls00;                     // increase count of read rows

        if (SP5VF_TEXT == VFile.vfDataType_ls00)
        {
            lFilePosIncrement += NEW_LINE_LENGTH;
        }
        ls07AddLargeNumbers(VFile.vfFilePos_ls00, lFilePosIncrement);
    }
    
    return STATIC_CAST(tsp00_Int4, ReadLen);        // IA64
}
//  ls30VFReadLnEx()


/*
  -----------------------------------------------------------------------------
  function:     ls30VFReadLongBin
  -----------------------------------------------------------------------------
*/
tsp00_Longint
ls30VFReadLongBin(tsp00_Addr                pszBuffer,
                  tls00_VFile              &VFile,
                  tsp00_Longint             lBytes,
                  const tls00_LargeNumber  *plReadPosition,
                  tsp00_Addr                pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls30VFReadLongBin");

    tsp5_rte_fileerror RteError;
    tsp00_Longint      ReadLen = 0;
    tsp00_Int4         rc      = errOK_els00;

    // Seek to right position
    rc = ls30VFSeekLong(VFile, plReadPosition, pszErrText);
    if (errOK_els00 == rc)
    {
        // Read data
        sqlfreadc(VFile.vfFileHndl_ls00, pszBuffer, lBytes, &ReadLen, &RteError);
        if (RteError.sp5fe_result  != vf_ok)
        {
            if (vf_eof == RteError.sp5fe_result)
            {
                VFile.vfFilePos_ls00.lnuRemainder_ls00 = EOF_LS00;

                ReadLen = EOF_LS00;
            }
            else
            {
                tsp00_C40c pszTmpLZUText;
                pszTmpLZUText.p2c(RteError.sp5fe_text);
                RepServMessages::Instance()->ls98Msg(layVFile_els98, pszErrText,
                                                                     errFileRead_els98,
                                                                     VFile.vfFileName_ls00,
                                                                     pszTmpLZUText.asCharp());
                ReadLen = errFileRead_els98;
            }
        }
        else
        {            
            ls07AddLargeNumbers(VFile.vfFilePos_ls00, ReadLen);    // adapt file position
        }
    }
    else
    {
        ReadLen = rc;   // In case of error the error code is negative
    }

    return ReadLen;
}
//  ls30VFReadLongBin()


/*
  -----------------------------------------------------------------------------
  function:     ls30VFWriteLn
  -----------------------------------------------------------------------------
*/
int
ls30VFWriteLn(tls00_VFile   &VFile,
              tsp00_Addr     pszBuffer,
              tsp00_Longint  lCount,
              tsp00_Addr     pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls30VFWriteLn");

    tsp00_Int4          rc = errOK_els00;
    tsp5_rte_fileerror  RteError;

    sqlfwritec(VFile.vfFileHndl_ls00, pszBuffer, lCount, &RteError);
    if (vf_ok != RteError.sp5fe_result)
    {
        tsp00_C40c pszTmpLZUText;
        pszTmpLZUText.p2c(RteError.sp5fe_text);
        RepServMessages::Instance()->ls98Msg(layVFile_els98, pszErrText,
                                                             errFileWrite_els98,
                                                             VFile.vfFileName_ls00,
                                                             pszTmpLZUText.asCharp());
        rc = errFileWrite_els98;
    }
    else
    {
        if (SP5VF_TEXT == VFile.vfDataType_ls00)
        {
            ++VFile.vfFileLineNum_ls00;
            lCount += NEW_LINE_LENGTH;
        }

        ls07AddLargeNumbers(VFile.vfFileLen_ls00, lCount);
        VFile.vfFilePos_ls00 = VFile.vfFileLen_ls00;        // file pos and file len are identical
    }

    return rc;
}
//  ls30VFWriteLn()


/*
  -----------------------------------------------------------------------------
   function:     ls30VFWriteHeader
  -----------------------------------------------------------------------------
*/
int
ls30VFWriteHeader(tls00_VFile & VFile,
                  tsp00_Int4    lRecordLength,
                  tsp00_Addr    pszCommand,
                  tsp00_Addr    pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls30VFWriteHeader");

    tsp00_Int4          rc = errOK_els00;

    assert(VFile.vfDataType_ls00 == SP5VF_BINARY);  // opened in binary mode?

    tls30_DatafileHeader HeaderStruct;              // only for length information

    tsp00_Int4 lLengthOfBuffer = STATIC_CAST( tsp00_Int4, (sizeof(tls30_DatafileHeader) + strlen(pszCommand)) );
    lLengthOfBuffer = ALIGN_EO00(lLengthOfBuffer, 1024);    // align to full kB
    tsp00_Addr pszBuffer       = NULL;

    pszBuffer = new _TCHAR[lLengthOfBuffer];
    tsp00_Addr pszBuffer2 = pszBuffer;
    memset(pszBuffer, 0, lLengthOfBuffer);      // initialize

    // Define swap kind and log it
    tsp00_SwapKind_Enum Swap;
    Swap = ls07DefineSwap(NULL);

    ls07Swap4BIntToChar(lLengthOfBuffer, Swap, sw_normal, pszBuffer2);
    pszBuffer2 += HeaderStruct.dhHeaderLength_ls30.size();

    memcpy(pszBuffer2, HEADER_IDENTIFIER_LS00, sizeof(HEADER_IDENTIFIER_LS00));
    pszBuffer2 += HeaderStruct.dhIdentifier_ls30.size();

    ls07Swap4BIntToChar(lRecordLength, Swap, sw_normal, pszBuffer2);
    pszBuffer2 += HeaderStruct.dhRecordLength_ls30.size();

    memcpy(pszBuffer2, pszCommand, strlen(pszCommand));

    rc = ls30VFWriteLn(VFile, pszBuffer, lLengthOfBuffer, pszErrText);  // error is already logged by ls30VFWriteLn

    return rc;
}
// ls30VFWriteHeader()

/*
  -----------------------------------------------------------------------------
  function:     ls30VFReadHeader
  -----------------------------------------------------------------------------
*/
int
ls30VFReadHeader(tls00_VFile & VFile,
                 tsp00_Int4  & lRecordLength,
                 tsp00_Addr    pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls30VFReadHeader");

    tsp00_Int4 rc = errOK_els00;

    assert(VFile.vfDataType_ls00 == SP5VF_BINARY);  // opened in binary mode

    tsp00_Int4 lLengthOfBuffer = sizeof(tls30_DatafileHeader);
    tsp00_Addr pszBuffer       = NULL;

    pszBuffer = new _TCHAR[lLengthOfBuffer];
    memset(pszBuffer, 0, lLengthOfBuffer);      // initialize

    tsp00_Int4 ReadLen = STATIC_CAST( tsp00_Int4, ls30_ReadLnBin(pszBuffer, VFile, lLengthOfBuffer, pszErrText) );
    if ( ReadLen <= 0 )         // In case of error length corresponds to error number
    {
        if (EOF_LS00 == ReadLen)  // file is empty
        {
            rc = errNoDatafileHeader_ls98;
        }
        else
        {
            rc = ReadLen;
        }
    }
    else
    {
        tls30_DatafileHeader HeaderStruct;
        tsp00_Int4           lPosition = HeaderStruct.dhHeaderLength_ls30.size();
        if (0 != memcmp(&pszBuffer[lPosition], HEADER_IDENTIFIER_LS00, sizeof(HEADER_IDENTIFIER_LS00)))
        {
            // We need to seek back to the beginning of file if we do not find a header
            ls30VFSeek(VFile, 0, pszErrText);     // not expecting errors

            rc = errNoDatafileHeader_ls98;
        }
        else
        {
            tsp00_SwapKind_Enum Swap;
            Swap = ls07DefineSwap(NULL);

            // At this point we assume that the header may not exceed 2GB so that we use a simple long int
            tsp00_Longint lfilepos = ls07Swap4BHexToLong(pszBuffer, sw_normal, Swap);

            lPosition     = HeaderStruct.dhHeaderLength_ls30.size() + HeaderStruct.dhIdentifier_ls30.size();
            lRecordLength = ls07Swap4BHexToLong(&pszBuffer[lPosition], sw_normal, Swap);

            // After work is done we need to seek to the beginning of data; the
            // seek function updates the vfFilePos member in VFile, too.
            ls30VFSeek(VFile, lfilepos, pszErrText);     // not expecting errors
        }
    }

    if (NULL != pszBuffer)
    {
        delete [] pszBuffer;
        pszBuffer = NULL;
    }

    return rc;
}
//  ls30VFReadHeader()


/*
  -----------------------------------------------------------------------------
  function:     ls30VFWriteMetaHeader
  -----------------------------------------------------------------------------
*/
tsp00_Int4
ls30VFWriteMetaHeader(tls00_VFile       &VFile,
                      tls00_DextractRec *pDERec,
                      tls00_DatExtrCmd  *pDECmd,
                      tsp00_Addr         pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls30VFWriteMetaHeader");

    tsp00_Int4 rc = errOK_els00;

    assert(VFile.vfDataType_ls00 == SP5VF_BINARY);  // opened in binary mode?

    // Define version
    tsp00_Versionc pszVersion;
    sp100_GetVersionString(COMP_NAME_REPL_MAN_SP100, s100buildnumber, pszVersion);

    
    //version (contains some id, too)|swap|headerlengh|header|longpos|command
                                        //  full header length including preheader information -> full kB
    // Define header length without information describing the header
    size_t hlen = sizeof(tls30_FileMetaDataHeader)              +
                  pDERec->derLongColCnt_ls00*sizeof(tsp00_Int4) +
                  _tcslen(pDECmd->dexFullUserInput_ls00);
    
    // Define overall length of buffer to write
    size_t len = pszVersion.size() + sizeof(unsigned char) + sizeof(tsp00_Int4) + hlen;

    len = ALIGN_EO00(len, 1024);                    // align file header to full kB
    tsp00_Addr pszBuffer = new _TCHAR[len];
    memset(pszBuffer, 0, len);

    tsp00_Addr pszKeepBufPtr = pszBuffer;

    memcpy(pszBuffer, pszVersion.asCharp(), pszVersion.length());
    pszBuffer += pszVersion.size();

    *pszBuffer = STATIC_CAST(unsigned char, ls07DefineSwap(NULL));
    pszBuffer += sizeof(unsigned char);

    // We store the full header length in header
    memcpy(pszBuffer, &len, sizeof(tsp00_Int4));
    pszBuffer += sizeof(tsp00_Int4);

    
    // define meta information store
    tls30_FileMetaDataHeader *pMeta = REINTERPRET_CAST(tls30_FileMetaDataHeader*, pszBuffer);

    pMeta->mdRecordLength_ls30 = pDERec->derRecordLen_ls00;
    pMeta->mdFieldCount_ls30   = pDERec->derTableInfo_ls00.tdFieldCount_ls00;
    pMeta->mdLongCount_ls30    = pDERec->derLongColCnt_ls00;
    pMeta->mdMetaData_ls30     = pDECmd->dexDatFile_ls00.fsFormat_ls00;
    
    pszBuffer += sizeof(tls30_FileMetaDataHeader);      // move pointer behind tls30_FileMetaDataHeader to
                                                        // write additional information

    // Log position of longs in data record behind written header
    for (tsp00_Int4 i = 0; i < pDECmd->dexOutSpec_ls00.mcsCount_ls00; ++i)
    {
        if (1 == szMap_LongDataType_ls00[pDERec->pderColumnData_ls00[i].cdSQLDataType_ls00])
        {
            tsp00_Int4 pos = i+1;
            memcpy(pszBuffer, &pos, sizeof(tsp00_Int4));
            pszBuffer += sizeof(tsp00_Int4);
        }
    }

    // Write command to buffer
    memcpy(pszBuffer, pDECmd->dexFullUserInput_ls00, _tcslen(pDECmd->dexFullUserInput_ls00));


    // Write structure to file; updates file pos
    rc = ls30VFWriteLn(VFile, pszKeepBufPtr, len, pszErrText); // error is already logged by ls30VFWriteLn
    if (NULL != pszKeepBufPtr)
    {
        delete [] pszKeepBufPtr;
        pszKeepBufPtr = NULL;
    }

    return rc;
}
// ls30VFWriteMetaHeader()

/*
  -----------------------------------------------------------------------------
  function:     ls30VFReadMetaHeader
  -----------------------------------------------------------------------------
*/
tsp00_Int4
ls30VFReadMetaHeader(tls00_VFile               &VFile,
                     tls30_FileMetaDataHeader  &FileMetaDataHeader,
                     tsp00_Uint4              *&pLongPos,
                     tsp00_Addr               *&pszCmd,
                     tsp00_SwapKind_Enum      *pSSwap,
                     tsp00_Addr                 pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls30VFReadMetaHeader");

    tsp00_Int4 rc = errOK_els00;

    assert(VFile.vfDataType_ls00 == SP5VF_BINARY);  // opened in binary mode?

    // First read length of buffer from file
    tsp00_Versionc Version;
    tsp00_Longint rlen = 0;

    tsp00_Int4 buflen = Version.size() + sizeof(unsigned char) + sizeof(tsp00_Int4);

    tsp00_Addr pszBuffer = new _TCHAR[buflen];
    tsp00_Addr pszKeepBufPtr = pszBuffer;

    rlen = ls30_ReadLnBin(pszBuffer, VFile, buflen, pszErrText);
    if (rlen <= 0)         // In case of error length corresponds to error number
    {
        rc = STATIC_CAST(tsp00_Int4, rlen);
    }
    else
    {
        // For now skip version info
        pszBuffer += Version.size();

        tsp00_SwapKind_Enum DSwap = ls07DefineSwap(NULL);
        *pSSwap = STATIC_CAST(tsp00_SwapKind_Enum, *pszBuffer);
        pszBuffer += sizeof(unsigned char);

        // Define header length for buffer to read; we have to take into account the fact
        // that the header is aligned to full kB written -> so we need to read at least
        // the full kB
        tsp00_Int4 hlen = *(REINTERPRET_CAST(tsp00_Int4*, pszBuffer));
        hlen = ls07Swap4BInt(hlen, *pSSwap, DSwap);
        hlen -= buflen;             // header length contains the length of preheader info, too

        tsp00_Addr pszHeader = new _TCHAR[hlen];
        memset(pszHeader, 0, hlen);
        tsp00_Addr pszKeepHeaderPtr = pszHeader;

        rlen = ls30_ReadLnBin(pszHeader, VFile, hlen, pszErrText);
        if (rlen <= 0)                  
        {
            rc = STATIC_CAST(tsp00_Int4, rlen);
        }
        else
        {
            FileMetaDataHeader.mdRecordLength_ls30 = ls07Swap4BInt(*(REINTERPRET_CAST(tsp00_Uint4*, pszHeader)), *pSSwap, DSwap);
            pszHeader += sizeof(tsp00_Uint4);

            FileMetaDataHeader.mdFieldCount_ls30 = ls07Swap4BInt(*(REINTERPRET_CAST(tsp00_Uint4*, pszHeader)), *pSSwap, DSwap);
            pszHeader += sizeof(tsp00_Uint4);

            FileMetaDataHeader.mdLongCount_ls30 = ls07Swap4BInt(*(REINTERPRET_CAST(tsp00_Uint4*, pszHeader)), *pSSwap, DSwap);
            pszHeader += sizeof(tsp00_Uint4);

            FileMetaDataHeader.mdMetaData_ls30 = *(REINTERPRET_CAST(tls00_FileFormat*, pszHeader));
            pszHeader += sizeof(tls00_FileFormat);

            if (0 != FileMetaDataHeader.mdLongCount_ls30)
            {
                // Read long positions
                pLongPos = new tsp00_Uint4[FileMetaDataHeader.mdLongCount_ls30];
                tsp00_Uint4 i = 0;
                pLongPos[0] = ls07Swap4BInt(*(REINTERPRET_CAST(tsp00_Uint4*, pszHeader)), *pSSwap, DSwap);

                for (i = 1; i < FileMetaDataHeader.mdLongCount_ls30; ++i)
                {
                    pszHeader += sizeof(tsp00_Uint4);
                    pLongPos[i] = ls07Swap4BInt(*(REINTERPRET_CAST(tsp00_Uint4*, pszHeader)), *pSSwap, DSwap);
                }
            }

            // Not yet implemented to read command from header and to swap the structure tls00_FileFormat

            if (NULL != pszKeepHeaderPtr)
            {
                delete [] pszKeepHeaderPtr;
                pszKeepHeaderPtr = NULL;
                pszHeader        = NULL;
            }

            // At the end adapt the file position
            ls07AddLargeNumbers(VFile.vfFilePos_ls00, hlen + buflen);
        }

        if (pszKeepBufPtr != NULL)
        {
            delete pszKeepBufPtr;
            pszKeepBufPtr = NULL;
            pszBuffer     = NULL;
        }
    }

    return rc;
        
}
//  ls30VFReadMetaHeader()


/*
  -----------------------------------------------------------------------------
  function:     ls30VFSeek
  -----------------------------------------------------------------------------
*/
int ls30VFSeek(tls00_VFile &VFile, tsp00_Longint FilePos, tsp00_Addr ErrText)
{
    ROUTINE_DBG_MEO00 ("ls30VFSeek");

    tsp00_Int4          rc = errOK_els00;
    tsp5_rte_fileerror  RteError;


    if ( 0 != VFile.vfFileHndl_ls00 )
    {
        sqlfseekc(VFile.vfFileHndl_ls00, FilePos, SP5VF_SEEK_BEGIN, &RteError);
        if (vf_ok != RteError.sp5fe_result)
        {
            tsp00_C40c pszTmpLZUText;
            pszTmpLZUText.p2c(RteError.sp5fe_text);
            RepServMessages::Instance()->ls98Msg(layVFile_els98, ErrText,
                                                                 errFileRead_els98,
                                                                 VFile.vfFileName_ls00,
                                                                 pszTmpLZUText.asCharp());
            rc = errFileRead_els98;
        }
        else
        {
            // Adapt file pos in VFile structure
            ls07AddLargeNumbers(VFile.vfFilePos_ls00, FilePos);
        }
    }

    return rc;
}
// ls30VFSeek()


/*
  -----------------------------------------------------------------------------
  function:     ls30VFSeekLong (overloaded)
  -----------------------------------------------------------------------------
*/
tsp00_Int4
ls30VFSeekLong(tls00_VFile             &VFile,
               const tls00_LargeNumber *pFilePos,
               tsp00_Addr               pszErrText,
               tsp5_rte_seekkind        lOrigin)
{
    ROUTINE_DBG_MEO00 ("ls30VFSeekLong");

    tsp00_Int4          rc = errOK_els00;
    tsp5_rte_fileerror  RteError;
    RteError.sp5fe_result.becomes(vf_ok);           // initialize to no-error
    
    if (0 != VFile.vfFileHndl_ls00)
    {
        tls00_LargeNumber distance = *pFilePos;

        // Unterscheiden, ob vom anfang gesucht wird oder von der current pos
        if (SP5VF_SEEK_CUR == lOrigin)
        {
            distance = ls07SubtractLargeNumbers(*pFilePos, VFile.vfFilePos_ls00);
        }
        //else if (SP5VF_SEEK_BEGIN == lOrigin)   // adapting distance here is done by initialization

        // position to seek to is higher than the 2 GB limit in both directions - pos and neg
        if (0 != distance.lnu2GB_Blocks_ls00)
        {
            tsp00_Longint maxdistance = (distance.lnu2GB_Blocks_ls00 < 0) ? -MAX_INT4_SP00 : MAX_INT4_SP00;
            tsp00_Longint cnt         = distance.lnu2GB_Blocks_ls00;
            if (distance.lnu2GB_Blocks_ls00 < 0)
            {
                cnt *= -1;
            }

            sqlfseekc(VFile.vfFileHndl_ls00, maxdistance, lOrigin, &RteError);
            if ( (vf_ok == RteError.sp5fe_result) && (cnt > 1) )
            {
                for (tsp00_Longint i = 1; ( (i < cnt) && (vf_ok == RteError.sp5fe_result) ); ++i)
                {
                    sqlfseekc(VFile.vfFileHndl_ls00, maxdistance, SP5VF_SEEK_CUR, &RteError);
                }
            }

            // taking into account pos and neg movements in file
            if ( (0 != distance.lnuRemainder_ls00) && (vf_ok == RteError.sp5fe_result) )
            {
                sqlfseekc(VFile.vfFileHndl_ls00, distance.lnuRemainder_ls00, SP5VF_SEEK_CUR, &RteError);
            }
        }
        else    // position to seek to is lower than the 2GB limit
        {
            if (0 != distance.lnuRemainder_ls00)        // taking into account pos and neg movements in file
            {
                sqlfseekc(VFile.vfFileHndl_ls00, distance.lnuRemainder_ls00, lOrigin, &RteError);
            }
        }
    }

    // Error handling
    if (vf_ok != RteError.sp5fe_result)
    {
        tsp00_C40c pszTmpLZUText;
        pszTmpLZUText.p2c(RteError.sp5fe_text);
        RepServMessages::Instance()->ls98Msg(layVFile_els98, pszErrText, errFileRead_els98,
                                                                         VFile.vfFileName_ls00,
                                                                         pszTmpLZUText.asCharp());
        rc = errFileRead_els98;
    }
    else
    {
        VFile.vfFilePos_ls00.lnu2GB_Blocks_ls00 = pFilePos->lnu2GB_Blocks_ls00;
        VFile.vfFilePos_ls00.lnuRemainder_ls00  = pFilePos->lnuRemainder_ls00;
    }

    return rc;
}
// ls30VFSeekLong()


/*
  -----------------------------------------------------------------------------
  function:     ls30VFSeekLong (overloaded)
  -----------------------------------------------------------------------------
*/
tsp00_Int4
ls30VFSeekLong(tls00_VFile         &VFile,
               const tsp00_Longint  lDistance,
               tsp00_Addr           pszErrText,
               tsp5_rte_seekkind    lOrigin)
{
    ROUTINE_DBG_MEO00 ("ls30VFSeekLong II");

    tsp00_Int4          rc = errOK_els00;
    tsp5_rte_fileerror  RteError;

    if (0 != VFile.vfFileHndl_ls00)
    {
        sqlfseekc(VFile.vfFileHndl_ls00, lDistance, lOrigin, &RteError);

        // Error handling
        if (vf_ok != RteError.sp5fe_result)
        {
            tsp00_C40c pszTmpLZUText;
            pszTmpLZUText.p2c(RteError.sp5fe_text);
            RepServMessages::Instance()->ls98Msg(layVFile_els98, pszErrText, errFileRead_els98,
                                                                             VFile.vfFileName_ls00,
                                                                             pszTmpLZUText.asCharp());
            rc = errFileRead_els98;
        }
        else
        {
            // Adapt file position
            ls07AddLargeNumbers(VFile.vfFilePos_ls00, lDistance);
        }
    }

    return rc;
}
// ls30VFSeekLong()


/*
  -----------------------------------------------------------------------------
  function:     ls30VFTruncate
  -----------------------------------------------------------------------------
*/
tsp00_Int4
ls30VFTruncate(tls00_VFile &VFile, tsp00_Addr ErrText, tsp00_Longint lNewSize)
{
    ROUTINE_DBG_MEO00 ("ls30VFTruncate");

    tsp00_Int4          rc = errOK_els00;
    tsp5_rte_fileerror  RteError;

    if ( 0 != VFile.vfFileHndl_ls00 )
    {
        sqlftruncatec(VFile.vfFileHndl_ls00, lNewSize, &RteError);
        if (vf_ok != RteError.sp5fe_result)
        {
            tsp00_C40c pszTmpLZUText;
            pszTmpLZUText.p2c(RteError.sp5fe_text);
            RepServMessages::Instance()->ls98MsgToString(layVFile_els98, ErrText,
                                                                         errFileTruncate_els98,
                                                                         VFile.vfFileName_ls00,
                                                                         pszTmpLZUText.asCharp());
            rc = errFileTruncate_els98;
        }
        else
        {
            VFile.vfFilePos_ls00.lnu2GB_Blocks_ls00 = 0;
            VFile.vfFilePos_ls00.lnuRemainder_ls00  = lNewSize;
            VFile.vfFileLen_ls00.lnu2GB_Blocks_ls00 = 0;
            VFile.vfFileLen_ls00.lnuRemainder_ls00  = lNewSize;
        }
    }

    return rc;
}
// ls30VFTruncate()


/*
  -----------------------------------------------------------------------------
  function:     ls30VFGetFileSize
  -----------------------------------------------------------------------------
*/
tsp00_Int4
ls30VFGetFileSize(tsp00_Addr pszFileName, tls00_LargeNumber *plFileSize)
{
    ROUTINE_DBG_MEO00 ("ls30VFGetFileSize");

    tsp00_Int4 rc = errOK_els00;

    tsp5_rte_fileerror  RteError;
    tsp5_rte_fileinfo   RteInfo;

    sqlfinfoc(pszFileName, &RteInfo, &RteError);
    if (vf_ok == RteError.sp5fe_result)
    {
        if (RteInfo.sp5fi_exists)
        {
            plFileSize->lnu2GB_Blocks_ls00 = RteInfo.sp5fi_size_x;
            plFileSize->lnuRemainder_ls00  = RteInfo.sp5fi_size;            
        }
    }
    else
    {
        rc = errFileRead_els98;
    }

    return rc;
}
//  ls30VFGetFileSize()


/*
  -----------------------------------------------------------------------------
  function:     ls30VFMoveToStartPosition
  -----------------------------------------------------------------------------
*/
tsp00_Int4
ls30VFMoveToStartPosition(tls00_VFile              &VFile,
                          const tls00_LargeNumber  *plStartPos,
                          tsp00_Addr                pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls30VFMoveToStartPosition");

    tsp00_Int4    rc      = errOK_els00;
    tsp00_Longint ReadLen = 0;

    // In case of binary data file the lStartPos is the physical pos
    // In case of text file it is the line/record number, so we have to read until the right
    //  line number is reached
    if (SP5VF_BINARY == VFile.vfDataType_ls00)
    {
        rc = ls30VFSeekLong(VFile, plStartPos, pszErrText, SP5VF_SEEK_BEGIN);
    }
    else
    {
        tsp00_Longint i = 1;        // because working in c/c++
        _TCHAR        szLineBuffer[LINELENGTH_MXSP00];

        for (i; ( (i < plStartPos->lnuRemainder_ls00) && (rc >= 0) ); ++i)
        {
            ReadLen = ls30VFReadLnEx(VFile, szLineBuffer, pszErrText, 0);
        }
        // VFile.vfFilePos_ls00 is adapted with every call of ls30VFReadLnEx

        if (ReadLen < 0)        // adapt error code
        {
            rc = STATIC_CAST(tsp00_Int4, ReadLen);
        }
        VFile.vfFileLineNum_ls00 = 0;       // nothing read here!
    }

    return rc;
}
// ls30VFMoveToStartPosition()


/*
  -----------------------------------------------------------------------------
  function:     ls30VFClose
  -----------------------------------------------------------------------------
*/
int ls30VFClose(tls00_VFile &VFile, tsp00_Addr ErrText)
{
    ROUTINE_DBG_MEO00 ("ls30VFClose");

    tsp00_Int4          rc = errOK_els00;
    tsp5_rte_fileerror  RteError;

    if ( 0 != VFile.vfFileHndl_ls00 )
    {
        sqlfclosec(VFile.vfFileHndl_ls00, VFile.vfFileCOpt_ls00, &RteError);
        if (vf_ok != RteError.sp5fe_result)
        {
            tsp00_C40c pszTmpLZUText;
            pszTmpLZUText.p2c(RteError.sp5fe_text);
            RepServMessages::Instance()->ls98Msg(layVFile_els98, ErrText,
                                                                 errFileClose_els98,
                                                                 VFile.vfFileName_ls00,
                                                                 pszTmpLZUText.asCharp());
            rc = errFileClose_els98;
        }
        else
        {
            VFile.vfFileHndl_ls00 = 0;      // Reset file handle
        }
    }

    VFile.vfFilePos_ls00.lnuRemainder_ls00 = EOF_LS00;        // Set file position to eof

    return rc;
}
// ls30VFClose()


/*
  -----------------------------------------------------------------------------
  function:     ls30VFErase
  -----------------------------------------------------------------------------
*/
tsp00_Int4
ls30VFErase(tls00_VFile &VFile, tsp00_Addr pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls30VFClose");

    tsp00_Int4          rc = errOK_els00;
    tsp5_rte_fileerror  RteError;

    // Close file is still open
    ls30VFClose(VFile, pszErrText);

    // Erase the file
    sqlferasec(VFile.vfFileName_ls00, &RteError);
    if (vf_ok != RteError.sp5fe_result)
    {
        tsp00_C40c pszTmpLZUText;
        pszTmpLZUText.p2c(RteError.sp5fe_text);
        RepServMessages::Instance()->ls98Msg(layVFile_els98, pszErrText,
                                                             errFileErase_els98,
                                                             VFile.vfFileName_ls00,
                                                             pszTmpLZUText.asCharp());
        rc = errFileErase_els98;
    }

    return rc;
}

/*
  -----------------------------------------------------------------------------
  Local functions: implementation
  -----------------------------------------------------------------------------
*/

/*
  -----------------------------------------------------------------------------
  function:     ls30_ReadLnTxt
  -----------------------------------------------------------------------------
*/
static tsp00_Longint
ls30_ReadLnTxt(tsp00_Addr     pszBuffer,
               tls00_VFile   &VFile,
               tsp00_Longint  lBytes,
               tsp00_Addr     pszErrorText,
               tsp00_Longint  MaxReadLen)
{
    ROUTINE_DBG_MEO00 ("ls30_ReadLnTxt");

    RepServMessages *pMsgObj = RepServMessages::Instance();    // error reporting obj

    tsp5_rte_fileerror RteError;
    tsp00_Longint      ReadLen = 0;

    do
    {
        sqlfreadc(VFile.vfFileHndl_ls00, pszBuffer, MaxReadLen, &ReadLen, &RteError);
        if (vf_ok != RteError.sp5fe_result)     // error???
        {
            if (vf_eof == RteError.sp5fe_result)
            {
                VFile.vfFilePos_ls00.lnuRemainder_ls00 = ReadLen = EOF_LS00;
            }
            else
            {
                tsp00_C40c pszTmpLZUText;
                pszTmpLZUText.p2c(RteError.sp5fe_text);
                pMsgObj->ls98Msg(layVFile_els98, pszErrorText, errFileRead_els98,
                                                               VFile.vfFileName_ls00,
                                                               pszTmpLZUText.asCharp());
                ReadLen = errFileRead_els98;
            }
        }
        else
        {
            if (vf_ok != RteError.sp5fe_warning)    // warning  --> buffer not sufficient
            {
                pMsgObj->ls98Msg(layVFile_els98, pszErrorText, errFileRead_els98,
                                                               VFile.vfFileName_ls00,
                                                               "Buffer (32 kByte) too small to hold whole data row");
                ReadLen = errFileRead_els98;
            }
        }
    }
    while (0 == ReadLen);
    
    return ReadLen;
}
//  ls30_ReadLnTxt()


/*
  -----------------------------------------------------------------------------
  function:     ls30_ReadLnBin
  -----------------------------------------------------------------------------
*/
static tsp00_Longint
ls30_ReadLnBin(tsp00_Addr    &pszBuffer,
               tls00_VFile   &VFile,
               tsp00_Longint  lBytes,
               tsp00_Addr     pszErrorText)
{
    ROUTINE_DBG_MEO00 ("ls30_ReadLnBin");

    tsp5_rte_fileerror RteError;
    tsp00_Longint      ReadLen = 0;

    sqlfreadc(VFile.vfFileHndl_ls00, pszBuffer, lBytes, &ReadLen, &RteError);
    if (vf_ok != RteError.sp5fe_result)
    {
        if (vf_eof == RteError.sp5fe_result)
        {
            VFile.vfFilePos_ls00.lnuRemainder_ls00 = EOF_LS00;

            ReadLen = EOF_LS00;
        }
        else
        {
            tsp00_C40c pszTmpLZUText;
            pszTmpLZUText.p2c(RteError.sp5fe_text);
            RepServMessages::Instance()->ls98Msg(layVFile_els98, pszErrorText,
                                                                 errFileRead_els98,
                                                                 VFile.vfFileName_ls00,
                                                                 pszTmpLZUText.asCharp());
            ReadLen = errFileRead_els98;
        }
    }

    return ReadLen;
}
//  ls30_ReadLnBin()
