/*
  -----------------------------------------------------------------------------

  module: vls31_catalogextract.cpp

  -----------------------------------------------------------------------------

  responsible:  Hans-GeorgB and SteffenS
  special area: Replication Manager
  description:  Implementation of the CATALOGEXTRACT/CATALOGLOAD functionality
  version:      7.3.

  -----------------------------------------------------------------------------

  ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end
*/

#include "gls00.h"
#include "hls03.h"
#include "hls15.h"
#include "hls30.h"
#include "hls98msg.h"
#include "hls99.h"
//#include "hls31_oi_Result.h"
#include "hls31_odbc_Result.h"
#include "hls31_catalogextract.h"

static SAPDB_Int4
ls31_ExtrCatalogCmd_ODBC(tls00_CatExtrCmd *pCatExtrCmd,
                         tls00_DBInfo     *pDBInfo,
                         tsp00_Addr        pszErrText);


static SAPDB_Int4
ls31_LoadCatalogCmd_ODBC(tls00_CatLoadCmd *pCatLoadCmd,
                         tls00_DBInfo     *pDBInfo,
                         tsp00_Addr        pszErrText);


/* -----------------------------------------------------------------------------
   ls31ExecCatalogExtract
/-------------------------------------------------------------------------------*/
SAPDB_Int4
ls31ExecCatalogExtract(tls00_CatExtrCmd *pCatExtrCmd, tls00_DBInfo *pDBInfo, tsp00_Addr pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls31ExecCatalogExtract");

    SAPDB_Int4 rc = errOK_els00;

    // Database Unicode ( ODBC supports only ASCII )
    if ( SAPDB_TRUE == ls00IsDBUnicode( pDBInfo ) )
    {
        RepServMessages::Instance()->ls98Msg(layCatalogExtract_els98, pszErrText, errIncompatibleEncodings_els98);
        rc = errIncompatibleEncodings_els98;
    }
    else
    {
        // interpret <catalogextract statement>::=
        // catalogextract <part_spec> OUTFILE <external_file_spec>;
        // <part_spec> :: = all | user | table

        rc = ls31_ExtrCatalogCmd_ODBC(pCatExtrCmd, pDBInfo, pszErrText);
    }
    return rc;
}
// ls31ExecCatalogExtract()


/* -----------------------------------------------------------------------------
   ls31ExecCatalogLoad
/-------------------------------------------------------------------------------*/
SAPDB_Int4
ls31ExecCatalogLoad(tls00_CatLoadCmd *pCatLoadCmd, tls00_DBInfo *pDBInfo, tsp00_Addr pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls31ExecCatalogLoad");

    SAPDB_Int4 rc = errOK_els00;
    
    // Database Unicode ( ODBC supports only ASCII )
    if ( SAPDB_TRUE == ls00IsDBUnicode( pDBInfo ) )
    {
        RepServMessages::Instance()->ls98Msg(layCatalogLoad_els98, pszErrText, errIncompatibleEncodings_els98);
        rc = errIncompatibleEncodings_els98;
    }
    else
    {
        // interpret <catalogload statement>::=
        // catalogextract <part_spec> INFILE <external_file_spec>;
        // <part_spec> :: = all | user | table

        rc = ls31_LoadCatalogCmd_ODBC(pCatLoadCmd, pDBInfo, pszErrText);
    }
    return rc;
}
// ls31ExecCatalogLoad()


/* -----------------------------------------------------------------------------*
   ExtrCatalogCmd_ODBC
/-------------------------------------------------------------------------------*/
static SAPDB_Int4
ls31_ExtrCatalogCmd_ODBC(tls00_CatExtrCmd *pCatExtrCmd, tls00_DBInfo *pDBInfo, tsp00_Addr pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls31_ExtrCatalogCmd_ODBC");

    SAPDB_Int4  rc = errOK_els00;

    if (SAPDB_TRUE == pCatExtrCmd->cexPartSpec_ls00.psSelAll_ls00)
    {
        rc = ls15CheckUserMode(pDBInfo,
                               pDBInfo->dbiSession_ls00.siUserName_ls00.asCharp(),
                               dbuSysdba_els00,
                               pszErrText);
        if (errOK_els00 != rc)
        {
            if (dbRowNotFound_els00 == rc)
            {
                RepServMessages::Instance()->ls98Msg(layTableExtract_els98, pszErrText, errRestrictedStmt_els98);
            }
        }
    }


    if ( (errOK_els00 == rc) && (true == pDBInfo->dbiSession_ls00.siSqlStat_ls00) )
    {
        // close oi database connection
        ls15CommitUser (pDBInfo, pszErrText);

        // establish odbc database connection
        CatalogDefODBC  *pCatalogDefODBC = new CatalogDefODBC();

        pCatalogDefODBC->setPartSpec(pCatExtrCmd->cexPartSpec_ls00);

        if (SAPDB_TRUE == pCatExtrCmd->cexOutfile_ls00.fsFormat_ls00.ffoXml_ls00)
        {
            pCatalogDefODBC->setExtrMode( emExtractAsXML_els31 );
        }
        else
        {
            pCatalogDefODBC->setExtrMode( emExtractAsDDL_els31 );
        }

        if ( pCatalogDefODBC->Connect( pDBInfo ) )
        {
#ifdef MEMTEST
            pCatalogDefODBC->Disconnect();
            delete pCatalogDefODBC;
#endif
            // establish file connection
            tls00_FileInfo aFileInfo ;
            aFileInfo.fileSpec = pCatExtrCmd->cexOutfile_ls00;

            ls30VFInitialize(aFileInfo.vFile, pCatExtrCmd->cexOutfile_ls00.fsFileName_ls00, SP5VF_TEXT, SP5VF_WRITE);
            rc = ls30VFOpen (aFileInfo.vFile, pszErrText);
            if (errOK_els00 == rc)
            {
                pCatalogDefODBC->setFileConnection(aFileInfo);

                // execute commands
                if ( (SAPDB_FALSE == pCatExtrCmd->cexPartSpec_ls00.psSelAll_ls00) &&
                     (SAPDB_FALSE == pCatExtrCmd->cexPartSpec_ls00.psSelUser_ls00) )
                {
                    // catalogextract table outfile

                    if (0 != pCatExtrCmd->cexPartSpec_ls00.psTabName_ls00.length())
                    {
                        rc = pCatalogDefODBC->ExtrTableCmd(pCatExtrCmd->cexPartSpec_ls00);
                    }
                }
                else
                {
                    StudioTD_String  username;

                    username.ConvertFromASCII_Latin1(pDBInfo->dbiSession_ls00.siUserName_ls00.asCharp(),
                                                     pDBInfo->dbiSession_ls00.siUserName_ls00.asCharp() +
                                                      pDBInfo->dbiSession_ls00.siUserName_ls00.length());

                    if (SAPDB_TRUE == pCatExtrCmd->cexPartSpec_ls00.psSelAll_ls00)
                    {
                        // catalogextract all outfile
                        rc = pCatalogDefODBC->ExtrCatalogCmd(username);
                    }
                    else
                    {
                        // catalogextract user outfile
                        rc = pCatalogDefODBC->ExtrSchemaCmd(username);
                    }
                }

                // close file Connection
                SAPDB_Int4 rc2 = ls30VFClose(aFileInfo.vFile, pszErrText);
            
            }   // end if (errOK_els00 == (rc = ls30VFOpen (aFileInfo.vFile, pszErrText))

            // close odbc database Connection
            pCatalogDefODBC->Disconnect();
        
        }   // end if ( pCatalogDefODBC->Connect( pDBInfo ) )

        if (NULL != pCatalogDefODBC)
        {
            delete pCatalogDefODBC;
            pCatalogDefODBC = NULL;
        }

        // establish oi database connection
        ls03SQLConnect (pDBInfo, pszErrText);
        ls15ConnectUser(pDBInfo, pszErrText);
    }

    return rc;
}
// ls31_ExtrCatalogCmd_ODBC()

/* -----------------------------------------------------------------------------
   ls31ExecCatalogLoad
/-------------------------------------------------------------------------------*/
static SAPDB_Int4
ls31_LoadCatalogCmd_ODBC(tls00_CatLoadCmd *pCatLoadCmd, tls00_DBInfo *pDBInfo, tsp00_Addr pszErrText)
{
    ROUTINE_DBG_MEO00 ("ls31_LoadCatalogCmd_ODBC");

    SAPDB_Int4 rc = errOK_els00;

    if (SAPDB_TRUE == pCatLoadCmd->cloPartSpec_ls00.psSelAll_ls00)
    {
        rc = ls15CheckUserMode(pDBInfo,
                               pDBInfo->dbiSession_ls00.siUserName_ls00.asCharp(),
                               dbuSysdba_els00,
                               pszErrText);
        if (errOK_els00 != rc)
        {
            if (dbRowNotFound_els00 == rc)
            {
                RepServMessages::Instance()->ls98Msg(layTableExtract_els98, pszErrText, errRestrictedStmt_els98);
            }
        }
    }

    if ( (errOK_els00 == rc) && (SAPDB_TRUE == pDBInfo->dbiSession_ls00.siSqlStat_ls00) )
    {
        // close oi database connection
        ls15CommitUser(pDBInfo, pszErrText);

        // establish odbc database connection
        CatalogDefODBC *pCatalogDefODBC = new CatalogDefODBC();

        pCatalogDefODBC->setPartSpec(pCatLoadCmd->cloPartSpec_ls00);

        if (SAPDB_TRUE == pCatLoadCmd->cloInfile_ls00.fsFormat_ls00.ffoXml_ls00)
        {
            pCatalogDefODBC->setExtrMode( emLoadAsXML_els31 );
        }
        else
        {
            pCatalogDefODBC->setExtrMode( emLoadAsDDL_els31 );
        }

        if ( pCatalogDefODBC->Connect( pDBInfo ) )
        {
            // establish file connection
            tls00_FileInfo aFileInfo ;
            aFileInfo.fileSpec = pCatLoadCmd->cloInfile_ls00;

            ls30VFInitialize(aFileInfo.vFile, pCatLoadCmd->cloInfile_ls00.fsFileName_ls00, SP5VF_TEXT);
            rc = ls30VFOpen (aFileInfo.vFile, pszErrText);
            if ( errOK_els00 == rc )
            {
                pCatalogDefODBC->setFileConnection( aFileInfo );

                // execute commands
                if ( (SAPDB_FALSE == pCatLoadCmd->cloPartSpec_ls00.psSelAll_ls00) &&
                     (SAPDB_FALSE == pCatLoadCmd->cloPartSpec_ls00.psSelUser_ls00) )
                {
                    // catalogload table infile
                    rc = pCatalogDefODBC->LoadTableCmd( pCatLoadCmd->cloPartSpec_ls00 );
                }
                else
                {
                    StudioTD_String  username;
                    username.ConvertFromASCII_Latin1(pDBInfo->dbiSession_ls00.siUserName_ls00.asCharp(),
                                                     pDBInfo->dbiSession_ls00.siUserName_ls00.asCharp() +
                                                      pDBInfo->dbiSession_ls00.siUserName_ls00.length() );

                    if (SAPDB_TRUE == pCatLoadCmd->cloPartSpec_ls00.psSelAll_ls00)
                    {
                        // catalogload all infile
                        rc = pCatalogDefODBC->LoadCatalogCmd(username);
                    }
                    else
                    {
                        // catalogload user infile
                        rc = pCatalogDefODBC->LoadSchemaCmd(username);
                    }
                }

                if (errOK_els00 != rc)
                {
                    sapdb_Error *pError = pCatalogDefODBC->getError();
                    RepServMessages::Instance()->ls98MsgToString(layTableExtract_els98, pszErrText, 
                                                                                        STATIC_CAST(ErrorNumbers, rc),
                                                                                        pError->getErrorText().StrPtr());
                }
                // close file vonnection
                SAPDB_Int4 rc2 = ls30VFClose(aFileInfo.vFile, pszErrText);

            }   // end if (errOK_els00 == (rc = ls30VFOpen (aFileInfo.vFile, pszErrText))

            // close odbc database connection
            pCatalogDefODBC->Disconnect();

        }   // end if ( pCatalogDefODBC->Connect( pDBInfo ) )

        if (NULL != pCatalogDefODBC)
        {
            delete pCatalogDefODBC;
            pCatalogDefODBC = NULL;
        }

        // establish oi database connection
        ls03SQLConnect (pDBInfo, pszErrText); 
        ls15ConnectUser(pDBInfo, pszErrText);
    }

    return rc;
}
// ls31_LoadCatalogCmd_ODBC()

/*===========================================================================*
 *     END
 *===========================================================================*/
