/*! 
  -----------------------------------------------------------------------------
 
  module: vls40bload.cpp
 
  -----------------------------------------------------------------------------
 
  responsible:  SteffenS and d033893
 
  special area: Replication Manager
    
  description:  implementation of command line replication server access

  version:      7.3.

  -----------------------------------------------------------------------------
 
  copyright:    SAP AG, 1997-2002
 
  -----------------------------------------------------------------------------
*/

/*


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end


*/
/*!
  -----------------------------------------------------------------------------
  includes
  -----------------------------------------------------------------------------
*/

#include <stdio.h>
#include <stdlib.h>

#include "gsp00.h"

#include "gsp01.h"
#include "vsp001.h"
#include "gsp09.h"

#include "geo47.h"

#include "heo01.h"      // function to define db root directory
#include "heo02.h"
#include "heo03.h"
#include "hsp100.h"
#include "hcn14.h"
#include "hls40cli.h"


/*! 
  -----------------------------------------------------------------------------
  Specification
  -----------------------------------------------------------------------------
*/
externPascal int s02applencrypt( char * pw_clear, char * pw_crypt);     // modul vsp002


// Defines the length of options specifier (like -u <username>,<password>)
// computed as cSpecOptLength[<specifier char> - 'A']
static const char cSpecOptLength[66] = {1,                                      // index 0
                                  1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
                                  1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
                                  1,1,1,1,1,2,2,1,1,1,1,1,1,1,1,1,1,1,1,1,
                                  1,1,1,1,1};


/*!
  -----------------------------------------------------------------------------
 
  function:     ls40_Initialize

  -----------------------------------------------------------------------------

  description:  Initializes the structure with all information
                for execution.

  arguments:    opt         [IN]  - structure to initialize

  returnvalue:  No
  -----------------------------------------------------------------------------
*/
static void
ls40_Initialize(ls40_Options& opt);


/*!
  -----------------------------------------------------------------------------
 
  function:     ls40_Usage

  -----------------------------------------------------------------------------

  description:  Prints usage information.

  arguments:    No

  returnvalue:  No
  -----------------------------------------------------------------------------
*/
static void ls40_Usage();


/*!
  -----------------------------------------------------------------------------
 
  function:     ls40_ShowPrgVersion

  -----------------------------------------------------------------------------

  description:  Print version info of client program to specified outfile.

  arguments:    OutFile     [IN]   - file pointer of file to print to

  returnvalue:  No
  -----------------------------------------------------------------------------
*/
static void
ls40_ShowPrgVersion(FILE *OutFile);


/*!
  -----------------------------------------------------------------------------
 
  function:     ls40_ShowServerVersion

  -----------------------------------------------------------------------------

  description:  Print version info of replication server.
                To get the version information the function builds a command and
                sends it to the server. Then it 'analyzes' the reply.

  arguments:    options     [IN]    - structure with all information needed

  returnvalue:  PRG_OK              - success
                PRG_CMD_ERROR       - error getting version, error message is written
                PRG_SERVER_ERROR    - fatal server error getting version,
                                      error message is written
  -----------------------------------------------------------------------------
*/
static int
ls40_ShowServerVersion(ls40_Options & options);


/*!
  -----------------------------------------------------------------------------
 
  function:     ls40_ParseOptions

  -----------------------------------------------------------------------------

  description:  Parses command line.

  arguments:    options     [IN]        - structure to initialize
                argc        [IN]        - command line argument
                argv        [IN]        - command line argument

  returnvalue:  PRG_OK      - success
                PRG_FATAL   - either error in command line or only the help
                              information should be displayed
  -----------------------------------------------------------------------------
*/
static int
ls40_ParseOptions(ls40_Options *options, int argc, char *argv[]);


/*!
  -----------------------------------------------------------------------------
 
  function:     ls40_ServerConnect

  -----------------------------------------------------------------------------

  description:  Connects the client to the replication server.
                The server node to connect to is chosen in the following way:
                1. No server node given in command line -> the local host is used
                or
                2. command option indicates to use the database server node. If there
                   is no db server node given in the command line anyway the local host
                   is used for the replication server.
                or
                3. An own server node is given for the replication server which will
                   be used.

                In case there are more than one installations of db software of the
                same version on the server where to start the replication server the
                user has to give parameter DBROOT in the command line call which will
                be given through the connect to the C-server.

                Additionally the function fills in the max. packet length in the
                option structure.

  arguments:    *options        [IN]    - structure containing all needed information
                                          to connect to replication server

  returnvalue:  PRG_OK      - success
                PRG_FATAL   - could not connect to server; error message is given
  -----------------------------------------------------------------------------
*/
static int ls40_ServerConnect(ls40_Options *options);


/*!
  -----------------------------------------------------------------------------
 
  function:     ls40_UserConnect

  -----------------------------------------------------------------------------

  description:  Connect user to db over the replication manager. User name
                and db name are delivered as options to the program.
                In case either the user name or the db name options are not
                specified the function returns simply PRG_OK without doing anything.
                In this case the user must make sure to specify the user name
                and db name in the script file to the replication server.

  arguments:    opt     [IN]        - structure containing all needed information
                                       to logon user

  returnvalue:  PRG_OK              - success
                PRG_CMD_ERROR       - error executing the command, error message is written
                PRG_SERVER_ERROR    - fatal server error executing the command,
                                      error message is written
  -----------------------------------------------------------------------------
*/
static int ls40_UserConnect(ls40_Options &opt);


/*!
  -----------------------------------------------------------------------------
 
  function:     ls40_ServerCommand

  -----------------------------------------------------------------------------

  description:  Executes the command delivered in command buffer of options 
                structure. Checks the answer of server.

  arguments:    opt     [IN]        - structure containing all needed information
                                       to send the command

  returnvalue:  PRG_OK              - success
                PRG_CMD_ERROR       - error executing the command
                PRG_SERVER_ERROR    - fatal server error executing the command,
                                      error message is written
  -----------------------------------------------------------------------------
*/
static int ls40_ServerCommand(ls40_Options *options);


/*!
  -----------------------------------------------------------------------------
 
  function:     ls40_CheckAnswer

  -----------------------------------------------------------------------------

  description:  Checks the answer of the server after command execution. Removes
                the server answer header from the reply string. Extracts in case
                of error the error number from the reply string and returns it.

  arguments:    opt     [IN]    - structure containing all needed information

  returnvalue:  PRG_OK          - no error on execution of the command
                value > 0       - error executing the command, return value is 
                                  the error number returned by the server
                The reply string contains the reply from the server after returning.
  -----------------------------------------------------------------------------
*/
static int ls40_CheckAnswer(ls40_Options & options, tsp00_Int4 ReplyLen);


/*!
  -----------------------------------------------------------------------------
 
  function:     ls40_Batch

  -----------------------------------------------------------------------------

  description:  Reads the replication server command file, separates commands and
                sends them to the server.

                Reads 32k from command file in a buffer and works on buffer;
                leaves free space of at least ReplyBufferLen byte in packet for
                answer of server
                At the beginning allocates a buffer of size packet size. For every
                command the buffer is initialized.

  arguments:    opt     [IN]        - structure containing all needed information

  returnvalue:  PRG_OK              - success
                PRG_CMD_ERROR       - error executing the command, error message is written
                PRG_SERVER_ERROR    - fatal server error executing the command,
                                      error message is written
  -----------------------------------------------------------------------------
*/
static int ls40_Batch(ls40_Options &opt);


/*!
  -----------------------------------------------------------------------------
 
  function:     ls40_InitBuffers

  -----------------------------------------------------------------------------

  description:  Initializes used command and reply buffers in options structure.

  arguments:    opt     [IN]        - structure to initialize

  returnvalue:  No
  -----------------------------------------------------------------------------
*/
static void ls40_InitBuffers(ls40_Options & opt);


/*!
  -----------------------------------------------------------------------------
 
  function:     ls40_IsCmdSeparator

  -----------------------------------------------------------------------------

  description:  Checks if an null terminated buffer (read line from an command
                file) evaluates to an command separator.
                The only recognized separator is a double slash at the beginning
                of a line.

  arguments:    p       [IN]        - buffer

  returnvalue:  true  = comment
                false = no comment
  -----------------------------------------------------------------------------
*/
static bool ls40_IsCmdSeparator(_TCHAR *p);


/*
  -----------------------------------------------------------------------------
 
  function:     ls40_Substitute

  -----------------------------------------------------------------------------

  description:  Substitutes parameters of the form &<digit> if such parameter
                was defined at the command line when calling the client.

  arguments:    opt           [IN]      - structure holding all information
                pszBuffer     [IN|OUT]  - string where to substitute the
                                          parameters
                lBufferLength [IN|OUT]  - length of buffer, will be
                                          adapted if parameters are found

  returnvalue:  0  = successful substitution or nothing to substitute; the string
                     can be used as command to send to the server
                -1 = substititution creates a string that exceeds the possible
                     length of command strings
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls40_Substitute(ls40_Options &opt, _TCHAR *pszBuffer, tsp00_Longint & lBufferLength);


/****************************************************************************** 
  -----------------------------------------------------------------------------
  START OF IMPLEMENTATION
  -----------------------------------------------------------------------------
******************************************************************************/

/* 
  -----------------------------------------------------------------------------
  function: ls40_ShowPrgVersion
  -----------------------------------------------------------------------------
*/
static void ls40_ShowPrgVersion(FILE *OutFile)
{
    tsp100_VersionID1 Version;

    sp100_GetVersionID(VersionIDType1_esp100, s100buildnumberEx, &Version );

    fprintf(OutFile, "Version: %d.%d.%.2d.%.2d\n",
                     STATIC_CAST(tsp00_Int4, Version.MajorVersion_sp100),
                     STATIC_CAST(tsp00_Int4, Version.MinorVersion_sp100),
                     STATIC_CAST(tsp00_Int4, Version.CorrLevel_sp100),
                     STATIC_CAST(tsp00_Int4, Version.BuildNumberPrefix_sp100));
}
//  ls40_ShowPrgVersion()


/* 
  -----------------------------------------------------------------------------
  function: ls40_ShowServerVersion
  -----------------------------------------------------------------------------
*/
static int ls40_ShowServerVersion(ls40_Options & options)
{
    short rc        = PRG_OK;

    ls40_InitBuffers(options);

    options.CmdBufferLen = 8*sizeof(_TCHAR);
    memcpy( options.pszCmdBuffer, "version", 7);
    options.pszCmdBuffer[7] = '\0';

    rc = ls40_ServerCommand (&options);
    if (PRG_OK == rc)
    {
        fprintf(options.outfile, "Server version information:\n%s\n", options.pszReplyBuffer);
    }
    else if (PRG_CMD_ERROR == rc)
    {
        fprintf(options.outfile, "Error getting server version information: %s\n", options.pszReplyBuffer);
    }
    // else means PRG_SERVER_ERROR error msg is already written by ls40_ServerCommand

    return rc;
}


/*
  -----------------------------------------------------------------------------
  function:     ls40_Usage
  -----------------------------------------------------------------------------
*/
static void ls40_Usage()
{
    char szString[3096];
  
    sprintf(szString,   "usage: repmcli [<options>]\n"
                        " <options>:\n"
                        "  -u <username>,<passwd>           :: connect user\n"
// PTS 1107858
//                      "  -U <userkey>            :: instead of Username/Password\n"
//                      "                          :: combination\n"
                        "  -d <serverdb>                    :: database name\n"
                        "  -n <db node>                     :: database server node (default localhost)\n"
                        "  -r [<node>]                      :: replication manager server node (default localhost)\n"
                        "  -R <INSTROOT>                    :: INSTROOT directory\n"
                        "  -b <file name>                   :: command input file\n"
                        "  -p<number> <substitution string> :: string that substitutes place holder '&<number>'\n"
                        "                                      in command input file (up to 9 parameters possible)\n"
                        "  -o[w|a] <file name>              :: output protocol file (default: stdout)\n"
                        "  -E <number>                      :: defines the number of errors after that the client\n"
                        "                                      stops executing (1 = default and 0 = ignore errors)\n"
                        "  -V                               :: show version of REPMServer\n"
                        "  -v                               :: show version of REPMServer client\n"
                        "  -h (or none)                     :: help-information\n"
                        "\n"
                        "Examples:  repmcli -u MyName,MyPassword -d MyDB -b MyFile\n"
// PTS 1107858          "           repmcli -U MyKey -n <db server node> -r <repm server node> -d MyDB -b MyFile\n");
                        "           repmcli -u MyName,MyPassword -n <db server node> -r <repm server node> -d MyDB -b MyFile\n");
    printf(szString);

    return;
}
//  ls40_Usage()



/*
  -----------------------------------------------------------------------------
   function:     ls40_Initialize
  -----------------------------------------------------------------------------
*/
static void ls40_Initialize(ls40_Options& opt)
{
    opt.dbuser             = NULL;
    opt.dbname             = NULL;
    
    opt.szDBRoot_ls40[0]   = '\0';

    opt.db_node            = "";

    opt.NodeLoc            = rncLocal_els40;
    opt.repm_node          = NULL;
    
    opt.serverpgm          = DEFAULT_PGM_L;
    opt.infile             = NULL;
    opt.outfile            = stdout;
    opt.showServerVersion  = false;
    opt.session            = NULL;

    opt.nMaxDataSize       = 0;
    opt.pszCmdBuffer       = NULL;
    opt.CmdBufferLen       = 0;

    opt.pszReplyBuffer     = NULL;
    opt.ReplyBufferLen     = RPM_REPLY_BUFFER_LEN_CN14;

    opt.CmdExecError       = 0;
    opt.BatchErrorCnt_ls40 = 1;     // default: stop after 1 error

    opt.bNoParam_ls40      = true;  // default: no parameters to consider

    tsp00_Int2 i = 0;
    
    for (i; i < MAX_PARAM_LS40; ++i)
    {
        opt.pszParamList_ls40[i] = NULL;
    }
}
// ls40_Initialize()


/*
  -----------------------------------------------------------------------------
   function:     ls40_InitBuffers
  -----------------------------------------------------------------------------
*/
static void ls40_InitBuffers(ls40_Options & opt)
{
    opt.pszCmdBuffer[0] = '\0';
    opt.CmdBufferLen    = 0;

    memset(opt.pszReplyBuffer, '\0', opt.ReplyBufferLen);
}


/* 
  -----------------------------------------------------------------------------
  function: ls40_ParseOptions
  -----------------------------------------------------------------------------
*/
static int ls40_ParseOptions(ls40_Options *options, int argc, char *argv[])
{
    short rc        = PRG_OK;

    bool  inOptions = true;
    char  cOptChar  = '\0';
    int   argI      = 1;

    if ( (1 == argc) ||
         ( (2 == argc) && ( ('h' == argv[1][1]) || ('H' == argv[1][1]) ) ) )
    {
        ls40_Usage();
        inOptions = false;
    }
    else
    {
        while ( (true == inOptions) && (argI < argc) )
        {
            if ('-' != argv[argI][0])                   // first char MUST BE '-'
            {
                ls40_Usage();
                inOptions = false;
            } 
            else
            {
                cOptChar = argv[argI][1];

                if (STATIC_CAST(size_t, cSpecOptLength[cOptChar - 'A']) < strlen(&argv[argI][1]) )
                {
                    ls40_Usage();
                    inOptions = false;
                }
                else
                {
                    switch (cOptChar)
                    {
                        case 'u':                       // User name and Password
                        {
                            ++argI;
                            options->dbuser = argv[argI];

                            size_t l = strlen(options->dbuser);
                            for (char *p = options->dbuser; p < options->dbuser + l; *p = _totupper(*p), ++p)
                                ;

                            break;
                        }
                        // Username and Password from xUser NOT IMPLEMENTED YET !!!!!!!!!!!!!!!!!!
                        case 'd':                       // DB name
                        {
                            ++argI;
                            options->dbname = argv [argI];

                            size_t l = strlen(options->dbname);
                            for (char *p = options->dbname; p < options->dbname + l; *p = _totupper(*p), ++p)
                                ;
                            
                            break;
                        }
                        case 'n':                       // Database node name
                        {
                            ++argI;
                            options->db_node = argv [argI];
                            break;
                        }
                        case 'r':                       // Replication server node name
                        {
                            ++argI;
                            if ( (argI >= argc) || ('-' == argv[argI][0]) || ('/' == argv[argI][0]) )
                            {
                                --argI;
                                options->NodeLoc = rncDBNode_els40;
                            }
                            else
                            {
                                options->repm_node = argv [argI];
                                options->NodeLoc = rncDifferent_els40;
                            }
                            break;
                        }
                        case 'R':                       // name of dbroot
                        {
                            ++argI;
                            if ( (argI >= argc) || ('-' == argv[argI][0]) )
                            {
                                --argI;     // only option provided, no name of directory found
                            }
                            else
                            {
                                // Is the given path name longer than var generate error
                                size_t lDBRootLen = strlen(argv [argI]);
                                if (lDBRootLen <= (sizeof(tsp00_Pathc)-1) )
                                {
                                    memcpy(options->szDBRoot_ls40, argv [argI], lDBRootLen);
                                    options->szDBRoot_ls40[lDBRootLen] = '\0';
                                }
                                else
                                {
                                    printf("Path name of dbroot path too long.\n\n");
                                    inOptions = false;
                                }
                            }
                            break;
                        }
                        case 'b':                       // command file
                        {
                            ++argI;

                            // Open file binary for later operations like fseek
                            options->pszInfileName = argv[argI];

                            options->infile = fopen(options->pszInfileName, "rb");
                            if (NULL == options->infile)
                            {
                                printf("Cannot open input file %s", options->pszInfileName);
                                inOptions = false;
                            }
                            break;
                        }
                        case 'o':                       // log file
                        {
                            char pszSwitch[2] = "w";                    // default is write

                            if ( strlen(argv[argI]) > 2 )               // open mode given by user
                            {
                                if (!(argv[argI][2] == 'a' || argv[argI][2] == 'w'))
                                {
                                    printf("Wrong parameter for access type of output file\n", argv[argI][2]);
                                    ls40_Usage();
                                    inOptions = false;
                                }
                                else
                                {
                                    pszSwitch[0] = argv[argI][2];
                                }               
                            }

                            if (true == inOptions)                  // no error in parameter specification
                            {
                                ++argI;
                        
                                options->pszOutfileName = argv[argI];

                                options->outfile = fopen(options->pszOutfileName, pszSwitch);
                                if (NULL == options->outfile) 
                                {
                                    printf("Cannot open output file %s", options->pszOutfileName);
                                    inOptions = false;
                                }
                            }
                            break;
                        }
                        case 'E':                       // error number
                        {
                            ++argI;

                            inOptions = false;          // initialize to error
                            bool bInvalid   = false;

                            char   *pszNumber = argv[argI];
                            size_t  lNumberLen = strlen(pszNumber);
                    
                            if (lNumberLen < 10)
                            {
                                if ( (1 == lNumberLen) && ('0' == pszNumber[0]) )
                                {
                                    options->BatchErrorCnt_ls40 = 0;
                                    inOptions = true;
                                }
                                else
                                {
                                    options->BatchErrorCnt_ls40 = atoi(pszNumber);
                                    if (0 == options->BatchErrorCnt_ls40)
                                    {
                                        bInvalid = true;
                                    }
                                    else
                                    {
                                        inOptions = true;
                                    }
                                }
                            }
                            else if (lNumberLen == 10)
                            {
                                if (0 < memcmp("2147483647", pszNumber, 10))
                                {
                                    options->BatchErrorCnt_ls40 = atoi(pszNumber);
                                    if (0 == options->BatchErrorCnt_ls40)
                                    {
                                        bInvalid = true;
                                    }
                                    else
                                    {
                                        inOptions = true;
                                    }
                                }
                            }
                            // Is the given value longer than 10 bytes or GT or EQ 2147483647? --> error
                            if (false == inOptions)
                            {
                                if (true == bInvalid)
                                {
                                    printf("The value '%s' for the number of errors is not a valid number.\n\n",
                                           pszNumber);
                                }
                                else
                                {
                                    printf("Error numbers may not exceed 2147483646\n\n");
                                }
                                ls40_Usage();
                            }
                            break;
                        }
                        case 'v':                       // client version
                        {
                            ls40_ShowPrgVersion(options->outfile);
                            inOptions = false;
                            break;
                        }
                        case 'V':                       // server version
                        {
                            options->showServerVersion = true;
                            break;
                        }
                        case 'p':
                        {
                            tsp00_Int4 Index = atoi(&argv[argI][2]);
                            if ( (0 == Index) || (Index > MAX_PARAM_LS40) )
                            {
                                printf("Wrong parameter specification\n");
                                inOptions = false;
                            }
                            else
                            {
                                options->bNoParam_ls40 = false;

                                Index -= 1;

                                ++argI;

                                size_t lStrLen = strlen(argv[argI]);
                                if (NULL != options->pszParamList_ls40[Index])
                                {
                                    delete [] options->pszParamList_ls40[Index];
                                }
                                options->pszParamList_ls40[Index] = new _TCHAR[lStrLen + 1];
                                memcpy(options->pszParamList_ls40[Index], argv[argI], lStrLen);
                                options->pszParamList_ls40[Index][lStrLen] = '\0';
                            }
                            break;
                        }
                        default:                                            // unknown option
                        {
                            printf("Unknown option \"%c\"\n", cOptChar);
                            inOptions = false;
                        }
                    }       // end switch (cOptChar)
                }
                ++argI;
            }       // if (argv[argI][0] != '-')
        }       // while ( (true == inOptions) && (argI < argc) )
    }       // if ( (1 == argc) || ...


    if (false == inOptions)
    {
        rc = PRG_FATAL;
    }

    return rc;
}
//  ls40_ParseOptions()


/* 
  -----------------------------------------------------------------------------
  function: ls40_ServerConnect
  -----------------------------------------------------------------------------
*/
static int ls40_ServerConnect(ls40_Options *options)
{
    tsp00_ErrTextc   errtext;
    tsp00_Int4       rc = PRG_OK;
    tsp00_NodeIdc    szNode;
    tsp00_DbNamec    szDbName;
    tsp00_VFilenamec szDbRoot;

    switch (options->NodeLoc)       // repm_node-option empty ? --> log on at local host
    {
        case rncLocal_els40:
        {
            options->repm_node = "";                        // use local machine
            break;
        }
        case rncDBNode_els40:
        {
            if ('\0' != options->db_node[0])
            {
                options->repm_node = options->db_node;      // use node running the db for repm server
            }
            else
            {
                options->repm_node = "";                    // use local machine
            }
            break;
        }
        default:
        {
            break;  // Default is rncDifferent_els40 -> repm_node has already right value
        }
    }
    

    char pszNodeString[1024] = "local host";
    if ('\0' != options->repm_node[0])
    {
        strcpy(&pszNodeString[0], options->repm_node);
    }

    szDbName.Init();
    szNode.Init();
    szDbRoot.Init();

    if ('\0' != options->repm_node[0])
    {
        szNode.rawAssign(options->repm_node);
    }

    //*
    //*     Adjust calling parameter at connect to server
    //*
    //* - user may have specified: repm_node, db_node, dbname, dbroot
    //* - name of replication server host is ALWAYS specified; empty string means <local host>
    //*
    //* 1. if db_node != repm_node
    //*     ---> dbname (ALWAYS)= leer, dbroot = [<name>]
    //*
    //* 2. if db_node == repm_node
    //*     - dbname given, no dbroot given     ---> dbname = <name>, dbroot = leer
    //*     - dbname given, dbroot given        ---> dbname = leer,   dbroot = <name>
    //*     - no dbname given, dbroot given     ---> dbname = leer,   dbroot = <name>
    //*     - no dbname given, no dbroot given  ---> dbname = leer,   dbroot = leer

    if (0 != _tcsicmp(options->repm_node, options->db_node) )
    {
        if ('\0' != options->szDBRoot_ls40[0])
        {
            szDbRoot.rawAssign(options->szDBRoot_ls40.asCharp());
        }
    }
    else
    {
        if ('\0' != options->szDBRoot_ls40[0])
        {
            szDbRoot.rawAssign(options->szDBRoot_ls40.asCharp());
        }
        else
        {
            if (NULL != options->dbname)
            {
                szDbName.rawAssign(options->dbname);
            }
        }
    }

    char pszReply[2*PATH_MXSP00];
    rc = cn14connectRPM(szNode, szDbName, szDbRoot, pszReply, &(options->session), errtext);
    if (PRG_OK == rc)
    {
        fprintf(options->outfile,
                "Opened connection to REPLICATION SERVER at node %s.\n", pszNodeString);
        if (NULL != pszReply)
        {
            fprintf(options->outfile, "%s\n", pszReply);
        }
    } 
    else 
    {
        fprintf(options->outfile,
                "Connection to REPLICATION SERVER at node %s failed: %s\n",
                pszNodeString, STATIC_CAST(char*,errtext) );

        rc = PRG_FATAL;
    }

    fflush(options->outfile);


    //*
    //*     Get maximum packet length and build command buffer
    //*
    if (PRG_OK == rc)
    {
        tsp00_Int4 nDummy1, nDummy2;
        cn14packetstats(options->session, &(options->nMaxDataSize), &nDummy1, &nDummy2);

        options->pszCmdBuffer = new _TCHAR[MAX_COMMAND_LENGTH_LS40];
        memset(options->pszCmdBuffer, '\0', options->nMaxDataSize);

        options->pszReplyBuffer = new _TCHAR[options->ReplyBufferLen];
        memset(options->pszReplyBuffer, '\0', options->ReplyBufferLen);
    }

    return rc;
}
//  ls40_ServerConnect()




/*
  -----------------------------------------------------------------------------
  function:     ls40_UserConnect
  -----------------------------------------------------------------------------
*/
static int ls40_UserConnect(ls40_Options &opt)
{
    int         rc = PRG_OK;

    const char *pszPassword;
    size_t      nUserNameLength;
    
    if (NULL == opt.dbname)      // db name given ?
    {
        if (false == opt.showServerVersion)
        {
            fprintf(opt.outfile, "No database name given to connect to.\n");
        }
        return rc;
    }
    

    ls40_InitBuffers(opt);

    if (NULL == opt.dbuser)
    {
        // connects only the replication server to db but does not logon user
        // (password and user name missing)
        sprintf(opt.pszCmdBuffer, "USE SERVERDB %s", opt.dbname);
        if ('\0' != opt.db_node[0])
        {
            strcat(opt.pszCmdBuffer, " ON ");
            strcat(opt.pszCmdBuffer, opt.db_node);
        }
    }
    else
    {
        // Separate user name from password
        pszPassword = strchr (opt.dbuser, ',');
        if (NULL != pszPassword)
        {
            nUserNameLength = pszPassword - opt.dbuser;
            pszPassword++;
        }
        else
        {
            nUserNameLength = strlen(opt.dbuser);
            pszPassword = "";
        }

        // Mask DB name with double quotes to make special identifiers (like
        // hyphen -) possible in DB names

        sprintf(opt.pszCmdBuffer, "USE USER \"%.*s\" %s SERVERDB \"%s\"",
                                  nUserNameLength, opt.dbuser, pszPassword, opt.dbname);
        if ('\0' != opt.db_node[0])
        {
            strcat(opt.pszCmdBuffer, " ON ");
            strcat(opt.pszCmdBuffer, opt.db_node);
        }
    }

    char pszNode[1024] = "local host";
    if ('\0' != opt.db_node[0])
    {
        strcpy(&pszNode[0], opt.db_node);
    }


    opt.CmdBufferLen = _tcslen(opt.pszCmdBuffer);

    rc = ls40_ServerCommand(&opt);
    if (PRG_OK == rc)
    {
        if (NULL == opt.dbuser)
        {
            fprintf(opt.outfile, "Connection to database %s on %s established.\n",
                                 opt.dbname, pszNode);
        }
        else
        {
            fprintf(opt.outfile, "User %.*s connected to database %s on %s.\n",
                                 nUserNameLength, opt.dbuser, opt.dbname, pszNode);
        }
    }
    else if (PRG_CMD_ERROR == rc)
    {
        if (NULL == opt.dbuser)
        {
            fprintf(opt.outfile, "Error establishing connection to database %s on %s: %s\n",
                                 opt.dbname, pszNode, opt.pszReplyBuffer);
        }
        else
        {
            fprintf(opt.outfile, "Error connecting user %.*s to database %s on %s: %s\n",
                                 nUserNameLength, opt.dbuser, opt.dbname, pszNode, opt.pszReplyBuffer);
        }
    }
    // else means PRG_SERVER_ERROR error msg is already written by ls40_ServerCommand
    
    fflush(opt.outfile);

    return rc;
}
//  ls40_UserConnect()


/*
  -----------------------------------------------------------------------------
  function:     ls40_IsCmdSeparator
  -----------------------------------------------------------------------------
*/
static bool ls40_IsCmdSeparator(_TCHAR *p)
{
    bool rc             = false;

    size_t len = strlen(p);
    size_t i   = 0;

    for (i = 0; i < len; ++i)
    {
        if ( (0 == isspace(p[i])) && (0 == iscntrl(p[i])) )
        {
            if ('/' == p[i])
            {
                if ('/' == p[i+1])
                {
                    rc = true;
                }
            }
            break;
        }
    }

    return rc;
}
//  ls40_IsCmdSeparator()



/*
  -----------------------------------------------------------------------------
  function:     ls40_Substitute
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls40_Substitute(ls40_Options &opt, _TCHAR *pszBuffer, tsp00_Longint & lBufferLength)
{
    tsp00_Int4 rc          = 0;

    _TCHAR        *pszBufStart = pszBuffer;
    tsp00_Longint  lMyBufLen   = lBufferLength;

    
    tsp00_Int4    lIndex    = 0;
    tsp00_Longint lRestLen  = 0;
    tsp00_Longint lParamLen = 0;

    _TCHAR *pszFound  = NULL;
    bool    bNotFound = true;

    while (NULL != (pszFound = strchr(pszBuffer, '&') ) )
    {
        bNotFound = true;

        if ( 0 != isdigit(pszFound[1]) )            // We treat only parameters with one digit
        {
            _TCHAR pszTemp[2] = {'\0'};
            pszTemp[0] = pszFound[1];
            if (0 != (lIndex = atoi(pszTemp) ) )    // the char following & must be a digit
            {
                lIndex -= 1;                        // lIndex moves by -1, because we start at 0

                if (NULL != opt.pszParamList_ls40[lIndex])  // No parameter defined -> jump over it
                {
                    bNotFound = false;

                    lParamLen = _tcslen(opt.pszParamList_ls40[lIndex]);
                    
                    if ( (lMyBufLen + lParamLen - 3) > opt.nMaxDataSize)
                    {
                        rc = -1;                     // entire length of string to generate may not exceed
                        break;                       // allowed length (considering terminating 0, too)
                    }

                    lRestLen = lMyBufLen - ((pszFound - pszBufStart) + 2);

                    memmove(&pszFound[lParamLen], pszFound + 2, lRestLen);
                    memcpy(pszFound, opt.pszParamList_ls40[lIndex], lParamLen);
                    
                    pszBuffer = pszFound + lParamLen;       // move buf ptr to new position behind inserted
                                                            // parameter string
                    lMyBufLen += (lParamLen - 2);           // adapt length
                    
                    pszBufStart[lMyBufLen] = '\0';          // memmove destroys '\0'
                }
            }
        }

        if (true == bNotFound)
        {
            pszBuffer = &pszFound[1];
        }
    }

    if (lMyBufLen != lBufferLength)
    {
        lBufferLength = lMyBufLen;
    }

    return rc;
}
// ls40_Substitute()


/*
  -----------------------------------------------------------------------------
  function:     ls40_Batch
  -----------------------------------------------------------------------------
*/
static int
ls40_Batch(ls40_Options &opt)
{
    int rc = PRG_OK;
    
    _TCHAR        *pszBuffer  = new _TCHAR[MAXBUFFERSIZE_LS40];
    tsp00_Longint  lBufferLen = 0;

    int            lErrNum    = 0;
    char          *dummy      = NULL;
    bool           cmdfound   = false;

    //*
    //*     Read command file until eof is reached and process found commands
    //*

    ls40_InitBuffers(opt);      // Initialize buffer
    while ( (PRG_OK == rc) && ( 0 == feof(opt.infile) ) )
    {
        opt.CmdBufferLen = 0;

        //*
        //*     Find rows belonging to command until we find a comment or the end of file is
        //*     reached or the command becomes too long and must be sent in parts
        //*
        while ( (PRG_OK == rc) && (NULL   != (dummy = fgets(pszBuffer, MAXBUFFERSIZE_LS40, opt.infile) ) ) )
        {
            lBufferLen = strlen(pszBuffer);
            if ( true == ls40_IsCmdSeparator(pszBuffer) )
            {
                if (true == cmdfound)
                {
                    break;
                }
            }
            else
            {
                opt.CmdBufferLen += lBufferLen;
                if (opt.CmdBufferLen > MAX_COMMAND_LENGTH_LS40)
                {
                    //*
                    //*     Error: command becomes too long
                    //* Log the command to file and exit
                    //*
                    if ('\n' == opt.pszCmdBuffer[opt.CmdBufferLen - lBufferLen - 1])
                    {
                        fprintf(opt.outfile, "%s", opt.pszCmdBuffer);
                    }
                    else
                    {
                        fprintf(opt.outfile, "%s\n", opt.pszCmdBuffer);
                    }
                    fprintf(opt.outfile, "Error processing command; command is too long\n"
                                         "Maximum command length: %d\n", MAX_COMMAND_LENGTH_LS40); 
                    rc = PRG_FATAL;
                }
                else
                {
                    //*
                    //*     Copy command extract to command buffer
                    //*
                    strcat(opt.pszCmdBuffer, pszBuffer);
                    if (false == cmdfound)
                    {
                        cmdfound = true;
                    }
                }
            }

        }   // end while ( (PRG_OK == rc) && ...

        //*
        //*     Check for reading error
        //*
        if ( (PRG_OK == rc) && (NULL == dummy) )
        {
            int nError = ferror(opt.infile);
            if (0 != nError)
            {
                fprintf(opt.outfile, "Error reading from command file: %d", nError);
                rc = PRG_FATAL;
            }
        }

        //*
        //*     Process found command
        //*
        if ( (PRG_OK == rc) && (true == cmdfound) )
        {
            //*
            //*     Substitute parameters and log command to outfile
            //*
            if (false == opt.bNoParam_ls40)
            {
                ls40_Substitute(opt, opt.pszCmdBuffer, opt.CmdBufferLen);
            }
            if ('\n' == opt.pszCmdBuffer[opt.CmdBufferLen - 1])
            {
                fprintf(opt.outfile, "%s", opt.pszCmdBuffer);   // Log command to file
            }
            else
            {
                fprintf(opt.outfile, "%s\n", opt.pszCmdBuffer);   // Log command to file
            }


            //*
            //*     Send command to server and wait for reply
            //*
            rc = ls40_ServerCommand(&opt);
            if (PRG_OK == rc)
            {
                if ('\0' != opt.pszReplyBuffer[0])
                {
                    fprintf(opt.outfile, "%s\n", opt.pszReplyBuffer);
                }
                fprintf(opt.outfile, "Successfully executed\n\n");
            }
            else if (PRG_CMD_ERROR == rc)
            {
                fprintf(opt.outfile, "Error during execution\n-->%s\n\n", opt.pszReplyBuffer);
            
                ++lErrNum;
                if ( (0 != opt.BatchErrorCnt_ls40) && (lErrNum == opt.BatchErrorCnt_ls40) )
                {
                    fprintf(opt.outfile, "Defined maximum number of errors (%d) reached\n",
                                         opt.BatchErrorCnt_ls40);
                }
                else
                {
                    rc = PRG_OK;
                }
            }
        }   // end if ( (PRG_OK == rc) && (true == cmdfound) )

        cmdfound = false;           // reinitialize
        ls40_InitBuffers(opt);      // reinitialize buffer

        fflush(opt.outfile);
    }   // end while ( (PRG_OK == rc) && ( 0 == feof(opt.infile) ) )


    if (NULL != pszBuffer)
    {
        delete [] pszBuffer;
        pszBuffer = NULL;
    }

    // In case of error while processing commands print summary to stderr 
    if (0 != lErrNum)
    {
        fprintf(stderr, "\n\n%d error(s) during execution of command file '%s'.\n"
                        "---> See log file(s) for more information.\n",
                        lErrNum, opt.pszInfileName);
    }
    fflush(opt.outfile);

    return rc;
}
//  ls40_Batch()


/*
  -----------------------------------------------------------------------------
  function: ls40_CheckAnswer
  -----------------------------------------------------------------------------
*/
static int ls40_CheckAnswer(ls40_Options & options, tsp00_Int4 ReplyLen)
{
    short   i               = 0;
    int     nErrorNumber    = PRG_OK;


    if (0 == strncmp(options.pszReplyBuffer, REPMAN_ERR_STR, strlen(REPMAN_ERR_STR)))       // error
    {
        // Skip indicator "err" and read error number. The error is given in
        // the form "E<error number>:" . The first and the last char need to be removed.
        nErrorNumber = atol(&options.pszReplyBuffer[REPLY_HEADER_SIZE_LS40 + 1]);

        options.CmdExecError = (nErrorNumber == 0) ? PRG_FATAL : nErrorNumber;
        nErrorNumber = PRG_CMD_ERROR;
    }
    
    // Remove answer header in any case
    tsp00_Int4 len = ReplyLen - REPLY_HEADER_SIZE_LS40;

    memmove(options.pszReplyBuffer, &options.pszReplyBuffer[REPLY_HEADER_SIZE_LS40], len);
    options.pszReplyBuffer[len] = '\0';
    
    return nErrorNumber;
}
//  ls40_CheckAnswer()


/* 
  -----------------------------------------------------------------------------
  function: ls40_ServerCommand
  -----------------------------------------------------------------------------
*/
static int ls40_ServerCommand(ls40_Options  *options)
{
    tsp00_Int4      rc        = PRG_OK;
    
    tsp00_ErrTextc  errtext;
    tsp00_Int4      commErr   = CtrlCmdOk_csp09;
    tsp00_Int4      nReplyLen = options->ReplyBufferLen;

    commErr = cn14ExecuteLoaderCmd(options->session,
                                   options->pszCmdBuffer,
                                   STATIC_CAST(tsp00_Int4, options->CmdBufferLen),     // IA64
                                   options->pszReplyBuffer,
                                   &nReplyLen,
                                   errtext);
    // nReplyLen has new value (after returning) which is set by the server;
    // we use this value when checking the answer string for command processing from server

    // return codes of server: CtrlCmdOk_csp09      0
    //                         CtrlCmdErr_csp09     1
    //                         CtrlRelease_csp09    2

    if (CtrlCmdOk_csp09 == commErr)         // no server error; check for command
    {                                       // processing return code
        rc = ls40_CheckAnswer(*options, nReplyLen);
    }
    else if (CtrlRelease_csp09 == commErr)      // relaese command; nothing to do
    {
        ;        
    }
    else                                        // server error
    {
        fprintf (options->outfile, "Replication server error (%d : '%s')",
                                   STATIC_CAST(int, commErr),
                                   STATIC_CAST(char*, errtext) );
        rc = PRG_SERVER_ERROR;
    }

    return rc;
}
//  ls40_ServerCommand()


/* 
  -----------------------------------------------------------------------------
  function: main
  -----------------------------------------------------------------------------
*/
int main (int argc, char* argv[])
{    int             MainRetCode = PRG_OK;
    ls40_Options    options;


    tsp_compname c_component;           // initialize connection to Xserver
    strcpy(c_component, C_COMPONENT);
    sqlinit(c_component, NULL);

    ls40_Initialize(options);

    MainRetCode = ls40_ParseOptions(&options, argc, argv);
    if (PRG_OK != MainRetCode)
    {
        exit(PRG_FATAL);
    }

    // Connect to replication server;
    // this function fills options->nMaxDataSize und options-session and allocates
    // space for command buffer;
    // in case of error we need to release this connection
    if (PRG_OK == MainRetCode)
    {
        MainRetCode = ls40_ServerConnect(&options);
    }

    if ( (MainRetCode == PRG_OK) && (true == options.showServerVersion) )
    {
        MainRetCode = ls40_ShowServerVersion(options);
    }


    //*
    //*     User logon
    //*
    if (PRG_OK == MainRetCode)
    {
        MainRetCode = ls40_UserConnect(options);
    }

    
    //*     Command file
    //* Read Batch-File into variable 'cmd'
    if ( (PRG_OK == MainRetCode) && (NULL != options.infile) )
    {
        MainRetCode = ls40_Batch(options);
    }

    // Print general error msg to stderr in case of error
    if (PRG_OK != MainRetCode)
    {
        fprintf(stderr, "\n\nSome error(s) occurred running command.\n"
                        "---> See log file(s) for more information.\n");
    }
    
    //*     Disconnect Server
    if (NULL != options.session)
    {
        cn14release(&options.session);
    }

    sqlfinish(false);           // Undo LZU-initialization (sqlinit() )

    if ( NULL != options.pszCmdBuffer)
    {
        delete [] options.pszCmdBuffer;
        options.pszCmdBuffer = NULL;
    }
    if ( NULL != options.pszReplyBuffer)
    {
        delete [] options.pszReplyBuffer;
        options.pszReplyBuffer = NULL;
    }

    if ( NULL != options.infile )
    {
        fclose (options.infile);
        options.infile = NULL;
    }
    if ( (options.outfile != stdout) && (NULL != options.outfile) )
    {
        fclose (options.outfile);
        options.outfile = NULL;
    }

    tsp00_Int2 i = 0;
    
    for (i; i < MAX_PARAM_LS40; ++i)
    {
        if (NULL != options.pszParamList_ls40[i])
        {
            delete [] options.pszParamList_ls40[i];
            options.pszParamList_ls40[i] = NULL;
        }
    }

    exit(MainRetCode);
    return MainRetCode;
}
//  main()
