/*!**********************************************************************

  module: vwd206dbfs.cpp

  -----------------------------------------------------------------------

  responsible:  Holger Bischoff
				Markus zgen

  special area: WebAgent

  description:  FTP Interface to database-filesystem

  see also:

  -----------------------------------------------------------------------

                         Copyright by SAP AG, 1999



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

************************************************************************/

/*=======================================================================
 *  INCLUDES
 *=====================================================================*/

#include <stdio.h>

#include "hwd26err.h"
#include "hwd206dbfs.h"
#include "hwd208err.h"

/*!**********************************************************************

  Chapter: twd206_Intern
  
************************************************************************/

class twd206_Intern
{
public:
    // default contructor
    twd206_Intern( sapdbwa_SessionPoolP &pool ) : pool(pool) {};
    void ResetError();

    // make new path string. Absolute paths begin with DBFS_PATH_SEP_WD101.
    // calculate the path reached, if we are in actPath and make a
    // 'cd absOrRelPath'. Two dots '..' goes to the parent path.
    // example: actPath = '/usr/pgm'; relPath = '../wrk/test'
    //          => newPath = '/usr/wrk/test'
    sapdbwa_Bool MakePath( const ZString &absOrRelPath, ZString &newPath );

    void ResetErr();

    class DirStream : public twd207_BasicInputStream {
    public:
        DirStream();
        ~DirStream();

        void Init( twd101DBFSP dbfs );
        sapdbwa_Bool Open( const ZString &path, sapdbwa_Bool details );
        sapdbwa_Bool Close();
        sapdbwa_Bool IsOpened();
        sapdbwa_Int4 Read( char *buf, sapdbwa_Int4 bufSize );

    private:
        sapdbwa_Bool   isOpened;
        twd101DBFSP  dbfs;
        twd101DBFile fno;       // file handle for opened directory
        sapdbwa_Bool   details;   // show detailed file info (date, time, length)
    }; // class DirStream

    // used to retrieve a file from the dbfs
    class FileInStream : public twd207_BasicInputStream {
    public:
        FileInStream();
        ~FileInStream() {};

        void Init( twd101DBFSP dbfs );
        sapdbwa_Bool Open( const ZString &path );
        sapdbwa_Bool Close();
        sapdbwa_Bool IsOpened();
        sapdbwa_Int4 Read( char *buf, sapdbwa_Int4 bufSize );        
    private:
        sapdbwa_Bool   isOpened;
        twd101DBFSP  dbfs;
        twd101DBFile fno;       // file handle for opened file
    }; // class FileInStream

    // used to store a file to the dbfs
    class FileOutStream : public twd207_BasicOutputStream {
    public:
        FileOutStream();
        ~FileOutStream() {};

        void Init( twd101DBFSP dbfs );
        sapdbwa_Bool Open( const ZString &path );
        sapdbwa_Bool Close();
        sapdbwa_Bool IsOpened();
        sapdbwa_Int4 Write( char *buf, sapdbwa_Int4 bufSize );
    private:
        sapdbwa_Bool   isOpened;
        twd101DBFSP  dbfs;
        twd101DBFile fno;       // file handle for opened file
    }; // class FileOutStream

    twd101DBFSP     dbfs;          // dbfs-"object" for the C-API
    DirStream       dirStream;     // only one dir can be read at once
    FileInStream    fileInStream;  // only one file can be read at once
    FileOutStream   fileOutStream; // only one file can be written at once
    ZString         workingDir;    // actual working directory
    sapdbwa_SessionPoolP &pool;          // session-pool used by dbfs
    sapdbwa_Int4      lastError;
    ZString         lastErrorMsg;
}; // class twd206Intern


void twd206_Intern::ResetError()
{
    this->lastError = 0; // NO_ERROR
    this->lastErrorMsg = "";
}; // ResetError

/* internal helper functions */
ZString wd206_GetWord( const ZString &path, sapdbwa_Int4 &pos );
sapdbwa_Bool wd206_CdUp( ZString &path, sapdbwa_Int4 &pos );
void wd206_Cd( ZString &path, const ZString &cdPath, sapdbwa_Int4 &pos );


sapdbwa_Bool twd206_Intern::MakePath( const ZString &absOrRelPath,
                                    ZString       &newPath )
{
    sapdbwa_Bool  ok = sapdbwa_True;

    if (absOrRelPath[0] == DBFS_PATH_SEP_WD101) {
        /* absolute path */
        newPath = absOrRelPath;
    } else { /* relative path */
        ZString     tmpPath = this->workingDir;
        sapdbwa_Int4  tmpPathPos = tmpPath.GetSize()-1;
        sapdbwa_Int4  relPathPos = 0;
        ZString     word;

        do {
            word = wd206_GetWord( absOrRelPath, relPathPos );
            if (word == "..") {
                if (!wd206_CdUp( tmpPath, tmpPathPos )) {
                    // invalid dir
                    ok = sapdbwa_False;
                }; // if
            } else if (word.GetSize() > 0) {
                wd206_Cd( tmpPath, word, tmpPathPos );
            }; // else
        } while (ok && word.GetSize() > 0);
        if (ok) {
            newPath = tmpPath;
        } else {
            this->lastErrorMsg = ERR_MSG_INVALID_FILE_OR_DIR_WD208;
            this->lastError = ERR_INVALID_FILE_OR_DIR_WD208;
        }; // else
    }; // else

    return ok;
}; // MakePath

/*!**********************************************************************

  EndChapter: twd206_Intern
  
************************************************************************/

/*!**********************************************************************

  Chapter: twd206_Intern::DirStream
  
************************************************************************/

twd206_Intern::DirStream::DirStream()
    : isOpened(sapdbwa_False), dbfs( NULL ), details( sapdbwa_False )
{
}; // DirStream 


void twd206_Intern::DirStream::Init( twd101DBFSP dbfsP )
{
    this->dbfs = dbfsP;
}; // Init


twd206_Intern::DirStream::~DirStream()
{
}; // ~DirStream


sapdbwa_Bool twd206_Intern::DirStream::Open( const ZString &path,
                                           sapdbwa_Bool     showDetails )
{
    sapdbwa_Bool openOk;

    this->details = showDetails;
    this->fno = wd101OpenDir( this->dbfs, (char*) path, name_asc );
    openOk = (this->fno != -1);
    if (!openOk) {
        this->lastErrorMsg = wd101GetErrorMsg( this->dbfs );
        this->lastError = wd101GetErrorId( this->dbfs );
    }; // if
    this->isOpened = openOk;

    return openOk;
}; // Open


sapdbwa_Bool twd206_Intern::DirStream::Close()
{
    sapdbwa_Bool closeOk = sapdbwa_True;

    if (this->isOpened) {
        closeOk = wd101CloseDir( this->dbfs, this->fno );
        if (!closeOk) {
            this->lastErrorMsg = wd101GetErrorMsg( this->dbfs );
            this->lastError = wd101GetErrorId( this->dbfs );
        }; // if
        this->isOpened = !closeOk;
    }; // if

    return closeOk;
}; // Close


sapdbwa_Bool twd206_Intern::DirStream::IsOpened()
{
    return this->isOpened;
}; // IsOpened


sapdbwa_Int4 twd206_Intern::DirStream::Read( char *buf, sapdbwa_Int4 bufSize )
{
    sapdbwa_Int4 bytesRead;

    bytesRead = wd101GetDirEntry( this->dbfs, this->fno, buf, bufSize,
                                  this->details );
    if (bytesRead < 0) {
        this->lastErrorMsg = wd101GetErrorMsg( this->dbfs );
        this->lastError = wd101GetErrorId( this->dbfs );
        if (this->lastError == ERR_NO_MORE_ENTRIES_WD26) {
            /* not really an error */
            bytesRead = 0;
        };
    }; // if

    return bytesRead;
}; // Read

/*!**********************************************************************

  EndChapter: twd206_Intern::DirStream
  
************************************************************************/
/*!**********************************************************************

  Chapter: twd206_Intern::FileInStream
  
************************************************************************/

twd206_Intern::FileInStream::FileInStream()
    : isOpened(sapdbwa_False), dbfs( NULL )
{
}; // FileInStream 


void twd206_Intern::FileInStream::Init( twd101DBFSP dbfsP )
{
    this->dbfs = dbfsP;
}; // FileInStream


sapdbwa_Bool twd206_Intern::FileInStream::Open( const ZString &path )
{
    sapdbwa_Bool openOk;

    this->fno = wd101Open( this->dbfs, (char*) path, access_read );
    openOk = (this->fno != -1);
    if (!openOk) {
        this->lastErrorMsg = wd101GetErrorMsg( this->dbfs );
        this->lastError = wd101GetErrorId( this->dbfs );
    }; // if
    this->isOpened = openOk;

    return openOk;
}; // Open


sapdbwa_Bool twd206_Intern::FileInStream::Close()
{
    sapdbwa_Bool closeOk = sapdbwa_True;

    if (this->isOpened) {
        closeOk = wd101Close( this->dbfs, this->fno );
        if (!closeOk) {
            this->lastErrorMsg = wd101GetErrorMsg( this->dbfs );
            this->lastError = wd101GetErrorId( this->dbfs );
        }; // if
        this->isOpened = !closeOk;
    }; // if

    return closeOk;
}; // Close


sapdbwa_Bool twd206_Intern::FileInStream::IsOpened()
{
    return this->isOpened;
}; // IsOpened


sapdbwa_Int4 twd206_Intern::FileInStream::Read( char *buf, sapdbwa_Int4 bufSize )
{
    sapdbwa_Int4 bytesRead;

    bytesRead = wd101Read( this->dbfs, this->fno, buf, bufSize );
    if (bytesRead<0) {
        // error
        this->lastErrorMsg = wd101GetErrorMsg( this->dbfs );
        this->lastError = wd101GetErrorId( this->dbfs );
    }; // else

    return bytesRead;
}; // Read


/*!**********************************************************************

  EndChapter: twd206_Intern::FileInStream
  
************************************************************************/
/*!**********************************************************************

  Chapter: twd206_Intern::FileOutStream
  
************************************************************************/

twd206_Intern::FileOutStream::FileOutStream()
    : isOpened(sapdbwa_False), dbfs( NULL )
{
}; // FileOutStream 


void twd206_Intern::FileOutStream::Init( twd101DBFSP dbfsP )
{
    this->dbfs = dbfsP;
}; // Init


sapdbwa_Bool twd206_Intern::FileOutStream::Open( const ZString &path )
{
    sapdbwa_Bool openOk;

    this->fno = wd101Create( this->dbfs, (char*) path,
                             access_write_compressed );
    openOk = (this->fno != -1);
    if (!openOk) {
        this->lastErrorMsg = wd101GetErrorMsg( this->dbfs );
        this->lastError = wd101GetErrorId( this->dbfs );
    }; // if
    this->isOpened = openOk;

    return openOk;
}; // Open


sapdbwa_Bool twd206_Intern::FileOutStream::Close()
{
    sapdbwa_Bool closeOk = sapdbwa_True;

    if (this->isOpened) {
        closeOk = wd101Close( this->dbfs, this->fno );
        if (!closeOk) {
            this->lastErrorMsg = wd101GetErrorMsg( this->dbfs );
            this->lastError = wd101GetErrorId( this->dbfs );
        }; // if
        this->isOpened = !closeOk;
    }; // if

    return closeOk;
}; // Close


sapdbwa_Bool twd206_Intern::FileOutStream::IsOpened()
{
    return this->isOpened;
}; // IsOpened


sapdbwa_Int4 twd206_Intern::FileOutStream::Write( char *buf, sapdbwa_Int4 bufSize )
{
    sapdbwa_Int4 bytesWritten;

    bytesWritten = wd101Write( this->dbfs, this->fno, buf, bufSize );
    if (bytesWritten < 0) {
        // error
        this->lastErrorMsg = wd101GetErrorMsg( this->dbfs );
        this->lastError = wd101GetErrorId( this->dbfs );
    }; // else

    return bytesWritten;
}; // Write

/*!**********************************************************************

  EndChapter: twd206_Intern::FileOutStream
  
************************************************************************/
/*!**********************************************************************

  Chapter: twd206_DBFS
  
************************************************************************/

///////////////////////////////////
// constructor
///////////////////////////////////

twd206_DBFS::twd206_DBFS( sapdbwa_SessionPoolP pool )
    : twd205_BasicFS()
{
    this->in = new twd206_Intern( pool );

    if (this->in != NULL) {
        this->in->workingDir   = "/";
        this->in->dbfs         = wd101CreateDBFS( pool, DBFS_PARTITION_ID_WD101 );
        this->in->lastError    = 0;
        this->in->lastErrorMsg = "";
        if (this->in->dbfs != NULL) {
            this->in->dirStream.Init( this->in->dbfs );
            this->in->fileInStream.Init( this->in->dbfs );
            this->in->fileOutStream.Init( this->in->dbfs );
        };
    }; // if
}; // twd206_DBFS


///////////////////////////////////
// destructor
///////////////////////////////////

twd206_DBFS::~twd206_DBFS()
{
    if (this->in->dbfs != NULL) {
        wd101DestroyDBFS( this->in->dbfs );
        this->in->dbfs = NULL;
    }; // if
}; // ~twd206_DBFS

    
///////////////////////////////////
// public
///////////////////////////////////

sapdbwa_Bool twd206_DBFS::Delete( const ZString &path )
{
    sapdbwa_Bool deleteOk = sapdbwa_True;
    ZString    absolutePath;

    this->in->ResetError();
    if (this->in->MakePath( path, absolutePath )) {
        if (!wd101Delete( this->in->dbfs, (char*) absolutePath )) {
            this->in->lastErrorMsg = wd101GetErrorMsg( this->in->dbfs );
            this->in->lastError = wd101GetErrorId( this->in->dbfs );
            deleteOk = sapdbwa_False;
        }; // if
    } else {
        deleteOk = sapdbwa_False;        
    }; // else
        
    return deleteOk;
}; // Delete


sapdbwa_Bool twd206_DBFS::GetDir( ZString &dir )
{
    this->in->ResetError();
    dir = this->in->workingDir;

    return sapdbwa_True;
}; // GetDir


twd207_BasicInputStream* twd206_DBFS::GetDirStream( const ZString &path,
                                                    sapdbwa_Bool     details )
{
    twd206_Intern::DirStream *returnStream = NULL;
    ZString absolutePath = this->in->workingDir;

    this->in->ResetError();
    if (path[0] == DBFS_PATH_SEP_WD101) {
        /* path is absolute path */
        absolutePath = path;
    } else {
        /* path is relativ path */
        absolutePath += path;
    }; // else
    this->in->ResetError();
    if (this->in->dirStream.IsOpened()) {
        // already one dir opened
        this->in->lastErrorMsg = ERR_MSG_DIR_ALREADY_OPEN_WD208;
        this->in->lastError    = ERR_DIR_ALREADY_OPEN_WD208;
    } else {
        if (this->in->dirStream.Open( absolutePath, details )) {
            returnStream = &this->in->dirStream;
        } else {
            this->in->lastErrorMsg = wd101GetErrorMsg( this->in->dbfs );
            this->in->lastError    = wd101GetErrorId( this->in->dbfs );
        }; // else
    }; // else

    return returnStream;
}; // GetDirStream


// get stream to retrieve a file
twd207_BasicInputStream* twd206_DBFS::GetFileInStream( const ZString &path )
{
    twd206_Intern::FileInStream *returnStream = NULL;
    ZString                      absolutePath;

    this->in->ResetError();
    if (this->in->MakePath( path, absolutePath )) {
        if (this->in->fileInStream.IsOpened()) {
            // already one file opened
            this->in->lastErrorMsg = ERR_MSG_FILE_ALREADY_OPEN_WD208;
            this->in->lastError    = ERR_FILE_ALREADY_OPEN_WD208;
        } else {
            if (this->in->fileInStream.Open( absolutePath )) {
                returnStream = &this->in->fileInStream;
            } else {
                this->in->lastErrorMsg = wd101GetErrorMsg( this->in->dbfs );
                this->in->lastError    = wd101GetErrorId( this->in->dbfs );
            }; // else
        }; // else
    }; /* if */

    return returnStream;
}; // GetFileInStream


// get stream to store a file
twd207_BasicOutputStream* twd206_DBFS::GetFileOutStream( const ZString &path )
{
    twd206_Intern::FileOutStream *returnStream = NULL;

    this->in->ResetError();
    if (this->in->fileOutStream.IsOpened()) {
        // already one file opened
        this->in->lastErrorMsg = ERR_MSG_FILE_ALREADY_OPEN_WD208;
        this->in->lastError    = ERR_FILE_ALREADY_OPEN_WD208;
    } else {
        if (this->in->fileOutStream.Open( path )) {
            returnStream = &this->in->fileOutStream;
        } else {
            this->in->lastErrorMsg = wd101GetErrorMsg( this->in->dbfs );
            this->in->lastError    = wd101GetErrorId( this->in->dbfs );
        }; // else
    }; // else

    return returnStream;
}; // GetFileOutStream


sapdbwa_Bool twd206_DBFS::GetSize( const ZString &path, sapdbwa_Int4 &size )
{
    sapdbwa_Bool         ok = sapdbwa_True;
    struct st_db_stat  dbStat;
    ZString            absolutePath;

    this->in->ResetError();
    if (this->in->MakePath( path, absolutePath )) {
        if (wd101Stat( this->in->dbfs, absolutePath, &dbStat )) {
            size = dbStat.size;
        } else {
            ok = sapdbwa_False;
            this->in->lastErrorMsg = wd101GetErrorMsg( this->in->dbfs );
            this->in->lastError    = wd101GetErrorId( this->in->dbfs );
        }; // else
    } else {
        // invalid dir
        ok = sapdbwa_False;
    };

    return ok;
}; // GetSize


sapdbwa_Bool twd206_DBFS::GetStatStr( const ZString &path, ZString &statStr )
{
    sapdbwa_Bool        ok = sapdbwa_True;
    struct st_db_stat dbStat;
    char              sizeStr[100];
    ZString           absolutePath;

    this->in->ResetError();
    if (this->in->MakePath( path, absolutePath )) {
        if (wd101Stat( this->in->dbfs, absolutePath, &dbStat )) {
            switch (dbStat.mode) {
            case DB_DIR:
                statStr = "dir ";
                break;
            case DB_FIL:
                statStr = "fil ";
                break;
            case DB_LIN:
                statStr = "lin ";
                break;
            default:
                statStr = "*** ";
                break;
            }; // switch
            statStr += "\t";
            sprintf( sizeStr, "%d", dbStat.size );
            statStr += sizeStr;
            statStr += "\t";
            statStr += path;
        } else {
            ok = sapdbwa_False;
            this->in->lastErrorMsg = wd101GetErrorMsg( this->in->dbfs );
            this->in->lastError    = wd101GetErrorId( this->in->dbfs );
        }; // else
    } else {
        ok = sapdbwa_False;
    }; // else        

    return ok;
}; // GetStatStr


sapdbwa_Bool twd206_DBFS::MkDir( const ZString &newDir )
{
    twd101DBFile  newDirFD;
    sapdbwa_Bool    mkDirOk = sapdbwa_True;
    ZString       absoluteNewDir;
    
    this->in->ResetError();
    if (this->in->MakePath( newDir, absoluteNewDir )) {
        newDirFD = wd101MkDir( this->in->dbfs, (char*) absoluteNewDir );
        if (newDirFD < 0) {
            this->in->lastErrorMsg = wd101GetErrorMsg( this->in->dbfs );
            this->in->lastError = wd101GetErrorId( this->in->dbfs );
            mkDirOk = sapdbwa_False;
        }; // if
    } else {
        mkDirOk = sapdbwa_False;        
    }; // else

    return mkDirOk;
}; // MkDir


sapdbwa_Bool twd206_DBFS::CheckPath( const ZString &path, sapdbwa_Bool &pathExists )
{
    sapdbwa_Bool        checkPathOk = sapdbwa_True;
    struct st_db_stat dbStat;
    ZString           absolutePath;

    this->in->ResetError();
    if (this->in->MakePath( path, absolutePath )) {
        pathExists = wd101Stat( this->in->dbfs,
                                (char*) absolutePath, &dbStat );
        if (!pathExists && wd101GetErrorId( this->in->dbfs )
            != ERR_FILE_OR_DIR_NOT_FOUND_WD26) {
            // something went wrong
            this->in->lastErrorMsg = wd101GetErrorMsg( this->in->dbfs );
            this->in->lastError = wd101GetErrorId( this->in->dbfs );
            checkPathOk = sapdbwa_False;
        }; // if
    } else {
        checkPathOk = sapdbwa_False;
    }; // else

    return checkPathOk;
}; // PathExists


sapdbwa_Bool twd206_DBFS::RenameFromTo( const ZString &fromName,
                                      const ZString &toName )
{
    sapdbwa_Bool renameOk = sapdbwa_True;
    ZString    absoluteFromName;
    ZString    absoluteToName;

    this->in->ResetError();
    if (this->in->MakePath( fromName, absoluteFromName )
        && this->in->MakePath( toName, absoluteToName )) {
        renameOk = wd101Rename( this->in->dbfs,
                                (char*) absoluteFromName,
                                (char*) absoluteToName );
        if (!renameOk) {
            this->in->lastErrorMsg = wd101GetErrorMsg( this->in->dbfs );
            this->in->lastError = wd101GetErrorId( this->in->dbfs );
        }; // if
    } else {
        renameOk = sapdbwa_False;
    }; // else
        
    return renameOk;
}; // RenameFromTo


sapdbwa_Bool twd206_DBFS::RmDir( const ZString &dir )
{
    sapdbwa_Bool rmDirOk = sapdbwa_True;
    ZString    absoluteDir;

    this->in->ResetError();
    if (this->in->MakePath( dir, absoluteDir )) {
        if (!wd101RmDir( this->in->dbfs, (char*) absoluteDir, sapdbwa_False )) {
            this->in->lastErrorMsg = wd101GetErrorMsg( this->in->dbfs );
            this->in->lastError = wd101GetErrorId( this->in->dbfs );
            rmDirOk = sapdbwa_False;
        }; // if
    } else {
        rmDirOk = sapdbwa_False;        
    }; // else
        
    return rmDirOk;
}; // RmDir


sapdbwa_Bool twd206_DBFS::ChDir( const ZString &dir )
{
    sapdbwa_Bool        chDirOk = sapdbwa_True;
    ZString           tmpDir  = "";
    struct st_db_stat dbStat;

    this->in->ResetError();
    // make new directory string -> tmpPath
    if (!this->in->MakePath( dir, tmpDir )) {
        chDirOk = sapdbwa_False;
    }; // if

    if (chDirOk) {
        if (wd101Stat( this->in->dbfs, (char*) tmpDir, &dbStat )) {
            if (dbStat.mode == DB_DIR) {
                // chDir succesful
                this->in->workingDir = tmpDir;
            } else {
                twd26ErrP err = wd26CreateErr();
                wd26SetErr( err, ERR_NOT_DIR_WD26, (char*) dir, NULL );
                this->in->lastErrorMsg = wd26GetMsg( err );
                this->in->lastError = wd26GetId( err );
                wd26DestroyErr( err );
                chDirOk = sapdbwa_False;
            }; // else
        } else {
            this->in->lastErrorMsg = wd101GetErrorMsg( this->in->dbfs );
            this->in->lastError = wd101GetErrorId( this->in->dbfs );
            chDirOk = sapdbwa_False;
        }; // if
    }; // if
        
    return chDirOk;
}; // ChDir


sapdbwa_Bool twd206_DBFS::CdUp()
{
    sapdbwa_Bool cdUpOk = sapdbwa_True;

    this->in->ResetError();
    cdUpOk = this->ChDir( ".." );

    return cdUpOk;
}; // CdUp


const ZString& twd206_DBFS::GetLastErrorMsg()
{
    return this->in->lastErrorMsg;
}; // GetLastErrorMsg


sapdbwa_Int4 twd206_DBFS::GetLastError()
{
    return this->in->lastError;
}; // GetLastError


///////////////////////////////////
// private
///////////////////////////////////

/*!**********************************************************************

  EndChapter: twd206_DBFS
  
************************************************************************/
/*!**********************************************************************

  Chapter: internal helper functions
  
************************************************************************/

#define MAX_WORD_LEN 255

ZString wd206_GetWord( const ZString &path, sapdbwa_Int4 &pos )
{
    char        res[MAX_WORD_LEN];
    sapdbwa_Int4  resPos = 0;

    while ( resPos < sizeof(res)-2
            && pos < path.GetSize()
            && path[pos] != DBFS_PATH_SEP_WD101) {
        res[resPos] = path[pos];
        pos++;
        resPos++;
    }; // while
    res[resPos] = '\0';
    // skip over DBFS_PATH_SEP_WD101
    if (path[pos] == DBFS_PATH_SEP_WD101) {
        pos++;
    }; // if

    return ZString(res);
}; // wd206_GetWord    


sapdbwa_Bool wd206_CdUp( ZString &path, sapdbwa_Int4 &pos )
{
    sapdbwa_Bool ok = sapdbwa_True;

    if (path!="/") {
        // dont skip over the root DBFS_PATH_SEP_WD101
        while (pos > 1 && path[pos] != DBFS_PATH_SEP_WD101) {
            pos--;
        }; // while
    } else {
        ok = sapdbwa_False;
    }; // else
    path[pos] = '\0';
    path.SetSize( pos );

    return ok;
}; // wd206_CdUp


void wd206_Cd( ZString &path, const ZString &cdPath, sapdbwa_Int4 &pos )
{
    if (path[path.GetSize()-1] != DBFS_PATH_SEP_WD101) {
        path += DBFS_PATH_SEP_WD101;
    }; // if
    path += cdPath;
    pos = path.GetSize()-1;
}; // wd206_Cd

/*!**********************************************************************

  EndChapter: internal helper functions
  
************************************************************************/
