/*!**********************************************************************

  module: vwd210inetd.cpp

  -----------------------------------------------------------------------

  responsible:  Holger Bischoff
				Markus zgen

  special area: WebAgent

  description:  Inet-daemon

  see also:

  -----------------------------------------------------------------------

                         Copyright by SAP AG, 1999



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

************************************************************************/

#include <stdio.h>

#include "geo102.h"
#include "hwd04wases.h"
#include "hwd20control.h"
#include "hwd32hses.h"
#include "hwd201sock.h"
#include "hwd203util.h"
#include "hwd206dbfs.h"
#include "hwd220ftp.h"

#ifndef IPPORT_FTP
#define IPPORT_FTP	21
#endif

sapdbwa_Bool wd210InitLstnSocket( sapdbwa_Int4         port,
                                twd201_Socket     *pLstnSocket,
                                twd201_SockAddrIn *pLstnSockName );

void wd210Error( const char *where );

sapdbwa_Bool wd210FtpConnection( twd201_Socket     socket,
                               twd201_SockAddrIn sockName,
                               sapdbwa_Bool        debugMode );

static sapdbwa_SessionPoolP wd210SessionPool;

int main( int argc, char *argv[] )
{
    twd201_SockRet sockRet;
    sapdbwa_Bool ok;
    twd201_Socket     lstnSocket;
    twd201_SockAddrIn lstnSockName;
    twd26ErrP         err = wd26CreateErr();
    sapdbwa_Bool      debugMode = sapdbwa_False; /* set by -v option	*/

    if (argc > 1) {
        if (strcmp(argv[1], "-v" ) == 0) {
            debugMode = sapdbwa_True;
        }; 
    };
    wd210SessionPool = wd20ApiDBFSInit( err );
    if (wd210SessionPool == NULL) {
        printf( "main; wd20ApiDBFSInit: %s", wd26GetMsg( err ));
        return -1;
    };

    sockRet = wd201SocketStartup();
    if (sockRet != 0) {
        /* error	*/
        printf( "SocketStartup failed!!!\n");
        return -1;
    } else {
        printf( "SocketStartup ok\n" );
    };

    ok = wd210InitLstnSocket( IPPORT_FTP, &lstnSocket, &lstnSockName );
    if (ok) {
        twd201_Socket     incomingSocket;
        twd201_SockAddrIn incomingSockName;

        printf( "InitSocket ok\n" );
        while (1) { /* never stop listening	*/
            /* this is blocking	*/
            ok = wd203Accept( lstnSocket, &incomingSocket, &incomingSockName );
            if (ok) {
                printf("main: incoming request :-)\n");
                
                ok = wd210FtpConnection( incomingSocket,
                                         incomingSockName,
                                         debugMode );
                printf("main: request finished\n");
            };
        };
    };

    sockRet = wd201CloseSocket( lstnSocket );
    if (sockRet != 0) {
        wd203SockErr( "main", "CloseSocket" );
    } else {
        printf( "CloseSocket ok\n" );
    };

    sockRet = wd201SocketCleanup();
    if (sockRet != 0) {
        wd203SockErr( "main", "SocketCleanup" );
    } else {
        printf( "Cleanup ok\n" );
    };

    wd20ApiDBFSExit();
    wd26DestroyErr( err );
    return sockRet;
};


sapdbwa_Bool wd210InitLstnSocket( sapdbwa_Int4         port,
                                twd201_Socket     *pLstnSocket,
                                twd201_SockAddrIn *pLstnSockName )
{
    twd201_SockRet sockRet   = SOCKET_ERROR;
    sapdbwa_Bool     initOk    = sapdbwa_True;

    *pLstnSocket = wd201Socket( AF_INET, SOCK_STREAM, 0 );
    if (*pLstnSocket == INVALID_SOCKET) {
        wd203SockErr( "wd210InitSocket", "Socket" );
        initOk = sapdbwa_False;
    } else {
        pLstnSockName->sin_family = PF_INET;
        pLstnSockName->sin_port   = (u_short) htons((u_short) port);
        pLstnSockName->sin_addr.s_addr = INADDR_ANY; // use local address
        sockRet = wd201Bind( *pLstnSocket,
                             (twd201_SockAddr*) pLstnSockName,
                             SOCKADDR_LEN );
        if (sockRet == SOCKET_ERROR) {
            wd203SockErr( "wd210InitSocket", "Bind" );
            initOk = sapdbwa_False;
        } else {
            sockRet = wd201Listen( *pLstnSocket, 5 );
            if (sockRet == SOCKET_ERROR) {
                wd203SockErr( "wd210InitSocket", "Listen" );
                initOk = sapdbwa_False;
            };
        };
    };

    return initOk;
};

#define FTP_ID 0

sapdbwa_Bool wd210FtpConnection( twd201_Socket      socket,
                               twd201_SockAddrIn  sockName,
                               sapdbwa_Bool         debugMode )
{
    sapdbwa_Int4      serviceRet;

    twd206_DBFS    *dbfs       = new twd206_DBFS( wd210SessionPool );

    twd220_FtpService *pFtpService =
        new twd220_FtpService( socket, sockName, *dbfs, debugMode );

    serviceRet = pFtpService->Service();

    /* cleanup	*/
    delete pFtpService;
    delete dbfs;

    return sapdbwa_True;
};
