/*!**********************************************************************

  module: vwd21req.c

  -----------------------------------------------------------------------

  responsible:  Holger Bischoff
				Markus Oezgen

  special area: WebAgent

  description:  HttpRequest functions

  see also:

  -----------------------------------------------------------------------

                         Copyright by SAP AG, 1998



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

************************************************************************/

/*=======================================================================
 *  INCLUDES
 *=====================================================================*/

#include "hwd12hpool.h"
#include "hwd21req.h"
#include "hwd28dict.h"

void * wd21_RequestConstructor();

void wd21_RequestDestructor( void *requestHandle );

void wd21_RequestInit( sapdbwa_HttpRequestP     req,
                       twd00ServerHandle   serverHandle,
                       twd00ReadBodyFunc   readBody,
                       twd00GetHeaderFunc  getHeader,
                       twd00IsSecureFunc   isSecure );

sapdbwa_Bool wd21_InsertCookies( sapdbwa_HttpRequestP req );

sapdbwa_Bool wd21_InsertParams( sapdbwa_HttpRequestP req );

sapdbwa_Bool wd21_InsertValues( twd28DictionaryP dict, char *queryString );

sapdbwa_Bool wd21_InsertCookieValues( twd28DictionaryP  dict,
                                    const char       *queryString );

void wd21_PlusToSpace( char *str, sapdbwa_Int4 len );

sapdbwa_Bool wd21_ReadBodyData( sapdbwa_HttpRequestP req );

void wd21_UnescapeUrl( char *url, sapdbwa_Int4 len );

sapdbwa_Int4 wd21_Base64Decode( char       *bufplain,
                              const char *bufcoded);

/*!**********************************************************************

  Chapter: sapdbwa_HttpRequestP (functions from hwd03waapi.h)

  description:
  
************************************************************************/

struct st_wa_http_req
{
    const char	*Uri;		        /* Url without the nodename  */
    const char	*IfModifiedSince;   /* HTTP Date, Only if requested
                                             * is newer */
    const char	*PathInfo;          /* More parameters               */
    const char	*RequestMethod;	    /* Method of request(HEAD or POST)*/
    const char	*ContentType;	    /* Encoding type	             */
    const char	*ContentLength;	    /* Length of message body	     */
    const char	*PhysicalPath;
    const char	*ServerName;
    
    twd28DictionaryP   cookieValues;    
    twd28DictionaryP   paramValues;
    sapdbwa_Bool         cookiesInserted;
    sapdbwa_Bool         paramsInserted;
    char              *bodyData;      /* for POST requests body data is
                                       * copied to bodyData and paramBodyData */
    char              *paramBodyData; /* copy of bodyData used for method
                                       * POST. InsertParams replaces '='
                                       * and '&' with '\0' */
    sapdbwa_Int4         bodyDataLen;   /* == contentLength */
    sapdbwa_Int4         bodyDataPos;   /* actual position in bodyData for
                                       * WAReadBody */
    char              *queryString;
    char              *paramQueryString;
    char              *authHeader;
    
    twd00ReadBodyFunc    ReadBody;
    twd00GetHeaderFunc   GetHeader;
    twd00IsSecureFunc    IsSecure;
    
    twd00ServerHandle    serverHandle;
    twd26ErrP            err;
}; /* struct st_wa_http_req */


static twd12HandlePoolP wd21requestPool = NULL;

sapdbwa_Bool wd21CreateRequestHandlePool()
{
    wd21requestPool = wd12CreateHandlePool( wd21_RequestConstructor,
                                            wd21_RequestDestructor );

    return (wd21requestPool != NULL);
} /* wd21CreateRequestHandlePool */


void wd21DestroyRequestHandlePool()
{
    wd12DestroyHandlePool( wd21requestPool );
} /* wd21DestroyRequestHandlePool */


twd21HttpRequestP wd21CreateRequest( twd00ServerHandle  serverHandle,
                                     twd00ReadBodyFunc  readBody,
                                     twd00GetHeaderFunc getHeader,
                                     twd00IsSecureFunc  isSecure )
{
    twd21HttpRequestP newRequest = NULL;

    newRequest = wd12GetHandle( wd21requestPool );
    if (newRequest != NULL) {
        wd21_RequestInit( newRequest,
                          serverHandle,
                          readBody,
                          getHeader,
                          isSecure );
    };
    
    return newRequest;
} /* wd21CreateRequest */


void wd21DestroyRequest( twd21HttpRequestP req )
{
    wd12PutHandle( wd21requestPool, req );
} /* wd21DestroyRequest */


sapdbwa_ErrP WINAPI sapdbwa_ReqGetErr( sapdbwa_HttpRequestP req )
{
    return req->err;
} /* sapdbwa_ReqGetErr */


sapdbwa_Bool WINAPI sapdbwa_GetAuthorization( sapdbwa_HttpRequestP   req,
                                       const char      **userName,
                                       const char      **password )
{
    const char *authHeader = NULL;
    sapdbwa_Bool  ok = true;
    sapdbwa_Bool  allocatOk = true;
    char       *colonPos = NULL;

    *userName = NULL;
    *password = NULL;
    authHeader = sapdbwa_GetHeader( req, "HTTP_AUTHORIZATION" );
    if (authHeader == NULL) {
        authHeader = sapdbwa_GetHeader( req, "HTTP_PROXY_AUTHORIZATION" );
    };
    if (authHeader != NULL) {
        if (req->authHeader != NULL) {
            sapdbwa_SQLFREE( (sapdbwa_UInt1*) req->authHeader );
        };
        sapdbwa_SQLALLOCAT( strlen(authHeader)+1,
                       (sapdbwa_UInt1**) &req->authHeader,
                       &allocatOk );
        if (allocatOk) {
            const char *startPos = authHeader;
            const char *base64Pos = NULL;
            /* skip leading spaces */
            while (*startPos != '\0' && *startPos == ' ')
                startPos++;
            ok = (startPos[0] == 'B'
                  && startPos[1] == 'a'
                  && startPos[2] == 's'
                  && startPos[3] == 'i'
                  && startPos[4] == 'c'
                  && startPos[5] == ' ');
            if (ok) {
                /* find space between Basic and base64-encoded string */
                base64Pos = startPos + 5;
                /* skip leading spaces */
                while (base64Pos != NULL && *base64Pos != '\0' && *base64Pos == ' ')
                    base64Pos++;
                if (base64Pos != NULL) {
                    wd21_Base64Decode( req->authHeader, base64Pos );
                    colonPos = strchr( req->authHeader, ':' );
                    if (colonPos != NULL) {
                        *colonPos = '\0';
                        *userName = req->authHeader;
                        *password = colonPos+1;
                    } else {
                        /* wrong header format */
                        ok = false;
                    };
                } else {
                    /* wrong header format */
                    ok = false;
                };
            };
        } else {
            /* memory alloc failure */
            ok = false;
        };
    } else {
        /* Authorization Header does not exist */
        ok = false;
    };

    return ok;
} /* sapdbwa_GetAuthorization */


sapdbwa_Bool WINAPI sapdbwa_GetCookieNames( sapdbwa_HttpRequestP req,
                                     sapdbwa_StringSeqP   names )
{
    sapdbwa_Bool  ok = true;

    if (!req->cookiesInserted) {
        wd21_InsertCookies( req );
        req->cookiesInserted = true;
    }; /* if */

    ok = wd28GetKeys( req->cookieValues, names );
    
    return ok;
} /* sapdbwa_GetCookieNames */


const char* WINAPI sapdbwa_GetCookieValue( sapdbwa_HttpRequestP req, const char *name )
{
    const char *value = NULL;
    
    if (!req->cookiesInserted) {
        wd21_InsertCookies( req );
        req->cookiesInserted = true;
    }; /* if */

    value = wd28GetValueByKey( req->cookieValues, name );

    return value;
} /* sapdbwa_GetCookieValue */


const char* WINAPI sapdbwa_GetHeader( sapdbwa_HttpRequestP req, const char *fieldName )
{
    return req->GetHeader( req->serverHandle, fieldName );
} /* sapdbwa_GetHeader */


const char* WINAPI sapdbwa_GetRequestURI( sapdbwa_HttpRequestP req )
{
    return req->Uri;
} /* sapdbwa_GetRequestURI */


const char* WINAPI sapdbwa_GetIfModifiedSince( sapdbwa_HttpRequestP req )
{
    return req->IfModifiedSince;
} /* sapdbwa_GetIfModifiedSince */


const char* WINAPI sapdbwa_GetQueryString( sapdbwa_HttpRequestP req )
{
    return req->queryString;
} /* sapdbwa_GetQueryString */


const char* WINAPI sapdbwa_GetPathInfo( sapdbwa_HttpRequestP req )
{
    return req->PathInfo;
} /* sapdbwa_GetPathInfo */


const char* WINAPI sapdbwa_GetMethod( sapdbwa_HttpRequestP req )
{
    return req->RequestMethod;
} /* sapdbwa_GetRequestMethod */


const char* WINAPI sapdbwa_GetContentType( sapdbwa_HttpRequestP req )
{
    return req->ContentType;
} /* sapdbwa_GetContentType */


const char* WINAPI sapdbwa_GetContentLength( sapdbwa_HttpRequestP req )
{
    return req->ContentLength;
} /* sapdbwa_GetContentLength */


const char* WINAPI sapdbwa_GetPathTranslated( sapdbwa_HttpRequestP req )
{
    return req->PhysicalPath;
} /* sapdbwa_GetPhysicalPat */


twd00ServerHandle wd21GetServerHandle( twd21HttpRequestP req )
{
    return req->serverHandle;
} /* wd21GetServerHandle */


const char* WINAPI sapdbwa_GetServerName( sapdbwa_HttpRequestP req )
{
    return req->ServerName;
} /* sapdbwa_GetServerName */


sapdbwa_Bool WINAPI sapdbwa_IsSecure( sapdbwa_HttpRequestP  req )
{
    return req->IsSecure( req->serverHandle );
} /* sapdbwa_IsSecure */


sapdbwa_UInt4 WINAPI sapdbwa_ReadBody( sapdbwa_HttpRequestP  req,
                                char            *buff,
                                sapdbwa_UInt4      len )
{
    if (strcmp( sapdbwa_GetMethod( req ), "POST" ) == 0
        && (req->bodyData == NULL)) {
        if (!wd21_ReadBodyData( req )) {
            return -1;
        }; /* if */
    }; /* if */

    if (req->bodyData != NULL) {
        sapdbwa_Int4 copyBytes = min( req->bodyDataLen - req->bodyDataPos,
                                    len );
        memcpy( buff, req->bodyData + req->bodyDataPos, copyBytes );
        req->bodyDataPos += copyBytes;
        return copyBytes;
    } else {
        return req->ReadBody( req->serverHandle, buff, len );
    }; /* else */
} /* sapdbwa_ReadBody */


sapdbwa_Bool WINAPI sapdbwa_GetParameterNames( sapdbwa_HttpRequestP  req,
                                        sapdbwa_StringSeqP    names )
{
    sapdbwa_Bool ok = true;

    if (!req->paramsInserted) {
        wd21_InsertParams( req );
        req->paramsInserted = true;
    };
    wd11FreeElems( names );
    ok = wd28GetKeys( req->paramValues, names );

    return ok;
} /* sapdbwa_GetParameterNames */


sapdbwa_Bool WINAPI sapdbwa_GetParameterValues( sapdbwa_HttpRequestP  req,
                                         const char      *parameterName,
                                         sapdbwa_StringSeqP    values )
{
    sapdbwa_Bool ok = true;
    
    if (!req->paramsInserted) {
        wd21_InsertParams( req );
        req->paramsInserted = true;
    };
    wd11FreeElems( values );
    ok = wd28GetValuesByKey( req->paramValues, parameterName, values );

    return ok;
} /* sapdbwa_GetParameterValues */


void wd21SetErr( twd21HttpRequestP  req, twd26ErrP err )
{
    req->err = err;
} /* wd21SetErr */


void wd21SetUri( twd21HttpRequestP  req,
                 const char        *uri )
{
    req->Uri             = uri;
} /* wd21SetUri */


void wd21SetIfModifiedSince( twd21HttpRequestP  req,
                             const char        *ifModifiedSince )
{
    req->IfModifiedSince = ifModifiedSince;
} /* wd21SetIfModifiedSince */


void wd21SetQueryString( twd21HttpRequestP  req,
                         const char        *queryString )
{
    sapdbwa_Int4  queryStringLen = 0;
    sapdbwa_Bool  allocatOK = true;

    if (queryString != NULL && strlen(queryString) != 0) {
        if (req->queryString != NULL) {
            sapdbwa_SQLFREE( (sapdbwa_UInt1*) req->queryString );
            req->queryString = NULL;
        }; /* if */
        /* ouer parameters are not valid anymore */
        if (req->paramValues != NULL) {
            wd28FreeAllEntries( req->paramValues );
            req->paramsInserted = false;
        }; /* if */
    
        queryStringLen = strlen( queryString );
        sapdbwa_SQLALLOCAT( queryStringLen+1,
                       (sapdbwa_UInt1**) &req->queryString,
                       &allocatOK );
        if (allocatOK) {
            sapdbwa_SQLALLOCAT( queryStringLen+1,
                           (sapdbwa_UInt1**) &req->paramQueryString,
                           &allocatOK );
            if (allocatOK) {
                strcpy( req->queryString, queryString );
                wd21_PlusToSpace( req->queryString, queryStringLen );
                wd21_UnescapeUrl( req->queryString, queryStringLen );
                strcpy( req->paramQueryString, req->queryString );
            } else {
                req->queryString = NULL;
            }; /* else */
        } else {
            req->queryString = NULL;
        }; /* else */
    }; /* if */
} /* wd21SetQueryString */


void wd21SetPathInfo( twd21HttpRequestP  req,
                      const char        *pathInfo )
{
    req->PathInfo        = pathInfo;
} /* wd21SetPathInfo */

void wd21SetRequestMethod( twd21HttpRequestP  req,
                           const char        *requestMethod )
{
    req->RequestMethod   = requestMethod;
} /* wd21SetRequestMethod */

void wd21SetContentType( twd21HttpRequestP  req,
                         const char        *contentType )
{
    req->ContentType     = contentType;
} /* wd21SetContentType */

void wd21SetContentLength( twd21HttpRequestP  req,
                           const char        *contentLength )
{
    req->ContentLength   = contentLength;
} /* wd21SetContentLength */

void wd21SetPhysicalPath( twd21HttpRequestP  req,
                          const char        *physicalPath )
{
    req->PhysicalPath    = physicalPath;
} /* wd21SetPhysicalPath */

void wd21SetServerName( twd21HttpRequestP  req,
                        const char        *serverName )
{
    req->ServerName      = serverName;
} /* wd21SetServerName */

/*!**********************************************************************

  EndChapter: sapdbwa_HttpRequestP (functions from hwd03waapi.h)

************************************************************************/

void * wd21_RequestConstructor()
{
    twd21HttpRequestP newRequest = NULL;
    sapdbwa_Bool        allocatOK;

    sapdbwa_SQLALLOCAT( sizeof( struct st_wa_http_req ),
                   (sapdbwa_UInt1**) &newRequest,
                   &allocatOK );
    if (allocatOK) {
        /* these are allocated later if necessary */
        newRequest->authHeader        = NULL;
        newRequest->cookieValues      = NULL;
        newRequest->paramValues       = NULL;
        newRequest->bodyData          = NULL;
        newRequest->paramBodyData     = NULL;
        newRequest->queryString       = NULL;
        newRequest->paramQueryString  = NULL;
        newRequest->cookieValues = wd28CreateDictionary();
        newRequest->paramValues = wd28CreateDictionary();
    } else {
        newRequest = NULL;
    };
    
    return newRequest;
} /* wd21_RequestConstructor */


void wd21_RequestDestructor( void *handle )
{
    sapdbwa_HttpRequestP req = (sapdbwa_HttpRequestP) handle;
    
    if (req->paramValues != NULL) {
        wd28DestroyDictionary( req->paramValues );
    }; /* if */
    /* req->err is destroyed by caller! */
    if (req->authHeader != NULL) {
        sapdbwa_SQLFREE( (sapdbwa_UInt1*) req->authHeader );
    };
    if (req->cookieValues != NULL) {
        wd28DestroyDictionary( req->cookieValues );
    }; /* if */
    if (req->bodyData != NULL) {
        sapdbwa_SQLFREE( (sapdbwa_UInt1*) req->bodyData );
    }; /* if */
    if (req->paramBodyData != NULL) {
        sapdbwa_SQLFREE( (sapdbwa_UInt1*) req->paramBodyData );
    }; /* if */
    if (req->queryString != NULL) {
        sapdbwa_SQLFREE( (sapdbwa_UInt1*) req->queryString );
    }; /* if */
    if (req->paramQueryString != NULL) {
        sapdbwa_SQLFREE( (sapdbwa_UInt1*) req->paramQueryString );
    }; /* if */
    
    sapdbwa_SQLFREE( (sapdbwa_UInt1*) req );
} /* wd21_RequestDestructor */


void wd21_RequestInit( sapdbwa_HttpRequestP    req,
                       twd00ServerHandle  serverHandle,
                       twd00ReadBodyFunc  readBody,
                       twd00GetHeaderFunc getHeader,
                       twd00IsSecureFunc  isSecure )
{
    if (req->authHeader != NULL) {
        sapdbwa_SQLFREE( (sapdbwa_UInt1*) req->authHeader );
        req->authHeader = NULL;
    };
    req->serverHandle    = serverHandle;
    req->ReadBody        = readBody;
    req->GetHeader       = getHeader;
    req->IsSecure        = isSecure;
    req->Uri             = NULL;
    req->IfModifiedSince = NULL;
    req->PathInfo        = NULL;
    req->RequestMethod   = NULL;
    req->ContentType     = NULL;
    req->ContentLength   = NULL;
    req->PhysicalPath    = NULL;
    req->ServerName      = NULL;
    if (req->bodyData != NULL) {
        sapdbwa_SQLFREE( (sapdbwa_UInt1*) req->bodyData );
        req->bodyData        = NULL;
    }; /* if */
    if (req->paramBodyData != NULL) {
        sapdbwa_SQLFREE( (sapdbwa_UInt1*) req->paramBodyData );
        req->paramBodyData   = NULL;
    }; /* if */
    if (req->queryString != NULL) {
        sapdbwa_SQLFREE( (sapdbwa_UInt1*) req->queryString );
        req->queryString     = NULL;
    }; /* if */
    if (req->paramQueryString != NULL) {
        sapdbwa_SQLFREE( (sapdbwa_UInt1*) req->paramQueryString );
        req->paramQueryString = NULL;
    }; /* if */
    req->bodyDataLen     = 0;
    req->bodyDataPos     = 0;
    req->err             = NULL;
    wd28FreeAllEntries( req->cookieValues );
    req->cookiesInserted = false;
    wd28FreeAllEntries( req->paramValues );
    req->paramsInserted  = false;
} /* wd21_RequestInit */


sapdbwa_Bool wd21_InsertCookies( sapdbwa_HttpRequestP req )
{
    sapdbwa_Bool  insertOK = true;    
    const char *cookieStr = NULL;

    cookieStr = sapdbwa_GetHeader( req, "HTTP_COOKIE" );
    if (cookieStr != NULL) {
        wd21_InsertCookieValues( req->cookieValues, cookieStr );
    }; /* if */

    return insertOK;
} /* wd21_InsertCookies */    


sapdbwa_Bool wd21_InsertParams( sapdbwa_HttpRequestP req )
{
    sapdbwa_Bool  insertOK = true;
    
    if (strcmp( sapdbwa_GetMethod( req ), "POST" ) == 0
        && (req->bodyData == NULL)) {
        insertOK = wd21_ReadBodyData( req );
    }; /* if */
    if (insertOK) {
        insertOK = wd21_InsertValues( req->paramValues, req->paramBodyData );
    }; /* if */
        
    if (insertOK) {
        insertOK = wd21_InsertValues( req->paramValues,
                                      req->paramQueryString );
    }; /* if */

    return insertOK;
} /* wd21_InsertParams */

void wd21_SkipSpaces( char **pos )
{
	char *tpos = *pos;

    while (*tpos == ' ')
        tpos++;

	*pos = tpos;
} /* wd21_SkipSpaces */


sapdbwa_Bool wd21_InsertCookieValues( twd28DictionaryP  dict,
                                    const char       *cookieString )
{
    sapdbwa_Bool  insertOK = true;
    char       *pos      = NULL;
    const char *name     = NULL;
    const char *value    = NULL;
    char       *tmpCookie = NULL;
    sapdbwa_Bool  allocatOK = true;
    
    sapdbwa_SQLALLOCAT( strlen(cookieString)+1,
                   (sapdbwa_UInt1**) &tmpCookie,
                   &allocatOK );
    if (allocatOK) {
        strcpy( tmpCookie, cookieString );
        pos = tmpCookie;
        while (insertOK && (pos != NULL) && (*pos != '\0')) {
            name = pos;
            pos = strchr( pos, '=' );
            if (pos != NULL) {
                *pos = '\0'; /* '=' -> '\0' */
                pos++;
                value = pos;
                pos = strchr( pos, ';' );
                if (pos != NULL) {
                    *pos = '\0'; /* ';' -> '\0' */
                    pos++;
                    wd21_SkipSpaces(&pos);
                }; /* if */
                insertOK = wd28AddCopyEntry( dict, name, value );
            }; /* if */
        }; /* while */
        sapdbwa_SQLFREE( (sapdbwa_UInt1*) tmpCookie );
    } else {
        /* memory allocation error */
        insertOK = false;
    }; /* else */
    
    return insertOK;
} /* wd21_InsertCookieValues */
    

sapdbwa_Bool wd21_InsertValues( twd28DictionaryP dict, char *queryString )
{
    sapdbwa_Bool	 insertOK = true;
    char			*pos      = queryString;
    char			*name     = NULL;
    char			*value    = NULL;
	sapdbwa_UInt4	 strStack = 0;
    
	/* Scan query string */
    while ((pos != NULL) && (*pos != '\0')) {
        name = pos;

		/* Search value */
        pos = strchr( pos, '=' );
        if (pos != NULL) {
			/* Terminate name */
            *pos = '\0'; /* '=' -> '\0' */
            pos++;

			/* Is value a string? */
			if ( *pos == '"' ) {
				strStack=1;
				pos++;

				value = pos;

				if ( *pos != '\0' ) {

					do {

						pos = strchr( pos, '"');
						if ( pos != NULL ) {
							if ( *(pos-1) == '=' )
								strStack++;
							else
								strStack--;
						}

						if ( strStack > 0 )
							pos++;

					} while ( strStack && *pos != '\0' );

					/* Terminate value */
					*pos = '\0';
					pos++;
				}
			} else {
				value = pos;
			}

			/* Search next parameter */
			if ( *pos != '\0' ) {
				pos = strchr( pos, '&' );
				if (pos != NULL) {
					/* Terminate value */
					*pos = '\0'; /* '&' -> '\0' */
					pos++;
				}; /* if */
			}

			/* Add parameter/value to list */
            wd28AddEntry( dict, name, value );
        }; /* if */
    }; /* while */
    
    return insertOK;
} /* wd21_InsertValues */


char wd21_x2c(char *what) {
    char digit;

    digit = (what[0] >= 'A' ? ((what[0] & 0xdf) - 'A')+10 : (what[0] - '0'));
    digit *= 16;
    digit += (what[1] >= 'A' ? ((what[1] & 0xdf) - 'A')+10 : (what[1] - '0'));
    return(digit);
} /* wd21_x2c */


void wd21_UnescapeUrl( char *url, sapdbwa_Int4 len ) {
    sapdbwa_Int4 x,y;

    for(x=0,y=0; y<len; ++x,++y) {
        if((url[x] = url[y]) == '%') {
            url[x] = wd21_x2c(&url[y+1]);
            y+=2;
        }
    }
    url[x] = '\0';
} /* wd21_UnescapeUrl */


sapdbwa_Bool wd21_ReadBodyData( sapdbwa_HttpRequestP req )
{
    sapdbwa_Bool  readOK = true;
    sapdbwa_Int4  contentLength = atoi( sapdbwa_GetContentLength( req ));
        
    /* insert values from the body */
    sapdbwa_SQLALLOCAT( contentLength+1,
                   (sapdbwa_UInt1**) &req->bodyData,
                   &readOK );
    if (readOK) {
        sapdbwa_SQLALLOCAT( contentLength+1,
                       (sapdbwa_UInt1**) &req->paramBodyData,
                       &readOK );
        if (readOK) {
            sapdbwa_Int4 bytesRead = 0;
            
            bytesRead = req->ReadBody( req->serverHandle,
                                       req->bodyData,
                                       contentLength );
            readOK = (bytesRead == contentLength);
            if (readOK) {
                wd21_PlusToSpace( req->bodyData, bytesRead );
                wd21_UnescapeUrl( req->bodyData, bytesRead );
                req->bodyDataLen = contentLength;
                req->bodyDataPos = 0;
                memcpy( req->paramBodyData, req->bodyData, contentLength );
                /* we need a null termination, bacause paramBodyData is parsed */
                req->paramBodyData[contentLength] = '\0';
            }; /* if */
        }; /* if */
    }; /* if */

    return readOK;
} /* wd21_ReadBodyData */


void wd21_PlusToSpace( char *str, sapdbwa_Int4 len ) {
    sapdbwa_Int4 x;

    for( x=0; x<len; x++) if(str[x] == '+') str[x] = ' ';
} /* wd21_PlusToSpace */

/* stolen from Apache */
sapdbwa_Int4 wd21_Base64DecodeBinary( char       *bufplain,
                                    const char *bufcoded);

sapdbwa_Int4 wd21_Base64Decode( char       *bufplain,
                              const char *bufcoded)
{
    sapdbwa_Int4 len;
    
    len = wd21_Base64DecodeBinary( (char*) bufplain, bufcoded);

    return len;
} /* wd21_Base64Decode */


/* aaaack but it's fast and const should make it shared text page. */
static const unsigned char pr2six[256] =
{
    /* ASCII table */
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 62, 64, 64, 64, 63,
    52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 64, 64, 64, 64, 64, 64,
    64,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14,
    15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 64, 64, 64, 64, 64,
    64, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40,
    41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 64, 64, 64, 64, 64,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64
};

sapdbwa_Int4 wd21_Base64DecodeBinary( char       *bufplain,
                                    const char *bufcoded)
{
    sapdbwa_Int4           nbytesdecoded;
    register const char *bufin;
    register char       *bufout;
    register sapdbwa_Int4  nprbytes;

    bufin = (const char *) bufcoded;
    while (pr2six[*(bufin++)] <= 63);
    nprbytes = (bufin - (const char *) bufcoded) - 1;
    nbytesdecoded = ((nprbytes + 3) / 4) * 3;
    
    bufout = (char *) bufplain;
    bufin = (const char *) bufcoded;
    
    while (nprbytes > 4) {
        *(bufout++) =
            (unsigned char) (pr2six[*bufin] << 2 | pr2six[bufin[1]] >> 4);
        *(bufout++) =
            (unsigned char) (pr2six[bufin[1]] << 4 | pr2six[bufin[2]] >> 2);
        *(bufout++) =
            (unsigned char) (pr2six[bufin[2]] << 6 | pr2six[bufin[3]]);
        bufin += 4;
        nprbytes -= 4;
    }

    /* Note: (nprbytes == 1) would be an error, so just ingore that case */
    if (nprbytes > 1) {
        *(bufout++) =
            (unsigned char) (pr2six[*bufin] << 2 | pr2six[bufin[1]] >> 4);
    }
    if (nprbytes > 2) {
        *(bufout++) =
            (unsigned char) (pr2six[bufin[1]] << 4 | pr2six[bufin[2]] >> 2);
    }
    if (nprbytes > 3) {
        *(bufout++) =
	    (unsigned char) (pr2six[bufin[2]] << 6 | pr2six[bufin[3]]);
    }
    
    *(bufout++) = '\0';
    nbytesdecoded -= (4 - nprbytes) & 3;
    return nbytesdecoded;
} /* wd21_Base64DecodeBinary */
