/*!**********************************************************************

  module: vwd220ftp.c

  -----------------------------------------------------------------------

  responsible:	Holger Bischoff
				Markus zgen

  special area: WebAgent

  description:  FTP service

  see also:

  -----------------------------------------------------------------------

                         Copyright by SAP AG, 1999



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

************************************************************************/

/*=======================================================================
 *  INCLUDES
 *=====================================================================*/

#include <stdio.h>

#include "hwd203util.h"
#include "hwd220ftp.h"
#include "hwd221ftpcmd.h"

/*
 * Type codes
 */
const sapdbwa_Int4 TYPE_A = twd221_Token::TOK_A;   /* ASCII */
const sapdbwa_Int4 TYPE_E = twd221_Token::TOK_E;   /* EBCDIC */
const sapdbwa_Int4 TYPE_I = twd221_Token::TOK_I;   /* image */
const sapdbwa_Int4 TYPE_L = twd221_Token::TOK_L;   /* local byte size */

/*
 * Timeout intervals for retrying connections
 * to hosts that don't accept PORT cmds.  This
 * is a kludge, but given the problems with TCP...
 */
#define	SWAITMAX	90	/* wait at most 90 seconds */
#define	SWAITINT	5	/* interval between retries */

sapdbwa_Int4 wd220SWaitMax = SWAITMAX;
sapdbwa_Int4 wd220SWaitInt = SWAITINT;

// convert <carriage return><linefeed> to <linefeed>.
// outBuf must be at least as big as inBuf.
void wd220_CrLfToLf( char        *inBuf,
                     sapdbwa_Int4   inLen,
                     char        *outBuf,
                     sapdbwa_Int4  *outLen,
                     sapdbwa_Int4  *bareLineFeeds );

// convert <linefeed> to <carriage return><linefeed>.
// outBuf must at least have the double size as inBuf.
void wd220_LfToCrLf( char        *inBuf,
                     sapdbwa_Int4   inLen,
                     char        *outBuf,
                     sapdbwa_Int4  *outLen );

///////////////////////////////////
// public
///////////////////////////////////


const char *twd220_FtpService::ftpVersion = "V 0.0";

twd220_FtpService::twd220_FtpService( twd201_Socket      ctrlSocket,
                                      twd201_SockAddrIn  ctrlSockName,
                                      twd205_BasicFS    &fs,
                                      sapdbwa_Bool         debugMode )
    : debug(debugMode), loggedIn(sapdbwa_False), askPassword(sapdbwa_True), userName(""),
      dataSocket( INVALID_SOCKET ), pasvSocket( INVALID_SOCKET ),
      useDefault(sapdbwa_True),
      fs(fs),
      dataOutStream( NULL ),
      dataConnType( TYPE_A ),
      formCode( twd221_Token::FORM_N ),
      fromName(""),
      twd211_BasicService( ctrlSocket, ctrlSockName )
{
    this->dataSourceAddr.sin_family = 0;
    this->dataDestAddr.sin_family = 0;
}; // twd220_FtpService


twd220_FtpService::~twd220_FtpService()
{
}; // twd220_FtpService 

#define MAX_HOSTNAME_LEN 1024

sapdbwa_Int4 twd220_FtpService::Service()
{
    twd201_SockRet sockRet;
    char           hostname[MAX_HOSTNAME_LEN];
    char           addrBuf[1024];
#if 0
    sapdbwa_Int4     addrLen;
    addrLen = sizeof( this->sockName );
    sockRet = wd201GetPeername( this->socket, this->sockName, &addrLen );
    if (sockRet == SOCKET_ERROR) {
        if (this->debug) {
            syslog( LOG_ERR,  "getsockname (%s)","ftp");
        };
        return SOCKET_ERROR;
    }; // if
#endif
    if (this->debug) {
        wd203makeIPAddr( this->sockName, addrBuf, sizeof(addrBuf) );
        syslog( LOG_DEBUG, "remote: %s", addrBuf );
    }; // if

    sockRet = wd201GetHostname( hostname, sizeof( hostname ));

    reply( 220, "%s FTP server (%s) ready.", hostname, ftpVersion );

    this->StartSession();

    return 0;
}; // Service

///////////////////////////////////
// private
///////////////////////////////////

void twd220_FtpService::reply( int n, const char *fmt, ...)
{
    va_list  ap;
    char     localBuf[MAX_REPLY_LEN];

    va_start(ap, fmt);
	(void)sprintf( this->replyBuf, "%d ", n);
	(void)vsprintf( localBuf, fmt, ap);
    strcat( this->replyBuf, localBuf );
	strcat( this->replyBuf, "\r\n" );
    this->sendReply( this->replyBuf );

	if (this->debug) {
		syslog( LOG_DEBUG, "<--- %d ", n);
		vsyslog( LOG_DEBUG, fmt, ap);
	}; // if
}; // reply


void twd220_FtpService::lreply( int n, const char *fmt, ...)
{
    va_list  ap;
    char     localBuf[MAX_REPLY_LEN];

    va_start(ap, fmt);
	(void)sprintf( this->replyBuf, "%d- ", n);
	(void)vsprintf( localBuf, fmt, ap);
    strcat( this->replyBuf, localBuf );
	strcat( this->replyBuf, "\r\n" );
    this->sendReply( this->replyBuf );

	if (this->debug) {
		syslog( LOG_DEBUG, "<--- %d- ", n);
		vsyslog( LOG_DEBUG, fmt, ap);
	}; // if
}; // lreply


void twd220_FtpService::perror_reply( int n, const char *str, const char *msg )
{
    reply( n, "%s: %s.", str, msg );
}; // perror_reply


void twd220_FtpService::sendReply( const char *replyString )
{
    twd201_SockRet sockRet;

    sockRet = wd201Send( this->socket, replyString, strlen(replyString), 0 );
}; // sendReply


void twd220_FtpService::acknowledge( const char *cmd )
{
    reply(250, "%s command successful.", cmd);
}; // ackknowledge


void twd220_FtpService::StartSession()
{
    twd221_FtpParser ftpParser( this->socket, this->debug );
    sapdbwa_Bool       endSession = sapdbwa_False;
    twd221_FtpCmd    ftpCmd;
    sapdbwa_Bool       timeout = sapdbwa_False;

    while (!endSession) {
        if (!ftpParser.GetNextCmd( ftpCmd, timeout )) {
            // error
            if (this->debug) {
                syslog( LOG_DEBUG, "internal error!!!\n", 0 );
            };
            endSession = sapdbwa_True;
        } else if (timeout) {
            reply( 421,
                   "Timeout (%d seconds): closing control connection.", 300 );
            endSession = sapdbwa_True;
            this->DoLogout(0);
        } else {
            // handle command
            switch (ftpCmd.CmdType()) {
            case twd221_Token::USER :
                {
                    this->User( ftpCmd.Arg1() );
                    break;
                }
            case twd221_Token::PASS :
                {
                    this->Pass( ftpCmd.Arg1() );
                    break;
                }
            case twd221_Token::PORT :
                {
                    this->SetRemoteDataPort( ftpCmd.SockAddrArg() );
                    reply(200, "PORT command successful.");
                    break;
                }
            case twd221_Token::QUIT :
                {
                    reply( 221, "Goodbye.");
                    endSession = sapdbwa_True;
                    this->DoLogout(0);
                    break;
                }
            case twd221_Token::SYST :
                {
                    reply( 215, "UNIX" );
                    break;
                }
            case twd221_Token::NOOP :
                {
                    reply(200, "NOOP command successful.");
                    break;
                }
            case twd221_Token::PWD :
                {
                    if (this->checkLogin()) {
                        this->Pwd();
                    }; // if
                    break;
                }
            case twd221_Token::CWD :
                {
                    if (this->checkLogin()) {
                        this->Cwd( ftpCmd.Arg1() );
                    }; // if
                    break;
                }
            case twd221_Token::CDUP :
                {
                    if (this->checkLogin()) {
                        this->CdUp();
                    }; // if
                    break;
                }
            case twd221_Token::RMD :
                {
                    if (this->checkLogin()) {
                        this->RemoveDir( ftpCmd.Arg1() );
                    }; // if
                    break;
                }
            case twd221_Token::MKD :
                {
                    if (this->checkLogin()) {
                        this->MakeDir( ftpCmd.Arg1() );
                    }; // if
                    break;
                }
            case twd221_Token::PASV :
                {
                    this->PassiveMode();
                    break;
                }
            case twd221_Token::LIST :
                {
                    if (this->checkLogin()) {
                        ZString dir = ftpCmd.Arg1();
                        if (dir[0] == '-') {
                            dir = ""; // ignore arguments
                        }; // if
                        this->ListDir( dir, sapdbwa_True );
                    }; // if
                    break;
                }
            case twd221_Token::NLST :
                {
                    if (this->checkLogin()) {
                        this->ListDir( ftpCmd.Arg1(), sapdbwa_False );
                    }; // if
                    break;
                }
            case twd221_Token::STAT :
                {
                    const ZString path = ftpCmd.Arg1();

                    if (path != "") {
                        if (this->checkLogin()) {
                            this->StatFileCmd( ftpCmd.Arg1() );
                        }; // if
                    } else {
                        this->StatCmd();
                    }; // else
                    break;
                }
            case twd221_Token::DELE :
                {
                    if (this->checkLogin()) {
                        this->Delete( ftpCmd.Arg1() );
                    }; // if
                    break;
                }
            case twd221_Token::ABOR :
                {
                    reply(225, "ABOR command successful.");
                    break;
                }
            case twd221_Token::RETR :
                {
                    if (this->checkLogin()) {
                        this->Retrieve( ftpCmd.Arg1() );
                    }; // if
                    break;
                }
            case twd221_Token::STOR :
                {
                    if (this->checkLogin()) {
                        this->Store( ftpCmd.Arg1() );
                    }; // if
                    break;
                }
            case twd221_Token::MODE :
                {
                    sapdbwa_Int4 modeCode = ftpCmd.CodeArg1();

                    this->SetModeCode( modeCode );
                    break;
                }
            case twd221_Token::STRU :
                {
                    sapdbwa_Int4 structCode = ftpCmd.CodeArg1();

                    this->SetStructCode( structCode );
                    break;
                }
            case twd221_Token::TYPE :
                {
                    sapdbwa_Int4 type = ftpCmd.CodeArg1();
                    sapdbwa_Int4 form = ftpCmd.CodeArg2();

                    this->SetTypeCode( type, form );
                    break;
                }
            case twd221_Token::RNFR :
                {
                    this->RenameFrom( ftpCmd.Arg1() );
                    break;
                }
            case twd221_Token::RNTO :
                {
                    this->RenameTo( ftpCmd.Arg1() );
                    break;
                }
            case twd221_Token::SIZE :
                {
                    syslog( LOG_DEBUG, "SIZE\n" );
                    this->Size( ftpCmd.Arg1() );
                    break;
                }
            case twd221_Token::LEXERR :
                {
                    if (this->debug) {
                        syslog( LOG_DEBUG, "Parse error\n" , 0);
                    };
                    ZString command = ftpCmd.Arg1();
                    // we have to strip off '\r\n'
                    command[command.GetSize()-1] = '\0';
                    reply(500, "'%s': command not understood.",
                          (char*) command );
                    break;
                }
            case twd221_Token::ERR :
                {
                    // some fatal error occured session ends
                    endSession = sapdbwa_True;
                    this->DoLogout(0);
                    break;
                }
            default:
                {
                    if (this->debug) {
                        syslog( LOG_DEBUG, "Parse error\n", 0);
                    };
                    reply( 500, "Command not understood." );
                    break;
                }
            }; // case
        }; // else
    }; // while

}; // StartSession


sapdbwa_Bool twd220_FtpService::checkLogin()
{
    if (!this->loggedIn) {
        reply(530, "Please login with USER and PASS.");
    }; // if

    return this->loggedIn;
}; // checkLogin


sapdbwa_Bool twd220_FtpService::GetDataSock()
{
    twd201_SockRet sockRet;
    twd201_Socket  newDataSocket = INVALID_SOCKET;
    sapdbwa_Bool     ok            = sapdbwa_True;
    sapdbwa_Int4     tries; 

    if (this->dataSocket == INVALID_SOCKET) {
        // create new socket
        newDataSocket = wd201Socket( AF_INET, SOCK_STREAM, 0);
        if (newDataSocket != INVALID_SOCKET) {
            int on = 1;
            sockRet = wd201SetSockOpt( newDataSocket,
                                       SOL_SOCKET, SO_REUSEADDR,
                                       (char*) &on, sizeof(on) );
            if (sockRet == SOCKET_ERROR) {
                ok = sapdbwa_False;
            }
        } else {
            ok = sapdbwa_False;
        };

        /* anchor socket to avoid multi-homing problems */
        this->dataSourceAddr.sin_family = AF_INET;
        /* use INADDR_ANY (Quinn p. 543) (PTS 1106196) */
        this->dataSourceAddr.sin_addr.s_addr = INADDR_ANY; //this->sockName.sin_addr;

        for (tries = 1; (ok == (sapdbwa_UInt1) sapdbwa_True); tries++) {
            sockRet = wd201Bind( newDataSocket,
                                 (twd201_SockAddr*) &this->dataSourceAddr,
                                 sizeof( this->dataSourceAddr ));
            if (sockRet == 0) {
                break;
            } else {
                if (wd201GetLastError() != EADDRINUSE_WD201 || tries > 10) {
                    ok = sapdbwa_False;
                } else {
                    // wait before next try
                    wd203Sleep( tries );
                }; // else
            }; // else
        }; // for
        if (ok) {
            this->dataSocket = newDataSocket;
        }; // if
    }; // if

    return ok;
}; // GetDataSock


twd201_Socket* twd220_FtpService::OpenDataConn( const ZString &name )
{
    sapdbwa_Bool     openOK = sapdbwa_True;
    twd201_Socket *dataSockPtr = NULL;

    if (this->pasvSocket != INVALID_SOCKET) {
        // we are in passive mode
        twd201_SockAddrIn incomingAddr;
        twd201_Socket     incomingSocket = INVALID_SOCKET;

        openOK = wd203Accept( this->pasvSocket,
                              &incomingSocket,
                              &incomingAddr );
        if (!openOK) {
            openOK = sapdbwa_False;
			reply( 425, "Can't open data connection.");
            this->pasvSocket = INVALID_SOCKET;
        } else {
            this->pasvSocket = incomingSocket;
            reply( 150, "Opening %s mode data connection for '%s'.",
                   dataConnType == TYPE_A ? "ASCII" : "BINARY", (char*) name );
        }; // else
        dataSockPtr = (openOK) ? &this->pasvSocket : NULL;
    } else {
        if (this->dataSocket != INVALID_SOCKET) {
            reply( 125, "Using existing data connection for '%s'.",
                   (char*) name );
            this->useDefault = sapdbwa_True;
        } else {
            if (this->useDefault) {
                this->dataDestAddr = this->sockName;
            }; // if
            this->useDefault = sapdbwa_True;
            // try to build up new data socket
            if (this->GetDataSock()) {
                sapdbwa_Int4 retry = 0;

                while (wd201Connect( this->dataSocket,
                                     (twd201_SockAddr*) &this->dataDestAddr,
                                     SOCKADDR_LEN ) < 0) {
                    if ( wd201GetLastError() == EADDRINUSE_WD201
                         && retry < wd220SWaitMax) {
                        wd203Sleep( (sapdbwa_UInt4) wd220SWaitInt );
                        retry += wd220SWaitInt;
                    } else {
                        perror_reply( 425, "Can't build data connection", "" );
                        this->CloseDataConn( &this->dataSocket );
                        openOK = sapdbwa_False;
                    }; // else
                }; // while
                if (openOK) {
                    reply( 150, "Opening %s mode data connection for '%s'.",
                           dataConnType == TYPE_A ? "ASCII" : "BINARY",
                           (char*) name );
                }; // if
            } else {
                openOK = sapdbwa_False;
                reply( 425, "Can't create data socket (%s,%d): %s.",
                       inet_ntoa( this->dataSourceAddr.sin_addr ),
                       ntohs( this->dataSourceAddr.sin_port ),
                       (char*) wd201StrError( wd201GetLastError() ));
            }; // else
        }; // else
        dataSockPtr = (openOK) ? &this->dataSocket : NULL;
    }; // else
    
    return dataSockPtr;
}; // OpenDataConn


void twd220_FtpService::CloseDataConn( twd201_Socket *dataSockPtr )
{
    twd201_SockRet sockRet;
    
    if (*dataSockPtr != INVALID_SOCKET) {
        sockRet = wd201CloseSocket( *dataSockPtr );
        *dataSockPtr = INVALID_SOCKET;
        if (sockRet != 0) {
            wd203SockErr( "CloseDataConn", "CloseSocket" );
        } else {
            if (this->debug) {
                syslog( LOG_DEBUG, "CloseDataConn ok\n", 0 );
            };
        }; // else
    }; // if
}; // CloseDataConn

#define UC(b) (((int) b) & 0xff)

void twd220_FtpService::SetRemoteDataPort
( const twd201_SockAddrIn &dataSockAddr )
{
    this->useDefault = sapdbwa_False;
    if (this->pasvSocket != INVALID_SOCKET) {
        // close open passive connection
        this->CloseDataConn( &this->pasvSocket );
    }; // if

    this->dataDestAddr = dataSockAddr;
    this->dataSocket = INVALID_SOCKET;
}; // SetRemoteDataPort

#define SEND_DATA_BUFSIZE 8192

void  twd220_FtpService::SendData( twd201_Socket             dataSock,
                                   twd207_BasicInputStream  &fromStream )
{
    char           buf[SEND_DATA_BUFSIZE];
    char           convBuf[2*SEND_DATA_BUFSIZE]; /* if buf contains only \n
                                                  * we need the double size
                                                  * for conversion to \r\n */
    sapdbwa_Int4     bytesRead = 0;
    sapdbwa_Int4     bytesConverted = 0;
    twd201_SockRet sockRet = 0;

    switch (this->dataConnType) {
    case TYPE_I:
    case TYPE_L:
        // raw binary data
        do {
            bytesRead = fromStream.Read( buf, sizeof(buf) );
            if (bytesRead > 0) {
                sockRet = wd201Send( dataSock, buf, bytesRead, 0 );
            }; // if
        } while ((bytesRead > 0) && (sockRet >= 0));

        if (sockRet < 0) {
            perror_reply( 426, "Data connection",
                          (const char*) wd201StrError( wd201GetLastError()) );
        } else if (bytesRead < 0) {
            perror_reply( 451, "File system",
                          (char*) fromStream.GetLastErrorMsg() );
        } else {
            reply(226, "Transfer complete.");
        }; // else
        break;
    case TYPE_A:
        // ascii data
        do {
            bytesRead = fromStream.Read( buf, sizeof(buf) );
            if (bytesRead > 0) {
                wd220_LfToCrLf( buf, bytesRead, convBuf, &bytesConverted );
                sockRet = wd201Send( dataSock, convBuf, bytesConverted, 0 );
            }; // if
        } while ((bytesRead > 0) && (sockRet >= 0));

        if (sockRet < 0) {
            perror_reply( 426, "Data connection",
                          (const char*) wd201StrError( wd201GetLastError()) );
        } else if (bytesRead < 0) {
            perror_reply( 451, "File system",
                          (char*) fromStream.GetLastErrorMsg() );
        } else {
            reply(226, "Transfer complete.");
        }; // else
        break;
    default:
        // unknown type
		reply( 550, "Unimplemented TYPE %d in send_data",
               this->dataConnType);
        break;
    }; /* switch */
}; // SendData

#define RECEIVE_DATA_BUFSIZE 8192

void twd220_FtpService::ReceiveData( twd201_Socket              dataSock,
                                     twd207_BasicOutputStream  &toStream )
{
    char           buf[RECEIVE_DATA_BUFSIZE];
    char           convBuf[RECEIVE_DATA_BUFSIZE];// buffer for ASCII conversion
    sapdbwa_Int4     bytesConverted = 0; // number of bytes in convBuf
    sapdbwa_Int4     bareLineFeeds  = 0; // counter for bare '\n's
    sapdbwa_Int4     bytesReceived = 0;
    sapdbwa_Int4     bytesWritten  = 0;

    switch (this->dataConnType) {
    case TYPE_I:
    case TYPE_L:
        // raw binary data
        do {
            bytesReceived = wd201Recv( dataSock, buf, sizeof(buf), 0 );
            if (bytesReceived > 0) {
                bytesWritten = toStream.Write( buf, bytesReceived );
            }; // if
        } while (bytesReceived > 0 && bytesWritten > 0);
        if (bytesReceived < 0) {
            perror_reply( 550, "Data connection",
                          (const char*) wd201StrError( wd201GetLastError()) );
        } else if (bytesWritten < 0) {
            perror_reply( 550, "Data connection",
                          (char*) toStream.GetLastErrorMsg() );
        } else {
            reply(226, "Transfer complete.");
        }; // else
        break;
    case TYPE_E:
        reply(553, "TYPE E not implemented.");
        break;
    case TYPE_A:
        // ascii data
        do {
            bytesReceived = wd201Recv( dataSock, buf, sizeof(buf), 0 );
            if (bytesReceived > 0) {
                wd220_CrLfToLf( buf, bytesReceived,
                                convBuf, &bytesConverted,
                                &bareLineFeeds );
                bytesWritten = toStream.Write( convBuf, bytesConverted );
            }; // if
        } while (bytesReceived > 0 && bytesWritten > 0);
        if (bytesReceived < 0) {
            perror_reply( 550, "Data connection",
                          (const char*) wd201StrError( wd201GetLastError()) );
        } else if (bytesWritten < 0) {
            perror_reply( 550, "Data connection",
                          (char*) toStream.GetLastErrorMsg() );
        } else {
            if (bareLineFeeds > 0) {
                lreply( 226,
                        "WARNING! %d bare linefeeds received in ASCII mode",
                        bareLineFeeds );
            }; // if
            reply(226, "Transfer complete.");
        }; // else
        break;
    default:
        // unknown type
		reply(550, "Unimplemented TYPE %d in receive_data",
              this->dataConnType);
        break;
    }; // switch
}; // ReceiveData


void twd220_FtpService::CdUp()
{
    if (!this->fs.CdUp()) {
        perror_reply(550, "", (char*) this->fs.GetLastErrorMsg() );
    } else {
        this->acknowledge( "CDUP" );
    }; // else
}; // Cwd


void twd220_FtpService::Cwd( const ZString &path )
{
    if (!this->fs.ChDir( path )) {
        perror_reply(550, (char*) path, (char*) this->fs.GetLastErrorMsg() );
    } else {
        this->acknowledge( "CWD" );
    }; // else
}; // Cwd


void twd220_FtpService::Delete( const ZString &path )
{
    if (!this->fs.Delete( path )) {
        perror_reply(550, (char*) path, (char*) this->fs.GetLastErrorMsg() );
    } else {
        this->acknowledge( "DELE" );
    }; // else
}; // Delete


void twd220_FtpService::DoLogout( sapdbwa_Int4 status )
{
    this->loggedIn    = sapdbwa_False;
    this->askPassword = sapdbwa_True;
    // ...
}; // DoLogout


void twd220_FtpService::PassiveMode()
{
    twd201_SockRet sockRet = SOCKET_ERROR;    
    sapdbwa_Bool openConnOK  = sapdbwa_True;

    this->pasvSocket = wd201Socket( AF_INET, SOCK_STREAM, 0 );
    if (this->pasvSocket == INVALID_SOCKET) {
        openConnOK = sapdbwa_False;
    } else {
        this->pasvSockAddr = this->sockName;
        this->pasvSockAddr.sin_port = 0;
        sockRet = wd201Bind( this->pasvSocket,
                             (twd201_SockAddr*) &this->pasvSockAddr,
                             SOCKADDR_LEN );
        if (sockRet < 0) {
            openConnOK = sapdbwa_False;
        } else {
            sapdbwa_Int4 addrLen = sizeof( this->pasvSockAddr );
            sockRet = wd201GetSockName( this->pasvSocket,
                                        (twd201_SockAddr*) &this->pasvSockAddr,
                                        &addrLen );
            if (sockRet < 0) {
                openConnOK = sapdbwa_False;
            } else {
                sockRet = wd201Listen( this->pasvSocket, 1 );
                if (sockRet < 0) {
                    openConnOK = sapdbwa_False;
                }; // if
            }; // else
        }; // else
    }; // else

    if (openConnOK) {
        char *a = (char *) &this->pasvSockAddr.sin_addr;
        char *p = (char *) &this->pasvSockAddr.sin_port;
        reply( 227, "Entering Passive Mode (%d,%d,%d,%d,%d,%d)",
               UC(a[0]), UC(a[1]), UC(a[2]), UC(a[3]),
               UC(p[0]), UC(p[1]));
    } else {
		perror_reply( 425, "Can't open passive connection",
                      (const char*) wd201StrError( wd201GetLastError() ));
    }; // if
}; // PassiveMode


void twd220_FtpService::ListDir( const ZString &path, sapdbwa_Bool details )
{
    twd207_BasicInputStream *istream = this->fs.GetDirStream( path, details );
    twd201_Socket           *dataSockPtr = NULL;

    if (istream == NULL) {
		perror_reply( 551, "Error on input file.",
                      (const char*) this->fs.GetLastErrorMsg() );
    } else {
        dataSockPtr = this->OpenDataConn( path );
        if (dataSockPtr != NULL) {
            this->SendData( *dataSockPtr, *istream );
            istream->Close();
            this->CloseDataConn( dataSockPtr );
        }; // if
    }; // else
}; // ListDir


void  twd220_FtpService::MakeDir( const ZString &newDir )
{
    if (!this->fs.MkDir( newDir )) {
        perror_reply(550, (char*) newDir, (char*) this->fs.GetLastErrorMsg() );
    } else {
        this->acknowledge( "MKD" );
    }; // else
}; // MakeDir


void twd220_FtpService::Pwd()
{
    ZString path;
    
    if (this->fs.GetDir( path )) {
        reply( 257, "\"%s\" is current directory.",
               (char*) path);
    } else {
        perror_reply( 550, "Can't get path.",
                      this->fs.GetLastErrorMsg());
    }; // else
}; // Pwd


void twd220_FtpService::RemoveDir( const ZString &dir )
{
    if (!this->fs.RmDir( dir )) {
        perror_reply(550, (char*) dir, (char*) this->fs.GetLastErrorMsg() );
    } else {
        this->acknowledge( "RMD" );
    }; // else
}; // RemoveDir


void twd220_FtpService::RenameFrom( const ZString &oldName )
{
    sapdbwa_Bool pathExists = sapdbwa_False;

    if (!this->fs.CheckPath( oldName, pathExists )) {
        this->fromName = "";
        perror_reply( 550, oldName, (char*) this->fs.GetLastErrorMsg() );
    } else if (!pathExists) {
        this->fromName = "";
        perror_reply( 550, oldName, "No such file or directory." );
    } else {
        this->fromName = oldName;
        reply(350, "File exists, ready for destination name");
    }; // else
}; // RenameFrom


void twd220_FtpService::RenameTo( const ZString &newName )
{
    if (!this->fs.RenameFromTo( this->fromName, newName )) {
        perror_reply( 550, "rename", (const char*) this->fs.GetLastErrorMsg() );
    } else {
        this->acknowledge("RNTO");
    }; // else
}; // RenameTo


void twd220_FtpService::Retrieve( const ZString &path )
{
    twd207_BasicInputStream *istream = this->fs.GetFileInStream( path );
    twd201_Socket           *dataSockPtr = NULL;

    if (istream == NULL) {
		perror_reply( 550, "",
                      (const char*) this->fs.GetLastErrorMsg() );
    } else {
        dataSockPtr = this->OpenDataConn( path );
        if (dataSockPtr != NULL) {
            this->SendData( *dataSockPtr, *istream );
            this->CloseDataConn( dataSockPtr );
        }; // if
        if (!istream->Close()) {
            perror_reply( 550, "istream->Close",
                          (const char*) istream->GetLastErrorMsg() );
        }; // if
    }; // else
}; // Retrieve


void twd220_FtpService::Size( const ZString &path )
{
    sapdbwa_Int4  size = 0;

    switch (this->dataConnType) {
    case TYPE_I:
    case TYPE_L:
        // raw binary data
        if (this->fs.GetSize( path, size )) {
            reply( 213, "%d", size );
        } else {
            // not found or other error
            perror_reply( 550, "",
                          (const char*) this->fs.GetLastErrorMsg() );
        };
        break;
    case TYPE_A:
        // ascii data
        if (this->fs.GetSize( path, size )) {
            sapdbwa_Int4  bytesRead = 0;
            sapdbwa_Int4  linefeeds = 0;
            sapdbwa_Int4  i = 0;
            char        buf[SEND_DATA_BUFSIZE];

            twd207_BasicInputStream *istream =
                this->fs.GetFileInStream( path );

            if (istream == NULL) {
                perror_reply( 550, "",
                              (const char*) this->fs.GetLastErrorMsg() );
            } else {
                // count linefeeds to get real transfer size
                do {
                    bytesRead = istream->Read( buf, sizeof(buf) );
                    for ( i = 0; i < bytesRead; i++ ) {
                        if (buf[i] == '\n' && (i == 0 || buf[i-1] != '\r')) {
                            linefeeds++;
                        };
                    }; // for
                } while (bytesRead > 0);
                reply( 213, "%d", size + linefeeds );
            }; // else
        } else {
            // not found or other error
            perror_reply( 550, "",
                          (const char*) this->fs.GetLastErrorMsg() );
        };
        break;
    default:
        // unknown type
		reply(550, "Unimplemented TYPE %d in Size",
              this->dataConnType);
        break;
    }; // switch
}; // Size


/*
#define LINE_MAX 2048
*/

void twd220_FtpService::StatCmd()
{
    char           hostname[MAX_HOSTNAME_LEN];
    twd201_SockRet sockRet;
    char           localBuf[MAX_REPLY_LEN];

    hostname[0] = '\0';
    sockRet = wd201GetHostname( hostname, sizeof( hostname ));

	lreply(211, "%s FTP server status:", hostname);
    if (this->loggedIn) {
        sprintf( localBuf, "     Logged in as %s\r\n",
                 (char*) this->userName );
    } else if (this->askPassword) {
		sprintf( localBuf, "     Waiting for password\r\n");        
    } else {
        sprintf( localBuf, "     Waiting for user name\r\n");
    }; // else
    this->sendReply( localBuf );
    // ...

	reply(211, "End of status");
}; // StatCmd


void twd220_FtpService::StatFileCmd( const ZString &path )
{
    ZString  statStr;

    if (this->fs.GetStatStr( path, statStr )) {
        lreply( 211,
                "status of %s: %s",
                (const char*) path,
                (const char*) statStr );
    } else {
		perror_reply( 551, "",
                      (const char*) this->fs.GetLastErrorMsg() );
    }; // else
}; // StatFileCmd


void twd220_FtpService::Store( const ZString &path )
{
    twd207_BasicOutputStream *ostream = this->fs.GetFileOutStream( path );
    twd201_Socket            *dataSockPtr = NULL;

    if (ostream == NULL) {
		perror_reply( 553, "",
                      (const char*) this->fs.GetLastErrorMsg() );
    } else {
        dataSockPtr = this->OpenDataConn( path );
        if (dataSockPtr != NULL) {
            this->ReceiveData( *dataSockPtr, *ostream );
            this->CloseDataConn( dataSockPtr );
        }; // if
        if (!ostream->Close()) {
            perror_reply( 553, "ostream->Close",
                          (const char*) ostream->GetLastErrorMsg() );
        }; // if            
    }; // else
}; // Store


void twd220_FtpService::User( const ZString &user )
{
    if (user == "wa") {
        reply(331, "Password required for %s.", (char*) user );
        this->askPassword = sapdbwa_True;
        this->userName    = user;
    } else if (user == "anonymous" ) {
        reply( 230, "Guest login ok." );
        this->askPassword = sapdbwa_False;
        this->loggedIn    = sapdbwa_True;
        this->userName    = user;
    } else {
        reply( 530, "Access denied." );
    }; // else
}; // User


void twd220_FtpService::Pass( const ZString &passwd )
{
    if (this->loggedIn || !this->askPassword) {
        reply(503, "Login with USER first.");
		return;
	}
	this->askPassword = sapdbwa_False;
    this->loggedIn    = sapdbwa_True;

    reply(230, "Guest login ok, access restrictions apply.");
	//this->CompleteLogin( passwd );
}; // Pass


void twd220_FtpService::SetTypeCode( sapdbwa_Int4 type,
                                     sapdbwa_Int4 form )
{
    switch (type) {
        
    case TYPE_A:
        if (form == twd221_Token::FORM_N) {
            reply(200, "Type set to A.");
            this->dataConnType = type;
            this->formCode     = form;
        } else
            reply(504, "Form must be N.");
        break;

    case TYPE_E:
        reply(504, "Type E not implemented.");
        break;
        
    case TYPE_I:
        reply(200, "Type set to I.");
        this->dataConnType = type;
				break;
                
    case TYPE_L:
        reply(504, "Type L not implemented.");
        break;
#if 0
#if defined (NBBY) && NBBY == 8
        if (cmd_bytesz == 8) {
            reply(200,
                  "Type set to L (byte size 8).");
            type = cmd_type;
        } else
            reply(504, "Byte size must be 8.");
#else /* NBBY == 8 */
        UNIMPLEMENTED for NBBY != 8;
#endif /* NBBY == 8 */
#endif
    default:
        perror_reply( 451, "Internal error", "Wrong typeCode" );
        break;
    }; // switch
    
}; // SetTypeCode


void twd220_FtpService::SetModeCode( sapdbwa_Int4 modeCode )
{
    switch (modeCode) {

    case twd221_Token::TOK_S:
        reply(200, "MODE S ok.");
        break;

    default:
        reply(502, "Unimplemented MODE type.");
    }; // switch
}; // SetModeCode


void twd220_FtpService::SetStructCode( sapdbwa_Int4 structCode )
{
    switch (structCode) {

    case twd221_Token::TOK_F:
        reply(200, "STRU F ok.");
        break;
        
    default:
        reply(504, "Unimplemented STRU type.");
    }; // switch
}; // SetStructCode


void wd220_CrLfToLf( char        *inBuf,
                     sapdbwa_Int4   inLen,
                     char        *outBuf,
                     sapdbwa_Int4  *outLen,
                     sapdbwa_Int4  *bareLineFeeds )
{
    sapdbwa_Int4 inPos = 0;
    sapdbwa_Int4 outPos = 0;
    char       ch;

    while (inPos < inLen) {
        ch = inBuf[inPos];
        inPos++;
        switch (ch) {
        case '\n':
            *bareLineFeeds++;
            break;
        case '\r':
            if ((inPos < inLen) && (inBuf[inPos] == '\n')) {
                // write '\n' instead of '\r\n'
                ch = '\n';
                inPos++; // skip '\n'
            }; // if
            break;
        default:
            break;
        }; // switch
        outBuf[outPos++] = ch;
    }; // while

    *outLen = outPos;
}; // wd220_CrLfToLf


// convert <linefeed> to <carriage return><linefeed>.
// outBuf must at least have the double size as inBuf.
void wd220_LfToCrLf( char        *inBuf,
                     sapdbwa_Int4   inLen,
                     char        *outBuf,
                     sapdbwa_Int4  *outLen )
{
    sapdbwa_Int4 inPos = 0;
    sapdbwa_Int4 outPos = 0;
    char       ch;

    while (inPos < inLen) {
        ch = inBuf[inPos];
        if (ch == '\n' && (inPos == 0 || inBuf[inPos-1] != '\r')) {
            // write '\r\n' instead of '\n'
            outBuf[outPos++] = '\r';
        }; // if
        outBuf[outPos++] = ch;
        inPos++;
    }; // while

    *outLen = outPos;
}; /* wd220_LfToCrLf */
