/*!**********************************************************************

  module: hwd221ftpcmd.h

  -----------------------------------------------------------------------

  responsible:	Holger Bischoff
				Markus zgen

  special area: WebAgent

  description:  Parser for ftp commands

  see also:

  -----------------------------------------------------------------------

                         Copyright by SAP AG, 1999



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

************************************************************************/

/*=======================================================================
 *  INCLUDES
 *=====================================================================*/

#include <stdio.h>
#include <ctype.h>    /* isdigit */
#include <stdlib.h>   /* atoi    */

#include "hwd203util.h"
#include "hwd221ftpcmd.h"

sapdbwa_Bool wd221_ContainsChar( char       *buf,
                               sapdbwa_Int4  from,
                               sapdbwa_Int4  to,
                               char        c );

/*!**********************************************************************

  Chapter: twd221_FtpCmd

************************************************************************/

twd221_FtpCmd::twd221_FtpCmd( twd221_FtpCmdType        iCmdType,
                              const ZString           &iArg1,
                              const ZString           &iArg2,
                              const twd201_SockAddrIn *iSockAddrArg,
                              const sapdbwa_Int4         iCodeArg1,
                              const sapdbwa_Int4         iCodeArg2 )
{
    this->Init( iCmdType, iArg1, iArg2, iSockAddrArg, iCodeArg1, iCodeArg2 );
}; // twd221_FtpCmd


twd221_FtpCmd::~twd221_FtpCmd() {
}; // ~twd221_FtpCmd 


void twd221_FtpCmd::Init( twd221_FtpCmdType        iCmdType,
                          const ZString           &iArg1,
                          const ZString           &iArg2,
                          const twd201_SockAddrIn *iSockAddrArg,
                          const sapdbwa_Int4         iCodeArg1,
                          const sapdbwa_Int4         iCodeArg2 )
{
    this->cmdType = iCmdType;
    this->arg1 = iArg1;
    this->arg2 = iArg2;
    if (iSockAddrArg != NULL) { 
        this->sockAddrArg = *iSockAddrArg;
    }; // if
    this->codeArg1 = iCodeArg1;
    this->codeArg2 = iCodeArg2;
}; // Init


/*!**********************************************************************

  EndChapter: twd221_FtpCmd

************************************************************************/

/*!**********************************************************************

  Chapter: twd221_FtpParser
  
************************************************************************/

// scanner states
#define	CMD     0	// beginning of command
#define	ARGS	1	// expect miscellaneous arguments
#define	STR1	2	// expect SP followed by STRING
#define	STR2	3	// expect STRING
#define	OSTR	4	// optional SP then STRING
#define	ZSTR1	5	// SP then optional STRING
#define	ZSTR2	6	// optional STRING after SP
#define	SITECMD	7	// SITE command
#define	NSTR	8	// Number followed by a string

#define CR  '\r'

#define SELECT_TIMEOUT 300 // in seconds

struct ParserCmdEntry {
    const char *name;
    sapdbwa_Int2  token;
    sapdbwa_Int2  state;
    sapdbwa_Bool  implemented; // sapdbwa_True if command is implemented
    const char *help;        // help text
}; // struct ParserCmdEntry


///////////////////////////////////
// public
///////////////////////////////////

sapdbwa_Bool twd221_FtpParser::GetNextCmd( twd221_FtpCmd  &ftpCmd,
                                         sapdbwa_Bool     &timeoutOccured  )
{
    sapdbwa_Int4         tokenType;
    twd221_FtpCmdType  cmdType = twd221_Token::ERR;
    ZString            arg1 = "";
    ZString            arg2 = "";
    twd201_SockAddrIn  sockAddr;
    sapdbwa_Int4         code1 = -1;
    sapdbwa_Int4         code2 = -1;

    this->NewCmd();
    tokenType = this->GetNextToken( arg1 );
    cmdType = tokenType; // may be overidden
    switch (tokenType) {
    case twd221_Token::USER: // USER <sp> <username>
        tokenType = this->GetNextToken( arg1 );
        if (tokenType != twd221_Token::STRING) {
            cmdType = twd221_Token::LEXERR;
        }; // if
        break;
    case twd221_Token::PASS: // PASS <sp> <password>
        tokenType = this->GetNextToken( arg1 );
        break;
    case twd221_Token::PORT: // PORT <sp> b0,b1,b2,b3,b4,b5
        if (!this->GetPort( sockAddr )) {
            cmdType = twd221_Token::LEXERR;
        }; // if            
        break;
    case twd221_Token::PASV: // PASV
        break;
    case twd221_Token::TYPE: // TYPE <sp> <type-code>
        if (!this->GetTypeCode( code1, code2 )) {
            cmdType = twd221_Token::LEXERR;
        }; // if
        break;
    case twd221_Token::STRU: // STRU <sp> <struct-code>
        if (!this->GetStructCode( code1 )) {
            cmdType = twd221_Token::LEXERR;
        }; // if
        break;
    case twd221_Token::MODE: // MODE <sp> <mode-code>
        if (!this->GetModeCode( code1 )) {
            cmdType = twd221_Token::LEXERR;
        }; // else
        break;
    case twd221_Token::ALLO: // ALLO ...
        // ignore arguments
        this->SkipCmd();
        break;
    case twd221_Token::RETR: // RETR <sp> <pathname>
        tokenType = this->GetNextToken( arg1 );
        if (tokenType != twd221_Token::STRING) {
            cmdType = twd221_Token::LEXERR;
        }; // if
        break;
    case twd221_Token::STOR: // STOR <sp> <pathname>
        tokenType = this->GetNextToken( arg1 );
        if (tokenType != twd221_Token::STRING) {
            cmdType = twd221_Token::LEXERR;
        }; // if
        break;
    case twd221_Token::APPE: // APPE <sp> <pathname>
        tokenType = this->GetNextToken( arg1 );
        if (tokenType != twd221_Token::STRING) {
            cmdType = twd221_Token::LEXERR;
        }; // if
        break;
    case twd221_Token::NLST: // NLST <sp> [<pathname>]
        // get optional pathname
        tokenType = this->GetNextToken( arg1 );
        break;
    case twd221_Token::LIST: // LIST <sp> [<pathname>]
        // get optional pathname
        tokenType = this->GetNextToken( arg1 );
        if (tokenType != twd221_Token::TOK_CRLF
            && tokenType != twd221_Token::STRING) {
            cmdType = twd221_Token::LEXERR;
        }; // if
        break;
    case twd221_Token::STAT: // STAT <sp> [<pathname>]
        // get optional pathname
        tokenType = this->GetNextToken( arg1 );
        if (tokenType != twd221_Token::TOK_CRLF
            && tokenType != twd221_Token::STRING) {
            cmdType = twd221_Token::LEXERR;
        }; // if
        break;
    case twd221_Token::DELE: // DELE <sp> [<pathname>]
        // get optional pathname
        tokenType = this->GetNextToken( arg1 );
        if (tokenType != twd221_Token::TOK_CRLF
            && tokenType != twd221_Token::STRING) {
            cmdType = twd221_Token::LEXERR;
        }; // if
        break;
    case twd221_Token::RNTO: // RNTO <sp> <pathname>
        tokenType = this->GetNextToken( arg1 );
        if (tokenType != twd221_Token::STRING) {
            cmdType = twd221_Token::LEXERR;
        }; // if
        break;
    case twd221_Token::ABOR: // ABOR
        break;
    case twd221_Token::CWD: // CWD <sp> <pathname>
        tokenType = this->GetNextToken( arg1 );
        if (tokenType != twd221_Token::STRING) {
            cmdType = twd221_Token::LEXERR;
        }; // if
        break;
    case twd221_Token::HELP: // HELP [<string>]
        // get optional help string
        tokenType = this->GetNextToken( arg1 );
        if (tokenType != twd221_Token::TOK_CRLF
            && tokenType != twd221_Token::STRING) {
            cmdType = twd221_Token::LEXERR;
        }; // if
        break;
    case twd221_Token::NOOP: // NOOP
        break;
    case twd221_Token::MKD: // MKD <sp> <pathname>
        tokenType = this->GetNextToken( arg1 );
        if (tokenType != twd221_Token::STRING) {
            cmdType = twd221_Token::LEXERR;
        }; // if
        break;
    case twd221_Token::RMD: // RMD <sp> <pathname>
        tokenType = this->GetNextToken( arg1 );
        if (tokenType != twd221_Token::STRING) {
            cmdType = twd221_Token::LEXERR;
        }; // if
        break;
    case twd221_Token::PWD: // PWD
        break;
    case twd221_Token::CDUP: // CDUP
        break;
    case twd221_Token::SITE: // SITE <sp> HELP [<string>] ...
        // ...
        break;
    case twd221_Token::STOU: // STOU <sp> <pathname>
        tokenType = this->GetNextToken( arg1 );
        if (tokenType != twd221_Token::STRING) {
            cmdType = twd221_Token::LEXERR;
        }; // if
        break;
    case twd221_Token::SYST: // SYST
        break;
    case twd221_Token::SIZE: // SIZE <sp> <pathname>
        tokenType = this->GetNextToken( arg1 );
        if (tokenType != twd221_Token::STRING) {
            cmdType = twd221_Token::LEXERR;
        }; // if
        break;
    case twd221_Token::MDTM: // MDTM <sp> <pathname>
        tokenType = this->GetNextToken( arg1 );
        if (tokenType != twd221_Token::STRING) {
            cmdType = twd221_Token::LEXERR;
        }; // if
        break;
    case twd221_Token::QUIT: // QUIT
        break;
    case twd221_Token::RNFR: // RNFR <sp> <pathname>
        tokenType = this->GetNextToken( arg1 );
        if (tokenType != twd221_Token::STRING) {
            cmdType = twd221_Token::LEXERR;
        }; // if
        break;
    case twd221_Token::REST: // REST ...
        // ignore arguments
        this->SkipCmd();
        break;
    case twd221_Token::LEXERR:
        break;
    case twd221_Token::ERR:
        // error (timeout etc.)
        timeoutOccured = this->timeout;
        break;
    default:
        // unexpected token
        cmdType = twd221_Token::LEXERR;
        break;
    }; // switch

    ftpCmd.Init( cmdType, arg1, arg2, &sockAddr, code1, code2 );

    return sapdbwa_True;
}; // GetNextCmd


///////////////////////////////////
// private
///////////////////////////////////

/* In order defined in RFC 765 */
static ParserCmdEntry cmdTab[] = {
	{ "USER",twd221_Token::USER, STR1,  sapdbwa_True,	"<sp> username" },
	{ "PASS",twd221_Token::PASS, ZSTR1, sapdbwa_True,	"<sp> password" },
	{ "ACCT",twd221_Token::ACCT, STR1, sapdbwa_False,	"(specify account)" },
	{ "SMNT",twd221_Token::SMNT, ARGS, sapdbwa_False,	"(structure mount)" },
	{ "REIN",twd221_Token::REIN, ARGS, sapdbwa_False,	"(reinitialize server state)" },
	{ "QUIT",twd221_Token::QUIT, ARGS,  sapdbwa_True,	"(terminate service)", },
	{ "PORT",twd221_Token::PORT, ARGS,  sapdbwa_True,	"<sp> b0, b1, b2, b3, b4" },
	{ "PASV",twd221_Token::PASV, ARGS,  sapdbwa_True,	"(set server in passive mode)" },
	{ "TYPE",twd221_Token::TYPE, ARGS,  sapdbwa_True,	"<sp> [ A | E | I | L ]" },
	{ "STRU",twd221_Token::STRU, ARGS,  sapdbwa_True,	"(specify file structure)" },
	{ "MODE",twd221_Token::MODE, ARGS,  sapdbwa_True,	"(specify transfer mode)" },
	{ "RETR",twd221_Token::RETR, STR1,  sapdbwa_True,	"<sp> file-name" },
	{ "STOR",twd221_Token::STOR, STR1,  sapdbwa_True,	"<sp> file-name" },
	{ "APPE",twd221_Token::APPE, STR1,  sapdbwa_True,	"<sp> file-name" },
	{ "MLFL",twd221_Token::MLFL, OSTR, sapdbwa_False,	"(mail file)" },
	{ "MAIL",twd221_Token::MAIL, OSTR, sapdbwa_False,	"(mail to user)" },
	{ "MSND",twd221_Token::MSND, OSTR, sapdbwa_False,	"(mail send to terminal)" },
	{ "MSOM",twd221_Token::MSOM, OSTR, sapdbwa_False,
      "(mail send to terminal or mailbox)" },
	{ "MSAM",twd221_Token::MSAM, OSTR, sapdbwa_False,
      "(mail send to terminal and mailbox)" },
	{ "MRSQ",twd221_Token::MRSQ, OSTR, sapdbwa_False,
      "(mail recipient scheme question)" },
	{ "MRCP",twd221_Token::MRCP, STR1, sapdbwa_False,	"(mail recipient)" },
	{ "ALLO",twd221_Token::ALLO, ARGS,  sapdbwa_True,	"allocate storage (vacuously)" },
	{ "REST",twd221_Token::REST, ARGS,  sapdbwa_True,	"<sp> offset (restart command)" },
	{ "RNFR",twd221_Token::RNFR, STR1,  sapdbwa_True,	"<sp> file-name" },
	{ "RNTO",twd221_Token::RNTO, STR1,  sapdbwa_True,	"<sp> file-name" },
	{ "ABOR",twd221_Token::ABOR, ARGS,  sapdbwa_True,	"(abort operation)" },
	{ "DELE",twd221_Token::DELE, STR1,  sapdbwa_True,	"<sp> file-name" },
	{ "CWD", twd221_Token::CWD,  OSTR,  sapdbwa_True,	"[ <sp> directory-name ]" },
	{ "XCWD",twd221_Token::CWD,	OSTR,  sapdbwa_True,	"[ <sp> directory-name ]" },
	{ "LIST",twd221_Token::LIST, OSTR,  sapdbwa_True,	"[ <sp> path-name ]" },
	{ "NLST",twd221_Token::NLST, OSTR,  sapdbwa_True,	"[ <sp> path-name ]" },
	{ "SITE",twd221_Token::SITE, SITECMD, sapdbwa_True, "site-cmd [ <sp> arguments ]" },
	{ "SYST",twd221_Token::SYST, ARGS,  sapdbwa_True,
      "(get type of operating system)" },
	{ "STAT",twd221_Token::STAT, OSTR,  sapdbwa_True,	"[ <sp> path-name ]" },
	{ "HELP",twd221_Token::HELP, OSTR,  sapdbwa_True,	"[ <sp> <string> ]" },
	{ "NOOP",twd221_Token::NOOP, ARGS,  sapdbwa_True,	"" },
	{ "MKD", twd221_Token::MKD,  STR1,  sapdbwa_True,	"<sp> path-name" },
	{ "XMKD",twd221_Token::MKD,  STR1,  sapdbwa_True,	"<sp> path-name" },
	{ "RMD", twd221_Token::RMD,  STR1,  sapdbwa_True,	"<sp> path-name" },
	{ "XRMD",twd221_Token::RMD,  STR1,  sapdbwa_True,	"<sp> path-name" },
	{ "PWD", twd221_Token::PWD,  ARGS,  sapdbwa_True,	"(return current directory)" },
	{ "XPWD",twd221_Token::PWD,  ARGS,  sapdbwa_True,	"(return current directory)" },
	{ "CDUP",twd221_Token::CDUP, ARGS,  sapdbwa_True,	"(change to parent directory)" },
	{ "XCUP",twd221_Token::CDUP, ARGS,  sapdbwa_True,	"(change to parent directory)" },
	{ "STOU",twd221_Token::STOU, STR1,  sapdbwa_True,	"<sp> file-name" },
	{ "SIZE",twd221_Token::SIZE, OSTR,  sapdbwa_True,	"<sp> path-name" },
	{ "MDTM",twd221_Token::MDTM, OSTR,  sapdbwa_True,	"<sp> path-name" },
	{ NULL,   0,    0,    sapdbwa_False,	NULL }
}; // cmdTab


ParserCmdEntry* twd221_FtpParser::GetCmdEntry()
{
    char               cmd[5];
    ParserCmdEntry    *cmdEntry = NULL;
    
    if (this->recvPos >= 4) {
        if (this->recvBuf[3] == ' '
            || this->recvBuf[3] == '\0'
            || this->recvBuf[3] == CR) { // three bytes command
            memcpy( cmd, this->recvBuf, 3 );
            cmd[3] = '\0';
            this->parsePos += 3;
        } else {                       // four bytes command
            memcpy( cmd, this->recvBuf, 4 );
            cmd[4] = '\0';
            this->parsePos += 4;
        }; // else

        cmdEntry = this->FindCmdEntry( cmd );
    }; // else

    return cmdEntry;
}; // GetCmdEntry


ParserCmdEntry* twd221_FtpParser::FindCmdEntry( const ZString &cmd )
{
    sapdbwa_Int4      tabIndex;
    ParserCmdEntry *cmdEntry = NULL;

    for (tabIndex = 0;
         (cmdTab[tabIndex].name != NULL) && (cmdEntry == NULL);
         tabIndex++) {
        if (cmd == cmdTab[tabIndex].name ) {
            cmdEntry = &cmdTab[tabIndex];
        }; // if
    }; // for

    return cmdEntry;
}; // FindCmdEntry


sapdbwa_Int4 twd221_FtpParser::GetNextToken( ZString &value )
{
    twd201_SockRet     sockRet      = 0;
    ParserCmdEntry    *cmdEntry     = NULL;
    sapdbwa_Bool         leaveControl = sapdbwa_False;
    sapdbwa_Int4         tokenType    = twd221_Token::LEXERR;

    value = "";
    while (!leaveControl) {
        switch (this->state) {
        case CMD: // expect command
            sockRet = this->Read(); // read one line
            leaveControl = sapdbwa_True;
            if (sockRet > 0) {
                cmdEntry = this->GetCmdEntry();
                if (cmdEntry != NULL) {
                    this->state = cmdEntry->state;
                    tokenType   = cmdEntry->token;
                } else {
                    tokenType = twd221_Token::LEXERR;
                    value = this->recvBuf;
                }; // else
            } else {
                tokenType = twd221_Token::ERR;
            }; // else
            break;
        case OSTR: // read optional <sp> followed by a string
            leaveControl = sapdbwa_True;
            if (this->recvBuf[this->parsePos] == CR) {
                tokenType = twd221_Token::TOK_CRLF;
            } else if (this->SkipSpace() && this->GetString( value )) {
                tokenType = twd221_Token::STRING;
            } else {
                tokenType = twd221_Token::LEXERR;
            }; // else
            break;
        case STR1: // expect space followed by string
            if (this->SkipSpace()) {
                this->state = STR2;
            } else {
                tokenType = twd221_Token::LEXERR;
                value = this->recvBuf;
                leaveControl = sapdbwa_True;
            }; // else
            break;
        case STR2: // expect string
            leaveControl = sapdbwa_True;
            if (this->GetString( value )) {
                tokenType = twd221_Token::STRING;
            } else {
                tokenType = twd221_Token::LEXERR;
                value = this->recvBuf;
            }; // else
            break;
        case ARGS: // read arguments
            leaveControl = sapdbwa_True;
            if (isdigit(this->recvBuf[this->parsePos])) {
                // read number
                if (this->GetNumber( value )) {
                    tokenType = twd221_Token::NUMBER;
                } else {
                    tokenType = twd221_Token::LEXERR;
                    value = this->recvBuf;
                }; // else
            } else {
                // read other argument
                switch (this->recvBuf[this->parsePos++]) {
                case '\n':
                    leaveControl = sapdbwa_False;
                    state = CMD;
                    break;
                case ' ':
                    leaveControl = sapdbwa_False;
                    break;
                case ',':
                    tokenType = twd221_Token::COMMA;
                    break;
                case 'A':
                case 'a':
                    tokenType = twd221_Token::TOK_A;
                    break;
                case 'B':
                case 'b':
                    tokenType = twd221_Token::TOK_B;
                    break;
                case 'C':
                case 'c':
                    tokenType = twd221_Token::TOK_C;
                    break;
                case 'E':
                case 'e':
                    tokenType = twd221_Token::TOK_E;
                    break;
                case 'F':
                case 'f':
                    tokenType = twd221_Token::TOK_F;
                    break;
                case 'I':
                case 'i':
                    tokenType = twd221_Token::TOK_I;
                    break;
                case 'L':
                case 'l':
                    tokenType = twd221_Token::TOK_L;
                    break;
                case 'N':
                case 'n':
                    tokenType = twd221_Token::TOK_N;
                    break;
                case 'P':
                case 'p':
                    tokenType = twd221_Token::TOK_P;
                    break;
                case 'R':
                case 'r':
                    tokenType = twd221_Token::TOK_R;
                    break;
                case 'S':
                case 's':
                    tokenType = twd221_Token::TOK_S;
                    break;
                case 'T':
                case 't':
                    tokenType = twd221_Token::TOK_T;
                    break;
                }
            }; // else
            break;
        default:
            // unknown state
            leaveControl = sapdbwa_True;
            tokenType = twd221_Token::ERR;        
        }; // switch
    }; // while

    return tokenType;
}; // GetNextToken


sapdbwa_Bool twd221_FtpParser::GetPort( twd201_SockAddrIn &sockAddr )
{
    sapdbwa_Bool  ok = sapdbwa_True;
    sapdbwa_Int4  i;
    sapdbwa_Int4  token;
    ZString     value;
    sapdbwa_Int4  numVal;
    char       *addr = (char*) &sockAddr.sin_addr;
    char       *port = (char*) &sockAddr.sin_port;

    // read ip address
    for (i = 0; (i<4 && ok); i++) {
        token = this->GetNextToken( value );
        if (token == twd221_Token::NUMBER) {
            numVal = atoi( (char*) value );
            if ((numVal<=255) && (numVal>=0)) {
                addr[i] = numVal;
                token = this->GetNextToken( value );
                if (token != twd221_Token::COMMA) {
                    // COMMA expected
                    ok = sapdbwa_False;
                }; // if
            } else {
                // invalid number
                ok = sapdbwa_False;
            }; // else
        } else {
            // number expected
            ok = sapdbwa_False;
        }; // else
    }; // for

    // read port number
    for (i = 0; (i<2 && ok); i++) {
        token = this->GetNextToken( value );
        if (token == twd221_Token::NUMBER) {
            numVal = atoi( (char*) value );
            if ((numVal<=255) && (numVal>=0)) {
                port[i] = numVal;
                if (i==0) { // need comma
                    token = this->GetNextToken( value );
                    if (token != twd221_Token::COMMA) {
                        // COMMA expected
                        ok = sapdbwa_False;
                    }; // if
                }; // if
            } else {
                // invalid number
                ok = sapdbwa_False;
            }; // else
        } else {
            // number expected
            ok = sapdbwa_False;
        }; // else
    }; // for
    sockAddr.sin_family = AF_INET;

    return ok;
}; // GetPort


sapdbwa_Bool twd221_FtpParser::GetTypeCode( sapdbwa_Int4 &typeCode,
                                          sapdbwa_Int4 &formCode )
{
    sapdbwa_Bool  ok = sapdbwa_True;
    sapdbwa_Int4  token;
    ZString     dummyValue;

    token = this->GetNextToken( dummyValue );
    switch (token) {
    case twd221_Token::TOK_A :
    case twd221_Token::TOK_E :
    case twd221_Token::TOK_I :
    case twd221_Token::TOK_L :
        typeCode = token;
        break;
    default:
        ok = sapdbwa_False;
        break;
    }; // switch
    formCode = twd221_Token::FORM_N;

    return ok;
}; // GetTypeCode


sapdbwa_Bool twd221_FtpParser::GetStructCode( sapdbwa_Int4 &structCode )
{
    sapdbwa_Bool  ok = sapdbwa_True;
    sapdbwa_Int4  token;
    ZString     dummyValue;

    token = this->GetNextToken( dummyValue );
    switch (token) {
    case twd221_Token::TOK_F:
    case twd221_Token::TOK_R:
    case twd221_Token::TOK_P:
        structCode = token;
        break;
    default:
        ok = sapdbwa_False;
        break;
    }; // switch

    return ok;
}; // GetStructCode


sapdbwa_Bool twd221_FtpParser::GetModeCode( sapdbwa_Int4 &modeCode )
{
    sapdbwa_Bool  ok = sapdbwa_True;
    sapdbwa_Int4  token;
    ZString     dummyValue;

    token = this->GetNextToken( dummyValue );
    switch (token) {
    case twd221_Token::TOK_S:
    case twd221_Token::TOK_B:
    case twd221_Token::TOK_C:
        modeCode = token;
        break;
    default:
        ok = sapdbwa_False;
        break;
    }; // switch

    return ok;
}; // GetModeCode


sapdbwa_Bool twd221_FtpParser::SkipSpace()
{
    sapdbwa_Bool ok;

    if (this->recvBuf[this->parsePos] == ' ') {
        this->parsePos++;
        ok = sapdbwa_True;
    } else {
        ok = sapdbwa_False;
    }; // else

    return ok;
}; // SkipSpace


void twd221_FtpParser::SkipCmd()
{
    sapdbwa_Int4 bufIndex = this->parsePos;

    while ((this->parsePos < sizeof(this->recvBuf))
           && (this->recvBuf[this->parsePos] != '\0')
           && (this->recvBuf[this->parsePos] != CR))
        this->parsePos++;
    if (this->recvBuf[this->parsePos] == CR)
        this->parsePos++;           
}; // SkipCmd


sapdbwa_Bool twd221_FtpParser::GetString( ZString &value )
{
    char *endPos;
    sapdbwa_Bool ok;

    endPos = strchr( &this->recvBuf[this->parsePos], CR );
    if (endPos != NULL) {
        *endPos = '\0';
        value = &this->recvBuf[this->parsePos];
        *endPos = '\n';
        this->parsePos += strlen( value );
        ok = sapdbwa_True;
    } else {
        ok = sapdbwa_False;
    }; // else

    return ok;
}; // GetString


sapdbwa_Bool twd221_FtpParser::GetNumber( ZString &value )
{
    sapdbwa_Int4 valPos = 0;
    char       num[5];

    while (valPos+1 < sizeof(num)
           && isdigit( this->recvBuf[this->parsePos] )
           && this->parsePos < sizeof( this->recvBuf )) {
        num[valPos] = this->recvBuf[this->parsePos];
        valPos++;
        this->parsePos++;
    }; // while
    num[valPos] = '\0';
    value = num;

    return (valPos > 0);
}; // GetNumber


void twd221_FtpParser::NewCmd()
{
    this->recvPos = 0;
    this->parsePos = 0;
    this->recvBuf[0] = '\0';
    this->state = CMD;
}; // NewCmd

sapdbwa_Int4 twd221_FtpParser::Read()
{
    twd201_SockRet sockRet;
    sapdbwa_Int4     bytesRead = 0;
    sapdbwa_Bool     selectOK  = sapdbwa_False;

    while (selectOK = this->Select()) {
        sockRet = this->Receive( &this->recvBuf[this->recvPos],
                                 sizeof( this->recvBuf ) - this->recvPos );
        if (sockRet > 0) {
            this->recvPos += sockRet;
            bytesRead += sockRet;
            // did we receive a CRLF ?
            if (wd221_ContainsChar( this->recvBuf,
                                    this->recvPos-sockRet,
                                    this->recvPos-1, CR )) {
                // then give control to the parser/scanner
                break;
            }; // if
                              
        } else if (sockRet < 0) {
            wd203SockErr( "twd221_FtpParser::Read", "recv" );

            return sockRet;
        } else { // sockRet == 0
            break;
        }; // else
    }; // while
    if (selectOK) {
        this->recvBuf[this->recvPos] = '\0';
    } else {
        return -1;
    }; // else

    return bytesRead;
}; // Read


sapdbwa_Bool twd221_FtpParser::Select()
{
    fd_set stReadFDS;
    fd_set stXcptFDS;
    struct timeval stTimeout;
    twd201_SockRet sockRet;
    
    FD_ZERO( &stReadFDS );
    FD_ZERO( &stXcptFDS );
    FD_SET( this->socket, &stReadFDS );
    FD_SET( this->socket, &stXcptFDS );

    stTimeout.tv_sec = SELECT_TIMEOUT;
    stTimeout.tv_usec = 0;

    sockRet = wd201Select( this->socket+1, /* only required for UNIX */
                           &stReadFDS,
                           NULL,
                           &stXcptFDS,
                           &stTimeout );
    if (sockRet == SOCKET_ERROR) {
        // error
        wd203SockErr( "twd221_FtpParser::Select", "select" );

        return sapdbwa_False;
    } else if (sockRet == 0) {
        // timeout
        this->timeout = sapdbwa_True;
        wd203Error( "twd221_FtpParser::Select", "select", "TimeOut" );

        return sapdbwa_False;
    } else if (sockRet != 0) {
        if ((FD_ISSET( this->socket, &stXcptFDS))) {
            // exception
            twd201_SockErr errCode;
            sapdbwa_Int4     errLen = sizeof(twd201_SockErr);

            sockRet = wd201GetSockOpt( this->socket,
                                       SOL_SOCKET, SO_ERROR,
                                       (char*) &errCode, &errLen );
            if (sockRet != SOCKET_ERROR) {
                wd203Error( "twd221_FtpParser::Select", "getsockopt",
                            (char*) wd201StrError( errCode ));
            } else {
                wd203SockErr( "twd221_FtpParser::Select", "getsockopt" );
            }; // else

            return sapdbwa_False;
        }; // if
        if (!(FD_ISSET( this->socket, &stReadFDS))) {
            // should never happen
            wd203Error( "twd221_FtpParser::Select", "getsockopt",
                        "should never happen\n." );

            return sapdbwa_False;
        }; // if
    }; // else

    // now we can read
    return sapdbwa_True;
}; // Select


sapdbwa_Int4 twd221_FtpParser::Receive( char *buf, sapdbwa_Int4 bufLen )
{
    twd201_SockRet sockRet;

    sockRet = wd201Recv( this->socket, buf, bufLen-1, 0 );
    if (sockRet>0) {
        buf[sockRet] = '\0';
        if (this->debug) {
            syslog( LOG_DEBUG, buf, 0 );
        };
    }; // if

    return sockRet;
}; // Receive


/*!**********************************************************************

  EndChapter: twd221_FtpParser
  
************************************************************************/

sapdbwa_Bool wd221_ContainsChar( char       *buf,
                               sapdbwa_Int4  from,
                               sapdbwa_Int4  to,
                               char        c )
{
    sapdbwa_Int4 pos = from;

    while (pos <= to) {
        if (buf[pos] == c)
            return sapdbwa_True;
        pos++;
    }; // while

    return sapdbwa_False;
}; // ContainsChar
