/*!**********************************************************************

  module: vwd52ms.c

  -----------------------------------------------------------------------

  responsible:  Holger Bischoff
				Markus zgen

  special area: WebAgent

  description:  Functions for Microsoft's Web Server

  see also:

  -----------------------------------------------------------------------

                         Copyright by SAP AG, 1998-2000



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

************************************************************************/

/*=======================================================================
 *  INCLUDES
 *=====================================================================*/

#include <WINDOWS.H>
/* for IISAPI */
#include <httpfilt.h>
#include <httpext.h>
/* for IISAPI till here */

#include "gwd00.h"
#include "hwd03waapi.h"
#include "hwd20control.h"
#include "hwd21req.h"
#include "hwd22rep.h"
#include "hwd28dict.h"
#include "hwd29date.h"

sapdbwa_TraceDefModule( "vwd52ms.c" );

/*!**********************************************************************

  Chapter: Filter-Functions

  description: These functions are called by the Microsoft Web-Server.
  
************************************************************************/

/* GetFilterVersion is called at startup of the webserver.
 * Here we initialize the WebAgent.
 */
BOOL WINAPI GetFilterVersion( HTTP_FILTER_VERSION *pVer );


/* This functions is called by the WebServer for each HTTP-request
 * that comes in. Here we retrieve the request information and call
 * the WebAgent service function.
 */
DWORD WINAPI HttpFilterProc( HTTP_FILTER_CONTEXT   *pfc,
                             DWORD                  NotificationType,
                             VOID                  *pvData );


/*!**********************************************************************

  EndChapter: Filter-Functions

************************************************************************/

/*!**********************************************************************

  Chapter: DLL-Main function

  description: We have to clean up resources when exiting
  
************************************************************************/

BOOL APIENTRY DllMain( HANDLE	hModule,
                       DWORD    dwReason,
                       LPVOID   lpReserved)
{
    switch (dwReason) {
    case DLL_PROCESS_ATTACH:
        break;
    case DLL_PROCESS_DETACH:
        wd20ApiExit();
        break;
    default:
        break;
    }; /* switch */
}; /* DllMain */

/*!**********************************************************************

  EndChapter: DLL-Main function

************************************************************************/

/*!**********************************************************************

  Chapter: Call-Back Functions

  description: These functions are stored in the WAHttpRequest and
               WAHttpReply structures. They are used by the WebAgent
               to retrieve body-data from, and send the header and
               content to the web-server.
  
************************************************************************/

sapdbwa_UInt4 wd52BillsReadBody( twd00ServerHandle   handle,
                               char               *b,
                               sapdbwa_UInt4         len );


const char* wd52BillsGetHeader( twd00ServerHandle   handle,
                                const char         *fieldName );

sapdbwa_Bool wd52BillsSendHeader( twd00ServerHandle   handle,
                                sapdbwa_HttpReplyP       rep );


sapdbwa_UInt4 wd52BillsSendBody( twd00ServerHandle   handle,
                               const void         *Buf,
                               sapdbwa_UInt4         len );


sapdbwa_Bool wd52BillsIsSecure( twd00ServerHandle   handle );

/*!**********************************************************************

  EndChapter: Call-Back Functions

************************************************************************/

/*!**********************************************************************

  Chapter: Internal Functions

************************************************************************/

/* Increment when allocating buffers (e.g. for query string). */
#define CGI_CHUNK	2048


/* Bill does not support anything like ReadClient in case request-method
 * is POST.
 * The only way to access the data in the body of the HTTP request is
 * to access the request in 'raw' form.
 */
struct st_wa_raw_data
{
    size_t      BufSz;
    size_t      BufLg;
    void       *Buf;
    void       *BufBody;
    size_t      CL;
	size_t		HeaderLg;
};


typedef struct st_wa_bills_server_handle {
    HTTP_FILTER_CONTEXT *httpFilterContext;
    HTTP_FILTER_URL_MAP *httpFilterUrlMap;
    char                *queryString;
    sapdbwa_UInt4          contentLength;
	sapdbwa_Bool           we_malloced_Uri;
	sapdbwa_Bool           we_malloced_IfModifiedSince;
	sapdbwa_Bool           we_malloced_QueryString;
	sapdbwa_Bool           we_malloced_PathInfo;
	sapdbwa_Bool           we_malloced_RequestMethod;
	sapdbwa_Bool           we_malloced_ContentType;
	sapdbwa_Bool           we_malloced_ContentLength;
	sapdbwa_Bool           we_malloced_PhysicalPath;
	sapdbwa_Bool           we_malloced_ServerName;

    twd28DictionaryP headerFields; /* allocated header fields of the
                                    * request allocated by
                                    * wd52GetHeaderField are stored here */
} *twd52BillsServerHandle;


/* Set content-length in req. */
static size_t wd52_BillsContentLength( twd21HttpRequestP  req,
                                       size_t             BufferSize );

/* Set content-type in req. */
static size_t wd52_BillsContentType( twd21HttpRequestP  req,
                                     size_t             BufferSize );

/* Set if-modified-since in req. */
static size_t wd52_BillsIfModifiedSince( twd21HttpRequestP req );


/* Set path-info in req. */
static size_t wd52_BillsPathInfo( twd21HttpRequestP  req,
                                  size_t             BufferSize );

/* Set physical (translated) path in req. */
static size_t wd52_BillsPhysicalPath( twd21HttpRequestP  req,
                                      size_t             BufferSize );

/* Read body-data from web-server. */
static size_t wd52_BillsReadPost( twd52BillsServerHandle  handle,
                                  size_t                  BufferSize,
                                  char                   *Buffer );

/* Prepare reading of body data. */
static sapdbwa_Bool wd52_BillsReadPostInit( twd21HttpRequestP req );

/* Free buffers used by wd52_BillsReadRaw */
static DWORD WINAPI wd52_BillsClearReadRaw( HTTP_FILTER_CONTEXT  *pfc,
                                            HTTP_FILTER_RAW_DATA *pvData );

/* Set query-string in req. */
static size_t wd52_BillsQueryString( struct st_wa_http_req *req );

/* Read raw data from server into internal buffers. */
static DWORD WINAPI wd52_BillsReadRaw( HTTP_FILTER_CONTEXT   *pfc,
                                       HTTP_FILTER_RAW_DATA  *pvData );

/* Create request and reply handles and call service function */
static DWORD WINAPI wd52_BillsRequestMap( HTTP_FILTER_CONTEXT  *pfc,
                                          HTTP_FILTER_URL_MAP  *pvData );

/* Set server-name in req. */
static size_t wd52_BillsServerName( twd21HttpRequestP req, size_t BufferSize );

/* Send reply body to web-server. */
static int wd52_BillsWrite( twd52BillsServerHandle  handle,
                            void                   *b,
                            size_t                  n );

/* Send reply header to web-server. */
static sapdbwa_Bool wd52_BillsWriteHeader( twd52BillsServerHandle handle,
                                         sapdbwa_HttpReplyP          rep );

/* Create bills-server-handle. This handle is passed to the
 * call-back-functions. */
twd52BillsServerHandle wd52_CreateServerHandle
                                  ( HTTP_FILTER_CONTEXT *httpFilterContext,
                                    HTTP_FILTER_URL_MAP *httpFilterUrlMap );

/* Free memory allocated for the request. */
void wd52_DestroyRequest( twd52BillsServerHandle handle,
                          twd21HttpRequestP      req );

/* Free memory for the server-handle */
void wd52_DestroyServerHandle( twd52BillsServerHandle handle );

/*!**********************************************************************

  EndChapter: Internal Functions

************************************************************************/

/*!**********************************************************************

  Chapter: Filter-Functions

************************************************************************/

BOOL WINAPI GetFilterVersion( HTTP_FILTER_VERSION * pVer )
{

	sapdbwa_TraceDefFunction( "GetFilterVersion" );

    if (!wd20ApiInit(NULL, API_IIS_WD20 )) {
        return false;
    }; /* if */

	sapdbwa_TraceAdd( "WebAgent API Init successfull." );
    
    pVer->dwFilterVersion = HTTP_FILTER_REVISION;

    pVer->dwFlags = (/*SF_NOTIFY_PREPROC_HEADERS    |*/
                     SF_NOTIFY_READ_RAW_DATA    |
                     SF_NOTIFY_URL_MAP      |
                     SF_NOTIFY_LOG      |
                     SF_NOTIFY_END_OF_NET_SESSION|
                     SF_NOTIFY_ORDER_DEFAULT);

    strcpy( pVer->lpszFilterDesc, "WebAgent Filter, version 1.0" );
    
    return true;
}; /* GetFilterVersion */


DWORD WINAPI HttpFilterProc( HTTP_FILTER_CONTEXT   *pfc,
                             DWORD                  NotificationType,
                             VOID                  *pvData )
{

    DWORD   rtc = SF_STATUS_REQ_NEXT_NOTIFICATION;

	sapdbwa_TraceDefFunction( "HttpFilterProc" );

    switch (NotificationType) {
    case SF_NOTIFY_READ_RAW_DATA:
        rtc = wd52_BillsReadRaw( pfc, pvData );
        break;
    case SF_NOTIFY_URL_MAP:
        rtc = wd52_BillsRequestMap( pfc, pvData );
        wd52_BillsClearReadRaw( pfc, pvData );
        break;
    case SF_NOTIFY_LOG:
    case SF_NOTIFY_END_OF_NET_SESSION:
        rtc = wd52_BillsClearReadRaw( pfc, pvData );
        break;
    default:
        rtc = SF_STATUS_REQ_NEXT_NOTIFICATION;
    }; /* switch */

    return(rtc);
}; /* HttpFilterProc */

/*!**********************************************************************

  EndChapter: Filter-Functions

************************************************************************/

/*!**********************************************************************

  Chapter: Call-Back-Functions

************************************************************************/


sapdbwa_UInt4 wd52BillsReadBody( twd00ServerHandle   handle,
                               char               *b,
                               sapdbwa_UInt4         len )
{
    return wd52_BillsReadPost( handle, len, b );
}; /* wd52BillsReadBody */

#define MAX_VALUE_LEN 1024
#define MAX_KEY_LEN 255

const char* wd52BillsGetHeader( twd00ServerHandle   handle,
                                const char         *fieldName )
{
    twd52BillsServerHandle  billsHandle;
    HTTP_FILTER_CONTEXT    *pfc;
    const char *result;
    char key[MAX_KEY_LEN+1];
    char value[MAX_VALUE_LEN+1];
    sapdbwa_Bool allocatOK;
    sapdbwa_UInt4 valueSize;
    
    billsHandle = (twd52BillsServerHandle) handle;

    /* did we already retrieve this field ? */
    result = wd28GetValueByKey( billsHandle->headerFields, fieldName );
    if (result == NULL) {
        /* try to get value from web server */
        pfc = (HTTP_FILTER_CONTEXT*) billsHandle->httpFilterContext;

        valueSize = MAX_VALUE_LEN;
        strncpy( key, fieldName, MAX_KEY_LEN );
        /* get the value */
        if (pfc->GetServerVariable( pfc,
                                    (char*) fieldName,
                                    value,
                                    &valueSize)) {
            if (!wd28AddCopyEntry( billsHandle->headerFields, key, value )) {
                result = NULL;
            } else {
                result = wd28GetValueByKey( billsHandle->headerFields, key );
            }; /* else */
        } else {
            result = NULL;
        }; /* else */
    }; /* if */
    
    return result;
}; /* wd52BillsGetHeader */


sapdbwa_Bool wd52BillsSendHeader( twd00ServerHandle   handle,
                                sapdbwa_HttpReplyP       rep )
{
    return wd52_BillsWriteHeader( handle, rep );
}; /* twd52BillsSendHeader */ 


sapdbwa_UInt4 wd52BillsSendBody( twd00ServerHandle   handle,
                               const void         *Buf,
                               sapdbwa_UInt4         len )
{
    return wd52_BillsWrite( handle, (void*) Buf, len );
}; /* wd52BillsSendBody */


sapdbwa_Bool wd52BillsIsSecure( twd00ServerHandle   handle )
{
    const char *isSecure = NULL;
    
    isSecure = wd52BillsGetHeader( handle, "SERVER_PORT_SECURE" );

    return (isSecure != NULL && (strcmp( isSecure, "1" ) == 0));
} /* wd52BillsIsSecure */

/*!**********************************************************************

  EndChapter: Call-Back-Functions

************************************************************************/

/*!**********************************************************************

  Chapter: Internal Functions

************************************************************************/


twd52BillsServerHandle
wd52_CreateServerHandle( HTTP_FILTER_CONTEXT *httpFilterContext,
                         HTTP_FILTER_URL_MAP *httpFilterUrlMap )
{
    twd52BillsServerHandle handle;
    sapdbwa_Bool             allocatOK;
    
    sapdbwa_SQLALLOCAT( sizeof( struct st_wa_bills_server_handle),
                (char**) &handle, &allocatOK );
    if (!allocatOK) {
        /* memory allocation error */
        wd20WriteLogMsg( ERR_MSG_MEMORY_WD26 );
         return NULL;
    }; /* if */

    handle->headerFields = wd28CreateDictionary();
    if (handle->headerFields == NULL) {
        sapdbwa_SQLFREE( (sapdbwa_UInt1*) handle );
        wd20WriteLogMsg( ERR_MSG_MEMORY_WD26 );
        return NULL;
    }; /* if */
    
    handle->httpFilterContext = httpFilterContext;
    handle->httpFilterUrlMap  = httpFilterUrlMap;
    handle->contentLength = 0;
    handle->queryString = NULL;
	handle->we_malloced_Uri = false;
	handle->we_malloced_IfModifiedSince = false;
	handle->we_malloced_QueryString = false;
	handle->we_malloced_PathInfo = false;
	handle->we_malloced_RequestMethod = false;
	handle->we_malloced_ContentType = false;
	handle->we_malloced_ContentLength = false;
	handle->we_malloced_PhysicalPath = false;
	handle->we_malloced_ServerName = false;

    return handle;
}; /* wd52_CreateServerHandle */


void wd52_DestroyServerHandle( twd52BillsServerHandle handle )
{
    wd28DestroyDictionary( handle->headerFields );        
    sapdbwa_SQLFREE( (sapdbwa_UInt1*) handle );
}; /* wd52_DestroyServerHandle */


void wd52_DestroyRequest( twd52BillsServerHandle handle,
                          twd21HttpRequestP      req )
{
    if (handle->we_malloced_Uri) {
        sapdbwa_SQLFREE( (char*) sapdbwa_GetRequestURI( req ));
    }; /* if */
    if (handle->we_malloced_IfModifiedSince) {
        sapdbwa_SQLFREE( (char*) sapdbwa_GetIfModifiedSince( req ));
    }; /* if */
    if (handle->we_malloced_QueryString) {
        sapdbwa_SQLFREE( handle->queryString );
    }; /* if */
    if (handle->we_malloced_PathInfo) {
        sapdbwa_SQLFREE( (char*) sapdbwa_GetPathInfo( req ));
    }; /* if */
    if (handle->we_malloced_RequestMethod) {
        sapdbwa_SQLFREE( (char*) sapdbwa_GetMethod( req ));
    }; /* if */
    if (handle->we_malloced_ContentType) {
        sapdbwa_SQLFREE( (char*) sapdbwa_GetContentType( req ));
    }; /* if */
    if (handle->we_malloced_ContentLength) {
        sapdbwa_SQLFREE( (char*) sapdbwa_GetContentLength( req ));
    }; /* if */
    if (handle->we_malloced_PhysicalPath) {
        sapdbwa_SQLFREE( (char*) sapdbwa_GetPathTranslated( req ));
    }; /* if */
    if (handle->we_malloced_ServerName) {
        sapdbwa_SQLFREE( (char*) sapdbwa_GetServerName( req ));
    }; /* if */
    wd21DestroyRequest( req );
}; /* wd52_DestroyRequest */

/*********************************************************/

static int wd52_BillsWrite( twd52BillsServerHandle  handle,
                            void                   *b,
                            size_t                  n )
{
    DWORD                Lg = n;
    HTTP_FILTER_CONTEXT *Pfc = handle->httpFilterContext;
    BOOL                 rtc;

    rtc = Pfc->WriteClient( Pfc, b, &n, HSE_IO_SYNC );

    if (!rtc)
        return(0);
    else
        return(Lg);
}; /* wd52_BillsWrite */


/*********************************************************/

#define HEADER_BUFF_SIZE 1024

static sapdbwa_Bool wd52_BillsWriteHeader( twd52BillsServerHandle handle,
                                         sapdbwa_HttpReplyP          rep )
{
    HTTP_FILTER_CONTEXT *Pfc = handle->httpFilterContext;
    DWORD                n = 0;
    DWORD                Lg = 0;
    BOOL                 rtc;
    char                 Date[MAX_DATE_LEN_WD29] = "";
    char                *Buf;
    char                *Buf2;
    int                  buflen = 0;
    int                  buflen2 = 0;
    int                  flen;
    int                  urllen = 0;
    sapdbwa_Bool           allocatOK;
    
    sapdbwa_UInt4       iHeader;
    twd28DictionaryP  headers = wd22GetHeaders( rep );
    const char       *name;
    char             *value;
    
    if(wd22GetLocation(rep) != NULL)
        urllen = strlen(wd22GetLocation(rep));

    sapdbwa_SQLALLOCAT( HEADER_BUFF_SIZE + urllen, &Buf, &allocatOK );
    if(!allocatOK) {
        /* WebDbLog("bills_WriteHeader : could not malloc"); */
        return 0;
    }; /* if */

    Buf[0] = '\0';
        
    if( wd22GetLocation(rep) != NULL
        && (flen = strlen(wd22GetLocation(rep))) != 0) {
        sprintf( Buf + buflen, "Location:%s\r\n", wd22GetLocation(rep) );
        buflen += (flen + 11);
    };
    if( wd22GetLastModified(rep) != NULL
        && (flen = wd29WADateToHttpDate(wd22GetLastModified(rep), Date)) != 0)
        {
            sprintf(Buf + buflen, "Last-Modified:%s\r\n", Date);
            buflen += (flen + 16);
        };
    if( wd22GetContentLength(rep) != NULL
        && (flen = strlen(wd22GetContentLength(rep))) != 0) {
        sprintf( Buf + buflen, "Content-Length:%s\r\n",
                 wd22GetContentLength(rep));
        buflen += (flen + 17);
    };
    if ( wd22GetContentType(rep) != NULL
         && (flen = strlen(wd22GetContentType(rep))) != 0) {
        sprintf(Buf + buflen, "Content-Type:%s\r\n", wd22GetContentType(rep));
        buflen += (flen + 15);
    };
    if ( wd22GetExpires( rep ) != NULL
         && (flen = wd29WADateToHttpDate( wd22GetExpires(rep), Date)) != 0) {
        sprintf(Buf + buflen, "Expires:%s\r\n", Date);
        buflen += (flen + 10);
    };

    for (iHeader = 0; iHeader < wd28NumEntries( headers ); iHeader++) {
        if (wd28GetEntryByIndex( headers, iHeader, &name, &value )) {
            flen = strlen( value );
            /* there must not be a space between the ':' and the value;
             * else IE 5 is confused */
            sprintf( Buf + buflen, "%s:%s\r\n", name, value );
            buflen += (flen + strlen(name) + 3);
        }; /* if */
    }; /* for */
    
    sprintf( Buf + buflen, "\r\n");
    buflen += 2;

    switch (wd22GetStatus(rep)) {
    case sapdbwa_Status_Ok:
        if (!Pfc->ServerSupportFunction( Pfc,
                                         SF_REQ_SEND_RESPONSE_HEADER,
                                         NULL,
                                         (DWORD)Buf,
                                         strlen(Buf))) {
            sapdbwa_SQLFREE(Buf);
            wd20WriteLogMsg( "vwd52ms.c: SF_REQ_SEND_RESPONSE_HEADER failed.\n" );
            return(false);
        };
        sapdbwa_SQLFREE(Buf);
        return(true);
        break;
    default:                    /* Thu, 11 Jun 1998 11:46:58 GMT */
        wd29HttpDate(Date);
        buflen2 = strlen("HTTP/1.0 %d \r\nDate: %s\r\n%s\r\n")
            + 10 + strlen(Date) + buflen + strlen("\r\n");
        sapdbwa_SQLALLOCAT( buflen2+1, &Buf2, &allocatOK );
        Buf2[0] = 0;
        sprintf( Buf2, "HTTP/1.0 %d \r\nDate: %s\r\n%s\r\n",
                 wd22GetStatus(rep), Date, Buf );
        Lg = n = strlen(Buf2);
        if (!(rtc = Pfc->WriteClient(Pfc, Buf2, &n, HSE_IO_SYNC))) {
            sapdbwa_SQLFREE(Buf);
            sapdbwa_SQLFREE(Buf2);
            wd20WriteLogMsg( "vwd52ms.c: WriteClient failed.\n" );
            return(false);
        }; /* if */
        return(true);
        break;
    }; /* switch */
    return(false);
}; /* wd52_BillsWriteHeader */

/*********************************************************/

static size_t wd52_BillsQueryString( struct st_wa_http_req *req )
{
    char                   *queryString = NULL;
    size_t                  Lg = 0;
    DWORD                   err;
    unsigned long           bills_query_string_size = 0;
    twd52BillsServerHandle  handle;
    HTTP_FILTER_CONTEXT    *pfc;
    sapdbwa_Bool              allocatOK;
    
    handle = wd21GetServerHandle( req );
    pfc = (HTTP_FILTER_CONTEXT*) handle->httpFilterContext;

    if (sapdbwa_GetQueryString(req) != NULL) {
        return(strlen(sapdbwa_GetQueryString(req))+1);
    }; /* if */

    bills_query_string_size = CGI_CHUNK;
    sapdbwa_SQLALLOCAT( bills_query_string_size, &queryString, &allocatOK );
    if (!allocatOK) {
        /* memory allocation failure */
        wd20WriteLogMsg( ERR_MSG_MEMORY_WD26 );
        return 0;
    }; /* if */
    do {
        if (!pfc->GetServerVariable( pfc,
                                     "QUERY_STRING",
                                     queryString,
                                     &bills_query_string_size)) {
            err = GetLastError();
            if ( err == ERROR_INSUFFICIENT_BUFFER
                 || err == ERROR_MORE_DATA ) {
                bills_query_string_size += CGI_CHUNK;
                sapdbwa_SQLREALLOCAT( bills_query_string_size,
                              &queryString,
                              &allocatOK );
                if (!allocatOK) {
                    sapdbwa_SQLFREE(queryString);     /* will crash, here we
                                            * have still sth to do
                                            */
                    handle->we_malloced_QueryString = false;
                    return (0);
                }
            } else {
                sapdbwa_SQLFREE(queryString);    /* will crash, here
                                          * we have still sth
                                          * to do */
                handle->we_malloced_QueryString = false;
                return (0);
            }
        } else
            break;
    } while (true);
    wd21SetQueryString( req, queryString );
    handle->queryString = queryString;
    handle->we_malloced_QueryString = true;
    
    return(strlen(queryString)+1);
}; /* wd52_BillsQueryString */


/*********************************************************/

static sapdbwa_Bool wd52_BillsReadPostInit( twd21HttpRequestP req )
{
    struct st_wa_raw_data  *pFilterContext;
    const char          *tmp;
    HTTP_FILTER_CONTEXT *pfc;
    twd52BillsServerHandle serverHandle = wd21GetServerHandle( req );
    
    pfc = (HTTP_FILTER_CONTEXT*) serverHandle->httpFilterContext;
    pFilterContext = pfc ? pfc->pFilterContext : NULL;

    tmp = sapdbwa_GetMethod( req );

    if (!stricmp("POST", tmp) || !stricmp("PUT", tmp))
        {
            tmp = sapdbwa_GetContentLength( req );
            serverHandle->contentLength = strtoul( tmp, NULL, 10);
            if (serverHandle->contentLength == ULONG_MAX)
                serverHandle->contentLength = 0;

            /* find begin of body in pFilterContext */
            if ( pFilterContext
                 && pFilterContext->Buf
                 && pFilterContext->BufLg > 0)
                {
                    pFilterContext->BufBody =
                        strstr( pFilterContext->Buf, "\r\n\r\n");
                    if (pFilterContext->BufBody)
                        pFilterContext->BufBody =
                            (char*) pFilterContext->BufBody
                            + strlen("\r\n\r\n");
                    else
                        serverHandle->contentLength = 0;
                }
        }
    else
        serverHandle->contentLength = 0;

    return(true);
}; /* wd52_BillsReadPostInit */

/*********************************************************/

static size_t wd52_BillsReadPost( twd52BillsServerHandle  handle,
                                  size_t                  BufferSize,
                                  char                   *Buffer )
{
    HTTP_FILTER_CONTEXT *pfc;
    struct st_wa_raw_data  *pFilterContext;
    size_t               rtc = 0;
    size_t               tmp = 0;

    pfc = (HTTP_FILTER_CONTEXT*) handle->httpFilterContext;

    pFilterContext = pfc ? pfc->pFilterContext : NULL;

    if ( handle->contentLength == 0
         || !pFilterContext
         || !pFilterContext->BufBody)
        return(0);

    BufferSize = min( BufferSize, handle->contentLength );

    if (BufferSize > 0)
        {
            memcpy( Buffer, pFilterContext->BufBody, BufferSize );
            pFilterContext->BufBody =
                (char*) pFilterContext->BufBody + BufferSize;

            handle->contentLength -= BufferSize;
            rtc = BufferSize;
        }

    return(rtc);
}; /* wd52_BillsReadPost */

/*********************************************************/

static size_t wd52_BillsContentType( twd21HttpRequestP  req,
                                     size_t             BufferSize )
{
    HTTP_FILTER_CONTEXT *pfc;
    char *contentType;
    sapdbwa_Bool allocatOK;
    twd52BillsServerHandle handle;

    sapdbwa_SQLALLOCAT( BufferSize, &contentType, &allocatOK );
    if (!allocatOK) {
        /* memory allocation failure */
        wd20WriteLogMsg( ERR_MSG_MEMORY_WD26 );
        return 0;
    }; /* if */

    handle = wd21GetServerHandle( req );
    pfc = (HTTP_FILTER_CONTEXT*) handle->httpFilterContext;
    if (!pfc->GetServerVariable( pfc,
                                 "CONTENT_TYPE",
                                 contentType,
                                 &BufferSize)) {
        sapdbwa_SQLFREE( contentType );
        return(0);
    } else {
        handle->we_malloced_ContentType = true;
        wd21SetContentType( req, contentType );
        return(BufferSize);
    }; /* else */
}; /* wd52_BillsContentType */

/*********************************************************/

static size_t wd52_BillsRequestMethod( twd21HttpRequestP  req,
                                       size_t             BufferSize )
{
    HTTP_FILTER_CONTEXT *pfc;
    char *requestMethod;
    sapdbwa_Bool allocatOK;
    twd52BillsServerHandle handle;
    
    sapdbwa_SQLALLOCAT( BufferSize, &requestMethod, &allocatOK );
    if (!allocatOK) {
        /* memory allocation failure */
        wd20WriteLogMsg( ERR_MSG_MEMORY_WD26 );
        return 0;
    }; /* if */

    handle = wd21GetServerHandle(req);
    pfc = (HTTP_FILTER_CONTEXT*) handle->httpFilterContext;
    if (!pfc->GetServerVariable( pfc,
                                 "REQUEST_METHOD",
                                 requestMethod,
                                 &BufferSize )) {
        sapdbwa_SQLFREE( requestMethod );
        return(0);
    } else {
        handle->we_malloced_RequestMethod = true;
        wd21SetRequestMethod( req, requestMethod );
        return(BufferSize);
    }; /* else */
}; /* wd52_BillsRequestMethod */


/*********************************************************/

static size_t wd52_BillsIfModifiedSince( twd21HttpRequestP req )
{
    char                    *tmp = NULL;
    size_t                   Lg = 0;
    DWORD                    err;
    unsigned long            bills_if_modified_since_size = 0;
    char                    *iscontained;
    char                    *endpos;
    char                    *mybuf = NULL;
    char                    *ifModifiedSince;
    HTTP_FILTER_CONTEXT     *pfc;
    twd52BillsServerHandle   handle;
    sapdbwa_Bool               allocatOK;
    
    handle = wd21GetServerHandle( req );
    pfc = (HTTP_FILTER_CONTEXT*) handle->httpFilterContext;

    sapdbwa_SQLALLOCAT( 15, &ifModifiedSince, &allocatOK );
    if (!allocatOK) {
        /* memory allocation failure */
        wd20WriteLogMsg( ERR_MSG_MEMORY_WD26 );
        return 0;
    }; /* if */    

    sapdbwa_SQLALLOCAT( CGI_CHUNK, &mybuf, &allocatOK );
    if (!allocatOK) {
        /* memory allocation failure */
        wd20WriteLogMsg( ERR_MSG_MEMORY_WD26 );
        sapdbwa_SQLFREE( ifModifiedSince );
        return 0;
    }; /* if */    
    do {
        if (!pfc->GetServerVariable( pfc,
                                     "ALL_HTTP",
                                     mybuf,
                                     &bills_if_modified_since_size)) {
            err = GetLastError();
            if ( err == ERROR_INSUFFICIENT_BUFFER
                 || err == ERROR_MORE_DATA) {
                tmp = mybuf;
                bills_if_modified_since_size += CGI_CHUNK;
                sapdbwa_SQLREALLOCAT( bills_if_modified_since_size,
                              &mybuf,
                              &allocatOK );
                if (!allocatOK) {
                    sapdbwa_SQLFREE( ifModifiedSince );
                    sapdbwa_SQLFREE( tmp );
                    return(0);
                }; /* if */
            } else {
                sapdbwa_SQLFREE( mybuf );
                sapdbwa_SQLFREE( ifModifiedSince );
                return(0);
            }; /* else */
        } else
            break;
    } while (true);

    iscontained = strstr( mybuf, "HTTP_IF_MODIFIED_SINCE:" );
    if(iscontained != NULL) {
        endpos = strpbrk( iscontained, ";\r\n");
        *endpos = '\0';   /* now iscontained consists of the
                           * HTTP-Date (and nothing else)    */
        wd29HttpDateToWADate( iscontained + 23, ifModifiedSince );
        sapdbwa_SQLFREE( mybuf );
        wd21SetIfModifiedSince( req, ifModifiedSince );
        handle->we_malloced_IfModifiedSince = true;
        return( strlen(ifModifiedSince)+1 );
    } else {
        sapdbwa_SQLFREE( mybuf );
        sapdbwa_SQLFREE( ifModifiedSince );
        return 0;
    }; /* else */
}; /* wd52_BillsIfModifiedSince */

/*********************************************************/

static size_t wd52_BillsPathInfo( twd21HttpRequestP  req,
                                  size_t             BufferSize )
{
    HTTP_FILTER_CONTEXT   *pfc;
    char                  *pathInfo;
    sapdbwa_Bool             allocatOK;
    twd52BillsServerHandle handle;
    
    sapdbwa_SQLALLOCAT( BufferSize, &pathInfo, &allocatOK );
    if (!allocatOK) {
        /* memory allocation failure */
        wd20WriteLogMsg( ERR_MSG_MEMORY_WD26 );
        return 0;
    }; /* if */

    handle = wd21GetServerHandle( req );
    pfc = (HTTP_FILTER_CONTEXT*) handle->httpFilterContext;
    if (!pfc->GetServerVariable( pfc, "PATH_INFO", pathInfo, &BufferSize)) {
        sapdbwa_SQLFREE( pathInfo );
        return(0);
    } else {
        wd21SetPathInfo( req, pathInfo );
        handle->we_malloced_PathInfo = true;
        return( BufferSize );
    }; /* else */
}; /* wd52_BillsPathInfo */

/*********************************************************/

static size_t wd52_BillsContentLength( twd21HttpRequestP  req,
                                       size_t             BufferSize )
{
    HTTP_FILTER_CONTEXT   *pfc;
    char                  *contentLength;
    sapdbwa_Bool             allocatOK;
    twd52BillsServerHandle handle;
    
    sapdbwa_SQLALLOCAT( BufferSize, &contentLength, &allocatOK );
    if (!allocatOK) {
        /* memory allocation failure */
        return 0;
    }; /* if */

    handle = wd21GetServerHandle( req );
    pfc = (HTTP_FILTER_CONTEXT*) handle->httpFilterContext;
    if (!pfc->GetServerVariable( pfc,
                                 "CONTENT_LENGTH",
                                 contentLength,
                                 &BufferSize)) {
        sapdbwa_SQLFREE( contentLength );
        return(0);
    } else {
        wd21SetContentLength( req, contentLength );
        handle->we_malloced_ContentLength = true;
        return(BufferSize);
    }; /* else */
}; /* wd52_BillsContentLength */

/*********************************************************/

static size_t wd52_BillsPhysicalPath( twd21HttpRequestP  req,
                                      size_t             BufferSize )
{
    twd52BillsServerHandle handle;
    char                  *physicalPath = NULL;
    sapdbwa_Bool             allocatOK;
    
    handle = wd21GetServerHandle( req );
    physicalPath = handle->httpFilterUrlMap->pszPhysicalPath;

    if(physicalPath == NULL || strlen(physicalPath) == 0) {
        sapdbwa_SQLALLOCAT( sizeof(char), &physicalPath, &allocatOK );
        handle->we_malloced_PhysicalPath = true;           
        *physicalPath = '\0';
        wd21SetPhysicalPath( req, physicalPath );
        return(0);
    } else {
        wd21SetPhysicalPath( req, physicalPath );    
        handle->we_malloced_PhysicalPath = false;
        return strlen(physicalPath);
    }; /* else */
}; /* wd52_BillsPhysicalPath */


/*********************************************************/

static size_t wd52_BillsServerName( twd21HttpRequestP req, size_t BufferSize )
{
    HTTP_FILTER_CONTEXT   *pfc;
    char                  *serverName;
    sapdbwa_Bool             allocatOK;
    twd52BillsServerHandle handle;
    
    sapdbwa_SQLALLOCAT( BufferSize, &serverName, &allocatOK );
    if (!allocatOK) {
        /* memory allocation failure */
        wd20WriteLogMsg( ERR_MSG_MEMORY_WD26 );
        return 0;
    }; /* if */

    handle = wd21GetServerHandle( req );
    pfc = (HTTP_FILTER_CONTEXT*) handle->httpFilterContext;
    if (!pfc->GetServerVariable( pfc,
                                 "SERVER_NAME",
                                 serverName,
                                 &BufferSize) ) {
        sapdbwa_SQLFREE( serverName );
        return(0);
    } else {
        wd21SetServerName( req, serverName );
        handle->we_malloced_ServerName = true;
        return(BufferSize);
    }; /* else */
}; /* wd52_BillsServerName */


#define MAX_URI_SZ 1024
#define MAX_PATH_INFO_SZ 1024
#define MAX_REQUEST_METHOD_SZ 1024
#define MAX_CONTENT_TYPE_SZ 1024
#define MAX_CONTENT_LENGTH_SZ 1024
#define MAX_PHYSICALPATH_SZ 1024
#define MAX_SERVERNAME_SZ 1024

static DWORD WINAPI wd52_BillsRequestMap( HTTP_FILTER_CONTEXT  *pfc,
                                          HTTP_FILTER_URL_MAP  *pvData )
{
    DWORD      rtc = SF_STATUS_REQ_NEXT_NOTIFICATION;
    sapdbwa_Bool allocatOK;
    
    twd21HttpRequestP  req = NULL;
    sapdbwa_HttpReplyP      rep = NULL;
    twd52BillsServerHandle handle = NULL;

    handle = wd52_CreateServerHandle( pfc, pvData );
    if (handle == NULL) {
        /* memory allocation error */
        wd20WriteLogMsg( ERR_MSG_MEMORY_WD26 );
        return rtc;
    }; /* if */
        
    req = wd21CreateRequest( handle,
                             wd52BillsReadBody,
                             wd52BillsGetHeader,
                             wd52BillsIsSecure );
    if (req == NULL) {
        /* memory allocation error */
        wd20WriteLogMsg( ERR_MSG_MEMORY_WD26 );
        wd52_DestroyServerHandle( handle );
        return rtc;
    }; /* if */

    rep = wd22CreateReply( handle, wd52BillsSendHeader, wd52BillsSendBody );
    if (rep == NULL) {
        /* memory allocation error */
        wd20WriteLogMsg( ERR_MSG_MEMORY_WD26 );
        wd21DestroyRequest( req );        
        wd52_DestroyServerHandle( handle );
        return rtc;
    }; /* if */
    
    wd21SetUri( req, (char*) pvData->pszURL );

    //wd52_BillsIfModifiedSince( req );
    wd52_BillsQueryString( req );
    wd52_BillsPathInfo( req, MAX_PATH_INFO_SZ );
    wd52_BillsRequestMethod( req, MAX_REQUEST_METHOD_SZ );
    wd52_BillsContentType( req, MAX_CONTENT_TYPE_SZ );
    wd52_BillsContentLength( req, MAX_CONTENT_LENGTH_SZ );
    wd52_BillsPhysicalPath( req, MAX_PHYSICALPATH_SZ );
    wd52_BillsServerName( req, MAX_SERVERNAME_SZ );

    wd52_BillsReadPostInit( req );

    switch ( wd20CallServiceFunc( req, rep ))
        {
        case sapdbwa_RTC_Ignored:
            rtc = SF_STATUS_REQ_NEXT_NOTIFICATION;
            break;
        case sapdbwa_RTC_Accepted:
            rtc = SF_STATUS_REQ_FINISHED_KEEP_CONN;
            break;
        case sapdbwa_RTC_Error:
            rtc = SF_STATUS_REQ_ERROR;
            break;
        default:
            rtc = SF_STATUS_REQ_ERROR;
            break;
        }; /* switch */

    wd52_DestroyRequest( handle, req );
    wd22DestroyReply( rep );
    wd52_DestroyServerHandle( handle );
    
    return(rtc);
}; /* wd52_BillsRequestMap */


static DWORD WINAPI wd52_BillsReadRaw( HTTP_FILTER_CONTEXT   *pfc,
                                       HTTP_FILTER_RAW_DATA  *pvData )
    /*
     * NOTE !!!!!
     * dont forget to free the context.
     */
{
    struct			st_wa_raw_data  *pFilterContext = pfc->pFilterContext;
    char           *tmp = NULL;
    char           *n = NULL;
    sapdbwa_Bool    allocatOK;
    
    if (!pFilterContext) {
        sapdbwa_SQLALLOCAT( sizeof(*pFilterContext),
                       (char**) &pFilterContext,
                       &allocatOK );
        if (!allocatOK) {
            wd20WriteLogMsg( ERR_MSG_MEMORY_WD26 );
            return(SF_STATUS_REQ_ERROR);
        }; /* if */
        memset( pFilterContext, 0, sizeof(*pFilterContext) );
        pfc->pFilterContext = pFilterContext;
    }; /* if */

	/* Check buffer if it is already allocated	*/
	/* We have to use this check, because our RTE SQLREALLOCAT	*/
	/* works like the posix defintion only on Unix platforms.	*/
	/* On Windows it uses a heap pool to allocat memory and		*/
	/* the according memory function doesn't work like the		*/
	/* posix definition.										*/
	if (pFilterContext->Buf) {
		/* Reallocat memory	*/
		sapdbwa_SQLREALLOCAT( pvData->cbInData,
					(char**) &pFilterContext->Buf,
					&allocatOK );
		if (!allocatOK) {
			sapdbwa_SQLFREE( (char*) pFilterContext->Buf);
			sapdbwa_SQLFREE( (char*) pFilterContext);
			return(SF_STATUS_REQ_ERROR);
		}; /* if */
	} else {
		/* Allocat new memory	*/
		sapdbwa_SQLALLOCAT( pvData->cbInData,
					(char**) &pFilterContext->Buf,
					&allocatOK );
		if (!allocatOK) {
			sapdbwa_SQLFREE( (char*) pFilterContext);
			return(SF_STATUS_REQ_ERROR);
		}; /* if */
	}

    memcpy( (char*) pFilterContext->Buf, pvData->pvInData, pvData->cbInData );
    pFilterContext->BufLg = pvData->cbInData;
    pFilterContext->BufSz = pvData->cbInData;

	if (!pFilterContext->BufBody) {
		tmp = pvData->pvInData;
		if (tmp) {
			tmp = strstr(tmp, "\r\n\r\n");
			if (tmp) {
				pFilterContext->BufBody = strstr(tmp, "\r\n\r\n");
				pFilterContext->HeaderLg = tmp + sizeof("\r\n\r\n") - 1 - (char*)pvData->pvInData;
			}
		}
	}

	if (pFilterContext->HeaderLg > 0 && pFilterContext->CL == 0) {
		tmp = pvData->pvInData;
		if ((n = strstr(tmp, "Content-length")) || (n = strstr(tmp, "Content-Length")))	{
			n += strlen("Content-length");
			while (*n && *n == ' ') n++;
			if (*n == ':') {
				n++;
				pFilterContext->CL = strtoul(n, NULL, 10);
				if (pFilterContext->CL == ULONG_MAX)
					pFilterContext->CL = 0;
			}
		} else {
			pFilterContext->CL = 0;
		}; /* else */
	}
 
    if (pFilterContext->BufBody &&											/* Header has been read			*/
		pFilterContext->CL + pFilterContext->HeaderLg > pvData->cbInData)	/* BodyLg + HeaderLg > ReqLg	*/
        return(SF_STATUS_REQ_READ_NEXT);
    else
        return(SF_STATUS_REQ_NEXT_NOTIFICATION);
}; /* wd52_BillsReadRaw */


static DWORD WINAPI wd52_BillsClearReadRaw( HTTP_FILTER_CONTEXT  *pfc,
                                            HTTP_FILTER_RAW_DATA *pvData )
{
    if (pfc && pfc->pFilterContext)
        {
            struct st_wa_raw_data  *pFilterContext = pfc->pFilterContext;

            if (pFilterContext->Buf) {
                sapdbwa_SQLFREE( (char*) pFilterContext->Buf );
            }; /* if */
            sapdbwa_SQLFREE( (char*) pFilterContext );
    
            pfc->pFilterContext = NULL;
        }
    return(SF_STATUS_REQ_NEXT_NOTIFICATION);
}; /* wd52_BillsClearReadRaw */


/*!**********************************************************************

  EndChapter: Internal Functions

************************************************************************/
