/*!**********************************************************************

  module: vwd53ap.c

  -----------------------------------------------------------------------

  responsible:  Holger Bischoff
				Markus Oezgen

  special area: WebAgent

  description:  Functions for Apache Web Server

  see also:

  -----------------------------------------------------------------------

                         Copyright by SAP AG, 1998



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

************************************************************************/

/*=======================================================================
 *  INCLUDES
 *=====================================================================*/

#include <httpd.h>
#include <http_config.h>
#include <http_core.h>
#include <http_prot.h> /* original http_protocol.h */
#include <http_request.h>
#include <http_log.h>
#include <util_script.h>

#include "gwd00.h"
#include "hwd03waapi.h"
#include "hwd20control.h"
#include "hwd21req.h"
#include "hwd22rep.h"
#include "hwd28dict.h"
#include "hwd29date.h"

module WebAgentModule;


static sapdbwa_Bool initialized = false;

#ifdef WIN32
/* under NT we have to clean on process detach PTS 1106610 */
BOOL APIENTRY DllMain( HANDLE	hModule,
                       DWORD    dwReason,
                       LPVOID   lpReserved)
{
    switch (dwReason) {
    case DLL_PROCESS_ATTACH:
        break;
    case DLL_PROCESS_DETACH:
        if (initialized) {
            wd20ApiExit();
            initialized = false;
        };
        break;
    default:
        break;
    }; /* switch */
}; /* DllMain */
#endif

/*!**********************************************************************

  Chapter: module-Functions

  description: These functions are called by Apache
  
************************************************************************/

static void wd53ApacheWebAgentInit( server_rec *s, pool *p)
{
    if (!initialized) {
        if (wd20ApiInit(NULL, API_APACHE_WD20)) {
            initialized = true;
            wd20WriteLogMsg( "WebAgent Apache module loaded\n" );            
        }; /* else */
    };
} /* wd53ApacheWebAgentInit */


static int wd53ApacheWebAgentHandler( request_rec *r);


static void wd53WebAgentExit( server_rec *s, pool *p)
{
    wd20WriteLogMsg( "WebAgent Apache module unloaded\n" );
    if (initialized) {
        wd20ApiExit();
        initialized = false;
    };
} /* wd53WebAgentExit */

/*!**********************************************************************

  EndChapter: module-functions

************************************************************************/

/*!**********************************************************************

  Chapter: Call-Back Functions

  description: These functions are stored in the WAHttpRequest and
               WAHttpReply structures. They are used by the WebAgent
               to retrieve body-data from, and send the header and
               content to the web-server.
  
************************************************************************/

sapdbwa_UInt4 wd53ApacheReadBody( twd00ServerHandle   handle,
                                char               *b,
                                sapdbwa_UInt4         len );


const char* wd53ApacheGetHeader( twd00ServerHandle   handle,
                                 const char         *fieldName );

sapdbwa_Bool wd53ApacheSendHeader( twd00ServerHandle   handle,
                                 sapdbwa_HttpReplyP       rep );


sapdbwa_UInt4 wd53ApacheSendBody( twd00ServerHandle   handle,
                                const void         *Buf,
                                sapdbwa_UInt4         len );


sapdbwa_Bool wd53ApacheIsSecure( twd00ServerHandle   handle );

/*!**********************************************************************

  EndChapter: Call-Back Functions

************************************************************************/

/*!**********************************************************************

  Chapter: Internal Functions

************************************************************************/

typedef struct st_apache_server_handle
{
    request_rec *reqRec; 
} *twd53ApacheServerHandle;

twd53ApacheServerHandle wd53_CreateServerHandle( request_rec *r );

void wd53_DestroyServerHandle( twd53ApacheServerHandle handle );

int wd53_ApacheRequestMap( twd53ApacheServerHandle handle );

/*!**********************************************************************

  EndChapter: Internal Functions

************************************************************************/

/*!**********************************************************************

  Chapter: module-Functions

************************************************************************/

static int wd53ApacheWebAgentHandler( request_rec *r)
{
    twd53ApacheServerHandle handle;
    int rtc = OK;

    handle = wd53_CreateServerHandle( r );
    if (handle == NULL) {
        /* ... log error */
        return rtc;
    }; /* if */
    
    rtc = wd53_ApacheRequestMap( handle );
    wd53_DestroyServerHandle( handle );

    return rtc;
} /* wd53ApacheWebAgentHandler */

/*!**********************************************************************

  EndChapter: module-Functions

************************************************************************/


/*!**********************************************************************

  Chapter: Call-Back Functions

************************************************************************/

sapdbwa_UInt4 wd53ApacheReadBody( twd00ServerHandle   handle,
                                char               *b,
                                sapdbwa_UInt4         len )
{
    sapdbwa_UInt4 retcode;
    sapdbwa_Int4 read;
    twd53ApacheServerHandle apacheHandle = (twd53ApacheServerHandle) handle;
    
    if (ap_should_client_block( apacheHandle->reqRec )) {
        if ((read = ap_get_client_block( apacheHandle->reqRec,
                                         b, len)) < 0) {
            retcode = 0;
        } else {
            retcode = (sapdbwa_UInt4) read;
        }; /* else */
    } else {
        retcode = 0;
    }; /* else */

    return retcode;
} /* wd53ApacheReadBody */


const char* wd53ApacheGetHeader( twd00ServerHandle   handle,
                                 const char         *fieldName )
{
    twd53ApacheServerHandle apacheHandle = (twd53ApacheServerHandle) handle;
    const char *result;

    result = ap_table_get( apacheHandle->reqRec->subprocess_env, fieldName );
    /* Authorization headers are not handled by ap_add_common_vars to prevent
     * cgi from stealing passwords with ps -e. So they need special handling. */
    if (result == NULL
        && (strcmp( fieldName, "HTTP_AUTHORIZATION" ) == 0)) {
        result = ap_table_get( apacheHandle->reqRec->headers_in, "Authorization" );
    } else if (result == NULL
               && (strcmp( fieldName, "HTTP_PROXY_AUTHORIZATION" ) == 0)) {
        result = ap_table_get( apacheHandle->reqRec->headers_in, "Proxy-Authorization" );        
    };
        
    return result;
} /* wd53ApacheGetHeader */


sapdbwa_Bool wd53ApacheSendHeader( twd00ServerHandle   handle,
                                 sapdbwa_HttpReplyP       rep )
{
    twd53ApacheServerHandle apacheHandle = (twd53ApacheServerHandle) handle;
    const char *contentLength = wd22GetContentLength( rep );
    const char *lastModified  = wd22GetLastModified( rep );
    const char *expires       = wd22GetExpires( rep );
    const char *location      = wd22GetLocation( rep );
    
    sapdbwa_UInt4 iHeader;
    twd28DictionaryP headers = wd22GetHeaders( rep );
    const char *name;
    char       *value;
    
    if (wd22GetContentType( rep ) != NULL) {
        apacheHandle->reqRec->content_type = wd22GetContentType( rep );
    }; /* if */

    if (contentLength != NULL) {
        ap_table_set( apacheHandle->reqRec->headers_out,
                      "Content-Length", contentLength );
    };
    if (lastModified != NULL) {
        ap_table_merge( apacheHandle->reqRec->err_headers_out,
                        "Last-Modified", lastModified );
    };
    if (expires != NULL) {
        ap_table_merge( apacheHandle->reqRec->err_headers_out,
                        "Expires", expires );
    };
    if (location != NULL) {
        ap_table_merge( apacheHandle->reqRec->err_headers_out,
                        "Location", location );
    };
    
    for (iHeader = 0; iHeader < wd28NumEntries( headers ); iHeader++) {
        if (wd28GetEntryByIndex( headers, iHeader, &name, &value )) {
            ap_table_merge( apacheHandle->reqRec->err_headers_out,
                            name, value );
        }; /* if */
    }; /* for */
            
/* ap_table_set( r->headers_out, data, value);
       ap_table_add(r->err_headers_out, data, value);
       ap_table_merge(r->err_headers_out, data, value);
       */
    apacheHandle->reqRec->status = wd22GetStatus( rep );
    ap_send_http_header( apacheHandle->reqRec );

    return true;
} /* wd53ApacheSendHeader */


sapdbwa_UInt4 wd53ApacheSendBody( twd00ServerHandle   handle,
                                const void         *Buf,
                                sapdbwa_UInt4         len )
{
    twd53ApacheServerHandle apacheHandle = (twd53ApacheServerHandle) handle;
    sapdbwa_Int4   written;
    
    if ((written = ap_rwrite( (void*) Buf,
                              len,
                              apacheHandle->reqRec )) == EOF) {
        /* error */
        written = 0;
    }; /* if */
   
    return (sapdbwa_UInt4) (written > 0 ? written : 0 );
} /* wd53ApacheSendBody */


sapdbwa_Bool wd53ApacheIsSecure( twd00ServerHandle   handle )
{
    return (wd53ApacheGetHeader( handle, "HTTPS" ) != NULL );
} /* wd53ApacheIsSecure */

/*!**********************************************************************

  EndChapter: Call-Back Functions

************************************************************************/

twd53ApacheServerHandle wd53_CreateServerHandle( request_rec *r )
{
    sapdbwa_Bool              allocatOK;
    twd53ApacheServerHandle newHandle;

    sapdbwa_SQLALLOCAT( sizeof( struct st_apache_server_handle ),
                (sapdbwa_UInt1**) &newHandle,
                &allocatOK );
    if (allocatOK) {
        newHandle->reqRec = r;
    } else {
        newHandle = NULL;
    }; /* else */

    return newHandle;
} /* wd53_CreateServerHandle */


void wd53_DestroyServerHandle( twd53ApacheServerHandle handle )
{
    sapdbwa_SQLFREE( (sapdbwa_UInt1*) handle );
} /* wd53_DestroyServerHandle */


int wd53_ApacheRequestMap( twd53ApacheServerHandle handle )
{
    int rtc  = OK;
    table *e = handle->reqRec->subprocess_env;
    twd21HttpRequestP  req = NULL;
    sapdbwa_HttpReplyP      rep = NULL;
    sapdbwa_Int2         apacheRetcode;
    
    /* Set up variables */
    ap_add_common_vars( handle->reqRec );
    ap_add_cgi_vars( handle->reqRec );
    
    req = wd21CreateRequest( handle,
                             wd53ApacheReadBody,
                             wd53ApacheGetHeader,
                             wd53ApacheIsSecure );
    if (req == NULL) {
        /* memory allocation error */
        /* ... log error */
        return HTTP_INTERNAL_SERVER_ERROR;
    }; /* if */

    rep = wd22CreateReply( handle, wd53ApacheSendHeader, wd53ApacheSendBody );
    if (rep == NULL) {
        /* memory allocation error */
        /* ... log error */
        wd21DestroyRequest( req );        
        return HTTP_INTERNAL_SERVER_ERROR;
    }; /* if */
    
    wd21SetUri            ( req, handle->reqRec->uri );
    wd21SetIfModifiedSince( req, ap_table_get( e, "IF_MODIFIED_SINCE"));
    wd21SetQueryString    ( req, ap_table_get( e, "QUERY_STRING"));
    wd21SetPathInfo       ( req, ap_table_get( e, "PATH_INFO"));
    wd21SetRequestMethod  ( req, handle->reqRec->method );
    wd21SetContentType    ( req, ap_table_get( e, "CONTENT_TYPE"));
    wd21SetContentLength  ( req, ap_table_get( e, "CONTENT_LENGTH"));
    wd21SetPhysicalPath   ( req, ap_table_get(e, "PATH_TRANSLATED"));
    wd21SetServerName     ( req, ap_table_get( e, "REMOTE_HOST"));

    /* prepare reading of the body data */
    /* REQUEST_CHUNKED_ERROR : send 411 (LENGTH REQUIRED) if body
     * without content-length */
    if ( (apacheRetcode = ap_setup_client_block( handle->reqRec,
                                                 REQUEST_CHUNKED_ERROR))
         != OK) {
        wd21DestroyRequest( req );
        wd22DestroyReply( rep );
        return apacheRetcode;
    }; /* if */
    
    switch ( wd20CallServiceFunc( req, rep )) {
    case sapdbwa_RTC_Ignored:
        rtc = DECLINED;
        break;
    case sapdbwa_RTC_Accepted:
        rtc = OK;
        break;
    case sapdbwa_RTC_Error:
        rtc = HTTP_INTERNAL_SERVER_ERROR;
        /* error ... */
        break;
    default:
        rtc = HTTP_INTERNAL_SERVER_ERROR;
        /* error ... */
        break;
    }; /* switch */
    wd21DestroyRequest( req );
    wd22DestroyReply( rep );

    return rtc;
} /* wd53_ApacheRequestMap */


#if 0
/* 
 * List of directives specific to our module.
 */
static const command_rec wd53WebAgentCmds[] =
{
    {
        "Example",              /* directive name */
        cmd_example,            /* config action routine */
        NULL,                   /* argument to include in call */
        OR_OPTIONS,             /* where available */
        NO_ARGS,                /* arguments */
        "Example directive - no arguments"
                                /* directive description */
    },
    {NULL}
}; /* wd53WebAgentCmds */
#endif /* 0 */

static const handler_rec wd53WebAgentHandlers[] =
{
    {"webagent", wd53ApacheWebAgentHandler},
    {NULL}
}; /* wd53WebAgentHandlers */


module WebAgentModule =
{
    STANDARD_MODULE_STUFF,
    wd53ApacheWebAgentInit,     /* module initializer */
    NULL,                       /* per-directory config creator */
    NULL,                       /* dir config merger */
    NULL,                       /* server config creator */
    NULL,                       /* server config merger */
    NULL,                       /* command table */
    wd53WebAgentHandlers,       /* [7] list of handlers */
    NULL,                       /* [2] filename-to-URI translation */
    NULL,                       /* [5] check/validate user_id */
    NULL,                       /* [6] check user_id is valid *here* */
    NULL,                       /* [4] check access by host address */
    NULL,                       /* [7] MIME type checker/setter */
    NULL,                       /* [8] fixups */
    NULL,                       /* [10] logger */
#if MODULE_MAGIC_NUMBER >= 19970103
    NULL,                       /* [3] header parser */
#endif
#if MODULE_MAGIC_NUMBER >= 19970719
    NULL,                       /* process initializer */
#endif
#if MODULE_MAGIC_NUMBER >= 19970728
    wd53WebAgentExit,           /* process exit/cleanup */
#endif
#if MODULE_MAGIC_NUMBER >= 19970902
    NULL                        /* [1] post read_request handling */
#endif
}; /* WebAgentModule */

