/*!**********************************************************************

  module: vwd55ap2.c

  -----------------------------------------------------------------------

  responsible:  Holger Bischoff
				Markus Oezgen

  special area: WebAgent

  description:  Functions for Apache Web Server

  see also:

  -----------------------------------------------------------------------

                         Copyright by SAP AG, 2000



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

************************************************************************/

/*=======================================================================
 *  INCLUDES
 *=====================================================================*/

#include <httpd.h>
#include <http_config.h>
#include <http_core.h>
#include <http_prot.h> /* original http_protocol.h */
#include <http_request.h>
#include <http_log.h>
#include <util_script.h>

#include "gwd00.h"
#include "hwd03waapi.h"
#include "hwd20control.h"
#include "hwd21req.h"
#include "hwd22rep.h"
#include "hwd28dict.h"
#include "hwd29date.h"

module WebAgentModule;

/*!**********************************************************************

  Chapter: module-Functions

  description: These functions are called by Apache
  
************************************************************************/

static sapdbwa_Bool initialized = false;

#ifdef WIN32
/* under NT we have to clean on process detach PTS 1106610 */
BOOL APIENTRY DllMain( HANDLE	hModule,
                       DWORD    dwReason,
                       LPVOID   lpReserved)
{
    switch (dwReason) {
    case DLL_PROCESS_ATTACH:
        break;
    case DLL_PROCESS_DETACH:
        if (initialized) {
            wd20ApiExit();
            initialized = false;
        };
        break;
    default:
        break;
    }; /* switch */
}; /* DllMain */
#endif

static void wd55ApacheWebAgentInit( ap_pool_t    *p,
                                    server_rec   *s)
{
    if (!initialized) {
        if (wd20ApiInit(NULL,API_APACHE_2_WD20)) {
            initialized = true;
            wd20WriteLogMsg( "WebAgent Apache module loaded\n" );
        }; /* else */
    };
} /* wd55ApacheWebAgentInit */


static int wd55ApacheWebAgentHandler( request_rec *r);

#if 0
static void wd55WebAgentExit( server_rec *s, pool *p)
{
    wd20WriteLogMsg( "ApacheWebAgentExit\n" );
    if (initialized) {
        wd20ApiExit();
        initialized = false;
    };
} /* wd55WebAgentExit */
#endif

/*!**********************************************************************

  EndChapter: module-functions

************************************************************************/

/*!**********************************************************************

  Chapter: Call-Back Functions

  description: These functions are stored in the WAHttpRequest and
               WAHttpReply structures. They are used by the WebAgent
               to retrieve body-data from, and send the header and
               content to the web-server.
  
************************************************************************/

sapdbwa_UInt4 wd55ApacheReadBody( twd00ServerHandle   handle,
                                char               *b,
                                sapdbwa_UInt4         len );


const char* wd55ApacheGetHeader( twd00ServerHandle   handle,
                                 const char         *fieldName );

sapdbwa_Bool wd55ApacheSendHeader( twd00ServerHandle   handle,
                                 sapdbwa_HttpReplyP       rep );


sapdbwa_UInt4 wd55ApacheSendBody( twd00ServerHandle   handle,
                                const void         *Buf,
                                sapdbwa_UInt4         len );


sapdbwa_Bool wd55ApacheIsSecure( twd00ServerHandle   handle );

/*!**********************************************************************

  EndChapter: Call-Back Functions

************************************************************************/

/*!**********************************************************************

  Chapter: Internal Functions

************************************************************************/

typedef struct st_apache_server_handle
{
    request_rec *reqRec; 
} *twd55ApacheServerHandle;

twd55ApacheServerHandle wd55_CreateServerHandle( request_rec *r );

void wd55_DestroyServerHandle( twd55ApacheServerHandle handle );

int wd55_ApacheRequestMap( twd55ApacheServerHandle handle );

/*!**********************************************************************

  EndChapter: Internal Functions

************************************************************************/

/*!**********************************************************************

  Chapter: module-Functions

************************************************************************/

static int wd55ApacheWebAgentHandler( request_rec *r)
{
    twd55ApacheServerHandle handle;
    int rtc = OK;

    wd20WriteLogMsg( "1\n" );
    handle = wd55_CreateServerHandle( r );
    wd20WriteLogMsg( "2\n" );
    if (handle == NULL) {
        wd20WriteLogMsg( "3\n" );
        /* ... log error */
        return rtc;
    }; /* if */
    wd20WriteLogMsg( "4\n" );
    
    rtc = wd55_ApacheRequestMap( handle );
    wd20WriteLogMsg( "5\n" );
    wd55_DestroyServerHandle( handle );
    wd20WriteLogMsg( "6\n" );

    return rtc;
} /* wd55ApacheWebAgentHandler */

/*!**********************************************************************

  EndChapter: module-Functions

************************************************************************/


/*!**********************************************************************

  Chapter: Call-Back Functions

************************************************************************/

sapdbwa_UInt4 wd55ApacheReadBody( twd00ServerHandle   handle,
                                char               *b,
                                sapdbwa_UInt4         len )
{
    sapdbwa_UInt4 retcode;
    sapdbwa_Int4 read;
    twd55ApacheServerHandle apacheHandle = (twd55ApacheServerHandle) handle;
    
    if (ap_should_client_block( apacheHandle->reqRec )) {
        if ((read = ap_get_client_block( apacheHandle->reqRec,
                                         b, len)) < 0) {
            retcode = 0;
        } else {
            retcode = (sapdbwa_UInt4) read;
        }; /* else */
    } else {
        retcode = 0;
    }; /* else */

    return retcode;
} /* wd55ApacheReadBody */


const char* wd55ApacheGetHeader( twd00ServerHandle   handle,
                                 const char         *fieldName )
{
    twd55ApacheServerHandle apacheHandle = (twd55ApacheServerHandle) handle;
    const char *result;

    result = ap_table_get( apacheHandle->reqRec->subprocess_env, fieldName );
    /* Authorization headers are not handled by ap_add_common_vars to prevent
     * cgi from stealing passwords with ps -e. So they need special handling. */
    if (result == NULL
        && (strcmp( fieldName, "HTTP_AUTHORIZATION" ) == 0)) {
        result = ap_table_get( apacheHandle->reqRec->headers_in, "Authorization" );
    } else if (result == NULL
               && (strcmp( fieldName, "HTTP_PROXY_AUTHORIZATION" ) == 0)) {
        result = ap_table_get( apacheHandle->reqRec->headers_in, "Proxy-Authorization" );        
    };
        
    return result;
} /* wd55ApacheGetHeader */


sapdbwa_Bool wd55ApacheSendHeader( twd00ServerHandle   handle,
                                 sapdbwa_HttpReplyP       rep )
{
    twd55ApacheServerHandle apacheHandle = (twd55ApacheServerHandle) handle;
    const char *contentLength = wd22GetContentLength( rep );
    const char *lastModified  = wd22GetLastModified( rep );
    const char *expires       = wd22GetExpires( rep );
    const char *location      = wd22GetLocation( rep );
    
    sapdbwa_UInt4 iHeader;
    twd28DictionaryP headers = wd22GetHeaders( rep );
    const char *name;
    char       *value;
    
    if (wd22GetContentType( rep ) != NULL) {
        apacheHandle->reqRec->content_type = wd22GetContentType( rep );
    }; /* if */

    if (contentLength != NULL) {
        ap_table_set( apacheHandle->reqRec->headers_out,
                      "Content-Length", contentLength );
    };
    if (lastModified != NULL) {
        ap_table_merge( apacheHandle->reqRec->err_headers_out,
                        "Last-Modified", lastModified );
    };
    if (expires != NULL) {
        ap_table_merge( apacheHandle->reqRec->err_headers_out,
                        "Expires", expires );
    };
    if (location != NULL) {
        ap_table_merge( apacheHandle->reqRec->err_headers_out,
                        "Location", location );
    };
    
    for (iHeader = 0; iHeader < wd28NumEntries( headers ); iHeader++) {
        if (wd28GetEntryByIndex( headers, iHeader, &name, &value )) {
            ap_table_merge( apacheHandle->reqRec->err_headers_out,
                            name, value );
        }; /* if */
    }; /* for */
            
/* ap_table_set( r->headers_out, data, value);
       ap_table_add(r->err_headers_out, data, value);
       ap_table_merge(r->err_headers_out, data, value);
       */
    apacheHandle->reqRec->status = wd22GetStatus( rep );
    ap_send_http_header( apacheHandle->reqRec );

    return true;
} /* wd55ApacheSendHeader */


sapdbwa_UInt4 wd55ApacheSendBody( twd00ServerHandle   handle,
                                const void         *buf,
                                sapdbwa_UInt4         len )
{
    twd55ApacheServerHandle apacheHandle = (twd55ApacheServerHandle) handle;
    sapdbwa_Int4   written;
    
    if ((written = ap_rwrite( (void*) buf,
                              len,
                              apacheHandle->reqRec )) == EOF) {
        /* error */
        written = 0;
    }; /* if */
   
    return (sapdbwa_UInt4) (written > 0 ? written : 0 );
} /* wd55ApacheSendBody */


sapdbwa_Bool wd55ApacheIsSecure( twd00ServerHandle   handle )
{
    return (wd55ApacheGetHeader( handle, "HTTPS" ) != NULL );
} /* wd55ApacheIsSecure */

/*!**********************************************************************

  EndChapter: Call-Back Functions

************************************************************************/

twd55ApacheServerHandle wd55_CreateServerHandle( request_rec *r )
{
    sapdbwa_Bool              allocatOK;
    twd55ApacheServerHandle newHandle;

    sapdbwa_SQLALLOCAT( sizeof( struct st_apache_server_handle ),
                (sapdbwa_UInt1**) &newHandle,
                &allocatOK );
    if (allocatOK) {
        newHandle->reqRec = r;
    } else {
        newHandle = NULL;
    }; /* else */

    return newHandle;
} /* wd55_CreateServerHandle */


void wd55_DestroyServerHandle( twd55ApacheServerHandle handle )
{
    sapdbwa_SQLFREE( (sapdbwa_UInt1*) handle );
} /* wd55_DestroyServerHandle */


int wd55_ApacheRequestMap( twd55ApacheServerHandle handle )
{
    int rtc  = OK;
    ap_table_t        *e = handle->reqRec->subprocess_env;
    twd21HttpRequestP  req = NULL;
    sapdbwa_HttpReplyP      rep = NULL;
    sapdbwa_Int2         apacheRetcode;
    
    /* Set up variables */
    ap_add_common_vars( handle->reqRec );
    ap_add_cgi_vars( handle->reqRec );
    
    req = wd21CreateRequest( handle,
                             wd55ApacheReadBody,
                             wd55ApacheGetHeader,
                             wd55ApacheIsSecure );
    if (req == NULL) {
        /* memory allocation error */
        return HTTP_INTERNAL_SERVER_ERROR;
    }; /* if */

    rep = wd22CreateReply( handle, wd55ApacheSendHeader, wd55ApacheSendBody );
    if (rep == NULL) {
        /* memory allocation error */
        wd21DestroyRequest( req );        
        return HTTP_INTERNAL_SERVER_ERROR;
    }; /* if */
    
    wd21SetUri            ( req, handle->reqRec->uri );
    wd21SetIfModifiedSince( req, ap_table_get( e, "IF_MODIFIED_SINCE"));
    wd21SetQueryString    ( req, ap_table_get( e, "QUERY_STRING"));
    wd21SetPathInfo       ( req, ap_table_get( e, "PATH_INFO"));
    wd21SetRequestMethod  ( req, handle->reqRec->method );
    wd21SetContentType    ( req, ap_table_get( e, "CONTENT_TYPE"));
    wd21SetContentLength  ( req, ap_table_get( e, "CONTENT_LENGTH"));
    wd21SetPhysicalPath   ( req, ap_table_get( e, "PATH_TRANSLATED"));
    wd21SetServerName     ( req, ap_table_get( e, "REMOTE_HOST"));

    /* prepare reading of the body data */
    /* REQUEST_CHUNKED_ERROR : send 411 (LENGTH REQUIRED) if body
     * without content-length */
    if ( (apacheRetcode = ap_setup_client_block( handle->reqRec,
                                                 REQUEST_CHUNKED_ERROR))
         != OK) {
        wd21DestroyRequest( req );
        wd22DestroyReply( rep );
        return apacheRetcode;
    }; /* if */
    
    switch ( wd20CallServiceFunc( req, rep )) {
    case sapdbwa_RTC_Ignored:
        rtc = DECLINED;
        break;
    case sapdbwa_RTC_Accepted:
        rtc = OK;
        break;
    case sapdbwa_RTC_Error:
        rtc = HTTP_INTERNAL_SERVER_ERROR;
        /* error ... */
        break;
    default:
        rtc = HTTP_INTERNAL_SERVER_ERROR;
        /* error ... */
        break;
    }; /* switch */
    wd21DestroyRequest( req );
    wd22DestroyReply( rep );

    return rtc;
} /* wd55_ApacheRequestMap */


static void wd55RegisterHooks( void )
{
    ap_hook_child_init( wd55ApacheWebAgentInit, NULL, NULL,
                        AP_HOOK_MIDDLE );
} /* wd55RegisterHooks */


static const handler_rec wd55WebAgentHandlers[] =
{
    {"webagent", wd55ApacheWebAgentHandler},
    {NULL}
}; /* wd55WebAgentHandlers */


API_VAR_EXPORT module WebAgentModule =
{
    STANDARD20_MODULE_STUFF,
    NULL,           /* dir config creater */
    NULL,           /* dir merger --- default is to override */
    NULL,           /* server config */
    NULL,           /* merge server config */
    NULL,           /* command ap_table_t */
    wd55WebAgentHandlers,  /* handlers */
    wd55RegisterHooks      /* register hooks */
};
