/*!**********************************************************************

  module: vwd56fcgi.c

  -----------------------------------------------------------------------

  responsible:  Holger Bischoff
				Markus Oezgen

  special area: WebAgent

  description:  FastCGI -> WebAgent interface

  see also:

  -----------------------------------------------------------------------

  copyright:              Copyright by SAP AG, 2000



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

************************************************************************/

/*=======================================================================
 *  INCLUDES
 *=====================================================================*/

#include "wd/fcgi_config.h"

#include <stdlib.h> /* getenv */
#include <string.h>
#include <fcntl.h>

#include "hwd20control.h"
#include "hwd21req.h"
#include "hwd22rep.h"
#include "hwd28dict.h"
#include "hwd29date.h"

#include "wd/fcgi_stdio.h"

/*!**********************************************************************

  Chapter: Call-Back Functions

  description: These functions are stored in the WAHttpRequest and
               WAHttpReply structures. They are used by the WebAgent
               to retrieve body-data from, and send the header and
               content to the web-server.
  
************************************************************************/

sapdbwa_UInt4 wd56CgiReadBody( twd00ServerHandle   handle,
                             char               *buf,
                             sapdbwa_UInt4         len );

const char* wd56CgiGetHeader( twd00ServerHandle   handle,
                              const char         *fieldName );

tsp00_Bool wd56CgiSendHeader( twd00ServerHandle   handle,
                              sapdbwa_HttpReplyP       rep );


sapdbwa_UInt4 wd56CgiSendBody( twd00ServerHandle   handle,
                             const void         *buf,
                             sapdbwa_UInt4         len );


tsp00_Bool wd56CgiIsSecure( twd00ServerHandle   handle );

/*!**********************************************************************

  EndChapter: Call-Back Functions

************************************************************************/

/*!**********************************************************************

  Chapter: twd56CgiHandle

************************************************************************/

typedef struct st_cgi_handle *twd56CgiHandle;

twd56CgiHandle wd56_CreateCgiHandle();

void wd56_DestroyCgiHandle( twd56CgiHandle cgiHandle );

twd21HttpRequestP wd56_CreateRequest( twd56CgiHandle cgiHandle );

sapdbwa_HttpReplyP wd56_CreateReply( twd56CgiHandle cgiHandle );

/*!**********************************************************************

  EndChapter: twd56CgiHandle

************************************************************************/


/*!**********************************************************************

  Chapter: main

************************************************************************/


sapdbwa_Int4 main( int argc, char *argv[] )
{
    twd56CgiHandle     cgiHandle = NULL;
    twd21HttpRequestP  req       = NULL;
    sapdbwa_HttpReplyP      rep       = NULL;
    sapdbwa_Int4         rtc       = 0;

    int count = 0;

    if (!wd20ApiInit(NULL, API_FCGI_WD20 )) {
        /* API init failed */
        printf("api init failed\n");
        return 3;
    }; /* if */
    
    while(FCGI_Accept() >= 0) {
#if 0        
        char *contentLength = getenv("CONTENT_LENGTH");
        char *ldpath = getenv("LD_LIBRARY_PATH" );
        int len;

        printf("Content-type: text/html\r\n"
               "\r\n"
               "<title>FastCGI echo</title>"
               "<h1>FastCGI echo</h1>\n"
               "Request number %d,  Process ID: %d<p>\n", ++count, getpid());

        printf( " LD_PATH = %s<p>\n", ldpath != NULL ? ldpath : "" );
        if (contentLength != NULL) {
            len = strtol(contentLength, NULL, 10);
        }
        else {
            len = 0;
        }
        
        if (len <= 0) {
            printf("No data from standard input.<p>\n");
        }
        else {
            int i, ch;

            printf("Standard input:<br>\n<pre>\n");
            for (i = 0; i < len; i++) {
                if ((ch = getchar()) < 0) {
                    printf("Error: Not enough bytes received on standard input<p>\n");
                    break;
                }
                putchar(ch);
            }
            printf("\n</pre><p>\n");
        }

#endif        
        /* create cgiHandle */
        cgiHandle = wd56_CreateCgiHandle();
        if (cgiHandle == NULL) {
            wd20WriteLogMsg("Creation of cgiHandle failed.");
            return 4;
        }; /* if */
    
        /* create request-handle */
        req = wd56_CreateRequest( cgiHandle );
        if (cgiHandle == NULL) {
            wd20WriteLogMsg("Creation of requestHandle failed.");
            return 5;
        }; /* if */
    
        /* create reply-handle */
        rep = wd56_CreateReply( cgiHandle );
        if (cgiHandle == NULL) {
            wd20WriteLogMsg("Creation of replyHandle failed.");
            return 6;
        }; /* if */
        
        /* call user-dll service function */
        switch ( wd20CallServiceFunc( req, rep )) {
        case sapdbwa_RTC_Ignored:
            sapdbwa_InitHeader( rep, sapdbwa_Status_Ok, "text/html",
                          NULL, NULL, NULL, NULL );
            sapdbwa_SendHeader( rep );
            sapdbwa_SendBody( rep, "Unhandled fastcgi request!", 0 );
            rtc = 0;
            break;
        case sapdbwa_RTC_Accepted:
            rtc = 0; /* ok */
            break;
        case sapdbwa_RTC_Error:
            wd20WriteLogMsg("User service func returned sapdbwa_RTC_Error.");
            rtc = -1;
            break;
        default:
            rtc = -1;
            break;
        }; /* switch */

        wd21DestroyRequest( req );
        wd22DestroyReply( rep );
        wd56_DestroyCgiHandle( cgiHandle );
        
    }; /* while */
    wd20ApiExit();

    
    return rtc;
} /* main */

/*!**********************************************************************

  Chapter: twd56CgiHandle

************************************************************************/


struct st_cgi_handle {
    tsp00_Bool  sendHeaderCalled;
    char       *uri; /* = scriptName + pathInfo + queryString */
}; /* struct st_cgi_handle */


twd56CgiHandle wd56_CreateCgiHandle()
{
    tsp00_Bool allocatOK;
    twd56CgiHandle newHandle;

    sapdbwa_SQLALLOCAT( sizeof( struct st_cgi_handle ),
                (sapdbwa_UInt1**) &newHandle,
                &allocatOK );
    if (allocatOK) {
        newHandle->sendHeaderCalled = false;
    } else {
        newHandle = NULL;
    }; /* else */

    return newHandle;
} /* wd56_CreateCgiHandle */


void wd56_DestroyCgiHandle( twd56CgiHandle cgiHandle )
{
    if (cgiHandle->uri != NULL) {
        sapdbwa_SQLFREE( (sapdbwa_UInt1*) cgiHandle->uri );
    }; /* if */
    sapdbwa_SQLFREE( (sapdbwa_UInt1*) cgiHandle );
} /* wd56_DestroyCgiHandle */


twd21HttpRequestP wd56_CreateRequest( twd56CgiHandle cgiHandle )
{
    twd21HttpRequestP req;
    tsp00_Bool allocatOK;
    sapdbwa_UInt4 uriLen;
    
    char *scriptName;
    char *pathInfo;
    char *queryString;
    char *requestMethod;
    char *contentType;
    char *contentLength;
    char *pathTranslated;
    char *serverName;
    
    req = wd21CreateRequest( cgiHandle,
                             wd56CgiReadBody,
                             wd56CgiGetHeader,
                             wd56CgiIsSecure );
    if (req != NULL) {
        /* scriptName = getenv( "SCRIPT_NAME" ); */
        scriptName = "";
        pathInfo = getenv( "PATH_INFO" );
        queryString = getenv("QUERY_STRING");
        uriLen = (scriptName != NULL) ? strlen( scriptName ) + 1 : 1;
        if (pathInfo != NULL) {
            uriLen += strlen(pathInfo);
        }; /* if */
        if (queryString != NULL) {
            uriLen += strlen( queryString );
        }; /* if */
        sapdbwa_SQLALLOCAT( uriLen,
                    (sapdbwa_UInt1**) &cgiHandle->uri,
                    &allocatOK );
        if (allocatOK) {
            strcpy( cgiHandle->uri,  (scriptName != NULL) ? scriptName : "" );
            if (pathInfo != NULL) {
                strcat( cgiHandle->uri, pathInfo );
            }; /* if */
/* Why? (Bernd)
            if (queryString != NULL) {
                strcat( cgiHandle->uri, queryString );
            };*/ /* if */
            wd21SetUri( req, cgiHandle->uri );
        } else {
            cgiHandle->uri = NULL;
        }; /* else */
        
        wd21SetIfModifiedSince( req, "19000101000000" );
        wd21SetQueryString( req, queryString );
        wd21SetPathInfo( req, pathInfo );
        requestMethod = getenv("REQUEST_METHOD");
        wd21SetRequestMethod( req, requestMethod );
        contentType = getenv( "CONTENT_TYPE" );
        wd21SetContentType( req, contentType );
        contentLength = getenv("CONTENT_LENGTH");
        wd21SetContentLength( req, contentLength );
        pathTranslated = getenv("PATH_TRANSLATED");
        wd21SetPhysicalPath( req, pathTranslated );
        serverName = getenv( "SERVER_NAME" );
        wd21SetServerName( req, serverName );
    }; /* if */

    return req;
} /* wd56_CreateRequest */


sapdbwa_HttpReplyP wd56_CreateReply( twd56CgiHandle cgiHandle )
{
    sapdbwa_HttpReplyP rep;
    rep = wd22CreateReply( cgiHandle, wd56CgiSendHeader, wd56CgiSendBody );

    return rep;
} /* wd56_CreateReply */

/*!**********************************************************************

  EndChapter: twd56CgiHandle

************************************************************************/

/*!**********************************************************************

  Chapter: Call-Back Functions

************************************************************************/

void wd56_SendHeader( sapdbwa_HttpReplyP    rep );

sapdbwa_UInt4 wd56CgiReadBody( twd00ServerHandle   handle,
                             char               *buf,
                             sapdbwa_UInt4         len )
{
    sapdbwa_UInt4    bytesRead;

#ifdef WIN32    
    /* use binary mode to preserve end of line characters */
    setmode( fileno( stdin ), _O_BINARY );
#endif    
    bytesRead = fread( buf, 1, len, stdin );
    
    return bytesRead;
} /* wd56CgiReadBody */


const char* wd56CgiGetHeader( twd00ServerHandle   handle,
                              const char         *fieldName )
{
    return getenv( fieldName );
} /* wd56CgiGetHeader */


tsp00_Bool wd56CgiSendHeader( twd00ServerHandle  handle,
                              sapdbwa_HttpReplyP      rep )
{
    twd56CgiHandle cgiHandle = (twd56CgiHandle) handle;
    tsp00_Bool retcode = true;
    
    if (!cgiHandle->sendHeaderCalled) {
        wd56_SendHeader( rep );
        cgiHandle->sendHeaderCalled = true;
    } else {
        retcode = false;
    }; /* else */

    return retcode;
} /* wd56CgiSendHeader */


sapdbwa_UInt4 wd56CgiSendBody( twd00ServerHandle   handle,
                             const void         *buf,
                             sapdbwa_UInt4         len )
{
    sapdbwa_UInt4    bytesWritten;

    bytesWritten = fwrite( (void*) buf, 1, len, stdout );
    
    return bytesWritten;
} /* wd56CgiSendBody */


tsp00_Bool wd56CgiIsSecure( twd00ServerHandle   handle )
{
    return false;
} /* wd56CgiIsSecure */


void wd56_SendHeader( sapdbwa_HttpReplyP    rep )
{
    char *location      = wd22GetLocation( rep );
    char *lastModified  = wd22GetLastModified( rep );
    char *contentLength = wd22GetContentLength( rep );
    char *contentType   = wd22GetContentType( rep );
    char *expires       = wd22GetExpires( rep );

    char  date[MAX_DATE_LEN_WD29] = "";
    
    sapdbwa_UInt4       iHeader;
    twd28DictionaryP  headers = wd22GetHeaders( rep );
    const char       *name;
    char             *value;
    
    if (location != NULL) {
        printf( "location: %s \n", location );
    }; /* if */
    if (lastModified != NULL) {
        if (wd29WADateToHttpDate( lastModified, date ) != 0) {
            printf( "last-modified: %s \n", date );
        }; /* if */
    }; /* if */
    if (contentLength != NULL) {
        printf( "content-length: %s \n", contentLength );
    }; /* if */
    if ( contentType != NULL) {
        printf( "content-type: %s \n", contentType );
    }; /* if */
    if ( expires != NULL) {
        if (wd29WADateToHttpDate( expires, date ) != 0) {
            printf( "expires: %s \n", date );
        }; /* if */
    }; /* if */

    for (iHeader = 0; iHeader < wd28NumEntries( headers ); iHeader++) {
        if (wd28GetEntryByIndex( headers, iHeader, &name, &value )) {
            printf( "%s: %s\n", name, value );
        }; /* if */
    }; /* for */
    
    printf("\n"); /* separating header from body */
} /* wd54_SendHeader */

/*!**********************************************************************

  EndChapter: Call-Back Functions

************************************************************************/
