/*!**********************************************************************

  module: vwd59sim.c

  -----------------------------------------------------------------------

  responsible:  Holger Bischoff
				Markus Oezgen

  special area: WebAgent

  description:  simple simulation of a web-server

  see also:

  -----------------------------------------------------------------------

                         Copyright by SAP AG, 1998



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

************************************************************************/

/*=======================================================================
 *  INCLUDES
 *=====================================================================*/

#include <stdio.h>

#include "gwd00.h"
#include "hwd03waapi.h"
#include "hwd20control.h"
#include "hwd21req.h"
#include "hwd22rep.h"
#include "hwd28dict.h"
#include "hwd29date.h"

/*!**********************************************************************

  Chapter: module-Functions

  description: These functions are called by the main function of the
               server simulator.
  
************************************************************************/

static void wd59SimInit()
{
    if (!wd20ApiInit(NULL, API_WASIM_WD20)) {
        /* init failed */
        printf(" init failed\n");
    } else {
        printf(" init success\n");
        wd20WriteLogMsg( "wd59SimInit\n" );
    }; /* else */
} /* wd59SimInit */

sapdbwa_Int2 wd59SimHandler( twd28DictionaryP reqDict, const char *fname );

static void wd59SimExit()
{
    wd20WriteLogMsg( "wd59SimExit\n" );
    wd20ApiExit();
} /* wd59SimExit */

/*!**********************************************************************

  EndChapter: Filter-Functions

************************************************************************/

/*!**********************************************************************

  Chapter: Call-Back Functions

  description: These functions are stored in the WAHttpRequest and
               WAHttpReply structures. They are used by the WebAgent
               to retrieve body-data from, and send the header and
               content to the simulated web-server.
  
************************************************************************/

sapdbwa_UInt4 wd59SimReadBody( twd00ServerHandle   handle,
                             char               *b,
                             sapdbwa_UInt4         len );


const char* wd59SimGetHeader( twd00ServerHandle   handle,
                              const char         *fieldName );

sapdbwa_Bool wd59SimSendHeader( twd00ServerHandle   handle,
                              sapdbwa_HttpReplyP       rep );


sapdbwa_UInt4 wd59SimSendBody( twd00ServerHandle   handle,
                             const void         *Buf,
                             sapdbwa_UInt4         len );


sapdbwa_Bool wd59SimIsSecure( twd00ServerHandle   handle );

/*!**********************************************************************

  EndChapter: Call-Back Functions

************************************************************************/

/*!**********************************************************************

  Chapter: Internal Functions

************************************************************************/

typedef struct st_sim_server_handle
{
    sapdbwa_UInt4       reqNo;
    twd28DictionaryP  reqDict;
    const char       *inputFileName;
    FILE             *inputFile;
} *twd59SimServerHandle;

twd59SimServerHandle wd59_CreateServerHandle( sapdbwa_UInt4       reqNo,
                                              twd28DictionaryP  reqDict,
                                              const char       *fname );

void wd59_DestroyServerHandle( twd59SimServerHandle handle );

int wd59_SimRequestMap( twd59SimServerHandle handle );

void wd59_ShowRetcode( sapdbwa_Int4 retcode );

/*!**********************************************************************

  EndChapter: Internal Functions

************************************************************************/

/*!**********************************************************************

  Chapter: main

************************************************************************/

void wd59_PrintUsage()
{
    printf("Usage: wasim <GET|POST|PUT|...> <uri> <input-file>");
} /* wd59_PrintUsage */

int main( int argc, char **argv )
{
    twd28DictionaryP  reqDict;
    sapdbwa_Int4        retcode;
    const char       *fname = NULL;
    
    if (argc >=3) {
        if (stricmp( argv[1], "GET" ) == 0
            || stricmp( argv[1], "PUT" ) == 0
            || stricmp( argv[1], "POST" ) == 0
            || stricmp( argv[1], "DELETE" ) == 0
            || stricmp( argv[1], "PROPFIND" ) == 0
            || stricmp( argv[1], "OPTIONS" ) == 0
            || stricmp( argv[1], "PROPPATCH" ) == 0
            || stricmp( argv[1], "MKCOL" ) == 0
            || stricmp( argv[1], "COPY" ) == 0
            || stricmp( argv[1], "MOVE" ) == 0
            || stricmp( argv[1], "LOCK" ) == 0
            || stricmp( argv[1], "UNLOCK" ) == 0
            ) {
            /* create request data */
            reqDict = wd28CreateDictionary();
        
            wd59SimInit();
            wd28AddEntry( reqDict, "METHOD", argv[1] );
            wd28AddEntry( reqDict, "URI", argv[2] );
            wd28AddEntry( reqDict, "CONTENT_LENGTH", "123" );

            if (argc >= 4) {
                /* input file for POST, PUT, ... specified */
                fname = argv[3];
            };
            /* call handler */
            retcode = wd59SimHandler( reqDict, fname );
            printf( "wd59SimHandler retcode : ");
            wd59_ShowRetcode( retcode );
            wd59SimExit();
        } else {
            wd59_PrintUsage();
        }; /* else */
    } else {
        wd59_PrintUsage();
    }; /* else */    
    
    return 0;
} /* main */

/*!**********************************************************************

  EndChapter: main

************************************************************************/

/*!**********************************************************************

  Chapter: module-Functions

************************************************************************/

sapdbwa_Int2 wd59SimHandler( twd28DictionaryP reqDict, const char *fname )
{
    twd59SimServerHandle handle;
    sapdbwa_Int2 rtc = sapdbwa_RTC_Accepted;
    
    handle = wd59_CreateServerHandle( 0, reqDict, fname );
    if (handle == NULL) {
        printf( "CreateServerHandle failed\n" );
        
        return sapdbwa_RTC_Error;
    }; /* if */
    
    rtc = wd59_SimRequestMap( handle );
    wd59_DestroyServerHandle( handle );

    return rtc;
} /* wd59SimWebAgentHandler */

/*!**********************************************************************

  EndChapter: module-Functions

************************************************************************/


/*!**********************************************************************

  Chapter: Call-Back Functions

************************************************************************/

#ifdef WIN32
#define FILENO _fileno
#else
#define FILENO fileno
#endif

sapdbwa_UInt4 wd59SimReadBody( twd00ServerHandle   handle,
                             char               *b,
                             sapdbwa_UInt4         len )
{
    sapdbwa_UInt4 retcode = 0;
    sapdbwa_Int4 read;
    twd59SimServerHandle serverHandle = (twd59SimServerHandle) handle;

    if (!serverHandle->inputFile && serverHandle->inputFileName) {
        serverHandle->inputFile = fopen( serverHandle->inputFileName, "r" );
    };
    if (serverHandle->inputFile) {
        retcode = fread( b, 1, len, serverHandle->inputFile );
    } else {
        retcode = 0;
    };
    
    return retcode;
} /* wd59SimReadBody */


const char* wd59SimGetHeader( twd00ServerHandle   handle,
                              const char         *fieldName )
{
    twd59SimServerHandle simHandle = (twd59SimServerHandle) handle;
    const char *result = NULL;

    result = wd28GetValueByKey( simHandle->reqDict, fieldName );
    
    return result;
} /* wd59SimGetHeader */


sapdbwa_Bool wd59SimSendHeader( twd00ServerHandle   handle,
                              sapdbwa_HttpReplyP       rep )
{
    char *location      = wd22GetLocation( rep );
    char *lastModified  = wd22GetLastModified( rep );
    char *contentLength = wd22GetContentLength( rep );
    char *contentType   = wd22GetContentType( rep );
    char *expires       = wd22GetExpires( rep );

    char  date[MAX_DATE_LEN_WD29] = "";
    
    sapdbwa_UInt4       iHeader;
    twd28DictionaryP  headers = wd22GetHeaders( rep );
    const char       *name;
    char             *value;
    
    if (location != NULL) {
        printf( "location: %s \n", location );
    }; /* if */
    if (lastModified != NULL) {
        if (wd29WADateToHttpDate( lastModified, date ) != 0) {
            printf( "last-modified: %s \n", date );
        }; /* if */
    }; /* if */
    if (contentLength != NULL) {
        printf( "content-length: %s \n", contentLength );
    }; /* if */
    if ( contentType != NULL) {
        printf( "content-type: %s \n", contentType );
    }; /* if */
    if ( expires != NULL) {
        if (wd29WADateToHttpDate( expires, date ) != 0) {
            printf( "expires: %s \n", date );
        }; /* if */
    }; /* if */

    for (iHeader = 0; iHeader < wd28NumEntries( headers ); iHeader++) {
        if (wd28GetEntryByIndex( headers, iHeader, &name, &value )) {
            printf( "%s: %s\n", name, value );
        }; /* if */
    }; /* for */
    
    printf("\n"); /* separating header from body */

    return true;
} /* wd59SimSendHeader */


sapdbwa_UInt4 wd59SimSendBody( twd00ServerHandle   handle,
                             const void         *Buf,
                             sapdbwa_UInt4         len )
{
    sapdbwa_UInt4    bytesWritten;

    bytesWritten = fwrite( (void*) Buf, 1, len, stdout );
    
    return bytesWritten;
} /* wd59SimSendBody */


sapdbwa_Bool wd59SimIsSecure( twd00ServerHandle   handle )
{
    return false;
} /* wd59SimIsSecure */

/*!**********************************************************************

  EndChapter: Call-Back Functions

************************************************************************/

twd59SimServerHandle wd59_CreateServerHandle( sapdbwa_UInt4       reqNo,
                                              twd28DictionaryP  reqDict,
                                              const char       *fname )
{
    sapdbwa_Bool           allocatOK;
    twd59SimServerHandle newHandle;

    sapdbwa_SQLALLOCAT( sizeof( struct st_sim_server_handle ),
                (sapdbwa_UInt1**) &newHandle,
                &allocatOK );
    if (allocatOK) {
        newHandle->reqNo          = reqNo;
        newHandle->reqDict        = reqDict;
        newHandle->inputFileName  = fname;
        newHandle->inputFile      = NULL;
    } else {
        newHandle = NULL;
    }; /* else */

    return newHandle;
} /* wd59_CreateServerHandle */


void wd59_DestroyServerHandle( twd59SimServerHandle serverHandle )
{
    if (serverHandle->inputFile) {
        fclose( serverHandle->inputFile );
    };
    sapdbwa_SQLFREE( (sapdbwa_UInt1*) serverHandle );
} /* wd53_DestroyServerHandle */


int wd59_SimRequestMap( twd59SimServerHandle handle )
{
    int rtc  = sapdbwa_RTC_Accepted;
    sapdbwa_HttpReplyP      rep = NULL;
    twd21HttpRequestP  req;
    
    req = wd21CreateRequest( handle,
                             wd59SimReadBody,
                             wd59SimGetHeader,
                             wd59SimIsSecure );
    if (req == NULL) {
        /* memory allocation error */
        return sapdbwa_RTC_Error;
    }; /* if */

    rep = wd22CreateReply( handle, wd59SimSendHeader, wd59SimSendBody );
    if (rep == NULL) {
        /* memory allocation error */
        wd21DestroyRequest( req );        
        return sapdbwa_RTC_Error;
    }; /* if */
    
    wd21SetUri            ( req, wd28GetValueByKey( handle->reqDict,
                                                    "URI" ));
    wd21SetIfModifiedSince( req, wd28GetValueByKey( handle->reqDict,
                                                    "IF_MODIFIED_SINCE"));
    wd21SetQueryString    ( req, wd28GetValueByKey( handle->reqDict,
                                                    "QUERY_STRING"));
    wd21SetPathInfo       ( req, wd28GetValueByKey( handle->reqDict,
                                                    "PATH_INFO"));
    wd21SetRequestMethod  ( req, wd28GetValueByKey( handle->reqDict,
                                                    "METHOD"));
    wd21SetContentType    ( req, wd28GetValueByKey( handle->reqDict,
                                                    "CONTENT_TYPE"));
    wd21SetContentLength  ( req, wd28GetValueByKey( handle->reqDict,
                                                    "CONTENT_LENGTH"));
    wd21SetPhysicalPath   ( req, wd28GetValueByKey( handle->reqDict,
                                                    "PATH_TRANSLATED"));
    wd21SetServerName     ( req, wd28GetValueByKey( handle->reqDict,
                                                    "SERVER_NAME"));
    
    rtc = wd20CallServiceFunc( req, rep );

    return rtc;
} /* wd59_SimRequestMap */


void wd59_ShowRetcode( sapdbwa_Int4 retcode )
{
    switch (retcode) {
    case sapdbwa_RTC_Error:
        printf("sapdbwa_RTC_Error\n");
        break;
    case sapdbwa_RTC_Ignored:
        printf("sapdbwa_RTC_Ignored\n");
        break;
    case sapdbwa_RTC_Accepted:
        printf("sapdbwa_RTC_Accepted\n");
        break;
    default:
        printf("Unknown retcode!!!\n");
        break;
    }; /* switch */
} /* wd59_ShowRetcode */
