/*!**********************************************************************

  module: vwd82fs.c

  -----------------------------------------------------------------------

  responsible:  Holger Bischoff
				Markus Oezgen

  special area: WebAgent

  description:  UserDll for access to the database file system

  see also:

  -----------------------------------------------------------------------

                         Copyright by SAP AG, 1998



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

************************************************************************/

/*=======================================================================
 *  INCLUDES
 *=====================================================================*/

#include <WINDOWS.H>

#include "heo102.h"

#include "hwd03waapi.h"
#include "hwd05wases.h"
#include "hwd20control.h"
#include "hwd101dbfs.h"
#include "hwd91mime.h"

static twd91MimeTypesP wd82mimeTypes = NULL;

const char *wd82_GetMimeTypeFile( sapdbwa_Handle wa );

static twd101DBFSP  wd82dbfs = NULL;

sapdbwa_Int2 wd82FSInit( sapdbwa_Handle wa )
{
    sapdbwa_SessionPoolP  pool;
    sapdbwa_Int4       rtc = sapdbwa_RTC_Ok;
    
    if ((wd82mimeTypes =
         wd91CreateMimeTypes( wd82_GetMimeTypeFile(wa) ))
        != NULL) {
        sapdbwa_WriteLogMsg( wa, "create mime-types success\n" );
        pool = sapdbwa_GetSessionPool( wa );
        if (pool != NULL) {
            if ((wd82dbfs = wd101CreateDBFS( pool, DBFS_PARTITION_ID_WD101 )) != NULL) {
                sapdbwa_WriteLogMsg( wa, "create dbfs success\n" );
            } else {
                sapdbwa_WriteLogMsg( wa, "create dbfs failed\n" );
                wd91DestroyMimeTypes( wd82mimeTypes );
                rtc = sapdbwa_RTC_Error;
            }; /* else */
        } else {
            sapdbwa_WriteLogMsg( wa, "get sessionpool failed\n" );
            wd91DestroyMimeTypes( wd82mimeTypes );
            rtc = sapdbwa_RTC_Error;
        }; /* else */
    } else {
        sapdbwa_WriteLogMsg( wa, "create mime-types failed\n" );
        rtc = sapdbwa_RTC_Error;
    }; /* else */
    
    return (rtc);
} /* wd82FSInit */


sapdbwa_Int2 wd82FSExit( sapdbwa_Handle wa )
{
    if (wd82mimeTypes != NULL) {
        wd91DestroyMimeTypes( wd82mimeTypes );
        wd82mimeTypes = NULL;
    };
    if (wd82dbfs != NULL) {
        wd101DestroyDBFS( wd82dbfs );
        wd82dbfs = NULL;
    };
    
    return 1;
} /* wd82FSExit */


#define BUFF_SIZE 4096

sapdbwa_Bool wd82MakePathFromURI( const char **uri );

const char *wd82_GetExtension( const char *path );

sapdbwa_Int2 wd82GetFile( sapdbwa_Handle         wa,
                        sapdbwa_HttpRequestP   request,
                        sapdbwa_HttpReplyP     reply,
                        twd101DBFSP       dbfs );

sapdbwa_Int2 wd82PutFile( sapdbwa_Handle         wa,
                        sapdbwa_HttpRequestP   request,
                        sapdbwa_HttpReplyP     reply,
                        twd101DBFSP      dbfs );


#define INIT_FAILED_MSG "Initialization failed!"

sapdbwa_Int2 wd82FSService( sapdbwa_Handle         wa,
                          sapdbwa_HttpRequestP   request,
                          sapdbwa_HttpReplyP     reply )
{
    /* TODO: handle DELETE request */
    if (strcmp( sapdbwa_GetMethod( request ), "GET" ) == 0) {
        return wd82GetFile( wa, request, reply, wd82dbfs );
    } else if (strcmp( sapdbwa_GetMethod( request ), "PUT" ) == 0) {
        /* TODO: Send status CREATED if new file was created,
         * or OK if existing file was updated */
        return wd82PutFile( wa, request, reply, wd82dbfs );
    } else {
        return sapdbwa_RTC_Ignored;
    }; /* else */
} /* vwd82FSService */

#define DEFAULT_MIME_TYPE "text/html"

const char *wd82_GetExtension( const char *path )
{
    sapdbwa_UInt2 pos;
    
    for (pos = strlen( path ); (pos > 0 && path[pos-1] != '.'); pos--) {};
    if ( path[pos-1] == '.') {
        return &path[pos];
    } else {
        return NULL;
    }; /* else */
} /* wd82_GetExtension */


sapdbwa_Int2 wd82GetFile( sapdbwa_Handle         wa,
                        sapdbwa_HttpRequestP   request,
                        sapdbwa_HttpReplyP     reply,
                        twd101DBFSP      dbfs )
{
    const char     *errMsg = NULL;
    const char     *path = NULL;
    char            buffer[BUFF_SIZE];
    sapdbwa_Int4      docLen;
    twd101DBFile    file;
    sapdbwa_Bool      ok = true;
    const char     *mimeType;
    const char     *extension;
    
    path = sapdbwa_GetRequestURI( request );
    if ((ok = wd82MakePathFromURI( &path ))) {
        if ( (file = wd101Open( dbfs, path, access_read )) == -1) {
            errMsg = wd101GetErrorMsg( dbfs );
            ok = false;
            mimeType = DEFAULT_MIME_TYPE;            
        }; /* if */
    } else {
        sapdbwa_InitHeader( reply, sapdbwa_Status_Ok, "text/html",
                      NULL, NULL, NULL, NULL );
        sapdbwa_SendHeader( reply );
        sapdbwa_SendBody( reply, "make path failed!", 0 );
        mimeType = DEFAULT_MIME_TYPE;
        ok = false;
    }; /* else */

    if (ok) {
        extension = wd82_GetExtension( path );
        if (extension != NULL) {
            mimeType = wd91GetMimeType( wd82mimeTypes, extension );
            if (mimeType == NULL) {
                mimeType = DEFAULT_MIME_TYPE;
            }; /* if */
        } else {
            mimeType = DEFAULT_MIME_TYPE;
        }; /* else */
        sapdbwa_InitHeader( reply, sapdbwa_Status_Ok, mimeType,
                      NULL, NULL, NULL, NULL );
        sapdbwa_SendHeader( reply );
        
        do {
            docLen = wd101Read( dbfs, file, buffer, sizeof( buffer ));
            if (docLen > 0) {
                sapdbwa_SendBody( reply, buffer, docLen );
            }; /* if */
        } while (docLen > 0 /* = sizeof(buffer) */);
        wd101Close( dbfs, file );
    } else {
        sapdbwa_InitHeader( reply, sapdbwa_Status_Ok, mimeType,
                      NULL, NULL, NULL, NULL );
        sapdbwa_SendHeader( reply );
        if (errMsg != NULL) {
            sapdbwa_SendBody( reply, errMsg, 0 );
        } else {
            sapdbwa_SendBody( reply, "open failed!", 0 );
        }; /* else */            
    }; /* if */
    
    return sapdbwa_RTC_Accepted;
} /* wd82GetFile */


const char *wd82_GetMimeTypeFile( sapdbwa_Handle wa )
{
    static char mimeTypeFileName[1024];

    if (!sapdbwa_GetParam( wa,
                      "MimeTypeFile",
                      mimeTypeFileName,
                      sizeof(mimeTypeFileName) )) {
        /* use default */
#ifdef WIN32    
        strcpy( mimeTypeFileName, "d:\\V72\\usr\\WebAgent\\conf\\mime.types" );
#else
        strcpy( mimeTypeFileName, "mime.types" );
#endif
    };
    
    return mimeTypeFileName;
} /* wd82_GetMimeTypeFile */


sapdbwa_Int2 wd82PutFile( sapdbwa_Handle         wa,
                        sapdbwa_HttpRequestP   request,
                        sapdbwa_HttpReplyP     reply,
                        twd101DBFSP      dbfs )
{
    const char     *errMsg = NULL;
    const char     *path = NULL;
    char           *buffer;
    twd101DBFile    file;
    sapdbwa_UInt4     docLen;
    sapdbwa_Bool      ok = true;
    sapdbwa_Bool      allocatOK;
    sapdbwa_UInt4     contentLength = 0;
    sapdbwa_UInt4     read;    /* bytes read from webserver */
    sapdbwa_UInt4     written; /* bytes written to file     */
    
    sapdbwa_InitHeader( reply, sapdbwa_Status_Ok, "text/html", NULL, NULL, NULL, NULL );
    sapdbwa_SendHeader( reply );

    if (sapdbwa_GetContentLength( request ) != NULL) {
        contentLength = atoi( sapdbwa_GetContentLength( request ) );
    }; /* if */
    if (contentLength > 0) {
        sapdbwa_SQLALLOCAT( contentLength+1, (sapdbwa_UInt1**) &buffer, &allocatOK );
        if (allocatOK) {
            path = sapdbwa_GetRequestURI( request );
            if ((ok = wd82MakePathFromURI( &path ))) {
                read = sapdbwa_ReadBody( request, buffer, contentLength );
                if (read == contentLength) {
                    char text[100];
                    file = wd101Create( dbfs, path, access_write_compressed );
                    if (file != -1) {
                        written = wd101Write( dbfs, file, buffer, read );
                        ok = wd101Close( dbfs, file );
                        sapdbwa_SendBody( reply, "File written!", 0 );
                    } else {
                        sapdbwa_SendBody( reply, "Create file failed! ", 0 );
                        sapdbwa_SendBody( reply, wd101GetErrorMsg( dbfs ), 0 );
                    }; /* else */
                } else {
                    sapdbwa_SendBody( reply, "Could not read body!", 0 );
                }; /* else */                    
            } else {
                sapdbwa_SendBody( reply, "invalid path!", 0 );
            }; /* else */
            sapdbwa_SQLFREE( (sapdbwa_UInt1*) buffer );
        } else {
            sapdbwa_SendBody( reply, "sapdbwa_SQLALLOCAT failed!", 0 );
        }; /* else */
    } else {
        sapdbwa_SendBody( reply, "contentLength == 0!", 0 );
    }; /* else */
    
    return sapdbwa_RTC_Accepted;
} /* wd82PutFile */


sapdbwa_Bool wd82MakePathFromURI( const char **uri )
{
    const char *path = *uri;
    sapdbwa_Bool  ok = false;
    
    if (path != NULL) {
        path = strchr( path, '/' );
        if (path != NULL) {
            path++; /* step over first / */
            path = strchr( path, '/' );
            if (path != NULL) {
                path++; /* step over second / */
                ok = true;
            }; /* if */
        }; /* if */
    }; /* if */
    if (ok) {
        *uri = path;
    }; /* if */
    
    return ok;
} /* wd82MakePathFromURI */
